/*---------------------------------------------------------------------------*
  Project:  CtrBrom - library - init
  File:     crt0.c

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/
#include <brom/code32.h>
#include <brom/os.h>

#define BROM_ENABLE_FCRAM
#define BROM_ENABLE_VRAM
#define BROM_ENABLE_DSP_WRAM

int main( void );

/*---------------------------------------------------------------------------*
  Name:         _start

  Description:  Start up

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void _start( void )
{
        PRESERVE8

        CODE32  // for _start reference

#ifdef BROM_ENABLE_FCRAM

        // FCRAMNbNJn
        ldr     r3, =REG_FCRAM_ADDR
        ldr     r0, =REG_PDN_FCRAM_CLK_E_MASK
        str     r0, [r3]

        // FCRAM 512MB & Wait
        // hCoOixt@[Xg & hCoxʏ
        ldr     r3, =REG_FCRAM_CNT0_ADDR
        ldr     r0, =(0 << REG_MI_FCRAM_CNT0_PASR_SHIFT) | REG_MI_FCRAM_CNT0_WT_MASK
        str     r0, [r3]

        // FCRAMZbg
        ldr     r3, =REG_FCRAM_ADDR
        ldr     r1, =REG_PDN_FCRAM_RSTB_MASK
        orr     r0, r0, r1
        str     r0, [r3]

#endif // BROM_ENABLE_FCRAM

#ifdef BROM_ENABLE_VRAM

        // VRAMNbN
        ldr     r3, =REG_VRAM_ADDR
        ldr     r0, =REG_PDN_VRAM_CLK_E_MASK
        strb    r0, [r3]

#endif // BROM_ENABLE_VRAM

#ifdef BROM_ENABLE_DSP_WRAM

        // DSP-WRAM}bsO
        ldr     r3, =REG_DSPP0_ADDR
        mov     r0, #(0 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #0]
        mov     r0, #(1 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #1]
        mov     r0, #(2 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #2]
        mov     r0, #(3 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #3]
        mov     r0, #(4 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #4]
        mov     r0, #(5 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #5]
        mov     r0, #(6 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #6]
        mov     r0, #(7 << REG_SCFG_DSPP0_OF_SHIFT) | (0 << REG_SCFG_DSPP0_M_MASK) | REG_SCFG_DSPP0_E_MASK
        strb    r0, [r3, #7]

        ldr     r3, =REG_DSPD0_ADDR
        mov     r0, #(0 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #0]
        mov     r0, #(1 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #1]
        mov     r0, #(2 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #2]
        mov     r0, #(3 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #3]
        mov     r0, #(4 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #4]
        mov     r0, #(5 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #5]
        mov     r0, #(6 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #6]
        mov     r0, #(7 << REG_SCFG_DSPD0_OF_SHIFT) | (0 << REG_SCFG_DSPD0_M_MASK) | REG_SCFG_DSPD0_E_MASK
        strb    r0, [r3, #7]

#endif // BROM_ENABLE_DSP_WRAM

LSYM(10)
        b       BSYM(10)
}

