/*---------------------------------------------------------------------------*
  Project:  CtrBrom - library - init
  File:     crt0_excpHandler.c

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/
#include <brom/code32.h>
#include <brom/os.h>
#include <brom/swi.h>


/*---------------------------------------------------------------------------*
  Name:         stupInitExceptions

  Description:  Initialize Exceptions

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
asm void stupInitExceptions( void )
{
        adr     r2, i_stupExcpVennerCode
        ldr     r3, =HW_EXCP_VENEER_BUF
        add     r12, r3, #HW_EXCP_VENEER_BUF_SIZE
LSYM(0)
        ldr     r0, [r2], #4
        ldm     r0, {r0-r1}
        stm     r3!,{r0-r1}
        cmp     r3, r12
        blt     BSYM(0)

#ifdef SDK_ARM11
        // disable VFP
#ifdef SDK_MG20EMU
        fmrx    r12, fpexc
        bic     r12, r12, #HW_FPEXC_VFP_ENABLE
        fmxr    fpexc, r12
#endif // SDK_MG20EMU
#endif // SDK_ARM11

        bx      lr

i_stupExcpVennerCode
        DCD     i_stupExcpTerminateCode  // IRQ
        DCD     i_stupExcpTerminateCode  // FIQ
        DCD     i_stupSwiVeneerCode      // SWI    osInitOprinto\邽߂Őݒ
#ifdef BROM_ENABLE_SCATLD_VFP
        INASM_EXTERN( i_osUndefInstHandler )
        DCD     i_osUndefInstHandler     // UNDEF
#else // BROM_ENABLE_SCATLD_VFP
        DCD     i_stupExcpTerminateCode  // UNDEF
#endif // BROM_ENABLE_SCATLD_VFP
        DCD     i_stupIAbtTerminateCode  // IABT   ߃A{[gƃf[^A{[gʂ邽ߕʊ֐
        DCD     i_stupExcpTerminateCode  // DABT

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupExcpTerminateCode

  Description:  exception veneer code

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
asm void i_stupExcpTerminateCode( void )
{
        ldr     pc, =i_stupExcpTerminateCode
        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupIAbtTerminateCode

  Description:  Prefetch Abort veneer code

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
asm void i_stupIAbtTerminateCode( void )
{
        ldr     pc, =i_stupIAbtTerminateCode
        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupSwiVeneerCode

  Description:  SWI veneer code

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
asm void i_stupSwiVeneerCode( void )
{
        ldr     pc, =i_stupSwiHandler
        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupIrqVeneer

  Description:  IRQ veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupIrqVeneer( void )
{
        ldr     pc, =HW_INTR_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupFiqVeneer

  Description:  FIQ veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupFiqVeneer( void )
{
        ldr     pc, =HW_FIQ_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupSwiVeneer

  Description:  SWI veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupSwiVeneer( void )
{
        ldr     pc, =HW_SWI_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupUndefVeneer

  Description:  Undefine Instruction veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupUndefVeneer( void )
{
        ldr     pc, =HW_UNDEF_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupIAbtVeneer

  Description:  Prefetch Abort veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupIAbtVeneer( void )
{
        ldr     pc, =HW_IABT_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupDAbtVeneer

  Description:  Data Abort veneer

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupDAbtVeneer( void )
{
        ldr     pc, =HW_DABT_VENEER_BUF

        LTORG
}

/*---------------------------------------------------------------------------*
  Name:         i_stupSwiHandler

  Description:  SWI handler

  Arguments:    None

  Returns:      None.
 *---------------------------------------------------------------------------*/
asm void i_stupSwiSemihosting( void )
{
        // 荞݋֎~
        mrs     r1, cpsr
        orr     r1, r1, #(HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        msr     cpsr_fsxc, r1
        ldmfd   sp!, {r0-r1, lr}
        add     sp, sp, #12
        msr     cpsr_fsxc, #(HW_PSR_SVC_MODE | HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        // SVC[h
        ldmfd   sp!, {lr}                      // SP_svc
        msr     spsr_fsxc, lr
        ldmfd   sp!, {lr}
semi_vector
        // Z~zXeBOAhXSVC[hłȂARMfobKG[o
        movs    pc, lr
}

asm void i_stupSwiHandler( void )
{
        INASM_EXTERN( _start )

        // smf[^A{[gLiclear HW_PSR_IMPR_ABORT_DISABLEj
        // FIQnhłSWIgpSP_svc㏑h~̂ߍŏFIQ֎~iIRQHW֎~j
        msr     cpsr_fsxc, #(HW_PSR_SVC_MODE | HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)

        // SVC[h
        mrs     sp, spsr                        // SP_svc
        and     sp, sp, #HW_PSR_CPU_MODE_MASK
        // [U[hȂVXe[h
        cmp     sp, #HW_PSR_USR_MODE
        moveq   sp, #HW_PSR_SYS_MODE
        // 荞݋֎~
        orr     sp, sp, #(HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        msr     cpsr_fsxc, sp

        // Ăяõ[h
        sub     sp, sp, #12
        stmfd   sp!, {r0-r1, lr}                // ̃^C~OŃf[^A{[gnhSWIgpSVCWX^͕As\
        add     r1, sp, #24                     // ĂяõX^bN֊mۂSPSR_svcALR_svcpX^bNAhXRs[
        msr     cpsr_fsxc, #(HW_PSR_SVC_MODE | HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)

        // SVC[h
        mov     sp, r1                          // SP_svc
        mov     r1, lr
        mrs     r0, spsr
        stmfd   sp!, {r0, lr}                   // SPSR_svcALR_svcޔiLR_svc͍ŌPOP̂ňSj
        // 荞ݏԌp
        and     r0, r0, #(HW_PSR_CPU_MODE_MASK | HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        // [U[hȂ犄荞݋֎~ŃVXe[h
        // i荞݋֎~ɂ͕̂ʃXbhX^bN֑ޔPC
        //  āAVXe[hȂ悤ɂ邽߁B
        //  AADMAɂnbLO֑Ήɂ̓[U[hSVC[h
        //  X^bN𕪂KviOS̃bp[őΉ\jBj
        tst     r0, #(HW_PSR_SYS_MODE ^ HW_PSR_USR_MODE)
        orreq   r0, r0, #(HW_PSR_SYS_MODE | HW_PSR_IRQ_DISABLE)
        msr     cpsr_fsxc, r0

        // Ăяõ[h
        ldrh    r0, [r1, #-2]                  // VXeR[No擾
        mov     r0, r0, lsl #1
        and     r0, r0, #SVC_ID_PREMASK<<SVC_ID_SHIFT

        cmp     r0, #(SVC_ID_SEMIHOST_ARM   & SVC_ID_PREMASK)<<SVC_ID_SHIFT  // SVC_ID_SEMIHOST_ARM   = 0x12
        cmpne   r0, #(SVC_ID_SEMIHOST_THUMB & SVC_ID_PREMASK)<<SVC_ID_SHIFT  // SVC_ID_SEMIHOST_THUMB = 0xab
        beq     i_stupSwiSemihosting

        adr     r1, SWI_Table                  // VXeR[AhXi2oCge[uj擾
        ldrh    r0, [r1, r0]

        ldr     r1, =HW_BROM
        orr     r0, r0, r1

        blx     r0

swi_return
        // 荞݋֎~
        mrs     r1, cpsr
        orr     r1, r1, #(HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        msr     cpsr_fsxc, r1
        ldmfd   sp!, {r0-r1, lr}
        add     sp, sp, #12
        msr     cpsr_fsxc, #(HW_PSR_SVC_MODE | HW_PSR_IRQ_DISABLE | HW_PSR_FIQ_DISABLE)
        // SVC[h
        ldmfd   sp!, {lr}                      // SP_svc
        msr     spsr_fsxc, lr
        ldmfd   sp!, {lr}
        movs    pc, lr

        LTORG

#include <../build/libraries/swi/common/swi_table.c>
}

