/*---------------------------------------------------------------------------*
  Project:  CTR - NAND driver
  File:     nandif.c

  Copyright 2006 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include <brom.h>
#include <ioreg_MI.h>
#include <brom/nand/nand.h>

#include "nandif_ip.h"
#include "nandif_reg.h"

//#define PRINTDEBUG  osTPrintf
//#define PRINTLEVEL2 osTPrintf
#define PRINTDEBUG( ...) ((void)0)
#define PRINTLEVEL2( ...) ((void)0)


/***********************************************************************

***********************************************************************/
#define NAND_STACK_SIZE                 (8192)
#define NAND_THREAD_PRIO                (10)

//NAND_OPERATION_INIT͂Ȃ(InitŃ^XN𗧂Ă邽߁j
#define NAND_OPERATION_ENABLE           (0)
#define NAND_OPERATION_FORMAT           (1)
#define NAND_OPERATION_READ             (2)
#define NAND_OPERATION_WRITE            (3)
#define NAND_OPERATION_FLUSH            (4)

#define NAND_FLUSH_RETRY_COUNT          (8)

/*i_nandFlushPageCache֐̕Ԃl*/
#define I_NAND_ERR_SUCCESS		    (0)  //
#define I_NAND_ERR_VERIFY_ERR       (-1) //}[Wŏ
#define I_NAND_ERR_FATAL_ERR        (-2) //sԍ
#define I_NAND_ERR_IMPOSSIBLE_ERR   (-3) //freeubNȂ

/**/
#define I_NAND_FLUSH_FALSE          (-1)
//#define I_NAND_FLUSH_TRUE           (1)
#define I_NAND_FLUSH_CASHED         (-2)

extern u32 nand_ecc[4];


/*---------------------------------------------------------------------------*
    API(ʑw^XN) - nand^XNԒʐMp\
 *---------------------------------------------------------------------------*/
typedef struct {
    NandPageCacheFormat*   cache;
    void*                   buf;
    u32                     sector;
    u32                     sector_count;
    u32                     operation;
//    ID                      tskid;          /*nandAPIĂ񂾃^XNID*/
} NandMsg;


/*---------------------------------------------------------------------------*
    _ubN-ubN ϊe[u
 *---------------------------------------------------------------------------*/
u16 nandBlkTbl[1024];   /*physical_blk = nandBlkTbl[logical_blk]*/

//i_nandWritePage,i_nandEraseBlock̕ԂlBadBlockFɂ́A
//XyÄȊȌꏊBadBlockǗKv
//NandSysFormat nandSys[1024];


/*---------------------------------------------------------------------------*
   ubNXe[^X̃NXg
 *---------------------------------------------------------------------------*/
NandBlockStatInfo   NandBlockStat[1024];
NandStatInfo        NandStat;


/*---------------------------------------------------------------------------*
    globalAvP[Vɂ͔J̊֐
 *---------------------------------------------------------------------------*/
void i_nandEnable( NandPageCacheFormat* NandPageCache);
void i_nandFormat( void);
void i_nandReset( void);
void i_nandReadSector( NandPageCacheFormat* NandPageCache, u32* dest, u32 logical_sector);
void i_nandWriteSector( NandPageCacheFormat* NandPageCache, u32* src, u32 logical_sector);
BOOL i_nandFlush( NandPageCacheFormat* NandPageCache);

int i_nandFlushPageCache( NandPageCacheFormat* NandPageCache);
void i_nandMarkingBadBlock( u16 physical_blk, NandSpareFormat* NandSpareOrg);


/*---------------------------------------------------------------------------*
    static֐
 *---------------------------------------------------------------------------*/
static u8   i_nandCheckBlock( u16 physical_blk, u16* logical_adr, u32* erase_count);
static BOOL i_nandCheckFollowCluster( u16 physical_blk, u16* logical_adr, u32* erase_count);
static u8   i_nandCountBitDifferent( u8 dat1, u8 dat2);

void i_nandInsertStatList( NandBlockStatInfo** NandBlockStatList, NandBlockStatInfo* NandBlockStat);
void i_nandRemoveStatList( NandBlockStatInfo** NandBlockStatList, NandBlockStatInfo* NandBlockStat);
static BOOL i_nandCheckECC( u32* data, u32* stored_ecc, u32 new_ecc);

void nandFillPageCache( NandPageCacheFormat* NandPageCache, u32 physical_page);
static void nandReadPageCache( NandPageCacheFormat* NandPageCache, u32* dest, u16 cache_sect);
static void nandWritePageCache( NandPageCacheFormat* NandPageCache, u32* src, u16 cache_sect);

BOOL i_nandMergeBlock( NandPageCacheFormat* NandPageCache);

static void i_nandThread( void* arg);


/*---------------------------------------------------------------------------*
    staticϐ
 *---------------------------------------------------------------------------*/
static BOOL     nand_tsk_created = FALSE;

/*---------------------------------------------------------------------------*
    globalϐ
 *---------------------------------------------------------------------------*/
OSThread       nand_tsk;
OSMessageQueue nand_dtq;
OSMessage      nand_dtq_array[1];
OSMessageQueue nand_result_dtq;
OSMessage      nand_result_dtq_array[1];

u64 nand_stack[NAND_STACK_SIZE / sizeof(u64)];

NandPageCacheFormat*       g_nandPageCache;


//u64                           nand_stack[NAND_STACK_SIZE / sizeof(u64)];





/*---------------------------------------------------------------------------*
  Name:         mystdCompareString

  Description:  compare strings. same to strcmp

  Arguments:    str1, str2 : strings

  Returns:      0 if same
 *---------------------------------------------------------------------------*/
int mystdCompareString(const char *str1, const char *str2);
int mystdCompareString(const char *str1, const char *str2)
{
    while (*str1 == *str2 && *str1)
    {
        str1++;
        str2++;
    }
    return (*str1 - *str2);
}


/*---------------------------------------------------------------------------*
  Name:         nandInit

  Description:

  Arguments:

  Returns:
 *---------------------------------------------------------------------------*/
void nandInit( void)
{
    NandMsg     nandMsg;
    u32         init_msg;

    if( nand_tsk_created == FALSE) {
        /*---------- OS ----------*/
        osInitMessageQueue( &nand_dtq,        &nand_dtq_array[0],        1);
        osInitMessageQueue( &nand_result_dtq, &nand_result_dtq_array[0], 1);

        osCreateThread( &nand_tsk, i_nandThread, NULL, (nand_stack+NAND_STACK_SIZE / sizeof(u64)),
                        NAND_STACK_SIZE, NAND_THREAD_PRIO);
      
        osWakeupThreadDirect( &nand_tsk);
        /*----------------------------*/
        nand_tsk_created = TRUE;
    }

    i_nandReset();
}


/*---------------------------------------------------------------------------*
  Name:         i_nandEnable

  Description:  ubNe[u쐬

  Arguments:

  Returns:
 *---------------------------------------------------------------------------*/
void i_nandEnable( NandPageCacheFormat* NandPageCache)
{
    u16             i;
    u16             physical_blk;
    u16             logical_blk;
    NandPageFormat  NandPage;
    u16             ctrdg_reg;
    u8              block_stat;
    u32             erase_count;
    u16             free_blk_count = 0;
    u16             busy_blk_count = 0;
    u16             bad_blk_count = 0;

    miCpuFill8( nandBlkTbl, 0xFF, 2048);    //SĖ蓖ďԂ
    NandStat.busy = NULL;
    NandStat.free = NULL;

    osTPrintf( "nand block checking\n");
    for( i=0; i<1024; i++) {
        if( (i % 16) == 0) {
            osTPrintf( ".");
        }
        block_stat = i_nandCheckBlock( i, &logical_blk, &erase_count);
        physical_blk = i;

        /*e[u̍\z*/
        if( (logical_blk < 1024)&&(block_stat == 0xFF)) {//蓖(0xFFFF)͂
            nandBlkTbl[logical_blk] = physical_blk;
            if( logical_blk == 0) {
                PRINTDEBUG( "nandEnable : physical:0x%x, logical:0x%x\n", physical_blk, logical_blk);
            }
        }

        /*Xe[^X̃NXg\z*/
        NandBlockStat[physical_blk].physical_blk = physical_blk;
        NandBlockStat[physical_blk].block_stat_bad = block_stat;
        NandBlockStat[physical_blk].erase_count = erase_count;
        NandBlockStat[physical_blk].next = NULL;
        if( (logical_blk < 1024)&&(block_stat == 0xFF)) {
            i_nandInsertStatList( &(NandStat.busy), &(NandBlockStat[physical_blk]));
            PRINTLEVEL2( "busy [physical_blk:0x%x, erase_count:0x%x]\n", physical_blk, NandBlockStat[physical_blk].erase_count);
            busy_blk_count++;
        }else{
            if( block_stat == 0xFF) {
                i_nandInsertStatList( &(NandStat.free), &(NandBlockStat[physical_blk]));
                PRINTLEVEL2( " free [physical_blk:0x%x, erase_count:0x%x]\n", physical_blk, NandBlockStat[physical_blk].erase_count);
                free_blk_count++;
            }else{
                bad_blk_count++;
                PRINTLEVEL2( "%s : bad block found ... blk 0x%x(physical)\n", __FUNCTION__, physical_blk);
            }
        }
    }
    osTPrintf( "\n");
    PRINTLEVEL2( "%s : NandStat(LinkList) busy:0x%x, free:0x%x, and bad:0x%x\n",
                __FUNCTION__, busy_blk_count, free_blk_count, bad_blk_count);

    /**/
    for( i=0; i<64; i++) {
        NandPageCache->page_flushed[i] = I_NAND_FLUSH_FALSE;
    }
    NandPageCache->page_flushed_count = 0;

    
    NandPageCache->valid = FALSE;

    NandPageCache->OnTheFlush = 0xFFFF;
    NandPageCache->BlockDirty = FALSE;

    /*globalϐɓo^*/
    g_nandPageCache = NandPageCache;
}

/*NXgNandBlockStat}i񐔏j*/
void i_nandInsertStatList( NandBlockStatInfo** NandBlockStatList, NandBlockStatInfo* NandBlockStat)
{
    NandBlockStatInfo   NandBlockStatDmy;
    NandBlockStatInfo*  CurrentNandBlockStat;

    /*O*/
    NandBlockStat->next = NULL;
    
    if( *NandBlockStatList == NULL) {
        *NandBlockStatList = NandBlockStat;
        return;
    }else{
        NandBlockStatDmy.next = *NandBlockStatList;
        CurrentNandBlockStat = &NandBlockStatDmy;
        while( CurrentNandBlockStat->next != NULL) {
            if( NandBlockStat->erase_count < CurrentNandBlockStat->next->erase_count) {
                break;
            }
            CurrentNandBlockStat = CurrentNandBlockStat->next;
        }
        NandBlockStat->next = CurrentNandBlockStat->next;
        CurrentNandBlockStat->next = NandBlockStat;
    }

    /*擪ɑ}ꍇ́ANXg̊Jnʒûς*/
    if( CurrentNandBlockStat == &NandBlockStatDmy) {
        *NandBlockStatList = CurrentNandBlockStat->next;
    }
}

/*NXgNandBlockStat𔲂o*/
void i_nandRemoveStatList( NandBlockStatInfo** NandBlockStatList, NandBlockStatInfo* NandBlockStat)
{
    NandBlockStatInfo   NandBlockStatDmy;
    NandBlockStatInfo*  CurrentNandBlockStat;

    if( *NandBlockStatList == NULL) {
        return;
    }
    NandBlockStatDmy.next = *NandBlockStatList;
    CurrentNandBlockStat = &NandBlockStatDmy;
    while( CurrentNandBlockStat->next != NandBlockStat) {
        if( CurrentNandBlockStat->next == NULL) {
            return;
        }else{
            CurrentNandBlockStat = CurrentNandBlockStat->next;
        }
    }
    CurrentNandBlockStat->next = CurrentNandBlockStat->next->next;
    *NandBlockStatList = NandBlockStatDmy.next;
}

/*---------------------------------------------------------------------------*
  Name:         i_nandFormat

  Description:  tH[}bgiXyÄ\zj

  Arguments:

  Returns:
 *---------------------------------------------------------------------------*/
void i_nandFormat( void)
{
    u16             i, j;
    u32             physical_page_addr;
    NandPageFormat  NandPage;           /*ݗpƏꏊ*/
    u16             good_blk_count;
    u8              block_stat;
    u16             logical_blk;
    u32             erase_count;

    miCpuFill8( &(NandPage.data.data[0][0]), 0xFF, NAND_PAGE_DATASIZE);
    miCpuFill8( &(NandPage.spare), 0xFF, NAND_PAGE_SPARESIZE);
    miCpuCopy8( "CTR\0", &(NandPage.spare.sig0[0]), 4);
    miCpuCopy8( "CTR\0", &(NandPage.spare.sig1[0]), 4);

    good_blk_count = 0;
    osTPrintf( "formatting");
    for( i=0; i<1024; i++) {
        if( (i % 16) == 0) {
            osTPrintf( ".");
        }
        /*obhubNǂ`FbN*/
        block_stat = i_nandCheckBlock( i, &logical_blk, &erase_count);
        if( block_stat == 0xFF) {
            NandPage.spare.block_stat_bad     = block_stat;
            NandPage.spare.block_erase_count0 = erase_count;
            NandPage.spare.block_erase_count1 = erase_count;
//            if( good_blk_count > NAND_GUARANTEED_BLOCKS) {
                NandPage.spare.logical_adr0       = 0xFFFF; //蓖
                NandPage.spare.logical_adr1       = 0xFFFF;
//            }else{
//                NandPage.spare.logical_adr0       = good_blk_count;
//                NandPage.spare.logical_adr1       = good_blk_count;
//            }
            
            if( i_nandEraseBlock( i) == FALSE) {
                PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                i_nandMarkingBadBlock( i, &(NandPage.spare));
            }else{
                good_blk_count++;
                for( j=0; j<64; j++) {
                    physical_page_addr = ((u32)(i)<<6) + j;
                    //TODO:i_nandWritePagẽG[R[h
                    if( i_nandWritePage( &NandPage, physical_page_addr, 1) == FALSE) {
                        PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                        i_nandMarkingBadBlock( i, &(NandPage.spare));
                        good_blk_count--;
                        break;
                    }
                }
            }
        }
    }

    osTPrintf( "\n");
    osTPrintf( "%s : good_blk_count = 0x%x\n", __FUNCTION__, good_blk_count);
}


/*---------------------------------------------------------------------------*
  Name:         i_nandCheckBlock

  Description:  ubN`FbN

  Arguments:    physical_blk : `FbNubNԍ

  Returns:      Ԃl0xFF(good block)̂ƂAlogical_adr
                erase_countlLBԂl0x00(bad block)̂Ƃ
                ǂlƂȂ܂B
 *---------------------------------------------------------------------------*/
u8 i_nandCheckBlock( u16 physical_blk, u16* logical_adr, u32* erase_count)
{
    u32             physical_page_addr;
    NandPageFormat  NandDevPage1;       /*[hy[W0̃f[^*/
    NandSpareFormat NandDevSpare0;      /*[hy[W1̃XyÄ*/
    u8              new_block_stat;
    u32             new_erase_count = 0;

    /**/
    physical_page_addr = (u32)((u32)(physical_blk) << 6);
    i_nandReadPage( &NandDevPage1, physical_page_addr, 1);
    miCpuCopy8( &(NandDevPage1.spare), &NandDevSpare0, NAND_PAGE_SPARESIZE);
    i_nandReadPage( &NandDevPage1, (physical_page_addr+1), 1);

    /*----- BBtO0xFF̂Ƃ -----*/
    if( (NandDevSpare0.block_stat_bad == 0xFF) &&
        (NandDevPage1.spare.block_stat_bad == 0xFF)) {

        new_block_stat = 0xFF;
        if( (mystdCompareString( NandDevSpare0.sig0, "CTR\0") == 0) ||
            (mystdCompareString( NandDevSpare0.sig1, "CTR\0") == 0)) {
            new_erase_count = NandDevSpare0.block_erase_count0; //todo:parity check
        }else{/*--- pre slip writeς݂post slip writeĂȂƂɑΉ ---*/
            if( i_nandCheckFollowCluster( physical_blk, logical_adr,
                                            erase_count) == TRUE) {
                return( new_block_stat);
            }
        }
    }else{
        /*----- BBtO1bit0Ƃ -----*/
        if( (i_nandCountBitDifferent( NandDevSpare0.block_stat_bad, 0xFF) <= 1) &&
            (i_nandCountBitDifferent( NandDevPage1.spare.block_stat_bad, 0xFF) <= 1)) {
            if( (mystdCompareString( NandDevSpare0.sig0, "CTR\0") == 0) ||
                (mystdCompareString( NandDevSpare0.sig1, "CTR\0") == 0)) {
                new_block_stat = 0xFF;
                new_erase_count = NandDevSpare0.block_erase_count0; //todo:parity check
            }else{/*--- pre slip writeς݂post slip writeĂȂƂɑΉ ---*/
                if( i_nandCheckFollowCluster( physical_blk, logical_adr,
                                                erase_count) == TRUE) {
                    new_block_stat = 0xFF;
                    return( new_block_stat);
                }else{
                    //signatureȂ΃obhubN
                    new_block_stat = 0x00;
                }
            }
        }else{/*----- BBtO2bitȏ0Ƃ -----*/
            new_block_stat = 0x00;
        }
    }

    *logical_adr = NandDevSpare0.logical_adr0;  //todo:parity check
    *erase_count = new_erase_count;
    return( new_block_stat);
}

/*---------------------------------------------------------------------------*
  Name:         i_nandCheckFollowCluster

  Description:  wubNɂNX^1`7signature`FbN

  Arguments:    physical_blk : `FbNubNԍ

  Returns:      Ԃl0xFF(good block)̂ƂAlogical_adr
                erase_countlLBԂl0x00(bad block)̂Ƃ
                ǂlƂȂ܂B
 *---------------------------------------------------------------------------*/
static BOOL i_nandCheckFollowCluster( u16 physical_blk, u16* logical_adr, u32* erase_count)
{
    u16             i;
    u32             physical_page_addr;
    NandPageFormat  NandDevPage;

    for( i=1; i<8; i++) {
        physical_page_addr = (u32)(((u32)(physical_blk) << 6) + (i * 8));
        i_nandReadPage( &NandDevPage, physical_page_addr, 1);

        if( (mystdCompareString( NandDevPage.spare.sig0, "CTR\0") == 0) ||
            (mystdCompareString( NandDevPage.spare.sig1, "CTR\0") == 0)) {
            *logical_adr = NandDevPage.spare.logical_adr0;  //todo:parity check
            *erase_count = NandDevPage.spare.block_erase_count0;    //todo:parity check
            return( TRUE);
        }
    }

    return( FALSE);
}

/* 8bitf[^rāAقȂrbg̐Ԃ */
static u8 i_nandCountBitDifferent( u8 dat1, u8 dat2)
{
    u8 dif_count;
    u8 i;

    dif_count = 0;
    for( i=0; i<8; i++) {
        if( ((dat1>>i)&0x01) != ((dat2>>i)&0x01)) {
            dif_count++;
        }
    }

    return( dif_count);
}

/*---------------------------------------------------------------------------*
  Name:         i_nandReadSector

  Description:  _ZN^ǂݏo

  Arguments:    logical_sector : _ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void i_nandReadSector( NandPageCacheFormat* NandPageCache, u32* dest, u32 logical_sector)
{
    u16     logical_blk;
    u16     physical_blk;
    u32     logical_page;
    u32     physical_page;
    u16     cache_sect;     /*PageCachẽZN^ԍ*/
    u16     cache_offset;
    u32*    cache_sect_adr;
    int     result;
    NandBlockStatInfo* AssignNandBlockStat;

    logical_blk = (u16)(logical_sector / 256);      //1blk = 64*4 sectors
    logical_page = (u32)(logical_sector / 4);
    cache_sect = (u16)(logical_sector & 0x3);  //0`3

    /*----- }[W -----*/
    if( (NandPageCache->valid == TRUE)&&(NandPageCache->dirty == TRUE)) {
        result = i_nandFlushPageCache( NandPageCache);
        if( (result == I_NAND_ERR_VERIFY_ERR) ||
            (result == I_NAND_ERR_SUCCESS)) {
            i_nandMergeBlock( NandPageCache);
        }else{
            PRINTDEBUG( "%s(%d) : fatal error!\n", __FUNCTION__, __LINE__);
            while( 1) {};
        }
    }else{
        if( NandPageCache->BlockDirty == TRUE) {
            i_nandMergeBlock( NandPageCache);
        }else{
            if( NandPageCache->page_flushed[(logical_page & 0x3F)] != I_NAND_FLUSH_FALSE) {
                PRINTDEBUG( "%s(%d) : fatal error!\n", __FUNCTION__, __LINE__);
                while( 1) {};
            }
        }
    }
    /*----------------------*/
    
    /*}[WŃe[uXVĂ邩mȂ̂Ń}[W̌ŌvZ*/
    physical_blk = nandBlkTbl[logical_blk];
    physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));

    /*----- AhX蓖Ă̂Ƃ -----*/
    //miCpuFill8( dest, 0xFF, 512);łȂ
    if( physical_blk == 0xFFFF) {
        if( (NandStat.free)&& //t[ubNΊ蓖
            (logical_blk < NAND_GUARANTEED_BLOCKS)) {
            AssignNandBlockStat = NandStat.free;
            nandBlkTbl[logical_blk] = NandStat.free->physical_blk;
            
            physical_blk = nandBlkTbl[logical_blk];
            physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));
            
            i_nandRemoveStatList( &(NandStat.free), AssignNandBlockStat);
            i_nandInsertStatList( &(NandStat.busy), AssignNandBlockStat);
        }else{
            return;
        }
    }/*-----------------------------------------*/
    
    /*PageCachȅꍇ*/
    if( NandPageCache->valid == FALSE) {
        nandFillPageCache( NandPageCache, physical_page);
        NandPageCache->logical_page = logical_page;  //PageCache̘_y[WԍXV
    }else{
        /*PageCacheLqbgƂ*/
        if( (NandPageCache->valid == TRUE) &&
            (NandPageCache->logical_page == logical_page)) {
            PRINTDEBUG( "%s : PageCache Hit! (physical:0x%x, logical:0x%x)\n", __FUNCTION__, physical_page, logical_page);
        }else{
            /*PageCacheLqbgȂƂ*/
            if( NandPageCache->dirty == TRUE) {
                result = i_nandFlushPageCache( NandPageCache); //܂ł̂ԂFlush
                if( (result == I_NAND_ERR_SUCCESS)||(result == I_NAND_ERR_VERIFY_ERR)) {
                    i_nandMergeBlock( NandPageCache); //}[W
				    /*}[WŃe[uXVĂ邩mȂ̂Ń}[W̌ŌvZ*/
				    physical_blk = nandBlkTbl[logical_blk];
				    physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));
                }else{
                    while( 1) {};
                }
            }
            nandFillPageCache( NandPageCache, physical_page);
            NandPageCache->logical_page = logical_page;  //PageCache̘_y[WԍXV
        }
    }

    /*ECCr & */
    cache_offset = cache_sect;//(u16)(cache_sect % 4);
    cache_sect_adr = (u32*)((u32)&(NandPageCache->NandPage[0].data.data[cache_offset][0]));

    if( i_nandCheckECC( cache_sect_adr,
                        &(NandPageCache->NandPage[0].spare.ecc[cache_offset]),
                        NandPageCache->ecc[0][cache_offset]) == FALSE) {
        PRINTLEVEL2( "ECC error : bad block generated. %s, %d\n", __FUNCTION__, __LINE__);
        /*G[iłǂɊւ炸j*/
        NandPageCache->NandSpareWriteBack.block_stat_bad = 0;  /*obhubNE}[LO*/
        NandBlockStat[physical_blk].block_stat_bad = 0;
        NandPageCache->dirty = TRUE;   /*dirtyɂĂȂƃobhubNE}[LO߂Ȃ*/
    }

    /*PageCachef[^ǂݏo*/
    nandReadPageCache( NandPageCache, dest, cache_sect);

    return;
}

/*---------------------------------------------------------------------------*
  Name:         i_nandCheckECC

  Description:  ECCpăf[^̌

  Arguments:    data : 512oCgf[^̃AhX
                stored_ecc :
                new_ecc :

  Returns:      TRUE : good ( No Error)
                FALSE : bad ( ECC correction or Uncorrectable Error)
 *---------------------------------------------------------------------------*/
static BOOL i_nandCheckECC( u32* data, u32* stored_ecc, u32 new_ecc)
{
#if 0
    return TRUE;
#else
    u32 xored_ecc;
    u16 i;
    u16 bit_count;
    u16 bit_adr;
    u16 word_adr;

    /*vȂrbg̐JEg*/
    xored_ecc = (*stored_ecc) ^ new_ecc;
    if( xored_ecc == 0) {
        PRINTDEBUG( "%s : no err\n", __FUNCTION__);
        return TRUE;        /*G[Ȃ*/
    }
    bit_count = 0;
    for( i=0; i<24; i++) {
        if( xored_ecc & ( (u32)(1)<<i)) {
            bit_count++;
        }
    }

    /*\ȃG[*/
    if( bit_count == 12) {
        bit_adr = (u16)(
            ((xored_ecc >> 1) & 0x01) |     /* CP9,CP8,CP7,CP6,CP5,CP4,CP3,CP2,CP1 */
            ((xored_ecc >> 2) & 0x02) |     /*     CP9,CP8,CP7,CP6,CP5,CP4,CP3,CP2 */
            ((xored_ecc >> 3) & 0x04) |     /*         CP9,CP8,CP7,CP6,CP5,CP4,CP3 */
            ((xored_ecc >> 4) & 0x08) |     /*             CP9,CP8,CP7,CP6,CP5,CP4 */
            ((xored_ecc >> 5) & 0x10) );    /*                 CP9,CP8,CP7,CP6,CP5 */
        word_adr = (u16)(
            ((xored_ecc >> 11) & 0x01) |    /* LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6,LP5,LP4,LP3,LP2,LP1 */
            ((xored_ecc >> 12) & 0x02) |    /*     LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6,LP5,LP4,LP3,LP2 */
            ((xored_ecc >> 13) & 0x04) |    /*         LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6,LP5,LP4,LP3 */
            ((xored_ecc >> 14) & 0x08) |    /*             LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6,LP5,LP4 */
            ((xored_ecc >> 15) & 0x10) |    /*                 LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6,LP5 */
            ((xored_ecc >> 16) & 0x20) |    /*                     LP13,LP12,LP11,LP10,LP9,LP8,LP7,LP6 */
            ((xored_ecc >> 17) & 0x40) );   /*                         LP13,LP12,LP11,LP10,LP9,LP8,LP7 */

        data[word_adr] = data[word_adr] ^ (0x01 << bit_adr);
        *(u32*)stored_ecc = new_ecc;
        PRINTDEBUG( "%s : error corrected !!!\n", __FUNCTION__);
        return FALSE;
    }
    /*ECCR[ĥ̃G[*/
    if( bit_count == 1) {
        *(u32*)stored_ecc = new_ecc;
        PRINTDEBUG( "%s : err of ecc own!!!\n", __FUNCTION__);
        return FALSE;
    }
    /*s\ȏꍇ*/
    PRINTDEBUG( "%s : fatal error!!!\n", __FUNCTION__);
    return FALSE;
#endif
}

/*---------------------------------------------------------------------------*
  Name:         i_nandWriteSector

  Description:  _ZN^

  Arguments:    logical_sector : _ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void i_nandWriteSector( NandPageCacheFormat* NandPageCache, u32* src, u32 logical_sector)
{
    u16 logical_blk;
    u16 physical_blk;
    u32 logical_page;
    u32 physical_page;
    u16 cache_sect;
    int result;
    NandBlockStatInfo* AssignNandBlockStat;

    logical_blk = (u16)(logical_sector / 256);  //1blk = 64*4 sectors
    logical_page = (u32)(logical_sector / 4);
    cache_sect = (u16)(logical_sector & 0x3);  //0`255

    /*----- }[W -----*/
    if( (NandPageCache->page_flushed[(logical_page & 0x3F)] != I_NAND_FLUSH_FALSE) ||
        ((NandPageCache->logical_page >> 6) != logical_blk)) {

        if( (NandPageCache->valid == TRUE)&&(NandPageCache->dirty == TRUE)) {    //܂Flush
            result = i_nandFlushPageCache( NandPageCache); //܂ł̂ԂFlush
            if( (result == I_NAND_ERR_VERIFY_ERR) ||
                (result == I_NAND_ERR_SUCCESS)) {
                i_nandMergeBlock( NandPageCache);
            }else{
                PRINTDEBUG( "%s : fatal error!\n", __FUNCTION__);
                while( 1) {};
            }
        }else{
            if( NandPageCache->BlockDirty == TRUE) {
                i_nandMergeBlock( NandPageCache);
            }
        }
    }
    /*----------------------*/

    /*}[WŃe[uXVĂ邩mȂ̂Ń}[W̌ŌvZ*/
    physical_blk = nandBlkTbl[logical_blk];
    physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));

    /*----- AhX蓖Ă̂Ƃ -----*/
    if( physical_blk == 0xFFFF) {
        if( (NandStat.free)&& //t[ubNΊ蓖
            (logical_blk < NAND_GUARANTEED_BLOCKS)) {
            AssignNandBlockStat = NandStat.free;
            nandBlkTbl[logical_blk] = NandStat.free->physical_blk;
            
            physical_blk = nandBlkTbl[logical_blk];
            physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));
            
            i_nandRemoveStatList( &(NandStat.free), AssignNandBlockStat);
            i_nandInsertStatList( &(NandStat.busy), AssignNandBlockStat);
        }else{
            return;
        }
    }/*-----------------------------------------*/
    
    /*PageCachȅꍇ*/
    if( NandPageCache->valid == FALSE) {
        nandFillPageCache( NandPageCache, physical_page);
        NandPageCache->logical_page = logical_page;  //PageCache̘_ubNԍXV
    }else{
        /*PageCacheLqbgƂ*/
        if( (NandPageCache->valid == TRUE) &&
            (NandPageCache->logical_page == logical_page)) {
            PRINTDEBUG( "%s : PageCache Hit! (physical:0x%x, logical:0x%x)\n", __FUNCTION__, physical_page, logical_page);
        }else{
            /*PageCacheLqbgȂƂ*/
            if( NandPageCache->dirty == TRUE) {
                result = i_nandFlushPageCache( NandPageCache); //܂ł̂ԂFlush
                if( (result == I_NAND_ERR_FATAL_ERR)||
                    (result == I_NAND_ERR_IMPOSSIBLE_ERR)) {
                    while( 1) {};
                }
                if( result == I_NAND_ERR_VERIFY_ERR) { //Flushs͂Ƀ}[WăJo
                    i_nandMergeBlock( NandPageCache);
                    /*}[WŃe[uXVĂ邩mȂ̂Ń}[W̌ŌvZ*/
                    physical_blk = nandBlkTbl[logical_blk];
                    physical_page = (u32)(((u32)(physical_blk) << 6) + (logical_page & 0x3F));
                }
            }
            nandFillPageCache( NandPageCache, physical_page);
            NandPageCache->logical_page = logical_page;  //PageCache̘_ubNԍXV
        }
    }

    nandWritePageCache( NandPageCache, src, cache_sect);
    
    NandPageCache->dirty = TRUE;
    return;
}

/*---------------------------------------------------------------------------*
  Name:         nandFillPageCache

  Description:  ubNPageCacheɓǂݏo

  Arguments:    physical_blk : ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void nandFillPageCache( NandPageCacheFormat* NandPageCache, u32 physical_page)
{
    u16 i;
    u32 physical_page_addr;

    PRINTDEBUG( "%s : Fill PageCache! (physical_page:0x%x)\n", __FUNCTION__, physical_page);
   
    for( i=0; i<NAND_CACHE_PAGES; i++) {
        physical_page_addr = (physical_page + i);
        i_nandReadPage( &(NandPageCache->NandPage[i]), physical_page_addr, 1);

        /*ꂽECCʂ̗̈ɑޔĂ*/
        miCpuCopy8( &(nand_ecc[0]), &(NandPageCache->ecc[i][0]), 4);
        miCpuCopy8( &(nand_ecc[1]), &(NandPageCache->ecc[i][1]), 4);
        miCpuCopy8( &(nand_ecc[2]), &(NandPageCache->ecc[i][2]), 4);
        miCpuCopy8( &(nand_ecc[3]), &(NandPageCache->ecc[i][3]), 4);
    }

    /*oldubNɃCgobNƂ̂߂̃XyÄ*/
    miCpuCopy8( &(NandPageCache->NandPage[0].spare), &(NandPageCache->NandSpareWriteBack),
                NAND_PAGE_SPARESIZE);

    NandPageCache->valid = TRUE;
    NandPageCache->dirty = FALSE;
    /*ĂяoNandPageCache->logical_blkXV邱*/
}


/*---------------------------------------------------------------------------*
  Name:         i_nandFlush

  Description:

  Arguments:

  Returns:
 *---------------------------------------------------------------------------*/
BOOL i_nandFlush( NandPageCacheFormat* NandPageCache)
{
    int result;

    result = 0;
    /*--- y[WLbṼtbV ---*/
    if( (NandPageCache->valid == TRUE) && (NandPageCache->dirty == TRUE)) {
        result = i_nandFlushPageCache( NandPageCache);
        if( (result == I_NAND_ERR_FATAL_ERR)||
            (result == I_NAND_ERR_IMPOSSIBLE_ERR)) {
            PRINTDEBUG( "%s : fatal error!\n", __FUNCTION__);
            while( 1) {};
//            return( FALSE);
        }else{
            i_nandMergeBlock( NandPageCache);
        }
    }/*----------------------------------*/

    /*--- Block̃}[W ---*/
    if( NandPageCache->BlockDirty == TRUE) {
        if( i_nandMergeBlock( NandPageCache) == FALSE) {
            PRINTDEBUG( "%s : fatal error!\n", __FUNCTION__);
            while( 1) {};
//            return( FALSE);
        }
    }/*--------------------*/
    
    return( TRUE);
}


/*---------------------------------------------------------------------------*
  Name:         i_nandFlushPageCache

  Description:  PageCache畨ubNɏo

  Arguments:    physical_blk : ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
int i_nandFlushPageCache( NandPageCacheFormat* NandPageCache)
{
    u16                 i, k;
    u16                 new_physical_blk;
    u32                 new_physical_page_addr;
//    u16                 old_physical_blk;
    u32                 old_physical_page_addr;
    NandPageFormat      NandPage;
    BOOL                rslt_writepage;

    PRINTDEBUG( "%s\n", __FUNCTION__);

    /**/
//    miCpuFill8( bad_physical_blk, 0xFF, (NAND_FLUSH_RETRY_COUNT*sizeof(bad_physical_blk)));

    /*---------- FlushNewBlockT[` ----------*/
    if( NandPageCache->OnTheFlush == 0xFFFF) {
        if( NandStat.free) {
            do {
                NandPageCache->TargetNandBlockStat = NandStat.free;
                new_physical_blk = NandStat.free->physical_blk;
                if( new_physical_blk >= 1024) {
                    PRINTDEBUG( "%s : fatal error! illegal block number.\n", __FUNCTION__);
                    return( I_NAND_ERR_FATAL_ERR); //sԍ
                }
                /*-----*/
                /*newubN̏񐔓ǂݏo(TODO:peB`FbN)*/
                new_physical_page_addr = (u32)(new_physical_blk << 6);
                i_nandReadPage( &NandPage, new_physical_page_addr, 1);
                /*񐔂CNgWriteForwardobt@ɓKp,_AhXKp*/
                miCpuCopy8( &(NandPage.spare), &(NandPageCache->NandSpareWriteForward), NAND_PAGE_SPARESIZE);
/*                NandPageCache->NandSpareWriteForward.logical_adr0 = 0xFFFF;//(u16)(NandPageCache->logical_page >> 6);
                NandPageCache->NandSpareWriteForward.logical_adr1 = 0xFFFF;//(u16)(NandPageCache->logical_page >> 6);
                NandPageCache->NandSpareWriteForward.block_erase_count0 = NandPage.spare.block_erase_count0 + 1;
                NandPageCache->NandSpareWriteForward.block_erase_count1 = NandPage.spare.block_erase_count1 + 1;
*/
                /*newubN(TODO:eraseG[`FbN)*/
                if( i_nandEraseBlock( new_physical_blk) == FALSE) {
                    PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                    i_nandMarkingBadBlock( new_physical_blk, &(NandPageCache->NandSpareWriteForward));
                    i_nandRemoveStatList( &(NandStat.free), NandPageCache->TargetNandBlockStat);
                }else{
                    break;
                }
            }while( 1);

            /*FlushpubNƂnew_physical_blko^*/
            NandPageCache->OnTheFlush = new_physical_blk;

            for( i=0; i<64; i++) {
                NandPageCache->page_flushed[i] = I_NAND_FLUSH_FALSE;
            }
            NandPageCache->page_flushed_count = 0;
            
            /*merge֐mergepubNfreeXgT̂ŁAړĂ*/
            //TODO:ړɁAmerge֐ŒTƂɂ悯ׂ?
            i_nandRemoveStatList( &(NandStat.free), NandPageCache->TargetNandBlockStat);
            i_nandInsertStatList( &(NandStat.busy), NandPageCache->TargetNandBlockStat);

            /*-----*/
        }else{  /*freeubNȂꍇ͏ݕs\Ȃ̂ŃG[*/
            PRINTDEBUG( "%s : fatal error! cannot found free block.\n", __FUNCTION__);
            return( I_NAND_ERR_IMPOSSIBLE_ERR);
        }
    }else{ /* 2ڈȍ~Flush */
        new_physical_blk = NandPageCache->OnTheFlush;
        if( new_physical_blk >= 1024) {
            PRINTDEBUG( "%s : fatal error! illegal block number.\n", __FUNCTION__);
            return( I_NAND_ERR_FATAL_ERR); //sԍ
        }
    }
    /*-------------------------------------------------------*/

  PRINTDEBUG( "new_physical_blk : 0x%x, flushed_count : 0x%x\n", new_physical_blk,
              NandPageCache->page_flushed_count);

    /*newubN̊Yy[WPageCache*/
    for( i=0; i<NAND_CACHE_PAGES; i++) {
//        new_physical_page_addr = (u32)((new_physical_blk << 6) + ((NandPageCache->logical_page) & 0x3F));
        //擪y[W珇ɏĂi_y[W݋֎~̂߁j
        new_physical_page_addr = (u32)((new_physical_blk << 6) + NandPageCache->page_flushed_count);
        /*LbṼXyÄnewubN̏񐔂Ƙ_AhXKp*/
        miCpuCopy8( &(NandPageCache->NandSpareWriteForward), &(NandPageCache->NandPage[i].spare), NAND_PAGE_SPARESIZE);
        //TODO:i_nandWritePagẽG[R[h
        rslt_writepage = i_nandWritePage( &(NandPageCache->NandPage[i]), new_physical_page_addr, 1);

        /*FlushedtO*/
        if( NandPageCache->page_flushed[((NandPageCache->logical_page) & 0x3F)] != I_NAND_FLUSH_FALSE) {
            PRINTDEBUG( "%s : kabutta!\n", __FUNCTION__);
            while( 1) {};
        }
        NandPageCache->page_flushed[((NandPageCache->logical_page) & 0x3F)] = NandPageCache->page_flushed_count;
        NandPageCache->page_flushed_count++;
    }


    /*ȈՃxt@C(FLUSHŃobhubNȂ`FbN)*/
    for( i=0; i<NAND_CACHE_PAGES; i++) {
//        new_physical_page_addr = (u32)((new_physical_blk << 6) + ((NandPageCache->logical_page) & 0x3F));
        i_nandReadPage( &NandPage, new_physical_page_addr, 1);

        for( k=0; k<4; k++) {
            if( (i_nandCheckECC( (u32*)(&(NandPage.data.data[k][0])), &(NandPage.spare.ecc[k]),
                                *(u32*)((u32)(&(nand_ecc[0]))+(k*4))) == FALSE) ||
                ( rslt_writepage == FALSE)) {
                //(tbVłȂubNobhubN̂merge֐)
                //ł̓}[LÔ
//                i_nandRemoveStatList( &(NandStat.busy), NandPageCache->TargetNandBlockStat); //BadBlockfreeXgO
                NandPageCache->NandSpareWriteForward.block_stat_bad = 0;
                NandPageCache->TargetNandBlockStat->block_stat_bad = 0;
//                NandPageCache->OnTheFlush = 0xFFFF;
//                NandPageCache->BlockDirty = FALSE;
                NandPageCache->page_flushed[((NandPageCache->logical_page) & 0x3F)] = I_NAND_FLUSH_CASHED;
                PRINTLEVEL2( "%s : FLUSH failed!(bad block generated)\n", __FUNCTION__);
                //FlushsɂnewubNɃCgȂ̂ŁABlockDirtyTRUEɂȂ
                return( I_NAND_ERR_VERIFY_ERR);
            }
        }
    }

    /*PageCache𖳌*/
    NandPageCache->valid = FALSE;
    NandPageCache->dirty = FALSE;

    NandPageCache->BlockDirty = TRUE;
    
//    PRINTDEBUG( "old : 0x%x,   new : 0x%x\n", old_physical_blk, new_physical_blk);
    return( I_NAND_ERR_SUCCESS);
}


/*---------------------------------------------------------------------------*
  Name:         i_nandMergeBlock

  Description:  

  Arguments:    

  Returns:      None
 *---------------------------------------------------------------------------*/
BOOL i_nandMergeBlock( NandPageCacheFormat* NandPageCache)
{
    u16                i, k;
    u16                retry_count = 0;
    u16                logical_blk;
    u16                old_physical_blk;
    u32                old_physical_page_addr;
    u16                new_physical_blk;
    u32                new_physical_page_addr;
    u16                merge_physical_blk;
    u32                merge_physical_page_addr;
    NandPageFormat     NandPage;
    NandBlockStatInfo* MergeNandBlockStat;
    u32                bad_physical_page_addr;
    u16                bad_physical_blk[NAND_FLUSH_RETRY_COUNT];
    BOOL               rslt_writepage;
    
    /*NewBlockECC`FbNFLUSHŒsς*/
    
    PRINTDEBUG( "%s\n", __FUNCTION__);

    if( NandPageCache->OnTheFlush == 0xFFFF) {
        return( FALSE);
    }

    do {
        /*----- MergepNewBlockT[` -----*/
        do {
            MergeNandBlockStat = NandStat.free;
            do {
                if( MergeNandBlockStat == NULL) { /*freeubNȂꍇ*/
                    PRINTDEBUG( "%s : fatal error! cannot found free block.\n", __FUNCTION__);
                    while( 1) {};  //return( FALSE);
                }else{
                    break; /**/
                }
    //            MergeNandBlockStat = MergeNandBlockStat->next;
            }while( 1);
        
            merge_physical_blk = MergeNandBlockStat->physical_blk;
            if( merge_physical_blk >= 1024) {
                PRINTDEBUG( "%s : fatal error! illegal block number.\n", __FUNCTION__);
                while( 1) {};  //return( FALSE); //sԍ
            }
            /*mergeubN̏񐔓ǂݏo(TODO:peB`FbN)*/
            merge_physical_page_addr = (u32)(merge_physical_blk << 6);
            i_nandReadPage( &NandPage, merge_physical_page_addr,1 );
            /*񐔂CNgWriteMergeobt@ɓKp,_AhXKp*/
            miCpuCopy8( &(NandPage.spare), &(NandPageCache->NandSpareWriteMerge), NAND_PAGE_SPARESIZE);
            NandPageCache->NandSpareWriteMerge.logical_adr0 = (u16)(NandPageCache->logical_page >> 6);
            NandPageCache->NandSpareWriteMerge.logical_adr1 = (u16)(NandPageCache->logical_page >> 6);
            NandPageCache->NandSpareWriteMerge.block_erase_count0 = NandPage.spare.block_erase_count0 + 1;
            NandPageCache->NandSpareWriteMerge.block_erase_count1 = NandPage.spare.block_erase_count1 + 1;

            /*mergeubN*/
            if( i_nandEraseBlock( merge_physical_blk) == FALSE) {
                PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                i_nandMarkingBadBlock( merge_physical_blk, &(NandPageCache->NandSpareWriteMerge));
                i_nandRemoveStatList( &(NandStat.free), MergeNandBlockStat);
            }else{
                break;
            }
        }while( 1);

    
        /*oldubNZo(newubNƂ͕KقȂĂ)*/
        logical_blk = (u16)((NandPageCache->logical_page) >> 6);
        old_physical_blk = nandBlkTbl[logical_blk];
    
        for( i=0; i<64; i++) {
            merge_physical_page_addr = (u32)((merge_physical_blk << 6) + i);
            switch( NandPageCache->page_flushed[i]) {
              case I_NAND_FLUSH_FALSE:/*OldBlock}[Wf[^MergeBlockɃRs[*/
                old_physical_page_addr = (u32)((old_physical_blk<<6) + i);
                i_nandReadPage( &NandPage, old_physical_page_addr, 1);
                /*sparë(,_AhX)ύX*/
                miCpuCopy8( &(NandPageCache->NandSpareWriteMerge), &(NandPage.spare), NAND_PAGE_SPARESIZE);
                //TODO:i_nandWritePagẽG[R[h
                rslt_writepage = i_nandWritePage( &NandPage, merge_physical_page_addr, 1);
                break;
              case I_NAND_FLUSH_CASHED:/*LbV}[Wf[^MergeBlockɃRs[*/
                /*sparë(,_AhX)ύX*/
                miCpuCopy8( &(NandPageCache->NandSpareWriteMerge), &(NandPageCache->NandPage[0].spare),
                            NAND_PAGE_SPARESIZE);
                //TODO:i_nandWritePagẽG[R[h
                rslt_writepage = i_nandWritePage( &(NandPageCache->NandPage[0]), merge_physical_page_addr, 1);
                break;
              default:/*NewBlock}[Wf[^MergeBlockɃRs[*/
                if( NandPageCache->page_flushed[i] >= 0) {
                    new_physical_page_addr = (u32)(((NandPageCache->OnTheFlush)<<6) +
                                                   NandPageCache->page_flushed[i]);
                    i_nandReadPage( &NandPage, new_physical_page_addr, 1);
                    /*sparë(,_AhX)ύX*/
                    miCpuCopy8( &(NandPageCache->NandSpareWriteMerge), &(NandPage.spare), NAND_PAGE_SPARESIZE);
                    //TODO:i_nandWritePagẽG[R[h
                    rslt_writepage = i_nandWritePage( &NandPage, merge_physical_page_addr, 1);
                    break;
                }else{
                    PRINTDEBUG( "%s : fatal error! invalid status(0x%x).\n", NandPageCache->page_flushed[i]);
                    while( 1) {};
                }
            }
            if( rslt_writepage == FALSE) {
                PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                i_nandRemoveStatList( &(NandStat.free), MergeNandBlockStat); //BadBlockfreeXgO
                bad_physical_blk[retry_count] = merge_physical_blk;
                goto retry;
            }
        }
        
        /* mergeretryƂēxfreeXgT̂ŁAbusyɈړĂ*/
        i_nandRemoveStatList( &(NandStat.free), MergeNandBlockStat);
        i_nandInsertStatList( &(NandStat.busy), MergeNandBlockStat);
        
        /*ȈՃxt@C(mergeŃobhubNȂ`FbN)*/
            for( i=0; i<64; i++) {
                merge_physical_page_addr = (u32)((merge_physical_blk << 6) + i);
                i_nandReadPage( &NandPage, merge_physical_page_addr, 1);
                for( k=0; k<4; k++) {
                    if( i_nandCheckECC( (u32*)(&(NandPage.data.data[k][0])), &(NandPage.spare.ecc[k]),
                                        *(u32*)((u32)(&(nand_ecc[0]))+(k*4))) == FALSE) {
                        PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
                        i_nandRemoveStatList( &(NandStat.busy), MergeNandBlockStat); //BadBlockfreeXgO
                        bad_physical_blk[retry_count] = merge_physical_blk;
                        goto retry;
                    }
                }
            }
        break; /* Merge */
retry:
        retry_count++;
        if( retry_count >= NAND_FLUSH_RETRY_COUNT) {
            break;
        }
        PRINTDEBUG( "%s : retry merge!( blk 0x%x is bad.)\n", __FUNCTION__,
                    merge_physical_blk);
    }while( 1);

    /*PageCache𖳌*/
    NandPageCache->valid = FALSE;
    NandPageCache->dirty = FALSE;
    
    //TODO:newblockAubN߂
      //(TODO:e[uNANDɎ悤ɂ΂͕̏KvȂ)
    
    /*newubN*/
    new_physical_blk = NandPageCache->OnTheFlush;
    if( i_nandEraseBlock( new_physical_blk) == FALSE) {
        PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
        NandPageCache->TargetNandBlockStat->block_stat_bad = 0;
        NandPageCache->NandSpareWriteForward.block_stat_bad = 0;
    }
    /*newubNɏXyÄ\z*/
    NandPageCache->NandSpareWriteForward.logical_adr0 = 0xFFFF;   //gp}[LO
    NandPageCache->NandSpareWriteForward.logical_adr1 = 0xFFFF;
    NandPageCache->NandSpareWriteForward.block_erase_count0++;
    NandPageCache->NandSpareWriteForward.block_erase_count1++;
    /*newubNɏf[^NandPageɍ쐬*/
    miCpuFill8( &(NandPage.data), 0xFF, NAND_PAGE_DATASIZE);
    miCpuCopy8( &(NandPageCache->NandSpareWriteForward), &(NandPage.spare), NAND_PAGE_SPARESIZE);
    /*newubNɏ*/
    for( i=0; i<64; i++) {
        new_physical_page_addr = (u32)((new_physical_blk<<6) + i);
        //TODO:i_nandWritePagẽG[R[h
        if( i_nandWritePage( &NandPage, new_physical_page_addr, 1) == FALSE) {
            PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
            i_nandMarkingBadBlock( new_physical_blk, &(NandPage.spare));
            break;
        }
    }
    /*-----------------------------------------------------*/
    
    
    /*----- OldBlockAubN߂ -----*/
      //(TODO:e[uNANDɎ悤ɂ΂͕̏KvȂ)
    
    /*oldubN*/
    if( i_nandEraseBlock( old_physical_blk) == FALSE) {
        PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
        NandBlockStat[old_physical_blk].block_stat_bad = 0;
        NandPageCache->NandSpareWriteBack.block_stat_bad = 0;
    }
    /*oldubNɏXyÄ\z*/
    NandPageCache->NandSpareWriteBack.logical_adr0 = 0xFFFF;   //gp}[LO
    NandPageCache->NandSpareWriteBack.logical_adr1 = 0xFFFF;
    NandPageCache->NandSpareWriteBack.block_erase_count0++;
    NandPageCache->NandSpareWriteBack.block_erase_count1++;
    /*oldubNɏf[^NandPageɍ쐬*/
    miCpuFill8( &(NandPage.data), 0xFF, NAND_PAGE_DATASIZE);
    miCpuCopy8( &(NandPageCache->NandSpareWriteBack), &(NandPage.spare), NAND_PAGE_SPARESIZE);
    /*oldubNɏ*/
    for( i=0; i<64; i++) {
        old_physical_page_addr = (u32)((old_physical_blk<<6) + i);
        //TODO:i_nandWritePagẽG[R[h
        if( i_nandWritePage( &NandPage, old_physical_page_addr, 1) == FALSE) {
            PRINTLEVEL2( "bad block generated : %s, %d\n", __FUNCTION__, __LINE__);
            i_nandMarkingBadBlock( old_physical_page_addr, &(NandPage.spare));
            break;
        }
    }
    /*-----------------------------------------------------*/
    
    /*tbVłȂmergeubNCɃobhubN*/
    for( i=0; i<retry_count; i++) {
        i_nandReadPage( &NandPage, bad_physical_blk[i]<<6, 1);
        /**/
        miCpuFill8( &(NandPage.data), 0xFF, NAND_PAGE_DATASIZE);
        
        NandPage.spare.logical_adr0 = 0xFFFF;  //gp}[LO
        NandPage.spare.logical_adr1 = 0xFFFF;
        NandPage.spare.block_stat_bad = 0;
        NandBlockStat[bad_physical_blk[i]].block_stat_bad = 0;
        /*ubN*/
        i_nandEraseBlock( bad_physical_blk[i]);
        /*badubNɏ*/
        for( k=0; k<64; k++) {
            bad_physical_page_addr = (u32)((bad_physical_blk[i] << 6) + k);
            i_nandWritePage( &NandPage, bad_physical_page_addr, 1);
        }
    }
    /*-----------------------------------------------------------*/

    
    /*----- RAMnandBlkTblXV -----*/
    nandBlkTbl[logical_blk] = merge_physical_blk;
    
    PRINTDEBUG( "%s : logical_blk:0x%x ... old_physical_blk:0x%x, new_physical_blk:0x%x\n",
                __FUNCTION__, logical_blk, old_physical_blk, NandPageCache->OnTheFlush);
    PRINTDEBUG( "merge_physical_blk:0x%x\n", merge_physical_blk);
    /*---------------------------------*/

    {
        //u16 reload_logical_adr;
        //u32 reload_erase_count;
        
    /*BlockStatNXgXV*/
    /*new_physical_blk͏񐔂XVꂽ̂ŁAbusyXgOfreeXgɓ꒼*/
    //memo:NandPageCache->TargetNandBlockStatNandBlockStat[NandPageCache->OnTheFlush]łǂ͂
    i_nandRemoveStatList( &(NandStat.busy), NandPageCache->TargetNandBlockStat);
    NandPageCache->TargetNandBlockStat->erase_count++;
    if( NandPageCache->TargetNandBlockStat->block_stat_bad == 0xFF) { //BadBlockłȂ
        i_nandInsertStatList( &(NandStat.free), NandPageCache->TargetNandBlockStat);
    }

    /*merge_physical_blkbusyXgObusyXgɓ꒼i\[gj*/
    i_nandRemoveStatList( &(NandStat.busy), MergeNandBlockStat);
        MergeNandBlockStat->erase_count++; //NXg̏񐔂XV
    i_nandInsertStatList( &(NandStat.busy), MergeNandBlockStat);

    /*old_physical_blkbusyXgOfreeXgɓ*/
    //i_nandCheckBlock( old_physical_blk, &reload_logical_adr, &reload_erase_count);
    i_nandRemoveStatList( &(NandStat.busy), &(NandBlockStat[old_physical_blk]));
    NandBlockStat[old_physical_blk].erase_count++;
    if( NandBlockStat[old_physical_blk].block_stat_bad == 0xFF) { //BadBlockłȂ
        i_nandInsertStatList( &(NandStat.free), &(NandBlockStat[old_physical_blk]));
    }
    }

    /**/
    for( i=0; i<64; i++) {
        NandPageCache->page_flushed[i] = I_NAND_FLUSH_FALSE;
    }
    NandPageCache->page_flushed_count = 0;

    
    NandPageCache->OnTheFlush = 0xFFFF;
    NandPageCache->BlockDirty = FALSE;

    return( TRUE);
}


/*㔭BadBlockɃ}[LOigpATODO:eĂяoj*/
void i_nandMarkingBadBlock( u16 physical_blk, NandSpareFormat* NandSpareOrg)
{
    u16            i;
    u32            physical_page_addr;
    NandPageFormat NandPage;

    miCpuFill8( &(NandPage.data.data[0][0]), 0xFF, NAND_PAGE_DATASIZE);
    miCpuFill8( &(NandPage.spare), 0xFF, NAND_PAGE_SPARESIZE);
    miCpuCopy8( "CTR\0", &(NandPage.spare.sig0[0]), 4);
    miCpuCopy8( "CTR\0", &(NandPage.spare.sig1[0]), 4);
    
    NandPage.spare.block_stat_bad = 0x00; /* Bad Marking */
    NandPage.spare.block_erase_count0 = NandSpareOrg->block_erase_count0;
    NandPage.spare.block_erase_count1 = NandSpareOrg->block_erase_count1;
    NandPage.spare.logical_adr0 = 0xFFFF; /* 蓖 */
    NandPage.spare.logical_adr1 = 0xFFFF; /* 蓖 */
    
    i_nandEraseBlock( physical_blk);
    for( i=0; i<64; i++) {
        physical_page_addr = ((u32)(physical_blk)<<6) + i;
        i_nandWritePage( &NandPage, physical_page_addr, 1);
    }
}


/*---------------------------------------------------------------------------*
  Name:         nandReadPageCache

  Description:  PageCacheZN^ǂݏo

  Arguments:    

  Returns:      None
 *---------------------------------------------------------------------------*/
static void nandReadPageCache( NandPageCacheFormat* NandPageCache, u32* dest, u16 cache_sect)
{
    u16     cache_offset;
    u32*    cache_sect_adr;

    cache_offset = (u16)(cache_sect % 4);   //PageCachẽy[Wł̃ItZbg

    cache_sect_adr = (u32*)((u32)&(NandPageCache->NandPage[0].data.data[cache_offset][0]));
    
    miCpuCopy8( cache_sect_adr, dest, 512);
}

/*---------------------------------------------------------------------------*
  Name:         nandWritePageCache

  Description:  PageCacheɃZN^

  Arguments:    physical_blk : ubNԍ

  Returns:      None
 *---------------------------------------------------------------------------*/
static void nandWritePageCache( NandPageCacheFormat* NandPageCache, u32* src, u16 cache_sect)
{
//    u16     cache_page;
    u16     cache_offset;
    u32*    cache_sect_adr;

//    cache_page = (u16)(cache_sect / 4); //PageCacheł̃y[Wԍ
    cache_offset = (u16)(cache_sect % 4);   //PageCachẽy[Wł̃ItZbg

    cache_sect_adr = (u32*)((u32)&(NandPageCache->NandPage[0].data.data[cache_offset][0]));
    
    miCpuCopy8( src, cache_sect_adr, 512);

    //ECCČvZ
}





/*---------------------------------------------------------------------------*
  Name:         nandEnable

  Description:

  Arguments:

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void nandEnable( NandPageCacheFormat* NandPageCache)
{
    NandMsg nandMsg;
    u32     recv_dat;

    g_nandPageCache = NandPageCache;
//    i_nandEnable( NandPageCache);
    nandMsg.cache           = NandPageCache;
    nandMsg.buf             = NULL;
    nandMsg.sector          = 0;
    nandMsg.sector_count    = 0;
    nandMsg.operation       = NAND_OPERATION_ENABLE;

//    PRINTDEBUG( "%s(%d) : snd_dtq begin\n", __FUNCTION__, __LINE__);
    osSendMessage( &nand_dtq, (OSMessage*)&nandMsg, OS_MESSAGE_BLOCK);

//    PRINTDEBUG( "%s(%d) : rcv_dtq begin\n", __FUNCTION__, __LINE__);
    osReceiveMessage( &nand_result_dtq, (OSMessage*)&recv_dat, OS_MESSAGE_BLOCK);

    return;
}

/*---------------------------------------------------------------------------*
  Name:         nandFormat

  Description:  tH[}bgiXyÄ\zj

  Arguments:

  Returns:
 *---------------------------------------------------------------------------*/
void nandFormat( void)
{
    NandMsg nandMsg;
    u32     recv_dat;

    nandMsg.cache           = g_nandPageCache;
    nandMsg.buf             = NULL;
    nandMsg.sector          = 0;
    nandMsg.sector_count    = 0;
    nandMsg.operation       = NAND_OPERATION_FORMAT;

//    PRINTDEBUG( "%s(%d) : snd_dtq begin\n", __FUNCTION__, __LINE__);
    osSendMessage( &nand_dtq, (OSMessage*)&nandMsg, OS_MESSAGE_BLOCK);

//    PRINTDEBUG( "%s(%d) : rcv_dtq begin\n", __FUNCTION__, __LINE__);
    osReceiveMessage( &nand_result_dtq, (OSMessage*)&recv_dat, OS_MESSAGE_BLOCK);

    return;
}

/*---------------------------------------------------------------------------*
  Name:         nandReadSector

  Description:  _ZN^ǂݏo

  Arguments:    logical_sector : _ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void nandReadSector( u32* dest, u32 logical_sector, u32 sector_count)
{
    NandMsg nandMsg;
    u32     recv_dat;

    nandMsg.cache           = g_nandPageCache;
    nandMsg.buf             = dest;
    nandMsg.sector          = logical_sector;
    nandMsg.sector_count    = sector_count;
    nandMsg.operation       = NAND_OPERATION_READ;

//    PRINTDEBUG( "%s(%d) : snd_dtq begin\n", __FUNCTION__, __LINE__);
    osSendMessage( &nand_dtq, (OSMessage*)&nandMsg, OS_MESSAGE_BLOCK);

//    PRINTDEBUG( "%s(%d) : rcv_dtq begin\n", __FUNCTION__, __LINE__);
    osReceiveMessage( &nand_result_dtq, (OSMessage*)&recv_dat, OS_MESSAGE_BLOCK);

    return;
}

/*---------------------------------------------------------------------------*
  Name:         nandWriteSector

  Description:  _ZN^

  Arguments:    logical_sector : _ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
void nandWriteSector( u32* src, u32 logical_sector, u32 sector_count)
{
    NandMsg nandMsg;
    u32     recv_dat;
//    ID      tskid;

    nandMsg.cache           = g_nandPageCache;
    nandMsg.buf             = src;
    nandMsg.sector          = logical_sector;
    nandMsg.sector_count    = sector_count;
    nandMsg.operation       = NAND_OPERATION_WRITE;
  /*
    if( get_tid( &tskid) == E_OK) {
        nandMsg.tskid       = tskid;
    }else{
        nandMsg.tskid       = 0;
    }*/

//    PRINTDEBUG( "%s(%d) : snd_dtq begin\n", __FUNCTION__, __LINE__);
    osSendMessage( &nand_dtq, (OSMessage*)&nandMsg, OS_MESSAGE_BLOCK);

//    PRINTDEBUG( "%s(%d) : rcv_dtq begin\n", __FUNCTION__, __LINE__);
    osReceiveMessage( &nand_result_dtq, (OSMessage*)&recv_dat, OS_MESSAGE_BLOCK);

    return;
}

/*---------------------------------------------------------------------------*
  Name:         nandFlush

  Description:

  Arguments:

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
BOOL nandFlush( void)
{
    NandMsg nandMsg;
    u32     recv_dat;

    nandMsg.cache           = g_nandPageCache;
    nandMsg.buf             = NULL;
    nandMsg.sector          = 0;
    nandMsg.sector_count    = 0;
    nandMsg.operation       = NAND_OPERATION_FLUSH;

//    PRINTDEBUG( "%s(%d) : snd_dtq begin\n", __FUNCTION__, __LINE__);
    osSendMessage( &nand_dtq, (OSMessage*)&nandMsg, OS_MESSAGE_BLOCK);

//    PRINTDEBUG( "%s(%d) : rcv_dtq begin\n", __FUNCTION__, __LINE__);
    osReceiveMessage( &nand_result_dtq, (OSMessage*)&recv_dat, OS_MESSAGE_BLOCK);
    if( recv_dat == 1) {
        return( TRUE);
    }else{
        return( FALSE);
    }
}

/*^XNĂȂ̂ŉB֐*/
void nandEraseChip( void)
{
    u16 i;
    for( i=0; i<1024; i++) {
        i_nandEraseBlock( i);
    }
}

/*---------------------------------------------------------------------------*
  Name:         i_nandThread

  Description:

  Arguments:

  Returns:      None
 *---------------------------------------------------------------------------*/
static void i_nandThread( void* arg)
{
    NandMsg*        nandMsg;
    BOOL            result;
    u32             current_dat;
    u32             i;
    u32*            buf_pointer;

    result = TRUE;
    while( TRUE) {
        osReceiveMessage( &nand_dtq, (OSMessage*)&current_dat, OS_MESSAGE_BLOCK);
        nandMsg = (NandMsg*)current_dat;
        PRINTDEBUG( "nand task : receive command : %d\n", nandMsg->operation);

        switch( nandMsg->operation) {
          case NAND_OPERATION_ENABLE:
            i_nandEnable( nandMsg->cache);
            break;
          case NAND_OPERATION_FORMAT:
            i_nandFormat();
            break;
          case NAND_OPERATION_READ:
            PRINTDEBUG( "from:%x, sectors:%x\n", nandMsg->sector, nandMsg->sector_count);
            for( i=0; i<nandMsg->sector_count; i++) {
                buf_pointer = (u32*)((u32)(nandMsg->buf) + (512 * i));
                i_nandReadSector( nandMsg->cache, buf_pointer, (nandMsg->sector) + i);
            }
            break;
          case NAND_OPERATION_WRITE:
            PRINTDEBUG( "from:%x, sectors:%x\n", nandMsg->sector, nandMsg->sector_count);
            /*---------- ʏ̏ ----------*/
            for( i=0; i<nandMsg->sector_count; i++) {
                buf_pointer = (u32*)((u32)(nandMsg->buf) + (512 * i));
                i_nandWriteSector( nandMsg->cache, buf_pointer, (nandMsg->sector) + i);
            }/*-----------------------------------*/
            break;
          case NAND_OPERATION_FLUSH:
            result = i_nandFlush( nandMsg->cache);
            break;
          default:
            break;
        }

        PRINTDEBUG( "nand task : operation ends, send message begin\n");
        //bZ[Wԑ
        if( result) {
            current_dat = 1;
        }else{
            current_dat = 0;
        }
        osSendMessage( &nand_result_dtq, (OSMessage)current_dat, OS_MESSAGE_BLOCK);
    }
}

