/*---------------------------------------------------------------------------*
  Project:  CTR - NAND driver
  File:     nand.h

  Copyright 2006 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#ifndef    __NAND_H__
#define    __NAND_H__


#include "nandif_ip.h"


#ifdef __cplusplus
extern "C" {
#endif


/*---------------------------------------------------------------------------*
    `
 *---------------------------------------------------------------------------*/
#define NAND_GUARANTEED_BLOCKS  (999)
#define NAND_GUARANTEED_SECTORS (NAND_GUARANTEED_BLOCKS*256)    //ۏ؂ꂽZN^(=ge)

#define NAND_CACHE_PAGES        (1) /* 1ȊO͐ݒ֎~ */
#define NAND_ANTI_READDISTURB   (1) /* TODO */

/*---------------------------------------------------------------------------*
    \
 *---------------------------------------------------------------------------*/
typedef struct NandPageCacheFormat {            //ubNLbV
    u16                       valid;               /*L or */
    u16                       dirty;               /*_[eB or N[*/
    u16                       OnTheFlush;          /*FlushisNewPhysicalBlkԍ*/
    u16                       reserve0;
    struct NandBlockStatInfo* TargetNandBlockStat; /*FlushisNewPhysicalBlkւ̃|C^*/
    BOOL                      BlockDirty;          /*ubNE_[eB or N[*/
//    u16                      logical_blk;        /*_ubNԍ*/
    u32                       logical_page;        /*_y[Wԍ*/
    u8                        page_flushed_count;  /*Flush̕y[WItZbg(0JEgAbv)*/
    u8                        reserve1[3];
    int                       page_flushed[64];    /*Flushɂď܂ꂽy[W̃ItZbg*/
    struct NandPageFormat     NandPage[NAND_CACHE_PAGES]; /*1y[W̃LbV*/
    u32                       ecc[NAND_CACHE_PAGES][4];   /*ǂݏoɎꂽECC*/
    struct NandSpareFormat    NandSpareWriteForward;      /*newubNɃCgXyÄ*/
    struct NandSpareFormat    NandSpareWriteBack;  /*oldubNɃCgobNXyÄ*/
    struct NandSpareFormat    NandSpareWriteMerge; /*mergeubNɃCgXyÄ*/
} NandPageCacheFormat;


typedef struct NandBlockStatInfo {                //ubNԃNXg
    u16                            physical_blk;
    u16                            block_stat_bad;
    u32                            erase_count;
    struct NandBlockStatInfo*    next;
} NandBlockStatInfo;


typedef struct NandStatInfo {
    struct NandBlockStatInfo*    busy;    /*_blkԍ蓖čς*/
    struct NandBlockStatInfo*    free;    /*_blkԍ蓖*/
} NandStatInfo;



/*RTFSp FATp[^*/
typedef struct {
    u32    device_capacity;          //foCXŜ̃TCY(512ByteP)
    u32    adjusted_device_capacity;

    u32    memory_capacity;          //data areãTCY(512ByteP)
    u32    adjusted_memory_capacity; //memory_capacityV_(heads*secptrack)̔{ɒTCY(cylinders*heads*secptrackɂȂ)
    u16    volume_cylinders;
    
    u16    heads;
    u16    secptrack;
    u16    cylinders;
    u16    SC;                       //sectors per cluster
    u16    BU;
    u16    RDE;                      //number of root dir entries(512 fix)
    u32    SS;                       //sector size(512 fix)
    u32    RSC;                      //reserved sector count(1 fix)
    u16    FATBITS;                  //16 or 32
    u16    SF;                       //sectors per FAT
    u32    SSA;                      //sectors in system area
    u32    NOM;                      //sectors in master boot record

    u32    begin_sect;
} NandSpec;



/*---------------------------------------------------------------------------*
    API
 *---------------------------------------------------------------------------*/
BOOL nandRtfsAttach( int driveno, int partition);

void nandInit( void);
void nandEnable( NandPageCacheFormat* NandPageCache);
void nandFormat( void);

void nandReadSector( u32* dest, u32 logical_sector, u32 sector_count);
void nandWriteSector( u32* src, u32 logical_sector, u32 sector_count);

BOOL nandFlush( void);
BOOL nandCheckMedia( void);


#ifdef __cplusplus
}    /* extern "C" */
#endif


#endif    /*__NAND_H__*/
