/*---------------------------------------------------------------------------*
  Project:  CTR - NAND driver
  File:     nandif.c

  Copyright 2006 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#include <brom.h>
#include "nandif_ip.h"
#include "nandif_reg.h"

#if 0
#define PRINTDEBUG  osTPrintf
#else
#define PRINTDEBUG( ...) ((void)0)
#endif


/*---------------------------------------------------------------------------*
    static֐
 *---------------------------------------------------------------------------*/
static void i_nandWait( void);
static void i_nandStart( u16 mode);
static void i_nandSetEccDirection( u32 direction);
static void i_nandChangeBuffer( void);

static u32 CalcECC( u32* buf);
static u32 CalcCP( u32* buf, u32 st, u32 cmp, u32 logic);
static u32 CalcLP( u32* buf, u32 st, u32 cmp, u32 logic);

u32 nand_ecc[4];


/*NAND჌xw*/
/*---------------------------------------------------------------------------*
  Name:         i_nandWait

  Description:  Iy[VI܂ő҂

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
static void i_nandWait( void)
{
    u32 stat;
    
    while( 1) {
        stat = *(vu16*)REG_NAND_COMMAND;
        if( (stat & NAND_BUSY) == 0) {
            return;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         i_nandStart

  Description:  w肵[hX^[g

  Arguments:    mode : 

  Returns:      None
 *---------------------------------------------------------------------------*/
static void i_nandStart( u16 mode)
{
    i_nandWait();
    *(vu32*)REG_NAND_COMMAND = (mode | NAND_START);
}

/*---------------------------------------------------------------------------*
  Name:         i_nandSetEccDirection

  Description:  ECCZow肷

  Arguments:    direction : NAND_CTRL_ECC_WRITE or NAND_CTRL_ECC_READ

  Returns:      None
 *---------------------------------------------------------------------------*/
static void i_nandSetEccDirection( u32 direction)
{
}

/*---------------------------------------------------------------------------*
  Name:         i_nandChangeBuffer

  Description:  obt@̗\؂ւ

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
static void i_nandChangeBuffer( void)
{
}

/*---------------------------------------------------------------------------*
  Name:         i_nandReset

  Description:  HW-IPԂɂ

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/
void i_nandReset( void)
{
    /*--- HW ECC OFF ---*/
    *(vu16*)REG_NAND_ECC &= ~NAND_ENABLE_ECC;
  
    i_nandWait();
    *(u16*)REG_NAND_COMMAND = (NAND_RESET | NAND_START);
    i_nandWait();
}

/*---------------------------------------------------------------------------*
  Name:         i_nandEraseBlock

  Description:  ubN

  Arguments:    physical_blk : ubNԍ

  Returns:      FALSE : s
                TRUE : 
 *---------------------------------------------------------------------------*/
BOOL i_nandEraseBlock( u16 physical_blk)
{
//    PRINTDEBUG( ">>>@ erase physical blk : 0x%x\n", physical_blk);
    
    i_nandWait();        /*foCX҂*/
    
    /*C[X邽߂̃WX^ݒ*/
    *(u16*)REG_NAND_ADR_LO = 0;
    *(u16*)REG_NAND_ADR_HI = (physical_blk << 1);
    
    /*݊Jn*/
    *(u16*)REG_NAND_COMMAND = (NAND_ERASE_BLOCK | NAND_START);

//    osSleep( 1);        /* 90nm(typ:2ms, max:3ms),60nm1.2ms */
    
    i_nandWait();       /*foCX쒆*/
#if 0
    /*READYԂɂȂPASSFAIL𔻒肷*/
    while( 1) {
        /*Xe[^X[h*/
        *(vu32*)NAND_IP_CTRL = NAND_CTRL_READSTAT | NAND_CTRL_ECC_WRITE |
                                (*(vu32*)NAND_IP_CTRL & NAND_CTRL_PAGE_MASK) | NAND_CTRL_SPEED_HI;
        *(u32*)NAND_IP_CTRL |= NAND_CTRL_START;
        /*Xe[^X[h*/
        i_nandWait();
        /*Xe[^XWX^`FbN*/
        if( *(vu16*)NAND_IP_STAT & NAND_STAT_READY) {
            if( *(vu16*)NAND_IP_STAT & NAND_STAT_FAIL) {
                return FALSE;
            }else{
                return TRUE;
            }
        }
    }
#else
    return TRUE;
#endif
}

/*---------------------------------------------------------------------------*
  Name:         i_nandReadPage

  Description:  y[W[h

  Arguments:    physical_page_addr : y[Wԍ

  Returns:      
 *---------------------------------------------------------------------------*/
BOOL i_nandReadPage( NandPageFormat* dest, u32 physical_page_addr, BOOL sync)
{
    u32 byte_addr;
    PRINTDEBUG( ">>>@ read physical page : 0x%x\n", physical_page_addr);
    
    /*IP쒆*/
    i_nandWait();

    /*[h邽߂̃WX^ݒ*/
    byte_addr = (physical_page_addr * NAND_PAGE_DATASIZE);
    *(u16*)REG_NAND_ADR_LO = (u16)(byte_addr);
    *(u16*)REG_NAND_ADR_HI = (u16)(((u32)byte_addr >> 16));

    /*ŏ1y[W[h*/
    *(u16*)REG_NAND_COMMAND = (NAND_READ_PAGE | NAND_START);

    /*IP쒆*/
    i_nandWait();

    if( sync) {
        /*IP쒆*/
        i_nandWait();
        //Ńy[Wobt@Ƀf[^
    }

    miCpuCopy16( (u16*)REG_NAND_DATA, &(dest->data.data[0][0]), NAND_PAGE_DATASIZE);
    miCpuCopy16( (u16*)REG_NAND_SPARE, (u32*)&(dest->spare), NAND_PAGE_SPARESIZE);

    /*---------- \tgEFAECC ----------*/
    {
        u32 q_addr = (u32)(&(dest->data.data[0][0]));
        nand_ecc[0] = CalcECC( (u32*)q_addr);
        nand_ecc[1] = CalcECC( (u32*)(q_addr + 512));
        nand_ecc[2] = CalcECC( (u32*)(q_addr + 1024));
        nand_ecc[3] = CalcECC( (u32*)(q_addr + 1536));
    }
    /*-------------------------------------*/

    return TRUE;
}


/*---------------------------------------------------------------------------*
  Name:         i_nandWritePage

  Description:  y[Wɏ
                \ߊYy[WC[XĂƁB

  Arguments:    physical_page_addr : y[Wԍ

  Returns:      
 *---------------------------------------------------------------------------*/
BOOL i_nandWritePage( NandPageFormat* src, u32 physical_page_addr, BOOL sync)
{
    u32 byte_addr;
    PRINTDEBUG( ">>>@ write physical page : 0x%x\n", physical_page_addr);
    
    /*IP쒆*/
    i_nandWait();

    /*DataSpare*/
    miCpuCopy16( &(src->data.data[0][0]), (u16*)REG_NAND_DATA, NAND_PAGE_DATASIZE); //2KB
    /*---------- \tgEFAECC ----------*/
    {
        u32 q_addr = (u32)(&(src->data.data[0][0]));
        nand_ecc[0] = CalcECC( (u32*)(q_addr));
        nand_ecc[1] = CalcECC( (u32*)(q_addr + 512));
        nand_ecc[2] = CalcECC( (u32*)(q_addr + 1024));
        nand_ecc[3] = CalcECC( (u32*)(q_addr + 1536));
    }
    /*-------------------------------------*/
    miCpuCopy16( &(nand_ecc[0]), &(src->spare.ecc[0]), 4);
    miCpuCopy16( &(nand_ecc[1]), &(src->spare.ecc[1]), 4);
    miCpuCopy16( &(nand_ecc[2]), &(src->spare.ecc[2]), 4);
    miCpuCopy16( &(nand_ecc[3]), &(src->spare.ecc[3]), 4);
    miCpuCopy16( &(src->spare), (u16*)REG_NAND_SPARE, NAND_PAGE_SPARESIZE);
    
    /*Cg邽߂̃WX^ݒ*/
    byte_addr = (physical_page_addr * NAND_PAGE_DATASIZE);
    *(u16*)REG_NAND_ADR_LO = (u16)(byte_addr);
    *(u16*)REG_NAND_ADR_HI = (u16)(((u32)byte_addr >> 16));
  
    *(u16*)REG_NAND_COMMAND = (NAND_PROGRAM_PAGE | NAND_START);

    if( sync) {
        /*IP쒆*/
        i_nandWait();
#if 0
        /*READYԂɂȂPASSFAIL𔻒肷*/
        while( 1) {
            /*Xe[^X[h*/
            *(vu32*)NAND_IP_CTRL = NAND_CTRL_READSTAT | NAND_CTRL_ECC_WRITE |
                                    (*(vu32*)NAND_IP_CTRL & NAND_CTRL_PAGE_MASK) | NAND_CTRL_SPEED_HI;
            *(u32*)NAND_IP_CTRL |= NAND_CTRL_START;

            /*IP쒆*/
            i_nandWait();
            /*Xe[^XWX^`FbN*/
            if( *(vu16*)NAND_IP_STAT & NAND_STAT_READY) {
                if( *(vu16*)NAND_IP_STAT & NAND_STAT_FAIL) {
                    return FALSE;
                }else{
                    return TRUE;
                }
            }
        }
#else
      return TRUE;
#endif
    }else{
        return TRUE;
    }
}


static u32 CalcECC( u32* buf)
{
    u32 cp[10];
    u32 lp[14];
    u32 cym;

    cp[0] = CalcCP( buf, 0, 0x1, 0);
    cp[1] = CalcCP( buf, 1, 0x1, 1);
    cp[2] = CalcCP( buf, 0, 0x2, 0);
    cp[3] = CalcCP( buf, 2, 0x2, 1);
    cp[4] = CalcCP( buf, 0, 0x4, 0);
    cp[5] = CalcCP( buf, 4, 0x4, 1);
    cp[6] = CalcCP( buf, 0, 0x8, 0);
    cp[7] = CalcCP( buf, 8, 0x8, 1);
    cp[8] = CalcCP( buf, 0x00, 0x10, 0);
    cp[9] = CalcCP( buf, 0x10, 0x10, 1);

    lp[0] = CalcLP( buf, 0, 0x1, 0);
    lp[1] = CalcLP( buf, 1, 0x1, 1);
    lp[2] = CalcLP( buf, 0, 0x2, 0);
    lp[3] = CalcLP( buf, 2, 0x2, 1);
    lp[4] = CalcLP( buf, 0, 0x4, 0);
    lp[5] = CalcLP( buf, 4, 0x4, 1);
    lp[6] = CalcLP( buf, 0, 0x8, 0);
    lp[7] = CalcLP( buf, 8, 0x8, 1);
    lp[8] = CalcLP( buf, 0x00, 0x10, 0);
    lp[9] = CalcLP( buf, 0x10, 0x10, 1);
    lp[10] = CalcLP( buf, 0x00, 0x20, 0);
    lp[11] = CalcLP( buf, 0x20, 0x20, 1);
    lp[12] = CalcLP( buf, 0x00, 0x40, 0);
    lp[13] = CalcLP( buf, 0x40, 0x40, 1);

    cym = ((lp[13] & 0x1) << 23) +
          ((lp[12] & 0x1) << 22) +
          ((lp[11] & 0x1) << 21) + 
          ((lp[10] & 0x1) << 20) + 
          ((lp[ 9] & 0x1) << 19) + 
          ((lp[ 8] & 0x1) << 18) + 
          ((lp[ 7] & 0x1) << 17) + 
          ((lp[ 6] & 0x1) << 16) + 
          ((lp[ 5] & 0x1) << 15) + 
          ((lp[ 4] & 0x1) << 14) + 
          ((lp[ 3] & 0x1) << 13) + 
          ((lp[ 2] & 0x1) << 12) + 
          ((lp[ 1] & 0x1) << 11) + 
          ((lp[ 0] & 0x1) << 10) + 
          ((cp[ 9] & 0x1) <<  9) + 
          ((cp[ 8] & 0x1) <<  8) + 
          ((cp[ 7] & 0x1) <<  7) + 
          ((cp[ 6] & 0x1) <<  6) + 
          ((cp[ 5] & 0x1) <<  5) + 
          ((cp[ 4] & 0x1) <<  4) + 
          ((cp[ 3] & 0x1) <<  3) + 
          ((cp[ 2] & 0x1) <<  2) + 
          ((cp[ 1] & 0x1) <<  1) + 
          ((cp[ 0] & 0x1) <<  0);
//    osTPrintf( "cym:0x%x\n", cym);
    return( cym);
}

static u32 CalcCP( u32* buf, u32 st, u32 cmp, u32 logic)
{
    u32 cp;
    u32 cmp0;
    u32 i, j;

    cp = 0;
    for( j=st; j<32; j++) {
        if( (j & cmp) == 0) {
            cmp0 = 0;
        }else{
            cmp0 = 1;
        }
        if( cmp0 == logic) {
            for( i=0; i<128; i++) {
                cp ^= ((buf[i] >> j) & 0x1);
            }
        }
    }
    return cp;
}

static u32 CalcLP( u32* buf, u32 st, u32 cmp, u32 logic)
{
    u32 lp;
    u32 cmp0;
    u32 i, j;

    lp = 0;
    for( i=st; i<128; i++) {
        if( (i & cmp) == 0) {
            cmp0 = 0;
        }else{
            cmp0 = 1;
        }
        if( cmp0 == logic) {
            for( j=0; j<32; j++) {
                lp ^= ((buf[i] >> j) & 0x1);
            }
        }
    }
    return lp;
}

