/*---------------------------------------------------------------------------*
  Project:  CTR - NAND driver
  File:     nandif_reg.h

  Copyright 2006 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.
 *---------------------------------------------------------------------------*/

#ifndef	__NAND_IP_REG_H__
#define	__NAND_IP_REG_H__




//REG_NAND_COMMAND
#define NAND_START        (0x0100)
#define NAND_BUSY         NAND_START
#define NAND_RESET        6
#define NAND_READ_STATUS  5
#define NAND_ERASE_BLOCK  4
#define NAND_PROGRAM_PAGE 3
#define NAND_READ_PAGE    1

//REG_NAND_ECC
#define NAND_ENABLE_ECC   1

#define REG_NAND_ADR_LO   (0x18019840)
#define REG_NAND_ADR_HI   (0x18019842)
#define REG_NAND_COMMAND  (0x1801984E)

#define REG_NAND_ECC      (0x18019850)

#define REG_NAND_DATA     (0x18019000)
#define REG_NAND_SPARE    (0x18019800)


#define NAND_IP_ECC_0		(0x0100) //_~[

#if 0
/*********************************************
 NAND IPWX^
    
 (R/W) : readable and writable
 (RO)  : read only
*********************************************/
#if (CTR_DEF_ENVIRONMENT_DSEMU == 1)
#define NAND_IP_BASE		(0x08020000)
#else
#define NAND_IP_BASE		(0x400D0000)	// IOP@ݒ
#endif


#define NAND_IP_CADDR		(NAND_IP_BASE + 0x0000)	//16bit
#define NAND_IP_PADDR		(NAND_IP_BASE + 0x0004)	//24bit + dummy8bit
#define NAND_IP_COMMAND		(NAND_IP_BASE + 0x0008)	//0,1,2,3
#define NAND_IP_CTRL		(NAND_IP_BASE + 0x000C)	//32bit
#define NAND_IP_ID			(NAND_IP_BASE + 0x0010)	//48bit
#define NAND_IP_ID_H		(NAND_IP_BASE + 0x0014)
#define NAND_IP_STAT		(NAND_IP_BASE + 0x0018)	//16bit
#define NAND_IP_ECC_0		(NAND_IP_BASE + 0x0100)
#define NAND_IP_ECC_1		(NAND_IP_BASE + 0x0104)
#define NAND_IP_ECC_2		(NAND_IP_BASE + 0x0108)
#define NAND_IP_ECC_3		(NAND_IP_BASE + 0x010C)
#define NAND_IP_SPARE		(NAND_IP_BASE + 0x0400)	//  64BYTE
#define NAND_IP_DATA		(NAND_IP_BASE + 0x0800)	//2048BYTE



/*R}hWX^ɓl̏ڍגl*/
#define NAND_COM_WRITE		(0x00701080)		//0x70:StatRead, 0x10:AutoProgram, 0x80:DataInput
#define NAND_COM_READ		(0x00003000)		//0x30:ReadStart, 0x00:ReadAdrInput
#define NAND_COM_ERASE		(0x0070D060)	    //0x70:StatRead, 0xD0:EraseStart, 0x60:EraseSetup
#define NAND_COM_READID     (0x00000090)

/*Rg[WX^ɓl̏ڍגl*/
#define NAND_CTRL_READSTAT		(0x0010)	//10000b
#define NAND_CTRL_READID		(0x0011)
#define NAND_CTRL_WRITE			(0x0012)
#define NAND_CTRL_ERASE			(0x0013)
#define NAND_CTRL_READ_SETUP	(0x0014)
#define NAND_CTRL_READ_2112		(0x000d)

#define NAND_CTRL_RDY_MASK		((u32)0x1<<15)

#define NAND_CTRL_ECC_WRITE		(0)
#define NAND_CTRL_ECC_READ		((u32)0x1<<24)

#define NAND_CTRL_PAGE_0		(0)
#define NAND_CTRL_PAGE_1		((u32)1<<25)
#define NAND_CTRL_PAGE_MASK		((u32)1<<25)
#define NAND_CTRL_PAGE_SHIFT	(25)

#define NAND_CTRL_W8MHZ			(0)
#define NAND_CTRL_W11MHZ		((u32)0x1<<26)
#define NAND_CTRL_W16MHZ		((u32)0x2<<26)

#define NAND_CTRL_R8MHZ			(0)
#define NAND_CTRL_R11MHZ		((u32)0x1<<28)

#define NAND_CTRL_START			((u32)0x1<<31)
#define NAND_CTRL_START_MASK	((u32)0x1<<31)


/**/
#define NAND_CTRL_SPEED_LO		(NAND_CTRL_W8MHZ | NAND_CTRL_R8MHZ)
#define NAND_CTRL_SPEED_MIDDLE	(NAND_CTRL_W11MHZ | NAND_CTRL_R11MHZ)
#define NAND_CTRL_SPEED_HI		(NAND_CTRL_W16MHZ | NAND_CTRL_R11MHZ)


/*Xe[^XWX^̏ڍגl*/
#define NAND_STAT_FAIL			(1)			//NAND_STAT_READYԂŗL
#define NAND_STAT_READY			(1<<6)		//60nmiΉ
#define NAND_STAT_NOPROTECT		(1<<7)


/*ID_HWX^ɓl̏ڍגl*/
#define NAND_ID_H_FIFTH_MASK	(0x8000)
#endif

#endif	/*__NAND_IP_REG_H__*/
