/*---------------------------------------------------------------------------*
  Project:  TwlBrom - MATH - include
  File:     math/math.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/
#ifndef BROM_MATH_MATH_H_
#define BROM_MATH_MATH_H_

#include <brom/misc.h>
#include <brom/types.h>

#ifdef __cplusplus
extern "C" {
#endif

//----------------------------------------------------------------------------
// Type definition
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Declaration of function
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Implementation of inline function
//----------------------------------------------------------------------------

/*---------------------------------------------------------------------------*
  Name:         MATH_ABS

  Description:  ΒlԂ}NB
                e͕]邽߁ApɒӁB
  
  Arguments:    a

  Returns:      a < 0 Ȃ -a, Ȃ a
 *---------------------------------------------------------------------------*/
#define MATH_ABS(a) ( ( (a) < 0 ) ? (-(a)) : (a) )

/*---------------------------------------------------------------------------*
  Name:         MATH_IAbs

  Description:  ΒlԂCC֐B
                CC֐ƂĎĂ̂ŁAp͂ȂB
  
  Arguments:    a

  Returns:      a < 0 Ȃ -a, Ȃ a
 *---------------------------------------------------------------------------*/
SDK_INLINE int MATH_IAbs(int a)
{
    return (a < 0) ? -a : a;
}


/*---------------------------------------------------------------------------*
  Name:         MATH_CLAMP

  Description:  lowhigh͈̔͂̒l擾}NB
                e͕]邽߁ApɒӁB
  
  Arguments:    x     
                low   ől
                hight ŏl
  
  Returns:      x < low Ȃ low,  x > high Ȃ high Ȃ x
 *---------------------------------------------------------------------------*/
#define MATH_CLAMP(x, low, high) ( ( (x) > (high) ) ? (high) : ( ( (x) < (low) ) ? (low) : (x) ) )


/*---------------------------------------------------------------------------*
  Name:         MATH_MIN

  Description:  2̈̂ǂ炩قԂB
                }NƂĎĂ̂ŁAsZq`Ă
                ^Ŏgp\Be͕]邽߁ApɒӁB

  Arguments:    a, b

  Returns:      a < b Ȃ a, Ȃ b
 *---------------------------------------------------------------------------*/
#define MATH_MIN(a,b) (((a) <= (b)) ? (a) : (b))


/*---------------------------------------------------------------------------*
  Name:         MATH_IMin

  Description:  2 int ^̂ǂ炩قԂB
                CC֐ƂĎĂ̂ŁAp͂ȂB

  Arguments:    a, b - int ^

  Returns:      a <= b Ȃ a, Ȃ b
 *---------------------------------------------------------------------------*/
SDK_INLINE int MATH_IMin(int a, int b)
{
    return (a <= b) ? a : b;
}

/*---------------------------------------------------------------------------*
  Name:         MATH_MAX

  Description:  2̈̂ǂ炩傫قԂB
                }NƂĎĂ̂ŁAsZq`Ă
                ^Ŏgp\Be͕]邽߁ApɒӁB

  Arguments:    a, b

  Returns:      a >= b Ȃ a, Ȃ b
 *---------------------------------------------------------------------------*/
#define MATH_MAX(a,b) (((a) >= (b)) ? (a) : (b))

/*---------------------------------------------------------------------------*
  Name:         MATH_IMax

  Description:  2 int ^̂ǂ炩傫قԂB
                CC֐ƂĎĂ̂ŁAp͂ȂB

  Arguments:    a, b - int ^

  Returns:      a >= b Ȃ a, Ȃ b
 *---------------------------------------------------------------------------*/
SDK_INLINE int MATH_IMax(int a, int b)
{
    return (a >= b) ? a : b;
}

/*---------------------------------------------------------------------------*
  Name:         MATH_DIVUP

  Description:  base Ŋ[؂グ}NB

  Arguments:    x:      l
                base:   2 ׂ̂̊

  Returns:      x  base Ŋ[؂グ
 *---------------------------------------------------------------------------*/
#define MATH_DIVUP(x, base) (((x) + ((base)-1)) / (base))

/*---------------------------------------------------------------------------*
  Name:         MATH_ROUNDUP

  Description:  ؂グlԂ}NB
  
  Arguments:    x
                base - 2 ׂ̂̊

  Returns:      x ȏ̍ŏ base ̔{
 *---------------------------------------------------------------------------*/
#define MATH_ROUNDUP(x, base) (((x) + ((base)-1)) & ~((base)-1))

/*---------------------------------------------------------------------------*
  Name:         MATH_ROUNDDOWN

  Description:  ؂艺lԂ}NB
  
  Arguments:    x
                base - 2 ׂ̂̊

  Returns:      x ȉ̍ő base ̔{
 *---------------------------------------------------------------------------*/
#define MATH_ROUNDDOWN(x, base) ((x) & ~((base)-1))

/*---------------------------------------------------------------------------*
  Name:         MATH_ROUNDUP32

  Description:  32̔{ɐ؂グlԂ}NB
  
  Arguments:    x

  Returns:      x ȏ̍ŏ32̔{
 *---------------------------------------------------------------------------*/
#define MATH_ROUNDUP32(x) MATH_ROUNDUP(x, 32)

/*---------------------------------------------------------------------------*
  Name:         MATH_ROUNDDOWN32

  Description:  32̔{ɐ؂艺lԂ}NB
  
  Arguments:    x

  Returns:      x ȉ̍ő32̔{
 *---------------------------------------------------------------------------*/
#define MATH_ROUNDDOWN32(x) MATH_ROUNDDOWN(x, 32)


#ifdef __cplusplus
}/* extern "C" */
#endif

/* BROM_MATH_MATH_H_ */
#endif
