/*---------------------------------------------------------------------------*
  Project:  CtrBrom - OS - include
  File:     interrupt_types.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/
#ifndef BROM_OS_INTERRUPT_TYPES_H_
#define BROM_OS_INTERRUPT_TYPES_H_

#include <ctr/arm_reg.h>

#ifdef __cplusplus
extern "C" {
#endif

#define reg_OS_IDR_SET_IE                                  (( REGType32v *) REG_IDR_SET_IE0_ADDR)
#define reg_OS_IDR_SET_IE_ALL                              (*( REGType128v *) REG_IDR_SET_IE0_ADDR)
#define reg_OS_IDR_CLR_IE                                  (( REGType32v *) REG_IDR_CLR_IE0_ADDR)
#define reg_OS_IDR_CLR_IE_ALL                              (*( REGType128v *) REG_IDR_CLR_IE0_ADDR)
#define reg_OS_IDR_SET_PND                                 (( REGType32v *) REG_IDR_SET_PND0_ADDR)
#define reg_OS_IDR_SET_PND_ALL                             (*( REGType128v *) REG_IDR_SET_PND0_ADDR)
#define reg_OS_IDR_CLR_PND                                 (( REGType32v *) REG_IDR_CLR_PND0_ADDR)
#define reg_OS_IDR_CLR_PND_ALL                             (*( REGType128v *) REG_IDR_CLR_PND0_ADDR)
#define reg_OS_IDR_ACT                                     (( REGType32v *) REG_IDR_ACT0_ADDR)
#define reg_OS_IDR_TGT                                     (( REGType8v *) REG_IDR_TGT0_ADDR)
#define reg_OS_IDR_PRIO                                    (( REGType8v *) REG_IDR_PRIO0_ADDR)
#define reg_OS_IDR_CFG                                     (( REGType32v *) REG_IDR_CFG0_ADDR)

#define HW_IDR_WORD_MASK                                   (~0)

// Interrupt configuration registers

#define HW_IDR_INTR_RISE_EDGE       0x2 // The interrupt line is rising edge sensitive
#define HW_IDR_INTR_LEVEL_HIGH      0x0 //                       level high active
#define HW_IDR_INTR_1_N_MODEL       0x1 // The interrupt line uses the 1-N software model
#define HW_IDR_INTR_N_N_MODEL       0x0 //                         the N-N software model

// Binary Point Register

#define HW_CPUIBP_CMP_PRIO_ALL_BITS 3   // All priority bits are compared for pre-emption
#define HW_CPUIBP_CMP_PRIO_7_5      4   // Only bits [7:5] of priority are compared for pre-emption
#define HW_CPUIBP_CMP_PRIO_7_6      5   // Only bits [7:6] of priority are compared for pre-emption
#define HW_CPUIBP_CMP_PRIO_7        6   // Only bits [7] of priority are compared for pre-emption
#define HW_CPUIBP_NO_PREEMPT        7   // No pre-emption is performed


typedef u128 OSIntrMask;

typedef enum
{
    OS_INTR_ID_IPI_0 = 0,
    OS_INTR_ID_IPI_1 = 1,
    OS_INTR_ID_IPI_2 = 2,
    OS_INTR_ID_IPI_3 = 3,
    OS_INTR_ID_IPI_4 = 4,
    OS_INTR_ID_IPI_5 = 5,
    OS_INTR_ID_IPI_6 = 6,
    OS_INTR_ID_IPI_7 = 7,
    OS_INTR_ID_IPI_8 = 8,
    OS_INTR_ID_IPI_9 = 9,
    OS_INTR_ID_IPI_10 = 10,
    OS_INTR_ID_IPI_11 = 11,
    OS_INTR_ID_IPI_12 = 12,
    OS_INTR_ID_IPI_13 = 13,
    OS_INTR_ID_IPI_14 = 14,
    OS_INTR_ID_IPI_15 = 15,

    OS_INTR_ID_IPI_MIN = OS_INTR_ID_IPI_0,
    OS_INTR_ID_IPI_MAX = OS_INTR_ID_IPI_15,

    OS_INTR_ID_TIMER = REG_OS_IDR_SET_IE0_TM_SHIFT, // 29
    OS_INTR_ID_WATCHDOG = REG_OS_IDR_SET_IE0_WDOG_SHIFT, // 30

    OS_INTR_ID_MIN = OS_INTR_ID_IPI_MIN,
    OS_INTR_ID_MAX = (32*5) + REG_OS_IDR_SET_IE3_PMUIRQ7_SHIFT,
    OS_INTR_ID_NUM
}
OSIntrID;


#ifdef __cplusplus
} // extern "C"
#endif

#endif // BROM_OS_INTERRUPT_TYPES_H_
