/*---------------------------------------------------------------------------*
  Project:  CtrBrom - OS - include
  File:     spinLock.h

  Copyright 2009 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/
#ifndef BROM_OS_SPINLOCK_H_
#define BROM_OS_SPINLOCK_H_

#include <brom/types.h>
#include <brom/memorymap.h>

#ifdef __cplusplus
extern "C" {
#endif

//======================================================================
//                      bNϐ
//
//EJ[h^J[gbW^botԋL[Nq`l^uq`l|b^c
//  Ȃǂ̂botԂł̋LX[XAbotԑMehen ̐L\[X
//  ̔r䂽߂ɂ̊֐gpB
//======================================================================

// bNhc

#define OS_UNLOCK_ID            0      // bNϐbNĂȂꍇ̂hc
#define OS_MAINP_LOCKED_FLAG    0x40   // CvZbTɂ郍bNmFtO
#define OS_MAINP_LOCK_ID_START  0x40   // CvZbTpbNhc̊蓖ĊJnԍ
#define OS_MAINP_LOCK_ID_END    0x6f   //                               蓖ďIԍ
#define OS_MAINP_DBG_LOCK_ID    0x70   //                               fobK\ԍ
#define OS_MAINP_SYSTEM_LOCK_ID 0x7f   //                               VXe\ԍ
#define OS_SUBP_LOCKED_FLAG     0x80   //   TuvZbTɂ郍bNmFtO
#define OS_SUBP_LOCK_ID_START   0x80   //   TuvZbTpbNhc̊蓖ĊJnԍ
#define OS_SUBP_LOCK_ID_END     0xaf   //                               蓖ďIԍ
#define OS_SUBP_DBG_LOCK_ID     0xb0   //                               fobK\ԍ
#define OS_SUBP_SYSTEM_LOCK_ID  0xbf   //                               VXe\ԍ

#define OS_LOCK_SUCCESS         0      // bN
#define OS_LOCK_ERROR           (-1)   // bNG[

#define OS_UNLOCK_SUCCESS       0      // bN
#define OS_UNLOCK_ERROR         (-2)   // bNG[

#define OS_LOCK_FREE            0      // bN

#define OS_LOCK_ID_ERROR        (-3)   // bNhcG[


//---- structure of lock variable 
typedef struct OSLockByte
{
    u8      lockFlag;
    u8      ownerID;
}
OSLockByte;

/*---------------------------------------------------------------------------*
  Name:         osInitLock

  Description:  initialize system lock variable 
                and privilege to access shared resources

                * cartridge exclusive control area is not cleared 
                  because debugger uses.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void    osInitLock(void);

//----------------------------------------------------------------------
//          XsbN
//
//EvZbTԂ⃂W[ԂŋL\[Xr䂷邽߂
//  bNϐ̃XsbNs܂B
//EbN܂Ŏs܂B
//EvZbTԂ̋L\[X͕KbNɎgpĉB
//EvZbTL̃\[X̓^C~OIɒł̂ł
//  bNȂĂ\܂B
//  L\[X̓fobOɌ肵ăbN邱Ƃł܂B
//
//EF
//  lockID              bNID
//  lockp               bNϐ̃|C^
//
//E߂lF
//  OS_LOCK_SUCCESS     bN
//----------------------------------------------------------------------
s32     osLockByte(u8 lockID, OSLockByte *lockp, void (*ctrlFuncp) (void));
s32     osLockCard(u8 lockID);

//----------------------------------------------------------------------
//          bN̉
//
//EbNAL\[X̃ANZXTuvZbTɓn܂B
//EbNĂȂW[sꍇɂ͉ꂸA
//  OS_UNLOCK_ERROR Ԃ܂B
//
//EF
//  lockID              bNID
//  lockp               bNϐ̃|C^
//
//E߂lF
//  OS_UNLOCK_SUCCESS   bN
//  OS_UNLOCK_ERROR     bNG[
//----------------------------------------------------------------------
s32     osUnlockByte(u8 lockID, OSLockByte *lockp, void (*ctrlFuncp) (void));
s32     osUnlockCard(u8 lockID);

//----------------------------------------------------------------------
//          bN̎s
//
//EP񂾂XsbNs܂B
//EvZbTԂ̋L\[X͕KbNɎgpĉB
//EvZbTL̃\[X̓^C~OIɒł̂ł
//  bNȂĂ\܂B
//  L\[X̓fobOɌ肵ăbN邱Ƃł܂B
//
//EF
//  lockID              bNID
//  lockp               bNϐ̃|C^
//  CtrlFuncp           \[X֐̃|C^
//
//E߂lF
//                    bNiOɊi[ĂIDj
//  OS_LOCK_SUCCESS     bN
//----------------------------------------------------------------------
s32     osTryLockByte(u8 lockID, OSLockByte *lockp, void (*crtlFuncp) (void));
s32     osTryLockCard(u8 lockID);

//----------------------------------------------------------------------
//          bNϐ̏LW[ID̓ǂݍ
//
//EbNϐ̏LW[ID ǂݍ݂܂B
//EW[ID 0 ̏ꍇ͂̎_łǂ̃vZbT
//  LĂ̂mFł܂B
//EL\[X̏ꍇ́uCvZbTLĂԁv
//  ݂̂荞݂֎~邱ƂɂĈێ邱Ƃł܂B
//  ̑̏Ԃ̓TuvZbTωĂ܂\܂B
//ELW[ID  0 łĂbNϐĂƂ͌܂B
//
//EF
//  lockp       bNϐ̃|C^
//
//E߂lF    LW[ID
//
//AACɑ΂Ă̓LbVoRłȂ
//  oCgANZX͂łȂƂɒӂĉB
//  ł̂ŁACł͊{I osReadOwnerOfLockWord() gpĉB
//----------------------------------------------------------------------
u8     osReadOwnerOfLockWord(OSLockByte *lockp);
#define osReadOwnerOfLockCard()       osReadOwnerOfLockWord( (OSLockByte *)HW_CARD_LOCK_BUF  )



/*---------------------------------------------------------------------------*
  Name:         osGetLockID

  Description:  get lock ID

  Arguments:    None.

  Returns:      OS_LOCK_ID_ERROR, if fail to get ID

                if ARM9
                   0x40`0x6f       lockID
                else if ARM7
                   0x80`0xaf       lockID
                endif

                *Notice:  ID is allocated only 48 pattern at a highest.

 *---------------------------------------------------------------------------*/
s32     osGetLockID(void);


/*---------------------------------------------------------------------------*
  Name:         osReleaseLockID

  Description:  release lock ID

  Arguments:    lockID : id to tend to release

  Returns:      None.
 *---------------------------------------------------------------------------*/
void    osReleaseLockID(u8 lockID);



//================================================================
//   private function.
//   This function is for internal use. Never call from application.
#define OSi_SYNCTYPE_SENDER 0
#define OSi_SYNCTYPE_RECVER 1

#define OSi_SYNCVAL_NOT_READY 0
#define OSi_SYNCVAL_READY     1

void OSi_SyncWithOtherProc( int type, void* syncBuf );

static inline void OSi_SetSyncValue( u8 n )
{
	*(vu8*)(HW_INIT_LOCK_BUF+4) = n;
}
static inline u8 OSi_GetSyncValue( void )
{
	return *(vu8*)(HW_INIT_LOCK_BUF+4);
}



//--------------------------------------------------------------------------------
#ifdef __cplusplus
} /* extern "C" */
#endif

/* BROM_OS_SPINLOCK_H_ */
#endif
