/*---------------------------------------------------------------------------*
  Project:  TWL - append for RTFS
  File:     rtfs_twl_append.c

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date:: #$
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/

#include <rtfs.h>

#define PRINTDEBUG( ...) ((void)0)
//#define PRINTDEBUG   OS_TPrintf

/***********************************************************************
 extern functions
***********************************************************************/
#if (INCLUDE_FAILSAFE_CODE)
extern dword fs_map_block(FAILSAFECONTEXT *pfscntxt, dword blockno);
#endif
extern DROBJ *pc_get_inode_ex( DROBJ *pobj, DROBJ *pmom, byte *filename, byte *fileext, int action, BOOL get_lfname);
extern BOOLEAN pc_findin_ex( DROBJ *pobj, byte *filename, byte *fileext, int action, BOOL get_lfname);

/***********************************************************************
 global functions
***********************************************************************/
BOOLEAN pc_fill_blk(BLKBUFF *pblk, int blocknum);
BOOLEAN block_devio_fill(BLKBUFF *pblk, int blocknum);
BOOLEAN devio_fill(int driveno, dword blockno, byte * buf, word n_to_write, BOOLEAN raw);
BOOLEAN rtfsi_pc_cache_clusters( PC_FILE *pfile, dword file_pointer, void* buf, u32 buf_size, u32* start_index);
static BOOL rtfsi_check_buffer_conflict( void* buf, u32 buf_size);
DROBJ *pc_mknode_ex(DROBJ *pmom ,byte *filename, byte *fileext, byte attributes, CLUSTERTYPE incluster);


/*---------------------------------------------------------------------------*
  Name:         pc_fill_blk

  Description:   prblock.c  pc_write_blk

  Arguments:    

  Returns:      TRUE / FALSE
 *---------------------------------------------------------------------------*/
BOOLEAN pc_fill_blk(BLKBUFF *pblk, int blocknum)
{
        if (!block_devio_fill(pblk, blocknum))
        {
            /* set errno to IO error unless devio set PEDEVICE */
            if (!get_errno())
                rtfs_set_errno(PEIOERRORWRITEBLOCK); /* pc_write_blk device write error */
            return(FALSE);
        }
        else
            return(TRUE);
}


/*---------------------------------------------------------------------------*
  Name:         block_devio_fill

  Description:   prblock.c  pc_write_blk (FAILSAFEgpj܂
                     prfscore.c  pc_write_blkiFAILSAFEgpj

  Arguments:    

  Returns:      TRUE / FALSE
 *---------------------------------------------------------------------------*/
#if (!INCLUDE_FAILSAFE_CODE)
/* prblock.c  block_devio_write */
BOOLEAN block_devio_fill(BLKBUFF *pblk, int blocknum)
{
/*    dword blockno;
    blockno = pblk->blockno;*/
    return(devio_fill(pblk->pdrive->driveno, pblk->blockno, pblk->data, (int)blocknum, FALSE));
}

#else
/* prfscore.c  block_devio_write */
BOOLEAN block_devio_fill(BLKBUFF *pblk, int blocknum)
{
dword mapped_block;
FAILSAFECONTEXT *pfscntxt;
    /* Write to the journal file if enabled */
    pfscntxt = (FAILSAFECONTEXT *) pblk->pdrive->pfscntxt;
    if (pfscntxt && pfscntxt->configuration_flags & FS_MODE_JOURNALING)
    {
        mapped_block = fs_map_block(pfscntxt, pblk->blockno);
        if (!mapped_block)  /* fs_map_block() set errno */
            return(FALSE);
        if (failsafe_write_nv_buffer(pfscntxt, mapped_block,pblk->data))
            return (TRUE);
        else
        {
            rtfs_set_errno(PEIOERRORWRITEJOURNAL);
            return(FALSE);
        }
    }
    else /* not journalling or not mapped so read from the volume */
        return(devio_fill(pblk->pdrive->driveno,pblk->blockno, pblk->data, (int)blocknum, FALSE));
}
#endif






/*---------------------------------------------------------------------------*
  Name:         pc_gfirst_ex

  Description:  tOԂpc_gfirst

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
void pc_upstat(DSTAT *statobj);
void pc_upstat_ex(DSTAT *statobj, u8* scf, BOOL get_lfname);

BOOLEAN pc_gfirst_ex(DSTAT *statobj, u8 *scf, byte *name, BOOL get_lfname)      /*__apifn__*/
{
    byte  *mompath;
    byte  *filename;
    byte  fileext[4];
    int driveno;
    DDRIVE *pdrive;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */

    rtfs_set_errno(0);  /* po_gfirst: clear error status */
    rtfs_memset((byte *) statobj,0,sizeof(*statobj));
/*    statobj->pobj = 0; */
/*    statobj->pmom = 0; */

    /* Get the drive and make sure it is mounted   */
    driveno = check_drive_name_mount(name);
    if (driveno < 0)
    {
        /* errno was set by check_drive */
        return(FALSE);
    }
    pdrive = pc_drno2dr(driveno);
    /* Use the buffers in the DRIVE structure. Access is locked via semaphore */
    mompath = (byte *)&(pdrive->pathname_buffer[0]);
    filename = (byte *)&(pdrive->filename_buffer[0]);

    /* Get out the filename and d:parent   */
    if (!pc_parsepath(mompath,filename,fileext,name))
    {
        release_drive_mount(driveno);/* Release lock, unmount if aborted */
        rtfs_set_errno(PEINVALIDPATH); /* pc_gfirst: Bad path name */
        return(FALSE);
    }

    /* Save the pattern. we will need it in pc_gnext   */
    copybuff(statobj->pname, filename, FILENAMESIZE_BYTES);
    copybuff(statobj->pext, fileext, 4);
    /* Copy over the path. we will need it later   */
    copybuff(statobj->path, mompath, EMAXPATH_BYTES);
    /* Find the file and init the structure   */
    statobj->pmom = (void *) pc_fndnode(mompath);
    /* pc_fndnode will set errno */
    if (statobj->pmom)
        /* Found it. Check access permissions   */
    {
        if(pc_isadir((DROBJ *)(statobj->pmom)))
        {
            /* Now find pattern in the directory   */
            statobj->pobj = (void *) pc_get_inode(0, (DROBJ *)(statobj->pmom), filename, (byte*) fileext, GET_INODE_WILD);
            if (statobj->pobj)
           {
               /*----- twl modified -----*/
               *scf = ((DROBJ*)(statobj->pobj))->finode->resarea[0]; //tO
               /* And update the stat structure   */
               pc_upstat_ex(statobj, scf, get_lfname);
               /*-----------------------*/

                /* remember the drive number. used by gnext et al.   */
                statobj->driveno = driveno;
/* 9-20-94 release the FINODE and allocate a dummy. This will keep everyone
    who expects the drobj to own a finode happy but will not leave the
    finode open which locks out unlink et al */
                pc_freei(((DROBJ *)(statobj->pobj))->finode); /* Release the current */
/* 3-07-07 - Change: Remove additional call to pc_alloci(). Was not needed and caused a leak
   on a hot swap event when a gfirst is outstanding */
                ((DROBJ *)(statobj->pobj))->finode = 0;
/* END 9-20-94   */
                /* Remember the unique number associated with the drive
                   mount. If the drive is closed before we call gnext or
                   gdone we'll know about it because this number won't match */
                statobj->drive_opencounter = pdrive->drive_opencounter;
                release_drive_mount(driveno);/* Release lock, unmount if aborted */
                return(TRUE);
            }
            else
            {
            /* pc_gfirst: if statobj->pobj is 0 pc_get_inode() has set errno to PENOENT or
               to an internal or IO error status
               if PENOENT set we will clear errno */
                if (get_errno() == PENOENT)
                    rtfs_set_errno(0); /* pc_gfirst: file not found in directory
                                      set errno to zero and return FALSE */
            }
        }
        else
            rtfs_set_errno(PEINVALIDDIR); /* pc_gfirst: Path not a directory, report not found */
    }
    /* If it gets here we had a problem   */
    if (statobj->pmom)
        pc_freeobj((DROBJ *)statobj->pmom);
    rtfs_memset((byte *) statobj,0,sizeof(*statobj));
    release_drive_mount(driveno);/* Release lock, unmount if aborted */

    return(FALSE);
}

/*---------------------------------------------------------------------------*
  Name:         pc_gnext_ex

  Description:  tOԂpc_gnext

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
BOOLEAN pc_gnext_ex(DSTAT *statobj, u8 *scf, BOOL get_lfname)     /*__apifn__*/
{
    DROBJ *nextobj;
    DDRIVE *pdrive;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */
    /* see if the drive is still mounted. Do not use pmom et al. since they
        may be purged */
    if (!statobj || !statobj->pmom)
    {
        rtfs_set_errno(PEINVALIDPARMS); /* pc_gnext: statobj is not valid */
        return(FALSE);
    }
    if (!check_drive_number_mount(statobj->driveno))
        return(FALSE);
    pdrive = pc_drno2dr(statobj->driveno);
    if (statobj->drive_opencounter != pdrive->drive_opencounter)
    { /* Card was removed and re-inserted since pc_gfirst() */
        rtfs_set_errno(PEINVALIDPARMS); /* pc_gnext: statobj is not valid */
        release_drive_mount(statobj->driveno);/* Release lock, unmount if aborted */
        return(FALSE);
    }

    rtfs_set_errno(0);  /* po_gnext: clear error status */

    /* Now find the next instance of pattern in the directory   */
    nextobj = pc_get_inode_ex((DROBJ *)(statobj->pobj), (DROBJ *)(statobj->pmom),
                              statobj->pname, statobj->pext, GET_INODE_WILD, FALSE);
    if (nextobj)
    {
        statobj->pobj = (void *)nextobj;
        /*----- twl modified -----*/
        *scf = ((DROBJ*)(statobj->pobj))->finode->resarea[0]; //tO
        /* And update the stat structure   */
        pc_upstat_ex(statobj, scf, get_lfname);
        /*-----------------------*/

/* 9-20-94 release the FINODE and allocate a dummy. This will keep everyone
    who expects the drobj to own a finode happy but will not leave the
    finode open which locks out unlink et al */
         pc_freei(((DROBJ *)(statobj->pobj))->finode); /* Release the current */
/* 3-07-07 - Change: Remove additional call to pc_alloci(). Was not needed and caused a leak
   on a hot swap event when a gfirst is outstanding */
         ((DROBJ *)(statobj->pobj))->finode = 0;
/* END 9-20-94   */
        release_drive_mount(statobj->driveno);/* Release lock, unmount if aborted */
        return(TRUE);
    }
    else
    {
        if (get_errno() == PENOENT)
            rtfs_set_errno(0); /* get_inode: file not found in directory
                                  set errno to zero and return FALSE */
       /* pc_gnext: nextobj is 0 pc_get_inode() has set errno to PENOENT or to an internal or IO error status */
        release_drive_mount(statobj->driveno);/* Release lock, unmount if aborted */
        return(FALSE);
    }
}

/* Copy internal stuf so the outside world can see it   */
void pc_upstat_ex(DSTAT *statobj, u8* scf, BOOL get_lfname)
{
    DROBJ *pobj;
    FINODE *pi;
    pobj = (DROBJ *)(statobj->pobj);

    pi = pobj->finode;

    copybuff( statobj->fname, pi->fname, 8);
    statobj->fname[8] = CS_OP_ASCII('\0');
    copybuff( statobj->fext, pi->fext, 3);
    statobj->fext[3] = CS_OP_ASCII('\0');
    /* put null termed file.ext into statobj   */
    pc_ascii_mfile((byte *)statobj->filename, (byte *)statobj->fname,
             (byte *)statobj->fext);

    statobj->fattribute = pi->fattribute;
    statobj->ftime = pi->ftime;
    statobj->fdate = pi->fdate;
    statobj->fsize = pi->fsize;

    if( get_lfname) {
        /* Get the lfn value for this object. If none available make
           sure it is a NULL string in ASCII or UNICODE */
        if (!pc_get_lfn_filename(pobj, (byte *)statobj->lfname))
        {
            statobj->lfname[0] = statobj->lfname[1] = 0;
            pc_cs_mfile((byte *)statobj->lfname, (byte *)statobj->fname,
                 (byte *)statobj->fext); //sfnlfnɃRs[
            *scf |= 0x80; //O݂Ȃ
        }
    }else{
        statobj->lfname[0] = statobj->lfname[1] = 0;
#if 0
        pc_cs_mfile((byte *)statobj->lfname, (byte *)statobj->fname,
             (byte *)statobj->fext); //sfnlfnɃRs[
#endif
//        *scf |= 0x80; //O݂Ȃ
    }
}

/* Give a directory mom. And a file name and extension.
    Find find the file or dir and initialize pobj.
    If pobj is NULL. We allocate and initialize the object otherwise we get the
    next item in the chain of dirents.
*/
DROBJ *pc_get_inode_ex( DROBJ *pobj, DROBJ *pmom, byte *filename, byte *fileext, int action, BOOL get_lfname) /*__fn__*/
{
    BOOLEAN  starting = FALSE;
    /* Create the child if just starting   */
    if (!pobj)
    {
        starting = TRUE;
        pobj = pc_mkchild(pmom);    /* If failure sets errno */
        if (!pobj)
            return(0);
    }
    else    /* If doing a gnext do not get stuck in and endless loop */
    {
        if ( ++(pobj->blkinfo.my_index) >= INOPBLOCK )
        {
            rtfs_set_errno(0);  /* Clear errno to be safe */
            if (!pc_next_block(pobj)) /* Will set errno illegal block id encountered */
            {
                if (!get_errno())
                    rtfs_set_errno(PENOENT); /* pc_get_inode: end of dir */
                if (starting)
                    pc_freeobj(pobj);
                return(0);
            }
            else
                pobj->blkinfo.my_index = 0;
        }
    }
    if (pc_findin_ex(pobj, filename, fileext,action, get_lfname))
    {
        return (pobj);
    }
    else
    {
        /* pc_findin set errno */
        if (starting)
            pc_freeobj(pobj);
        return (0);
    }
}

/*---------------------------------------------------------------------------*
  Name:         rtfs_pc_cache_clusters

  Description:  fd󂯎ăNX^`FCLbV

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
BOOL rtfs_pc_cache_clusters( PCFD fd, void* buf, u32 buf_size)
{
    BOOL    ret_val = FALSE;
    PC_FILE *pfile;
    DDRIVE  *pdrive;
    int     log2_bytespcluster;
    int     end_of_chain;
    int     i = 0;
    dword   eof_pointer;
    CLUSTERTYPE  n_clusters_to_seek, n_clusters;
    CLUSTERTYPE  first_cluster, next_cluster;
    u32 start;

    rtfs_set_errno( 0);

    /* buf4oCgACgKv */
    if( ((u32)buf) % 4) {
        rtfs_set_errno( PEINVALIDPARMS);
        return( FALSE);
    }

    /* ɃLbVobt@ƂĎgĂ */
    if( rtfsi_check_buffer_conflict( buf, buf_size) == TRUE) {
        rtfs_set_errno( PEINVALIDPARMS);
        return( FALSE);
    }

    /* Get the file structure and semaphore lock the drive */
    pfile = pc_fd2file(fd, 0);            /* lock */
    if (!pfile)
    { /* pc_fd2file set errno */
        rtfs_set_errno( PEINVALIDPARMS);
        return( FALSE);
    }else{
        pdrive = pfile->pobj->pdrive;
    }
  
    /* Cg\[hŃI[vĂ玸s */
    if( pfile->flag & (PO_WRONLY|PO_RDWR)) {
        rtfs_set_errno( PEINVALIDPATH);
        release_drive_mount(pdrive->driveno);/* Release lock, unmount if aborted */
        return( FALSE);
    }

    if( rtfsi_pc_cache_clusters( pfile, pfile->fptr, buf, buf_size, &start) == FALSE)
    {
        /* VoɃZbg */
        pfile->fcluster_cache.start_index = 0xFFFFFFFF;
        pfile->fcluster_cache.buf_size    = 0;
        pfile->fcluster_cache.buf         = NULL;
        ret_val = FALSE;
    }else{
        /* VoɃZbg */
        pfile->fcluster_cache.start_index = start;
        pfile->fcluster_cache.buf_size    = buf_size;
        pfile->fcluster_cache.buf         = buf;
        ret_val = TRUE;
    }

    release_drive_mount(pdrive->driveno);/* Release lock, unmount if aborted */
    return( ret_val);
}

/*---------------------------------------------------------------------------*
  Name:         rtfsi_check_buffer_conflict

  Description:  LbVobt@ɑ̃t@CɊ蓖ĂĂ邩𒲂ׂ

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
static BOOL rtfsi_check_buffer_conflict( void* buf, u32 buf_size)
{
    BOOL    ret_val = FALSE;
    PC_FILE *pfile;
    PCFD    i;
    u32     nowbuf, nowbuf_end, mybuf_end;
    u32     mybuf = (u32)buf;
    
    rtfs_set_errno(0);    /* clear errno */
    
    pfile = prtfs_cfg->mem_file_pool;
    for (i=0;i<pc_nuserfiles();i++, pfile++)
    {
        if( (pfile->is_free) == FALSE) //openH
        {
            if( !pfile->pobj) {
                //PECLOSEDG[ɑ(pc_fd2fileQ)
            }else{             //open
#if 1           /* 擪̂݃`FbN */
                if( pfile->fcluster_cache.buf == buf) {
                    ret_val = TRUE;
                    break;
                }
#else           /* ͈̓`FbN */
                mybuf_end = mybuf + buf_size;
                nowbuf = (u32)(pfile->fcluster_cache.buf);
                nowbuf_end = nowbuf + (pfile->fcluster_cache.buf_size);
                if( (nowbuf >= mybuf)&&(nowbuf <= mybuf_end)) {
                    ret_val = TRUE;
                    break;
                }
                if( (nowbuf_end >= mybuf)&&(nowbuf_end <= mybuf_end)) {
                    ret_val = TRUE;
                    break;
                }
                if( (mybuf >= nowbuf)&&(mybuf <= nowbuf_end)) {
                    ret_val = TRUE;
                    break;
                }
                if( (mybuf_end >= nowbuf)&&(mybuf_end <= nowbuf_end)) {
                    ret_val = TRUE;
                    break;
                }
#endif
            }
        }
    }
    if( ret_val == TRUE) {
        rtfs_set_errno( PEINVALIDPARMS);
        osTWarning( "(FS_SetSeekCache):Cache Buffer Conflict!\n");
    }
    return( ret_val);
}

/*---------------------------------------------------------------------------*
  Name:         rtfs_get_next_cluster_cache

  Description:  

  Arguments:    

  Returns:      TRUE / FALSE
 *---------------------------------------------------------------------------*/
BOOL rtfs_get_next_cluster_cache( PC_FILE *pfile, dword file_pointer, CLUSTERTYPE* clno)
{
    BOOL  result;
    DDRIVE *pdrive;
    int  log2_bytespcluster;
    u32 cluster_index, start_index, next_of_end_index;
    u32 *fat32_buf;
    u16 *fat16_buf;

    pdrive = pfile->pobj->pdrive;
    log2_bytespcluster = (int) (pdrive->log2_secpalloc + 9);

    cluster_index     = (file_pointer >> log2_bytespcluster);
    start_index       = pfile->fcluster_cache.start_index;
  
    if( pdrive->fasize == 8) {
        next_of_end_index = start_index + (pfile->fcluster_cache.buf_size / 4);
    }else{
        next_of_end_index = start_index + (pfile->fcluster_cache.buf_size / 2);
    }
  
    if( (cluster_index >= start_index) && (cluster_index < next_of_end_index)) {
        /* hit */
        result = TRUE;
        if( pdrive->fasize == 8) { //FAT32
            fat32_buf = (u32*)(pfile->fcluster_cache.buf);
            *clno = (CLUSTERTYPE)(fat32_buf[cluster_index - start_index]);
        }else{                     //FAT16,12
            fat16_buf = (u16*)(pfile->fcluster_cache.buf);
            *clno = (CLUSTERTYPE)(fat16_buf[cluster_index - start_index]);
        }
    }else{
        /* miss */
        result = FALSE;
    }
    return( result);
}

/*---------------------------------------------------------------------------*
  Name:         rtfsi_pc_cache_clusters

  Description:  PC_FILE\̂󂯎ăNX^`FCLbV

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
BOOL rtfsi_pc_cache_clusters( PC_FILE *pfile, dword file_pointer, void* buf, u32 buf_size, u32* start_index)
{
    BOOL    ret_val = FALSE;
    DDRIVE  *pdrive;
    u32     *fat32_buf = (u32*)buf;
    u16     *fat16_buf = (u16*)buf;
    u32     bytespcluster;
    int     log2_bytespcluster;
    int     end_of_chain;
    int     i;
    int     index = 0;
    CLUSTERTYPE  n_clusters_to_seek, n_clusters, n_clusters_to_eof;
    CLUSTERTYPE  first_cluster, next_cluster;
    u32 index_num, now_index;

    rtfs_set_errno( 0);

    /* bufNULLłȂ4oCgACgKv */
    if( (buf == NULL)||((u32)buf) % 4) {
        rtfs_set_errno( PEINVALIDPARMS);
        return( FALSE);
    }
    /* pfile`FbN */
    if (!pfile)
    {
        rtfs_set_errno( PEINVALIDPARMS);
        return( FALSE);
    }

    pdrive = pfile->pobj->pdrive;
    log2_bytespcluster = (int) (pdrive->log2_secpalloc + 9);
    bytespcluster = (1<<log2_bytespcluster);

    if( pfile->pobj->finode->fsize) {
        if( ((file_pointer % bytespcluster) == 0)&&(file_pointer == pfile->pobj->finode->fsize)) {
            n_clusters_to_seek = (CLUSTERTYPE)((file_pointer >> log2_bytespcluster));
        }else{
            n_clusters_to_seek = (CLUSTERTYPE)((file_pointer >> log2_bytespcluster) + 1);
        }
            first_cluster = pc_finode_cluster(pdrive,pfile->pobj->finode);
            now_index = 0;

        next_cluster = first_cluster; /* Set next cluster to first cluster so if n_clusters_to_seek is zero it's harmless. */

        { /*---------- obt@Jnʒu̓ ----------*/
            u32 new_index;
            if( ((file_pointer % bytespcluster) == 0)&&(file_pointer == pfile->pobj->finode->fsize)) {
                new_index = (file_pointer >> log2_bytespcluster) - 1;
            }else{
                new_index = (file_pointer >> log2_bytespcluster);
            }

            if( pdrive->fasize == 8) {
                index_num = (buf_size / 4);
            }else{
                index_num = (buf_size / 2);
            }
            if( index_num == 0) { //1Ȃ
                rtfs_set_errno( PEINVALIDPARMS);
                return( FALSE);
            }

            if( new_index >= index_num) {
                *start_index = ((new_index + 1) - index_num); //+1Ōݒn
            }else{ //߂肫ȂƂ0LbV
                *start_index = 0;
                n_clusters_to_eof = (CLUSTERTYPE)
                  ((pfile->pobj->finode->fsize + bytespcluster - 1) >> log2_bytespcluster);
//                  ((pfile->pobj->finode->fsize+(pdrive->secpalloc<<9)-1) >> (pdrive->log2_secpalloc+9));
                n_clusters_to_seek = (n_clusters_to_eof > index_num) ? index_num : n_clusters_to_eof;
            }
        } /*--------------------------------------------*/
      PRINTDEBUG( "CACHING -- (index_num:%d, start_index:%d)\n", index_num, *start_index);
      PRINTDEBUG( "           n_clusters_to_seek:%d\n", n_clusters_to_seek);
      PRINTDEBUG( "           fptr:%d -> file_pointer:%d\n", pfile->fptr, file_pointer);
      
        while( n_clusters_to_seek)
        {
            end_of_chain = 0;
            /* ̍Ōォn_clusters_to_seek܂łŌI֐ */
            n_clusters = FATOP(pdrive)->fatop_get_chain(pdrive, first_cluster,
                            &next_cluster, n_clusters_to_seek, &end_of_chain);
      PRINTDEBUG( "%s, n_clusters_to_seek:%d, n_clusters:%d\n", __FUNCTION__, n_clusters_to_seek, n_clusters);
            if (!n_clusters) /* get_chain stops when it reaches n_clusters_to_seek */
            {/* get_chain already set errno */
                rtfs_set_errno( PEINVALIDCLUSTER);
                goto return_error;
            }
            for( i=0; i<n_clusters; i++) {
      PRINTDEBUG( "after(i:%d, n_clusters:%d)\n", i, n_clusters);
                if( now_index >= *start_index) {
      PRINTDEBUG( "i:%d, now_index:%d, start_index:%d\n", i, now_index, *start_index);
                  if( index >= index_num) { PRINTDEBUG( "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");}
      PRINTDEBUG( "index:%d, index_num:%d\n", index, index_num);
                    if( pdrive->fasize == 8) { //FAT32
                        fat32_buf[index++] = first_cluster + i;
                    }else{
                        fat16_buf[index++] = first_cluster + i;
                    }
                }
                now_index++;
            }

            /* How many left to seek */
            n_clusters_to_seek = (CLUSTERTYPE) (n_clusters_to_seek - n_clusters);

            if( n_clusters_to_seek && end_of_chain)
            {
                rtfs_set_errno( PEINVALIDCLUSTER);
                goto return_error;
            }else{
                first_cluster = next_cluster; //̓
            }
        }
        ret_val = TRUE;
    }

return_error:
    return( ret_val);
}


/*---------------------------------------------------------------------------*
  Name:         twfs_po_open

  Description:  

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
PCFD twfs_po_open( byte *name, word flag, word mode)
{
    PCFD    fd;
    PC_FILE *pfile;
    DDRIVE  *pdrive;

    fd = po_open( name, flag, mode);
    if( fd >= 0) {
        /* Get the file structure and semaphore lock the drive */
        pfile = pc_fd2file(fd, 0);        /* lock */
        if( !pfile)
        {
            return( fd);
        }
        /* tV[NpVo𖳌ɃZbg */
        pfile->fcluster_cache.start_index = 0xFFFFFFFF;
        pfile->fcluster_cache.buf_size    = 0;
        pfile->fcluster_cache.buf         = NULL;

        pdrive = pfile->pobj->pdrive;
        release_drive_mount(pdrive->driveno);/* Release lock, unmount if aborted */
    }
    return( fd);
}


/*---------------------------------------------------------------------------*
  Name:         twfs_pc_get_fatbits

  Description:  

  Arguments:    

  Returns:      
 *---------------------------------------------------------------------------*/
int twfs_pc_get_fatbits( DDRIVE *pdrive)
{
    int ret_val;
    if (pdrive)
    {
        if( pdrive->fasize == 8) {
            ret_val = 32;
        }else if( pdrive->fasize == 4) {
            ret_val = 16;
        }else if( pdrive->fasize == 3) {
            ret_val = 12;
        }else{
            rtfs_set_errno(PEINVALIDBPB);
            ret_val = -1;
        }
    }
    else
    {
        ret_val = -1;
    }

    return(ret_val);
}


#if 0
/* dQƏԂȂɃf[^XN̉\pc_mv */
/* Rename a file */
BOOLEAN pc_mv_ex(byte *old_name, byte *new_name)  /*__apifn__*/
{
    int old_driveno;
    DROBJ *old_obj;
    DROBJ *old_parent_obj;
    byte  *path;
    byte  *filename;
    byte fileext[4];
    int new_driveno;
    DROBJ *new_obj;
    DROBJ *new_parent_obj;
    BOOLEAN ret_val;
    CLUSTERTYPE  cluster;
    DDRIVE *pdrive;
  
    byte bkup_fattribute;
    word   bkup_ftime;            /* time & date lastmodified */
    word   bkup_fdate;
    dword  bkup_fsize;            /* File size */
    

    int  p_errno;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */
    p_errno = 0;

    /* Drives must be the same */
    if (    !pc_parsedrive( &old_driveno, old_name ) ||
            !pc_parsedrive( &new_driveno, new_name ) ||
            old_driveno != new_driveno)
    {
        rtfs_set_errno(PEINVALIDDRIVEID);
        return(FALSE);
    }

    /* Get the drive and make sure it is mounted   */
    old_driveno = check_drive_name_mount(old_name);
    if (old_driveno < 0)
    {
        /* errno was set by check_drive */
        return(FALSE);
    }
    rtfs_set_errno(0);  /* pc_mv: clear error status */

    old_obj         = 0;
    old_parent_obj  = 0;
    new_obj         = 0;
    new_parent_obj  = 0;
    ret_val = FALSE;

    pdrive = pc_drno2dr(old_driveno);

    /* Use the buffers in the DRIVE structure. Access is locked via semaphore */
    path = (byte *)&(pdrive->pathname_buffer[0]);
    filename = (byte *)&(pdrive->filename_buffer[0]);

    /* Get out the filename and d:parent */
    if (!pc_parsepath(path, filename,fileext,old_name))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }
    /* Find the parent and make sure it is a directory  */
    old_parent_obj = pc_fndnode(path);
    if (!old_parent_obj)
        goto errex; /* pc_fndinode - set errno */

    if (!pc_isadir(old_parent_obj))
    {
        p_errno = PENOENT;
        goto errex;
    }
    /* Find the file */
    old_obj = pc_get_inode(0, old_parent_obj,
        filename, (byte*)fileext, GET_INODE_MATCH);
    if (!old_obj)
        goto errex; /* pc_get_inode - set errno */

     /* Be sure it exists and is a normal directory or file and is not open */
    if (pc_isroot(old_obj) || (old_obj->finode->opencount > 1) ||
       (old_obj->finode->fattribute&(ARDONLY|AVOLUME)))
    {
        p_errno = PEACCES;
        goto errex;
    }

    /* At this point old_obj contains the file we are renaming */

    /* See if the new directory entry already exists */
    new_obj = pc_fndnode(new_name);
    if (new_obj)
    {
        p_errno = PEEXIST;
        goto errex;
    }
    rtfs_set_errno(0);  /* pc_mv - clear errno condition after failed pc_fndnode */

    /* Get out the filename and d:parent */
    if (!pc_parsepath(path,filename,fileext,new_name))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }
    /* Find the parent and make sure it is a directory  */
    new_parent_obj = pc_fndnode(path);
    if (!new_parent_obj || !pc_isadir(new_parent_obj) ||  pc_isavol(new_parent_obj))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }


    /* Copy the old directory entry stuf over */
    bkup_fattribute = old_obj->finode->fattribute;
    bkup_ftime      = old_obj->finode->ftime;
    bkup_fdate      = old_obj->finode->fdate;
    bkup_fsize      = old_obj->finode->fsize;
    /* Set the old cluster value to zero */
    pc_pfinode_cluster(old_obj->pdrive,old_obj->finode,0);
    /* Delete the old but won't delete any clusters */
    if (!pc_rmnode(old_obj))
        goto errex;


    /* The cluster value old */
    cluster = pc_finode_cluster(old_obj->pdrive,old_obj->finode);

    /* Create the new entry and assign cluster to it. If it is a directory
       .. will be linked correctly */
    new_obj = pc_mknode( new_parent_obj, filename, fileext, bkup_fattribute, cluster);
    if (!new_obj)
        goto errex;

    /* Copy the old directory entry stuf over */
    new_obj->finode->fattribute = bkup_fattribute;
    new_obj->finode->ftime = bkup_ftime;
    new_obj->finode->fdate = bkup_fdate;
    new_obj->finode->fsize = bkup_fsize;

    /* Update the new inode. Do not set archive bit or change date */
     if (!pc_update_inode(new_obj, FALSE, FALSE))
        goto errex;

    p_errno = 0;
    ret_val = TRUE;

    /* Good conditions fall through here, error exits jump to here */
errex:
    if (old_parent_obj)
        pc_freeobj(old_parent_obj);
    if (old_obj)
        pc_freeobj(old_obj);
    if (new_parent_obj)
        pc_freeobj(new_parent_obj);
    if (new_obj)
        pc_freeobj(new_obj);
    /* Set errno if we have one and not set by lower level already */
    if ((p_errno) && !get_errno())
        rtfs_set_errno(p_errno);
    if (!release_drive_mount_write(old_driveno))/* Release lock, unmount if aborted */
        ret_val = FALSE;
    return(ret_val);
}
#else
/* fBNgt@CɂȂ^C~Ȏ݂Ȃpc_mv */
BOOLEAN pc_mv_ex(byte *old_name, byte *new_name)  /*__apifn__*/
{
    int old_driveno;
    DROBJ *old_obj;
    DROBJ *old_parent_obj;
    byte  *path;
    byte  *filename;
    byte fileext[4];
    int new_driveno;
    DROBJ *new_obj;
    DROBJ *new_parent_obj;
    BOOLEAN ret_val;
    CLUSTERTYPE  cluster;
    DDRIVE *pdrive;

    int  p_errno;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */
    p_errno = 0;

    /* Drives must be the same */
    if (    !pc_parsedrive( &old_driveno, old_name ) ||
            !pc_parsedrive( &new_driveno, new_name ) ||
            old_driveno != new_driveno)
    {
        rtfs_set_errno(PEINVALIDDRIVEID);
        return(FALSE);
    }

    /* Get the drive and make sure it is mounted   */
    old_driveno = check_drive_name_mount(old_name);
    if (old_driveno < 0)
    {
        /* errno was set by check_drive */
        return(FALSE);
    }
    rtfs_set_errno(0);  /* pc_mv: clear error status */

    old_obj         = 0;
    old_parent_obj  = 0;
    new_obj         = 0;
    new_parent_obj  = 0;
    ret_val = FALSE;

    pdrive = pc_drno2dr(old_driveno);

    /* Use the buffers in the DRIVE structure. Access is locked via semaphore */
    path = (byte *)&(pdrive->pathname_buffer[0]);
    filename = (byte *)&(pdrive->filename_buffer[0]);

    /* Get out the filename and d:parent */
    if (!pc_parsepath(path, filename,fileext,old_name))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }
    /* Find the parent and make sure it is a directory  */
    old_parent_obj = pc_fndnode(path);
    if (!old_parent_obj)
        goto errex; /* pc_fndinode - set errno */

    if (!pc_isadir(old_parent_obj))
    {
        p_errno = PENOENT;
        goto errex;
    }
    /* Find the file */
    old_obj = pc_get_inode(0, old_parent_obj,
        filename, (byte*)fileext, GET_INODE_MATCH);
    if (!old_obj)
        goto errex; /* pc_get_inode - set errno */

     /* Be sure it exists and is a normal directory or file and is not open */
    if (pc_isroot(old_obj) || (old_obj->finode->opencount > 1) ||
       (old_obj->finode->fattribute&(ARDONLY|AVOLUME)))
    {
        p_errno = PEACCES;
        goto errex;
    }

    /* At this point old_obj contains the file we are renaming */

    /* See if the new directory entry already exists */
    new_obj = pc_fndnode(new_name);
    if (new_obj)
    {
        p_errno = PEEXIST;
        goto errex;
    }
    rtfs_set_errno(0);  /* pc_mv - clear errno condition after failed pc_fndnode */

    /* Get out the filename and d:parent */
    if (!pc_parsepath(path,filename,fileext,new_name))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }
    /* Find the parent and make sure it is a directory  */
    new_parent_obj = pc_fndnode(path);
    if (!new_parent_obj || !pc_isadir(new_parent_obj) ||  pc_isavol(new_parent_obj))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }


    /* The cluster value old */
    cluster = pc_finode_cluster(old_obj->pdrive,old_obj->finode);

    /* Create the new entry and assign cluster to it. If it is a directory
       .. will be linked correctly */
    /*-------------------- twl modified --------------------*/
    if( old_obj->finode->fattribute & ADIRENT) {
        new_obj = pc_mknode_ex( new_parent_obj, filename, fileext, old_obj->finode->fattribute, cluster);
    }else{
        new_obj = pc_mknode( new_parent_obj, filename, fileext, old_obj->finode->fattribute, cluster);
    }
    /*------------------------------------------------------*/

    if (!new_obj)
        goto errex;

    /* Copy the old directory entry stuf over */
    new_obj->finode->fattribute = old_obj->finode->fattribute;
    new_obj->finode->ftime = old_obj->finode->ftime;
    new_obj->finode->fdate = old_obj->finode->fdate;
    new_obj->finode->fsize = old_obj->finode->fsize;

    /* Update the new inode. Do not set archive bit or change date */
     if (!pc_update_inode(new_obj, FALSE, FALSE))
        goto errex;

    /* Set the old cluster value to zero */
    pc_pfinode_cluster(old_obj->pdrive,old_obj->finode,0);
    /* Delete the old but won't delete any clusters */
    if (!pc_rmnode(old_obj))
        goto errex;

    p_errno = 0;
    ret_val = TRUE;

    /* Good conditions fall through here, error exits jump to here */
errex:
    if (old_parent_obj)
        pc_freeobj(old_parent_obj);
    if (old_obj)
        pc_freeobj(old_obj);
    if (new_parent_obj)
        pc_freeobj(new_parent_obj);
    if (new_obj)
        pc_freeobj(new_obj);
    /* Set errno if we have one and not set by lower level already */
    if ((p_errno) && !get_errno())
        rtfs_set_errno(p_errno);
    if (!release_drive_mount_write(old_driveno))/* Release lock, unmount if aborted */
        ret_val = FALSE;
    return(ret_val);
}
#endif


/* fBNgt@CɂȂ^C~O݂Ȃpc_mkdir */
BOOLEAN pc_mkdir_ex(byte  *name)    /*__apifn__*/
{
    DROBJ *pobj;
    DROBJ *parent_obj;
    byte  *path;
    byte  *filename;
    byte  fileext[4];
    BOOLEAN  ret_val;
    int driveno;
    DDRIVE *pdrive;
    int p_errno;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */

    parent_obj = 0;
    pobj = 0;
    p_errno = 0;
    ret_val = FALSE;
    rtfs_set_errno(0);  /* pc_mkdir: clear error status */

    /* Get the drive and make sure it is mounted   */
    driveno = check_drive_name_mount(name);
    if (driveno < 0)
    {
        /*  errno was set by check_drive */
        return(FALSE);
    }

    pdrive = pc_drno2dr(driveno);

    /* Use the buffers in the DRIVE structure. Access is locked via semaphore */
    path = (byte *)&(pdrive->pathname_buffer[0]);
    filename = (byte *)&(pdrive->filename_buffer[0]);

    /* Get out the filename and d:parent   */
    if (!pc_parsepath(path,filename,fileext,name))
    {
        p_errno = PEINVALIDPATH;
        goto errex;
    }
    /* Find the parent and make sure it is a directory \   */
    /* pc_fndnode set errno */
    parent_obj = pc_fndnode(path);
    if (!parent_obj)
        goto errex;

    /* Lock the parent      */
    if (!pc_isadir(parent_obj) ||  pc_isavol(parent_obj))
    {
        p_errno = PENOENT;      /* Path is not a directory */
        goto errex;
    }
    /* Fail if the directory exists   */
    pobj = pc_get_inode(0, parent_obj, filename,(byte*) fileext, GET_INODE_MATCH);
    if (pobj)
    {
        p_errno = PEEXIST;      /* Exclusive fail */
        goto errex;
    }
    else
    {
        if (get_errno() != PENOENT) /* If pobj is NULL we abort on abnormal errors */
            goto errex;
        rtfs_set_errno(0);  /* pc_mkdir: clear PENOENT error status */

        /*---------- twl modified ----------*/
        pobj = pc_mknode_ex( parent_obj,filename, fileext, ADIRENT, 0);
        /*----------------------------------*/

        if (pobj)
        {
            p_errno = 0;
            ret_val = TRUE;
        }
        else
        {
            /* pc_mknode has set errno */
            goto errex;
        }
    }

errex:
    if (pobj)
        pc_freeobj(pobj);
    if (parent_obj)
    {
        pc_freeobj(parent_obj);
    }
    if (p_errno)
        rtfs_set_errno(p_errno);
    
    /*---------- ctr modified ----------*/
    if( pdrive) {
        if( pdrive->dev_table_perform_device_ioctl) {
            pdrive->dev_table_perform_device_ioctl( driveno,
                                                    DEVCTL_FLUSH,
                                                    NULL);
        }
    }
    /*----------------------------------*/
    
    if (!release_drive_mount_write(driveno))/* Release lock, unmount if aborted */
        ret_val = FALSE;
    return(ret_val);
}

/* Make a node from path and attribs create and fill in pobj             */
/* Note: the parent directory is locked before this routine is called    */
DROBJ *pc_mknode_ex(DROBJ *pmom ,byte *filename, byte *fileext, byte attributes, CLUSTERTYPE incluster) /*__fn__*/
{
    DROBJ *pobj;
    BOOLEAN ret_val;
    DOSINODE *pdinodes;
    FINODE lfinode;
    CLUSTERTYPE cluster;
    CLUSTERTYPE cltemp;
    DATESTR crdate;
    BLKBUFF *pbuff;
    DDRIVE *pdrive;
    byte attr;
#if (RTFS_SUBDIRS)
    byte dot_str[4]; /* Make DOT and DOTDOT strings */
    byte null_str[4];
#endif

    ret_val = TRUE;

    if (!pmom || !pmom->pdrive)
    {
        rtfs_set_errno(PEINTERNAL); /* pc_mknode: Internal error*/
        return (0);
    }
    /* Make sure the file/directory name is legal */
    if (!validate_filename(filename, fileext))
    {
        rtfs_set_errno(PEINVALIDPATH);
        return (0);
    }


    pdrive = pmom->pdrive;
    cluster = incluster;
#if (RTFS_SUBDIRS)
    if (attributes & ADIRENT)
    {
        /*Unless renaming a directory we grab a cluster for a new dir and */
        /* clear it   */
        if (!incluster)
        {
            /* pc_alloc_dir will set errno */
            cluster = pc_alloc_dir(pdrive,pmom);
            if (!cluster)
                ret_val = FALSE;
            else if (!pc_clzero( pdrive , cluster ) )
            {
                /* clzero will set errno */
                FATOP(pdrive)->fatop_clrelease_dir(pdrive , cluster);
                ret_val = FALSE;
            }
        }
    }
#endif

    if (!ret_val)
    {
        return (0);
    }

    /* For a subdirectory. First make it a simple file. We will change the
        attribute after all is clean */
    attr = attributes;
#if (RTFS_SUBDIRS)
    /*---------- twl modified ----------*/
//    if (attr & ADIRENT)
//        attr = ANORMAL;
    /*----------------------------------*/
#endif
    /* Allocate an empty DROBJ and FINODE to hold the new file   */

    pobj = pc_allocobj();
    if (!pobj)
    {
        return (0);
    }
    /* Set up the drive link */
    pobj->pdrive = pmom->pdrive;

    /*---------- twl modified ----------*/
    //Ƃpc_insert_inodeꏊ
    /*----------------------------------*/

#if (RTFS_SUBDIRS)
    /* Now if we are creating subdirectory we have to make the DOT and DOT DOT
        inodes and then change pobj s attribute to ADIRENT
        The DOT and DOTDOT are not buffered inodes. We are simply putting
        the to disk  */
    if ( attributes & ADIRENT)
    {
        if (incluster)
        {
            pbuff = pc_read_blk( pdrive , pc_cl2sector(pdrive , incluster));
            if (!pbuff)
            {
                pc_freeobj(pobj);
                return (0);
            }
        }
        else
        {
            /* Set up a buffer to do surgery   */
            pbuff = pc_init_blk( pdrive , pc_cl2sector(pdrive , cluster));
            if (!pbuff)
            {
                pc_freeobj(pobj);
                FATOP(pdrive)->fatop_clrelease_dir(pdrive , cluster);
                return (0);
            }
        }
        pdinodes = (DOSINODE *) &pbuff->data[0];
        /* Load DOT and DOTDOT in native form                     */
        /* DOT first. It points to the begining of this sector    */
        dot_str[0] = CS_OP_ASCII('.');dot_str[1]=CS_OP_ASCII('\0');
        null_str[0] = CS_OP_ASCII('\0');

        /* Load the time and date stamp to be used for "." and ".."
           from the subdirectory that we are creating. In this way
           the three directory entries will all have the same timestamp
           value */

        /*---------- twl modified ----------*/
#if 1   //쐬fBNggpc_insert_inodeOȂ̂ŎRs[łȂߎ擾
        pc_getsysdate(&crdate);
#else
        crdate.time = pobj->finode->ftime;
        crdate.date = pobj->finode->fdate;
#endif
        /*----------------------------------*/
      
        pc_init_inode( &lfinode, dot_str,null_str, ADIRENT/*|ARCHIVE*/, //twl modified
                            cluster, /*size*/ 0L , &crdate);
        /* And to the buffer in intel form   */
        pc_ino2dos (pdinodes, &lfinode);

        /*---------- twl modified ----------*/
        //쐬fBNggpc_insert_inodeOȂ̂pc_get_parent_clusteȑOōs
        pobj->blkinfo.my_frstblock = pc_firstblock(pmom);
        /*----------------------------------*/

        /* Now DOTDOT points to mom s cluster   */
        cltemp = pc_get_parent_cluster(pdrive, pobj);
        dot_str[0] = CS_OP_ASCII('.');dot_str[1] = CS_OP_ASCII('.');
        dot_str[2] = CS_OP_ASCII('\0');
        null_str[0] = CS_OP_ASCII('\0');
        pc_init_inode( &lfinode, dot_str, null_str, ADIRENT/*|ARCHIVE*/, cltemp, //twl modified
                    /*size*/ 0L , &crdate);
        /* And to the buffer in intel form   */
        pc_ino2dos (++pdinodes, &lfinode );
        /* Write the cluster out   */
        if ( !pc_write_blk ( pbuff ) )
        {
            pc_discard_buf(pbuff);
            pc_freeobj(pobj);
            if (!incluster)
                FATOP(pdrive)->fatop_clrelease_dir(pdrive , cluster);
            return (0);
        }
        else
            pc_release_buf(pbuff);


      
    /* Load the inode copy name,ext,attr,cluster, size,datetime  */
    /* Convert pobj to native and stitch it in to mom   */
    if (!pc_insert_inode(pobj , pmom, attr, cluster, filename, fileext))
    {
        if (cluster && !incluster)
            FATOP(pdrive)->fatop_clrelease_dir(pdrive , cluster);
        pc_freeobj(pobj);
        return (0);
    }

#if 0   /*---------- twl modified ----------*/
        /* And write the node out with the original attributes   */
        pobj->finode->fattribute = (byte)(attributes/*|ARCHIVE*/);  //twl modified

        /* Convert to native. Overwrite the existing inode.Set archive/date  */
//        if (!pc_update_inode(pobj, TRUE, TRUE))
        if (!pc_update_inode(pobj, FALSE, TRUE))  //twl modified
        {
            pc_freeobj(pobj);
            if (!incluster)
                FATOP(pdrive)->fatop_clrelease_dir(pdrive , cluster);
            return (0);
        }
#endif  /*----------------------------------*/
    }
#endif
    ret_val = FATOP(pdrive)->fatop_flushfat(pdrive->driveno);

    if (ret_val)
    {
        return (pobj);
    }
    else
    {
        pc_freeobj(pobj);
        return (0);
    }
}

#if 1
BOOLEAN twfs_pc_set_properties( byte* path, byte attributes, DATESTR* st_mtime)
{
    DROBJ  *pobj;
    BOOLEAN ret_val;
    int driveno;
    CHECK_MEM(BOOLEAN, 0)   /* Make sure memory is initted */

    rtfs_set_errno(0);  /* pc_set_attributes: clear error status */
    if ((attributes&(0x40|0x80)) !=0 ) /* Illegal */
    {
        rtfs_set_errno(PEINVALIDPARMS);
        return(FALSE);
    }
    driveno = check_drive_name_mount(path);
    if (driveno < 0)
    {  /* errno was set by check_drive */
        return (FALSE);
    }
    ret_val = FALSE;
    /* pc_fndnode will set errno */
    pobj = pc_fndnode(path);
    if (pobj)
    {
        /*---------- twl modified ----------*/
        if( pobj->isroot) {
            rtfs_set_errno(PEACCES);
        }else{
        /*----------------------------------*/

            /* Change the time */
            if( st_mtime) {
                pobj->finode->ftime = st_mtime->time;
                pobj->finode->fdate = st_mtime->date;
            }
      
            /* Change the attributes if legal   */
            if (
                  (attributes & (AVOLUME|ADIRENT)) ==   /* Still same type */
                  (pobj->finode->fattribute & (AVOLUME|ADIRENT)))
            {
                pobj->finode->fattribute = attributes;
                /* Overwrite the existing inode. Do not set archive/date  */
                /* pc_update_inode() will set errno */
                ret_val = pc_update_inode(pobj, FALSE, FALSE);
            }
            else
            {
                rtfs_set_errno(PEACCES);
            }

        /*---------- twl modified ----------*/
        }
        /*----------------------------------*/
        pc_freeobj(pobj);
    }
    if (!release_drive_mount_write(driveno))/* Release lock, unmount if aborted */
        ret_val = FALSE;
    return (ret_val);

}
#endif
