/*---------------------------------------------------------------------------*
  Project:  Revolution WFS library
  File:     wfs_Server.h - Server API

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Server.h,v $
  Revision 1.4  2008/01/10 05:33:20  nakanose_jin
  validatePathName => validateDirectory

  Revision 1.3  2007/11/02 00:21:56  paul
  started logging changes in the file

*---------------------------------------------------------------------------*/

#ifndef __WFS_SERVER_H__
#define __WFS_SERVER_H__

#include <wfs.h>

WFSResult WFSSRVInit(void *pWorkSpaceBuffer, u32 nWorkSpaceSize);
WFSResult WFSSRVGetDeviceInfo(const utf8 *sDeviceName, WFSDeviceInfo *pDi);
WFSResult WFSSRVMountDevice(const utf8 *sDeviceName, utf8 *sVolumeIdBuffer);
WFSResult WFSSRVUnmountVolume(const utf8 *sVolumeId);
WFSResult WFSSRVInitializeDevice(const utf8 *sDeviceName);
WFSResult WFSSRVGetVolumeId(const utf8 *sDeviceName, utf8 *sVolumeId);
WFSResult WFSSRVGetFreeSpaceSize(const utf8 *sPathName, u64 *pSize);
WFSResult WFSSRVCreateDirectory(const utf8 *sPathName, u32 nFlags, u64 nQuota);
WFSResult WFSSRVCreateAndOpenFile(const utf8 *sPathName, u32 nFlags, u32 nPreAllocateSize, WFSFileHandle *pFh);
WFSResult WFSSRVGetAttributes(const utf8 *sPathName, WFSFileAttributes *pFa);
WFSResult WFSSRVChangePermissions(const utf8 *sPathName, u32 nFlags);
WFSResult WFSSRVDelete(const utf8 *sPathName);
WFSResult WFSSRVMoveLocal(const utf8 *sSrcPathName, const utf8 *sDstPathName);
WFSResult WFSSRVOpenFile(const utf8 *sPathName, WFSAccess nDesiredAccess, WFSFileHandle *pFh);
WFSResult WFSSRVReadFile(WFSFileHandle fh, void *pFileDataBuffer, u32 nSize);
WFSResult WFSSRVWriteFile(WFSFileHandle fh, const void *pFileData, u32 nSize);
WFSResult WFSSRVCloseFile(WFSFileHandle fh, u8 bTruncateFlag);
WFSResult WFSSRVSearchDirectoryFirst(const utf8 *sPattern, WFSSearchDirectoryHandle *pSdh, WFSFileInfo *pFi);
WFSResult WFSSRVSearchDirectoryNext(WFSSearchDirectoryHandle sdh, WFSFileInfo *pFi);
WFSResult WFSSRVSearchDirectoryClose(WFSSearchDirectoryHandle sdh);
WFSResult WFSSRVDebugSetTitleID(u64 nTID);
u64       WFSSRVDebugGetTitleID();



// Functions in the WFS API that are NOT in the WFS Server API:

// void WFSSetAttachDeviceCallback(utf8 *sDeviceNameBuffer, WFSDeviceCallback cbAttachDevice, void *pUserData);
// void WFSGetAttachDeviceCallback(utf8 **psDeviceName, WFSDeviceCallback *pCbAttachDevice, void **ppUserData);
// void WFSSetDetachDeviceCallback(utf8 *sDeviceNameBuffer, WFSDeviceCallback cbDetachDevice, void *pUserData);
// void WFSGetDetachDeviceCallback(utf8 **psDeviceName, WFSDeviceCallback *pCbDetachDevice, void **ppUserData);

// WFSResult WFSSetCurrentDirectory(const utf8 *sPathName);
// This function is replaced with:
WFSResult WFSSRVValidateDirectory(const utf8 *sPathName);



// WFSResult WFSGetCurrentDirectory(utf8 *sAbsPathNameBuffer, s32 nBufferSize);
// WFSResult WFSSetFilePosition(WFSFileHandle fh, WFSFileSize nFilePosition);
// WFSResult WFSGetFilePosition(WFSFileHandle fh, WFSFileSize *pFilePosition);
// WFSResult WFSGetFileSize(WFSFileHandle fh, WFSFileSize *pFileSize);




#endif //define __WFS_SERVER_H__
