/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_AsyncUtils.h - mutual exclusion for multi-thread case

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_AsyncUtils.h,v $
  Revision 1.6  2008/04/19 05:52:50  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.5  2007/11/09 01:19:54  paul
  *** empty log message ***

  Revision 1.4  2007/11/02 00:25:49  paul
  started logging changes in the file
 *---------------------------------------------------------------------------*/

#ifndef wfs_AsyncUtils_h
#define wfs_AsyncUtils_h

#include "wfs_PathNames.h"
#include "wfs_Mutex.h"

#ifdef _DEBUG
    #define _DEBUG_ASYNC_UTILS 0  // change to 1 if debug info required
#else
    #define _DEBUG_ASYNC_UTILS 0
#endif

#define WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(ptr,Type,var) *((Type*)ptr)=var; ptr=(void*)((size_t)pPtr + sizeof(Type))
#define WFSCLI_WRITE_ASYNC_PARAM(ptr,Type,var) *((Type*)ptr)=var
#define WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(ptr,Type,var) Type var=*((Type*)ptr); ptr=(void*)((size_t)pPtr + sizeof(Type))
#define WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(ptr,Type,var) Type var=*((Type*)ptr);
#define WFSCLI_READ_ASYNC_PARAM_AND_INC_PTR(ptr,Type,var) var=*((Type*)ptr); ptr=(void*)((size_t)pPtr + sizeof(Type))
#define WFSCLI_READ_ASYNC_PARAM(ptr,Type,var) var=*((Type*)ptr)


typedef struct WFSAsyncParamHdr       WFSAsyncParamHdr;
typedef struct WFSAsyncParamHdrAnchor WFSAsyncParamHdrAnchor;
typedef struct WFSAsyncParamHdrLink   WFSAsyncParamHdrLink;
typedef struct WFSAsyncParamGlobals   WFSAsyncParamGlobals;

struct WFSAsyncParamHdrLink {
    WFSAsyncParamHdr *pNext;
    WFSAsyncParamHdr *pPrev;
};

struct WFSAsyncParamHdrAnchor {
    WFSAsyncParamHdr *pHead;
    WFSAsyncParamHdr *pTail;
};

struct WFSAsyncParamHdr  {
    WFSAsyncParamHdrLink l;
    u32 nSize;  // Size of parameters including this header
    void *cb;
    void *pUserData;
    u8 nFunction;   // Values are defined in wfssrv_Defs.h
};

struct WFSAsyncGlobals {
    WFSAsyncParamHdrAnchor anchor;
    WFSThread asyncExecThread;
    void *pAsyncExecStack;
    WFSMutex mxAccessTheParamQueue;
    WFSCond condParamQueueEmpty;
    WFSCond condParamQueueNotEmpty;
};


void WFSAsyncUtilsInit();

void WFSAsyncUtilsClose();

WFSResult WFSPrepareAsyncCall(void **ppPtr, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize);

WFSResult WFSPrepareAsyncCallWithPathName(void **ppPtr, const utf8 *sInPathName,
    WFSPathNameType nPathType, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize);

WFSResult WFSPrepareMoveLocalAsyncCall(void **ppPtr, const utf8 *sSrcPathName, const utf8 *sDstPathName,
    WFSPathNameType nPathType, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize);

WFSResult WFSPrepareAsyncCallWithDeviceName(void **ppPtr, const utf8 *sDeviceName,
    u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize);

WFSResult WFSPrepareAsyncCallWithVolumeId(void **ppPtr, const utf8 *sVolumeId,
    u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize);

#endif //define wfs_AsyncUtils_h
