/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Config.h - project-wide configuration settings governing resource allocations etc.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Config.h,v $
  Revision 1.8  2008/11/04 06:07:16  ueno
  Changed the maximum number of file handles and search handles to 32.

  Revision 1.7  2008/09/29 01:06:04  kondo_masahiro
  Added several macros.

  Revision 1.6  2008/07/09 02:52:33  paul
  Increased maximum number of File & Directory Search handles

  Revision 1.5  2008/05/12 20:30:22  paul
  Added handle defs

  Revision 1.4  2007/11/05 18:38:42  paul
  WFSDEV flag is now defined in makefile / projectfile

  Revision 1.3  2007/11/02 00:25:29  paul
  started logging changes in the file
 *---------------------------------------------------------------------------*/

#ifndef wfs_Config_h
#define wfs_Config_h


// Define WFSDEV in the makefile/project settings to compile the WFS DEV version of the client
#ifdef WFSDEV
    #define _WFSDEV  1
#else
    #define _WFSDEV  0
#endif

// This determines the maximum number of asynchronous calls that can be outstanding before a busy error occurs
//#define WFS_MAX_OUTSTANDING_ASYNC_CALLS  4
// Now it is determined by the amount of memory the user supplies with WFSInit

#define WFSCLI_WORK_SPACE_SIZE    (40*1024)
#define WFS_ASYNC_EXEC_STACK_SIZE (16*1024)          // client common code async handler thread

// WFSDEV-specific
#define WFSDEV_CALLBACK_THREAD_STACK_SIZE (16*1024) // client thread for server-initiated callbacks; for WFSDEV implementation
#define WFSDEV_MAX_CLIENTS                (5)       // maximum number of clients that the server can support simultaneously

#define WFSCLI_FILE_HANDLE_IDX_BITS           5
#define WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS     5

#endif //wfs_Config_h
