/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Heap.h - memory allocation functions

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Heap.h,v $
  Revision 1.8  2008/04/19 05:52:37  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.7  2007/12/26 04:26:41  paul
  changed pAllocator to &allocator

  Revision 1.6  2007/12/26 04:13:56  paul
  Changed RVL version of WFSHeap

  Revision 1.5  2007/12/26 00:19:25  paul
  Added WFSHeapDestroy()

  Revision 1.4  2007/11/02 00:23:58  paul
  started logging changes in the file
 *---------------------------------------------------------------------------*/

#ifndef wfs_Heap_h
#define wfs_Heap_h


#include "wfs_Defs.h"

#ifdef _DEBUG
    #define _DEBUG_HEAP 0  // change to 1 if debug info required
#else
    #define _DEBUG_HEAP 0
#endif

#if _WIN32

typedef HANDLE WFSHeap;

#elif _TWL

#define WFS_HEAP_DEFAULT_ALIGNMENT 32

typedef OSHeapHandle WFSHeap;

#elif _RVL

#include <revolution/mem.h>
#define WFS_HEAP_DEFAULT_ALIGNMENT 32

typedef struct {
    MEMAllocator  allocator;
    MEMHeapHandle handle;
} WFSHeap;

#elif _IOP

#define WFSHeap IOSHeapId

#endif

void WFSCreateHeap(WFSHeap *pHeap, void *pBase, size_t nSize);
void *WFSHeapAlloc(WFSHeap *pHeap, size_t nSize);
void WFSHeapFree(WFSHeap *pHeap, void *pAddr);
void WFSHeapDestroy(WFSHeap *pHeap);

#endif //define wfs_Heap_h
