/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Names.h - global name types used by the WFS project

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Names.h,v $
  Revision 1.5  2008/02/25 05:32:50  paul
  Reverted to previous version - changed defs instead.

  Revision 1.3  2007/11/06 22:43:48  paul
  changed include

  Revision 1.2  2007/11/02 00:22:34  paul
  started logging changes in the file

 *---------------------------------------------------------------------------*/

#ifndef wfs_Names_h
#define wfs_Names_h

#include "wfs_Defs.h"


#if (WFS_MAX_PATH_NAME_SIZE<256)
    typedef u8 WFSPathNameLengthType;
#elif (WFS_MAX_PATH_NAME_SIZE<65536)
    typedef u16 WFSPathNameLengthType;
#else
    typedef u32 WFSPathNameLengthType;
#endif

typedef struct {
    WFSPathNameLengthType nLen;  // This is the length of the string not including the NULL terminator
    utf8 sStr[WFS_MAX_PATH_NAME_SIZE+1];
} WFSPathName;

WFSResult WFSSetPathName(WFSPathName *pPathName, const utf8 *sPathName);


#if (WFS_MAX_FILE_NAME_SIZE<256)
    typedef u8 WFSFileNameLengthType;
#elif (WFS_MAX_FILE_NAME_SIZE<65536)
    typedef u16 WFSFileNameLengthType;
#else
    typedef u32 WFSFileNameLengthType;
#endif

typedef struct {
    WFSFileNameLengthType nLen;  // This is the length of the string not including the NULL terminator
    utf8 sStr[WFS_MAX_FILE_NAME_SIZE+1];
} WFSFileName;

WFSResult WFSSetFileName(WFSFileName *pFileName, const utf8 *sFileName);


#if (WFS_MAX_DEVICE_NAME_SIZE<256)
    typedef u8 WFSDeviceNameLengthType;
#elif (WFS_MAX_DEVICE_NAME_SIZE<65536)
    typedef u16 WFSDeviceNameLengthType;
#else
    typedef u32 WFSDeviceNameLengthType;
#endif

typedef struct {
    WFSDeviceNameLengthType nLen;  // This is the length of the string not including the NULL terminator
    utf8 sStr[WFS_MAX_DEVICE_NAME_SIZE+1];
} WFSDeviceName;

WFSResult WFSSetDeviceName(WFSDeviceName *pDeviceName, const utf8 *sDeviceName);


#if (WFS_MAX_VOLUME_ID_SIZE<256)
    typedef u8 WFSVolumeIdLengthType;
#elif (WFS_MAX_VOLUME_ID_SIZE<65536)
    typedef u16 WFSVolumeIdLengthType;
#else
    typedef u32 WFSVolumeIdLengthType;
#endif

typedef struct {
    WFSVolumeIdLengthType nLen;  // This is the length of the string not including the NULL terminator
    utf8 sStr[WFS_MAX_VOLUME_ID_SIZE+1];
} WFSVolumeId;

WFSResult WFSSetVolumeId(WFSVolumeId *pVolumeId, const utf8 *sVolumeId);


#endif //define wfs_Names_h
