/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_PathNames.h - Functions for concatenating relative and absolute pathnames


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_PathNames.h,v $
  Revision 1.10  2008/08/19 14:05:40  nakanose_jin
  smash pathnames bug

  Revision 1.9  2008/04/19 20:44:58  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.8  2007/12/10 02:50:11  paul
  Added debug output

  Revision 1.7  2007/11/09 01:20:52  paul
  Added a new path name type to allow validation of path name depth for paths ending in directories

  Revision 1.6  2007/11/02 01:09:26  paul
  updated path type

  Revision 1.5  2007/11/02 00:20:16  paul
  started logging changes in the file

 *---------------------------------------------------------------------------*/

// File created by Paul Donnelly on 2007.10.05

#ifndef wfs_PathNames_h
#define wfs_PathNames_h

#ifdef _DEBUG
    #define _DEBUG_PATHNAMES 0  // change to 1 locally if debug info required. (should be 0 for checkin)
#else
    #define _DEBUG_PATHNAMES 0
#endif

#include "wfs_Mutex.h"

typedef enum {
    // Initial states
    PATH_TYPE_REL_OR_ABS=0,   // Path name can be a relative or absolute file or directory
    PATH_TYPE_REL_OR_ABS_DIR, // Path name can be a relative or absolute directory
    PATH_TYPE_ABS_DIR_ONLY,   // Path name must be an absolute directory
    PATH_TYPE_PATTERN         // Path name (pattern) can be relative or absolute and may include wild cards in file name part.
} WFSPathNameType;

typedef struct {
    u16  nNumParts;
    WFSPathNameLengthType  aPart[WFS_MAX_PATH_DEPTH+1];
} WFSPathPartOffsetArray;

typedef struct {
    WFSPathPartOffsetArray ppoa;
    WFSPathName pn;
} WFSPath;

typedef struct {
    WFSMutex mxAccessPathGlobals;
    WFSPath pathHome, pathCurrent;
} WFSPathGlobals;

void WFSPathInit();

WFSResult WFSPathResolve(WFSPathName *pPnOutAbs, WFSPathPartOffsetArray *pPpoaOutAbs, const utf8 *sInPathName, WFSPathNameType nPathType);

WFSResult WFSPathValidate(WFSPath *pPath);

void WFSPathCopy(WFSPathName *pPnDst, WFSPathPartOffsetArray *pPpoaDst, WFSPathName *pPnSrc, WFSPathPartOffsetArray *pPpoaSrc, bool bAppendSeparator, bool bNullTermination);

#endif //define wfs_PathNames_h
