/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Platform.h - low level definitions and types that are primarily platform dependent

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Platform.h,v $
  Revision 1.13  2008/08/28 00:20:35  nakanose_jin
  lets standard exception style

  Revision 1.11  2008/08/27 04:50:50  nakanose_jin
  add function and defines

  Revision 1.10  2008/07/17 05:56:25  kondo_masahiro
  Added #define SNPRINTF

  Revision 1.9  2008/04/24 23:19:15  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.8  2008/04/22 00:59:59  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.6  2007/11/21 04:17:28  wayne.wong
  Many changes that included single request/result RPC messages, bug fixes, code factoring, and English C comment adjustments.
    1. Combined RPC messages into a single request message and single result
        message. Except for file reads and writes.
    2. Fixed+factored some Endianess code.
    3. Enabled larger TCP/IP data transfer size. Previous setting was artifically small
        to test the code to perform large transfers.
    4. Changed the volume id parameters utf8*->WFSVolumeId for SrvGetVolumeId
        and SrvMountDevice. More consistent with the way we send this type at the
        RPC layer.
    5. Changed a couple Japanese comments to be compatible with English
        comments. In particular, the "//" comment with a backslash  at the end of
        the line causes the next line of code to be ignored.
    6. Removed Windows-specific debugging output defines for RVL build.

  Revision 1.5  2007/11/17 03:05:38  paul
  removed #include "WinOSReport.h" from NDEV version

  Revision 1.4  2007/11/05 18:39:52  paul
  removed kernel-specific definitions

  Revision 1.3  2007/11/02 00:22:23  paul
  started logging changes in the file

*---------------------------------------------------------------------------*/

#ifndef wfs_Platform_h
#define wfs_Platform_h

#include <firm.h>

#define TWL 1
//--twl modified
#if 1
    #define _OSX   0
    #define _WIN32 0
    #define _TWL   1
    #define _RVL   0
    #define _IOP   0
#else
//--twl modified
#ifdef OSX
    #define _OSX   1
    #define _WIN32 0
    #define _RVL   0
    #define _IOP   0
#else
 #ifdef WIN32
    #define _OSX   0
    #define _WIN32 1
    #define _RVL   0
    #define _IOP   0
 #else
  #ifdef IOP
    #define _OSX   0
    #define _WIN32 0
    #define _RVL   0
    #define _IOP   1
  #else // RVL
    #define _OSX   0
    #define _WIN32 0
    #define _RVL   1
    #define _IOP   0
  #endif
 #endif
#endif
#endif //twl modified

#if _WIN32
    #pragma warning(disable:4996)
//    #define UNALIGNED                      __unaligned
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define OSReport printf
    #include "WinOSReport.h"
    #define OSPanic  MyOSPanic
    #define SNPRINTF _snprintf
#elif  _OSX
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define OSReport printf
    #include "WinOSReport.h"
    #define OSPanic  MyOSReport
    #define SNPRINTF snprintf
//-- twl modified
#elif  _TWL
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define MyOSReport                     osTPrintf
    #define InitMyOSReport()
    #define SNPRINTF OS_SNPrintf
    #define STRNLEN( str, maxlen ) stdStrLen(str)
//-- twl modified
#elif  _RVL
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define max(a,b)                       (((a)>(b))?(a):(b))
    #define min(a,b)                       (((a)<(b))?(a):(b))
    #define MyOSReport                     OSReport
    #define InitMyOSReport()
    #define SNPRINTF snprintf
#elif  _IOP
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define max(a,b)                       (((a)>(b))?(a):(b))
    #define min(a,b)                       (((a)<(b))?(a):(b))
    #define MyOSReport                     printf
    #define InitMyOSReport()
    #define SNPRINTF snprintf
#endif // WIN32

#ifndef BIG_ENDIAN
    #ifdef WIN32
        #define BIG_ENDIAN 0
    #elif TWL
        #define BIG_ENDIAN 0
    #else
        #define BIG_ENDIAN 1
    #endif // WIN32
#endif // BIG_ENDIAN


#endif //define wfs_Platform_h
