/*---------------------------------------------------------------------------*
  Project:  Revolution WFS library
  File:     wfs_Debug.h - debugging routines for WFS implementation

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Debug.h,v $
  Revision 1.16  2008/10/30 06:10:29  kondo_masahiro
  Added WFSDEBUG_CMD_AREA_AND_DIR_CHECK_DISK

  Revision 1.15  2008/10/14 03:00:34  nakanose_jin
  (none)

  Revision 1.14  2008/09/18 10:25:33  nakanose_jin
  add new WFSDEBUG_CMD_*

  Revision 1.13  2008/08/29 01:14:57  kondo_masahiro
  Added a flag WFSDEBUG_CMD_KRN_INIT_AREA_SECTORS.

  Revision 1.11  2008/06/26 00:21:30  kondo_masahiro
  Added #define WFSDEBUG_CMD_SHIM_WORKSPACE.

  Revision 1.10  2008/05/23 02:00:33  nakanose_jin
  add new command

  Revision 1.9  2008/05/10 04:20:22  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.8  2008/05/08 11:40:58  nakanose_jin
  move WFSDebugSetTitleID wfsPerm.h => wfs_Debug.h

  Revision 1.7  2008/05/07 13:27:12  nakanose_jin
  fix comment

  Revision 1.6  2008/04/25 12:06:38  nakanose_jin
  add cmd (0x2)

  Revision 1.5  2008/04/18 07:53:33  nakanose_jin
  merging acl

  Revision 1.4  2008/04/03 07:11:05  nakanose_jin
  add extern C

  Revision 1.3  2008/03/31 04:24:58  nakanose_jin
  add comments

  Revision 1.2  2008/03/31 00:58:52  nakanose_jin
  add args for debugproc

  Revision 1.1  2008/03/26 01:11:50  nakanose_jin
  WFS Debug functions for tool develpment

  Revision 1.6  2008/02/20 08:10:10  nakanose_jin
  add WFSExecDebugProcedure

  Revision 1.5  2007/12/17 13:29:14  nakanose_jin
  add debug utility function

  Revision 1.4  2007/11/02 00:25:10  paul
  started logging changes in the file

*---------------------------------------------------------------------------*/

#include <revolution/wfs.h>
// #include "wfs_Defs.h"

#ifndef __WFS_DEBUG_H__
#define __WFS_DEBUG_H__

#ifdef __cplusplus
extern "C" {
#endif

//--------------------------------------------------------------------------------------------------------------
// WFSDebugSetTitleId
//--------------------------------------------------------------------------------------------------------------
WFSResult WFSDebugSetTitleId(WFSTitleId titleId);
// Arguments:
// ----------
// titleId             .. Title Id is assigned per application. Determines the ownership & access permissions of files.
//                        A value of WFS_ROOT_ID can be specified for SYSTEM level access to all files.
// groupId             .. Group Id determines the GROUP access to files. Usually corresponds to a Company Id.
//
// Return Values:
// --------------
// Possible return values for WFSDebugSetTitleId:
//      WFS_RESULT_OK                  .. The search handle has been closed successfully.
//      WFS_RESULT_INVALID             .. The function parameters were invalid.
//
// Description:
// ------------
// Allows the application to specify it's WFS Title Id, which will be used to determine access permissions to WFS files.
// NOTE: This function is for testing and debugging purposes only, and will be removed from the final API.
// NOTE: This function only affects WFS. It does not change the real Title Id.
// ToDo: This function should be removed before releasing to developers.

/*
 * for debug purpose only.
 */
WFSResult WFSDebugSrvTest(const char* deviceName);
WFSResult WFSDebugCloseServerSocket(void);
WFSResult WFSDebugCloseCallbackSocket(void);

const char *WFSDebug_GetResultString(WFSResult result);
WFSResult WFSExecDebugProcedure( u32 cmd , void *arg0 , void *arg1 );
const static u32 WFSDEBUG_CMD_DELETE_DEVICEROOT          = 0;
const static u32 WFSDEBUG_CMD_DELETE_PATH                = 1;
const static u32 WFSDEBUG_CMD_CLOSE_ALLFILE              = 2; 
const static u32 WFSDEBUG_CMD_DUMP_PERMISSION            = 3;
const static u32 WFSDEBUG_CMD_SET_AUTO_FORMAT_ON_ATTACH  = 4;
const static u32 WFSDEBUG_CMD_ATTACH                     = 5;
const static u32 WFSDEBUG_CMD_ATTACH_WRITE_PROTECTED     = 6;
const static u32 WFSDEBUG_CMD_DETACH                     = 7;
const static u32 WFSDEBUG_CMD_PUSH_PATH                  = 8;
const static u32 WFSDEBUG_CMD_EXEC_BAT                   = 9; // 
const static u32 WFSDEBUG_CMD_TOUCH                      = 10;


//#define WFSDEBUG_CMD_KRN_INIT_VIRTUAL_DISK           0x10
#define WFSDEBUG_CMD_KRN_INIT_AREA_SECTORS           0x10
#define WFSDEBUG_CMD_KRN_INIT_CACHE                  0x11
#define WFSDEBUG_CMD_DIR_CHECK_DNS_EMPTY             0x12
#define WFSDEBUG_CMD_SHIM_WORKSPACE                  0x13
#define WFSDEBUG_CMD_AREA_AND_DIR_CHECK_DISK         0x14

#define WFSDEBUG_CMD_DEVICE_ATTACH_DETACH            0x20
#define WFSDEBUG_CMD_DEVICE_PSEUDO_DETACH            0x21
#define WFSDEBUG_CMD_DEVICE_PSEUDO_HASH_INCONSISTENT 0x22
#define WFSDEBUG_CMD_DEVICE_SET_FLAGS                0x23

/*

	WFSDEBUG_CMD_DELETE_DEVICEROOT
        brief: wpX폜
		arg0 : ^[QbgfoCX ( "msc1"  )
		arg1 :  -

	WFSDEBUG_CMD_DELETE_PATH
        brief: wpX폜(}EgłȂĂ悢)
		arg0 : ^[QbgfoCX ( "msc1"     )
		arg1 : ^[QbgpX *1 ( "/titles"  )

	WFSDEBUG_CMD_CLOSE_ALLFILE
        brief: I[vꂽ܂܂̃t@CSĕ܂B
		arg0 : ^[Qbg{[
		arg1 :  - 

    WFSDEBUG_CMD_DUMP_PERMISSION
        brief: wpX̃p[~bV\ (}EgłȂĂ悢)
		arg0 : ^[QbgfoCX ( "msc1"     )
		arg1 : ^[QbgpX *1 ( "/titles"  )

    WFSDEBUG_CMD_SET_AUTO_FORMAT_ON_ATTACH
        brief: A^b`ɃfoCXtH[}bgłΕKvȃVXet/Ȃ (default:)
		arg0 : true/false
		arg1 :  - 

    WFSDEBUG_CMD_ATTACH
    WFSDEBUG_CMD_ATTACH_WRITE_PROTECTED
        brief: wfoCXA^b`܂BWFSDevAttach.exe [-p] arg0 arg1 ƓłB
		arg0 : ^[QbgfoCX ( ex. "msc1")
		arg1 : z{[[g̕pX ( ex. "D:\iifstest\msc01" )

    WFSDEBUG_CMD_DETACH
        brief: wfoCXf^b`܂BWFSDevAttach.exe -d arg0 ƓłB
		arg0 : ^[QbgfoCX ( ex. "msc1")
		arg1 :  - 

	WFSDEBUG_CMD_EXEC_BAT
        brief: Cӂ̃R}hsB
               gpɍۂĂ͈S̗RKvŏɂƂǂ߂邱Ƃ𐄏܂B
               ob`t@C̑Ōďdvt@C肵Ȃ悤dXӉB
		arg0 : ^[Qbgst@C  ( ΃pXŵ݁Bex. "H:\\iifstest\\bin\\ready_testLongPathName.bat" )
		arg1 : ^[QbgspX    ( ΃pXŵ݁Bw薳(NULL)łnjBex. "H:\\iifstest\\bin" )
 

    *1 : {[[g[gƂ鉼zpX
*/

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif // __WFS_DEBUG_H__
