/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Api.h - global types used by the WFS kernel

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Api.h,v $
  Revision 1.13  2008/12/04 00:42:37  ooizumi
  Changed definition name to enable permission.

  Revision 1.12  2008/09/28 23:30:00  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.11  2008/08/29 00:13:19  kondo_masahiro
  Removed gPtr and gSize.

  Revision 1.10  2008/07/25 02:54:16  paul
  Added TransInfo lists to wkg

  Revision 1.9  2008/07/17 22:43:08  paul
  removed SubBlkTracker

  Revision 1.8  2008/07/17 22:14:25  paul
  Added rootDirAttr

  Revision 1.7  2008/07/17 04:37:45  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.6  2008/07/09 01:06:24  paul
  Added some definitions to support lists of mounted volumes. Moved block cache memory buffer into wkg.

  Revision 1.5  2008/06/09 18:16:25  paul
  added devAttr

  Revision 1.4  2008/05/12 19:19:05  paul
  Added SubBlkTracker

  Revision 1.3  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:28:32  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:07  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.6  2008/04/23 00:55:23  paul
  Updated WFSKrnGlobals

  Revision 1.5  2008/04/19 05:50:11  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.4  2008/02/29 02:38:02  kondo_masahiro
  Fixed code to accept IOP compiler

  Revision 1.3  2008/02/28 02:38:02  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.2  2008/02/27 07:27:11  paul
  made areaHdrRoot a global so it could be accessed from debug code in wfskrn_Dir.cpp

  Revision 1.1  2007/12/27 11:40:03  paul
  *** empty log message ***

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Api_h
#define wfskrn_Api_h

#include "revolution/wfs.h"
#include "wfskrn_Config.h"
#include "wfskrn_Defs.h"
#include "wfskrn_Heap.h"
#include "wfskrn_Handles.h"
#include "wfskrn_Area.h"
#include "wfskrn_Bcache.h"
#include "wfskrn_Volume.h"
#include "wfskrn_PathCache.h"
#include "wfskrn_PTree.h"
#include "wfskrn_Device.h"
#include "wfskrn_Volume.h"


typedef struct VolumeLink_ {
    struct VolumeInfo_ *pNext;
    struct VolumeInfo_ *pPrev;
} VolumeLink;

typedef struct VolumeInfo_ {
    VolumeLink             vl;
    struct DeviceInfo_    *pDevInfo;
    VolumeHdr              vh;
#if _DEBUG
    WFSVolumeId            volId;
#endif
    AreaInfo               rootAreaInfo;
    DirEntryAttrHdr        rootAttr;
    u32                    nClientsMountedBy;  // A bit vector. Up to 32 clients represented by a bit each.
    bool                   bDetached;
} VolumeInfo;


#define WFSKRN_MAX_VOLUMES        (WFSKRN_MAX_DEVICES + WFSKRN_MAX_MOUNTED_VOLUMES)
#define WFS_STATIC_MEM_TOTAL_SIZE (WFS_BLK_CACHE_SIZE + WFS_BCACHE_META_DATA_SIZE + WFS_STATIC_HEAP_SIZE + WFS_STATIC_MEM_ALIGNMENT)

typedef struct {
    u8                 *pBlkCache;
    u8                 *pHeapData;
    bool                bKrnInitialized;
    bool                bMountedDeviceWasDetached;
    WFSKrnHeap          heap;

    //WFSKrnStrHashTbl    strHTbl;
    PathCache           pathCache;
    PTreeHdr            ptreeHdr;
    PTreeAllocator      ptreeAllocator;

    WFSBlkAdr           nRootDirBlkAdr;

    u32                 nNumDevices;
    DeviceInfo          aDevInfo[WFSKRN_MAX_DEVICES];
    VolumeInfo          aVolInfo[WFSKRN_MAX_VOLUMES];
    VolumeLink          mountedVolListAnchor;
    VolumeInfo         *pMountedVolListAnchor;  // double link list
    VolumeInfo         *pFreeVolList;           // single link list (stack)

    DirEntryAttrHdr     rootDirAttr;
    DirEntryAttrHdr     volDirAttr;
    DirEntryAttrHdr     devDirAttr;
    DirEntryAttrHdr     devAttr;

#if _WIN32 || PERM_TEST
    WFSTitleId          nDebugTitleId;
#endif
    TransInfo           aTransInfo[WFSKRN_MAX_ACTIVE_TRANSACTIONS];
    TransInfoLink       activeTransListAnchor;
    TransInfo          *pActiveTransListAnchor;
    TransInfo          *pTransInfoFreeList;

    WFSKrnHandles       hnd;
    WFSFileTime         time;

    u8                  aStaticMem[WFS_STATIC_MEM_TOTAL_SIZE]; // it includes block cache for meta block and heap data.
} WFSKrnGlobals;


extern WFSKrnGlobals wkg;

extern DirItr         gDi;

extern AreaInfo *pAhInit;
extern WFSBlkAdr  *pBaInit;

//extern u8  *gpPtr; // Pointer of vdsk memory
//extern u32  gSize; // Size of vdsk memory

extern u8  *gpHeapPtr; // Pointer of heap memory 
extern u32  gHeapSize; // Size of heap memory

#endif //define wfskrn_Api_h
