/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Config.h - kernel configuration settings governing resource allocations etc.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Config.h,v $
  Revision 1.18  2008/12/03 00:03:46  kondo_masahiro
  Added WFS_MAX_USER_BLK_CACHE_SIZE and WFS_MIN_USER_BLK_CACHE_SIZE

  Revision 1.17  2008/11/05 15:05:11  ueno
  Reduced WFSKRN_BCACHE_HASH_TABLE_SIZE from 2957 to 47.

  Revision 1.16  2008/11/04 06:11:07  ueno
  Reduced WFS heap size to 8KB.
  Changed the maximum number of file handles and search handles to 32.

  Revision 1.15  2008/10/21 09:50:09  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.14  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.13  2008/10/08 23:20:09  kondo_masahiro
  Added codes to access large size file data

  Revision 1.12  2008/10/03 08:37:18  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.11  2008/09/28 23:30:18  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.10  2008/08/27 23:45:02  paul
  Changed name from WFS_LOG2_HASH_BLK_SIZE to WFS_LOG2_MAX_HASHABLE_BLK_SIZE

  Revision 1.9  2008/08/14 08:08:40  kondo_masahiro
  Removed several definitions related on wfskrn_Device.

  Revision 1.8  2008/08/05 04:17:25  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.7  2008/07/25 02:53:10  paul
  added WFSKRN_MAX_ACTIVE_TRANSACTIONS

  Revision 1.6  2008/07/17 22:19:50  paul
  changed comment

  Revision 1.5  2008/07/09 01:25:15  paul
  Changes related to block cache memory.

  Revision 1.4  2008/06/09 18:08:48  paul
  Increased maximum number of open directory searches to reflect the maximum path depth x 2 devices. Made corresponding change to size of path cache.

  Revision 1.3  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:28:26  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:07  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.9  2008/04/26 01:08:45  wayne.wong
  Moved BCache parameters to here.

  Revision 1.8  2008/04/23 00:36:39  paul
  added defines for file/directory search handles

  Revision 1.7  2008/04/19 05:50:00  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.6  2008/04/05 03:07:07  paul
  *** empty log message ***

  Revision 1.4  2008/02/20 07:16:15  paul
  *** empty log message ***

  Revision 1.3  2007/12/27 11:38:32  paul
  added various config settings

  Revision 1.2  2007/11/16 05:45:54  paul
  Added definitions for areas and regions

  Revision 1.1  2007/11/05 19:03:13  paul
  first checkin

 *---------------------------------------------------------------------------*/

#ifndef wfskrn_Config_h
#define wfskrn_Config_h

#include "wfscli_Handles.h"
#include "wfssrv_Defs.h"


#define WFS_VERSION_MAJOR                1  // version of WFS  (1..255)
#define WFS_VERSION_MINOR                0  // (0..255)

// Block size is not static - can be set within the following range per device / area

#define WFS_LOG2_SMALL_BLK_SIZE         13  // Log to the base 2 of the small block size.  (13 would correspond to 8KB)
#define WFS_LOG2_MEDIUM_BLK_SIZE        16  // Log to the base 2 of the medium block size. (16 corresponds to 64KB)
#define WFS_LOG2_LARGE_BLK_SIZE         19  // Log to the base 2 of the large block size.  (19 corresponds to 512KB)

#define WFS_LOG2_SMALL_BLKS_PER_REGION  12  // A region is a fixed size subdivision of an area, which is split into either all small blocks or all large blocks
#define WFS_MIN_REGIONS_PER_AREA        2   // one for small and one for large blks
#define WFS_LOG2_MIN_REGION_SIZE        (WFS_LOG2_SMALL_BLK_SIZE+WFS_LOG2_SMALL_BLKS_PER_REGION)  // log2(bytes)
#define WFS_MIN_REGION_SIZE             (1<<(WFS_LOG2_SMALL_BLK_SIZE+WFS_LOG2_SMALL_BLKS_PER_REGION))
#define WFS_LOG2_MIN_AREA_SIZE          (WFS_MIN_REGION_SIZE*WFS_MIN_REGIONS_PER_AREA)      // in bytes
#define WFS_MIN_AREA_SIZE               (1<<(WFS_LOG2_MIN_AREA_SIZE))                       // in bytes

#define WFS_LOG2_MAX_CACHE_STORED_FILE_BLK_SIZE   13

#define WFS_LOG2_MAX_HASHABLE_BLK_SIZE  WFS_LOG2_MEDIUM_BLK_SIZE

#define WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT  16  // Number of DirNodeStack structs to add to the pool when it runs out. Also used as initial allocation quantity. (42*12B == 504B). DirNodeStack entries are linked together in stacks and used for traversing directories.

#define WFS_LOG2_MAX_FILE_SIZE  32 // 4GB

#define WFS_STATIC_HEAP_SIZE            (8 * 1024)

//#define WFS_BLK_CACHE_SIZE              (512 * 1024)
#define WFS_BLK_CACHE_SIZE              (512 * 256)
#define WFS_BCACHE_META_DATA_SIZE       (  8 * 1024) // Cache 512[KB], Block 8[KB] -> Metadata 3840[B]
#define WFS_STATIC_MEM_ALIGNMENT         512

#define WFS_MAX_USER_BLK_CACHE_SIZE     (512 * 1024)
#define WFS_MIN_USER_BLK_CACHE_SIZE     (144 * 1024)

// Special in-memory area at the top of WFS to hold "/dev", "/vol" directories
#define WFS_LOG2_PATH_CACHE_SIZE        15

#if _WIN32
    #define WFS_LOG2_MAX_AREA_SIZE      29
#else
    #define WFS_LOG2_MAX_AREA_SIZE      34  // Max area size of 16GB
#endif

#define WFSKRN_MAX_ACTIVE_TRANSACTIONS   4
#define WFSKRN_MAX_DEVICES               1
#define WFSKRN_MAX_MOUNTED_VOLUMES       1
#define WFSKRN_MAX_NUM_AREA_SIZES       10 // at least 5 (64MB, 256MB, 1GB, 4GB, and 16GB)
#define WFSKRN_MIN_AREA_SIZE            (2*WFS_MIN_REGION_SIZE)  // 64MB = 32MB + 32MB

#define WFSKRN_FILE_HANDLE_IDX_BITS       5
#define WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS 5

#define WFSKRN_FILE_BUFFER_MEM_ALIGNMENT_SIZE 32

// block cache parameters
#define WFSKRN_BCACHE_HASH_TABLE_SIZE    47   // use for 512KB w/ 8KB pages?
#define WFSKRN_BCACHE_HASH_FN(_key)      (_key % WFSKRN_BCACHE_HASH_TABLE_SIZE)
#define WFSKRN_BCACHE_HASH_LIST(_key)    (gHash[WFSKRN_BCACHE_HASH_FN(_key)])

// parameter for CATEGORY VERY LARGE
#define WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS  8

#define PERMISSION_ENABLED
#define PERM_TEST    1

#endif //wfskrn_Config_h
