/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Defs.h - kernel definitions

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Defs.h,v $
  Revision 1.5  2008/11/05 15:05:44  ueno
  Fixed a bug of WFS_ROUND_DOWN().

  Revision 1.4  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.3  2008/10/14 10:27:30  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.2  2008/07/09 01:29:34  paul
  Added some macros to convert C definition __DATE__ (object file compilation date) to an integer.

  Revision 1.1  2008/04/22 04:39:07  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/19 05:49:31  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2007/12/27 11:36:55  paul
  *** empty log message ***

  Revision 1.1  2007/11/05 19:03:13  paul
  first checkin

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Defs_h
#define wfskrn_Defs_h


#define _WIN32_WINNT 0x500

#include "revolution/wfs.h"
#include "wfskrn_Platform.h"
#include "wfskrn_Config.h"
#include "wfskrn_Errors.h"
#include "wfskrn_Types.h"

#if !_IOP
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#ifndef true
    #define true  1
#endif

#ifndef false
    #define false 0
#endif

#define WFS_ROUND_UP(v,x)   (((v)+(x)-1)&-((signed)x))
#define WFS_ROUND_DOWN(v,x) ((v)&-((signed)x))

#define WFS_NUM_BITS_REQUIRED(x) \
    (((x)<3)?1:((x)<5)?2:((x)<9)?3:\
     ((x)<0x11)?4:((x)<0x21)?5:((x)<0x41)?6:((x)<0x81)?7:\
     ((x)<0x101)?8:((x)<0x201)?9:((x)<0x401)?10:((x)<0x801)?11:\
     ((x)<0x1001)?12:((x)<0x2001)?13:((x)<0x4001)?14:((x)<0x8001)?15:\
     ((x)<0x10001)?16:((x)<0x20001)?17:((x)<0x40001)?18:((x)<0x80001)?19:\
     ((x)<0x100001)?20:((x)<0x200001)?21:((x)<0x400001)?22:((x)<0x800001)?23:\
     ((x)<0x1000001)?24:((x)<0x2000001)?25:((x)<0x4000001)?26:((x)<0x8000001)?27:\
     ((x)<0x10000001)?28:((x)<0x20000001)?29:((x)<0x40000001)?30:((x)<0x80000001)?31:32)

#define BYTES2BITS(b)        ((b)<<3)

#define BUILD_YEAR  ((((__DATE__[7]-'0')*10+(__DATE__[8]-'0'))*10+(__DATE__[9]-'0'))*10+(__DATE__[10]-'0'))
#define BUILD_MONTH (__DATE__[2]=='n'?0:__DATE__[2]=='b'?1:__DATE__[2]=='r'?(__DATE__[0]=='M'?2:3):__DATE__[2]=='y'?4:__DATE__[2]=='n'?5:__DATE__[2]=='l'?6:__DATE__[2]=='g'?7:__DATE__[2]=='p'?8:__DATE__[2]=='t'?9:__DATE__[2]=='v'?10:11)
#define BUILD_DAY   ((__DATE__[4]==' '?0:__DATE__[4]-'0')*10+(__DATE__[5]-'0'))
#define BUILD_DATE  (((BUILD_YEAR-2000)*12+BUILD_MONTH)*31+BUILD_DAY)

#ifndef swap_typ
    #define swap_typ(Typ,a,b) { Typ t=a; a=b; b=t; }
#endif

#ifdef _DEBUG
    #ifndef _ASSERT
        #define _ASSERT(x) ASSERT(x)
    #endif
#else
    #ifndef _ASSERT
        #define _ASSERT(x)
    #endif
#endif

#ifdef _DEBUG
    #define _DEBUG_BREAK_POINT 1
#else
    #define _DEBUG_BREAK_POINT 0
#endif


#endif // wfskrn_Defs_h
