/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Dir.h - Directory Module

  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Dir.h,v $
  Revision 1.28  2008/12/04 00:42:37  ooizumi
  Changed definition name to enable permission.

  Revision 1.27  2008/11/19 05:02:53  ooizumi
  Fixed typo.

  Revision 1.26  2008/11/18 04:57:11  saito_tomoya
  Modified SetBit() and added SetBitArray().

  Revision 1.25  2008/11/07 08:55:16  kondo_masahiro
  Added DirCalculateAttrSubBlkLog2SizeUpdateAllocSize()

  Revision 1.24  2008/11/04 00:25:07  kondo_masahiro
  Added DirSbaAllocAndRegistSubBlk()

  Revision 1.23  2008/10/31 09:51:42  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.22  2008/10/30 05:09:05  kondo_masahiro
  Added DirCheckDisk

  Revision 1.21  2008/10/24 09:02:25  kondo_masahiro
  Added decrease block process of DirResize.

  Revision 1.20  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.19  2008/10/16 09:25:53  ooizumi
  Fixed definitions for debug build.

  Revision 1.18  2008/10/14 10:27:44  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.17  2008/10/10 02:11:31  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.16  2008/10/09 04:15:46  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.15  2008/10/08 23:20:09  kondo_masahiro
  Added codes to access large size file data

  Revision 1.14  2008/10/07 10:30:34  ooizumi
  Added nObjCreator to DirEntryAttrHdr.

  Revision 1.13  2008/09/28 23:30:31  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.12  2008/08/27 09:49:46  paul
  Started coding for different size categories

  Revision 1.11  2008/08/05 04:17:43  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.10  2008/07/28 22:26:54  paul
  Improved updateMap processing. Added DirGetBlk().

  Revision 1.9  2008/07/25 02:56:12  paul
  Changed nTransIdx to *pTransInfo.  Changed WFSTransBlkAdr to TransBlkAdr

  Revision 1.8  2008/07/21 22:50:42  paul
  Changed name from DirItrPinBlkAndUpdatePtrs() -> DirItrPinBlkAndFixPtrs()

  Revision 1.7  2008/07/17 22:23:52  paul
  Changes for update counters

  Revision 1.6  2008/07/09 01:33:13  paul
  Added DirItrCloseNoUnpin alternative to DirItrClose to give the option of unpinning the cache block while closing an interator. Made other minor changes.

  Revision 1.5  2008/06/09 18:15:30  paul
  No longer storing the non-persistent state: number of open read/write handles per file, within file attributes - moved this info to path cache which is non-persistent.

  Revision 1.4  2008/05/12 19:20:28  paul
  Added a list of open directory iterators.

  Revision 1.3  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:28:21  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.14  2008/04/23 01:30:14  paul
  Continued updating and documenting Dir API

  Revision 1.13  2008/04/19 05:49:25  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.12  2008/04/18 00:09:02  paul
  Changed interface to Dir functions

  Revision 1.11  2008/04/15 18:35:38  paul
  Changed definition of DirFind to take separate nMaxBlkDepth and nFlags parameters.

  Revision 1.10  2008/04/05 03:10:40  paul
  Many changes for Delete and associated testing.

  Revision 1.9  2008/03/07 01:38:35  paul
  added DIR_PREFIX_SEARCH

  Revision 1.8  2008/02/29 02:29:46  paul
  Added Delete functions, fixed DirCreateAndInitNewBlk4 parameters

  Revision 1.7  2008/02/28 06:32:42  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.5  2008/02/21 04:30:54  paul
  reorganization

  Revision 1.4  2008/02/21 01:36:56  paul
  moved DIR_MAX_BLK_DEPTH to header file
  updated definition for DirFind

  Revision 1.3  2008/02/21 00:11:34  paul
  Moved SubBlk allocation out of the dir module making it more generic. Now DirBlkHdr pseudo-inherits from WFSSubBlkAllocHdr, which in turn inherits from WFSMetaDataHdr

  Revision 1.2  2008/02/20 07:46:21  paul
  Numerous changes

  Revision 1.1  2007/12/27 11:36:29  paul
  still at an early stage of development

  Revision 1.2  2007/11/08 05:29:16  paul
  continued with block layout initialization code

  Revision 1.1  2007/11/05 19:03:13  paul
  first checkin

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Dir_h
#define wfskrn_Dir_h

#include "wfskrn_Config.h"
#include "wfskrn_SubBlkAlloc.h"
#include "wfskrn_Trans.h"
#include "wfskrn_PTree.h"

#ifdef _DEBUG
    #define _DEBUG_DIR 1  // change to 1 if debug info required
#else
    #define _DEBUG_DIR 0
#endif

#define DIR_MAX_BLK_DEPTH           5
#define DIR_PREFIX_SEARCH           ((u32)-2)
#define DIR_PREFIX_SUPPRESSION      0

#define DIR_SUB_BLK_TYPE_RADIX_NODE 0x00
#define DIR_SUB_BLK_TYPE_BLK_PTR    0x01

#define DIR_SPLIT_BLK_ADR           0

// Flags which are store in the same variable as: WFS_FLAG_IS_A_DIRECTORY, WFS_FLAG_CONTIGUOUS_AREA
#define FILE_SIZE_CATEGORY_VERY_SMALL   0 // Files which fit in the directory leaf block. (up to about 960 bytes - also depends on file name length)
#define FILE_SIZE_CATEGORY_SMALL        1 // Files where the attr block points to between 1 and 5  small blocks.
#define FILE_SIZE_CATEGORY_MEDIUM       2 // Files where the attr block points to between 1 and 5  medium blocks.
#define FILE_SIZE_CATEGORY_LARGE        3 // Files where the attr block points to between 1 and 12 large blocks.
#define FILE_SIZE_CATEGORY_VERY_LARGE   4 // Files where the attr block points to indirect blocks which point to large blocks.



#if _DEBUG_DIR
    extern bool bCheckDirBlks;
#endif

#if _DEBUG_DIR
    extern u32 nDirTest;
    extern u32 nDirTestBreakPoint;
    extern u32 nDirTotalTests;
#endif

typedef struct {
    WFSSubBlkAllocHdr sbah;
    u16               nRootOfs;
    u16               nNumRecs;   // Number of records in the block, not the whole directory
} DirBlkHdr;

#pragma pack(1)
#define SIZE_OF_DIRSUBBLKHDR 3
typedef struct {
    u8              nStrLen;
    u8              nNumEntries;
    utf8            aFirstChar[1];
} _ATTRIBUTE_PACKED DirSubBlkHdr;
#pragma pack()

typedef struct DirNodeStack_ DirNodeStack;
struct DirNodeStack_ {
    DirNodeStack   *pParent;
    WFSBlkAdr       nBlkAdr;
    u32             nUpdateCtr; // Compared against counter stored in dir blocks to determine whether the block has been changed since this iterator was last used.
    u16             nSubBlkOfs;
    u8              nEntryIdx;
    u8              nStrIdx;
};

typedef struct {
    u16            *pOfs;         // A pointer to the offset pointer to the node. This would need to be updated if the sub block is relocated.
    DirSubBlkHdr   *pSubBlkHdr;   // A pointer to the sub block containing the node.
    u32             nLog2Size;    // The base 2 logarithm of the size of the sub block containing the node.
    utf8           *pStrPtr;      // A string pointer to the current cursor position within the search name. After DirFind(), it points to the first mismatch.
    utf8           *pNodeStrPtr;  // A string pointer to the current cursor position within the node string. After DirFind(), it points to the first mismatch.
    u32             nEntryIdx;    // The entry index of the choice within the current node which is closest but <= search name during DirFind().
    DirNodeStack   *pDns;         // A stack of locators within dir node sub blks recording the path taken through the name radix tree
    u32             nDnsDepth;    // Number of entries in the DirNodeStack
} DirNodeItr;

typedef struct {
    u32       nSize;
    u16       nContentIdx;
    u16       nVersion;
} DirFileAttr;

typedef struct {
    WFSBlkAdr nQuota;   // Quota in multiples of blocks
    WFSBlkAdr nSubDirBlkAdr;
} DirDirAttr;

typedef struct {
    u32 nAccessListIdx;
    u32 nAllocSize;
    WFSFileTime timeCreated;
    WFSFileTime timeUpdated;
#ifdef PERMISSION_ENABLED
    WFSTitleId nObjCreator;
#endif
    union {
        DirFileAttr file;
        DirDirAttr  dir;
    };
    u8  nAttrLog2Size;      // Log to the base 2 of the file attributes sub-block
    u8  nSizeCategory;      // Size category of the file
    u8  nNameLen;           // Length of the file name
    u8  aCaseBitArray[1];   // (variable size) A bit per character of the file name to indicate whether each character is upper or lower case
} DirEntryAttrHdr;

typedef struct {
    struct DirItr_ *pNext;
    struct DirItr_ *pPrev;
} DirItrLink;

typedef struct DirItr_ {
    DirItrLink       link;
    u32              nBlkDepth;       // The depth of the block currently pointed to by the iterator. Normally, set to 0 before calling DirFind* functions.
    TransBlkAdr      tba;             // The block currently pointed to by the iterator. Normally, set to the root block of the directory before calling DirFind* functions.
    DirEntryAttrHdr *pAttrHdr;        // Pointer to the file/dir attributes of the current file.
    DirBlkHdr       *pBlkHdr;         // Pointer to start of block tba.nBlkAdr in memory. (Start of the block is the block header).
    u32              nUpdateCtr;      // Update counter of the block referenced by tba.nBlkAdr
    DirNodeItr       dni;
    DirNodeStack   **ppDnsRootParent; // Pointer to the parent link within the root dir node stack entry - to allow for easy freeing of the whole dir node stack.
    WFSFileName      name;            // Name of the current entry pointed to by the iterator.
} DirItr;  // Directory Iterator

#define MAX_INDIRECT_FILE_DEPTH 3

typedef struct {
    //TransBlkAdr      tba;
    //u32              nPos;
    u8              *pDataPtr;
    u32              nRemainingBlkSize;
} DirFilePos;

typedef struct {
    //TransBlkAdr      tba;
    //u32              nPos;
    WFSFileBlkPtr   *pFileBlkPtr;
    u32              nRemainingBlkSize;
} DirFilePosMedium;

WFSKrnResult DirFindFilePos(DirItr *pDi, WFSFileSize nPos, DirFilePos *pFpos, u32 nTransFlags);
WFSKrnResult DirResizeFile(DirItr *pDi, WFSFileSize nNewSize);

// ToDo: Now that a WFSFileName member has been added to DirItr, I should check for performance problems
// introduced by indiscriminate copies of DirItr structs. Some Dir functions which define a local WFSFileName
// could possibly use the storage space within the DirItr instead (eDirLeafBlkSplit, DirNodeBlkSplit).

typedef struct {
    utf8        aToLower[256];
    utf8        aToUpper[256];
    DirItrLink  openDirItrListAnchor;
    DirItr     *pOpenDirItrListAnchor;
    PTreeHdr    updateMap;      // Maps from nBlkAdr to nUpdateCtr for directory blocks which are referenced by currently open directory iterators
    PTreeAllocator updateMapAllocator;
} DirGlobals;

extern DirGlobals dirGlobals;

u32 DirLeafGetSubBlkLog2Size_2(u32 nStrLen, u32 nNumEntries);
u32 DirNodeGetSubBlkLog2Size_3(u32 nStrLen, u32 nNumEntries, utf8 cFirstChoice);
u32 DirCalculateAttrSubBlkLog2Size(AreaInfo *pAreaInfo, u32 nAllocSize, u32 nNameLen, u8 *pSizeCategory);
u32 DirCalculateAttrSubBlkLog2SizeDecrease(AreaInfo *pAreaInfo, u32 nAllocSize, u32 nNameLen, u8 *pSizeCategory);
u32 DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(AreaInfo *pAreaInfo, u32 *pAllocSize, u32 nNameLen, u8 *pSizeCategory);
u32 DirCalculateAttrSubBlkLog2SizeDecreaseUpdateAllocSize(AreaInfo *pAreaInfo, u32 *pAllocSize, u32 nNameLen, u8 *pSizeCategory);

void        DirModuleInit();

void         DirLeafCheckBlk(DirItr *pDi);
void         DirNodeCheckBlk(DirItr *pDi);
void         DirCheckBlk(DirItr *pDi);

#define DIR_CHECK_DISK_FOR_DIR_TREE 0
#define DIR_CHECK_DISK_FOR_ACL_TREE 1 
WFSKrnResult DirCheckDisk(WFSBlkAdr rootBlkAdr, DirItr *pDi, WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 nFlag);

WFSKrnResult DirGetBlk(DirItr *pDi, u32 nTransFlags, DirBlkHdr **ppBlkHdr);

WFSKrnResult DirCreateAndInitNewBlk4(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, WFSMetaDataFlags nFlags, TransInfo *pTransInfo, DirBlkHdr **ppBlkHdr);
//WFSKrnResult DirCreateAndInitNewBlk(TransBlkAdr *pTba, WFSMetaDataFlags nFlags, DirBlkHdr **ppBlkHdr);
void         DirBlkCreateRootNode(DirBlkHdr *pBlkHdr);

WFSKrnResult DirReGetHashAdr(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, WFSBlkAdr nNewBlkAdr, DirBlkHdr* pNewBlkHdr);

WFSKrnResult DirSbaAllocAndRegistSubBlk(WFSFileName *pName, DirItr *pDi, u32 nLog2Size, DirEntryAttrHdr **ppNewAttrHdr);

WFSKrnResult DirFindNameWithoutCaseConversion(DirItr *pDi);


WFSKrnResult DirFindName(WFSFileName *pName, DirItr *pDi);
// Arguments:
// ----------
// pName                [IN/OUT] .. A pointer to a WFSFileName struct containing the utf8 name to search for.
// pDi                     [OUT] .. A pointer to a DirItr struct which will be updated with the search location.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The name was found. *pDi is overwritten with the contents of the directory iterator.
// WFSKRN_RESULT_NOT_FOUND       .. The specified name was not found.
// Other error codes are possible - passed back from lower level functions
//
// Description:
// ------------
// This function searches for a specific name in the directory, and if a match is found, returns the location of the
// file or sub-directory attributes for the associated entry.
// NOTE: Some parts of the directory iterator structure are allocated dynamically, and must be freed
//       by calling DirItrClose() when pDi is no longer needed.


WFSKrnResult DirFindRaw(DirItr *pDi);
// Arguments:
// ----------
// pDi                  [IN/OUT] .. A pointer to DirItr struct which will be updated with the details of the location.
//                                  pDi->name contains a raw input name to search for (will not be case-converted)
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The name was found. *ppAttr is overwritten with the contents of the associated attributes block.
// WFSKRN_RESULT_NOT_FOUND       .. The specified name was not found.
// Other error codes are possible - passed back from lower level functions
//
// Description:
// ------------
// This function searches for a specific raw name in the directory, and if a match is found, returns the location of the
// attributes for the associated entry.
// NOTE: This function is provided to support Access-lists, and should not be used for normal directories.
//       Some parts of the directory iterator structure are allocated dynamically, and must be freed
//       by calling DirItrClose() when pDi is no longer needed.


WFSKrnResult DirFindFirstNamePrefixMatch(WFSFileName *pName, DirItr *pDi);
// Arguments:
// ----------
// pName                [IN/OUT] .. A pointer to a WFSFileName struct containing the prefix of the utf8 name to search for.
// pDi                     [OUT] .. A pointer to a directory iterator struct which stores information about the file location.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The first name which starts with the prefix was found. *pDi contains the location of the file/dir attributes block.
// WFSKRN_RESULT_NOT_FOUND       .. No name was found which starts with the specified prefix.
// Other error codes are possible - passed back from lower level functions
//
// Description:
// ------------
// This function searches the directory for a name which starts with a specified prefix. If one or more matching names exsit,
// this function returns the location of the first match, and the corresponding entry attributes data in *pDi.
// pDi can be passed to DirFindNext() to find the next entry in the directory.
// NOTE: Some parts of the directory iterator structure are allocated dynamically, and must be freed
//       by calling DirItrClose() when pDi is no longer needed.


WFSKrnResult DirNextName(WFSFileName *pName, DirItr *pDi);
// Arguments:
// ----------
// pName                   [OUT] .. A pointer to a WFSFileName struct containing the file name to search for.
// pDi                  [IN/OUT] .. A pointer to a directory iterator struct which stores information about the file location.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. A next entry was found. *pName and *pDi were updated.
// WFSKRN_RESULT_NOT_FOUND       .. No further entries were found.
// Other error codes are possible - passed back from lower level functions
//
// Description:
// ------------
// This function increments the directory iterator to the next entry (in lexicographical order)
// The name of the next entry is returned in *pName. The directory location including file attributes is returned in *pDi.
// NOTE: Some parts of the directory iterator structure are allocated dynamically, and must be freed
//       by calling DirItrClose() when pDi is no longer needed.


WFSKrnResult DirNextRaw(WFSFileName *pName, DirItr *pDi);
// Arguments:
// ----------
// pName                   [OUT] .. A pointer to a WFSFileName struct containing the file name to search for.
// pDi                  [IN/OUT] .. A pointer to a directory iterator struct which stores information about the file location.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. A next entry was found. *pName and *pDi were updated.
// WFSKRN_RESULT_NOT_FOUND       .. No further entries were found.
// Other error codes are possible - passed back from lower level functions
//
// Description:
// ------------
// This function increments the directory iterator to the next entry (in lexicographical order)
// The name of the next entry is returned in *pName. The directory location including file attributes is returned in *pDi.
// NOTE: Some parts of the directory iterator structure are allocated dynamically, and must be freed
//       by calling DirItrClose() when pDi is no longer needed.


WFSKrnResult DirItrClose(DirItr *pDi);
// Arguments:
// ----------
// pDi                      [IN] .. A pointer to a directory iterator struct.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The directory iterator was successfully closed.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function closes the directory iterator and frees up dynamically allocated resources.
// Includes unpinning the currently referenced block


WFSKrnResult DirItrCloseNoUnpin(DirItr *pDi);
// Arguments:
// ----------
// pDi                      [IN] .. A pointer to a directory iterator struct.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The directory iterator was successfully closed.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function closes the directory iterator and frees up dynamically allocated resources.
// Does not include unpinning the currently referenced block



WFSKrnResult DirInsertName(WFSFileName *pName, DirItr *pDi, u32 nAllocSize);
// Arguments:
// ----------
// pName                    [IN] .. A pointer to a WFSFileName struct containing a utf-8 name string to insert.
// pDi                     [OUT] .. A pointer to a directory iterator struct which represents the location within the directory.
// nAllocSize               [IN] .. The initial size of the file. When creating sub-directories, use nAllocSize = 0.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The name was inserted successfully.
// WFSKRN_RESULT_ALREADY_EXISTS  .. A conflicting name already exists in the directory.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function inserts a utf-8 name into the directory. The case information for each character is stored, but is ingored
// when comparing with existing names, so the name will conflict with an existing name which differs only in capitalization.


WFSKrnResult DirInsertRaw(WFSFileName *pName, DirItr *pDi, u32 nAllocSize);
// Arguments:
// ----------
// pName                    [IN] .. A pointer to a WFSFileName struct containing a raw name to insert. (The name can use any 8-bit characters but must be 0 terminated).
// pDi                     [OUT] .. A pointer to a directory iterator struct.
// nAllocSize               [IN] .. The initial size of the file.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The name was inserted successfully.
// WFSKRN_RESULT_ALREADY_EXISTS  .. The same name already exists in the directory.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function inserts a raw string into the directory. No case manipulations are performed,
// so the comparison with existing names is case-sensitive.
// NOTE: This function is provided to support Access-lists, and should not be used for normal directories.


WFSKrnResult DirDeleteEntry(DirItr *pDi);
// Arguments:
// ----------
// pDi                     [OUT] .. A pointer to a directory iterator struct which is pointing to a particular entry.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The entry name and its associated attribute data was deleted successfully from the directory.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function deletes an entry from a directory. This includes deleting the attribute data for the file or sub-directory,
// but does not include deleting separate data blocks or meta-data blocks associated with the file or sub-directory.
// One of the DirFind* functions should be used to


WFSKrnResult DirDeleteFile(DirItr *pDi);
// Arguments:
// ----------
// pDi                      [IN] .. A pointer to a directory iterator struct which is pointing to a particular entry.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The file was deleted successfully.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function deletes the data blocks and meta-data blocks associated with a file entry, but does not delete
// the entry itself from its parent directory.


WFSKrnResult DirDeleteSubDir(DirItr *pDi);
// Arguments:
// ----------
// pDi                      [IN] .. A pointer to a directory iterator struct which is pointing to a particular entry.
//
// Return Values:
// --------------
// WFSKRN_RESULT_OK              .. The sub directory was deleted successfully.
// WFSKRN_RESULT_NOT_EMPTY       .. The sub directory is not empty, and was not deleted.
// Other error codes may be possible - passed back from lower level functions
//
// Description:
// ------------
// This function deletes the root block associated with a sub-directory entry, but does not delete
// the entry itself from its parent directory.


void DirInitDirEntryAttrHdr(DirEntryAttrHdr *pAttrHdr);
void DirInitFileEntryAttrHdr(DirEntryAttrHdr *pAttrHdr);
WFSKrnResult DirItrPinBlkAndFixPtrs(DirItr *pDi, u32 nTransFlags);
WFSKrnResult DirItrMapBlks(DirItr *pDi);
void DirItrUnmapBlks(DirItr *pDi);
WFSKrnResult DirItrCheckMapBlks(DirItr *pDi);
WFSKrnResult DirItrCheckAndUnmapBlks(DirItr *pDi);

void WFSKrnCopyFileName(WFSFileName *pDst, WFSFileName *pSrc);
void WFSKrnCopyStrAndConvertToLowerCase(utf8 *pDstStr, const utf8 *pSrcStr, u32 nLen);
void WFSKrnCopyFileNameAndConvertToLowerCase(WFSFileName *pDst, WFSFileName *pSrc);
void WFSKrnCopyFileNameAndRestoreCase(utf8 *pDst, WFSFileName *pSrc, u8 *pCaseBitArray);
WFSKrnResult WFSKrnValidateCopyStrAndConvertToLowerCase(utf8 *pDstStr, const utf8 *pSrcStr, u32 nLen, u32 nMaxLen);

void WFSKrnStoreCaseInformationStr(u8 *pCaseBitArray, const utf8 *pStr, u32 nLen);

static inline WFSKrnResult DirCreateAndInitNewBlk(TransBlkAdr *pTba, WFSMetaDataFlags nFlags, DirBlkHdr **ppBlkHdr) {
    return DirCreateAndInitNewBlk4(pTba->pAreaInfo, pTba->nBlkAdr, nFlags, pTba->pTransInfo, ppBlkHdr);
}

static inline void WFSKrnStoreCaseInformation(u8 *pCaseBitArray, WFSFileName *pName) {
    WFSKrnStoreCaseInformationStr(pCaseBitArray, pName->sStr, pName->nLen);
}

// Inline Functions for FILE_SIZE_CATEGORY_SMALL // 1 block = 8KB (small block size)
static inline WFSFileBlkPtr *WFSKrnGetFileBlkPtrForCategorySmall(DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nBlkIdx){
    return (WFSFileBlkPtr *)((size_t)pAttrHdr + (1<<nAttrLog2Size) - ((1 + nBlkIdx) * sizeof(WFSFileBlkPtr)));
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoForCategorySmall(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nBlkIdx){
    WFSFileBlkPtr  *pFileBlkPtr  = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, nAttrLog2Size, nBlkIdx);
    WFSFileBlkInfo fileBlkInfo;
    fileBlkInfo.nBlkAdr = pFileBlkPtr->nBlkAdr;
    fileBlkInfo.pHash   = &pFileBlkPtr->hash;
    return fileBlkInfo;
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoForCategorySmallCast(AreaInfo *pAreaInfo, void *pMetaData, u32 nLog2MetaSize, u32 nBlkIdx){
    return WFSKrnGetFileBlkInfoForCategorySmall(pAreaInfo, (DirEntryAttrHdr*)pMetaData, nLog2MetaSize, nBlkIdx);
}

// Inline Functions for FILE_SIZE_CATEGORY_MEDIUM // 1 block = 64KB (small block size)
#define WFSKrnGetFileBlkPtrForCategoryMedium        WFSKrnGetFileBlkPtrForCategorySmall
#define WFSKrnGetFileBlkInfoForCategoryMedium       WFSKrnGetFileBlkInfoForCategorySmall
#define WFSKrnGetFileBlkInfoForCategoryMediumCast   WFSKrnGetFileBlkInfoForCategorySmallCast

// Inline Functions for FILE_SIZE_CATEGORY_LARGE // Meddium Block Size = 64KB (Max hash size)
static inline WFSFileLargeBlkPtr *WFSKrnGetFileLargeBlkPtrForCategoryLarge(DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nLargeBlkIdx){
    return (WFSFileLargeBlkPtr *)( (size_t)pAttrHdr + (1<<nAttrLog2Size) - ((1 + nLargeBlkIdx) * sizeof(WFSFileLargeBlkPtr)) );
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoForCategoryLarge(AreaInfo *pAreaInfo, WFSFileLargeBlkPtr *pFileLargeBlkPtr, u32 nMedBlkIdx){
    WFSFileBlkInfo fileBlkInfo;
    fileBlkInfo.nBlkAdr = pFileLargeBlkPtr->nBlkAdr+(nMedBlkIdx<<pAreaInfo->nLog2BlksPerMediumBlk);
    fileBlkInfo.pHash   = &pFileLargeBlkPtr->aHash[nMedBlkIdx];
    return fileBlkInfo;
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoForCategoryLarge4(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nMedBlkIdx){
    return WFSKrnGetFileBlkInfoForCategoryLarge(pAreaInfo,
                                                WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr,
                                                                                         nAttrLog2Size,
                                                                                         nMedBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                                nMedBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk));
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoForCategoryLargeCast(AreaInfo *pAreaInfo, void *pMetaData, u32 nLog2MetaSize, u32 nMedBlkIdx){
    return WFSKrnGetFileBlkInfoForCategoryLarge4(pAreaInfo, (DirEntryAttrHdr*)pMetaData, nLog2MetaSize, nMedBlkIdx);
}

// Inline Functions for FILE_SIZE_CATEGORY_VERY_LARGE // Meddium Block Size = 64KB (Max hash size)
static inline WFSBlkAdr *WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nIndirectBlkIdx){
    return (WFSBlkAdr*)( (size_t)pAttrHdr + (1<<nAttrLog2Size) - ((1 + nIndirectBlkIdx) * sizeof(WFSBlkAdr)) );
}

static inline WFSBlkAdr *WFSKrnGetBlkAdrPtrFromAttrHdrAndLargeBlkIdxForCategoryVeryLarge(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nLargeBlkIdx){
    u32 nIndirectBlkIdx = (nLargeBlkIdx)/pAreaInfo->nNumLargeBlkPtrsPerBlk;
    return WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pAreaInfo, pAttrHdr, nAttrLog2Size, nIndirectBlkIdx);
}

static inline WFSBlkAdr *WFSKrnGetBlkAdrPtrFromAttrHdrAndMedBlkIdxForCategoryVeryLarge(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, u32 nAttrLog2Size, u32 nMedBlkIdx){
    return WFSKrnGetBlkAdrPtrFromAttrHdrAndLargeBlkIdxForCategoryVeryLarge(pAreaInfo, pAttrHdr, nAttrLog2Size, nMedBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
}

static inline WFSFileLargeBlkPtr *WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(u8 *pBlkPtr, u32 nLargeBlkIdx){
    return (WFSFileLargeBlkPtr *)( (size_t)pBlkPtr + sizeof(WFSMetaDataHdr) + (nLargeBlkIdx * sizeof(WFSFileLargeBlkPtr)) );
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(AreaInfo *pAreaInfo, u8 *pBlkPtr, u32 nMedBlkIdx){
    return WFSKrnGetFileBlkInfoForCategoryLarge(pAreaInfo,
                                                WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pBlkPtr,
                                                                                                            nMedBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                                nMedBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk));
}

static inline WFSFileBlkInfo WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLargeCast(AreaInfo *pAreaInfo, void *pMetaData, u32 nLog2MetaSize, u32 nMedBlkIdx){
    return WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pAreaInfo, (u8*)pMetaData, nMedBlkIdx);
}

// Inline Functions for Dir Module
static inline u32 DirLeafGetSubBlkLog2Size(DirSubBlkHdr *pSubBlkHdr) {
    return DirLeafGetSubBlkLog2Size_2(pSubBlkHdr->nStrLen, pSubBlkHdr->nNumEntries);
}

static inline u32 DirNodeGetSubBlkLog2Size(DirSubBlkHdr *pSubBlkHdr) {
    return DirNodeGetSubBlkLog2Size_3(pSubBlkHdr->nStrLen, pSubBlkHdr->nNumEntries, pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen]);
}

static inline u16* DirNodeGetOfsTblStartPtr_ote(DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
    if (pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen] == 0) {
        return &pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries-1];
    } else {
        return &pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries];
    }
}
static inline u16* DirNodeGetOfsTblStartPtr_l2s(DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    return DirNodeGetOfsTblStartPtr_ote(pSubBlkHdr, pOfsTblEnd);
}
static inline u16* DirNodeGetOfsTblStartPtr(DirSubBlkHdr *pSubBlkHdr) {
    u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    return DirNodeGetOfsTblStartPtr_l2s(pSubBlkHdr, nLog2Size);
}

static inline u16* DirNodeGetIntraBlkOfsPtr_ote(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
#if DIR_SPLIT_BLK_ADR
    return &pOfsTblEnd[-nIdx];
#else
    if (pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen] == 0) {
        return &pOfsTblEnd[-nIdx-1];
    } else {
        return &pOfsTblEnd[-nIdx];
    }
#endif
}

static inline u16* DirNodeGetIntraBlkOfsPtr_l2s(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    return DirNodeGetIntraBlkOfsPtr_ote(nIdx, pSubBlkHdr, pOfsTblEnd);
}
static inline u16* DirNodeGetIntraBlkOfsPtr(s32 nIdx, DirSubBlkHdr *pSubBlkHdr) {
    u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    return DirNodeGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size);
}
static inline u16 DirNodeGetIntraBlkOfs_ote(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
    return *DirNodeGetIntraBlkOfsPtr_ote(nIdx, pSubBlkHdr, pOfsTblEnd);
}
static inline u16 DirNodeGetIntraBlkOfs_l2s(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    return *DirNodeGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size);
}
static inline u16 DirNodeGetIntraBlkOfs(s32 nIdx, DirSubBlkHdr *pSubBlkHdr) {
    return *DirNodeGetIntraBlkOfsPtr(nIdx, pSubBlkHdr);
}
static inline void DirNodeSetIntraBlkOfs_ote(s32 nIdx, u16 nOfs, DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
    *DirNodeGetIntraBlkOfsPtr_ote(nIdx, pSubBlkHdr, pOfsTblEnd) = nOfs;
}
static inline void DirNodeSetIntraBlkOfs_l2s(s32 nIdx, u16 nOfs, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    *DirNodeGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size) = nOfs;
}
static inline void DirNodeSetIntraBlkOfs(s32 nIdx, u16 nOfs, DirSubBlkHdr *pSubBlkHdr) {
    *DirNodeGetIntraBlkOfsPtr(nIdx, pSubBlkHdr) = nOfs;
}

/*inline WFSBlkAdr DirNodeGetBlkAdr_ote(DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
#if DIR_SPLIT_BLK_ADR
    return (WFSBlkAdr)(pOfsTblEnd[-1] + (pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries-1]<<16));
#else
    DirSubBlkHdr *psbh = pSubBlkHdr; // for warning
    return ((WFSBlkAdr*)pOfsTblEnd)[-1];
#endif
}*/
#if DIR_SPLIT_BLK_ADR
    #define DirNodeGetBlkAdr_ote(pSubBlkHdr, pOfsTblEnd) ((WFSBlkAdr)(pOfsTblEnd[-1] + (pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries-1]<<16)));
#else
    #define DirNodeGetBlkAdr_ote(pSubBlkHdr, pOfsTblEnd) ((WFSBlkAdr*)pOfsTblEnd)[-1];
#endif


static inline WFSBlkAdr DirNodeGetBlkAdr_l2s(DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    return DirNodeGetBlkAdr_ote(pSubBlkHdr, pOfsTblEnd);
}
static inline WFSBlkAdr DirNodeGetBlkAdr(DirSubBlkHdr *pSubBlkHdr) {
    u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    return DirNodeGetBlkAdr_l2s(pSubBlkHdr, nLog2Size);
}

/*static inline void DirNodeSetBlkAdr_ote(WFSBlkAdr nBlkAdr, DirSubBlkHdr *pSubBlkHdr, u16 *pOfsTblEnd) {
#if DIR_SPLIT_BLK_ADR
    pOfsTblEnd[-1] = (u16)nBlkAdr;
    pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries-1] = (u16)(nBlkAdr>>16);
#else
    DirSubBlkHdr *psbh = pSubBlkHdr; // for warning
    ((u32*)pOfsTblEnd)[-1] = nBlkAdr;
#endif
}*/
#if DIR_SPLIT_BLK_ADR
    #define DirNodeSetBlkAdr_ote(nBlkAdr, pSubBlkHdr, pOfsTblEnd) { pOfsTblEnd[-1] = (u16)nBlkAdr; pOfsTblEnd[-(s32)pSubBlkHdr->nNumEntries-1] = (u16)(nBlkAdr>>16); }
#else
    #define DirNodeSetBlkAdr_ote(nBlkAdr, pSubBlkHdr, pOfsTblEnd) ((u32*)pOfsTblEnd)[-1] = nBlkAdr;
#endif


static inline void DirNodeSetBlkAdr_l2s(WFSBlkAdr nBlkAdr, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    DirNodeSetBlkAdr_ote(nBlkAdr, pSubBlkHdr, pOfsTblEnd);
}
static inline void DirNodeSetBlkAdr(WFSBlkAdr nBlkAdr, DirSubBlkHdr *pSubBlkHdr) {
    u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    DirNodeSetBlkAdr_l2s(nBlkAdr, pSubBlkHdr, nLog2Size);
}

static inline u16 *DirLeafGetIntraBlkOfsPtr_ote(s32 nIdx, u16 *pOfsTblEnd) {
    return &pOfsTblEnd[-nIdx];
}
static inline u16* DirLeafGetIntraBlkOfsPtr_l2s(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    return DirLeafGetIntraBlkOfsPtr_ote(nIdx, pOfsTblEnd);
}
static inline u16* DirLeafGetIntraBlkOfsPtr(s32 nIdx, DirSubBlkHdr *pSubBlkHdr) {
    u32 nLog2Size = DirLeafGetSubBlkLog2Size(pSubBlkHdr);
    return DirLeafGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size);
}
static inline u16 DirLeafGetIntraBlkOfs_ote(s32 nIdx, u16 *pOfsTblEnd) {
    return *DirLeafGetIntraBlkOfsPtr_ote(nIdx, pOfsTblEnd);
}
static inline u16 DirLeafGetIntraBlkOfs_l2s(s32 nIdx, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    return *DirLeafGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size);
}
static inline u16 DirLeafGetIntraBlkOfs(s32 nIdx, DirSubBlkHdr *pSubBlkHdr) {
    return *DirLeafGetIntraBlkOfsPtr(nIdx, pSubBlkHdr);
}
static inline void DirLeafSetIntraBlkOfs_ote(s32 nIdx, u16 nOfs, u16 *pOfsTblEnd) {
    *DirLeafGetIntraBlkOfsPtr_ote(nIdx, pOfsTblEnd) = nOfs;
}
static inline void DirLeafSetIntraBlkOfs_l2s(s32 nIdx, u16 nOfs, DirSubBlkHdr *pSubBlkHdr, u32 nLog2Size) {
    *DirLeafGetIntraBlkOfsPtr_l2s(nIdx, pSubBlkHdr, nLog2Size) = nOfs;
}
static inline void DirLeafSetIntraBlkOfs(s32 nIdx, u16 nOfs, DirSubBlkHdr *pSubBlkHdr) {
    *DirLeafGetIntraBlkOfsPtr(nIdx, pSubBlkHdr) = nOfs;
}

#endif //define wfskrn_Dir_h
