/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirNodeStack.h - Directory Module

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirNodeStack.h,v $
  Revision 1.3  2008/10/16 09:25:53  ooizumi
  Fixed definitions for debug build.

  Revision 1.2  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/17 17:34:24  paul
  Added some more debug functions

  Revision 1.2  2008/04/05 03:20:57  paul
  _DEBUG -> _DEBUG_DIR

  Revision 1.1  2008/02/27 09:48:02  paul
  *** empty log message ***


*---------------------------------------------------------------------------*/

#ifndef wfskrn_DirNodeStack_h
#define wfskrn_DirNodeStack_h

#include "wfskrn_Dir.h"

#if _DEBUG_DIR
    #define _DEBUG_DIR_NODE_STACK 1 // change to 1 if debug info required
#else
    #define _DEBUG_DIR_NODE_STACK 0
#endif

#if _DEBUG
    void DebugCheckNumFreeDnsPoolEntries(char *str);
#endif

#if _DEBUG_DIR_NODE_STACK
    extern bool bCheckDirNodeStack;
    void DirNodeStackCheckFreeList();
    WFSKrnResult _DirNodeStackGetEntry (char *sFile, u32 nLine, DirNodeItr *pDni);
    WFSKrnResult _DirPushNodeStackEntry(char *sFile, u32 nLine, DirNodeItr *pDni, WFSBlkAdr nBlkAdr);
    void         _DirNodeStackFree     (char *sFile, u32 nLine, DirNodeItr *pDni, DirNodeStack **ppParent);
    void         _DirPopNodeStackEntry (char *sFile, u32 nLine, DirNodeItr *pDni);
    void         _DirNodeStackCheckEmpty();

    #define DirNodeStackGetEntry(pDni)          _DirNodeStackGetEntry (_KRN_FILE_, __LINE__, pDni)
    #define DirPushNodeStackEntry(pDni,nBlkAdr) _DirPushNodeStackEntry(_KRN_FILE_, __LINE__, pDni, nBlkAdr)
    #define DirNodeStackFree(pDni,ppParent)     _DirNodeStackFree     (_KRN_FILE_, __LINE__, pDni, ppParent)
    #define DirPopNodeStackEntry(pDni)          _DirPopNodeStackEntry (_KRN_FILE_, __LINE__, pDni)
    #define DirNodeStackCheckEmpty              _DirNodeStackCheckEmpty();
#else
    WFSKrnResult DirNodeStackGetEntry (DirNodeItr *pDni);
    WFSKrnResult DirPushNodeStackEntry(DirNodeItr *pDni, WFSBlkAdr nBlkAdr);
    void         DirNodeStackFree(DirNodeItr *pDni, DirNodeStack **ppParent);
    void         DirPopNodeStackEntry(DirNodeItr *pDni);
    #define DirNodeStackCheckEmpty
#endif

WFSKrnResult DirAllocateMoreNodeStackEntries();

void         DirConstructEntryNameFromNodeStack(WFSFileName *pKey, DirItr *pDi);


#endif //wfskrn_DirNodeStack_h
