/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_EPTree.h


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_EPTree.h,v $
  Revision 1.4  2008/12/17 05:08:17  ueno
  Cleanup.

  Revision 1.3  2008/12/15 02:40:27  ueno
  Added support for multi p-tree.

  Revision 1.1  2008/12/02 04:55:39  ueno
  Added EPTree to support multi-PTree.


*---------------------------------------------------------------------------*/

#ifndef __WFSKRN_EPTREE_H__
#define __WFSKRN_EPTREE_H__

#include "wfskrn_Area.h"
#include "wfskrn_FTree.h"

#define EPTREEHDR_OFFSET          (BLOCK_SIZE - sizeof(EPTreeHdr))
#define GET_EPTREEHDR(pBlock)     ((EPTreeHdr*) ((u8*) (pBlock) + EPTREEHDR_OFFSET))
#define GET_BLOCK_FROM_EPTREEHDR(eptreeHdr)  ((u8*) (eptreeHdr) - EPTREEHDR_OFFSET)

#define EPTREE_GET_KEY(epti)    ((epti)->nKey)
#define EPTREE_GET_DATA(epti)   ((epti)->nData)

typedef struct
{
    PTreeHdr ptreeHdr; // 8 bytes.
    WFSBlkAdr addr;
    u8 height;         // height = 1, 2 ...
    u8 padding[11];
    SubBlockAllocatorHeader sbaHdr; // 8 bytes.
} EPTreeHdr; // sizeof(EPTreeHdr) must be 32 bytes. ( == sizeof(FTreeBlock));

typedef PTreeItr EPTreeItr;

#define EPTREE_MAX_HEIGHT  2

typedef struct
{
    EPTreeHdr* eptreeHdr[EPTREE_MAX_HEIGHT]; // inherited FBATreeItr
    EPTreeItr  epti[EPTREE_MAX_HEIGHT];      // inherited FBATreeItr.
} MEPTreeItr;

EPTreeHdr* EPTreeBlockInit(AreaInfo* area, void* block, WFSBlkAdr addr, u32 offset, u8 height);
EPTreeItr* EPTreeGetLastIter(MEPTreeItr* iter);
WFSKrnResult EPTreeFind(AreaInfo* area, MEPTreeItr* iter, u32 nKey);
WFSKrnResult EPTreePrev(AreaInfo* area, MEPTreeItr* iter);
WFSKrnResult EPTreeNext(AreaInfo* area, MEPTreeItr* iter);
WFSKrnResult EPTreeFirst(AreaInfo* area, MEPTreeItr* iter);
WFSKrnResult EPTreeLast(AreaInfo* area, MEPTreeItr* iter);
WFSKrnResult EPTreeInsert(AreaInfo* area, MEPTreeItr* iter, u32 key, u32 data); // [check] can remove area?
WFSKrnResult EPTreeDelete(AreaInfo* area, MEPTreeItr* iter, u32 key, WFSBlkAdr* freeBlks);
WFSKrnResult EPTreeInit(AreaInfo* area, u32 offset);

#endif // __WFSKRN_EPTREE_H__

