/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Errors_h

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Errors.h,v $
  Revision 1.19  2008/12/15 02:40:27  ueno
  Added support for multi p-tree.

  Revision 1.18  2008/11/07 08:54:35  kondo_masahiro
  Added WFSKrnExit4/5OnError()

  Revision 1.17  2008/10/30 05:08:19  kondo_masahiro
  Added error for DirCheckDisk

  Revision 1.16  2008/09/28 23:30:40  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.15  2008/08/27 10:03:58  paul
  added WFSKRN_RESULT_BLK_CACHE_ALLOC_FAILED

  Revision 1.14  2008/08/05 04:17:52  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.13  2008/07/17 22:25:49  paul
  Added WFSKRN_RESULT_DIR_ITR_INVALID
  Fixed value for WFSKRN_RESULT_FILE_OPEN

  Revision 1.12  2008/07/17 05:11:44  kondo_masahiro
  Added WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT

  Revision 1.11  2008/07/17 04:18:24  ueno
  Added errors for freeblock allocator.

  Revision 1.10  2008/07/11 18:05:17  paul
  Updated to match latest wfsError.h

  Revision 1.9  2008/07/09 01:35:43  paul
  Changed some result code names, and added new result codes for PTree.

  Revision 1.8  2008/06/10 20:13:04  kondo_masahiro
  Added WFSKrnCommandCountBreak.

  Revision 1.7  2008/06/09 18:05:48  paul
  Added support for logging calls to WFSSrv functions.
  Added new result codes.

  Revision 1.6  2008/06/05 08:09:18  ueno
  Revised permission APIs.

  Revision 1.5  2008/05/28 01:06:13  ueno
  Added error codes for AclCheckDisk().

  Revision 1.4  2008/05/12 19:21:31  paul
  Added an error code.

  Revision 1.3  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 12:07:44  nakanose_jin
  add new errors

  Revision 1.13  2008/04/26 01:13:53  wayne.wong
  Moved area allocation functionality from Volume to Area module.
  Changed error codes appropriately.

  Revision 1.12  2008/04/25 04:35:29  nakanose_jin
  add new errors

  Revision 1.11  2008/04/22 23:01:43  paul
  Added WFSKRN_RESULT_DEV_IN_USE to match wfsError.h. Added WFSKrnExitOnError

  Revision 1.10  2008/04/18 01:42:46  wayne.wong
  Added some area-specific error codes.

  Revision 1.9  2008/04/15 07:40:11  ueno
  Added error codes to WFSKrnResult for accesslist.

  Revision 1.8  2008/03/19 07:35:48  wayne.wong
  Added and fixed error codes.

  Revision 1.7  2008/03/11 03:09:34  wayne.wong
  Added some more BCache and Volume errors.

  Revision 1.6  2008/02/28 06:32:36  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.5  2008/02/27 00:46:26  paul
  Differentiated names of error functions

  Revision 1.4  2008/02/21 22:25:26  wayne.wong
  Added errors for the modules BCache, PMem,  Volume, and Transaction.

  Revision 1.3  2008/02/21 01:37:43  paul
  split into wfskrn_Errors.h and wfskrn_Errors.cpp

  Revision 1.2  2008/02/20 07:47:52  paul
  Updated error codes and added WFSKrnOutputError macro to output error code as a string

  Revision 1.1  2007/12/27 11:35:52  paul
  internal error definitions

 *---------------------------------------------------------------------------*/

#ifndef wfskrn_Errors_h
#define wfskrn_Errors_h

#define WFSKrnReturnOnError(x) { WFSKrnResult nResult = x; if (nResult < WFSKRN_RESULT_OK) return nResult; }
#define WFSKrnExitOnError(x)   {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit; }
#define WFSKrnExit1OnError(x)  {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit1; }
#define WFSKrnExit2OnError(x)  {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit2; }
#define WFSKrnExit3OnError(x)  {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit3; }
#define WFSKrnExit4OnError(x)  {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit4; }
#define WFSKrnExit5OnError(x)  {              nResult = x; if (nResult < WFSKRN_RESULT_OK) goto Exit5; }

#define WFSKrnReturnOnNullParameter(p) { if (p == NULL) return WFSKRN_RESULT_INVALID; }

typedef enum {
    WFSKRN_RESULT_OK                  =    0,  // Success

    // Errors that can usually be detected early, and if detected, would result in the Async callback not being called:
    WFSKRN_RESULT_BUSY                =   -1,  // Too many requests, try waiting then calling again.
    WFSKRN_RESULT_OUT_OF_MEMORY       =   -2,  // The library does not have enough memory to complete the requested operation. (May be due to too many outstanding aysnc calls).
    WFSKRN_RESULT_INVALID             =   -3,  // The function parameters were invalid, e.g. File name is too long, Path is too long or too deep.
    WFSKRN_RESULT_ACCESS              =   -4,  // Trying to write to a file which was opened for read-only access, or vice versa.
    WFSKRN_RESULT_LIB_NOT_INITIALIZED =   -5,  // The WFS library has not been initialized. Please call WFSInit() (This error is checked in functions that require the library to be initialized, but only in the debug versions).
    WFSKRN_RESULT_LIB_ALREADY_INITIALIZED = -6,  // The WFS library has already been initialized.
    WFSKRN_RESULT_FILE_TOO_BIG        =   -7,  // The current write/append request would push the file over the size limit.
    WFSKRN_RESULT_NO_CHANGE_SIZE      =   -8,  // The requested operation would cause the file size to change, but this file has WFS_PERM_NO_CHANGE_SIZE flag set.

    // Errors that cannot usually be detected early, and would be passed via nResult to the Async callbacks:
    WFSKRN_RESULT_MEDIA_ERROR         =   -9,  // The device or media is not attached, or was removed before the requested operation could be completed.
    WFSKRN_RESULT_DEV_UNUSABLE        =  -10,  // Attached device is not accessible, e.g. because it is an unsupported device, or has an unknown format.
    WFSKRN_RESULT_DEV_NOT_INITIALIZED =  -11,  // Device not formatted, or does not yet contain a WFS volume
    WFSKRN_RESULT_DEV_IN_USE          =  -12,  // Trying to initialize a device which is in use (mounted) by one or more clients.
    WFSKRN_RESULT_VOL_ID_ERROR        =  -13,  // Attempted to mount a volume with the same unique Id as an already mounted volume (This could happen if the User backs up an entire WFS volume using a PC)
    WFSKRN_RESULT_WRITE_PROTECTED     =  -14,  // Device is physically write protected (e.g. USB Flash memory, SD Card)
    WFSKRN_RESULT_ALREADY_MOUNTED     =  -15,  // The device is already mounted

    WFSKRN_RESULT_PERMISSION          =  -16,  // Permission flags do not permit the requested access from this application.
    WFSKRN_RESULT_PERMISSION_CL       =  -17,  // Permission error caused by control level problems.
    WFSKRN_RESULT_ACL_FULL            =  -18,  // The access list is full.
    WFSKRN_RESULT_ACL_ENTRY_NOT_FOUND =  -19,  // The specified entry does not exist in the access list.
    WFSKRN_RESULT_AUTHENTICATION      =  -20,  // Trying to access content on a different Wii, which does not have the required access rights.
    WFSKRN_RESULT_CORRUPTION          =  -21,  // A corrupted block was encountered which prevented the operation from being completed.

    WFSKRN_RESULT_DIRECTORY_QUOTA     =  -22,  // One of the ancestor directories would exceed its quota.

    WFSKRN_RESULT_MAX_HANDLES         =  -23,  // The maximum number of concurrent file handles / search handles are already in use.
    WFSKRN_RESULT_ALREADY_EXISTS      =  -24,  // The file or directory being created already exists.
    WFSKRN_RESULT_NOT_FOUND           =  -25,  // The specified file or directory or device does not exist.
    WFSKRN_RESULT_NOT_EMPTY           =  -26,  // The specified directory cannot be deleted because it is not empty.
    WFSKRN_RESULT_NOT_FILE            =  -27,  // The specified path is directory instead of a file.
    WFSKRN_RESULT_NOT_DIRECTORY       =  -28,  // The specified path is a file instead of a directory
    WFSKRN_RESULT_FILE_OPEN           =  -29,  // The file is open. Cannot delete, move, rename, change permissions of an open file. Cannot open a file more than once unless all opens are WFS_ACCESS_READ.
    WFSKRN_RESULT_LOCKED              =  -30,  // Cannot perform the requested operation since the file or directory is locked by an existing operation.
    WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED = -31,  // A resource limit prevents this function call. Try smaller values.

    // Directory find results - these can be errors or expected depending on whether we are trying to access an existing file or trying to create a new one:
    WFSKRN_RESULT_DIR_ENTRY_FOUND              =  -40,  // The search string was matched and pointed to an entry
    WFSKRN_RESULT_DIR_NODE_STRING_PREFIX       =  -41,  // The search string ended early during comparison with a node sub string. (So this name is a prefix of existing names)
    WFSKRN_RESULT_DIR_CHOICE_PREFIX            =  -42,  // The search string ended at the end of a node sub string, but the node did not have a termination choice. (This name is a prefix of existing names)
    WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH     =  -43,  // A mismatch was encountered while comparing the search string to a sub node prefix string.
    WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND    =  -44,  // A mismatch was encountered while comparing a character of the search string with a set of choices for that character.
    WFSKRN_RESULT_DIR_BLK_FULL                 =  -45,  // Unable to insert the specified string to the radix tree, due to lack of space in the block.
    WFSKRN_RESULT_DIR_ITR_INVALID              =  -46,  // One or more blocks referenced by the directory iterator and its node stack have been updated since the iterator was last used. DirFind must be called again.

    // These result codes are not errors, but normal return values for the path cache functions
    WFSKRN_RESULT_SRV_END_OF_PATH              =  -60,  // Last part of path name parsed
    WFSKRN_RESULT_SRV_PATH_DEPTH_1             =  -61,  // Path was of depth 1, such as "/something"
    WFSKRN_RESULT_SRV_PATH_DEV                 =  -62,  // Path was "/dev"
    WFSKRN_RESULT_SRV_PATH_VOL                 =  -63,  // Path was "/vol"
    WFSKRN_RESULT_SRV_PATH_VOL_ROOT            =  -64,  // Path was a volume root path, such as "/vol/xxxxxxx/"

    WFSKRN_RESULT_PTREE_ENTRY_FOUND            =  -70,
    WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND        =  -71,
    WFSKRN_RESULT_PTREE_FULL                   =  -72,
    WFSKRN_RESULT_PTREE_EMPTY                  =  -73,

    WFSKRN_RESULT_DEVICE_ERROR                 =  -80,
    WFSKRN_RESULT_DEVICE_INVALID_PARAMETER     =  -81,
    WFSKRN_RESULT_DEVICE_NOT_FOUND             =  -82,
    WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT     =  -83,

    WFSKRN_RESULT_BLK_CACHE_ALLOC_FAILED       =  -90,

    WFSKRN_RESULT_BCACHE_ERROR                 = -300,  // unspecified error
    WFSKRN_RESULT_BCACHE_RESOURCE_LIMIT        = -301,  // exhausted resources
    WFSKRN_RESULT_BCACHE_INVALID_PARAMETER     = -302,  // at least one argument is invalid
    WFSKRN_RESULT_BCACHE_NO_MEMORY             = -303,  // could not malloc
    WFSKRN_RESULT_BCACHE_NOT_FOUND             = -304,  // sector is not cached
    WFSKRN_RESULT_BCACHE_MAX_DEVICES           = -305,  // max devices alreay allocated
    WFSKRN_RESULT_BCACHE_INVALID_DEVICE        = -306,  // device not found/open
    WFSKRN_RESULT_BCACHE_INVALID_HANDLE        = -307,  // invalid device handle
    WFSKRN_RESULT_BCACHE_INVALID_VOLUME        = -308,  // invalid volume ID (not mapped)
    WFSKRN_RESULT_BCACHE_ALREADY_MAPPED        = -309,  // vol or devH already mapped
    WFSKRN_RESULT_BCACHE_ALLOC                 = -310,  // cannot alloc (no mem, or pmem error)
    WFSKRN_RESULT_BCACHE_HASH_BLK_NOT_MAPPED   = -311,  // hash block isn't mapped on cache

    WFSKRN_RESULT_VOLUME_ERROR                 = -400,  // unspecified volume error
    WFSKRN_RESULT_VOLUME_INVALID_PARAMETER     = -401,  // at least one argument is invalid
    WFSKRN_RESULT_VOLUME_BCACHE_ALLOC          = -402,  // cannot allocate a page/block in memory
    WFSKRN_RESULT_VOLUME_BCACHE_CONFIG         = -403,  // block cache not configured correctly
    WFSKRN_RESULT_VOLUME_ID                    = -404,  // bad volume ID or not initialized
    WFSKRN_RESULT_VOLUME_CORRUPT_MR            = -405,  // the MR (or its free lists) are corrupted
    WFSKRN_RESULT_VOLUME_AREA_ALLOC            = -406,  // area allocation cannot be satisfied
    WFSKRN_RESULT_VOLUME_DEVICE_PARAMETER      = -407,  // parameters incompatible with device
    WFSKRN_RESULT_VOLUME_NOT_MAPPED            = -408,  // volume is not mapped; invalid volume ID

    WFSKRN_RESULT_TRANSACTION_ERROR            = -500,  // unspecified transaction error
    WFSKRN_RESULT_TRANSACTION_INVALID_PARAMETER= -501,

    WFSKRN_RESULT_ACL_ERROR                    = -600,  // unspecified error
    WFSKRN_RESULT_ACL_INVALID_PARAMETER        = -601,  // at least one argument is invalid
    WFSKRN_RESULT_ACL_MAX_ENTRIES              = -602,  // max entries alreay in the accesslist.
    WFSKRN_RESULT_ACL_BUFFER_TOO_SMALL         = -603,  // buffer is too small to store an accesslist.
    WFSKRN_RESULT_ACL_CACHE                    = -610,  // cannot allocate a cache.
    WFSKRN_RESULT_ACL_FILE                     = -620,  // cannot create, modify or delete the accesslist file.
    WFSKRN_RESULT_ACL_ACLDIR_INCONSISTENT      = -621,  // accesslist files and indexlist are inconsistent.
    WFSKRN_RESULT_ACL_FILENAME                 = -630,  // cannot convert acl filename <=> acl content.
    WFSKRN_RESULT_ACL_NAMEDIR_INCONSISTENT     = -631,  // access name files and indexlist are inconsistent.
    WFSKRN_RESULT_ACL_HANDLE                   = -640,  // cannot allocate an accesslist handle.

    WFSKRN_RESULT_AREA_ERROR                   = -700,  // unspecified area error
    WFSKRN_RESULT_AREA_INVALID_PARAMETER       = -701,
    WFSKRN_RESULT_AREA_BCACHE_ALLOC            = -702,  // cannot allocate a page/block in memory
    WFSKRN_RESULT_AREA_CORRUPT_HDR             = -703,  // the area header is corrupted
    WFSKRN_RESULT_AREA_VOLUME_NOT_MAPPED       = -704,  // volume is not mapped; invalid volume ID
    WFSKRN_RESULT_AREA_ALLOC                   = -705,  // area allocation cannot be satisfied

    WFSKRN_RESULT_FTREE_RETRY                  = -801,  // retry insert.
    WFSKRN_RESULT_FTREE_SPLIT                  = -802,  // cannot split tree.
    WFSKRN_RESULT_FTREE_EMPTY                  = -803,  // no entry in the F-tree.

    WFSKRN_RESULT_CHECKDISK_INCONSISTENT       = -901,  // bit arrays does not consistent between free block allocator and dir module.
    WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR     = -902,  // bit arrays can not construct.

    WFSKRN_RESULT_NOT_IMPLEMENTED              = -1026,  // The requested function has not yet been implemented  //ToDo: This error should be removed from the final API
    WFSKRN_RESULT_UNKNOWN                      = -1027,  // Unexpected error - could be caused by a bug in the file system.
    WFSKRN_RESULT_FATAL_ERROR                  = -1028   // May be caused by critical metadata corruption.
} WFSKrnResult;

#define WFSSetLastError(nErr) (wkg.nLastErr = (nErr))

const char *WFSKrnGetErrorString( WFSKrnResult nResult );
WFSKrnResult WFSKrnOutputErrorCode(WFSKrnResult nResult);
WFSKrnResult WFSKrnOutputErrorStr(const char *pStr);
void WFSKrnCommandCountBreak();

#if _DEBUG
    #define DBG_OUTPUT_TO_LOG_FILE   0
    #define DBG_OUTPUT_TO_SCREEN     0
    #define DBG_OUTPUT_PREFIX "%5d. "
    extern u32 nDbgCommandCount;
    extern u32 nDbgCommandOutputStart;
    extern u32 nDbgCommandOutputEnd;
    extern u32 nDbgCommandCountBreak;
#define WFSKrnDbgOutputTest(x) if ((nDbgCommandCount>=nDbgCommandOutputStart)&&(nDbgCommandCount<nDbgCommandOutputEnd)) { x; WFSKrnCommandCountBreak(); }
#else
    #define WFSKrnDbgOutputTest(x)
#endif //_DEBUG

#if (DBG_OUTPUT_TO_LOG_FILE && _WIN32)
    extern FILE *fpLog;
    #define WFSKrnDbgOutputInit(fileName)  fpLog = fopen(fileName, "w");
    #define WFSKrnDbgOutputClose() fclose(fpLog);
    #if DBG_OUTPUT_TO_SCREEN
        #define WFSKrnDbgOutputCmdNum         { fprintf(fpLog, DBG_OUTPUT_PREFIX"WFSSrv", nDbgCommandCount); MyOSReport(DBG_OUTPUT_PREFIX, nDbgCommandCount); }
        #define WFSKrnDbgOutput1(fmt)         { fprintf(fpLog, fmt);          MyOSReport(fmt); }
        #define WFSKrnDbgOutput2(fmt,a)       { fprintf(fpLog, fmt, a);       MyOSReport(fmt, a); }
        #define WFSKrnDbgOutput3(fmt,a,b)     { fprintf(fpLog, fmt, a,b);     MyOSReport(fmt, a,b); }
        #define WFSKrnDbgOutput4(fmt,a,b,c)   { fprintf(fpLog, fmt, a,b,c);   MyOSReport(fmt, a,b,c); }
        #define WFSKrnDbgOutput5(fmt,a,b,c,d) { fprintf(fpLog, fmt, a,b,c,d); MyOSReport(fmt, a,b,c,d); }
    #else //DBG_OUTPUT_TO_SCREEN
        #define WFSKrnDbgOutputCmdNum         fprintf(fpLog, DBG_OUTPUT_PREFIX"WFSSrv", nDbgCommandCount)
        #define WFSKrnDbgOutput1(fmt)         fprintf(fpLog, fmt);
        #define WFSKrnDbgOutput2(fmt,a)       fprintf(fpLog, fmt, a);
        #define WFSKrnDbgOutput3(fmt,a,b)     fprintf(fpLog, fmt, a,b);
        #define WFSKrnDbgOutput4(fmt,a,b,c)   fprintf(fpLog, fmt, a,b,c);
        #define WFSKrnDbgOutput5(fmt,a,b,c,d) fprintf(fpLog, fmt, a,b,c,d);
    #endif //DBG_OUTPUT_TO_SCREEN
#else //(DBG_OUTPUT_TO_LOG_FILE && _WIN32)
    #define WFSKrnDbgOutputInit(fileName)
    #define WFSKrnDbgOutputClose()
    #if DBG_OUTPUT_TO_SCREEN
        #define WFSKrnDbgOutputCmdNum         MyOSReport(DBG_OUTPUT_PREFIX, nDbgCommandCount)
        #define WFSKrnDbgOutput1(fmt)         MyOSReport(fmt);
        #define WFSKrnDbgOutput2(fmt,a)       MyOSReport(fmt, a);
        #define WFSKrnDbgOutput3(fmt,a,b)     MyOSReport(fmt, a,b);
        #define WFSKrnDbgOutput4(fmt,a,b,c)   MyOSReport(fmt, a,b,c);
        #define WFSKrnDbgOutput5(fmt,a,b,c,d) MyOSReport(fmt, a,b,c,d);
    #else //DBG_OUTPUT_TO_SCREEN
        #define WFSKrnDbgOutputCmdNum
        #define WFSKrnDbgOutput1(fmt)
        #define WFSKrnDbgOutput2(fmt,a)
        #define WFSKrnDbgOutput3(fmt,a,b)
        #define WFSKrnDbgOutput4(fmt,a,b,c)
        #define WFSKrnDbgOutput5(fmt,a,b,c,d)
    #endif //DBG_OUTPUT_TO_SCREEN
#endif //(DBG_OUTPUT_TO_LOG_FILE && _WIN32)

#endif //wfskrn_Errors_h
