/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_FreeBlkAlloc.h


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_FreeBlkAlloc.h,v $
  Revision 1.16  2008/12/18 00:32:11  ueno
  Removed FBADirtyFtreeBlock().

  Revision 1.15  2008/12/17 05:08:17  ueno
  Cleanup.

  Revision 1.14  2008/12/15 02:40:27  ueno
  Added support for multi p-tree.

  Revision 1.13  2008/12/02 04:55:39  ueno
  Added EPTree to support multi-PTree.

  Revision 1.11  2008/11/18 04:57:11  saito_tomoya
  Modified SetBit() and added SetBitArray().

  Revision 1.10  2008/11/07 03:01:42  ueno
  Fixed FTreeExtentInfo{} for IOP.

  Revision 1.9  2008/10/14 11:25:13  ueno
  Disabled runtime check.

  Revision 1.8  2008/09/29 10:19:25  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.7  2008/08/13 04:14:00  ueno
  Modified FreeBlkAlloc() to allocate blocks near nRefBlkAdr.

  Revision 1.6  2008/08/13 00:14:09  paul
  Changed type of nStride from u32 to s32

  Revision 1.5  2008/08/09 03:34:29  ueno
  Disabled runtime check.

  Revision 1.4  2008/08/08 13:27:42  ueno
  Added UnpinFTreeBlk().

  Revision 1.3  2008/08/07 10:46:50  ueno
  Modified for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.2  2008/08/02 11:52:10  ueno
  Revised the positions of FTreeHdr{} and FBAPTreeHdr{} in 8KB-blocks to protect hash code.

  Revision 1.1  2008/08/01 06:57:39  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/

#ifndef __WFSKRN_FREEBLKALLOC_H__
#define __WFSKRN_FREEBLKALLOC_H__

#include "wfskrn_Area.h"
#if BTREE_BASED_ALLOCATOR

#include "wfskrn_FreeBlkAlloc_Type.h"
#include "wfskrn_FTree.h"
#include "wfskrn_EPTree.h"

#define FBA_ROOT_BLK           0 // default block for p-tree.
#define FBA_INITIAL_FTREE_BLK  1 // default block for f-tree.

#define GET_FBATREE_HDR(area) (&((area)->ah.freeBlkAllocHdr)) // [check]
#define FBA_SBA_OFFSET        (sizeof(AreaHdr) + sizeof(VolumeHdr))
#define FBA_GET_FTREE_BLK(pIter)  ((pIter)->epti[(pIter)->height - 1].nData)

typedef struct
{
    EPTreeHdr* eptreeHdr[EPTREE_MAX_HEIGHT]; // inherit MEPTreeItr.
    EPTreeItr  epti[EPTREE_MAX_HEIGHT];      // inherit MEPTreeItr.
    FTreeItr   fti;
    FTreeHdr*  ftreeHdr;
    WFSBlkAdr  ftreeAddr;

    WFSBool    isInserted;
    u8         height;
} FBATreeItr;

typedef struct _ExtentStorageHeader
{
    u32 prio; // priority
    u32 numPrev;
    u32 numNext;
} ExtentStorageHeader;

typedef struct _ExtentStorageEntry
{
    WFSBlkAdr start;
    WFSBlkAdr len; // block len.
    u16       treeType;
    u16       nNumBlk;
} ExtentStorageEntry;

typedef struct _FBASingleEntry
{
    ExtentStorageHeader header;
    ExtentStorageEntry entry;
} FBASingleEntry;

typedef struct _RemainderExtents
{
    FTreeExtent head;
    FTreeExtent fractionHead;
    FTreeExtent tail;
    FTreeExtent fractionTail;
} RemainderExtents;

typedef struct _FTreeExtentInfo
{
    WFSBlkAdr start;
    WFSBlkAdr len; // the unit is 8kB.
    FBATreeItr iter;
    WFSBool alignment;
} FTreeExtentInfo;

#define FBA_RESULT_KEY(pIter)  ((pIter)->fti.ptree.nKey)
#define FBA_RESULT_DATA(pIter)  ((pIter)->fti.ptree.nData)

EPTreeHdr* EpTreeBlockInit(AreaInfo* area, void* block, u32 offset);

WFSKrnResult FreeBlkInsert(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, WFSBlkAdr nNumBlks);
WFSKrnResult FreeBlkDelete(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, WFSBlkAdr nNumBlks, u32 nBlkLog2Size);

WFSKrnResult FreeBlkInit(AreaInfo *pAreaInfo, u32 offset);
WFSKrnResult FreeBlkAlloc(AreaInfo *pAreaInfo, WFSBlkAdr nRefBlkAdr, WFSBlkAdr nNumBlks, WFSBlkAdr *aBlkAdr, s32 nStride, u32 nBlkLog2Size);
WFSKrnResult FreeBlkFree(AreaInfo *pAreaInfo, WFSBlkAdr nNumBlks, WFSBlkAdr *aBlkAdr, s32 nStride, u32 nBlkLog2Size);

WFSKrnResult FreeBlkFirst(AreaInfo* area, FBATreeItr* iter, u32 nSubTreeBlkLog2Size);
WFSKrnResult FreeBlkLast(AreaInfo* area, FBATreeItr* iter, u32 nSubTreeBlkLog2Size);
WFSKrnResult FreeBlkNext(AreaInfo* area, FBATreeItr* iter);
WFSKrnResult FreeBlkPrev(AreaInfo* area, FBATreeItr* iter);

WFSKrnResult FreeBlkFind(AreaInfo* area, FBATreeItr* iter, u32 key, u32 nBlkLog2Size);
WFSKrnResult FreeBlkFindStrict(AreaInfo* area, FBATreeItr* iter, u32 nBlkAdr, u32 nBlkLog2Size);

WFSBlkAdr FreeBlkGetFreeSize(AreaInfo *pAreaInfo);

WFSBool FreeBlkVerify(AreaInfo* area);
WFSKrnResult FreeBlkCheckDisk(AreaInfo* area, WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize,  u32 size);
u32 FreeBlkGetFreeBlkLen(AreaInfo* area, u32 nBlkLog2SizeOfs);
WFSKrnResult FreeBlkDumpFreeblock(AreaInfo* area);

void FBAItrOpen(AreaInfo* area, FBATreeItr* iter);
void FBAItrClose(AreaInfo* area, FBATreeItr* iter);

PTreeHdr* FTreeSelectSubTree(FTreeHdr* header, WFSBlkAdr unitLen);

void FBAUnpinCache(AreaInfo* area, u32 diskAddr);
void FBAUnpinFTreeBlk(AreaInfo* area, FBATreeItr* iter);
void FBAFreeCache(AreaInfo* area, u32 diskAddr);
WFSKrnResult FBAGetCache(AreaInfo* area, u32 diskAddr, u32 flags, void** cache);

#ifdef _DEBUG
// #define FBA_DEBUG_VERBOSE
// #define WFSKRN_FBA_RUNTIME_CHECK
#endif // _DEBUG

#ifdef WFSKRN_FBA_RUNTIME_CHECK
#define FBA_STRICT_ASSERT(x)    {                               \
                                    if (FBADebugVerbose)        \
                                    {                           \
                                        WFSKRN_FTREE_ASSERT(x); \
                                    }                           \
                                }
#else
#define FBA_STRICT_ASSERT(x)
#endif

#endif // BTREE_BASED_ALLOCATOR

#endif // __WFSKRN_FREEBLKALLOC_H__

