/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_FreeBlkAlloc.h


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_FreeBlkAlloc_Type.h,v $
  Revision 1.4  2008/12/02 04:55:39  ueno
  Added EPTree to support multi-PTree.

  Revision 1.2  2008/08/09 03:34:29  ueno
  Disabled runtime check.

  Revision 1.1  2008/08/07 10:47:29  ueno
  Added for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.2  2008/08/02 11:52:10  ueno
  Revised the positions of FTreeHdr{} and FBAPTreeHdr{} in 8KB-blocks to protect hash code.

  Revision 1.1  2008/08/01 06:57:39  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/

#ifndef __WFSKRN_FREEBLKALLOC_TYPE_H__
#define __WFSKRN_FREEBLKALLOC_TYPE_H__

#include "wfskrn_Area.h"
#if BTREE_BASED_ALLOCATOR

#define GET_FBATREE_HDR(area) (&((area)->ah.freeBlkAllocHdr))

#define GET_EPTREEHDR_FROM_AREAINFO(area) ((EPTreeHdr*) ((u8*) ((area)->pAflh) + GET_FBATREE_HDR(area)->ofsEPtreeHdr))
//              ~

/*
                 root disk block in the area.
                     +-----------------+ <-- pVolInfo - - - - - - - - -A
                     |                 |                               |
                     +-----------------+ <-- pAreaHdr (pAreaInfo->ah)  |
                     |                 |                               |
                   A +-----------------+ <-- pAreaInfo->pAflh          |
                   | |                 | A                             |
                   | |                 | |                             |
                   | |                 | |                             | pAreaInfo->nBlkSize (4kB or 8kB?)
    ofsEPtreeHdr   | |                 | |                             |
                   | |                 | | nSizeRemaining              |
                   | |                 | |                             |
                   V +-----------------+ |                             |
                     |    EPTreeHdr    | |                             |
                     +-----------------+ V - - - - - - - - - - - - - - V
*/

// FBATreeHdr is stored in a disk.
typedef struct _FBATreeHdr
{
    unsigned  totalBlks; // freeblocks and metadata blocks in the Freeblock allocator.
    unsigned  ofsEPtreeHdr; // offset of FBAPTreeHdr from pAreaInfo->pAflh.
    unsigned  ptreeHeight; // height of ptree blocks. the min is FBA_PTREE_MIN_HEIGHT.
} FBATreeHdr;

#endif // BTREE_BASED_ALLOCATOR
#endif // __WFSKRN_FREEBLKALLOC_TYPE_H__

