/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Mutex.h - mutual exclusion for multi-thread case

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Mutex.h,v $
  Revision 1.2  2008/08/05 04:17:58  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/19 05:48:44  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2008/02/28 06:32:23  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.1  2007/12/27 11:34:29  paul
  similar to wfs_Mutex.h, but will eventually have to support IOP

 *---------------------------------------------------------------------------*/

#ifndef wfskrn_Mutex_h
#define wfskrn_Mutex_h

#include "wfskrn_Defs.h"
#define _TWL 1 //twl modified

#ifdef _DEBUG
    #define _DEBUG_WFSKRN_MUTEX 0  // change to 1 if debug info required
#else
    #define _DEBUG_WFSKRN_MUTEX 0
#endif

#if _WIN32

typedef struct {
    HANDLE hHandle;
} WFSKrnMutex;

typedef struct {
    HANDLE hAutoResetEvent;
    //HANDLE hManualResetEvent;
} WFSKrnCond;

#define WFSKrnSemaphore HANDLE

typedef struct {
    HANDLE hThread;
} WFSKrnThread;

void WFSKrnInitMutex(WFSKrnMutex *mutex);
void WFSKrnLockMutex(WFSKrnMutex *mutex);
bool WFSKrnTryLockMutex(WFSKrnMutex *mutex);
void WFSKrnUnlockMutex(WFSKrnMutex *mutex);

void WFSKrnInitCond(WFSKrnCond *pCond);
void WFSKrnSignalCond(WFSKrnCond *pCond);
void WFSKrnWaitCond(WFSKrnCond *pCond, WFSKrnMutex* pMx);
void WFSKrnResetCond(WFSKrnCond *pCond);

#define WFSKrnInitSemaphore(pSem, initNum)   *pSem = CreateSemaphore(NULL,initNum,1,NULL);
#define WFSKrnSignalSemaphore(pSem) ReleaseSemaphore(*pSem, 1, NULL);
#define WFSKrnWaitSemaphore(pSem)   WaitForSingleObject(*pSem, INFINITE);

s32 WFSKrnResumeThread(WFSKrnThread *pThread);
void WFSKrnCancelThread(WFSKrnThread *pThread);
void WFSKrnExitThread(void *pVal);
s32 WFSKrnSuspendThread(WFSKrnThread *pThread);
bool WFSKrnIsThreadTerminated(WFSKrnThread *thread);
bool WFSKrnJoinThread(WFSKrnThread *thread, void** val);

//-- twl modified
#elif _TWL
typedef u32 OSPriority;
#define WFSKrnMutex OSMutex
#define WFSKrnCond OSEvent
#define WFSKrnSemaphore OSSemaphore 
#define WFSKrnThread OSThread

#define WFSKrnInitMutex osInitMutex
#define WFSKrnLockMutex osLockMutex
#define WFSKrnTryLockMutex osTryLockMutex
#define WFSKrnUnlockMutex osUnlockMutex

//#define WFSKrnInitCond osInitCond
//#define WFSKrnSignalCond osSignalCond
//#define WFSKrnWaitCond osWaitCond
//#define WFSKrnResetCond(x) // This function does nothing on RVL, only needed for Windows
void WFSKrnInitCond(WFSCond *pCond);
void WFSKrnSignalCond(WFSCond *pCond);
void WFSKrnWaitCond(WFSCond *pCond, WFSMutex* pMx);
void WFSKrnResetCond(WFSCond *pCond);

#define WFSKrnInitSemaphore osInitSemaphore
#define WFSKrnSignalSemaphore osSignalSemaphore
#define WFSKrnWaitSemaphore osWaitSemaphore

#define WFSKrnResumeThread OSResumeThread
#define WFSKrnCancelThread OSCancelThread
#define WFSKrnExitThread OSExitThread
#define WFSKrnSuspendThread OSSuspendThread
#define WFSKrnIsThreadTerminated OSIsThreadTerminated
#define WFSKrnJoinThread OSJoinThread
//-- twl modified

#elif _RVL

#define WFSKrnMutex OSMutex
#define WFSKrnCond OSCond
#define WFSKrnSemaphore OSSemaphore 
#define WFSKrnThread OSThread

#define WFSKrnInitMutex OSInitMutex
#define WFSKrnLockMutex OSLockMutex
#define WFSKrnTryLockMutex OSTryLockMutex
#define WFSKrnUnlockMutex OSUnlockMutex

#define WFSKrnInitCond OSInitCond
#define WFSKrnSignalCond OSSignalCond
#define WFSKrnWaitCond OSWaitCond
#define WFSKrnResetCond(x) // This function does nothing on RVL, only needed for Windows

#define WFSKrnInitSemaphore OSInitSemaphore
#define WFSKrnSignalSemaphore OSSignalSemaphore
#define WFSKrnWaitSemaphore OSWaitSemaphore

#define WFSKrnResumeThread OSResumeThread
#define WFSKrnCancelThread OSCancelThread
#define WFSKrnExitThread OSExitThread
#define WFSKrnSuspendThread OSSuspendThread
#define WFSKrnIsThreadTerminated OSIsThreadTerminated
#define WFSKrnJoinThread OSJoinThread

#elif _IOP

#define WFSKRN_MUTEX_Q_SIZE 1
typedef s32 OSPriority;

typedef struct{
    IOSMessageQueueId   mqid;   // Message queue
    IOSMessage          mq[WFSKRN_MUTEX_Q_SIZE];
    s32                 count;  // lock count
    IOSThreadId         owner;
} WFSKrnMutex;

//#define WFSKrnCond

typedef struct {
    IOSThreadId tid;
    s32         count;
} WFSKrnThread;

void WFSKrnInitMutex(WFSKrnMutex *mutex);
void WFSKrnLockMutex(WFSKrnMutex *mutex);
bool WFSKrnTryLockMutex(WFSKrnMutex *mutex);
void WFSKrnUnlockMutex(WFSKrnMutex *mutex);

//#define WFSKrnInitCond
//#define WFSKrnSignalCond
//#define WFSKrnWaitCond
//#define WFSKrnResetCond

s32 WFSKrnResumeThread(WFSKrnThread *pThread);
void WFSKrnCancelThread(WFSKrnThread *pThread);
void WFSKrnExitThread(void *pVal);
s32 WFSKrnSuspendThread(WFSKrnThread *pThread);
bool WFSKrnIsThreadTerminated(WFSKrnThread *pThread);
bool WFSKrnJoinThread(WFSKrnThread *pThread, void** val);

#endif

typedef void *(*WFSKrnThreadFunc)(void *);

bool WFSKrnCreateThread(WFSKrnThread *thread,
                        WFSKrnThreadFunc threadFunc,
                        void*      param,
                        void*      stackBase,
                        u32        stackSize,
                        OSPriority priority);

#endif //define wfskrn_Mutex_h
