/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_PTree.h


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_PTree.h,v $
  Revision 1.2  2008/07/17 22:33:39  paul
  Added adapter functions to the Sba allocators for updateMap

  Revision 1.1  2008/07/09 01:42:35  paul
  u32->u32 mapping optimized B-Tree. (for use by new block allocator, and transaction block re-mapper)

*---------------------------------------------------------------------------*/

#ifndef wfskrn_PTree_h
#define wfskrn_PTree_h

#ifdef _DEBUG
    #define _DEBUG_PTREE 0  // change to 1 if debug info required
#else
    #define _DEBUG_PTREE 0
#endif


#include "wfskrn_Defs.h"


#define PTREE_LOG2_NODE_SIZE  5
#define PTREE_NODE_FAN_OUT    6   // (6 * 2 Byte Offset) + (5 * 4 Byte Key) = 32 Bytes
#define PTREE_LEAF_FAN_OUT    4   // 4 * (4 Byte Key + 4 Byte Value) = 32 Bytes
#define PTREE_MAX_HEIGHT      5


// Node Sizes:

typedef struct {
    u32         aKey[PTREE_NODE_FAN_OUT-1];
    u16         aOfs[PTREE_NODE_FAN_OUT];
} PTreeNode;

typedef struct {
    u32         aKey[PTREE_LEAF_FAN_OUT];
    u32         aData[PTREE_LEAF_FAN_OUT];
} PTreeLeaf;

typedef struct {
    u16        *pOfs;         // A pointer to the offset pointer to the node. This would need to be updated if the sub block is relocated.
    PTreeLeaf  *pLeaf;
    PTreeNode  *apNode[PTREE_MAX_HEIGHT-1];
    u32         aEntryIdx[PTREE_MAX_HEIGHT];
    //u32        *apKey[PTREE_MAX_HEIGHT];
    u32         nKey;
    u32         nData;
} PTreeItr;

typedef u16          (*PTreeAllocNodeFunc) (void *pBase);
typedef WFSKrnResult (*PTreeAllocNodesFunc)(void *pBase, u16 *pNodeOfs, u32 nNumNodes);
typedef void         (*PTreeFreeNodeFunc)  (void *pBase, u16 nOfs);

typedef struct {
    void                 *pBase;
    PTreeAllocNodeFunc  fpAllocNode;
    PTreeAllocNodesFunc fpAllocNodes;
    PTreeFreeNodeFunc   fpFreeNode;
} PTreeAllocator;

// Allocator functions which use SubBlkAlloc module
u16          PTreeAllocNodeFromSba(void *pBase);
WFSKrnResult PTreeAllocNodesFromSba(void *pBase, u16 aNodeOfs[], u32 nNumNodes);
void         PTreeFreeNodeToSba(void *pBase, u16 nOfs);

typedef struct {
    u32         nRootHeight;  // height of root node (opposite of depth).
    u16         nRootOfs;     // Offset to the root node of the tree.
    u16         nNumRecs;     // Number of records in the tree.
} PTreeHdr;

typedef struct {
    u16         nParentOfs;     // The offset the to parent of this node
    u8          nEntryIdx;      // The index of the entry in the parent block which points to this node
    u8          nNumEntries;    // The number of entries in this node
    u32         nLength;
} PTreeAttrHdr;

void         PTreeInit  (PTreeHdr *pPTreeHdr, PTreeAllocator *pAllocator);
WFSKrnResult PTreeFind  (PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, PTreeAllocator *pAllocator);
WFSKrnResult PTreeInsert(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, u32 nData, PTreeAllocator *pAllocator);
WFSKrnResult PTreeDelete(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator);
WFSKrnResult PTreeFirst (PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator);
WFSKrnResult PTreeNext  (PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator);

WFSKrnResult PTreeInsertOrUpdateRec(PTreeHdr *pPTreeHdr, u32 nKey, u32 nData, PTreeAllocator *pAllocator);

void         PTreeCheck(PTreeHdr *pPTreeHdr, u32 nLog2Size, PTreeAllocator *pAllocator);
void         PTreeTest();


#endif //define wfskrn_PTree_h
