/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Permission_AccessList.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_AccessList.h,v $
  Revision 1.23  2008/09/24 07:32:07  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.22  2008/08/14 08:02:57  kondo_masahiro
  Fixed definition of ACL_GET_CL_RELATIVE() from inline to static inline.

  Revision 1.21  2008/07/23 04:18:07  ueno
  Reduced memory size required for accesslist.

  Revision 1.20  2008/07/11 08:10:50  ueno
  Fixed control level check.

  Revision 1.19  2008/06/16 18:03:56  ueno
  Fixed comment.

  Revision 1.18  2008/06/05 14:04:11  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.17  2008/06/05 08:44:40  ueno
  Cleanup.

  Revision 1.16  2008/06/05 08:09:18  ueno
  Revised permission APIs.

  Revision 1.15  2008/06/02 15:09:54  ueno
  Revised WFSKrnAccessList{} to reduce the size.

  Revision 1.14  2008/05/29 14:08:13  ueno
  Modified _WFSAccessListControlBlock().

  Revision 1.13  2008/05/28 02:12:57  ueno
  Cleanup.

  Revision 1.12  2008/05/23 06:31:52  ueno
  Cleanup.

  Revision 1.11  2008/05/19 05:31:53  ueno
  Defined accesslist meta-data types.

  Revision 1.10  2008/05/15 13:33:46  ueno
  Implemented initialization using directory module.

  Revision 1.9  2008/05/06 02:27:21  ueno
  Fixed ifdef.

  Revision 1.8  2008/04/29 06:25:14  ueno
  Cleanup.

  Revision 1.7  2008/04/29 01:12:54  ueno
  Defined WFS_ROOT_ID and WFS_EVERYONE_ID.

  Revision 1.6  2008/04/25 19:15:14  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.4  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.3  2008/04/25 04:40:47  ueno
  Added support for WFSDEV.

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.23  2008/04/23 08:59:55  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.22  2008/04/21 06:07:54  ueno
  Fixed struct for C99.

  Revision 1.21  2008/04/15 07:35:29  ueno
  Fixed error codes.

  Revision 1.20  2008/04/15 06:45:47  ueno
  Changed return type to WFSKrnResult.

  Revision 1.19  2008/04/14 13:50:17  ueno
  Cleanup.

  Revision 1.18  2008/04/14 03:02:16  ueno
  Modified WFSKrnCreateRootAccessList() to create an accesslist only for the root.
  Added WFSKrnDebugCreateAccessList() for tests.

  Revision 1.17  2008/04/11 13:20:13  ueno
  Removed  WFSKrnAccessListSetInstallerPermission().

  Revision 1.16  2008/04/11 06:16:33  ueno
  Defined WFSAclFileHandle.
  Cleaned up headers.

  Revision 1.15  2008/04/11 02:52:11  ueno
  Renamed WFSKrnCreateAccessList() to make clear the purpose.

  Revision 1.14  2008/04/09 11:31:25  ueno
  Changed the maximum number of accesslist handles.

  Revision 1.13  2008/04/07 12:47:17  ueno
  Added support for group others.

  Revision 1.12  2008/04/07 09:18:54  ueno
  Modified WFSAccessListControlBlock() to keep permissions for the installer.

  Revision 1.11  2008/04/03 09:59:30  ueno
  Added manuals of WFS AccessList APIs.

  Revision 1.10  2008/04/03 06:40:35  ueno
  Removed WFSKRN_ACCESSLIST_FLAG_CONTENT_DIRECTORY flag.
  Cleanup.

  Revision 1.9  2008/04/02 14:15:13  ueno
  Revised permission inheritance.

  Revision 1.8  2008/03/31 14:06:49  ueno
  Removed WFSKrnAccessListAddUser() and RemoveUser().
  Removed WFSKrnAccessListCan*().

  Revision 1.7  2008/03/19 10:18:49  ueno
  Revised WFSKrnAddUser() and RemoveUser() to specify operator title ID.

  Revision 1.6  2008/03/18 13:42:28  ueno
  Added WFSKrnAccessListSetControlLevel(), GetControlLevel() to set/get a control level.

  Revision 1.5  2008/03/14 08:02:48  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.4  2008/03/12 06:37:34  ueno
  Added WFSAccessListCanRead(), Write()...

  Revision 1.3  2008/03/11 12:10:35  ueno
  Cleanup.

  Revision 1.2  2008/03/11 11:38:58  ueno
  Cleanup.

  Revision 1.1  2008/03/11 11:09:21  ueno
  Renamed.

  Revision 1.14  2008/03/11 09:51:51  ueno
  Fixed EscapeAccessList() to make a filename from UTF-16 characters (0x3300-0x33ff).

  Revision 1.13  2008/03/11 06:53:02  ueno
  Added WFSAccessListSetPermission() and GetPermission().

  Revision 1.12  2008/03/07 05:43:57  ueno
  Cleanup.

  Revision 1.10  2008/03/06 12:26:04  ueno
  Revised cache freelist (signgle -> double link list).

  Revision 1.9  2008/03/06 10:46:19  ueno
  Cleanup.

  Revision 1.8  2008/02/29 00:47:45  ueno
  Changed the attribute of index to inout (WFSAccessListAddEntry() and RemoveEntry()).

  Revision 1.7  2008/02/28 13:11:58  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.6  2008/02/28 04:36:47  ueno
  Cleanup.

  Revision 1.5  2008/02/27 10:05:26  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.4  2008/02/27 04:33:36  ueno
  Added root ID and alias permission flags.

  Revision 1.3  2008/02/26 14:12:06  ueno
  Added WFSAccessListAddEntry() and WFSAccessListRemoveEntry() to edit accesslist.

  Revision 1.2  2008/02/25 09:44:08  ueno
  Added support for accesslist cache.

  Revision 1.1  2008/02/22 05:15:55  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#ifndef __WFSKRN_PERMISSION_ACCESSLIST_H__
#define __WFSKRN_PERMISSION_ACCESSLIST_H__

#include <revolution/wfs.h>

#if !_IOP
#include "wfs_Client.h"
#endif

#include "wfskrn_Area.h"
#include "wfskrn_Mutex.h"
#include "wfskrn_Permission_Type.h"

#define WFSKRN_ACCESSLIST_MAX_CONTROL_LEVEL   3

#if !_IOP
namespace wfsdev_api{
  struct tagSAreaInfo;
}
#endif


#ifndef ROUNDUP32BYTE
#define ROUNDUP32BYTE(val) (((size_t)val+31)&~31)
#endif // ROUNDUP32BYTE

//
// Accesslist Cache
//

#define ACL_MAX_MEMORY_BLOCK  254 // [check]
#define ACL_NO_ENTRY          255 // [check]

typedef struct WFSKrnAccessListEntry
{
    u32 id; // title ID.
    u32 ecId; // entry creator's Id.
    u32 pm; // permissions and control levels.
} WFSKrnAccessListEntry;

typedef struct _WFSKrnAccessList WFSKrnAccessList;

struct _WFSKrnAccessList
{
    u8  num;  // number of blocks in this list.
    u8  type; // WFSKRN_ACCESSLIST_FREEBLOCK or 0.
    u16 ref;  // reference count of this cache.
    WFSKrnAclHandle handle; // access list handle and WFSKRN_ACCESSLIST_FLAG_* flags
    u8  prevFreeCacheBlock; // previous free cache.
    u8  nextFreeCacheBlock; // next free cache.
    u8  prevBlock;          // previous cache.
    u8  nextBlock;          // next cache.
};

// cache list, free cache list.
typedef struct WFSKrnAccessListLinkList
{
    WFSKrnAccessList* head;
    WFSKrnAccessList* tail;
} WFSKrnAccessListLinkList;

#define  WFSKRN_ACCESSLIST_FREEBLOCK                  ((u8) 0x1)  // Free memory block.
#define  WFSKRN_BLOCKS_IN_LIST(x)  (sizeof(WFSKrnAccessList)/sizeof(WFSKrnAccessListBlock) + \
                                    WFSKrnAccessListGetNum(x) * (sizeof(WFSKrnAccessListEntry)/sizeof(WFSKrnAccessListBlock)))

typedef struct WFSKrnAccessListBlock
{
    u8 num;   // number of blocks in this chunk.
    u8 type;  // WFSKRN_ACCESSLIST_FREEBLOCK or 0.
    u8 reserved;
    u8 nextBlock;  // the block number of the next chunk.
} WFSKrnAccessListBlock;

#define ACL_DEFAULT_CACHESIZE (128 * sizeof(WFSKrnAccessListBlock))

//
// Accesslist Index
//

#define ACL_RELATIVE_CL_FLAG        0x8000
#define ACL_CL_MASK                 0x0003
#define WFS_PERM_RELATIVE_CL_FLAG   0x8000

static inline WFSKrnResult ACL_GET_CL_RELATIVE(u32 perm, int* cl)
{
    WFSKrnResult result = WFSKRN_RESULT_INVALID;
    if (perm & ACL_RELATIVE_CL_FLAG)
    {
        u32 val = ((~perm >> WFS_PERM_CL_SHIFT) & ACL_CL_MASK);
        *cl =  -((int) val);
        result = WFSKRN_RESULT_OK;
    }
    return result;
}

#define  ACL_ENTRY_NET_SIZE  (sizeof(u32) + sizeof(u32) + sizeof(u16)) // smaller than sizeof(WFSKrnAccessListEntry).
#define  ACL_MAX_ENTRIES (WFS_MAX_FILE_NAME_SIZE - sizeof(u32)*2)/(ACL_ENTRY_NET_SIZE*2)
// flags for accesslist handle
#define  ACL_FLAG_DIRECTORY              (1<<31) // accesslist for directory.
#define  ACL_FLAG_RESERVED0              (1<<30)
#define  ACL_FLAG_RESERVED1              (1<<29)
#define  ACL_FLAG_RESERVED2              (1<<28)
#define  ACL_TYPE_MASK                   (~0xf0000000)
#define  ACL_TYPE(x)                     ((x)->handle & ~ACL_TYPE_MASK)
#define  ACL_INDEX(x)                    ((x)->handle & ACL_TYPE_MASK)

#define WFSKRN_ACCESSLIST_GROUP                        0x80000000

// escape sequence to convert a content of an accesslist to a filename.
#define ESCAPE                                         0x33
#define ESCAPED_NULL                                   0x58

//
// Index List
// which manages indices of free and used accesslists.
//
#define WFSKRN_ACCESSLIST_INDEXES_IN_PAGE              16
#define WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE           6
#define WFSKRN_ACCESSLIST_FREEINDEX                    0xffffffff

typedef struct IndexListHeader
{
    u32 numEntry;
} IndexListHeader;

typedef struct IndexListEntry
{
    u32 location;
} IndexListEntry;

//
// Control Block of Accesslists.
//

#if WFSDEV
namespace wfsdev_api{
  struct tagSAreaInfo;
}
#endif

#ifndef _struct_WFSAccessListControlBlock_
#define _struct_WFSAccessListControlBlock_
typedef struct _WFSAccessListControlBlock WFSAccessListControlBlock;
#endif

struct _WFSAccessListControlBlock
{
    WFSKrnAccessListBlock*   freeBlock;
    WFSKrnAccessListBlock*   memoryBlock; // heap for accesslist cache.
    u32                      numBlock;

    WFSKrnAccessListLinkList cache;
    WFSKrnAccessListLinkList cacheFreeList;
#ifdef WFSKRN_ACL_USE_MUTEX
    WFSKrnMutex              mxMutex;
#endif // WFSKRN_ACL_USE_MUTEX

#if WFSDEV
    struct wfsdev_api::tagSAreaInfo         *aclWrapper;
#endif
    WFSBool                  initialized;
};

//
// APIs
//

//--------------------------------------------------------------------------------------------------------------
// WFSKrnAccessListInit
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_ACL_ERROR      .. (see wfskrn_Errors.h)
//       WFSKRN_RESULT_INVALID
//       WFSKRN_RESULT_OUT_OF_MEMORY
//       WFSKRN_RESULT_LIB_ALREADY_INITIALIZED
//
// Description:
// ------------
// This function initializes accesslist cache and accesslist index list.
// WFSAccessListControlBlock{} in area (AreaInfo{}) must be properly initialized before calling this function.
// - Memory blocks for accesslist cache must be set in memoryBlock field of WFSAccessListControlBlock{}.
// - The number of blocks must be set in numBlock field.
// - initialized field must be set to FALSE.
WFSKrnResult WFSKrnAccessListInit(AreaInfo* area);


//--------------------------------------------------------------------------------------------------------------
// WFSKrnAccessListOpen
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_OUT_OF_MEMORY
//       WFSKRN_RESULT_LIB_ALREADY_INITIALIZED
//
// Description:
// ------------
// This function initializes accesslist cache and loads accesslist configuration from the device:
// - the root block for accesslist files.
// - the root block for accesslist namefiles and the index list.
// WFSAccessListControlBlock{} in area (AreaInfo{}) must be properly initialized before calling this function.
// - Memory blocks for accesslist cache must be set in memoryBlock field of WFSAccessListControlBlock{}.
// - The number of blocks must be set in numBlock field.
// - initialized field must be set to FALSE.
WFSKrnResult WFSKrnAccessListOpen(const AreaInfo* area);


//--------------------------------------------------------------------------------------------------------------
// WFSKrnAccessListExit
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
//
// Return Values:
// --------------
//       WFS_RESULT_OK             .. success
//
// Description:
// ------------
// This function is not implemented.
//
WFSKrnResult WFSKrnAccessListExit(const AreaInfo* area);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnDeletePermissions
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
// handle              .. [in] Accesslist handle.
// titleId             .. [in] Title Id to be deleted.
// entryCreatorId      .. [in] Entry creator's Id.
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_INVALID
//       WFSKRN_RESULT_ACL_ERROR
//       WFSKRN_RESULT_ACL_MAX_ENTRIES
//       WFSKRN_RESULT_ACL_FILE
//
// Description:
// ------------
// This function deletes the specified entry in the specified accesslist.

WFSKrnResult WFSKrnDeletePermissions(const AreaInfo* area,
    WFSKrnAclHandle* handle, WFSTitleId entryCreatorId, WFSTitleId titleId);

// Arguments:
// ----------
// area                .. [in] A WFS area.
// handle              .. [in] Accesslist handle.
// pAccessList         .. [out] Buffer for an accesslist.
// nMaxElements        .. [out] Size of the buffer.
//
// Return Values:
// --------------
//       The number of entries.             .. success
//       WFSKRN_RESULT_ACL_BUFFER_TOO_SMALL
//
// Description:
// ------------
// This function gets all entries in the specified accesslist.

s32 WFSKrnGetAccessList(const AreaInfo* area,
    WFSKrnAclHandle handle, WFSAccessListEntry *pAccessList, u32 nMaxElements);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnSetAccessList
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
// handle              .. [inout] Accesslist handle.
// caller              .. [in]
// nFlags              .. [in] Permissions and control levels to be set.
// nMask               .. [in] Mask of permissions.
// entryCreatorId      .. [in] Id of the title which created the entry.
// titleId             .. [in] Title Id of the entry to be modified.
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_INVALID
//       WFSKRN_RESULT_ACL_ERROR
//       WFSKRN_RESULT_ACL_MAX_ENTRIES
//       WFSKRN_RESULT_ACL_FILE
//
// Description:
// ------------
// This function chnages permissions of the specified entry, or creates a new entry.
WFSKrnResult WFSKrnSetAccessList(const AreaInfo* area, WFSKrnAclHandle* handle,
    u32 nFlags, u32 nMask, WFSTitleId entryCreatorId, WFSTitleId titleId);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnReleaseAccessList
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
// handle              .. [in] A handle of accesslist .
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_ACL_ERROR
//       WFSKRN_RESULT_ACL_CACHE
//       WFSKRN_RESULT_ACL_FILE
//
// Description:
// ------------
// This function decrements the reference count of the specified accesslist.
// When the count reduces to zero, the accesslist is removed.
// This is called when deleting a file or directory.
WFSKrnResult WFSKrnReleaseAccessList(const AreaInfo* area, WFSKrnAclHandle handle);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnAccessListCheckDirectoryPermission
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] WFS area.
// handle              .. [in] A handle of an accesslist for a directory.
// titleId             .. [in] Title ID.
// flag                .. [in] Permissions to be checked.
//
// Return Values:
// --------------
//       TRUE          .. The title has the specified permissions.
//       FALSE         .. The title does not have the permissions.
//
// Description:
// ------------
// This function checks if the title has the specified directory permissions.
//
// WFSBool WFSKrnAccessListCheckDirectoryPermission(const AreaInfo* area, WFSKrnAclHandle handle, WFSTitleId titleId, u32 flag);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnAccessListCheckDirectoryPermission
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] WFS area.
// handle              .. [in] A handle of an accesslist for a file.
// titleId             .. [in] Title ID.
// flag                .. [in] Permissions to be checked.
//
// Return Values:
// --------------
//       TRUE          .. The title has the specified permissions.
//       FALSE         .. The title does not have the permissions.
//
// Description:
// ------------
// This function checks if the title has the specified file permissions.
// WFSBool WFSKrnAccessListCheckFilePermission(const AreaInfo* area,
//     WFSKrnAclHandle handle, WFSTitleId titleId, u32 flag);

//
// Memory block allocator for accesslist cache.
//
WFSKrnAccessListBlock* WFSKrnMemoryBlockAlloc(const AreaInfo* area, int numBlocks);
int WFSKrnMemoryBlockFree(const AreaInfo* area, WFSKrnAccessListBlock* block, u32 numBlocks);

#endif // __WFSKRN_PERMISSION_ACCESSLIST_H__
