/*---------------------------------------------------------------------------*
  Project:  wfskrn
  File:     wfs_Permission.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_File.h,v $
  Revision 1.9  2008/07/23 04:18:07  ueno
  Reduced memory size required for accesslist.

  Revision 1.8  2008/05/23 06:31:52  ueno
  Cleanup.

  Revision 1.7  2008/05/22 06:36:01  ueno
  Ported permission APIs to IOP.

  Revision 1.6  2008/05/19 10:12:05  ueno
  Revised not to use hardlink.

  Revision 1.5  2008/05/15 13:33:46  ueno
  Implemented initialization using directory module.

  Revision 1.4  2008/04/29 06:25:14  ueno
  Cleanup.

  Revision 1.3  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.2  2008/04/25 01:22:46  ueno
  Moved from RVL_SDK to RM_SDK.

  Revision 1.12  2008/04/23 08:59:55  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.11  2008/04/21 06:07:55  ueno
  Fixed struct for C99.

  Revision 1.10  2008/04/15 06:45:48  ueno
  Changed return type to WFSKrnResult.

  Revision 1.9  2008/04/11 06:16:33  ueno
  Defined WFSAclFileHandle.
  Cleaned up headers.

  Revision 1.8  2008/03/14 08:02:49  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.7  2008/03/11 12:10:35  ueno
  Cleanup.

  Revision 1.6  2008/02/29 07:28:45  ueno
  Added Windows api wrappers.

  Revision 1.5  2008/02/27 13:12:23  ueno
  Revised to store a filename of an accesslist in a file (*.name).

  Revision 1.4  2008/02/27 10:05:26  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.3  2008/02/26 14:12:06  ueno
  Added WFSAccessListAddEntry() and WFSAccessListRemoveEntry() to edit accesslist.

  Revision 1.2  2008/02/25 09:44:08  ueno
  Added support for accesslist cache.

  Revision 1.1  2008/02/22 05:15:56  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#ifndef __WFSKRN_PERMISSION_FILE_H__
#define __WFSKRN_PERMISSION_FILE_H__

#include <revolution/wfs.h>
#include "wfssrv_Defs.h"
#include "wfskrn_Area.h"
#include "wfskrn_Permission_Type.h"

#ifndef WFSDEV
#define IndexDirInfo(x) (((x)->accessListCb->indexDirInfo))
#endif // WFSDEV

int WFSKrnConvertAccessListToFilename(const WFSKrnAccessList* list, utf8* filename, u32 size);
int WFSKrnConvertFilenameToAccessList(const utf8* filename, u32 size, WFSKrnAccessList* list);

int WFSKrnAccessListIndexListInit(const AreaInfo* area);
int WFSKrnAccessListIndexListExit(const AreaInfo* area);
int WFSKrnAccessListCreateFile(const AreaInfo* area, const utf8* filename, WFSKrnAccessList** list);
WFSKrnResult WFSKrnAccessListDeleteFile(const AreaInfo* area, u32 index);
WFSKrnAccessList* WFSKrnAccessListLookupFile(const AreaInfo* area, u32 index);
WFSKrnResult WFSKrnAccessListSetFileReferenceCount(const AreaInfo* area, u32 index, u32 ref);
WFSKrnResult WFSKrnAccessListGetFileReferenceCount(const AreaInfo* area, u32 index, u32* ref);
int WFSKrnAccessListOpenList(const AreaInfo* area, const utf8* filename, AclFileInfo* aclFile,
    WFSBool create, WFSAclFileHandle *pFh);

int GetAccessListIndex(const AreaInfo* area, WFSAclFileHandle indexListHandle, WFSKrnAclHandle handle, IndexListEntry* entry);
int OpenIndexList(const AreaInfo* area, WFSAclFileHandle* indexListHandle, AclFileInfo* indexListInfo);
int CloseIndexList(const AreaInfo* area, WFSAclFileHandle indexListHandle);

#endif // __WFSKRN_PERMISSION_FILE_H__
