/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfs_Permission_Iop.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_Iop.h,v $
  Revision 1.7  2008/09/24 07:32:07  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.6  2008/07/23 04:18:07  ueno
  Reduced memory size required for accesslist.

  Revision 1.5  2008/05/22 06:36:01  ueno
  Ported permission APIs to IOP.

  Revision 1.4  2008/05/19 10:12:05  ueno
  Revised not to use hardlink.

  Revision 1.3  2008/05/19 05:31:53  ueno
  Defined accesslist meta-data types.

  Revision 1.2  2008/05/15 13:33:46  ueno
  Implemented initialization using directory module.

  Revision 1.1  2008/05/06 07:38:27  ueno
  Initial Check-in.

  Revision 1.6  2008/04/29 06:25:14  ueno
  Cleanup.

  Revision 1.5  2008/04/29 01:12:54  ueno
  Defined WFS_ROOT_ID and WFS_EVERYONE_ID.

  Revision 1.4  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.3  2008/04/25 04:40:47  ueno
  Added support for WFSDEV.

  Revision 1.10  2008/04/23 08:59:55  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.9  2008/04/11 06:16:33  ueno
  Defined WFSAclFileHandle.
  Cleaned up headers.

  Revision 1.8  2008/04/07 09:17:44  ueno
  Modified WFSKrnAccessListDirectoryInit() to return an integer.

  Revision 1.7  2008/04/03 09:59:41  ueno
  Cleanup.

  Revision 1.6  2008/03/14 08:02:49  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.5  2008/02/29 07:28:45  ueno
  Added Windows api wrappers.

  Revision 1.4  2008/02/28 13:11:58  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.3  2008/02/27 13:12:24  ueno
  Revised to store a filename of an accesslist in a file (*.name).

  Revision 1.2  2008/02/27 10:05:26  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.1  2008/02/22 05:15:56  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#ifndef __WFSKRN_PERMISSION_IOP_H__
#define __WFSKRN_PERMISSION_IOP_H__

#include <revolution/wfs.h>
#include "wfssrv_Defs.h"
#include "wfskrn_Area.h"
#include "wfskrn_Permission_Type.h"

#define INDEXLIST_FILENAME  "index"
#define ACL_NAMEFILE_LEN    14 // including '\0'.

#define AclBlock(x)  (((x)->ah.aclHdr.aclBlock))
#define NameBlock(x) (((x)->ah.aclHdr.nameBlock))

#define ACL_FILEINFO_NAME(aclFileInfo)   ((aclFileInfo)->di.name.sStr)

int CreateNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const AclNameFileHeader* header, const utf8* filename);
int DeleteNamefile(const AreaInfo* area, u32 index);
int ReadNamefile(const AreaInfo* area, WFSKrnAclHandle handle, void* buf, WFSFileSize size, WFSFileSize offset);
int WriteNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const void* buf, WFSFileSize size, WFSFileSize offset);

int WFSKrnAccessListDirectoryInit(AreaInfo* area);

int WFSKrnAccessListGetAclFileInfo(const AreaInfo* area, const utf8* filename, u32 type, AclFileInfo* fileInfo);

int GetAccessListPath(const AreaInfo* area, AclFileInfo* aclFile);

//
// Wrapper
//

WFSResult AclGetFileSize(const AreaInfo* area, WFSAclFileHandle fh, WFSFileSize* fileSize);
WFSResult AclWriteFile(const AreaInfo* area, WFSAclFileHandle fh, const void *pFileData, WFSFileSize nSize, WFSFileSize offset);
s32 AclReadFile(const AreaInfo* area, WFSAclFileHandle fh, void *pFileDataBuffer, s32 nSize, WFSFileSize offset);

WFSResult AclCreateFile(const AreaInfo* area, const AclFileInfo* fileInfo, WFSBool create, WFSAclFileHandle *pFh);
WFSResult AclCloseFile(WFSAclFileHandle fh);

WFSResult AclDeleteAccessListFile(const AreaInfo* area, const AclFileInfo* aclFile);
WFSResult AclCreateAccessListFile(const AreaInfo* area, const AclFileInfo* aclFile, WFSBool create, WFSAclFileHandle *pFh);
WFSBool AclAccessListAlreadyExist(const AreaInfo* area, const AclFileInfo* aclFile);

#endif // __WFSKRN_PERMISSION_IOP_H__
