/*---------------------------------------------------------------------------*
  Project:  wfskrn
  File:     wfs_Permission_Private.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_Private.h,v $
  Revision 1.8  2008/06/05 14:04:11  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.7  2008/06/05 08:09:18  ueno
  Revised permission APIs.

  Revision 1.6  2008/05/28 01:06:52  ueno
  Added AclCheckDisk() to check acl meta-data.

  Revision 1.5  2008/05/23 06:31:52  ueno
  Cleanup.

  Revision 1.4  2008/05/22 06:36:01  ueno
  Ported permission APIs to IOP.

  Revision 1.3  2008/04/29 01:12:54  ueno
  Defined WFS_ROOT_ID and WFS_EVERYONE_ID.

  Revision 1.2  2008/04/25 01:22:46  ueno
  Moved from RVL_SDK to RM_SDK.

  Revision 1.11  2008/04/23 08:59:55  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.10  2008/04/14 03:02:16  ueno
  Modified WFSKrnCreateRootAccessList() to create an accesslist only for the root.
  Added WFSKrnDebugCreateAccessList() for tests.

  Revision 1.9  2008/04/11 06:16:33  ueno
  Defined WFSAclFileHandle.
  Cleaned up headers.

  Revision 1.8  2008/04/11 04:54:06  ueno
  Added WFSKRN_ACL_USE_MUTEX to switch mutex.

  Revision 1.7  2008/04/09 03:01:27  ueno
  Cleanup.

  Revision 1.6  2008/04/08 06:39:41  ueno
  Cleanup.

  Revision 1.5  2008/04/01 08:25:10  ueno
  Fixed cache reference count.

  Revision 1.4  2008/03/18 13:42:39  ueno
  Added a debug function.

  Revision 1.3  2008/03/14 08:02:49  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.2  2008/03/07 05:31:07  ueno
  Modified WFSCreateAccessList() to be able to create an accesslist for directories.

  Revision 1.1  2008/03/06 10:47:12  ueno
  New.


*---------------------------------------------------------------------------*/
#ifndef __WFSKRN_PERMISSION_PRIVATE_H__
#define __WFSKRN_PERMISSION_PRIVATE_H__

#include <revolution/wfs.h>
#include "wfssrv_Defs.h"
#include "wfskrn_Area.h"

// private APIs
#ifdef WFSDEV
int AclUtf8ToWchar(const utf8* utf8, LPWSTR wchar, u32 bytes);
int AclWcharToUtf8(const LPWSTR wchar, utf8* utf8, u32 bytes);
#endif // WFSDEV

void WFSKrnAccessListSetNum(WFSKrnAccessList* list, u32 num);
u32 WFSKrnAccessListGetNum(const WFSKrnAccessList* list);

WFSKrnResult WFSKrnAccessListAlloc(const AreaInfo* area, u32 numEntries, WFSKrnAccessList** list);
int WFSKrnAccessListFree(const AreaInfo* area, WFSKrnAccessList* list);

int WFSKrnAllocAccessListIndex(const AreaInfo* area, u32 indexFlags, u32* newIndex, u32 num);
int WFSKrnFreeAccessListIndex(const AreaInfo* area, u32 index);

void WFSKrnAccessListCacheAdd(const AreaInfo* area, WFSKrnAccessList* list);
void WFSKrnAccessListCacheRemove(const AreaInfo* area, WFSKrnAccessList* list);
void WFSKrnAccessListCacheFreeListRemove(const AreaInfo* area, WFSKrnAccessList* list);
WFSKrnAccessList* WFSKrnAccessListLookup(const AreaInfo* area, u32 index);
WFSKrnAccessList* WFSKrnAccessListLookupByIndex(const AreaInfo* area, u32 index);

// debug functions.
WFSBool VerifyMemoryBlock(const AreaInfo* area);
void PrintMemoryBlock(const AreaInfo* area);
void WFSKrnDebugPrintAccessList(const AreaInfo* area, u32 index);
s32 AclDirList(const AreaInfo* area, WFSBlkAdr block);
WFSKrnResult AclCheckDisk(const AreaInfo* area);

//--------------------------------------------------------------------------------------------------------------
// WFSKrnDebugCreateAccessList
//--------------------------------------------------------------------------------------------------------------
// Arguments:
// ----------
// area                .. [in] A WFS area.
// aclType             .. [in] ACL_FLAG_DIRECTORY (directory) or 0 (file).
// titleId             .. [in] The first title in the accesslist to be created.
// nFlags              .. [in] permissions and control levels (WFS_PERM_*)
// newHandle           .. [out] Pointer to the new accesslist handle.
//
// Return Values:
// --------------
//       WFSKRN_RESULT_OK             .. success
//       WFSKRN_RESULT_INVALID
//       WFSKRN_RESULT_ACL_ERROR
//       WFSKRN_RESULT_ACL_MAX_ENTRIES
//       WFSKRN_RESULT_ACL_FILE
//
// Description:
// ------------
// This function creates an accesslist.
WFSKrnResult WFSKrnDebugCreateAccessList(const AreaInfo* area, u32 aclType,
    WFSTitleId entryCreatorId, WFSTitleId titleId, u32 nFlags, WFSKrnAclHandle* newHandle);

#endif // __WFSKRN_PERMISSION_PRIVATE_H__
