/*---------------------------------------------------------------------------*
  Project:  wfskrn
  File:     wfs_Permission_Type.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_Type.h,v $
  Revision 1.12  2008/07/23 04:18:07  ueno
  Reduced memory size required for accesslist.

  Revision 1.11  2008/07/23 02:01:51  ueno
  Modified to use mutex on Windows.

  Revision 1.10  2008/05/30 08:50:35  ueno
  Disabled mutex for IOP.

  Revision 1.9  2008/05/23 06:31:52  ueno
  Cleanup.

  Revision 1.8  2008/05/22 06:36:01  ueno
  Ported permission APIs to IOP.

  Revision 1.7  2008/05/19 10:12:05  ueno
  Revised not to use hardlink.

  Revision 1.6  2008/05/19 05:31:53  ueno
  Defined accesslist meta-data types.

  Revision 1.5  2008/05/16 11:16:58  ueno
  Implemented permission APIs with directory module.

  Revision 1.4  2008/05/15 13:33:46  ueno
  Implemented initialization using directory module.

  Revision 1.3  2008/05/06 02:27:21  ueno
  Fixed ifdef.

  Revision 1.2  2008/04/29 06:25:14  ueno
  Cleanup.

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.1  2008/04/11 06:16:33  ueno
  Defined WFSAclFileHandle.
  Cleaned up headers.



*---------------------------------------------------------------------------*/
#ifndef __WFSKRN_PERMISSION_TYPE_H__
#define __WFSKRN_PERMISSION_TYPE_H__

#include <revolution/wfs.h>
#include "wfskrn_Dir.h"
#include "wfskrn_Handles.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"

// access list handle
typedef u32 WFSKrnAclHandle;

#if _WIN32
// The maximum path length of an emulation directory is WFSKRN_MAX_PATH_NAME_SIZE.
#define WFSKRN_ACCESSLIST_MAX_PATH_LEN ((WFS_MAX_PATH_NAME_SIZE+1)*2)
#else
#define WFSKRN_ACCESSLIST_MAX_PATH_LEN (WFS_MAX_PATH_NAME_SIZE+1)
#endif // _WIN32

// type of metadata.
#define ACL_TYPE_INDEXLIST   1
#define ACL_TYPE_NAMEFILE    2
#define ACL_TYPE_ACLFILE     3

typedef struct AclFileInfo
{
    u32 type; // ACL_TYPE_*
#ifdef WFSDEV
    utf8 path[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
#else
    DirItr di;
#endif // WIN32
} AclFileInfo;

typedef struct AclNameFileHeader
{
    u32 num;    // The number of titles in this list.
    u32 ref;    // The number of files and directories associated with this accesslist.
} AclNameFileHeader;

// file handle
#ifdef WFSDEV
typedef HANDLE WFSAclFileHandle;
#else // WFSDEV
typedef AclFileInfo* WFSAclFileHandle;
#endif // WFSDEV

// mutex
// #define WFSKRN_ACL_USE_MUTEX // [check] how to use mutex on IOP?

#ifdef WFSKRN_ACL_USE_MUTEX
#include "wfskrn_Mutex.h"
#define WFSAclInitMutex(x)    WFSKrnInitMutex(x)
#define WFSAclLockMutex(x)    WFSKrnLockMutex(x)
#define WFSAclUnlockMutex(x)  WFSKrnUnlockMutex(x)
#else
#define WFSAclInitMutex(x)
#define WFSAclLockMutex(x)
#define WFSAclUnlockMutex(x)
#endif

// print message
#define WFSKRN_ACL_DEBUG

#ifdef WFSKRN_ACL_DEBUG

#ifdef WIN32
#define WFSKRN_ACL_ASSERT(x)  assert(x)
#else  // Win32
#define WFSKRN_ACL_ASSERT(x)  ASSERT(x)
#endif // WIN32
#define WFSKRN_ACL_REPORT     printf

#else // WFSKRN_ACL_DEBUG

#define WFSKRN_ACL_ASSERT(x)
#define WFSKRN_ACL_REPORT

#endif // WFSKRN_ACL_DEBUG

#endif // __WFSKRN_PERMISSION_TYPE_H__
