/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Platform.h - low level definitions and types that are primarily platform dependent

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Platform.h,v $
  Revision 1.6  2008/08/27 04:05:21  nakanose_jin
  change ospanic definision

  Revision 1.5  2008/07/14 01:01:11  kondo_masahiro
  Removed _STRICMP

  Revision 1.4  2008/06/26 05:37:33  kondo_masahiro
  Added function _STRICMP().

  Revision 1.3  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:26:30  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.4  2008/04/26 05:36:11  paul
  Added _ATTRIBUTE_PACKED

  Revision 1.3  2008/04/19 05:47:36  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2008/02/28 06:32:17  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.1  2007/11/05 19:03:13  paul
  first checkin

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Platform_h
#define wfskrn_Platform_h

#include <firm.h>

//--twl modified
#define TWL 1
#if 1
    #define _OSX   0
    #define _WIN32 0
    #define _TWL   1
    #define _RVL   0
    #define _IOP   0
#else
//--twl modified
#ifdef OSX
    #define _OSX   1
    #define _WIN32 0
    #define _RVL   0
    #define _IOP   0
#else
 #ifdef WIN32
    #define _OSX   0
    #define _WIN32 1
    #define _RVL   0
    #define _IOP   0
 #else
  #ifdef IOP
    #define _OSX   0
    #define _WIN32 0
    #define _RVL   0
    #define _IOP   1
  #elif RVL
    #define _OSX   0
    #define _WIN32 0
    #define _RVL   1
    #define _IOP   0
  #endif
 #endif
#endif
#endif //twl modified

// Although the cache line size depends on the platform, it is fixed to
// the IOP cache line size, since the goal is not to optimize for different
// platforms, but rather to use other platforms as a development aid.
#define WFS_CPU_CACHE_LINE_SIZE        (1<<5)

#if _WIN32
    #pragma warning (3: 4101)
    #pragma warning (3: 4189)
    #pragma warning(disable:4996)
    #define _ATTRIBUTE_PACKED
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define MAX max
    #define MIN min
    //#define OSReport printf
    #include "WinOSReport.h"
    #define OSPanic  MyOSPanic
    #define SNPRINTF _snprintf
    #define STRNLEN( str, maxlen ) strlen(str)
    #define _ATTRIBUTE_PACKED
#elif  _OSX
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    //#define OSReport printf
    #include "WinOSReport.h"
    #define OSPanic  MyOSReport
    #define SNPRINTF snprintf
    #define STRNLEN( str, maxlen ) strlen(str)
    #define _ATTRIBUTE_PACKED
//-- twl modified
#elif  _TWL
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define MAX(a,b)                       (((a)>(b))?(a):(b))
    #define MIN(a,b)                       (((a)<(b))?(a):(b))
    #define MyOSReport                     osTPrintf
    #define InitMyOSReport()
    #define SNPRINTF OS_SNPrintf
    #define STRNLEN( str, maxlen ) stdStrNLen(str, maxlen)
    #define _ATTRIBUTE_PACKED
//-- twl modified
#elif  _RVL
    #pragma warn_unusedarg off
    #pragma warn_implicitconv off
    #pragma warn_possunwant off
    #include <ctype.h>
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define MAX(a,b)                       (((a)>(b))?(a):(b))
    #define MIN(a,b)                       (((a)<(b))?(a):(b))
    #define MyOSReport                     OSReport
    #define InitMyOSReport()
    #define SNPRINTF snprintf
    #define STRNLEN( str, maxlen ) strlen(str)
    #define _ATTRIBUTE_PACKED
#elif  _IOP
    #define BYTES_PER_INT                  4
    #define BITS_PER_INT                   32
    #define BIT_OFFSET_MASK                31
    #define BIT_OFFSET_TO_INT_OFFSET_SHIFT 5
    #define MAX(a,b)                       (((a)>(b))?(a):(b))
    #define MIN(a,b)                       (((a)<(b))?(a):(b))
    #define MyOSReport                     printf
    #define InitMyOSReport()
    #define SNPRINTF snprintf
    #define STRNLEN( str, maxlen ) strnlen( str, maxlen )
    #define _ATTRIBUTE_PACKED              __attribute__((packed))
#endif // WIN32

#ifndef BIG_ENDIAN
    #ifdef WIN32
        #define BIG_ENDIAN 0
    #elif TWL
        #define BIG_ENDIAN 0
    #else
        #define BIG_ENDIAN 1
    #endif // WIN32
#endif // BIG_ENDIAN


#endif //define wfskrn_Platform_h
