/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_SubBlkAlloc.h - Memory Managed Blocks, and functions for allocating sub blocks from them

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_SubBlkAlloc.h,v $
  Revision 1.6  2008/09/28 23:30:46  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.5  2008/08/27 10:01:32  paul
  Removed nTotalFree from block header

  Revision 1.4  2008/07/14 23:03:11  paul
  Changed@nOfs to use u16

  Revision 1.3  2008/06/09 18:11:14  paul
  Changed SbaBlkInit() to support blocks with different header sizes

  Revision 1.2  2008/04/25 17:26:24  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/19 05:47:18  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2008/04/05 03:23:05  paul
  Added extra check functions

  Revision 1.1  2008/02/21 00:06:19  paul
  *** empty log message ***

*---------------------------------------------------------------------------*/

#ifndef wfskrn_SubBlkAlloc_h
#define wfskrn_SubBlkAlloc_h


#include "wfskrn_Defs.h"
#include "wfskrn_Area.h"

#ifdef _DEBUG
    #define _DEBUG_SUB_BLK_ALLOC 0  // change to 1 if debug info required
#else
    #define _DEBUG_SUB_BLK_ALLOC 0
#endif

#define SBA_MIN_LOG2_SUB_BLK_SIZE             3
#define SBA_MAX_LOG2_SUB_BLK_SIZE            11
#define SBA_MAX_LOG2_SUB_BLK_SIZE_FOR_MIDDLE 10
#define SBA_NUM_SUB_BLK_SIZES      (SBA_MAX_LOG2_SUB_BLK_SIZE - SBA_MIN_LOG2_SUB_BLK_SIZE + 1)

//#define SBA_FREE_SUB_BLK_MAGIC  0xFEDCBAF8
#define SBA_FREE_SUB_BLK_MAGIC  0xFEDC // NOTE: This magic number is used to mark free sub blocks. It must not represent valid data for any in-use sub blocks using this allocator.

typedef struct {
    WFSIntraBlkOfs  nNumFree;
    WFSIntraBlkOfs  nFirstFree;
} WFSSubBlkList;

typedef struct {
    WFSMetaDataHdr  mdh;
    WFSSubBlkList   aSubBlkList[SBA_NUM_SUB_BLK_SIZES];
#if _DEBUG_SUB_BLK_ALLOC
    WFSIntraBlkOfs  nTotalFree;
    u16             pad;
#endif
} WFSSubBlkAllocHdr;

typedef struct {
    u16             nMagic;
    u16             nNext;
    u16             nPrev;
    u16             nLog2Size;
} WFSSubBlkFreeHdr;

void SbaCheckOfs(u16 nOfs);
void SbaRemoveSubBlkFromFreeList(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nLog2Size);
u16 SbaAllocSubBlk(WFSSubBlkAllocHdr *pBlkHdr, u32 nLog2Size);
WFSKrnResult SbaAllocSubBlks(WFSSubBlkAllocHdr *pBlkHdr, u32 nLog2Size, u16 aSubBlkOfs[], u32 nNumSubBlks);
void SbaFreeSubBlk(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nLog2Size);
void SbaReduceSubBlkSize(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nOldLog2Size, u32 nNewLog2Size);
WFSSubBlkAllocHdr *SbaBlkInit(void *pBlk, u32 nLog2BlkSize, u32 nLog2HdrSize);
u32 SbaCountFreeSpace(WFSSubBlkAllocHdr *pBlkHdr);
u32 SbaCheckFreeSpace(WFSSubBlkAllocHdr *pBlkHdr, u32 *aUtilization);


#endif //wfskrn_SubBlkAlloc_h
