/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Volume.h - handle volume-wide management of area allocation

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Volume.h,v $
  Revision 1.4  2008/07/17 05:21:49  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.3  2008/07/09 01:54:58  paul
  Many changes to Volume API to support mounting in namespace, and mounting root directory from an existing volume.

  Revision 1.2  2008/05/12 19:24:00  paul
  Merged in Wayne's previous changes

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/26 01:26:41  wayne.wong
  Moved area allocation functionality from Volume to Area module.
  Changed API appropriately.

  Revision 1.2  2008/03/19 07:41:22  wayne.wong
  Checkpoint. Added initial alloc/free area functions.

  Revision 1.1  2008/03/11 03:22:09  wayne.wong
  Checkpoint. Preliminary version.


*---------------------------------------------------------------------------*/

#ifndef _WFSKRN_VOLUME_H_
#define _WFSKRN_VOLUME_H_

#include "wfskrn_Device.h"

#define VOLUME_HDR_BLK_ADR  0  // Absolute block address of the volume header

typedef struct VolumeHdr_ {
    // This is the part of the VolumeInfo struct which gets stored on disk
    WFSMetaDataHdr  mdh;
    u32             nVolId;         // Randomly assigned 32 bit identifier
    u16             nWfsVersion;    // WFS version; major upper 8b, minor lower 8b
    u16             nBuildDate;     // Build date (in days since 2008.01.01)
} VolumeHdr;

void                VolumeModuleInit();
struct VolumeInfo_ *VolumeAlloc();
void                VolumeFree(struct VolumeInfo_ *pVolInfo);
void                VolumeRemoveFromMountedList(struct VolumeInfo_ *pVolInfo);
void                VolumeAddToMountedList(struct VolumeInfo_ *pVolInfo);
void                VolumeConvertU32ToVolId(WFSVolumeId *pVolId, u32 nVolId);

WFSKrnResult        VolumeInitRootArea(struct VolumeInfo_ *pVolInfo, u8 nLog2BlkSize);
WFSKrnResult        VolumeInit        (struct VolumeInfo_ *pVolInfo, u32 nLog2BlkSize, u32 nNumBlks);
WFSKrnResult        VolumeUnmount     (struct VolumeInfo_ *pVolInfo);

//#define _DEBUG_WFSKRN_VOLUME_

#ifdef _DEBUG_WFSKRN_VOLUME_
#define dbgv(_s)                 (_s)
#else
#define dbgv(_s)
#endif // _DEBUG_WFSKRN_VOLUME_

#endif // _WFSKRN_VOLUME_H_
