/*---------------------------------------------------------------------------*
  Project:  Revolution WFSSrv library API = shim layer in RVL version and RPC
            layer in WFSDEV version.
  File:     wfsdev_Defs.h - Definitions common to WFS client and server

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfssrv_Defs.h,v $
  Revision 1.34  2008/12/03 00:17:36  kondo_masahiro
  Added a argument of WFSSrvFlush().

  Revision 1.33  2008/10/21 10:45:09  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.32  2008/09/29 01:06:26  kondo_masahiro
  Added the argument of WFSSrvInit().

  Revision 1.31  2008/07/18 02:59:44  ueno
  Renamed WFSDeletePermissions() -> WFSDeleteAccessListEntry().

  Revision 1.30  2008/07/17 05:08:41  kondo_masahiro
  Added WFSSRV_ATTACH_DETACH

  Revision 1.29  2008/07/15 08:43:53  nakanose_jin
  change arg order

  Revision 1.28  2008/07/14 07:15:46  ueno
  Implemented WFSGetAccessListAsync(), WFSSetAccessListEntryAsync() and WFSDeletePermissionsAsync().

  Revision 1.27  2008/07/10 06:12:33  nakanose_jin
  add permissions API

  Revision 1.26  2008/05/22 08:05:49  ueno
  Added WFSSrvDebugAclTest to run accesslist test.

  Revision 1.25  2008/05/09 14:03:09  nakanose_jin
  canceling => reduce settitleid argument

  Revision 1.23  2008/04/19 05:52:03  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.22  2008/04/15 02:47:27  ueno
  Added WFSSrvGetPermissionForUser() and WFSSrvSetPermissionForUser().

  Revision 1.21  2008/03/28 14:14:20  nakanose_jin
  chance spec of debug proc

  Revision 1.20  2008/02/20 08:10:10  nakanose_jin
  add WFSExecDebugProcedure

  Revision 1.19  2008/02/08 04:42:00  paul
  Changed WFSSrvUnmountVolume parameter bForce to WFSBool

  Revision 1.18  2008/02/07 13:02:18  nakanose_jin
  WFSUnmountVolume : Add force-close option (but only in detach condition)

  Revision 1.17  2008/01/10 05:33:20  nakanose_jin
  validatePathName => validateDirectory

  Revision 1.16  2007/12/25 14:00:36  ueno
  Added WFSSrvConnectToServer() to connect to the server again.

  Revision 1.15  2007/12/12 07:51:29  paul
  Added WFSSrvGetFileSize

  Revision 1.14  2007/12/10 02:51:36  paul
  Added function codes for WFSSRV_SET_HOME_DIRECTORY, WFSSRV_SET_CURRENT_DIRECTORY. (These do not need to be implemented in the actual server, both call WFSSrvValidatePathName)

  Revision 1.13  2007/11/30 09:43:09  nakanose_jin
  refine for multi client

  Revision 1.12  2007/11/21 04:17:28  wayne.wong
  Many changes that included single request/result RPC messages, bug fixes, code factoring, and English C comment adjustments.
    1. Combined RPC messages into a single request message and single result
        message. Except for file reads and writes.
    2. Fixed+factored some Endianess code.
    3. Enabled larger TCP/IP data transfer size. Previous setting was artifically small
        to test the code to perform large transfers.
    4. Changed the volume id parameters utf8*->WFSVolumeId for SrvGetVolumeId
        and SrvMountDevice. More consistent with the way we send this type at the
        RPC layer.
    5. Changed a couple Japanese comments to be compatible with English
        comments. In particular, the "//" comment with a backslash  at the end of
        the line causes the next line of code to be ignored.
    6. Removed Windows-specific debugging output defines for RVL build.

  Revision 1.11  2007/11/16 21:54:26  paul
  Fixed WFSSrvCloseFile to pass nNewSize instead of bTruncateFlag

  Revision 1.10  2007/11/16 02:15:57  wayne.wong
  Added enumeration and stubs for WFSFlush call.

  Revision 1.9  2007/11/02 00:21:02  paul
  started logging changes in the file
  Added StartPosition to WFSSrvReadFile, WFSSrvWriteFile

 *---------------------------------------------------------------------------*/

#ifndef __WFSSRV_DEFS_H__
#define __WFSSRV_DEFS_H__

#include "wfs_Defs.h"

#ifdef __cplusplus
extern "C" {
#endif

// Enumeration of functions.
// Used by the client to specify functions implemented on the server, or
//      by the server to specify functions implemented on the client.

typedef enum {
  // server -> client

  WFSSRV_ATTACH_DEVICE,
  WFSSRV_DETACH_DEVICE,

  // client -> server

  WFSSRV_INIT,
  WFSSRV_EXIT,

  WFSSRV_GET_DEVICE_INFO,
  WFSSRV_MOUNT_DEVICE,
  WFSSRV_UNMOUNT_VOLUME,
  WFSSRV_INITIALIZE_DEVICE,
  WFSSRV_GET_VOLUME_ID,
  WFSSRV_GET_FREE_SPACE_SIZE,
  WFSSRV_FLUSH,

  WFSSRV_CREATE_DIRECTORY,
  WFSSRV_SEARCH_DIRECTORY_FIRST,
  WFSSRV_SEARCH_DIRECTORY_NEXT,
  WFSSRV_SEARCH_DIRECTORY_CLOSE,

  WFSSRV_SET_HOME_DIRECTORY,
  WFSSRV_SET_CURRENT_DIRECTORY,

  WFSSRV_VALIDATE_PATH_NAME,
  WFSSRV_DELETE,
  WFSSRV_MOVE_LOCAL,
  WFSSRV_GET_ATTRIBUTES,
  WFSSRV_CHANGE_PERMISSIONS,

  WFSSRV_CREATE_AND_OPEN_FILE,
  WFSSRV_OPEN_FILE,
  WFSSRV_GET_FILE_SIZE,
  WFSSRV_CLOSE_FILE,
  WFSSRV_READ_FILE,
  WFSSRV_WRITE_FILE,
  WFSSRV_WRITE_FILE_AND_DISCARD,

  WFSSRV_GET_PERMISSIONS,
  WFSSRV_SET_PERMISSIONS,

  WFSSRV_GET_ACL,
  WFSSRV_SET_ACL_ENTRY,
  WFSSRV_DELETE_ACL_ENTRY,

  WFSSRV_SET_TITLE_ID,
  WFSSRV_TEST,
  WFSSRV_EXEC_DEBUG,

  WFSSRV_ATTACH_DETACH,

  WFSSRV_DEBUG_ACLTEST // [check]

} WFSSrvFunc;


typedef u16 WFSSrvFileHandle;
typedef u16 WFSSrvSearchDirectoryHandle;

#ifndef WFSDEV_SERVER

// The WFSDev server must distinguish between multiple clients.
// This can be done by assigning a ClientId based on the client's IP address,
// which the server learns whenever it receives a request from a client via sockets.

// Note: ClientId is orthogonal to TitleId.
// There can be multiple clients with the same TitleId.
// The same client could change TitleId while the server is running.

// The home directory is only stored on the client side.
// The TitleId, however, is determined by a function in IOS.
// (except for the fake debug/wfsdev version)
//
// This allows a system application built on top of wfs to specify the home directory for
// an app prior to launching it.
//

// Path Names
// The client<->server API uses absolute WFS path names, becuase the server does not know about
// "Current Directory" or "Home Directory".
// The absolute path name does not include the full windows path name, it is a WFS path name.

// For example: "/vol/ABCDEFG/titles/1234/JAKK/gamesave/save1.bin"
// On windows, this could map to:
// "C:/my_project/my_wfs_device/titles/1234/JAKK/gamesave/save1.bin"
// There would be a file called:
// "C:/my_project/my_wfs_device/.Wfs" which contains: "Volume Name: ABCDEFG"


WFSResult WFSSrvInit(void *pWorkSpaceBuffer, u32 nWorkSpaceSize);
// Contact the WFSSrv server so that it can initialize state for this client.


WFSResult WFSSrvExit(void);
// Inform the WFSSrv server that this client is terminating.


WFSResult WFSSrvInitializeDevice(const WFSDeviceName *pDevName);
// pDevName            (IN)  pointer to device name struct.
// Server would create a volume (deleting any existing data within the device assigned to this folder).
// The server could have a GUI to allow device attach to be simulated.
// It depends on the usage model but this may not be necessary for game developers, so low priority.
// (Assuming a system app mounts the drive and assigns home directory in advance, so the
// game can access an already-mounted device from the start).


WFSResult WFSSrvGetDeviceInfo(const WFSDeviceName *pDevName, WFSDeviceInfo *pDi);
// pDevName            (IN)  A pointer to a device name struct.
// pDi                 (OUT) A pointer to a WFSDeviceInfo struct to receive the device information.


WFSResult WFSSrvMountDevice(const WFSDeviceName *pDevName, WFSVolumeId *pVolumeIdBuffer);
// pDevName            (IN)  A pointer to a device name struct.
// pVolumeIdBuffer     (OUT) A pointer to a WFSVolumeId buffer

WFSResult WFSSrvUnmountVolume(const WFSVolumeId *pVolumeId, WFSBool bForce);
// pVolumeId           (IN)  A pointer to a struct containing a mounted Volume Id.
// bForce              (IN)  If set to true, forces the volume to be unmounted regardless of open files.
//                           NOTE: This flag is only valid if the device is currently detached. Un-flushed data will be lost.

WFSResult WFSSrvGetVolumeId(const WFSDeviceName *pDevName, WFSVolumeId *sVolumeIdBuffer);
// pDevName            (IN)  A pointer to a device name struct.
// sVolumeIdBuffer     (OUT) A pointer to a WFSVolumeId buffer


WFSResult WFSSrvGetFreeSpaceSize(const WFSPathName *pAbsPathName, u64 *pSize);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of a file or directory.
// pSize               (OUT) A pointer to a u64 to receive the free space size value.


WFSResult WFSSrvFlush(const WFSVolumeId *pVolumeId);


WFSResult WFSSrvGetAttributes(const WFSPathName *pAbsPathName, WFSFileAttributes *pFa);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of a file or directory.
// pFa                 (OUT) A pointer to a WFSFileAttributes struct which will be overwritten with the attributes of the specified file or directory.

WFSResult WFSSrvChangePermissions(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of a file or directory.
// nFlags              (IN)  The new permissions flags that are to be set. Should be a combination of the permission flags: WFS_PERM_*
// nMask               (IN)  A mask to specify which flags to change.


WFSResult WFSSrvCreateDirectory(const WFSPathName *pAbsPathName, u32 nFlags, u64 nQuota);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of a file or directory.
// nFlags              (IN)  Should be a combination of the permission flags: WFS_PERM_*. Note: WFS_FLAG_DIRECTORY will be set
//                              whether it is included or not.
// nQuota              (IN)  This parameter that can be used to restrict the total size of the hierarchy beneath the
//                              new directory. A quota of 0 means there is no size restriction, i.e. the quota is Infinity.


WFSResult WFSSrvSearchDirectoryFirst(const WFSPathName *pPattern, WFSSrvSearchDirectoryHandle *pSsdh, WFSFileInfo *pFi);
// pPattern            (IN)  A pointer to a struct specifying the search pattern as an absolute path name.
// pSsdh               (OUT) A pointer to a WFSSrvSearchDirectoryHandle handle which is overwritten by the server.
// pFi                 (OUT) A pointer to a WFSFileInfo struct which will be overwritten with the details of the


WFSResult WFSSrvSearchDirectoryNext(WFSSrvSearchDirectoryHandle ssdh, WFSFileInfo *pFi);
// ssdh                (IN)  A WFSSrvSearchDirectoryHandle which was created by WFSSrvSearchDirectoryFirst.
// pFi                 (OUT) A pointer to a WFSFileInfo struct which will be overwritten with the details of the
//                              next file which matches the search pattern.
//
// If they have read permission on the directory, they can list all files/dirs in that directory. Directory read
// access was already checked by WFSSrvSearchDirectoryFirst.  While they can list the directory contents, they may
// not be able to access all contained files/dirs.


WFSResult WFSSrvSearchDirectoryClose(WFSSrvSearchDirectoryHandle ssdh);
// ssdh                (IN)  A WFSSrvSearchDirectoryHandle which was returned by WFSSrvSearchDirectoryFirst.


WFSResult WFSSrvValidateDirectory(const WFSPathName *pAbsPathName);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of a file or directory to validate.
// Returns WFS_RESULT_OK if the specified pathname exists, and is accessible to the current title


WFSResult WFSSrvDelete(const WFSPathName *pAbsPathName);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of the file or directory to delete.

WFSResult WFSSrvMoveLocal(const WFSPathName *pSrcAbsPathName, const WFSPathName *pDstAbsPathName);
// pSrcAbsPathName     (IN)  A pointer to a struct containing the absolute pathname of the source file or directory.
// pDstAbsPathName     (IN)  A pointer to a struct containing the absolute pathname of the destination file or directory.
// Moves or renames the source file/directory to the destination pathname. This function does not work across volumes.


WFSResult WFSSrvCreateAndOpenFile(const WFSPathName *pAbsPathName, u32 nFlags, WFSContentIdx nCidx, WFSFileSize nPreAllocateSize, WFSSrvFileHandle *pSfh);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of the file to create and open.
// nFlags              (IN)  Should be a combination of the permission flags: WFS_PERM_*
// nCidx               (IN)  The content idx, used to control access to purchased content.
// nPreAllocateSize    (IN)  The amount of disk-space to pre-allocate for the file. Hint to the file system.
// pSfh                (OUT) A pointer to a server file handle which is overwritten by the server.

WFSResult WFSSrvOpenFile(const WFSPathName *pAbsPathName, WFSAccess nDesiredAccess, WFSSrvFileHandle *pSfh, WFSFileAttributes *pFa);
// pAbsPathName        (IN)  A pointer to a struct containing the absolute pathname of the file to open.
// nDesiredAccess      (IN)  The kind of access that is required: should be one of the WFS_ACCESS_* options.
// pSfh                (OUT) A pointer to a server file handle, which is overwritten by the server.
// pFa                 (OUT) A pointer to a WFSFileAttributes struct which will be overwritten by the server,

WFSResult WFSSrvCloseFile(WFSSrvFileHandle sfh, WFSFileSize nNewSize);
// sfh                 (IN)  The server file handle which was returned when the file was opened.
// nNewSize            (IN)  This value may be set to truncate the size of a file.

WFSResult WFSSrvGetFileSize(WFSSrvFileHandle sfh, WFSFileSize *pNewSize);
// sfh                 (IN)  The server file handle which was returned when the file was opened.
// pNewSize            (OUT) A pointer to a WFSFileSize variable which is overwritten by the current file size.

s32 WFSSrvReadFile(WFSSrvFileHandle sfh, void *pFileDataBuffer, WFSFileSize nStartPosition, s32 nSize);
// sfh                 (IN)  The server file handle which was returned when the file was opened.
// pFileDataBuffer     (OUT) Points to a buffer which will be overwritten with data from the file.
// nSize               (IN)  The number of bytes to be read.


WFSResult WFSSrvWriteFile(WFSSrvFileHandle sfh, const void *pFileData, WFSFileSize nStartPosition, WFSFileSize nSize, WFSBool bDiscard);
// sfh                 (IN)  The server file handle which was returned when the file was opened.
// pFileData           (IN)  Points to the data to be written to the file.
// nSize               (IN)  The number of bytes to be written.
// bDiscard            (IN)  If it is true, the user buffer is used as work buffer.


WFSResult WFSSrvSetTitleId(WFSTitleId titleId,WFSGroupId groupId);
// titleId             (IN)  the title id which is the low 32 bits of the OSTitleId which can be obtained be ESP_GetTitleId()
// groupId             (IN)  the group id from the TMD, which can be obtained by ES_GetTmd()
// This function informs the server of the title id and group id of the application linked to this instance of the client.

WFSResult WFSSrvExecDebugProcedure( u32 cmd , void *arg0 , void *arg1 );
// cmd                 (IN)      special command No.
// arg0,arg1           (IN/OUT)  parameters of special command.
// This function let the server do special tasks.

WFSResult WFSSrvConnectToServer(void);

WFSResult WFSSrvGetPermissionForUser(const WFSPathName *pAbsPathName, u32 *pFlags, WFSTitleId titleId);
// pAbsPathName      (IN)  A pointer to a struct containing the absolute pathname of the file or the directory.
// pFlags            (OUT) A pointer to a u32 which will be overwritten with the permissions the specified title id has to
//                         access or modify the specified file or directory. See wfstypes.h for WFS_PERM* flag definitions.
// titleId           (IN)  Title id of the application for which permissions are being set.

WFSResult WFSSrvSetPermissionForUser(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask, WFSTitleId titleId);
// pAbsPathName      (IN)  A pointer to a struct containing the absolute pathname of the file or the directory.
// nFlags            (IN)  The new permissions flags that are to be set. Should be a combination of the permission flags: WFS_PERM_* (see wfstypes.h)
// nMask             (IN)  The mask determines which flags will be changed. NewAttributes = (OldAttributes & ~nMask) | (nFlags & nMask)
// titleId           (IN)  Title id of the application for which permissions are being set.

WFSResult WFSSrvDeleteAccessListEntry(const WFSPathName *pAbsPathName, WFSTitleId titleId, WFSTitleId entryCreatorId);
WFSResult WFSSrvSetAccessListEntry(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask, WFSTitleId titleId);
s32       WFSSrvGetAccessList(const WFSPathName *pAbsPathName, WFSAccessListEntry *pAccessList, u32 nMaxElements);



WFSResult WFSSrvSetWorkspace(void *pWorkSpaceBuffer, u32 nWorkSpaceSize);

#ifdef ACL_TEST
WFSResult WFSSrvDebugAclTest(u32 testNo, void *arg0, void *arg1);
#endif // ACL_TEST

#endif

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif //define __WFSSRV_DEFS_H__
