/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Client_Common.cpp - PPC side of the WFS API. Common to WFS and WFSDev


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Client_Common.cpp,v $
  Revision 1.74  2008/12/10 04:42:22  kondo_masahiro
  Added error handling for srcpath include current or home path.

  Revision 1.73  2008/12/04 00:02:09  kondo_masahiro
  Fixed WFSFlushAsync().

  Revision 1.72  2008/12/03 00:16:38  kondo_masahiro
  Fixed WFSFlush().

  Revision 1.71  2008/10/23 11:15:28  nakanose_jin
  Open and Search should give a invalid file handle if he fails to execute.

  Revision 1.70  2008/10/21 10:45:20  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.69  2008/10/21 04:14:55  nakanose_jin
  (none)

  Revision 1.68  2008/10/20 02:53:37  ueno
  Restored WFSExecDebugProcedure() for WFSDEV.

  Revision 1.67  2008/10/20 00:00:22  ueno
  Modified WFSExecDebugProcedure() to support WFSDEBUG_CMD_ATTACH and WFSDEBUG_CMD_DETACH options.

  Revision 1.66  2008/10/14 07:02:41  ueno
  Fixed WFSOpenFileAsync() to be able to return an error.

  Revision 1.65  2008/10/14 06:00:39  ueno
  Fixed WFSCreateAndOpenFileAsync() and WFSOpenFileAsync() to release client file info when WFSPrepareAsyncCallWithPathName() returns an error.

  Revision 1.64  2008/10/07 06:37:01  ueno
  Modified WFSExit() to close all open files.

  Revision 1.63  2008/09/29 01:06:19  kondo_masahiro
  Added the argument of WFSSrvInit().

  Revision 1.62  2008/09/05 02:27:19  ooizumi
  Fixed a bug WFSGetCurrentDirectory returns a broken path name.

  Revision 1.61  2008/08/19 14:05:43  nakanose_jin
  smash pathnames bug

  Revision 1.60  2008/08/08 05:22:29  nakanose_jin
  removed ASSERT and return WFS_RESULT_LIB_NOT_INITIALIZED.

  Revision 1.58  2008/07/18 02:59:46  ueno
  Renamed WFSDeletePermissions() -> WFSDeleteAccessListEntry().

  Revision 1.57  2008/07/15 08:43:57  nakanose_jin
  change arg order

  Revision 1.56  2008/07/14 07:15:23  ueno
  Implemented WFSGetAccessListAsync(), WFSSetAccessListEntryAsync() and WFSDeletePermissionsAsync().

  Revision 1.55  2008/07/11 09:20:37  ueno
  Fixed a typo.

  Revision 1.54  2008/07/10 06:12:37  nakanose_jin
  add permissions API

  Revision 1.52  2008/05/10 04:13:13  kondo_masahiro
  Fixed WFSExecDebugProcedure

  Revision 1.51  2008/05/09 14:03:12  nakanose_jin
  canceling => reduce settitleid argument

  Revision 1.49  2008/04/24 04:26:29  nakanose_jin
  byebye nPermissions param

  Revision 1.48  2008/04/18 07:45:03  nakanose_jin
  merge acl

  Revision 1.47  2008/04/10 13:35:38  ueno
  Added WFSSetPermissionsForUser[Async] and WFSGetPermissionsForUser[Async].

  Revision 1.46  2008/04/04 06:43:29  nakanose_jin
  link error

  Revision 1.45  2008/03/28 14:14:23  nakanose_jin
  chance spec of debug proc

  Revision 1.44  2008/02/25 08:47:32  paul
  Fixed WFSUnmountVolumeAsync

  Revision 1.43  2008/02/25 08:34:22  paul
  BOOL -> WFSBool

  Revision 1.42  2008/02/20 08:10:13  nakanose_jin
  add WFSExecDebugProcedure

  Revision 1.41  2008/02/13 01:49:37  nakanose_jin
  WFSGetCurrentDirectory : NULL arg

  Revision 1.40  2008/02/08 05:08:50  paul
  Updated WFSUnmountVolumeAsync

  Revision 1.39  2008/02/08 05:07:45  paul
  Changed BOOL to WFSBool

  Revision 1.38  2008/02/08 03:29:16  paul
  Fixed WFSGetCurrentDirectory to return '/' if current directory is '/' - normally trailing '/' is removed.

  Revision 1.37  2008/02/07 13:02:21  nakanose_jin
  WFSUnmountVolume : Add force-close option (but only in detach condition)

  Revision 1.36  2008/01/23 00:44:30  ueno
  Fixed to copy volume ID only when the result is WFS_RESULT_OK.

  Revision 1.35  2008/01/23 00:07:18  paul
  Fixed WFSGetVolumeId to only overwrite buffer when result==WFS_RESULT_OK

  Revision 1.34  2008/01/10 05:33:23  nakanose_jin
  validatePathName => validateDirectory

  Revision 1.33  2008/01/09 10:24:50  paul
  Added check for pWorkSpaceBuffer==NULL in WFSInit

  Revision 1.32  2008/01/08 12:17:37  ueno
  Fixed.

  Revision 1.31  2007/12/27 01:44:08  ueno
  Modified WFSExit() to call WFSSrvExit().

  Revision 1.30  2007/12/26 07:58:30  ueno
  Modified WFSInit() to start connecting WFSDEV server.

  Revision 1.29  2007/12/26 00:39:00  paul
  typo: DestroyHeap -> HeapDestroy

  Revision 1.28  2007/12/26 00:37:37  paul
  Added WFSHeapDestroy()

  Revision 1.27  2007/12/12 08:13:31  paul
  Added WFSSrvGetFileSize

  Revision 1.26  2007/12/12 01:08:10  ueno
  Modified WFSInit() to remove network initialization code.

  Revision 1.25  2007/12/10 02:48:49  paul
  Fixed WFSSetCurrentDirectoryAsync, WFSSetHomeDirectoryAsync

  Revision 1.24  2007/12/07 23:58:11  paul
  Changed lib initialization check. Now it always checks for WFSExit, WFSSetCurrentDirectory. For other functions, only check in debug version.

  Revision 1.23  2007/12/06 20:22:48  paul
  Fixed a bug in WFSSearchDirectoryNextAsync(). It was incorrectly allocating a search handle.

  Revision 1.22  2007/11/22 02:36:41  paul
  made WFSCloseFileAsync return WFS_RESULT_INVALID if nFileSize is > actual file size

  Revision 1.21  2007/11/22 01:59:58  paul
  Added nFileSize to WFSCloseFileAsync

  Revision 1.20  2007/11/21 04:17:45  wayne.wong
  Many changes that included single request/result RPC messages, bug fixes, code factoring, and English C comment adjustments.
    1. Combined RPC messages into a single request message and single result
        message. Except for file reads and writes.
    2. Fixed+factored some Endianess code.
    3. Enabled larger TCP/IP data transfer size. Previous setting was artifically small
        to test the code to perform large transfers.
    4. Changed the volume id parameters utf8*->WFSVolumeId for SrvGetVolumeId
        and SrvMountDevice. More consistent with the way we send this type at the
        RPC layer.
    5. Changed a couple Japanese comments to be compatible with English
        comments. In particular, the "//" comment with a backslash  at the end of
        the line causes the next line of code to be ignored.
    6. Removed Windows-specific debugging output defines for RVL build.

  Revision 1.19  2007/11/20 19:44:58  paul
  Added  WFSReturnIfLibNotInitialized to WFSFlush(), WFSFlushAsync()

  Revision 1.18  2007/11/16 21:55:33  paul
  Changed WFSSrvCloseFile to pass nNewSize instead of bTruncateFlag

  Revision 1.17  2007/11/16 02:04:32  paul
  Added WFSFlush()
  Removed previous change by Nakanose-san. (NULL pointer was already being checked by WFSSetDeviceName)

  Revision 1.16  2007/11/15 12:59:01  nakanose_jin
  TitleID -> TitleId
  add nullpointer check

  Revision 1.15  2007/11/09 01:17:01  paul
  Added sub-directory validation check for WFSMoveLocal
  Fixed validation of path name depth for paths ending in directories

  Revision 1.14  2007/11/06 22:42:37  paul
  Added validation checks for NULL pointers passed in

  Revision 1.13  2007/11/02 01:57:53  paul
  bugfix

  Revision 1.12  2007/11/02 00:54:31  paul
  now logging changes
  Changed WFSReadFileAsync, WFSWriteFileAsync to require explicit start position


*---------------------------------------------------------------------------*/

// File created by Paul Donnelly on 2007.08.29

#include "wfs_Client.h"
#include "wfs_Errors.h"
#include "wfssrv_Defs.h"
#include "wfs_AsyncUtils.h"
#include "private/wfs_Debug.h"

#define WFSReturnIfLibNotInitialized  if (!wcg.bLibInitialized) return WFS_RESULT_LIB_NOT_INITIALIZED;

WFSClientGlobals wcg ATTRIBUTE_ALIGN(4);

// ToDo: Check start and end of user-supplied buffers

WFSResult WFSInit(void *pWorkSpaceBuffer, u32 nWorkSpaceSize) {
    if (wcg.bLibInitialized) {
        return WFS_RESULT_LIB_ALREADY_INITIALIZED;
    }
    if ( (pWorkSpaceBuffer==NULL) || ((u32)pWorkSpaceBuffer&0x1ff)||(nWorkSpaceSize<WFS_MIN_WORK_SPACE_SIZE)||(nWorkSpaceSize>WFS_MAX_WORK_SPACE_SIZE) 
#if _RVL
         || (((u32)pWorkSpaceBuffer&0xf0000000)!=0x90000000)
#endif
        ) {
        return WFS_RESULT_INVALID;
    }
    wcg.pWorkSpaceBuffer = pWorkSpaceBuffer;
    WFSCreateHeap(&wcg.heap, pWorkSpaceBuffer, WFSCLI_WORK_SPACE_SIZE);
    pWorkSpaceBuffer = (void *)((size_t)pWorkSpaceBuffer + WFSCLI_WORK_SPACE_SIZE);
    nWorkSpaceSize   -= WFSCLI_WORK_SPACE_SIZE;
    WFSPathInit();
    WFSCliInitHandles();
    wcg.bLibInitialized = true;
    WFSAsyncUtilsInit();
    WFSReturnOnError(WFSSrvInit(pWorkSpaceBuffer, nWorkSpaceSize));
    return WFS_RESULT_OK;
}

static WFSResult CloseAllFile(void)
{
//    WFSFileHandle fh;
    WFSResult result = WFS_RESULT_OK;
    WFSCliFileInfo* pFi;
    u32 idx;
    for (idx=0; idx < WFSCLI_MAX_FILE_HANDLES; ++idx)
    {
        pFi = &wcg.hnd.aCliFileInfo[idx];
        if (pFi->nNext == WFSCLI_HANDLE_ALLOCATED)
        {
            // the handle is open.
            WFSFileHandle fh = pFi->nHandle;
            result = WFSCloseFile(fh, (WFSBool) FALSE); // [check] should wait file access?
            if (result != WFS_RESULT_OK &&
                result != WFS_RESULT_MEDIA_ERROR)
            {
                break;
            }
        }
    }

    if (result == WFS_RESULT_MEDIA_ERROR)
    {
        result = WFS_RESULT_OK;
    }
    return result;
}


WFSResult WFSExit() {
    if (!wcg.bLibInitialized) return WFS_RESULT_LIB_NOT_INITIALIZED;
    WFSReturnOnError(CloseAllFile());
    WFSReturnOnError(WFSSrvExit());
    wcg.bLibInitialized = false;
    WFSAsyncUtilsClose();
    WFSHeapDestroy(&wcg.heap);
    return WFS_RESULT_OK;
}

/*
WFSResult WFSExitAsync(void *pUserData, WFSResultCallback cb) {
    if (!wcg.bLibInitialized) return WFS_RESULT_LIB_NOT_INITIALIZED;
    wcg.bLibInitialized = false;
    return WFSPrepareAsyncCall(NULL, WFSSRV_EXIT, pUserData, cb, 0);
}
*/

void WFSSetAttachDeviceCallback(utf8 *sDeviceNameBuffer, WFSDeviceCallback cbAttachDevice, void *pUserData) {
    if (!wcg.dcg.bMutexInitialized) {
        wcg.dcg.bMutexInitialized = true;
        WFSInitMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    }
    WFSLockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    wcg.dcg.pAttachDeviceNameBuffer = sDeviceNameBuffer;
    wcg.dcg.cbAttachDevice          = cbAttachDevice;
    wcg.dcg.pAttachDeviceUserData   = pUserData;
    WFSUnlockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
}

void WFSGetAttachDeviceCallback(utf8 **psDeviceName, WFSDeviceCallback *pCbAttachDevice, void **ppUserData) {
    if (!wcg.dcg.bMutexInitialized) {
        wcg.dcg.bMutexInitialized = true;
        WFSInitMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    }
    WFSLockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    if (psDeviceName)    { *psDeviceName    = wcg.dcg.pAttachDeviceNameBuffer; }
    if (pCbAttachDevice) { *pCbAttachDevice = wcg.dcg.cbAttachDevice; }
    if (ppUserData)      { *ppUserData      = wcg.dcg.pAttachDeviceUserData; }
    WFSUnlockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
}

void WFSSetDetachDeviceCallback(utf8 *sDeviceNameBuffer, WFSDeviceCallback cbDetachDevice, void *pUserData) {
    if (!wcg.dcg.bMutexInitialized) {
        wcg.dcg.bMutexInitialized = true;
        WFSInitMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    }
    WFSLockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    wcg.dcg.pDetachDeviceNameBuffer = sDeviceNameBuffer;
    wcg.dcg.cbDetachDevice          = cbDetachDevice;
    wcg.dcg.pDetachDeviceUserData   = pUserData;
    WFSUnlockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
}

void WFSGetDetachDeviceCallback(utf8 **psDeviceName, WFSDeviceCallback *pCbDetachDevice, void **ppUserData) {
    if (!wcg.dcg.bMutexInitialized) {
        wcg.dcg.bMutexInitialized = true;
        WFSInitMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    }
    WFSLockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
    if (psDeviceName)    { *psDeviceName    = wcg.dcg.pDetachDeviceNameBuffer; }
    if (pCbDetachDevice) { *pCbDetachDevice = wcg.dcg.cbDetachDevice; }
    if (ppUserData)      { *ppUserData      = wcg.dcg.pDetachDeviceUserData; }
    WFSUnlockMutex(&wcg.dcg.mxAccessDeviceCallbackGlobals);
}


WFSResult WFSGetDeviceInfo(const utf8 *sDeviceName, WFSDeviceInfo *pDi) {
    WFSReturnIfLibNotInitialized;
    WFSDeviceName devName;
    WFSReturnOnNullParameter(pDi);
    WFSReturnOnError(WFSSetDeviceName(&devName, sDeviceName));
    return WFSSrvGetDeviceInfo(&devName, pDi);
}

WFSResult WFSGetDeviceInfoAsync(const utf8 *sDeviceName, WFSDeviceInfo *pDi, void *pUserData, WFSDeviceInfoCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(WFSDeviceInfo *);
    WFSReturnOnNullParameter(pDi);
    WFSReturnOnError(WFSPrepareAsyncCallWithDeviceName(&pPtr, sDeviceName, WFSSRV_GET_DEVICE_INFO, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSDeviceInfo *, pDi);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSMountDevice(const utf8 *sDeviceName, utf8 *sVolumeIdBuffer) {
    WFSReturnIfLibNotInitialized;
    WFSDeviceName devName;
    WFSVolumeId   volId;
    WFSResult     result;

    WFSReturnOnNullParameter(sVolumeIdBuffer);
    WFSReturnOnError(WFSSetDeviceName(&devName, sDeviceName));
    result = WFSSrvMountDevice(&devName, &volId);
    if (result==WFS_RESULT_OK) {
        strncpy(sVolumeIdBuffer, volId.sStr, WFS_MAX_VOLUME_ID_SIZE+1);
    }
    return result;
}

WFSResult WFSMountDeviceAsync(const utf8 *sDeviceName, utf8 *sVolumeIdBuffer, void *pUserData, WFSDeviceVolumeCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(const utf8 *);
    WFSReturnOnNullParameter(sVolumeIdBuffer);
    WFSReturnOnError(WFSPrepareAsyncCallWithDeviceName(&pPtr, sDeviceName, WFSSRV_MOUNT_DEVICE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, utf8 *, sVolumeIdBuffer);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSUnmountVolume(const utf8 *sVolumeId, WFSBool bForce) {
    WFSReturnIfLibNotInitialized;
    WFSVolumeId volumeId;
    WFSReturnOnError(WFSSetVolumeId(&volumeId, sVolumeId));
    return WFSSrvUnmountVolume(&volumeId, bForce);
}

WFSResult WFSUnmountVolumeAsync(const utf8 *sVolumeId, WFSBool bForce, void *pUserData, WFSUnmountVolumeCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(bForce);
    WFSReturnOnError(WFSPrepareAsyncCallWithVolumeId(&pPtr, sVolumeId, WFSSRV_UNMOUNT_VOLUME, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSBool, bForce);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSInitializeDevice(const utf8 *sDeviceName) {
    WFSReturnIfLibNotInitialized;
    WFSDeviceName devName;
    WFSReturnOnError(WFSSetDeviceName(&devName, sDeviceName));
    return WFSSrvInitializeDevice(&devName);
}

WFSResult WFSInitializeDeviceAsync(const utf8 *sDeviceName, void *pUserData, WFSInitializeDeviceCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareAsyncCallWithDeviceName(NULL, sDeviceName, WFSSRV_INITIALIZE_DEVICE, pUserData, cb, 0);
}

WFSResult WFSGetVolumeId(const utf8 *sDeviceName, utf8 *sVolumeIdBuffer) {
    WFSReturnIfLibNotInitialized;
    WFSDeviceName devName;
    WFSVolumeId   volId;
    WFSResult     result;
    WFSReturnOnNullParameter(sVolumeIdBuffer);
    WFSReturnOnError(WFSSetDeviceName(&devName, sDeviceName));
    result = WFSSrvGetVolumeId(&devName, &volId);
    if (result==WFS_RESULT_OK) {
        strncpy(sVolumeIdBuffer, volId.sStr, WFS_MAX_VOLUME_ID_SIZE+1);
    }
    return result;
}

WFSResult WFSGetVolumeIdAsync(const utf8 *sDeviceName, utf8 *sVolumeIdBuffer, void *pUserData, WFSDeviceVolumeCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(utf8 *);
    WFSReturnOnNullParameter(sVolumeIdBuffer);
    WFSReturnOnError(WFSPrepareAsyncCallWithDeviceName(&pPtr, sDeviceName, WFSSRV_GET_VOLUME_ID, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, utf8 *, sVolumeIdBuffer);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSGetFreeSpaceSize(const utf8 *sPathName, u64 *pSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pSize);
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvGetFreeSpaceSize(&pathNameAbs, pSize);
}

WFSResult WFSGetFreeSpaceSizeAsync(const utf8 *sPathName, void *pUserData, WFSGetFreeSpaceCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareAsyncCallWithPathName(NULL, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_GET_FREE_SPACE_SIZE, pUserData, cb, 0);
}

WFSResult WFSSetHomeDirectory(const utf8 *sAbsPathName) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSPathPartOffsetArray ppoa;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, &ppoa, sAbsPathName, PATH_TYPE_ABS_DIR_ONLY));
    WFSReturnOnError(WFSSrvValidateDirectory(&pathNameAbs));
    WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
    WFSPathCopy(&wcg.pg.pathHome.pn, &wcg.pg.pathHome.ppoa, &pathNameAbs, &ppoa, true, true);
    WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
    return WFS_RESULT_OK;
}

WFSResult WFSSetHomeDirectoryAsync(const utf8 *sAbsPathName, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareAsyncCallWithPathName(NULL, sAbsPathName, PATH_TYPE_ABS_DIR_ONLY, WFSSRV_SET_HOME_DIRECTORY, pUserData, cb, 0);
}

WFSResult WFSSetCurrentDirectory(const utf8 *sPathName) {
    if (!wcg.bLibInitialized) return WFS_RESULT_LIB_NOT_INITIALIZED;
    WFSPathName pathNameAbs;
    WFSPathPartOffsetArray ppoa;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, &ppoa, sPathName, PATH_TYPE_REL_OR_ABS));
    WFSReturnOnError(WFSSrvValidateDirectory(&pathNameAbs));
    WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
    WFSPathCopy(&wcg.pg.pathCurrent.pn, &wcg.pg.pathCurrent.ppoa, &pathNameAbs, &ppoa, true, true);
    WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
    return WFS_RESULT_OK;
}

WFSResult WFSSetCurrentDirectoryAsync(const utf8 *sPathName, void *pUserData, WFSResultCallback cb) {
    if (!wcg.bLibInitialized) return WFS_RESULT_LIB_NOT_INITIALIZED;
    return WFSPrepareAsyncCallWithPathName(NULL, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_SET_CURRENT_DIRECTORY, pUserData, cb, 0);
}

s32 WFSGetCurrentDirectory(utf8 *sAbsPathNameBuffer, s32 nBufferSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(sAbsPathNameBuffer);
    WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
    if (wcg.pg.pathCurrent.pn.nLen > nBufferSize) {
        s32 nLength = wcg.pg.pathCurrent.pn.nLen;   // This is enough since we remove the trailing '/'
        WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
        return nLength;
    }
    memcpy(sAbsPathNameBuffer, wcg.pg.pathCurrent.pn.sStr, wcg.pg.pathCurrent.pn.nLen);
    int nPos = wcg.pg.pathCurrent.pn.nLen;
    if (wcg.pg.pathCurrent.pn.nLen>1 && sAbsPathNameBuffer[wcg.pg.pathCurrent.pn.nLen-1] == '/') {
        --nPos;  // Removes the trailing '/' unless it is also the first '/'
    }
    sAbsPathNameBuffer[nPos]=0; // Ensure the return string is properly terminated
    WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
    return WFS_RESULT_OK;
}

WFSResult WFSCreateDirectory(const utf8 *sPathName, u32 nFlags, u64 nQuota) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS_DIR));
    return WFSSrvCreateDirectory(&pathNameAbs, nFlags, nQuota);
}

WFSResult WFSCreateDirectoryAsync(const utf8 *sPathName, u32 nFlags, u64 nQuota, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(nFlags) + sizeof(nQuota);
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS_DIR, WFSSRV_CREATE_DIRECTORY, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nFlags);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, u64, nQuota);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSGetAttributes(const utf8 *sPathName, WFSFileAttributes *pFa) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnNullParameter(pFa);
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvGetAttributes(&pathNameAbs, pFa);
}

WFSResult WFSGetAttributesAsync(const utf8 *sPathName, WFSFileAttributes *pFa, void *pUserData, WFSGetAttributesCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFa);
    void *pPtr;
    size_t nExtraSize = sizeof(WFSFileAttributes *);
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_GET_ATTRIBUTES, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileAttributes *, pFa);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSChangePermissions(const utf8 *sPathName, u32 nFlags, u32 nMask) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvChangePermissions(&pathNameAbs, nFlags, nMask);
}

WFSResult WFSChangePermissionsAsync(const utf8 *sPathName, u32 nFlags, u32 nMask, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(nFlags) + sizeof(nMask);
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_CHANGE_PERMISSIONS, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nFlags);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, u32, nMask);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSDelete(const utf8 *sPathName) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvDelete(&pathNameAbs);
}

WFSResult WFSDeleteAsync(const utf8 *sPathName, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareAsyncCallWithPathName(NULL, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_DELETE, pUserData, cb, 0);
}

WFSResult WFSDeleteAccessListEntry(const utf8 *sPathName, WFSTitleId titleId, WFSTitleId entryCreatorId) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvDeleteAccessListEntry(&pathNameAbs,titleId,entryCreatorId);
}

WFSResult WFSDeleteAccessListEntryAsync(const utf8 *sPathName, WFSTitleId titleId, WFSTitleId entryCreatorId, void *pUserData, WFSResultCallback cb)
{
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(WFSTitleId) * 2; // entryCreatorId and titleId.
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS,
        WFSSRV_DELETE_ACL_ENTRY, pUserData, cb, nExtraSize));

    // save arguments.
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSTitleId, titleId);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSTitleId, entryCreatorId);

    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSMoveLocal(const utf8 *sSrcPathName, const utf8 *sDstPathName) {
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbsSrc, pathNameAbsDst;
    WFSReturnOnError(WFSPathResolve(&pathNameAbsSrc, NULL, sSrcPathName, PATH_TYPE_REL_OR_ABS));
    WFSReturnOnError(WFSPathResolve(&pathNameAbsDst, NULL, sDstPathName, PATH_TYPE_REL_OR_ABS));
    // If moved directory is included to the current directory, invalid error is returned.
    if (strncmp(pathNameAbsSrc.sStr, wcg.pg.pathCurrent.pn.sStr, pathNameAbsSrc.nLen)==0) {
        return WFS_RESULT_INVALID;
    }
    // If moved directory is included to the home directory, invalid error is returned.
    if (strncmp(pathNameAbsSrc.sStr, wcg.pg.pathCurrent.pn.sStr, pathNameAbsSrc.nLen)==0) {
        return WFS_RESULT_INVALID;
    }    
    if (strncmp(pathNameAbsSrc.sStr, pathNameAbsDst.sStr, pathNameAbsSrc.nLen)==0) {
        if (pathNameAbsDst.sStr[pathNameAbsSrc.nLen]=='/') {
            // trying to move a directory to it's own sub-directory
            return WFS_RESULT_INVALID;
        }
    }
    return WFSSrvMoveLocal(&pathNameAbsSrc, &pathNameAbsDst);
}

WFSResult WFSMoveLocalAsync(const utf8 *sSrcPathName, const utf8 *sDstPathName, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareMoveLocalAsyncCall(NULL, sSrcPathName, sDstPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_MOVE_LOCAL, pUserData, cb, 0);
}

// ToDo: Operations which use file handles should lock the file handles

WFSResult WFSCreateAndOpenFile(const utf8 *sPathName, u32 nFlags, WFSContentIdx nCidx, WFSFileSize nPreAllocateSize, WFSFileHandle *pFh) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFh);
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSCliFileInfoAlloc(&pFi));
    WFSSrvFileHandle sfh;
    WFSResult nResult = WFSSrvCreateAndOpenFile(&pathNameAbs, nFlags, nCidx, nPreAllocateSize, &sfh);
    if (nResult != WFS_RESULT_OK) {
        WFSCliFileInfoFree(pFi);
        return nResult;
    }
    *pFh = pFi->nHandle |= sfh;
    pFi->fa.nFileSize = 0;
    pFi->fa.nFlags    = nFlags;
    pFi->fa.nCidx = nCidx;
    pFi->nAccess = WFS_ACCESS_RW;
    pFi->nPosition = 0;
    pFi->nPreAllocateSize = nPreAllocateSize;
    return WFS_RESULT_OK;
}

WFSResult WFSCreateAndOpenFileAsync(const utf8 *sPathName, u32 nFlags, WFSContentIdx nCidx, u32 nPreAllocateSize, void *pUserData, WFSFileOpCallback cb) {
    WFSResult nResult;
    WFSReturnIfLibNotInitialized;
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSCliFileInfoAlloc(&pFi));
    void *pPtr;
    size_t nExtraSize = sizeof(nFlags) + sizeof(nCidx) + sizeof(nPreAllocateSize) + sizeof(WFSCliFileInfo *);
    nResult = WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_CREATE_AND_OPEN_FILE, pUserData, cb, nExtraSize);
    if (nResult == WFS_RESULT_OK)
    {
        WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nFlags);
        WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSContentIdx, nCidx);
        WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nPreAllocateSize);
        WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSCliFileInfo *, pFi);
        WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    }
    else
    {
        WFSCliFileInfoFree(pFi);
    }
    return nResult;
}

WFSResult WFSOpenFile(const utf8 *sPathName, WFSAccess nDesiredAccess, WFSFileHandle *pFh) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFh);
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSCliFileInfoAlloc(&pFi));
    WFSSrvFileHandle sfh;
    WFSResult nResult = WFSSrvOpenFile(&pathNameAbs, nDesiredAccess, &sfh, &pFi->fa);
    if (nResult != WFS_RESULT_OK) {
        WFSCliFileInfoFree(pFi);
        *pFh = WFS_INVALID_HANDLE_VALUE;
        return nResult;
    }
    *pFh = pFi->nHandle |= sfh;
    pFi->nAccess = nDesiredAccess;
    pFi->nPosition = 0;
    return WFS_RESULT_OK;
}

WFSResult WFSOpenFileAsync(const utf8 *sPathName, WFSAccess nDesiredAccess, void *pUserData, WFSFileOpCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSResult nResult;
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSCliFileInfoAlloc(&pFi));
    void *pPtr;
    size_t nExtraSize = sizeof(nDesiredAccess) + sizeof(WFSCliFileInfo *);
    nResult = WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_OPEN_FILE, pUserData, cb, nExtraSize);
    if (nResult == WFS_RESULT_OK)
    {
        WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSAccess, nDesiredAccess);
        WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSCliFileInfo *, pFi);
        WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    }
    else
    {
        WFSCliFileInfoFree(pFi);
    }
    return nResult;
}

WFSResult WFSSetFilePosition(WFSFileHandle fh, WFSFileSize nFilePosition) {
    WFSReturnIfLibNotInitialized;
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if (nFilePosition>pFi->fa.nFileSize) {
        return WFS_RESULT_INVALID;
    }
    pFi->nPosition = nFilePosition;
    return WFS_RESULT_OK;
}

WFSResult WFSGetFilePosition(WFSFileHandle fh, WFSFileSize *pFilePosition) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFilePosition);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    *pFilePosition = pFi->nPosition;
    return WFS_RESULT_OK;
}

WFSResult WFSGetFileSize(WFSFileHandle fh, WFSFileSize *pFileSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileSize);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
    WFSReturnOnError(WFSSrvGetFileSize(sfh, pFileSize));
    return WFS_RESULT_OK;
}

WFSResult WFSGetFileSizeAsync(WFSFileHandle fh, void *pUserData, WFSGetFileSizeCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    void *pPtr;
    size_t nExtraSize = sizeof(fh);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_GET_FILE_SIZE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileHandle, fh);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

s32 WFSGetAccessList(const utf8 *sPathName, WFSAccessListEntry *pAccessList, u32 nMaxElements)
{
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pAccessList);
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));

    s32 nResultOrNumEntries = WFSSrvGetAccessList( &pathNameAbs , pAccessList , nMaxElements );
    if( nResultOrNumEntries >= 0)
    {
        // I
    }
    return nResultOrNumEntries;
}


s32 WFSReadFile(WFSFileHandle fh, void *pFileDataBuffer, s32 nSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileDataBuffer);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_READ) == 0) {
        return WFS_RESULT_ACCESS;
    }
    WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
    s32 nResultOrNumBytesRead = WFSSrvReadFile(sfh, pFileDataBuffer, pFi->nPosition, nSize);
    if (nResultOrNumBytesRead >= 0) {
        pFi->nPosition += nResultOrNumBytesRead;
    }
    return nResultOrNumBytesRead;
}

WFSResult WFSReadFileAsync(WFSFileHandle fh, void *pFileDataBuffer, WFSFileSize nStartPosition, s32 nSize, void *pUserData, WFSReadFileCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileDataBuffer);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_READ) == 0) {
        return WFS_RESULT_ACCESS;
    }
    void *pPtr;
    size_t nExtraSize = sizeof(fh) + sizeof(pFileDataBuffer) + sizeof(nStartPosition) + sizeof(nSize);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_READ_FILE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileHandle, fh);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, void *, pFileDataBuffer);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileSize, nStartPosition);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, s32, nSize);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSWriteFile(WFSFileHandle fh, const void *pFileData, u32 nSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileData);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_WRITE) == 0) {
        return WFS_RESULT_ACCESS;
    }
    WFSFileSize nNewPosition = pFi->nPosition + nSize;
    if (nNewPosition < pFi->nPosition) { // Detects u32 overflow
        return WFS_RESULT_FILE_TOO_BIG;
    }
    WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
    WFSResult nResult = WFSSrvWriteFile(sfh, pFileData, pFi->nPosition, nSize, FALSE);
    if (nResult == WFS_RESULT_OK) {
        pFi->nPosition = nNewPosition;
        if (pFi->fa.nFileSize < nNewPosition) {
            pFi->fa.nFileSize = nNewPosition;
        }
    }
    return nResult;
}

WFSResult WFSWriteFileAsync(WFSFileHandle fh, const void *pFileData, WFSFileSize nStartPosition, WFSFileSize nSize, void *pUserData, WFSFileOpCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileData);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_WRITE) == 0) {
        return WFS_RESULT_ACCESS;
    }
    WFSFileSize nNewPosition = nStartPosition + nSize;
    if (nNewPosition < nStartPosition) { // Detects u32 overflow
        return WFS_RESULT_FILE_TOO_BIG;
    }
    void *pPtr;
    size_t nExtraSize = sizeof(fh) + sizeof(pFileData) + sizeof(nStartPosition) + sizeof(nSize);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_WRITE_FILE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileHandle, fh);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, const void *, pFileData);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileSize, nStartPosition);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileSize, nSize);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSWriteFileAndDiscard(WFSFileHandle fh, const void *pFileData, u32 nSize) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileData);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_WRITE) == 0) {
        return WFS_RESULT_ACCESS;
    }
    WFSFileSize nNewPosition = pFi->nPosition + nSize;
    if (nNewPosition < pFi->nPosition) { // Detects u32 overflow
        return WFS_RESULT_FILE_TOO_BIG;
    }
    WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
    WFSResult nResult = WFSSrvWriteFile(sfh, pFileData, pFi->nPosition, nSize, TRUE);
    if (nResult == WFS_RESULT_OK) {
        pFi->nPosition = nNewPosition;
        if (pFi->fa.nFileSize < nNewPosition) {
            pFi->fa.nFileSize = nNewPosition;
        }
    }
    return nResult;
}

WFSResult WFSWriteFileAndDiscardAsync(WFSFileHandle fh, const void *pFileData, WFSFileSize nStartPosition, WFSFileSize nSize, void *pUserData, WFSFileOpCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFileData);
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    if ((pFi->nAccess & WFS_ACCESS_WRITE) == 0) {
        return WFS_RESULT_ACCESS;
    }
    WFSFileSize nNewPosition = nStartPosition + nSize;
    if (nNewPosition < nStartPosition) { // Detects u32 overflow
        return WFS_RESULT_FILE_TOO_BIG;
    }
    void *pPtr;
    size_t nExtraSize = sizeof(fh) + sizeof(pFileData) + sizeof(nStartPosition) + sizeof(nSize);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_WRITE_FILE_AND_DISCARD, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileHandle, fh);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, const void *, pFileData);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileSize, nStartPosition);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileSize, nSize);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSCloseFile(WFSFileHandle fh, WFSBool bTruncateFlag) {
    WFSReturnIfLibNotInitialized;
    if (fh==WFS_INVALID_HANDLE_VALUE) {
        return WFS_RESULT_INVALID;
    }
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
    u32 nNewSize = pFi->fa.nFileSize;
    if (bTruncateFlag) {
        if (pFi->nAccess & WFS_ACCESS_WRITE) {
            nNewSize = pFi->nPosition;
        } else {
            return WFS_RESULT_ACCESS;
        }
    }
    WFSResult nResult = WFSSrvCloseFile(sfh, nNewSize);
    if (nResult == WFS_RESULT_OK) {
        WFSCliFileInfoFree(pFi);
    }
    return nResult;
}

WFSResult WFSFlush(const utf8 *sVolumeId) {
    WFSReturnIfLibNotInitialized;
    WFSVolumeId volumeId;
    WFSReturnOnError(WFSSetVolumeId(&volumeId, sVolumeId));
    return WFSSrvFlush(&volumeId);
}

WFSResult WFSFlushAsync(const utf8 *sVolumeId, void *pUserData, WFSResultCallback cb) {
    WFSReturnIfLibNotInitialized;
    return WFSPrepareAsyncCallWithVolumeId(NULL, sVolumeId, WFSSRV_FLUSH, pUserData, cb, 0);
}


WFSResult WFSCloseFileAsync(WFSFileHandle fh, WFSBool bTruncateFlag, WFSFileSize nFileSize, void *pUserData, WFSFileOpCallback cb) {
    WFSReturnIfLibNotInitialized;
    if (fh==WFS_INVALID_HANDLE_VALUE) {
        return WFS_RESULT_INVALID;
    }
    WFSCliFileInfo *pFi;
    WFSReturnOnError(WFSGetCliFileInfo(&pFi, fh));
    u32 nNewSize = pFi->fa.nFileSize;
    if (bTruncateFlag) {
        if (pFi->nAccess & WFS_ACCESS_WRITE) {
            nNewSize = nFileSize;
            if (nNewSize > pFi->fa.nFileSize) {
                return WFS_RESULT_INVALID;
            }
        } else {
            return WFS_RESULT_ACCESS;
        }
    }
    void *pPtr;
    size_t nExtraSize = sizeof(fh) + sizeof(nNewSize);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_CLOSE_FILE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSFileHandle, fh);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileSize, nNewSize);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSSearchDirectoryFirst(const utf8 *sPattern, WFSSearchDirectoryHandle *pSdh, WFSFileInfo *pFi) {
    WFSReturnIfLibNotInitialized;
    if ((pSdh==NULL)||(pFi==NULL)) {
        return WFS_RESULT_INVALID;
    }
    WFSPathName pathNameAbsPattern;
    WFSReturnOnError(WFSPathResolve(&pathNameAbsPattern, NULL, sPattern, PATH_TYPE_PATTERN));
    WFSCliSearchDirInfo *pSdi;
    WFSReturnOnError(WFSCliSearchDirInfoAlloc(&pSdi));
    WFSSrvSearchDirectoryHandle ssdh;
    WFSResult nResult = WFSSrvSearchDirectoryFirst(&pathNameAbsPattern, &ssdh, pFi);
    if (nResult != WFS_RESULT_OK) {
        *pSdh = WFS_INVALID_HANDLE_VALUE;
        WFSCliSearchDirInfoFree(pSdi);
        return nResult;
    }
    *pSdh = pSdi->nHandle |= ssdh;
    return WFS_RESULT_OK;
}

WFSResult WFSSearchDirectoryFirstAsync(const utf8 *sPattern, WFSFileInfo *pFi, void *pUserData, WFSSearchDirectoryCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFi);
    WFSCliSearchDirInfo *pSdi;
    WFSReturnOnError(WFSCliSearchDirInfoAlloc(&pSdi));
    void *pPtr;
    size_t nExtraSize = sizeof(WFSCliSearchDirInfo *) + sizeof(pFi);
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPattern, PATH_TYPE_PATTERN, WFSSRV_SEARCH_DIRECTORY_FIRST, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSCliSearchDirInfo *, pSdi);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileInfo *, pFi);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSSearchDirectoryNext(WFSSearchDirectoryHandle sdh, WFSFileInfo *pFi) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFi);
    WFSCliSearchDirInfo *pSdi;
    WFSReturnOnError(WFSGetCliSearchDirInfo(&pSdi, sdh));
    WFSSrvSearchDirectoryHandle ssdh = (WFSSrvSearchDirectoryHandle)sdh;
    return WFSSrvSearchDirectoryNext(ssdh, pFi);
}

WFSResult WFSSearchDirectoryNextAsync(WFSSearchDirectoryHandle sdh, WFSFileInfo *pFi, void *pUserData, WFSSearchDirectoryCallback cb) {
    WFSReturnIfLibNotInitialized;
    WFSReturnOnNullParameter(pFi);
    void *pPtr;
    size_t nExtraSize = sizeof(sdh) + sizeof(WFSFileInfo *);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_SEARCH_DIRECTORY_NEXT, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSSearchDirectoryHandle, sdh);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSFileInfo *, pFi);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSSearchDirectoryClose(WFSSearchDirectoryHandle sdh) {
    WFSReturnIfLibNotInitialized;
    if (sdh==WFS_INVALID_HANDLE_VALUE) {
        return WFS_RESULT_INVALID;
    }
    WFSCliSearchDirInfo *pSdi;
    WFSReturnOnError(WFSGetCliSearchDirInfo(&pSdi, sdh));
    WFSSrvSearchDirectoryHandle ssdh = (WFSSrvSearchDirectoryHandle)sdh;
    WFSResult nResult = WFSSrvSearchDirectoryClose(ssdh);
    if (nResult == WFS_RESULT_OK) {
        WFSCliSearchDirInfoFree(pSdi);
    }
    return nResult;
}

WFSResult WFSSearchDirectoryCloseAsync(WFSSearchDirectoryHandle sdh, void *pUserData, WFSSearchDirectoryCloseCallback cb) {
    WFSReturnIfLibNotInitialized;
    if (sdh==WFS_INVALID_HANDLE_VALUE) {
        return WFS_RESULT_INVALID;
    }
    WFSCliSearchDirInfo *pSdi;
    WFSReturnOnError(WFSGetCliSearchDirInfo(&pSdi, sdh));
    void *pPtr;
    size_t nExtraSize = sizeof(sdh);
    WFSReturnOnError(WFSPrepareAsyncCall(&pPtr, WFSSRV_SEARCH_DIRECTORY_CLOSE, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSSearchDirectoryHandle, sdh);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}


WFSResult WFSGetPermissionsForUser(const utf8 *sPathName, u32 *pFlags, WFSTitleId titleId)
{
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvGetPermissionForUser(&pathNameAbs, pFlags, titleId);
}

WFSResult WFSGetPermissionsForUserAsync(const utf8 *sPathName, WFSTitleId titleId, void *pUserData, WFSGetPermissionsForUserCallback cb)
{
    WFSReturnIfLibNotInitialized;
    void *pPtr; // for other params.
    size_t nExtraSize = sizeof(WFSTitleId);

    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_GET_PERMISSIONS, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSTitleId, titleId);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSSetPermissionsForUser(const utf8 *sPathName, u32 nFlags, u32 nMask, WFSTitleId titleId)
{
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvSetPermissionForUser(&pathNameAbs, nFlags, nMask, titleId);
}

WFSResult WFSSetAccessListEntry(const utf8 *sPathName, u32 nFlags, u32 nMask, WFSTitleId titleId)
{
    WFSReturnIfLibNotInitialized;
    WFSPathName pathNameAbs;
    WFSReturnOnError(WFSPathResolve(&pathNameAbs, NULL, sPathName, PATH_TYPE_REL_OR_ABS));
    return WFSSrvSetAccessListEntry(&pathNameAbs, nFlags, nMask, titleId);
}

WFSResult WFSSetAccessListEntryAsync(const utf8 *sPathName, u32 nFlags, u32 nMask, WFSTitleId titleId, void *pUserData, WFSResultCallback cb)
{
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    size_t nExtraSize = sizeof(u32) * 2 + sizeof(WFSTitleId); // nFlags, nMask and titleId.
    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS,
        WFSSRV_SET_ACL_ENTRY, pUserData, cb, nExtraSize));

    // save arguments.
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nFlags);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nMask);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSTitleId, titleId);

    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSGetAccessListAsync(const utf8 *sPathName, WFSAccessListEntry *pAccessList, u32 nMaxElements, void *pUserData, WFSGetAccessListCallback cb)
{
    WFSReturnIfLibNotInitialized;
    void *pPtr;
    WFSReturnOnNullParameter(pAccessList);

    size_t nExtraSize = sizeof(WFSAccessListEntry*) + sizeof(u32); // pAccessList and nMaxElements.

    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_GET_ACL, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, WFSAccessListEntry*, pAccessList);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, u32, nMaxElements);

    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSSetPermissionsForUserAsync(const utf8 *sPathName, u32 nFlags, u32 nMask,
    WFSTitleId titleId, void *pUserData, WFSResultCallback cb)
{
    WFSReturnIfLibNotInitialized;
    void *pPtr; // for other params.
    size_t nExtraSize = sizeof(u32) + sizeof(u32) + sizeof(WFSTitleId);

    WFSReturnOnError(WFSPrepareAsyncCallWithPathName(&pPtr, sPathName, PATH_TYPE_REL_OR_ABS, WFSSRV_SET_PERMISSIONS, pUserData, cb, nExtraSize));
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nFlags);
    WFSCLI_WRITE_ASYNC_PARAM_AND_INC_PTR(pPtr, u32, nMask);
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, WFSTitleId, titleId);
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue); // Note: This is needed only if (nExtraSize != 0)
    return WFS_RESULT_OK;
}

WFSResult WFSDebugSetTitleId(WFSTitleId titleId);

WFSResult WFSDebugSetTitleId(WFSTitleId titleId) {
    WFSReturnIfLibNotInitialized;
    return WFSSrvSetTitleId(titleId,0);
}

WFSResult WFSExecDebugProcedure( u32 cmd, void *arg0, void *arg1 )
{
    /*if(cmd != WFSDEBUG_CMD_SHIM_WORKSPACE){
        WFSReturnIfLibNotInitialized;
    }*/
    return WFSSrvExecDebugProcedure( cmd , arg0 , arg1 );
}
