/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Heap.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Heap.cpp,v $
  Revision 1.8  2007/12/26 04:26:18  paul
  changed pAllocator to &allocator

  Revision 1.7  2007/12/26 04:14:14  paul
  Changed RVL version of WFSHeap

  Revision 1.6  2007/12/26 00:19:07  paul
  Added WFSHeapDestroy()

  Revision 1.5  2007/12/25 14:03:49  ueno
  Modified to initialize mutex, thread and heap only once.

  Revision 1.4  2007/11/02 00:55:27  paul
  now logging changes

*---------------------------------------------------------------------------*/

#include "wfs_Heap.h"

#undef dbg
#if _DEBUG_HEAP
#define dbg(s) s
#else
#define dbg(s)
#endif

#if _WIN32

dbg(u32 nSizeAvaliable=0;)

void WFSCreateHeap(WFSHeap *pHeap, void *pBase, size_t nSize) {
    *pHeap = HeapCreate(0, nSize, nSize);
    dbg(nSizeAvaliable = nSize);
}

void *WFSHeapAlloc(WFSHeap *pHeap, size_t nSize) {
    void *pAddr = HeapAlloc(*pHeap, 0, nSize);
    dbg(nSizeAvaliable -= nSize;
        MyOSReport("A:Heap:%d\n", nSizeAvaliable));
    return pAddr;
}

void WFSHeapFree(WFSHeap *pHeap, void *pAddr) {
    dbg(nSizeAvaliable += HeapSize(pHeap, NULL, pAddr);
        MyOSReport("F:Heap:%d\n", nSizeAvaliable));
    HeapFree(*pHeap, 0, pAddr);
}

void WFSHeapDestroy(WFSHeap *pHeap) {
    dbg(MyOSReport("WFSHeapDestroy()\n"));
    HeapDestroy(*pHeap);
}

//-- twl modified
#elif _TWL

void WFSCreateHeap(WFSHeap *pHeap, void *pBase, size_t nSize) {
    osTPrintf( "%s needs modify.\n", __FUNCTION__);
    //main.cōsĂ
}

void *WFSHeapAlloc(WFSHeap *pHeap, size_t nSize) {
//    void *pAddr = OS_AllocFromHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap, nSize);
    void *pAddr = osAlloc( nSize);
    if( pAddr == NULL) {
        osTPrintf( "%s error!\n", __FUNCTION__);
    }
    return pAddr;
}

void WFSHeapFree(WFSHeap *pHeap, void *pAddr) {
    osFree( pAddr);
  //    OS_FreeToHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap, pAddr);
}

void WFSHeapDestroy(WFSHeap *pHeap) {
    osTPrintf( "%s needs modify.\n", __FUNCTION__);
//    OS_DestroyHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap);
}
//-- twl modified

#elif _RVL

void WFSCreateHeap(WFSHeap *pHeap, void *pBase, size_t nSize) {
    pHeap->handle = MEMCreateExpHeap(pBase, nSize);
    MEMInitAllocatorForExpHeap(&pHeap->allocator, pHeap->handle, WFS_HEAP_DEFAULT_ALIGNMENT);
}

void *WFSHeapAlloc(WFSHeap *pHeap, size_t nSize) {
//    printf("==> %s:%d pHeap: 0x%08x size: %d\n", __FILE__, __LINE__, pHeap->pAllocator, nSize);
    void *pAddr = MEMAllocFromAllocator(&pHeap->allocator, nSize);
//    printf("==> %s:%d pAddr: 0x%08x\n", __FILE__, __LINE__, pAddr);
    return pAddr;
}

void WFSHeapFree(WFSHeap *pHeap, void *pAddr) {
    MEMFreeToAllocator(&pHeap->allocator, pAddr);
}

void WFSHeapDestroy(WFSHeap *pHeap) {
    MEMDestroyExpHeap(pHeap->handle);
}


#endif
