/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Names.cpp - Functions for setting and validating name types


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Names.cpp,v $
  Revision 1.5  2007/11/19 13:28:27  nakanose_jin
  bug fix for null pointer check

  Revision 1.4  2007/11/06 22:42:56  paul
  Added validation checks for NULL pointers passed in

  Revision 1.3  2007/11/02 01:01:00  paul
  now logging changes

*---------------------------------------------------------------------------*/

#include "wfs_Names.h"
#include <string.h>

WFSResult WFSSetPathName(WFSPathName *pPathName, const utf8 *sPathName) {
    WFSReturnOnNullParameter(sPathName);
    u32 nStrLen = strlen(sPathName);
    if (nStrLen > WFS_MAX_PATH_NAME_SIZE) {
        return WFS_RESULT_INVALID;
    }
    pPathName->nLen = (WFSPathNameLengthType)nStrLen;
    strcpy(pPathName->sStr, sPathName);
    return WFS_RESULT_OK;
}

WFSResult WFSSetFileName(WFSFileName *pFileName, const utf8 *sFileName) {
    WFSReturnOnNullParameter(sFileName);
    u32 nStrLen = strlen(sFileName);
    if (nStrLen > WFS_MAX_FILE_NAME_SIZE) {
        return WFS_RESULT_INVALID;
    }
    pFileName->nLen = (WFSFileNameLengthType)nStrLen;
    strcpy(pFileName->sStr, sFileName);
    return WFS_RESULT_OK;
}

WFSResult WFSSetDeviceName(WFSDeviceName *pDeviceName, const utf8 *sDeviceName) {
    WFSReturnOnNullParameter(sDeviceName);
    u32 nStrLen = strlen(sDeviceName);
    if (nStrLen > WFS_MAX_DEVICE_NAME_SIZE) {
        return WFS_RESULT_INVALID;
    }
    pDeviceName->nLen = (WFSDeviceNameLengthType)nStrLen;
    strcpy(pDeviceName->sStr, sDeviceName);
    return WFS_RESULT_OK;
}

WFSResult WFSSetVolumeId(WFSVolumeId *pVolumeId, const utf8 *sVolumeId) {
    WFSReturnOnNullParameter(sVolumeId);
    u32 nStrLen = strlen(sVolumeId);
    if (nStrLen > WFS_MAX_VOLUME_ID_SIZE) {
        return WFS_RESULT_INVALID;
    }
    pVolumeId->nLen = (WFSVolumeIdLengthType)nStrLen;
    strcpy(pVolumeId->sStr, sVolumeId);
    return WFS_RESULT_OK;
}
