/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Api.cpp - server side of the WFS API for WFSDEV


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Api.cpp,v $
  Revision 1.112  2008/12/18 09:44:36  ooizumi
  Fixed a typo in WFSSrvCreateDirectory().

  Revision 1.111  2008/12/15 02:40:07  ueno
  Added support for multi p-tree.

  Revision 1.110  2008/12/10 06:50:36  kondo_masahiro
  Added a argument WFSDevType nDevType in WfsKrnAttachDetachCallback.

  Revision 1.109  2008/12/05 06:02:08  ooizumi
  Added ifdef not to enable permission check in WFSSrvInitializeDevice().

  Revision 1.108  2008/12/04 00:45:55  ooizumi
  Modified WFSSrvInitializeDevice() to return PERMISSION error if caller is not ROOT.
  Changed definition name to enable permission.
  Modified error messages.

  Revision 1.107  2008/12/03 00:09:48  kondo_masahiro
  Fixed WFSSrvFlush().

  Revision 1.106  2008/12/01 10:05:42  kondo_masahiro
  Fixed error handling in WFSSrvCreateDirectory.

  Revision 1.105  2008/11/28 06:48:10  kondo_masahiro
  Fixed a bug around file edit in WFSSrvWriteFile.

  Revision 1.104  2008/11/28 04:43:21  kondo_masahiro
  Fixed a bug around file edit in WFSSrvWriteFile.

  Revision 1.103  2008/11/28 04:38:16  kondo_masahiro
  Fixed a bug around file edit in WFSSrvWriteFile.

  Revision 1.102  2008/11/28 04:33:42  kondo_masahiro
  Fixed a bug around file edit in WFSSrvWriteFile.

  Revision 1.101  2008/11/27 08:55:41  saito_tomoya
  Modified to invalidate BCache after check disk.

  Revision 1.100  2008/11/26 01:44:16  kondo_masahiro
  Fixed WFSSrvRead/Write in order to improve access speed.

  Revision 1.99  2008/11/18 07:35:45  kondo_masahiro
  Fixed WFSSrvMoveLocal()

  Revision 1.98  2008/11/18 04:53:46  saito_tomoya
  Modified WFSSrvCheckDisk() to use a bit array file.

  Revision 1.97  2008/11/17 04:45:46  kondo_masahiro
  Fixed WFSSrvMoveLocal()

  Revision 1.96  2008/11/14 06:51:19  ueno
  Fixed a typo.

  Revision 1.94  2008/11/12 13:30:20  ueno
  Modified WFSSrvMoveLocal(), Delete() and CreateDirectory() to return WFSKRN_RESULT_PERMISSION when they modify system paths (/, /dev, /vol).

  Revision 1.93  2008/11/12 09:08:12  kondo_masahiro
  Removed the debug print.

  Revision 1.92  2008/11/12 04:24:12  kondo_masahiro
  Fixed the codes that work buffer for buffer rotation method is obtained from stack.

  Revision 1.91  2008/11/12 01:45:27  kondo_masahiro
  Comment out the contents of WFSSrvCheckDisk().

  Revision 1.90  2008/11/10 01:40:10  ooizumi
  Added permission check sequence to MoveLocal().

  Revision 1.89  2008/11/10 01:00:24  ueno
  Added a workaround to WFSSrvMoveLocal().

  Revision 1.88  2008/11/07 08:52:21  kondo_masahiro
  Fixed WFSSrvMoveLocal()

  Revision 1.87  2008/11/07 04:22:45  ueno
  Fixed AdjustMemorySizeForBCache().

  Revision 1.86  2008/11/05 16:48:37  ueno
  Fixed.

  Revision 1.85  2008/11/05 15:08:20  ueno
  Modified DirItrMapBlks() to roll back the update map in case of error.

  Revision 1.84  2008/11/05 09:58:50  kondo_masahiro
  Fixed WFSMoveLocal()

  Revision 1.83  2008/11/05 05:54:51  kondo_masahiro
  Added Chainning system for indirect block in WFSWriteFile()

  Revision 1.82  2008/11/05 05:49:46  ooizumi
  Fixed WFSSrvMoveLocal() to insert a directory entry which has not cached yet to PathCache.

  Revision 1.81  2008/11/04 00:23:13  kondo_masahiro
  Added WFSSrvMoveLocal()

  Revision 1.80  2008/10/31 09:51:11  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.79  2008/10/30 06:34:18  kondo_masahiro
  Fixed WFSSrvCheckDisk

  Revision 1.78  2008/10/30 05:06:03  kondo_masahiro
  Added WFSSrvCheckDisk

  Revision 1.77  2008/10/27 13:46:11  ueno
  Modified WFSSrvInit() to enable attach/detach callbacks when to reinitialize WFS.

  Revision 1.76  2008/10/21 11:34:11  ueno
  Modified WFSSrvGetFreeSpaceSize() to return WFSKRN_RESULT_INVALID for virtual paths.

  Revision 1.75  2008/10/21 09:49:37  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.74  2008/10/20 09:27:58  ueno
  Added casts for porting IOP.

  Revision 1.73  2008/10/20 05:17:16  ueno
  Modified WFSSrvExecDebugProcedure() to support WFSDEBUG_CMD_ATTACH and WFSDEBUG_CMD_DETACH options.

  Revision 1.72  2008/10/17 08:50:48  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.71  2008/10/17 05:34:15  ooizumi
  Fixed to update pKfi->attr in WFSSrvOpenFile() and WFSSrvWriteFile().

  Revision 1.70  2008/10/16 09:32:22  ooizumi
  Fixed definitions for debug build.
  Fixed a bug failed to handle invalid nPos value error in WFSSrvReadFile() and
  WFSSrcWriteFile().

  Revision 1.69  2008/10/15 03:51:46  ooizumi
  Fixed a bug in WFSSrvReadFile().

  Revision 1.68  2008/10/15 00:14:44  kondo_masahiro
  Fixed a bug about 4GB-1

  Revision 1.67  2008/10/14 10:26:24  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.66  2008/10/14 10:23:28  ooizumi
  Fixed permission APIs to return WFSKRN_RESULT_PERMISSION error when
  the path which has no area info was specified.

  Revision 1.65  2008/10/14 08:53:45  ooizumi
  Fixed WFSSrvReadFile().

  Revision 1.64  2008/10/14 06:15:37  ueno
  Fixed WFSSrvReadFile to return WFSKRN_RESULT_INVALID when the specified position exceeds the end of the file.

  Revision 1.63  2008/10/14 04:15:41  ooizumi
  Modified WFSSrvSearchDirectoryFirst() to convert WFSKRN_RESULT_NOT_DIRECTORY
  to WFSKRN_RESULT_NOT_FOUND.

  Revision 1.62  2008/10/14 03:23:15  ooizumi
  Fixed WFSSrvUnmountVolume() to check bForce.
  Fixed WFSSrvGetVolumeId() to convert device name string to lower case.

  Revision 1.61  2008/10/13 07:25:26  ueno
  Fixed WFSSrvUnmountVolume() to return WFSKRN_RESULT_FILE_OPEN when 0 < pEntry->pcd.nOpenHandleCount.

  Revision 1.60  2008/10/10 10:53:33  ooizumi
  Fixed a bug WFSSearchDirectry() fails to search path cache entries.

  Revision 1.59  2008/10/10 04:50:35  ueno
  Fixed WFSSrvCloseFile() to reduce pin count in case of permission error.

  Revision 1.58  2008/10/10 04:45:41  ooizumi
  Modified to return not implemented error when searching path caching directries.

  Revision 1.57  2008/10/10 02:11:19  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.56  2008/10/10 01:19:46  ooizumi
  Fixed debug code in WFSSrvOpenFile().

  Revision 1.55  2008/10/09 07:18:24  ooizumi
  Modified WFSSrvCreateAndOpenFile() to check nFlags.

  Revision 1.54  2008/10/09 07:02:18  ooizumi
  Fixed a bug in WFSSrvCreateDirectry() and WFSSrvCreateAndOpenFile().

  Revision 1.53  2008/10/09 05:45:45  ooizumi
  Fixed to consider WFS_FLAG_CONTIGUOUS_AREA flag, but contiguous area is not yet implemented.

  Revision 1.52  2008/10/09 04:16:53  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.51  2008/10/09 03:27:25  ooizumi
  Implemented permission check sequence in WFSSrvValidateDirectory.

  Revision 1.50  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.49  2008/10/08 13:18:42  ueno
  Added AdjustMemorySizeForBCache() to adjust user data cache.

  Revision 1.48  2008/10/07 11:22:38  ueno
  Implemented WFSSrvGetFreeSpaceSize().
  Fixed content index and timeUpdated in WFSSrvGetFreeSpaceSize().
  Modified WFSSrvCreateDirectory() to check nFlags.

  Revision 1.47  2008/10/07 11:02:58  ooizumi
  Fixed to call WFSKrnDbgOutputInit first in WFSKrnInit.

  Revision 1.46  2008/10/07 10:41:56  ooizumi
  Several updates for permission functions.

  Revision 1.45  2008/10/07 05:55:36  ueno
  Added a workaround for PathCacheDecreaseEntryOpenHandleCount().

  Revision 1.44  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.43  2008/10/03 02:08:23  ueno
  Modified UnmountAllVolume() to call Srv APIs.
  Added a workaround for WFSSrvSearchDirectoryNext().

  Revision 1.42  2008/10/02 11:02:22  ueno
  Modified WFSSrvUnmountVolume() to return WFS_RESULT_FILE_OPEN when the volume has open files.

  Revision 1.41  2008/10/02 08:27:35  ueno
  Modified WFSSrvExit() to unmount all volumes.

  Revision 1.40  2008/09/30 02:01:19  ueno
  Modified to support large disk ( < 32TB).

  Revision 1.39  2008/09/29 10:19:36  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.38  2008/09/28 23:31:11  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.37  2008/09/26 11:06:24  ueno
  Fixed WFSSrvGetDeviceInfo() to work well even when the device is not mounted.

  Revision 1.36  2008/09/26 07:39:25  ooizumi
  Supported to search "/ *", "/dev/ *" and "/vol/ *".

  Revision 1.35  2008/09/05 02:05:00  ooizumi
  Fixed to always update pKfi->attr.file.nSize in WFSSrvWriteFile.

  Revision 1.34  2008/09/01 01:57:05  kondo_masahiro
  Minor fixed for IOP compiler.

  Revision 1.33  2008/08/29 00:08:12  kondo_masahiro
  Fixed WFSSrvGetVolumeId and WFSSrvExecDebugProcedure.

  Revision 1.32  2008/08/27 23:11:12  ooizumi
  Fixed.

  Revision 1.31  2008/08/27 23:10:20  ooizumi
  Fixed broken lines.

  Revision 1.30  2008/08/27 23:02:55  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.29  2008/08/27 09:36:15  paul
  Started coding for different size categories

  Revision 1.28  2008/08/27 04:13:46  nakanose_jin
  add assert

  Revision 1.27  2008/08/22 06:38:22  nakanose_jin
  kill warning

  Revision 1.26  2008/08/19 03:22:38  ooizumi
  Fixed not to free allocated info buffer before calling TransEnd.

  Revision 1.25  2008/08/07 09:10:34  ooizumi
  Implemented WFSSrvGetPermisson and WFSSrvSetPermission.
  Fixed a bug WFS_PERM_CHANGE_SIZE doesn't effect in WFSSrvWriteFile.
  Fixed a bug in WFSSrvWriteFile TransUnpinBlk is not called when DirResizeFile is failed.
  Fixed build errors.

  Revision 1.24  2008/08/05 04:07:58  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.23  2008/08/04 22:49:36  paul
  Fixed a bug in WFSSrvCloseFile() which caused an error when _CHECK_BLK_USAGE = 0.

  Revision 1.22  2008/07/31 23:59:19  ooizumi
  Fixed permission of special directories..

  Revision 1.21  2008/07/30 21:14:58  paul
  Minor changes (debug output)

  Revision 1.20  2008/07/30 06:11:57  ooizumi
  Implemented initial version of permission functions.

  Revision 1.19  2008/07/29 22:50:54  ooizumi
  Modified to call TransBegin and TransEnd.
  Fixed several bugs.

  Revision 1.18  2008/07/28 21:45:37  paul
  Improved updateMap processing - should be more correct, and no longer needs a global update counter.
  Fixed a bug in WFSSrvDelete()

  Revision 1.17  2008/07/25 02:51:28  paul
  Changed from nTransIdx to *pTransInfo

  Revision 1.16  2008/07/21 22:46:05  paul
  Fixed bugs in WFSSrvCloseFile(), WFSSrvWriteFile()
  Changed name from DirItrPinBlkAndUpdatePtrs() -> DirItrPinBlkAndFixPtrs()

  Revision 1.15  2008/07/18 02:10:00  ueno
  Rename WFSDeletePermissions() -> WFSDeleteAccessListEntry().

  Revision 1.14  2008/07/17 22:44:41  paul
  Fixed reading and writing small files via block cache. Changes to support updateMap (not completely finished yet).

  Revision 1.13  2008/07/17 05:30:21  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.12  2008/07/15 07:44:18  nakanose_jin
  change arg order

  Revision 1.11  2008/07/14 06:27:51  ueno
  Changed WFSSrvSetAccessList() to WFSSrvSetAccessListEntry().

  Revision 1.10  2008/07/14 00:39:22  ueno
  Added WFSSrvDeletePermissions(), WFSSrvGetAccessList() and WFSSrvSetAccessList().

  Revision 1.9  2008/07/08 23:39:05  paul
  Many changes - to support accessing a real drive via the block cache, and integrate the device & volume mount/unmount

  Revision 1.8  2008/06/09 18:03:00  paul
  Many changes including supporting new Path Cache.

  Revision 1.7  2008/05/22 07:57:25  ueno
  Fixed debug flag.

  Revision 1.6  2008/05/19 14:11:55  ueno
  Added WFSSrvDebugGetRootAreaInfo() for accesslist tests.

  Revision 1.5  2008/05/16 18:51:31  paul
  Changed WFSSrvSearchDirectoryNext so that it still works while adding/deleting files in the directory.

  Revision 1.4  2008/05/12 19:05:15  paul
  Changed names of PathCache defs

  Revision 1.3  2008/05/10 03:59:43  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:29:34  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.16  2008/04/23 00:23:31  paul
  Fixed a bug with pattern matcher

  Revision 1.15  2008/04/23 00:09:07  paul
  Updated WFSSrv calls to be closer to the real implementation

  Revision 1.14  2008/04/19 05:51:31  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.13  2008/04/18 00:08:33  paul
  Changed interface to Dir functions

  Revision 1.12  2008/04/15 18:36:59  paul
  Changed definition of DirFind.

  Revision 1.11  2008/04/15 00:27:04  ueno
  Added WFSSrvGetPermissionForUser() and WFSSrvSetPermissionForUser().

  Revision 1.10  2008/04/08 18:43:35  paul
  Changed WFS_PERM_OWNER_RW to WFS_PERM_F_ALL

  Revision 1.9  2008/04/04 21:07:20  paul
  Moved flag settings into WFSSrvCreateDirectory() instead of at call site.

  Revision 1.8  2008/04/03 23:46:01  kondo_masahiro
  change spec of debug proc

  Revision 1.7  2008/03/07 01:33:51  paul
  Changed WFSSrvGetVolumeId temporary implementation to use wkg.areHdrRoot
  Changed DirFind spec to use DIR_PREFIX_SEARCH for wildcard support.

  Revision 1.6  2008/02/28 06:30:42  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.5  2008/02/27 07:26:52  paul
  made areaHdrRoot a global so it could be accessed from debug code in wfskrn_Dir.cpp

  Revision 1.4  2008/02/27 00:41:33  paul
  Added WFSSrvExecDebugProcedure

  Revision 1.2  2007/12/27 12:06:17  paul
  *** empty log message ***

  Revision 1.1  2007/12/27 11:17:20  paul
  mostly dummy functions for now.. only what I needed to start working on Dir module

  Revision 1.7  2007/11/02 00:16:30  paul
  started logging changes in the file
  Added StartPosition to WFSSrvReadFile, WFSSrvWriteFile

*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "Api"

#include "wfskrn_Config.h"
#include "wfskrn_Api.h"
#include "wfskrn_Handles.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_DirRxTree.h"
#include "wfskrn_PathCache.h"
#include "wfskrn_PTree.h"
#include "wfskrn_Trans.h"
#ifdef PERMISSION_ENABLED
#include "wfskrn_Permission.h"
#include "wfskrn_Permission_AccessList.h"
#endif
#include "wfskrn_FreeBlkAlloc.h"
WFSKrnResult AreaPinFreeListHdr(AreaInfo *pAreaInfo, u32 nFlags, AreaFreeListHdr **ppAflh);
void AreaUnpinFreeListHdr(AreaInfo *pAreaInfo);

#include "randomlib.h"

#if _WIN32
#include "wfs_Client.h"
#endif

#include <drnand.h>

WFSKrnGlobals wkg ATTRIBUTE_ALIGN(512);

DirItr         gDi;

AreaInfo   *pAhInit;
WFSBlkAdr  *pBaInit;

//u8  *gpPtr = 0; // Pointer of vdsk memory
//u32  gSize = 0; // Size of vdsk memory

//u32 gNumSectors = 0; // Size of dsk memory

//u8  *gpCachePtr = 0; // Pointer of cache memory
//u32  gCacheSize = 0; // Size of cache memory

u8  *gpHeapPtr = 0; // Pointer of heap memory
u32  gHeapSize = 0; // Size of heap memory

////////////////////////////////////////////
#define MAX_ALLOC_NUM_SECTORS_FOR_LIST_ALLOCATOR  (32*1024*2) // for LIST_BASED_ALLOCATOR (32MB)
////////////////////////////////////////////

// #define MAX_ALLOC_NUM_SECTORS                     (1024*1024*2)
#define MAX_ALLOC_NUM_SECTORS                     (1024*1024*2*8/32)

#define TEST_LOG2_SECTOR_SIZE 9

#define CHECKDISKFILENAME "TestCheckBitArrayFile"

u32 WfsKrnGetCurrentClient() {
    // Return client index (0..31)
    return 0;   // ToDo: Support multiple clients.
}

// Arc file - IOP.


static void WfsKrnAttachDetachCallback(u32 bAttach, u32 nDevIdx, u32 nLog2SectorSize, u32 nNumSectors, WFSDevType nDevType, void *pUserData){
    //MyOSReport("WfsKrnAttachDetachCallback(), bAttach = %d, nDevIdx = %d\n", bAttach, nDevIdx);

    WFSKrnResult nResult;
    if(nDevIdx < 0 || nDevIdx >= WFSKRN_MAX_DEVICES){
        WFSKrnOutputErrorStr("Error: WfsKrnAttachDetachCallback(), nDevIdx is incorrect.\n");
        return;
    }
    if(bAttach){
        DeviceInfo *pDevInfo = &wkg.aDevInfo[nDevIdx];
#if !BTREE_BASED_ALLOCATOR
        u32 nMaxSectors = MAX_ALLOC_NUM_SECTORS_FOR_LIST_ALLOCATOR;
        if(nNumSectors > nMaxSectors){
            nNumSectors = nMaxSectors;
        }
#endif
        //osTPrintf("nLog2SectorSize = %d, nTotalSize>>nLog2SectorSize = %d\n", nLog2SectorSize, nNumSectors);
        nResult = DeviceAttach(pDevInfo, nDevType, nDevIdx, nLog2SectorSize, nNumSectors);
        if (nResult != WFSKRN_RESULT_OK)
        {
            WFSKrnOutputErrorStr("Device not found.\n");
            return;
        }

        pDevInfo->pVolInfo = 0;
        //PathCacheDebugOutput();
#if _WIN32
        // return callback
        if(wcg.dcg.pAttachDeviceNameBuffer){
            strcpy(wcg.dcg.pAttachDeviceNameBuffer, pDevInfo->devName.sStr);
        }
        if(wcg.dcg.cbAttachDevice){
            (*wcg.dcg.cbAttachDevice)(pDevInfo->devName.sStr, wcg.dcg.pAttachDeviceUserData);
        }
#endif
    }
    else{
        DeviceInfo *pDevInfo = &wkg.aDevInfo[nDevIdx];
        DeviceDetach(pDevInfo);
        //PathCacheDebugOutput();
#if _WIN32
        // return callback
        if(wcg.dcg.pDetachDeviceNameBuffer){
            strcpy(wcg.dcg.pDetachDeviceNameBuffer, pDevInfo->devName.sStr);
        }
        if(wcg.dcg.cbDetachDevice){
            (*wcg.dcg.cbDetachDevice)(pDevInfo->devName.sStr, wcg.dcg.pDetachDeviceUserData);
        }
#endif
    }
}

static WFSResult AdjustMemorySizeForBCache(u32* nWorkSpaceSize, u32* nBCacheMeta, u32* nBlkUser64K, u32* nBlkUser8K)
{
    u32 nBCacheMax = WFS_BCACHE_META_DATA_SIZE / (sizeof(BCacheEntry) + 2 * sizeof(BCacheListNode));
    *nBCacheMeta = WFS_BLK_CACHE_SIZE / (1<<WFS_LOG2_SMALL_BLK_SIZE);
    if (nBCacheMax <= *nBCacheMeta)
    {
        return WFS_RESULT_INVALID;
    }
    u32 ratio = 1<<(WFS_LOG2_MEDIUM_BLK_SIZE - WFS_LOG2_SMALL_BLK_SIZE);
    u32 nMemSizeMax = (1<<(WFS_LOG2_SMALL_BLK_SIZE)) * ((nBCacheMax - *nBCacheMeta)*(2 * ratio)/(ratio + 1));
    if (nMemSizeMax < *nWorkSpaceSize)
    {
        *nWorkSpaceSize = nMemSizeMax;
        *nBlkUser64K = (nBCacheMax - *nBCacheMeta) / (ratio + 1);
        *nBlkUser8K = *nBlkUser64K * ratio;
    }
    else
    {
        *nBlkUser64K = (*nWorkSpaceSize / 2) / (1<<WFS_LOG2_MEDIUM_BLK_SIZE);
        *nBlkUser8K = (*nWorkSpaceSize / 2) / (1<<WFS_LOG2_SMALL_BLK_SIZE);
        *nWorkSpaceSize = *nBlkUser64K * (1<<WFS_LOG2_MEDIUM_BLK_SIZE) +
                          *nBlkUser8K * (1<<WFS_LOG2_SMALL_BLK_SIZE);
    }
    return WFS_RESULT_OK;
}

static WFSResult SystemPathInit(void)
{
    PathCacheEntry* entry;
    entry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, (utf8*)""    );
    if (!entry)
    {
        return WFS_RESULT_RESOURCE_LIMIT_EXCEEDED;
    }
    entry->pcd.pData = (void*) 0;

    entry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, (utf8*)"/dev");
    if (!entry)
    {
        return WFS_RESULT_RESOURCE_LIMIT_EXCEEDED;
    }
    entry->pcd.pData = (void*)WFSKRN_RESULT_SRV_PATH_DEV;

    entry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, (utf8*)"/vol");
    if (!entry)
    {
        return WFS_RESULT_RESOURCE_LIMIT_EXCEEDED;
    }
    entry->pcd.pData = (void*)WFSKRN_RESULT_SRV_PATH_VOL;

    return WFS_RESULT_OK;
}

WFSResult WFSKrnInit(void *pWorkSpaceBuffer, u32 nWorkSpaceSize)
{
#if _DEBUG
    WFSKrnDbgOutputInit("CommandLog.txt");
#endif
    u32 nBCacheMeta;
    u32 nBlkUser64K;
    u32 nBlkUser8K;

    WFSResult nResult = AdjustMemorySizeForBCache(&nWorkSpaceSize, &nBCacheMeta, &nBlkUser64K, &nBlkUser8K);
    if (nResult != WFS_RESULT_OK)
    {
        return nResult;
    }

    //MyOSReport("pWorkSpaceBuffer = 0x%x, nWorkSpaceSize = %d\n", pWorkSpaceBuffer, nWorkSpaceSize);
    TransModuleInit();
    WFSKrnInitHandles();
    wkg.time = 0;
//    wkg.pBlkCache = (u8*)(((size_t)wkg.aStaticMem + WFS_STATIC_MEM_ALIGNMENT-1)&-WFS_STATIC_MEM_ALIGNMENT);
    wkg.pBlkCache = (u8*)(((size_t)wkg.aStaticMem + WFS_STATIC_MEM_ALIGNMENT-1)&~(WFS_STATIC_MEM_ALIGNMENT-1));

    // Prepare block cache meta data for all blocks.
    size_t nMemSize = (nBCacheMeta + nBlkUser64K + nBlkUser8K) * (sizeof(BCacheEntry) + 2 * sizeof(BCacheListNode));
    if (nMemSize > WFS_BCACHE_META_DATA_SIZE) {
        WFSKrnOutputErrorStr("BCache MetaData Size too small");
    }
    u8 *pBlkCacheMetaData = (u8*)((size_t)wkg.pBlkCache+WFS_BLK_CACHE_SIZE);
    BCacheInit(WFS_LOG2_SMALL_BLK_SIZE, pBlkCacheMetaData, nMemSize);

    // Prepare the block cache for meta data blocks on the memory for IOP.
    if (BCacheGroupInit(BCACHE_GROUP_ID_METABLK, wkg.pBlkCache, WFS_BLK_CACHE_SIZE, WFS_LOG2_SMALL_BLK_SIZE)) {
        return WFS_RESULT_INVALID;
    }

    // Prepare the block cache for user data blocks on the memory used by PPC.
    // 8KB blocks
    if (BCacheGroupInit(BCACHE_GROUP_ID_USERBLK_8K, (u8*)pWorkSpaceBuffer, nBlkUser8K * (1<<WFS_LOG2_SMALL_BLK_SIZE), WFS_LOG2_SMALL_BLK_SIZE)) {
        return WFS_RESULT_INVALID;
    }

    pWorkSpaceBuffer = (void*)((size_t)pWorkSpaceBuffer + nBlkUser8K * (1<<WFS_LOG2_SMALL_BLK_SIZE));
    // 64KB blocks
    if (BCacheGroupInit(BCACHE_GROUP_ID_USERBLK_64K, (u8*)pWorkSpaceBuffer, nBlkUser64K * (1<<WFS_LOG2_MEDIUM_BLK_SIZE), WFS_LOG2_MEDIUM_BLK_SIZE)) {
        return WFS_RESULT_INVALID;
    }

    // Create heap data
    wkg.pHeapData = (u8*)((size_t)pBlkCacheMetaData + WFS_BCACHE_META_DATA_SIZE);
    WFSKrnCreateHeap(&wkg.heap, wkg.pHeapData, WFS_STATIC_HEAP_SIZE);

    DirModuleInit();
    VolumeModuleInit();

    nResult = SystemPathInit();
    if (nResult != WFS_RESULT_OK)
    {
        return nResult;
    }

    DirInitDirEntryAttrHdr(&wkg.rootDirAttr);
    DirInitDirEntryAttrHdr(&wkg.volDirAttr);
    DirInitDirEntryAttrHdr(&wkg.devDirAttr);
    DirInitFileEntryAttrHdr(&wkg.devAttr);
    DeviceInitLib();

#if _DEBUG
    PathCacheDebugOutput();
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput1("Init() WFS_RESULT_OK\n"));
#endif
    return WFS_RESULT_OK;
}


WFSResult WFSSrvInit(void *pWorkSpaceBuffer, u32 nWorkSpaceSize) {
    // ToDo:
    // This code should be split into a part that is called when the Kernel is initialized when the firmware is loaded,
    // and a part which is done each time WFSInit() is called
    WFSResult nResult = WFS_RESULT_OK;
    if (!wkg.bKrnInitialized) {
        wkg.bKrnInitialized = true;
#if _WIN32
#define WFSSHIM_WORK_SPACE_SIZE (40*1024)
        nWorkSpaceSize -= WFSSHIM_WORK_SPACE_SIZE; // Adjust work space buffer because shim layer in RVL program use the buffer.
#endif
        ASSERT(nWorkSpaceSize >= WFS_MIN_WORK_SPACE_SIZE && nWorkSpaceSize <= WFS_MAX_WORK_SPACE_SIZE);
        nResult = WFSKrnInit(pWorkSpaceBuffer, nWorkSpaceSize);
    }
#if _WIN32
    //PTreeTest();
    //PathCacheTest();
    WfsKrnAttachDetachCallback(true, 0, TEST_LOG2_SECTOR_SIZE, MAX_ALLOC_NUM_SECTORS, WFS_DEVTYPE_USB_MSC_20, 0);
#elif _TWL
    {
        u32 sectors = 0;
        nandNfsCtrl( 0, NFS_DEVCTL_GET_GEOMETRY, &sectors);
        WfsKrnAttachDetachCallback(true, 0, TEST_LOG2_SECTOR_SIZE, sectors, WFS_DEVTYPE_USB_MSC_20, 0);
    }
#elif _IOP
    DeviceAttachDetachRegistCallback(WfsKrnAttachDetachCallback, 0);
#endif

    return nResult;
}


static WFSResult UnmountAllVolume(void)
{
    WFSResult result;
    WFSFileInfo wfi;
    WFSVolumeId volumeId;
    WFSPathName pathNameAbsPattern;
    WFSSrvSearchDirectoryHandle ssdh;

    osSNPrintf(pathNameAbsPattern.sStr, WFS_MAX_VOLUME_ID_SIZE, "/vol/*");
    pathNameAbsPattern.nLen = STRNLEN(pathNameAbsPattern.sStr, WFS_MAX_PATH_NAME_SIZE);

    result = WFSSrvSearchDirectoryFirst(&pathNameAbsPattern, &ssdh, &wfi);
    while (result == WFS_RESULT_OK)
    {
        volumeId.nLen = STRNLEN(wfi.sFileName, WFS_MAX_VOLUME_ID_SIZE);
        ASSERT(volumeId.nLen == WFS_MAX_VOLUME_ID_SIZE);
        memmove(volumeId.sStr, wfi.sFileName, volumeId.nLen+1);
        result = WFSSrvUnmountVolume(&volumeId, (WFSBool) TRUE);
        if (result != WFS_RESULT_OK)
        {
            goto UNMOUNTALL_RESULT;
        }
        break; // [check] workaround (WFSSrvSearchDirectoryNext() returns exception now).
        // result = WFSSrvSearchDirectoryNext(ssdh, &wfi); // [check] must call next().
    }

    if (result == WFS_RESULT_NOT_FOUND)
    {
        result = WFS_RESULT_OK;
    }

UNMOUNTALL_RESULT:
    WFSSrvSearchDirectoryClose(ssdh); // [check] always can call close()?
    return result;
}


WFSResult WFSSrvExit(void) {
    WFSResult result = UnmountAllVolume();
#if _DEBUG
    ++nDbgCommandCount;
    if (result == WFS_RESULT_OK)
    {
        WFSKrnDbgOutputTest(WFSKrnDbgOutput1("Exit() WFS_RESULT_OK\n"));
    }
#endif
    return result;
}


/*u32 WFSKrnCreateTransIdx() {
    static bool bFirstTime = true;
    static u32 nTransIdx = 0;
    static WFSKrnMutex mx;
    if (bFirstTime) {
        bFirstTime = false;
        WFSKrnInitMutex(&mx);
    }
    WFSKrnLockMutex(&mx);
    ++nTransIdx;
    WFSKrnUnlockMutex(&mx);
    return nTransIdx;
}*/


WFSResult WFSSrvInitializeDevice(const WFSDeviceName *pDevName) {
    WFSKrnResult nResult;
    TransBlkAdr tba;
    tba.pTransInfo = TransBegin();
#ifdef PERMISSION_ENABLED
    // ToDo: Now we cannot access from NDEV by ROOT. It is need to implement SetTitleId on WFSKRN.
    if(tba.pTransInfo->nTitleId != WFS_ROOT_ID)
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit;
    }
#endif
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(pDevName);
    if (pDevInfo == 0) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE) {
        nResult = WFSKRN_RESULT_DEV_UNUSABLE;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_WRITE_PROTECTED) {
        nResult = WFSKRN_RESULT_WRITE_PROTECTED;
        goto Exit;
    }
    VolumeInfo *pVolInfo;
    if (pDevInfo->pVolInfo == 0) {
        pDevInfo->pVolInfo = VolumeAlloc();
        VolumeAddToMountedList(pDevInfo->pVolInfo);
        pVolInfo = pDevInfo->pVolInfo;
        pVolInfo->nClientsMountedBy = 0;
        pVolInfo->pDevInfo = pDevInfo;
    } else {
        pVolInfo = pDevInfo->pVolInfo;
        if (pVolInfo->nClientsMountedBy) {
            nResult = WFSKRN_RESULT_DEV_IN_USE;
            goto Exit;
        }
    }
    nResult = VolumeInit(pVolInfo, WFS_LOG2_SMALL_BLK_SIZE, (pDevInfo->dh.nNumSectors>>WFS_LOG2_SMALL_BLK_SIZE) * pDevInfo->nSectorSize);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }
    tba.pAreaInfo = &pVolInfo->rootAreaInfo;
    tba.nBlkAdr   = pVolInfo->rootAreaInfo.ah.nRootBlkAdr;
    DirBlkHdr *pBlkHdr;
    nResult = DirCreateAndInitNewBlk(&tba, (WFSMetaDataFlags)(WFS_MDF_DIR_LEAF_BLK | WFS_MDF_DIR_ROOT_BLK), &pBlkHdr); // Map
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }
    DirBlkCreateRootNode(pBlkHdr);
    TransUnpinBlk(&tba);
    VolumeUnmount(pVolInfo);
    pDevInfo->nFlags &= ~WFS_DEVFLAG_NOT_INITIALIZED;
    nResult = WFSKRN_RESULT_OK;
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("InitializeDevice(\"%s\") WFS_RESULT_%s\n", pDevName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    TransEnd(tba.pTransInfo);
    return (WFSResult)nResult;
}


WFSResult WFSSrvGetDeviceInfo(const WFSDeviceName *pDevName, WFSDeviceInfo *pDi) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(pDevName);
    if (pDevInfo == 0) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    pDi->nDevFlags = pDevInfo->nFlags;
    if (0 == (pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED)) {
        VolumeInfo *pVolInfo = pDevInfo->pVolInfo;
        if (!pVolInfo || (0 == (pVolInfo->nClientsMountedBy & (1<<WfsKrnGetCurrentClient())))) {
            pDi->nDevFlags |= WFS_DEVFLAG_NOT_MOUNTED;
        }
    }
    pDi->nDevType          = pDevInfo->nDevType;
    pDi->nDevTotalCapacity = pDevInfo->nTotalCapacity;
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput2("GetDeviceInfo(\"%s\", pDi) WFS_RESULT_OK\n", pDevName->sStr));
#endif
    return (WFSResult) nResult;
}


WFSResult WFSSrvMountDevice(const WFSDeviceName *pDevName, WFSVolumeId *pVolumeIdBuffer) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(pDevName);
    if (pDevInfo == 0) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE) {
        nResult = WFSKRN_RESULT_DEV_UNUSABLE;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED) {
        nResult = WFSKRN_RESULT_DEV_NOT_INITIALIZED;
        goto Exit;
    }
    VolumeInfo *pVolInfo = pDevInfo->pVolInfo;
    u32 nClientBit = 1<<WfsKrnGetCurrentClient();
    if (pVolInfo) {
        if (pVolInfo->nClientsMountedBy & nClientBit) {
            nResult = WFSKRN_RESULT_ALREADY_MOUNTED;
            goto Exit;
        }
    } else {
        nResult = DeviceMountVolume(pDevInfo);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Exit;
        }
        pVolInfo = pDevInfo->pVolInfo;
    }
    WFSVolumeId volId;
    VolumeConvertU32ToVolId(&volId, pVolInfo->vh.nVolId);
    memcpy(pVolumeIdBuffer->sStr, volId.sStr, volId.nLen+1);
#if _DEBUG
    pVolInfo->volId = volId;
#endif
    pVolInfo->nClientsMountedBy |= nClientBit;
    if (pVolInfo->nClientsMountedBy == nClientBit) {
        utf8 sVolPath[WFS_MAX_VOLUME_ID_SIZE + 6];
        osSNPrintf(sVolPath, WFS_MAX_VOLUME_ID_SIZE + 6, "/vol/%s", volId.sStr);
        PathCacheEntry *pEntry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, sVolPath);
        if (pEntry)
        {
            pEntry->pcd.pData = pVolInfo;
            DirInitDirEntryAttrHdr(&pVolInfo->rootAttr);
        }
        else
        {
            nResult = WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED;
        }
    }
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput2("MountDevice(\"%s\", vid) ", pDevName->sStr);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("vid=%s ", pVolumeIdBuffer->sStr);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvUnmountVolume(const WFSVolumeId *pVolumeId, WFSBool bForce) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    utf8 sVolPath[WFS_MAX_VOLUME_ID_SIZE + 6];
    osSNPrintf(sVolPath, WFS_MAX_VOLUME_ID_SIZE + 6, "/vol/%s", pVolumeId->sStr);
    VolumeInfo *pVolInfo;
    RxtItr rxi;
    rxi.pStrPtr = sVolPath;
    PathCacheEntry *pEntry = PathCacheFind2(&rxi, (void **)&pVolInfo);
    if (!pEntry) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    if (0 < pEntry->pcd.nOpenHandleCount &&
        (!bForce || pVolInfo->pDevInfo->bInUse)) // Todo : the root entry will have to be unpined.
    {
#if _DEBUG
        PathCacheDebugOutput();
#endif
        nResult = WFSKRN_RESULT_FILE_OPEN;
        goto Exit;
    }
    u32 nClientBit = 1<<WfsKrnGetCurrentClient();
    if (0 == (pVolInfo->nClientsMountedBy & nClientBit)) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    pVolInfo->nClientsMountedBy &= ~nClientBit;
    if (pVolInfo->nClientsMountedBy == 0) {
        nResult = PathCacheDeleteSubTree(pEntry);
        VolumeUnmount(pVolInfo);
        // ToDo: For better test reproducibility, it may be good to detect special case of no devices mounted - then re-initialize path cache
    }
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput4("UnmountVolume(\"%s\", bForce=%d) WFS_RESULT_%s\n", pVolumeId->sStr, (int)bForce, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvGetVolumeId(const WFSDeviceName *pDevName, WFSVolumeId *pVolumeIdBuffer) {
/*
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(pDevName);
    if (pDevInfo == 0) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE) {
        nResult = WFSKRN_RESULT_DEV_UNUSABLE;
        goto Exit;
    }
    if (pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED) {
        nResult = WFSKRN_RESULT_DEV_NOT_INITIALIZED;
        goto Exit;
    }
    WFSVolumeId volId;
    VolumeInfo *pVolInfo = pDevInfo->pVolInfo;
    ASSERT(pVolInfo);
    VolumeConvertU32ToVolId(&volId, pVolInfo->vh.nVolId);
    //memcpy(pVolumeIdBuffer, volId.sStr, volId.nLen+1);
    *pVolumeIdBuffer = volId;
Exit:
*/
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSDeviceName devName;
    WFSKrnExitOnError(WFSKrnValidateCopyStrAndConvertToLowerCase(devName.sStr, pDevName->sStr, pDevName->nLen, WFS_MAX_DEVICE_NAME_SIZE));
    devName.nLen = pDevName->nLen;
    WFSKrnExitOnError(DeviceGetVolumeId(&devName, pVolumeIdBuffer));
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput2("GetVolumeId(\"%s\", vid) ", pDevName->sStr);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("vid=%s ", pVolumeIdBuffer->sStr);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvGetFreeSpaceSize(const WFSPathName *pAbsPathName, u64 *pSize) {
    WFSKrnResult nResult = WFSKRN_RESULT_INVALID;

    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
#ifdef PERMISSION_ENABLED
    // [check] permissions. NYI.
    /*
    u32 permissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((permissions & WFS_PERM_DELETE_OR_MOVE) == 0)
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    */
#endif
    AreaInfo *pAreaInfo = pci.di.tba.pAreaInfo;
    if (pAreaInfo)
    {
        nResult = AreaPinFreeListHdr(pAreaInfo, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaInfo->pAflh);
        if (nResult == WFSKRN_RESULT_OK)
        {
            u32 numBlk = FreeBlkGetFreeSize(pAreaInfo);
            *pSize = (u64) pAreaInfo->nBlkSize * numBlk;
            AreaUnpinFreeListHdr(pAreaInfo);
        }
    }
    else
    {
        nResult = WFSKRN_RESULT_INVALID;
    }
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("WFSSrvGetFreeSpaceSize(\"%s\", pSize) WFS_RESULT_%s\n", pAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvGetAttributes(const WFSPathName *pAbsPathName, WFSFileAttributes *pFa) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 144761) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
#ifdef PERMISSION_ENABLED
    // Check the permission of parent dir and specified entry
    u32 parentPermissions;
    u32 permissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) == 0 || !(permissions & WFS_PERM_LIST))
    {
        WFSKrnExit1OnError(WFSKrnGetParentInheritedPermissionForUserPci(&pci, &parentPermissions, pci.di.tba.pTransInfo->nTitleId));
        if(!(parentPermissions & WFS_PERM_LIST))
        {
            nResult = WFSKRN_RESULT_PERMISSION;
            goto Exit1;
        }
    }
#endif

#ifdef PERMISSION_ENABLED
    // Get inherited permissions
    pFa->nFlags    = permissions |
        (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) | (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_CONTIGUOUS_AREA);
#else
    pFa->nFlags    = (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY);   // ToDo: Get permission bits for current user
#endif
    if (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
        pFa->nNumEntries = 0;  // ToDo: Remove NumEntries parameter from WFS API
        pFa->nCidx     = 0; // A directory does not have a content index.
    } else {
        pFa->nFileSize = pci.di.pAttrHdr->file.nSize;
        pFa->nCidx     = pci.di.pAttrHdr->file.nContentIdx;
    }
    pFa->timeCreated = pci.di.pAttrHdr->timeCreated;
    pFa->timeUpdated = pci.di.pAttrHdr->timeUpdated;
Exit1:
    DirItrClose(&pci.di);
    if(pci.di.tba.pTransInfo == 0 || pci.di.tba.pTransInfo->link.pNext == 0){
        int a; a=0;
    }
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("GetAttributes(\"%s\", pFa) WFS_RESULT_%s\n", pAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvChangePermissions(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask) {
/*  WFSPathName *papn; papn = (WFSPathName*)pAbsPathName; // for warning
#if _IOP
    pAh = (AreaInfo*)((u32)pAh & 0x7fffffff);
    pBa = (WFSBlkAdr*)((u32)pBa & 0x7fffffff);
#endif
    //osTPrintf("pAreaInfo = 0x%x, pBlkAdr = 0x%x\n",pAh, pBa);
    ///////////////////////////////////////
    WfsInitRootArea(0, pAh, gpPtr, gSize); // for debug wfskrn
    WfsAreaAllocBlks(pAh, 0, 1, pBa, sizeof(WFSBlkAdr));
    ///////////////////////////////////////
#if _IOP
    IOS_FlushDCache(pBa, sizeof(WFSBlkAdr));
#endif*/
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    if (!pci.di.tba.pAreaInfo)
    {
        // no one can set permission to the entry which has no area info.
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    WFSKrnExit1OnError(DirGetBlk(&pci.di, TRANS_FLAG_RW_BLK, 0));
    WFSKrnExit1OnError(WFSKrnSetInheritedPermissionForUser(&pci, nFlags, nMask, pci.di.tba.pTransInfo->nTitleId, pci.di.tba.pTransInfo->nTitleId));
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        // Update acl index data in path cache tree
        pci.pEntry->pcd.nAccessListIdx = pci.di.pAttrHdr->nAccessListIdx;
    }
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("ChangePermission(\"%s\", nFlags, nMask) WFS_RESULT_%s\n", pAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
#else
    osTPrintf("WFSSrvChangePermissionForUser()\n");
    return WFS_RESULT_OK;
#endif
}


WFSResult WFSSrvCreateDirectory(const WFSPathName *pAbsPathName, u32 nFlags, u64 nQuota) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=144759){
        int a; a=0;
    }
#endif
    WFSKrnResult nResult=WFSKRN_RESULT_OK;
    PathCacheItr pci;
    WFSPathName path;
    if (nFlags & ~(WFS_FLAG_CONTIGUOUS_AREA | WFS_PERM_FULL))
    {
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit4;
    }
    WFSKrnExit4OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    pci.di.tba.pTransInfo = TransBegin();
    nResult = PathCacheFindParentDir(pAbsPathName, path.sStr, &pci);
    switch(nResult) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            break;
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1:
        case WFSKRN_RESULT_SRV_PATH_DEV:
        case WFSKRN_RESULT_SRV_PATH_VOL:
            nResult = PathCacheFindPath(pAbsPathName, path.sStr, &pci);
            switch(nResult)
            {
                case WFSKRN_RESULT_NOT_FOUND:
                    break;
                case WFSKRN_RESULT_OK:
                    nResult = WFSKRN_RESULT_ALREADY_EXISTS;
                default:
                    goto Exit3;
            }
            break;
        default:
            goto Exit3;
    }
#ifdef PERMISSION_ENABLED
    // Check the permission of parent dir and specified entry
    u32 permissions;
    WFSKrnExit3OnError(WFSKrnGetInheritedPermissionForUserPci(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if(!(permissions & WFS_PERM_ADD_OR_DELETE_SUBDIR))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit3;
    }
#else // PERMISSION_ENABLED
    if(pci.di.tba.pAreaInfo == NULL)
    {
        // system path.
        osTPrintf("WFSSrvCreateDirectory(): system path\n");
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit3;
    }
#endif // PERMISSION_ENABLED
    WFSBlkAdr nSubDirBlkAdr;
    // ToDo: Create sub area when WFS_FLAG_CONTIGUOUS_AREA is specified.
    WFSKrnExit3OnError(AreaAllocBlks(pci.di.tba.pAreaInfo, pci.di.tba.nBlkAdr, pci.di.tba.pAreaInfo->ah.nLog2BlkSize, 1, &nSubDirBlkAdr, sizeof(WFSBlkAdr)));
    WFSKrnExit2OnError(DirInsertRaw(&pci.di.name, &pci.di, 0));
    WFSKrnStoreCaseInformationStr(pci.di.pAttrHdr->aCaseBitArray, pci.pRawNameStart, pci.di.name.nLen);
    // Mark the file as open-for-writing
#ifdef PERMISSION_ENABLED
    // Get the correct Access List Idx
    pci.di.pAttrHdr->nAccessListIdx = WFS_FLAG_IS_A_DIRECTORY | (nFlags & WFS_FLAG_CONTIGUOUS_AREA); // the default is also the empty accesslist.
    pci.di.pAttrHdr->nObjCreator    = pci.di.tba.pTransInfo->nTitleId;
    WFSKrnExit1OnError(WFSKrnSetInheritedPermissionForUser(&pci, nFlags, WFS_PERM_FULL, pci.di.tba.pTransInfo->nTitleId, pci.di.tba.pTransInfo->nTitleId));
#else
    // ToDo: Add the calls to get the correct Access List Idx
    pci.di.pAttrHdr->nAccessListIdx = WFS_FLAG_IS_A_DIRECTORY | nFlags;
#endif
    pci.di.pAttrHdr->dir.nQuota = (WFSBlkAdr)(nQuota/pci.di.tba.pAreaInfo->nBlkSize);
    pci.di.pAttrHdr->dir.nSubDirBlkAdr = nSubDirBlkAdr;
    pci.di.pAttrHdr->timeCreated = ++wkg.time;
    pci.di.pAttrHdr->timeUpdated = wkg.time;
    TransUnpinBlk(&pci.di.tba);
    pci.di.tba.nBlkAdr = nSubDirBlkAdr;
    WFSKrnExit1OnError(DirCreateAndInitNewBlk(&pci.di.tba, (WFSMetaDataFlags)(WFS_MDF_DIR_LEAF_BLK | WFS_MDF_DIR_ROOT_BLK), &pci.di.pBlkHdr));
    DirBlkCreateRootNode(pci.di.pBlkHdr);
    goto Exit3;
Exit1:
    DirItrClose(&pci.di);
    if( PathCacheFindPath(pAbsPathName, path.sStr, &pci) != WFSKRN_RESULT_OK) {
        goto Exit2;
    }
    if( *pci.rxi.pStrPtr == 0) {
        PathCacheDeleteSubTree(pci.pEntry);
    }
    DirDeleteEntry(&pci.di);
Exit2:
    AreaFreeBlks(pci.di.tba.pAreaInfo, pci.di.tba.pAreaInfo->ah.nLog2BlkSize, 1, &nSubDirBlkAdr, sizeof(WFSBlkAdr));
Exit3:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit4:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput5("CreateDirectory(\"%s\", f=%x, q=%I64d) WFS_RESULT_%s\n", pAbsPathName->sStr, nFlags, nQuota, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


#define MAX_SEARCH_HANDLE 16

typedef struct{
    WFSSrvSearchDirectoryHandle sdh;
    WFSFileName                 filename;
}SearchBuf;

//static SearchBuf gSb[MAX_SEARCH_HANDLE];


void WFSKrnNormalizePattern(utf8 *sPattern) {
    utf8 *sTo = sPattern;
    while(*sPattern) {
        *sTo = *sPattern;
        if (*sPattern == '*') {
            // Convert substrings containing N '*' and M '?' characters only, to a string of M '?' characters followed by a single '*'.
            do {
                if (*sPattern == '?') {
                    *sTo++ = '?';
                }
                ++sPattern;
            } while ((*sPattern=='*')||(*sPattern=='?'));
            *sTo = '*';
        } else {
            ++sPattern;
        }
        ++sTo;
    }
    *sTo = 0;
}


bool WFSKrnNameMatchesPattern(utf8 *sName, utf8 *sPattern) {
    while ((*sPattern != '*') && (*sPattern != 0)) {
        if (*sPattern != *sName) {
            if ((*sPattern != '?') || (*sName == 0)) {
                return false;
            }
        }
        ++sPattern;
        ++sName;
    }
    if (*sPattern == 0) {
        return (*sName == 0);
    }
    // Skip the '*' char
    ++sPattern;
    // The corresponding sequence from sName can be any string of 0 or more characters
    do {
        while ((*sName != 0) && (*sName != *sPattern)) {
            ++sName;
        }
        if (*sName==0) {
            return (*sPattern == 0);
        }
        if (WFSKrnNameMatchesPattern(sName, sPattern)) {
            return true;
        }
        ++sName;
    } while (*sName != 0);
    return false;
}

void WfsKrnGetAttrPtr(void *pData, DirEntryAttrHdr **ppAttrHdr) {
    if (((s32)pData == WFSKRN_RESULT_SRV_PATH_DEV) ||
        ((s32)pData == WFSKRN_RESULT_SRV_PATH_VOL))
    {
        *ppAttrHdr = &wkg.devDirAttr;
    } else if (((DeviceInfo*)pData >= wkg.aDevInfo)&&((DeviceInfo*)pData < &wkg.aDevInfo[WFSKRN_MAX_DEVICES])) {
        *ppAttrHdr = &wkg.devAttr;
    } else if (((VolumeInfo*)pData >= wkg.aVolInfo)&&((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES])) {
        *ppAttrHdr = &((VolumeInfo*)pData)->rootAttr;
    }
}

WFSKrnResult WFSKrnSearchDirectoryCommon(WFSKrnResult nResult, WFSKrnSearchDirInfo *pSdi, WFSFileInfo *pFi, utf8 *sName, utf8 *sPattern) {
    while(1) {
        switch(nResult) {
            case WFSKRN_RESULT_OK:
            case WFSKRN_RESULT_DIR_ENTRY_FOUND:
            case WFSKRN_RESULT_DIR_NODE_STRING_PREFIX:
            case WFSKRN_RESULT_DIR_CHOICE_PREFIX:
                if (WFSKrnNameMatchesPattern(sName, sPattern)) {
                    if (pSdi->pci.di.tba.pAreaInfo) {
                        WFSKrnCopyFileNameAndRestoreCase(pFi->sFileName, &pSdi->pci.di.name, pSdi->pci.di.pAttrHdr->aCaseBitArray);
                    } else {
                        void *pData;
                        pSdi->pci.rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pSdi->pci.rxi.pNodeHdr));
                        pData = ((PathCacheEntry *)pSdi->pci.rxi.pNodeHdr)->pcd.pData;
                        WfsKrnGetAttrPtr(pData, &pSdi->pci.di.pAttrHdr);
                        pSdi->pci.di.name.nLen = pSdi->pci.rxi.pStrPtr - pSdi->pci.di.name.sStr;

                        memcpy(pFi->sFileName, pSdi->pci.di.name.sStr, pSdi->pci.di.name.nLen);
                        pFi->sFileName[pSdi->pci.di.name.nLen] = 0;
                    }
                    DirEntryAttrHdr *pAttrHdr = pSdi->pci.di.pAttrHdr;
                    pFi->attr.nCidx     = pAttrHdr->file.nContentIdx;
#ifdef PERMISSION_ENABLED
                    // Get inherited permissions
                    WFSKrnReturnOnError(WFSKrnGetInheritedPermissionForUserDi(&pSdi->pci, &pFi->attr.nFlags, pSdi->pci.di.tba.pTransInfo->nTitleId));
                    pFi->attr.nFlags   |= (pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) | (pAttrHdr->nAccessListIdx & WFS_FLAG_CONTIGUOUS_AREA);
#else
                    pFi->attr.nFlags    = pAttrHdr->nAccessListIdx;   // ToDo: Get permission bits for current user
#endif
                    if (pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
                        pFi->attr.nNumEntries = 0;  // ToDo: Remove NumEntries parameter from WFS API
                    } else {
                        pFi->attr.nFileSize = pAttrHdr->file.nSize;
                    }
                    pFi->attr.timeCreated = pAttrHdr->timeCreated;
                    pFi->attr.timeUpdated = pAttrHdr->timeUpdated;
#if _DEBUG_DIR
                    pFi->attr.timeCreated = pSdi->pci.di.tba.nBlkAdr;
#endif
                    return WFSKRN_RESULT_OK;
                }
                if (pSdi->pci.di.tba.pAreaInfo)
                {
                    nResult = DirNextRaw(&pSdi->pci.di.name, &pSdi->pci.di);
                }
                else
                {
                    nResult = RxtNext(wkg.pathCache.pRxtHdr, &pSdi->pci.rxi);
                }
                continue;
            case WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH:
            case WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND:
                return WFSKRN_RESULT_NOT_FOUND;
            default:
                return nResult;
        }
    }
}

WFSResult WFSSrvSearchDirectoryFirst(const WFSPathName *pAbsPattern, WFSSrvSearchDirectoryHandle *pSsdh, WFSFileInfo *pFi) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount >= 9390) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult;
    WFSPathName path;
    WFSKrnExit3OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPattern->sStr, pAbsPattern->nLen, WFS_MAX_PATH_NAME_SIZE));
    WFSKrnSearchDirInfo *pSdi;
    WFSKrnExit3OnError(WFSKrnSearchDirInfoAlloc(&pSdi));
    pSdi->pci.di.tba.pTransInfo = TransBegin();
    *pSsdh = pSdi->nHandle;
    nResult = PathCacheFindParentDir(pAbsPattern, path.sStr, &pSdi->pci);
    switch(nResult) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1:
        case WFSKRN_RESULT_SRV_PATH_DEV:
        case WFSKRN_RESULT_SRV_PATH_VOL:
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            break;
        case WFSKRN_RESULT_NOT_DIRECTORY:
            nResult = WFSKRN_RESULT_NOT_FOUND;
        default:
            goto Exit2;
    }
#ifdef PERMISSION_ENABLED
    // Check the permission of parent dir
    u32 permissions;
    WFSKrnExit2OnError(WFSKrnGetInheritedPermissionForUserPci(&pSdi->pci, &permissions, pSdi->pci.di.tba.pTransInfo->nTitleId));
    if(!(permissions & WFS_PERM_LIST))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit2;
    }
#endif
    //pPci->pEntry = PathCacheIncrementalInsertAfterCursorEntry(wkg.pathCache.pPinAnchor, &pPci->rxi, nPrefixLen, nSuffixLen);
#if _DEBUG
    memcpy(pSdi->pathDebug.sStr, path.sStr, pAbsPattern->nLen); pSdi->pathDebug.sStr[pAbsPattern->nLen]=0;
#endif
#if _DEBUG_PATH_CACHE
    PathCachePinEntry(pSdi->pci.pEntry, path.sStr);
    PathCacheIncreaseEntryOpenHandleCount(&pSdi->pci, WFS_ACCESS_READ, path.sStr);
#else
    PathCachePinEntry(pSdi->pci.pEntry);
    PathCacheIncreaseEntryOpenHandleCount(&pSdi->pci, WFS_ACCESS_READ);
#endif
    WFSKrnCopyStrAndConvertToLowerCase(pSdi->pattern.sStr, pSdi->pci.pRawNameStart, pSdi->pci.di.name.nLen);
    pSdi->pattern.nLen = pSdi->pci.di.name.nLen;
    WFSKrnNormalizePattern(pSdi->pattern.sStr);
    utf8 *pDstPtr=pSdi->pci.di.name.sStr;
    utf8 *pSrcPtr=pSdi->pattern.sStr;
    utf8 *pSrcEnd=&pSrcPtr[pSdi->pattern.nLen];
    for(; pSrcPtr<pSrcEnd; ++pSrcPtr, ++pDstPtr) {
        if ((*pSrcPtr == '*') || (*pSrcPtr == '?') || (*pSrcPtr == 0)) {
            break;
        }
        *pDstPtr = *pSrcPtr;
    }
    *pDstPtr = 0;
    if (pSdi->pci.di.tba.pAreaInfo == 0) {
        void *pData;
        s32 nStrLen = pSdi->pci.pNameEnd - pSdi->pci.rxi.pStrPtr;
        if (nStrLen == 0) {
            // special code for root '/' dir.
            pSdi->pci.di.pAttrHdr = &wkg.rootDirAttr;
            nResult = WFSKRN_RESULT_OK;
        } else {
            PathCacheIncrementalFindEntry(&pSdi->pci.rxi, 1, &pData);
            pSdi->pci.rxi.pStrPtr = pSdi->pci.di.name.sStr;
            nResult = RxtNextUntilSeparator(wkg.pathCache.pRxtHdr, &pSdi->pci.rxi, '/');
            if (nResult != WFSKRN_RESULT_OK) {
                goto Exit1;
            }
            pDstPtr = pSdi->pci.di.name.sStr;
            pSrcPtr = pSdi->pattern.sStr;
        }
    } else {
        pSdi->pci.di.nBlkDepth = 0;
#if _DEBUG_BREAK_POINT
        if (pSdi->pci.di.tba.nBlkAdr != pSdi->pci.nRootBlkAdr) {
            int a;a=0;
        }
#endif
        //pSdi->nRootBlkAdr = pSdi->pci.di.tba.nBlkAdr;
        nResult = DirFind(&pSdi->pci.di.name, &pSdi->pci.di, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_PREFIX_SEARCH);
    }
    nResult = WFSKrnSearchDirectoryCommon(nResult, pSdi, pFi, pDstPtr, pSrcPtr);
    if (nResult == WFSKRN_RESULT_OK) {
        if (pSdi->pci.di.tba.pAreaInfo) {
            TransUnpinBlk(&pSdi->pci.di.tba);
        }
        TransEnd(pSdi->pci.di.tba.pTransInfo);
    } else {
        if (pSdi->pci.di.tba.pAreaInfo)
        {
            DirItrClose(&pSdi->pci.di);
        }
Exit1:
    #if _DEBUG_PATH_CACHE
        PathCacheDecreaseEntryOpenHandleCount(&pSdi->pci, path.sStr);
    #else
        PathCacheDecreaseEntryOpenHandleCount(&pSdi->pci);
    #endif
Exit2:
        TransEnd(pSdi->pci.di.tba.pTransInfo);
        WFSKrnSearchDirInfoFree(pSdi);
    }
Exit3:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput2("SearchDirectoryFirst(\"%s\", pSsdh, pFi) ", pAbsPattern->sStr);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput3("sdh=%04x fn=\"%s\" ", *pSsdh, pFi->sFileName);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    return (WFSResult)nResult;
}

WFSResult WFSSrvSearchDirectoryNext(WFSSrvSearchDirectoryHandle ssdh, WFSFileInfo *pFi) {
#if _DEBUG_BREAK_POINT
    static int c=0; ++c; if (c>=0x7d68) {
        int a; a=0;
    }
#endif
    WFSKrnResult nResult;
    WFSKrnSearchDirInfo *pSdi;
    WFSKrnExitOnError(WFSKrnGetSearchDirInfo(&pSdi, ssdh));
    //nResult = DirNextRaw(&pSdi->pci.di.name, &pSdi->pci.di);   // This is unsafe since the directory may have changed
    // DirFindNext is slower than DirNextRaw becuase it starts from the root block of the directory every time.
    // ToDo: Add code to determine when it is necessary to use DirFindNext
    pSdi->pci.di.tba.pTransInfo = TransBegin();
    DirItrCloseNoUnpin(&pSdi->pci.di);
    pSdi->pci.di.tba.nBlkAdr = pSdi->pci.nRootBlkAdr;
    // ToDo: Add code for special dirs they have no DirItr.
    if (pSdi->pci.di.tba.pAreaInfo == 0) {
        s32 nStrLen = pSdi->pci.pNameEnd - pSdi->pci.rxi.pStrPtr;
        if (nStrLen == 0) {
            // special code for root '/' dir.
            nResult = WFSKRN_RESULT_NOT_FOUND;
            TransEnd(pSdi->pci.di.tba.pTransInfo);
            goto Exit;
        } else {
            // To make pSdi->pci.rxi.pNodeHdr to point RxtAttrHdr.
            pSdi->pci.rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pSdi->pci.rxi.pNodeHdr->nParentOfs);
            pSdi->pci.rxi.pStrPtr = pSdi->pci.di.name.sStr;
            nResult = RxtFindNext(wkg.pathCache.pRxtHdr, &pSdi->pci.rxi, '/');
            if (nResult != WFSKRN_RESULT_OK) {
                TransEnd(pSdi->pci.di.tba.pTransInfo);
                goto Exit;
            }
        }
    } else {
        nResult = DirFindNext(&pSdi->pci.di.name, &pSdi->pci.di);
    }
    nResult = WFSKrnSearchDirectoryCommon(nResult, pSdi, pFi, pSdi->pci.di.name.sStr, pSdi->pattern.sStr);
    if (pSdi->pci.di.tba.pAreaInfo) {
        TransUnpinBlk(&pSdi->pci.di.tba);
    }
    TransEnd(pSdi->pci.di.tba.pTransInfo);
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput2("SearchDirectoryNext(sdh=%04x, pFi) ", ssdh);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("fn=\"%s\" ", pFi->sFileName);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvSearchDirectoryClose(WFSSrvSearchDirectoryHandle ssdh) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 34840) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSKrnSearchDirInfo *pSdi;
    WFSKrnExitOnError(WFSKrnGetSearchDirInfo(&pSdi, ssdh));
    pSdi->pci.di.tba.pTransInfo = TransBegin();
    DirItrCloseNoUnpin(&pSdi->pci.di);
#if _DEBUG_PATH_CACHE
    PathCacheDecreaseEntryOpenHandleCount(&pSdi->pci, pSdi->pathDebug.sStr);
#else
    PathCacheDecreaseEntryOpenHandleCount(&pSdi->pci);
#endif
    TransEnd(pSdi->pci.di.tba.pTransInfo);
    WFSKrnSearchDirInfoFree(pSdi);
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("SearchDirectoryClose(sdh=%04x, pFi) WFS_RESULT_%s\n", ssdh, WFSKrnGetErrorString(nResult)));
#endif
    return WFS_RESULT_OK;
}


WFSResult WFSSrvValidateDirectory(const WFSPathName *pAbsPathName) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 15) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    if ((pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)==0) {
        nResult = WFSKRN_RESULT_NOT_DIRECTORY;
        goto Exit1;
    }
#ifdef PERMISSION_ENABLED
    u32 permissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((permissions & WFS_PERM_LIST) == 0)
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
#endif
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("ValidateDirectory(\"%s\") WFS_RESULT_%s\n", pAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvDelete(const WFSPathName *pAbsPathName) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 940958) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult;
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
#ifdef PERMISSION_ENABLED
    u32 permissions;
    u32 parentPermissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((permissions & WFS_PERM_DELETE_OR_MOVE) == 0)
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    WFSKrnExit1OnError(WFSKrnGetParentInheritedPermissionForUserPci(&pci, &parentPermissions, pci.di.tba.pTransInfo->nTitleId));
    if(pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)
    {
        if((parentPermissions & WFS_PERM_ADD_OR_DELETE_SUBDIR) == 0)
        {
            nResult = WFSKRN_RESULT_PERMISSION;
            goto Exit1;
        }
    }
    else
    {
        if((parentPermissions & WFS_PERM_ADD_OR_DELETE_FILE) == 0)
        {
            nResult = WFSKRN_RESULT_PERMISSION;
            goto Exit1;
        }
    }
#else // PERMISSION_ENABLED
    if(pci.di.tba.pAreaInfo == NULL)
    {
        // system path.
        osTPrintf("WFSSrvDelete(): system path\n");
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
#endif // PERMISSION_ENABLED

    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        if (pci.pEntry->pcd.nOpenHandleCount) {
            nResult = WFSKRN_RESULT_FILE_OPEN;
            goto Exit1;
        }
    }
#ifdef PERMISSION_ENABLED
    AreaInfo*       area   = pci.di.tba.pAreaInfo;
    WFSKrnAclHandle handle = pci.di.pAttrHdr->nAccessListIdx;
#endif
    if (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
        WFSKrnExit1OnError(DirDeleteSubDir(&pci.di));
    } else {
        WFSKrnExit1OnError(DirDeleteFile(&pci.di));
    }
#ifdef PERMISSION_ENABLED
    WFSKrnExit1OnError(WFSKrnReleaseAccessList(area, handle));
#endif
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        WFSKrnExit1OnError(PathCacheDeleteSubTree(pci.pEntry));
    }
    nResult = DirDeleteEntry(&pci.di);
Exit1:
    DirItrClose(&pci.di);
Exit2:
    TransEnd(pci.di.tba.pTransInfo);
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("Delete(\"%s\") WFS_RESULT_%s\n", pAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvMoveLocal(const WFSPathName *pSrcAbsPathName, const WFSPathName *pDstAbsPathName) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 0x75c8) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult;
    WFSPathName path;
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();

    WFSKrnExit5OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pSrcAbsPathName->sStr, pSrcAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    WFSKrnExit4OnError(PathCacheFindPath(pSrcAbsPathName, path.sStr, &pci));
#ifdef PERMISSION_ENABLED
    u32 permissions;
    u32 parentPermissions;
    WFSKrnExit4OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((permissions & WFS_PERM_DELETE_OR_MOVE) == 0)
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit4;
    }
    WFSKrnExit4OnError(WFSKrnGetParentInheritedPermissionForUserPci(&pci, &parentPermissions, pci.di.tba.pTransInfo->nTitleId));
    if(!(parentPermissions & ((pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) ? WFS_PERM_ADD_OR_DELETE_SUBDIR : WFS_PERM_ADD_OR_DELETE_FILE)))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit4;
    }
#else // PERMISSION_ENABLED
    if(pci.di.tba.pAreaInfo == NULL)
    {
        // system path.
        osTPrintf("WFSSrvMoveLocal(): system path\n");
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit4;
    }
#endif
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        if (pci.pEntry->pcd.nOpenHandleCount) {
            nResult = WFSKRN_RESULT_FILE_OPEN;
            goto Exit4;
        }
    }
    else
    {
        // When the entry is directory and it has not cached yet, it is need to cache
        // to calculate the sum of children's OpenHandleCount.
        if (pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)
        {
            pci.pEntry = PathCacheIncrementalInsertAfterCursorEntry(wkg.pathCache.pMruAnchor, &pci.rxi, pci.rxi.pStrPtr - path.sStr, pci.pNameEnd - pci.rxi.pStrPtr);
            if (pci.pEntry->pcd.nOpenHandleCount) {
                nResult = WFSKRN_RESULT_FILE_OPEN;
                goto Exit4;
            }
        }
        // The file entry doesn't have children. No need to calculate.
    }

    WFSBlkAdr  nBlkAdrTmp;
    void      *pBlkPtrTmp;
    AreaInfo *pAreaInfo = pci.di.tba.pAreaInfo;
    WFSKrnExit4OnError(AreaAllocBlks(pAreaInfo, pci.di.tba.nBlkAdr, pAreaInfo->ah.nLog2BlkSize, 1, &nBlkAdrTmp, sizeof(WFSBlkAdr)));
    WFSKrnExit3OnError(TransGetAndPinBlk5(pAreaInfo, nBlkAdrTmp, pci.di.tba.pTransInfo, 0, &pBlkPtrTmp));

    memcpy(pBlkPtrTmp, pci.di.pAttrHdr, (1<<pci.di.pAttrHdr->nAttrLog2Size));
    DirEntryAttrHdr *pSrcAttrHdr = (DirEntryAttrHdr*)pBlkPtrTmp;

    // Todo : Fixed bug for PathCacheRecursiveDelete() in cmp2_test_case3.
    if (*pci.rxi.pStrPtr==0){
        WFSKrnExit3OnError(PathCacheDeleteSubTree(pci.pEntry));
    }
    WFSKrnExit3OnError(DirDeleteEntry(&pci.di));

    DirItrClose(&pci.di);

    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pDstAbsPathName->sStr, pDstAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    nResult = PathCacheFindParentDir(pDstAbsPathName, path.sStr, &pci);
    switch(nResult) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            break;
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1:
        case WFSKRN_RESULT_SRV_PATH_DEV:
        case WFSKRN_RESULT_SRV_PATH_VOL:
            nResult = WFSKRN_RESULT_INVALID;
        default:
            goto Exit1;
    }
#ifdef PERMISSION_ENABLED
    // Check the permission of parent dir and specified entry
    WFSKrnExit2OnError(WFSKrnGetInheritedPermissionForUserPci(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if(!(permissions & ((pSrcAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) ? WFS_PERM_ADD_OR_DELETE_SUBDIR : WFS_PERM_ADD_OR_DELETE_FILE)))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
#endif
    if (pSrcAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY){
        // Process for moving directory .
        u8 nSizeCategory;
        u32 nAllocSize   = 0;
        u8 nAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(pAreaInfo, &nAllocSize, pci.di.name.nLen, &nSizeCategory);
#if _DEBUG
        if(nSizeCategory != FILE_SIZE_CATEGORY_VERY_SMALL){
            WFSKrnExit1OnError(WFSKRN_RESULT_FATAL_ERROR);
        }
#endif
        WFSKrnExit1OnError(DirSbaAllocAndRegistSubBlk(&pci.di.name, &pci.di, nAttrLog2Size, &pci.di.pAttrHdr));
        memcpy((u8*)pci.di.pAttrHdr , (u8*)pSrcAttrHdr, sizeof(DirEntryAttrHdr));
        pci.di.pAttrHdr->nAttrLog2Size = nAttrLog2Size;
        pci.di.pAttrHdr->nAllocSize    = nAllocSize;
        pci.di.pAttrHdr->nNameLen      = pci.di.name.nLen;
        WFSKrnStoreCaseInformationStr(pci.di.pAttrHdr->aCaseBitArray, pci.pRawNameStart, pci.di.name.nLen);
    } else {
        // Process for moving file.
        //
        // When the max sub block size of FILE_SIZE_CATEGORY_VERY_SMALL can not be record as new file name,
        // the src file resize by DirResizeFile and size category change to FILE_SIZE_CATEGORY_SMALL.
        if (pSrcAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_SMALL && pSrcAttrHdr->nAttrLog2Size == SBA_MAX_LOG2_SUB_BLK_SIZE) {
            u32 nNewAllocSize = (1<<SBA_MAX_LOG2_SUB_BLK_SIZE) - ((u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((pci.di.name.nLen+7)>>3));
            if (nNewAllocSize < pSrcAttrHdr->file.nSize){
                DirEntryAttrHdr *pNewSrcAttrHdr = (DirEntryAttrHdr*)((size_t)pBlkPtrTmp + (1<<SBA_MAX_LOG2_SUB_BLK_SIZE));
                memcpy((u8*)pNewSrcAttrHdr, (u8*)pSrcAttrHdr, ((u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((pSrcAttrHdr->nNameLen+7)>>3)));
                pNewSrcAttrHdr->nAllocSize = pAreaInfo->nBlkSize;
                pNewSrcAttrHdr->nAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2Size(pAreaInfo, pNewSrcAttrHdr->nAllocSize, pNewSrcAttrHdr->nNameLen, &pNewSrcAttrHdr->nSizeCategory);
                WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewSrcAttrHdr, pNewSrcAttrHdr->nAttrLog2Size, 0);
                memset((void*)pFileBlkPtr, 0, sizeof(WFSFileBlkPtr));
                WFSKrnExit1OnError(AreaAllocBlks(pAreaInfo, pci.di.tba.nBlkAdr, pAreaInfo->ah.nLog2BlkSize, 1, (WFSBlkAdr*)pFileBlkPtr, -(s32)sizeof(WFSFileBlkPtr)));
                void *pBlkPtr;
                u32 nValidSize = WFS_ROUND_UP(pSrcAttrHdr->file.nSize, pAreaInfo->pVolInfo->pDevInfo->dh.nLog2SectorSize);
                u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pBlkPtrTmp;
                nResult = TransGetAndPinFileBlk(&pci.di.tba, nBlkAdrTmp, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, &pBlkPtr);
                if(nResult < WFSKRN_RESULT_OK){
                    TransUnpinBlk3(pAreaInfo, pFileBlkPtr->nBlkAdr, pci.di.tba.pTransInfo);
                    AreaFreeBlks(pAreaInfo, pAreaInfo->ah.nLog2BlkSize, 1, (WFSBlkAdr*)pFileBlkPtr, -(s32)sizeof(WFSFileBlkPtr));
                    goto Exit1;
                }
                memcpy(pBlkPtr, &pSrcAttrHdr->aCaseBitArray[(pSrcAttrHdr->nNameLen+7)>>3], pSrcAttrHdr->file.nSize);
                WFSKrnExit1OnError(TransStoreBlk3(pAreaInfo, pFileBlkPtr->nBlkAdr, pci.di.tba.pTransInfo));
                WFSKrnExit1OnError(TransUnpinBlk3(pAreaInfo, pFileBlkPtr->nBlkAdr, pci.di.tba.pTransInfo));
                pSrcAttrHdr = pNewSrcAttrHdr;
            } else if (nNewAllocSize < pSrcAttrHdr->nAllocSize){
                pSrcAttrHdr->nAllocSize = nNewAllocSize;
            }
        }
        u8 nSizeCategory;
        u8 nAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2Size(pAreaInfo, pSrcAttrHdr->nAllocSize, pci.di.name.nLen, &nSizeCategory);
        if(nSizeCategory != pSrcAttrHdr->nSizeCategory){
            nAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2SizeDecrease(pAreaInfo, pSrcAttrHdr->nAllocSize, pci.di.name.nLen, &nSizeCategory);
#if _DEBUG
            if(nSizeCategory != pSrcAttrHdr->nSizeCategory){
                WFSKrnExit1OnError(WFSKRN_RESULT_FATAL_ERROR);
            }
#endif
        }
        // Alloc the new sub block and copy from src attributes to the new sub block.
        WFSKrnExit1OnError(DirSbaAllocAndRegistSubBlk(&pci.di.name, &pci.di, nAttrLog2Size, &pci.di.pAttrHdr));
        memcpy((u8*)pci.di.pAttrHdr , (u8*)pSrcAttrHdr, sizeof(DirEntryAttrHdr));
        switch(nSizeCategory)
        {
            case FILE_SIZE_CATEGORY_VERY_SMALL:{
                u32 nCopySize = pSrcAttrHdr->file.nSize;
                memcpy(&pci.di.pAttrHdr->aCaseBitArray[(pci.di.name.nLen+7)>>3], &pSrcAttrHdr->aCaseBitArray[(pSrcAttrHdr->nNameLen+7)>>3], nCopySize);
                break;
            }
            case FILE_SIZE_CATEGORY_SMALL:{
                u32 nCopySize = (pSrcAttrHdr->nAllocSize>>pAreaInfo->ah.nLog2BlkSize)*sizeof(WFSFileBlkPtr);
                memcpy((u8*)pci.di.pAttrHdr+(1<<nAttrLog2Size)-nCopySize, (u8*)pSrcAttrHdr+(1<<pSrcAttrHdr->nAttrLog2Size)-nCopySize, nCopySize);
                break;
            }
            case FILE_SIZE_CATEGORY_MEDIUM:{
                u32 nCopySize = (pSrcAttrHdr->nAllocSize>>pAreaInfo->ah.nLog2MediumBlkSize)*sizeof(WFSFileBlkPtr);
                memcpy((u8*)pci.di.pAttrHdr+(1<<nAttrLog2Size)-nCopySize, (u8*)pSrcAttrHdr+(1<<pSrcAttrHdr->nAttrLog2Size)-nCopySize, nCopySize);
                break;
            }
            case FILE_SIZE_CATEGORY_LARGE:{
                u32 nCopySize = (pSrcAttrHdr->nAllocSize>>pAreaInfo->ah.nLog2LargeBlkSize)*sizeof(WFSFileLargeBlkPtr);
                memcpy((u8*)pci.di.pAttrHdr+(1<<nAttrLog2Size)-nCopySize, (u8*)pSrcAttrHdr+(1<<pSrcAttrHdr->nAttrLog2Size)-nCopySize, nCopySize);
                break;
            }
            case FILE_SIZE_CATEGORY_VERY_LARGE:{
                u32 nCopySize = (((pSrcAttrHdr->nAllocSize>>pAreaInfo->ah.nLog2LargeBlkSize)-1)/pAreaInfo->nNumLargeBlkPtrsPerBlk+1)*sizeof(WFSBlkAdr);
                memcpy((u8*)pci.di.pAttrHdr+(1<<nAttrLog2Size)-nCopySize, (u8*)pSrcAttrHdr+(1<<pSrcAttrHdr->nAttrLog2Size)-nCopySize, nCopySize);
                break;
            }
        }
        pci.di.pAttrHdr->nAttrLog2Size = nAttrLog2Size;
        pci.di.pAttrHdr->nNameLen      = pci.di.name.nLen;
        WFSKrnStoreCaseInformationStr(pci.di.pAttrHdr->aCaseBitArray, pci.pRawNameStart, pci.di.name.nLen);
        WFSKrnExit3OnError(DirReGetHashAdr(pci.di.tba.pAreaInfo, pci.di.pAttrHdr, pci.di.tba.nBlkAdr, pci.di.pBlkHdr));
    }
    goto Exit3;

Exit1:;
    DirItrClose(&pci.di);
Exit2:;
    WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pSrcAbsPathName->sStr, pSrcAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE);
    switch(PathCacheFindParentDir(pSrcAbsPathName, path.sStr, &pci)) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            break;
        default:
            goto Exit3;
    }
    DirSbaAllocAndRegistSubBlk(&pci.di.name, &pci.di, pSrcAttrHdr->nAttrLog2Size, &pci.di.pAttrHdr);
    memcpy(pci.di.pAttrHdr, pSrcAttrHdr, (1<<pSrcAttrHdr->nAttrLog2Size));
Exit3:;
    TransUnpinBlk3(pAreaInfo, nBlkAdrTmp, pci.di.tba.pTransInfo);
    AreaFreeBlks(pAreaInfo, pAreaInfo->ah.nLog2BlkSize, 1, &nBlkAdrTmp, sizeof(WFSBlkAdr));
Exit4:;
    DirItrClose(&pci.di);
Exit5:;
    TransEnd(pci.di.tba.pTransInfo);
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput4("MoveLocal(\"%s\", \"%s\") WFS_RESULT_%s\n", pSrcAbsPathName->sStr, pDstAbsPathName->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvCreateAndOpenFile(const WFSPathName *pAbsPathName, u32 nFlags, WFSContentIdx nCidx, WFSFileSize nPreAllocateSize, WFSSrvFileHandle *pSfh) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 15529) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult;
    WFSPathName path;
    if (nFlags & ~(WFS_PERM_FULL))
    {
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit3;
    }
    WFSKrnExit3OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    WFSKrnFileInfo *pKfi;
    WFSKrnExit3OnError(WFSKrnFileInfoAlloc(&pKfi));
    pKfi->pci.di.tba.pTransInfo = TransBegin();
    nResult = PathCacheFindParentDir(pAbsPathName, path.sStr, &pKfi->pci);
    switch(nResult) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            break;
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1:
        case WFSKRN_RESULT_SRV_PATH_DEV:
        case WFSKRN_RESULT_SRV_PATH_VOL:
            nResult = PathCacheFindPath(pAbsPathName, path.sStr, &pKfi->pci);
            switch(nResult)
            {
                case WFSKRN_RESULT_NOT_FOUND:
                    break;
                case WFSKRN_RESULT_OK:
                    nResult = WFSKRN_RESULT_ALREADY_EXISTS;
                default:
                    goto Exit2;
            }
            break;
         default:
             goto Exit2;
    }
#ifdef PERMISSION_ENABLED
    // Check the permission of parent dir and specified entry
    u32 permissions;
    WFSKrnExit2OnError(WFSKrnGetInheritedPermissionForUserPci(&pKfi->pci, &permissions, pKfi->pci.di.tba.pTransInfo->nTitleId));
    if(!(permissions & WFS_PERM_ADD_OR_DELETE_FILE))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit2;
    }
#endif
    WFSKrnExit1OnError(DirInsertRaw(&pKfi->pci.di.name, &pKfi->pci.di, nPreAllocateSize));
    nResult = PathCacheIncrementalInsertPinnedEntry(&pKfi->pci, pAbsPathName->nLen);
    if (nResult == WFSKRN_RESULT_OK)
    {
#if _DEBUG
        memcpy(pKfi->pathDebug.sStr, path.sStr, pAbsPathName->nLen); pKfi->pathDebug.sStr[pAbsPathName->nLen]=0;
#endif
#if _DEBUG_PATH_CACHE
        nResult = PathCacheIncreaseEntryOpenHandleCount(&pKfi->pci, WFS_ACCESS_RW, path.sStr);
#else
        nResult = PathCacheIncreaseEntryOpenHandleCount(&pKfi->pci, WFS_ACCESS_RW);
#endif
        WFSKrnStoreCaseInformationStr(pKfi->pci.di.pAttrHdr->aCaseBitArray, pKfi->pci.pRawNameStart, pKfi->pci.di.name.nLen);
        // Mark the file as open-for-writing
#ifdef PERMISSION_ENABLED
        // Get the correct Access List Idx
        pKfi->pci.di.pAttrHdr->nAccessListIdx = 0; // the default is also the empty accesslist.
        pKfi->pci.di.pAttrHdr->nObjCreator    = pKfi->pci.di.tba.pTransInfo->nTitleId;
        WFSKrnExit1OnError(WFSKrnSetInheritedPermissionForUser(&pKfi->pci, nFlags, WFS_PERM_FULL, pKfi->pci.di.tba.pTransInfo->nTitleId, pKfi->pci.di.tba.pTransInfo->nTitleId));
#else
        // ToDo: Add the calls to get the correct Access List Idx
        pKfi->pci.di.pAttrHdr->nAccessListIdx = 1;// | WFSKRN_FILE_OPEN_FOR_WRITE_FLAG;
#endif
        pKfi->pci.di.pAttrHdr->file.nSize = 0;
        pKfi->pci.di.pAttrHdr->file.nContentIdx = nCidx;
        pKfi->pci.di.pAttrHdr->file.nVersion = 0;
        pKfi->pci.di.pAttrHdr->timeCreated = ++wkg.time;
        pKfi->pci.di.pAttrHdr->timeUpdated = wkg.time;
        pKfi->attr = *pKfi->pci.di.pAttrHdr;
        pKfi->nAccess = WFS_ACCESS_RW | WFS_ACCESS_CHANGE_SIZE; // Allowed read, write and change size access.
        *pSfh = pKfi->nHandle;

        nResult = DirItrMapBlks(&pKfi->pci.di);
        if (nResult == WFSKRN_RESULT_OK)
        {
            TransUnpinBlk(&pKfi->pci.di.tba);
            TransEnd(pKfi->pci.di.tba.pTransInfo);
            goto Exit3;
        }

        // ToDo: Need to decrease entry open handle when failed to set permission for user.
#if _DEBUG_PATH_CACHE
        PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci, pKfi->pathDebug.sStr);
#else
        PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci);
#endif
    }

Exit1:
    DirItrClose(&pKfi->pci.di);
Exit2:
    TransEnd(pKfi->pci.di.tba.pTransInfo);
    WFSKrnFileInfoFree(pKfi);
Exit3:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput5("CreateAndOpenFile(\"%s\", f=%x, c=%x, s=%u, ) ", pAbsPathName->sStr, nFlags, nCidx, nPreAllocateSize);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("fh=%04x ", *pSfh);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    if(nResult < WFSKRN_RESULT_OK){
        int a; a=0;
    }
    return (WFSResult)nResult;
}


WFSResult WFSSrvOpenFile(const WFSPathName *pAbsPathName, WFSAccess nDesiredAccess, WFSSrvFileHandle *pSfh, WFSFileAttributes *pFa) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount==2047) {
        int a;a=0;
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    if ((nDesiredAccess<WFS_ACCESS_READ) || (nDesiredAccess>WFS_ACCESS_RW)) {
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit2;
    }
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    WFSKrnFileInfo *pKfi;
    WFSKrnExit2OnError(WFSKrnFileInfoAlloc(&pKfi));
    pKfi->pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pKfi->pci));
    if (pKfi->pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
        nResult = WFSKRN_RESULT_NOT_FILE;
        goto Exit1;
    }
#ifdef PERMISSION_ENABLED
    // Check permission
    u32 flags;
    switch(nDesiredAccess)
    {
        case WFS_ACCESS_READ:
            flags = WFS_PERM_READ;
            break;
        case WFS_ACCESS_WRITE:
            flags = WFS_PERM_WRITE;
            break;
        case WFS_ACCESS_RW:
            flags = WFS_PERM_READ | WFS_PERM_WRITE;
            break;
        default:
            nResult = WFSKRN_RESULT_INVALID;
            goto Exit1;
    }
    u32 permissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pKfi->pci, &permissions, pKfi->pci.di.tba.pTransInfo->nTitleId));
    if(((permissions & WFS_PERM_FILE_FULL) | flags) != (permissions & WFS_PERM_FILE_FULL))
    {
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
#else // PERMISSION_ENABLED
    if (pKfi->pci.di.tba.pAreaInfo == NULL)
    {
        // system path.
        osTPrintf("WFSSrvOpenFile(): system path\n");
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
#endif // PERMISSION_ENABLED
    nResult = PathCacheIncrementalInsertPinnedEntry(&pKfi->pci, pAbsPathName->nLen);
    if (nResult != WFSKRN_RESULT_OK)
    {
        goto Exit1;
    }

#if _DEBUG
    memcpy(pKfi->pathDebug.sStr, path.sStr, pAbsPathName->nLen); pKfi->pathDebug.sStr[pAbsPathName->nLen]=0;
#endif
#if _DEBUG_PATH_CACHE
    WFSKrnExit1OnError(PathCacheIncreaseEntryOpenHandleCount(&pKfi->pci, nDesiredAccess, path.sStr));
#else
    WFSKrnExit1OnError(PathCacheIncreaseEntryOpenHandleCount(&pKfi->pci, nDesiredAccess));
#endif
    DirEntryAttrHdr *pAttrHdr = pKfi->pci.di.pAttrHdr;
    pFa->nCidx     = pAttrHdr->file.nContentIdx;
#ifdef PERMISSION_ENABLED
    pFa->nFlags    = permissions & WFS_PERM_FILE_FULL;
#else
    pFa->nFlags    = pAttrHdr->nAccessListIdx;   // ToDo: Get permission bits for current user
#endif
    if (pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
        pFa->nNumEntries = 0;  // ToDo: Remove NumEntries parameter from WFS API
    } else {
        pFa->nFileSize = pAttrHdr->file.nSize;
    }
    pFa->timeCreated = pAttrHdr->timeCreated;
    pFa->timeCreated = pAttrHdr->timeUpdated;
    // Update attributes in pKfi->attr.
    pKfi->attr = *pKfi->pci.di.pAttrHdr;
    *pSfh = pKfi->nHandle;
#ifdef PERMISSION_ENABLED
    pKfi->nAccess = nDesiredAccess | ((permissions & WFS_PERM_CHANGE_SIZE) ? WFS_ACCESS_CHANGE_SIZE : 0);
#else
    pKfi->nAccess = nDesiredAccess;
#endif
    //pKfi->fpos.nPos = 0;
    nResult = DirItrMapBlks(&pKfi->pci.di);
    if (nResult == WFSKRN_RESULT_OK)
    {
        TransUnpinBlk(&pKfi->pci.di.tba);
        TransEnd(pKfi->pci.di.tba.pTransInfo);
        goto Exit2;
    }

    // decrease entry open handle when failed to set permission for user.
#if _DEBUG_PATH_CACHE
    PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci, pKfi->pathDebug.sStr);
#else
    PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci);
#endif

Exit1:;
    DirItrClose(&pKfi->pci.di);
    TransEnd(pKfi->pci.di.tba.pTransInfo);
    WFSKrnFileInfoFree(pKfi);
Exit2:;
#if _DEBUG
    #if ((DBG_OUTPUT_TO_LOG_FILE && _WIN32) || DBG_OUTPUT_TO_SCREEN)
        char *aDA[4]={"?","R","W","RW"};
        if(nDesiredAccess>3)
            nDesiredAccess = (WFSAccess)0;
    #endif
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput3("OpenFile(\"%s\", %s, pSfh, pFa) ", pAbsPathName->sStr, aDA[nDesiredAccess]);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("fh=%04x ", *pSfh);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
    );
#endif
    return (WFSResult)nResult;
}


static inline WFSKrnResult DirItrUpdateAndPinBlk(DirItr* di, WFSBlkAdr nRootBlkAdr, int transFlag)
{
    WFSKrnResult result = DirItrCheckMapBlks(di);
    if (result == WFSKRN_RESULT_OK)
    {
        result = DirItrPinBlkAndFixPtrs(di, transFlag);
        return result;
    }

    DirItrUnmapBlks(di);
    DirItrCloseNoUnpin(di);
    di->nBlkDepth = 0;
    di->tba.nBlkAdr = nRootBlkAdr;

    result = DirFind(&di->name, di, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
    if (result == WFSKRN_RESULT_DIR_ENTRY_FOUND)
    {
        if (transFlag == TRANS_FLAG_RW_BLK)
        {
            result = DirGetBlk(di, TRANS_FLAG_RW_BLK, 0);
            if (result != WFSKRN_RESULT_OK)
            {
                return result;
            }
        }
        result = DirItrMapBlks(di);
    }
    else
    {
        WFSKrnOutputErrorCode(result);
        ASSERT(result == WFSKRN_RESULT_DIR_ENTRY_FOUND);
    }
    return result;
}


WFSResult WFSSrvCloseFile(WFSSrvFileHandle sfh, WFSFileSize nNewSize) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount==15531) {
        int a;a=0;
    }
#endif
    WFSKrnFileInfo *pKfi;
    WFSKrnResult nResult;
    WFSKrnExit2OnError(WFSKrnGetFileInfo(&pKfi, sfh));
    pKfi->pci.di.tba.pTransInfo = TransBegin();

    nResult = DirItrUpdateAndPinBlk(&pKfi->pci.di, pKfi->pci.nRootBlkAdr, TRANS_FLAG_READ_BLK);
    if (nResult != WFSKRN_RESULT_OK)
    {
        goto Exit1;
    }
    if (nResult == WFSKRN_RESULT_OK)
    {
        DirEntryAttrHdr *pAttrHdr = pKfi->pci.di.pAttrHdr;
        if (nNewSize != pAttrHdr->file.nSize) {
            if ((pKfi->nAccess & WFS_ACCESS_WRITE) == 0) {
                nResult = WFSKRN_RESULT_ACCESS;
                goto Exit1;
            }
#ifdef PERMISSION_ENABLED
            if ((pKfi->nAccess & WFS_ACCESS_CHANGE_SIZE) == 0) {
                TransEnd(pKfi->pci.di.tba.pTransInfo);
                TransUnpinBlk(&pKfi->pci.di.tba);
                return WFS_RESULT_NO_CHANGE_SIZE;
            }
#endif
            WFSKrnExit1OnError( DirGetBlk(&pKfi->pci.di, TRANS_FLAG_RW_BLK, 0) );
            DirResizeFile(&pKfi->pci.di, nNewSize);
        }
        DirItrUnmapBlks(&pKfi->pci.di);
        DirItrClose(&pKfi->pci.di);
    }

    if (nResult != WFSKRN_RESULT_MEDIA_ERROR) // [check] workaround.
                                              // PathCacheDecreaseEntryOpenHandleCount() throws exception when the device is unmounted.
    {
#if _DEBUG_PATH_CACHE
        PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci, pKfi->pathDebug.sStr);
#else
        PathCacheDecreaseEntryOpenHandleCount(&pKfi->pci);
#endif
    }
    TransEnd(pKfi->pci.di.tba.pTransInfo);
    WFSKrnFileInfoFree(pKfi);
    goto Exit2;
Exit1:
    TransEnd(pKfi->pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput4("CloseFile(fh=%04x, ns=%u) WFS_RESULT_%s\n", sfh, nNewSize, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;

}


WFSResult WFSSrvGetFileSize(WFSSrvFileHandle sfh, WFSFileSize *pSize) {
    WFSKrnFileInfo *pKfi;
    WFSKrnResult nResult;
    WFSKrnExitOnError(WFSKrnGetFileInfo(&pKfi, sfh));
    pKfi->pci.di.tba.pTransInfo = TransBegin();
    *pSize = pKfi->attr.file.nSize;
    TransEnd(pKfi->pci.di.tba.pTransInfo);
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput4("GetFileSize(fh=%04x, s=%u) WFS_RESULT_%s\n", sfh, *pSize, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


//static u8 aTempBuffer[WFS_STATIC_MEM_ALIGNMENT]; //__attribute__((aligned(32)));

void RotateBuffer(u8 *pBuffer, u32 nSize, s32 nRotationOffset, bool check) {
#if _DEBUG
    if ((nRotationOffset > WFS_STATIC_MEM_ALIGNMENT) || (nRotationOffset < -WFS_STATIC_MEM_ALIGNMENT)) {
        WFSKrnOutputErrorStr("Rotation Invalid");
    }
#endif
    u8 aTempBuffer[WFS_STATIC_MEM_ALIGNMENT];
    if (nRotationOffset > 0) {
        memcpy(aTempBuffer, pBuffer, nRotationOffset);
        memmove(pBuffer, pBuffer+nRotationOffset, nSize-nRotationOffset);
        memcpy(pBuffer+nSize-nRotationOffset, aTempBuffer, nRotationOffset);
    } else {
        nRotationOffset = -nRotationOffset;
        memcpy(aTempBuffer, pBuffer+nSize-nRotationOffset, nRotationOffset);
        memmove(pBuffer+nRotationOffset, pBuffer, nSize-nRotationOffset);
        memcpy(pBuffer, aTempBuffer, nRotationOffset);
        //nRotationOffset = -nRotationOffset;
    }
    /*
    if(check){
        int i;
        for(i=1;i<nSize;i++){
            if(pBuffer[0] != pBuffer[i]){
                osTPrintf("Rotation Error i = %d, buf = 0x%x, size = %d, rotation = %d\n", i, pBuffer, nSize, nRotationOffset);
                return;
            }
        }
        osTPrintf("Rotation OK : buf = 0x%x, size = %d, rotation = %d\n", pBuffer, nSize, nRotationOffset);
    }
    */
}


static inline WFSKrnResult WfsSrvTransReadFileBlksForCategorySmall(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo){
    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size, nBlkIdx);
    return TransReadFileBlks(pBuffer, pAreaInfo, nSize, pFileBlkPtr, -(signed)sizeof(*pFileBlkPtr), pTransInfo);
}

static inline WFSKrnResult WfsSrvTransReadFileBlksForCategoryMedium(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo){
    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size, nBlkIdx);
    return TransReadFileMediumBlks(pBuffer, pAreaInfo, nSize, pFileBlkPtr, -(signed)sizeof(*pFileBlkPtr), pTransInfo);
}

static inline WFSKrnResult WfsSrvTransReadFileBlksForCategoryLarge(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo){
    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size,
                                                                                    nBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
    return TransReadFileLargeBlks(pBuffer, pAreaInfo, nSize, nBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                  pFileLargeBlkPtr, -(signed)sizeof(*pFileLargeBlkPtr), pTransInfo);
}

static inline WFSKrnResult WfsSrvTransReadFileBlksForCategoryVeryLarge(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo){
    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge((u8*)pMetaData, nBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
    return TransReadFileLargeBlks(pBuffer, pAreaInfo, nSize, nBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                  pFileLargeBlkPtr, (s32)sizeof(*pFileLargeBlkPtr), pTransInfo);
}

s32 WFSSrvReadFile(WFSSrvFileHandle sfh, void *pBuffer, WFSFileSize nPos, s32 nSize) {
#if _DEBUG_BREAK_POINT
    if (nDbgCommandCount == 0x12e) {
        int a;a=0;
    }
  #if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
  #endif
#endif
#if _DEBUG
    #if ((DBG_OUTPUT_TO_LOG_FILE && _WIN32) || DBG_OUTPUT_TO_SCREEN)
        WFSFileSize nInPos = nPos;
        s32 nInSize = nSize;
    #endif
#endif
    //if((size_t)pBuffer%WFSKRN_FILE_BUFFER_MEM_ALIGNMENT_SIZE){
    //    return WFS_RESULT_INVALID;
    //}
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    s32 nReadSize=0;
    WFSKrnFileInfo *pKfi;
    WFSKrnExit3OnError(WFSKrnGetFileInfo(&pKfi, sfh));
    TransInfo *pTransInfo = pKfi->pci.di.tba.pTransInfo = TransBegin();
    if ((pKfi->nAccess & WFS_ACCESS_READ) == 0) {
        nResult = (WFSKrnResult)WFS_RESULT_ACCESS;
        goto Exit2;
    }
    nResult = DirItrUpdateAndPinBlk(&pKfi->pci.di, pKfi->pci.nRootBlkAdr, TRANS_FLAG_READ_BLK);
    if (nResult != WFSKRN_RESULT_OK)
    {
        goto Exit1;
    }

    //if(!memcmp(pKfi->pci.di.name.sStr,"araucarian",pKfi->pci.di.name.nLen)){
    //    int a; a=0;
    //}

    DirEntryAttrHdr *pAttrHdr = pKfi->pci.di.pAttrHdr;
    if (nPos > pAttrHdr->file.nSize){
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit1;
    }
    u64 nNewPos = (u64)nPos + (u64)nSize;
    if (nNewPos > pAttrHdr->file.nSize){
        nNewPos = pAttrHdr->file.nSize;
        nSize = pAttrHdr->file.nSize - nPos;
    }
    nReadSize = nSize;

    // Cases:
    // 1. Read a small block to a 512 Byte aligned memory address - direct read possible
    // 2. Read a large block, or a set of 64KB pieces from a large block to a 512 Byte aligned memory address - direct read possible
    // 3. Read a portion of a small block to a memory address (whether aligned or not) - must read to block cache first and copy
    // 4. Read a portion of a 64KB piece to a memory address (whether aligned or not) - read to an internal 64KB buffer first
    // Note: If a block is already in cache, use it regardless of the transfer size.
    // ToDo: It may be necessasry to make the block cache support allocating either 8KB or 64KB chunks, which would imply it
    //       needs to implement garbage collection as well.

    // ToDo: The root directory of the sub-area should store its hash code in its parent directory
    // ToDo: Hash-code initialization vector should be based on relative block address within the area XOR some intrinsic property of area like its size.
    AreaInfo *pAreaInfo = pKfi->pci.di.tba.pAreaInfo;
    TransBlkAdr *pTba = &pKfi->pci.di.tba;
    if(!nSize){
        nResult = WFSKRN_RESULT_OK;
        goto Exit1;
    }

    u32 nLog2UnitBlkSize = 0;
    WFSBlkAdr *pParentBlkAdr = 0;
    WFSFileBlkInfo (*pFuncWFSKrnGetFileBlkInfo)(AreaInfo *pAreaInfo, void *pMetaData, u32 nLog2MetaSize, u32 nBlkIdx) = 0;
    WFSKrnResult (*pFuncWfsSrvTransGetAndPinFileBlk)(TransBlkAdr *pTba, WFSBlkAdr nParentBlkAdr, u16 nHashOfs, WFSBlkAdr nFileBlkAdr, u32 nFlags, u32 nValidSize, void **ppBlkPtr) = 0;
    WFSKrnResult (*pFuncWfsSrvTransReadFileBlks)(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nNumBlks, TransInfo *pTransInfo) = 0;
    // for VERY_LARGE
    u32 nNumIndirectBlks = 0;
    s32 nIndirectIdx = 0;
    u8* pIndirectBlkPtr[WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS];

    switch(pAttrHdr->nSizeCategory) {
        case FILE_SIZE_CATEGORY_VERY_SMALL: {
            u8 *pFileDataPtr = &pAttrHdr->aCaseBitArray[(pAttrHdr->nNameLen+7)>>3] + nPos;
            memcpy(pBuffer, pFileDataPtr, nSize);
            nSize = 0;
            break;
        }
        case FILE_SIZE_CATEGORY_SMALL: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2BlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategorySmallCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileBlk;
            pFuncWfsSrvTransReadFileBlks     = WfsSrvTransReadFileBlksForCategorySmall;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_MEDIUM: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategoryMediumCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransReadFileBlks     = WfsSrvTransReadFileBlksForCategoryMedium;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategoryLargeCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransReadFileBlks     = WfsSrvTransReadFileBlksForCategoryLarge;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_VERY_LARGE: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLargeCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransReadFileBlks     = WfsSrvTransReadFileBlksForCategoryVeryLarge;
            pParentBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndLargeBlkIdxForCategoryVeryLarge(pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nPos>>pAreaInfo->ah.nLog2LargeBlkSize);
            nNumIndirectBlks = (nSize+nPos%(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize)-1)/(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize)+1;
            break;
        }
    }

    while(nSize > 0){
        void *pMetaData;
        void *pParentBlkPtr;
        u32 nLog2MetaSize;
        u32 nAccessSize, nAccessPos;

        if(pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_LARGE){
            if(!(nIndirectIdx&(WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS-1))){
                for(nIndirectIdx=0;nIndirectIdx<WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS && nIndirectIdx < nNumIndirectBlks;nIndirectIdx++){
                    WFSKrnExit1OnError(TransGetAndPinBlk5(pAreaInfo, pParentBlkAdr[-nIndirectIdx], pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr[nIndirectIdx]));
                }
                nNumIndirectBlks -= nIndirectIdx;
                nIndirectIdx=0;
            }
            pMetaData = pIndirectBlkPtr[nIndirectIdx];
            nLog2MetaSize = pAreaInfo->ah.nLog2BlkSize;
            pParentBlkPtr = pMetaData;
            nAccessPos    = nPos%(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize);
            nAccessSize   = (pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize) - nAccessPos;
            if(nAccessSize > nSize) {
                nAccessSize = nSize;
                nSize = 0;
            } else {
                nSize -= nAccessSize;
            }
        } else {
            pMetaData     = (void*)pAttrHdr;
            nLog2MetaSize = pAttrHdr->nAttrLog2Size;
            pParentBlkPtr = pKfi->pci.di.pBlkHdr;
            nAccessSize   = nSize;
            nAccessPos    = nPos;
            nSize = 0;
        }

        u8 *pBlkPtr;
        u32 nBlkMask           = (1<<nLog2UnitBlkSize)-1;
        s32 nStartBlkIdx       = nAccessPos >> nLog2UnitBlkSize;
        u32 nStartOfs          = nAccessPos & nBlkMask;
        u32 nStartBlkFraction  = (-(signed)nAccessPos) & nBlkMask;
        s32 nEndBlkIdx         = (nAccessPos + nAccessSize - 1) >> nLog2UnitBlkSize;
        u32 nEndOfs            = (nAccessPos + nAccessSize) & nBlkMask;
        u32 nEndBlkFraction    = 0;
        s32 nMaxBlkIdx         = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize - 1) >> nLog2UnitBlkSize) : 0;
        s32 nBufferRotationOfs = ((u32)pBuffer + nStartBlkFraction) & (WFS_STATIC_MEM_ALIGNMENT-1);
        if (nBufferRotationOfs > nStartBlkFraction) {
            nBufferRotationOfs = nBufferRotationOfs - WFS_STATIC_MEM_ALIGNMENT;
        }
        // If nStartOfs is not 0, nStartBlk is accessed path through block cache.
        if(nStartOfs){
            WFSFileBlkInfo fileBlkInfo = pFuncWFSKrnGetFileBlkInfo(pAreaInfo, pMetaData, nLog2MetaSize, nStartBlkIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pParentBlkPtr;
            if( !pAttrHdr->file.nSize || (nStartBlkIdx == nMaxBlkIdx && pAttrHdr->file.nSize&((1<<nLog2UnitBlkSize)-1)) ){
                WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, pAttrHdr->file.nSize&((1<<nLog2UnitBlkSize)-1), (void**)&pBlkPtr));
            } else {
                WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, 1<<nLog2UnitBlkSize, (void**)&pBlkPtr));
            }
            if (nStartBlkIdx == nEndBlkIdx){
                memcpy(pBuffer, pBlkPtr + nStartOfs, nAccessSize);
            } else {
                memcpy(pBuffer, pBlkPtr + nStartOfs, nStartBlkFraction);
            }
            WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
            ++nStartBlkIdx;
        }
        u32 nDirectReadSize = 0;
        u32 nEnableByteRotation = 0;
        // If nEndOfs is not 0, nEndBlk is accessed path through block cache. (just ready parameters)
        if ((nStartBlkIdx <= nEndBlkIdx) && nEndOfs){
            if(!(nEndOfs&((1<<pTba->pAreaInfo->pVolInfo->pDevInfo->dh.nLog2SectorSize)-1)) && (nBufferRotationOfs >= 0) && (nPos+nAccessSize == pAttrHdr->file.nSize) ){
                nDirectReadSize += nEndOfs;
                --nEndBlkIdx;
            } else {
                nEndBlkFraction += nEndOfs;
                nEnableByteRotation += nEndOfs;
                --nEndBlkIdx;
            }
        }
        // If nEndOfs and nStartBlkFraction is smaller than nBufferRotationOfs, the extra block is read. (just ready parameters)
        while ((nStartBlkIdx <= nEndBlkIdx) && (nBufferRotationOfs < 0) && (nEnableByteRotation < -nBufferRotationOfs)){
            nEnableByteRotation += 1<<nLog2UnitBlkSize;
            nEndBlkFraction += 1<<nLog2UnitBlkSize;
            --nEndBlkIdx;
        }
        if(nStartBlkIdx <= nEndBlkIdx){
            nDirectReadSize += (nEndBlkIdx-nStartBlkIdx+1)<<nLog2UnitBlkSize;
        }
        // Read remained blocks
        if(nDirectReadSize){
            if (nBufferRotationOfs) {
                RotateBuffer((u8*)pBuffer, nAccessSize, nBufferRotationOfs, true);
                //osTPrintf("rotation\n");
            }
            u8 *pAlignedPtr = (u8*)pBuffer + nStartBlkFraction - nBufferRotationOfs;
            WFSKrnExit1OnError(pFuncWfsSrvTransReadFileBlks(pAreaInfo, pMetaData, pAlignedPtr, nStartBlkIdx, nDirectReadSize, pTransInfo));
            //osTPrintf("nNumBlk = %d, nBlkAdr = %d, pHash = 0x%x\n", nEndBlkIdx-nStartBlkIdx+1, pFileBlkPtr->nBlkAdr, &pFileBlkPtr->hash);
            if (nBufferRotationOfs) {
                RotateBuffer((u8*)pBuffer, nAccessSize, -nBufferRotationOfs, false);
            }
        }
        // Read end block(s)
        //osTPrintf("nSize = %d, nEndBlkFraction = %d\n", nSize, nEndBlkFraction);
        void *pBufferEndBlk = (void*)((size_t)pBuffer + nAccessSize - nEndBlkFraction);
        while(nEndBlkFraction){
            ++nEndBlkIdx;
            WFSFileBlkInfo fileBlkInfo = pFuncWFSKrnGetFileBlkInfo(pAreaInfo, pMetaData, nLog2MetaSize, nEndBlkIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pParentBlkPtr;
            if( !pAttrHdr->file.nSize || (nEndBlkIdx == nMaxBlkIdx && pAttrHdr->file.nSize&((1<<nLog2UnitBlkSize)-1)) ){
                WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, pAttrHdr->file.nSize&((1<<nLog2UnitBlkSize)-1), (void**)&pBlkPtr));
            } else {
                WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, 1<<nLog2UnitBlkSize, (void**)&pBlkPtr));
            }
            if(nEndBlkFraction > (1<<nLog2UnitBlkSize)){
                memcpy(pBufferEndBlk, pBlkPtr, (1<<nLog2UnitBlkSize));
                pBufferEndBlk = (void*)((size_t)pBufferEndBlk+(1<<nLog2UnitBlkSize));
                nEndBlkFraction -= (1<<nLog2UnitBlkSize);
            } else{
                memcpy(pBufferEndBlk, pBlkPtr, nEndBlkFraction);
                nEndBlkFraction = 0;
            }
            //osTPrintf("nHashOfs = %d, nBlkAdr = %d, pBlkPtr = 0x%x\n", nHashOfs, pFileBlkPtr->nBlkAdr, pBlkPtr);
            WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
        }

        if(pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_LARGE){
            WFSKrnExit1OnError(TransUnpinBlk3(pAreaInfo, *pParentBlkAdr, pTransInfo));
            pParentBlkAdr = &pParentBlkAdr[-1];
            ++nIndirectIdx;
        }
        nPos   += nAccessSize;
        pBuffer = (void*)((size_t)pBuffer + nAccessSize);
    }

Exit1:
    TransUnpinBlk(&pKfi->pci.di.tba);
Exit2:
    TransEnd(pKfi->pci.di.tba.pTransInfo);
Exit3:
    if(nResult < 0)
    {
        nReadSize = (s32)nResult;
    }
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput4("ReadFile(fh=%04x, pBuf, p=%u, s=%d) ", sfh, nInPos, nInSize);
        if (nReadSize >= 0) {
            WFSKrnDbgOutput2("return %d\n", nReadSize);
        } else {
            WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString((WFSKrnResult)nReadSize));
        }
    );
#endif
    return nReadSize;
}


static inline WFSKrnResult WfsSrvTransWriteFileBlksForCategorySmall(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo, bool bDecryption){
    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size, nBlkIdx);
    return TransWriteFileBlks(pBuffer, pAreaInfo, nSize, pFileBlkPtr, -(signed)sizeof(*pFileBlkPtr), pTransInfo, bDecryption);
}

static inline WFSKrnResult WfsSrvTransWriteFileBlksForCategoryMedium(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo, bool bDecryption){
    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size, nBlkIdx);
    return TransWriteFileMediumBlks(pBuffer, pAreaInfo, nSize, pFileBlkPtr, -(signed)sizeof(*pFileBlkPtr), pTransInfo, bDecryption);
}

static inline WFSKrnResult WfsSrvTransWriteFileBlksForCategoryLarge(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo, bool bDecryption){
    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge((DirEntryAttrHdr*)pMetaData, ((DirEntryAttrHdr*)pMetaData)->nAttrLog2Size, nBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
    return TransWriteFileLargeBlks(pBuffer, pAreaInfo, nSize, nBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                  pFileLargeBlkPtr, -(signed)sizeof(*pFileLargeBlkPtr), pTransInfo, bDecryption);
}

static inline WFSKrnResult WfsSrvTransWriteFileBlksForCategoryVeryLarge(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nSize, TransInfo *pTransInfo, bool bDecryption){
    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge((u8*)pMetaData, nBlkIdx>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
    return TransWriteFileLargeBlks(pBuffer, pAreaInfo, nSize, nBlkIdx%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk),
                                   pFileLargeBlkPtr, (s32)sizeof(*pFileLargeBlkPtr), pTransInfo, bDecryption);
}


static u32 nGlobalChainId = 0;


WFSResult WFSSrvWriteFile(WFSSrvFileHandle sfh, const void *pBuffer, WFSFileSize nPos, WFSFileSize nSize, WFSBool bDiscard) {
#if _DEBUG_BREAK_POINT
    static int c = 0; c++; if (c == 0x2866) { //839702
        int a;a=0;
    }
#endif
#if _DEBUG
    #if ((DBG_OUTPUT_TO_LOG_FILE && _WIN32) || DBG_OUTPUT_TO_SCREEN)
        WFSFileSize nInPos = nPos;
        s32 nInSize = nSize;
    #endif
#endif
    //if((size_t)pBuffer%WFSKRN_FILE_BUFFER_MEM_ALIGNMENT_SIZE){
    //    return WFS_RESULT_INVALID;
    //}
    WFSKrnResult nResult = WFSKRN_RESULT_UNKNOWN;
    WFSKrnFileInfo *pKfi;
    WFSKrnExit3OnError(WFSKrnGetFileInfo(&pKfi, sfh));
    TransInfo *pTransInfo = pKfi->pci.di.tba.pTransInfo = TransBegin();
    if ((pKfi->nAccess & WFS_ACCESS_WRITE) == 0) {
        nResult = (WFSKrnResult)WFS_RESULT_ACCESS;
        goto Exit2;
    }
    nResult = DirItrUpdateAndPinBlk(&pKfi->pci.di, pKfi->pci.nRootBlkAdr, TRANS_FLAG_RW_BLK);
    if (nResult != WFSKRN_RESULT_OK)
    {
        goto Exit2;
    }
    DirEntryAttrHdr *pAttrHdr = pKfi->pci.di.pAttrHdr;
    TransBlkAdr *pTba = &pKfi->pci.di.tba;
    AreaInfo *pAreaInfo = pKfi->pci.di.tba.pAreaInfo;

    if (nPos > pAttrHdr->file.nSize){
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit1;
    }
    if(!nSize){
        nResult = WFSKRN_RESULT_OK;
        goto Exit1;
    }
    if (pAttrHdr->file.nSize < (u64) nPos)
    {
        nResult = WFSKRN_RESULT_INVALID;
        goto Exit1;
    }
    u32 nOldSize = pKfi->pci.di.pAttrHdr->file.nSize;
    u64 nNewSize = (u64)nPos + (u64)nSize;
    if(nNewSize>>WFS_LOG2_MAX_FILE_SIZE){
        nNewSize = (u32)-1;
        nSize = nNewSize - nPos;
    } else if (nOldSize > nNewSize){
        nNewSize = nOldSize;
    }
#ifdef PERMISSION_ENABLED
    if (nNewSize > pAttrHdr->file.nSize)
    {
        if((pKfi->nAccess & WFS_ACCESS_CHANGE_SIZE) == 0)
        {
            nResult = (WFSKrnResult)WFS_RESULT_NO_CHANGE_SIZE;
            goto Exit1;
        }
    }
#endif
#if _DEBUG_BREAK_POINT
    if (nNewSize == 0xfde71){
        int a; a=0;
    }
#endif
  osTPrintf( "nNewSize:0x%x, nAllocSize:0x%x\n", nNewSize, pKfi->pci.di.pAttrHdr->nAllocSize);
    if (nNewSize > pKfi->pci.di.pAttrHdr->nAllocSize){
        WFSKrnExit1OnError(DirResizeFile(&pKfi->pci.di, nNewSize));
        pAttrHdr = pKfi->pci.di.pAttrHdr;
    }

    u32 nLog2UnitBlkSize = 0;
    WFSBlkAdr *pParentBlkAdr = 0;
    WFSFileBlkInfo (*pFuncWFSKrnGetFileBlkInfo)(AreaInfo *pAreaInfo, void *pMetaData, u32 nLog2MetaSize, u32 nBlkIdx) = 0;
    WFSKrnResult (*pFuncWfsSrvTransGetAndPinFileBlk)(TransBlkAdr *pTba, WFSBlkAdr nParentBlkAdr, u16 nHashOfs, WFSBlkAdr nFileBlkAdr, u32 nFlags, u32 nValidSize, void **ppBlkPtr) = 0;
    WFSKrnResult (*pFuncWfsSrvTransWriteFileBlks)(AreaInfo *pAreaInfo, void *pMetaData, u8 *pBuffer, u32 nBlkIdx, u32 nNumBlks, TransInfo *pTransInfo, bool bDecryption) = 0;
    // for VERY_LARGE
    u32 nNumIndirectBlks = 0;
    s32 nIndirectIdx = 0;
    u8* pIndirectBlkPtr[WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS];
    u32 nChainId = 0;

    switch(pAttrHdr->nSizeCategory) {
        case FILE_SIZE_CATEGORY_VERY_SMALL: {
            u8 *pFileDataPtr = &pAttrHdr->aCaseBitArray[(pAttrHdr->nNameLen+7)>>3] + nPos;
            memcpy(pFileDataPtr, pBuffer, nSize);
            nPos  += nSize;
            nSize = 0;
            break;
        }
        case FILE_SIZE_CATEGORY_SMALL: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2BlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategorySmallCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileBlk;
            pFuncWfsSrvTransWriteFileBlks    = WfsSrvTransWriteFileBlksForCategorySmall;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_MEDIUM: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategoryMediumCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransWriteFileBlks    = WfsSrvTransWriteFileBlksForCategoryMedium;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoForCategoryLargeCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransWriteFileBlks    = WfsSrvTransWriteFileBlksForCategoryLarge;
            pParentBlkAdr = &pTba->nBlkAdr;
            break;
        }
        case FILE_SIZE_CATEGORY_VERY_LARGE: {
            nLog2UnitBlkSize = pAreaInfo->ah.nLog2MediumBlkSize;
            pFuncWFSKrnGetFileBlkInfo        = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLargeCast;
            pFuncWfsSrvTransGetAndPinFileBlk = TransGetAndPinFileMediumBlk;
            pFuncWfsSrvTransWriteFileBlks     = WfsSrvTransWriteFileBlksForCategoryVeryLarge;
            pParentBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndLargeBlkIdxForCategoryVeryLarge(pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nPos>>pAreaInfo->ah.nLog2LargeBlkSize);
            nNumIndirectBlks = (nSize+nPos%(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize)-1)/(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize)+1;
            if(TransGetChainId4(pAreaInfo, pParentBlkAdr[0], pTransInfo, &nChainId) < WFSKRN_RESULT_OK || nChainId == 0){
                ++nGlobalChainId;
                if(!nGlobalChainId){ nGlobalChainId = 1;}
                nChainId = nGlobalChainId;
            }
            break;
        }
    }

    while(nSize > 0){
        void *pMetaData;
        void *pHashBlkPtr;
        u32 nLog2MetaSize;
        u32 nAccessSize, nAccessPos;

        if(pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_LARGE){
            if(!(nIndirectIdx&(WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS-1))){
                for(nIndirectIdx=0;nIndirectIdx<WFSKRN_VERY_LARGE_MAP_NUM_INDIRECT_BLKS && nIndirectIdx < nNumIndirectBlks;nIndirectIdx++){
                    WFSKrnExit1OnError(TransGetAndPinBlk5(pAreaInfo, pParentBlkAdr[-nIndirectIdx], pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr[nIndirectIdx]));
                    WFSKrnExit1OnError(TransSetChainId4(pAreaInfo, pParentBlkAdr[-nIndirectIdx], pTransInfo, nChainId));
                }
                nNumIndirectBlks -= nIndirectIdx;
                nIndirectIdx = 0;
            }
            pMetaData = pIndirectBlkPtr[nIndirectIdx];
            nLog2MetaSize = pAreaInfo->ah.nLog2BlkSize;
            pHashBlkPtr   = pMetaData;
            nAccessPos    = nPos%(pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize);
            nAccessSize   = (pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->ah.nLog2LargeBlkSize) - nAccessPos;
            if(nAccessSize > nSize) {
                nAccessSize = nSize;
                nSize = 0;
            } else {
                nSize -= nAccessSize;
            }
        } else {
            pMetaData     = (void*)pAttrHdr;
            nLog2MetaSize = pAttrHdr->nAttrLog2Size;
            pHashBlkPtr   = pKfi->pci.di.pBlkHdr;
            nAccessSize   = nSize;
            nAccessPos    = nPos;
            nSize = 0;
        }

        u8 *pBlkPtr;
        u32 nBlkMask           =  (1<<nLog2UnitBlkSize)-1;
        s32 nStartBlkIdx       =  nAccessPos >> nLog2UnitBlkSize;
        u32 nStartOfs          =  nAccessPos & nBlkMask;
        u32 nStartBlkFraction  = (-(signed)nAccessPos) & nBlkMask;
        s32 nEndBlkIdx         = (nAccessPos + nAccessSize - 1) >> nLog2UnitBlkSize;
        u32 nEndOfs            = (nAccessPos + nAccessSize) & nBlkMask;
        u32 nEndBlkFraction    = 0;
        s32 nOldMaxBlkIdx      = nOldSize ? ((nOldSize - 1) >> nLog2UnitBlkSize) : 0;
        s32 nNewMaxBlkIdx      = nNewSize ? ((nNewSize - 1) >> nLog2UnitBlkSize) : 0;
        s32 nBufferRotationOfs = ((u32)pBuffer + nStartBlkFraction) & (WFS_STATIC_MEM_ALIGNMENT-1);
        if (nBufferRotationOfs > nStartBlkFraction) {
            nBufferRotationOfs = nBufferRotationOfs - WFS_STATIC_MEM_ALIGNMENT;
        }
        // If nStartOfs is not 0, nStartBlk is accessed path through block cache.
        if(nStartOfs){
            WFSFileBlkInfo fileBlkInfo = pFuncWFSKrnGetFileBlkInfo(pAreaInfo, pMetaData, nLog2MetaSize, nStartBlkIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pHashBlkPtr;
            u32 nValidSize;
            if(nStartBlkIdx < nOldMaxBlkIdx){
                nValidSize = 1<<nLog2UnitBlkSize;
            } else if (nStartBlkIdx == nOldMaxBlkIdx){
                nValidSize = ((nOldSize - 1)&((1<<nLog2UnitBlkSize)-1))+1;
            } else {
                nValidSize = 0;
            }
            WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_RW_BLK, nValidSize, (void**)&pBlkPtr));
            if (nStartBlkIdx == nEndBlkIdx){
                if(nStartBlkIdx == nNewMaxBlkIdx){
                    nValidSize = ((nNewSize-1)&((1<<nLog2UnitBlkSize)-1))+1;
                } else {
                    nValidSize = 1<<nLog2UnitBlkSize;
                }
                memcpy(pBlkPtr + nStartOfs, pBuffer, nAccessSize);
            } else {
                nValidSize = 1<<nLog2UnitBlkSize;
                memcpy(pBlkPtr + nStartOfs, pBuffer, nStartBlkFraction);
            }
            WFSKrnExit1OnError(TransSetValidSize4(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo, nValidSize));
            if (bDiscard){
                WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                WFSKrnExit1OnError(TransFlushBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
            } else {
                WFSKrnExit1OnError(TransStoreBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
            }
            ++nStartBlkIdx;
        }
        u32 nDirectWriteSize = 0;
        u32 nEnableByteRotation = 0;
        // If nEndOfs is not 0, nEndBlk is accessed path through block cache. (just ready parameters)
        if ((nStartBlkIdx <= nEndBlkIdx) && nEndOfs){
            if(!(nEndOfs&((1<<pTba->pAreaInfo->pVolInfo->pDevInfo->dh.nLog2SectorSize)-1)) && (nBufferRotationOfs >= 0) && (nPos+nAccessSize == nNewSize) ){
                nDirectWriteSize += nEndOfs;
                --nEndBlkIdx;
            } else {
                nEndBlkFraction += nEndOfs;
                nEnableByteRotation += nEndOfs;
                --nEndBlkIdx;
            }
        }
        // If nEndOfs and nStartBlkFraction is smaller than nBufferRotationOfs, the extra block is read. (just ready parameters)
        while ((nStartBlkIdx <= nEndBlkIdx) && (nBufferRotationOfs < 0) && (nEnableByteRotation < -nBufferRotationOfs)){
            nEnableByteRotation += 1<<nLog2UnitBlkSize;
            nEndBlkFraction += 1<<nLog2UnitBlkSize;
            --nEndBlkIdx;
        }
        if(nStartBlkIdx <= nEndBlkIdx){
            nDirectWriteSize += (nEndBlkIdx-nStartBlkIdx+1)<<nLog2UnitBlkSize;
        }
        // Write remained blocks
        if(nDirectWriteSize){
            if (nBufferRotationOfs) {
                RotateBuffer((u8*)pBuffer, nAccessSize, nBufferRotationOfs, true);
            }
            u8 *pAlignedPtr = (u8*)pBuffer + nStartBlkFraction - nBufferRotationOfs;
            //if((size_t)pAlignedPtr%0x200) osTPrintf("pAlignedPtr Error\n");
            WFSKrnExit1OnError(pFuncWfsSrvTransWriteFileBlks(pAreaInfo, pMetaData, pAlignedPtr, nStartBlkIdx, nDirectWriteSize, pTransInfo, !bDiscard));
            //osTPrintf("nNumBlk = %d, nBlkAdr = %d, pHash = 0x%x\n", nEndBlkIdx-nStartBlkIdx+1, pFileBlkPtr->nBlkAdr, &pFileBlkPtr->hash);
            if (nBufferRotationOfs) {
                RotateBuffer((u8*)pBuffer, nAccessSize, -nBufferRotationOfs, false);
            }
        }
        // Write end block(s)
        //osTPrintf("nSize = %d, nEndBlkFraction = %d\n", nSize, nEndBlkFraction);
        void* pBufferEndBlk = (void*)((size_t)pBuffer + nAccessSize - nEndBlkFraction);
        while(nEndBlkFraction){
            ++nEndBlkIdx;
            WFSFileBlkInfo fileBlkInfo = pFuncWFSKrnGetFileBlkInfo(pAreaInfo, pMetaData, nLog2MetaSize, nEndBlkIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pHashBlkPtr;
            //osTPrintf("nHashOfs = %d, nBlkAdr = %d, pBlkPtr = 0x%x\n", nHashOfs, pFileBlkPtr->nBlkAdr, pBlkPtr);
            u32 nValidSize;
            if(nEndBlkIdx < nOldMaxBlkIdx){
                nValidSize = 1<<nLog2UnitBlkSize;
            } else if (nEndBlkIdx == nOldMaxBlkIdx){
                nValidSize = ((nOldSize-1)&((1<<nLog2UnitBlkSize)-1))+1;
            } else {
                nValidSize = 0;
            }
            WFSKrnExit1OnError(pFuncWfsSrvTransGetAndPinFileBlk(pTba, *pParentBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_RW_BLK, nValidSize, (void**)&pBlkPtr));
            if(nEndBlkFraction > (1<<nLog2UnitBlkSize)){
                nValidSize = 1<<nLog2UnitBlkSize;
                memcpy(pBlkPtr, pBufferEndBlk, (1<<nLog2UnitBlkSize));
                pBufferEndBlk = (void*)((size_t)pBufferEndBlk+(1<<nLog2UnitBlkSize));
                nEndBlkFraction -= (1<<nLog2UnitBlkSize);
            } else{
                if(nEndBlkIdx == nNewMaxBlkIdx){
                    nValidSize = ((nNewSize-1)&((1<<nLog2UnitBlkSize)-1))+1;
                } else {
                    nValidSize = 1<<nLog2UnitBlkSize;
                }
                memcpy(pBlkPtr, pBufferEndBlk, nEndBlkFraction);
                nEndBlkFraction = 0;
            }
            WFSKrnExit1OnError(TransSetValidSize4(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo, nValidSize));
            if(bDiscard){
                WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                WFSKrnExit1OnError(TransFlushBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
            } else {
                WFSKrnExit1OnError(TransStoreBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                WFSKrnExit1OnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
            }
        }

        if(pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_LARGE){
            WFSKrnExit1OnError(TransUnpinBlk3(pAreaInfo, *pParentBlkAdr, pTransInfo));
            pParentBlkAdr = &pParentBlkAdr[-1];
            ++nIndirectIdx;
        }
        nPos   += nAccessSize;
        pBuffer = (void*)((size_t)pBuffer + nAccessSize);
    }

    //nPos += nSize;
    if (pKfi->pci.di.pAttrHdr->file.nSize < nPos) {
        pKfi->pci.di.pAttrHdr->file.nSize = nPos;
    }
    // Update attributes in pKfi->attr.
    pKfi->attr = *pKfi->pci.di.pAttrHdr;
Exit1:
    TransUnpinBlk(pTba);
Exit2:
    TransEnd(pKfi->pci.di.tba.pTransInfo);
Exit3:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput4("WriteFile(fh=%04x, pBuf, p=%u, s=%u) ", sfh, nInPos, nInSize);
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString((WFSKrnResult)nResult));
    );
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvFlush(const WFSVolumeId *pVolumeId){

    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    utf8 sVolPath[WFS_MAX_VOLUME_ID_SIZE + 6];
    osSNPrintf(sVolPath, WFS_MAX_VOLUME_ID_SIZE + 6, "/vol/%s", pVolumeId->sStr);
    VolumeInfo *pVolInfo;
    RxtItr rxi;
    rxi.pStrPtr = sVolPath;
    PathCacheEntry *pEntry = PathCacheFind2(&rxi, (void **)&pVolInfo);
    if (!pEntry) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    nResult = BCacheStoreVolume(pVolInfo);
Exit:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(WFSKrnDbgOutputCmdNum; WFSKrnDbgOutput3("Flush(\"%s\") WFS_RESULT_%s\n", pVolumeId->sStr, WFSKrnGetErrorString(nResult)));
#endif
    return (WFSResult)nResult;
}


WFSResult WFSSrvSetTitleId(WFSTitleId titleId, WFSGroupId groupId) {
#if _WIN32 || PERM_TEST
    wkg.nDebugTitleId = titleId;
#endif
    return WFS_RESULT_OK;
}


static WFSResult WFSSrvCheckDisk(WFSSrvFileHandle* pfh, utf8 *pDevName){

    WFSDeviceName devName;
    strncpy(devName.sStr, pDevName, WFS_MAX_DEVICE_NAME_SIZE);
    devName.nLen = STRNLEN(devName.sStr, WFS_MAX_DEVICE_NAME_SIZE);
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(&devName);
    AreaInfo *pAreaInfo = &pDevInfo->pVolInfo->rootAreaInfo;
    u32 nSize = pAreaInfo->ah.nNumBlks / 8;

    WFSSrvFileHandle fh;
    WFSFileAttributes fa;
    WFSPathName bitArrayFilePath;

    WFSVolumeId volId;
    VolumeConvertU32ToVolId(&volId, pAreaInfo->pVolInfo->vh.nVolId);
    osSNPrintf((char*)bitArrayFilePath.sStr, WFS_MAX_PATH_NAME_SIZE, "/vol/%s/%s", volId.sStr, CHECKDISKFILENAME);
    bitArrayFilePath.nLen = STRNLEN(bitArrayFilePath.sStr, WFS_MAX_PATH_NAME_SIZE);


    // Open bit array file
    // todo: check empty area size
    // todo: case: the number of blocks is not product of 8
    WFSResult result = WFSSrvCreateAndOpenFile(&bitArrayFilePath, WFS_PERM_FULL, 0, nSize, &fh);
    if(result != WFS_RESULT_OK){
        if(result == WFS_RESULT_ALREADY_EXISTS){
            result = WFSSrvOpenFile(&bitArrayFilePath, WFS_ACCESS_RW, &fh, &fa);
            if(result != WFS_RESULT_OK){
                osTPrintf("WFSSrvCheckDisk: failed to open file (already exists) %d\n", result);
                return (WFSResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
            }
        }
        osTPrintf("WFSSrvCheckDisk: failed to open file %d\n", result);
        return (WFSResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
    }

    //osTPrintf("fh %d\n", fh);
    //osTPrintf("nSize: %d\n", nSize);
    //osTPrintf("nNumBlks: %d\n", pAreaInfo->ah.nNumBlks);

    u8* pBuffer;
    WFSKrnResult krnResult = WFSKRN_RESULT_OK;

    // Alloc and pin
    WFSBlkAdr blkAdr;
    krnResult = AreaAllocBlks(pAreaInfo, 0, 16, 1, &blkAdr, 0); // reserve 64KB block x 1
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("WFSSrvCheckDisk: failed to alloc a block.\n");
        krnResult = (WFSKrnResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
        goto CLOSEFILE;
    }
    krnResult = TransGetAndPinFileBlk8(pAreaInfo, blkAdr, 0, blkAdr, NULL, (BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED), 1<<pAreaInfo->ah.nLog2MediumBlkSize, BCACHE_GROUP_ID_USERBLK_64K, (void**)&pBuffer);
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("WFSSrvCheckDisk: failed to pin the block.\n");

        krnResult = (WFSKrnResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
        goto UNPINBLOCK;
    }

    u32 bufferSize = 64*1024; // BCACHE_GROUP_ID_USERBLK_64K
    u8* pBuffer512 = pBuffer;

    // clear bit array file
    // todo: error handling
    memset(pBuffer512, 0, bufferSize);
    int nPos;
    for(nPos=0; nPos<nSize; nPos += bufferSize){
        int writeSize = bufferSize;
        if( nPos+bufferSize > nSize ) writeSize = nSize - nPos;
        WFSSrvWriteFile(fh, pBuffer512, nPos, writeSize, false);
    }

    // set 1 for blkAdr itself (USERBLK_64K -> 8KB block x 8 -> 1byte)
    pBuffer512[0] |= (u8) (0xff << (blkAdr % 8));
    pBuffer512[1] |= (u8)~(0xff << (blkAdr % 8));
    u32 writeSize = 1;
    if(blkAdr % 8 != 0) writeSize = 2;
    WFSSrvWriteFile(fh, pBuffer512, blkAdr/8, writeSize, false);


    //osTPrintf("LINE: %d\n", __LINE__);
    // Check disk for Free block allocator
    krnResult = AreaPinFreeListHdr(pAreaInfo, BCACHE_FLAG_READ , &pAreaInfo->pAflh);
    WFSKRN_FTREE_ASSERT(krnResult == WFSKRN_RESULT_OK);
    krnResult = FreeBlkCheckDisk(pAreaInfo, fh, pBuffer512, bufferSize, nSize);
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("Error : FreeBlkCheckDisk()\n");
    }
    AreaUnpinFreeListHdr(pAreaInfo);

    PathCacheItr pci;
    WFSPathName path, path1;
    osSNPrintf((char*)path1.sStr, WFS_MAX_PATH_NAME_SIZE, "/vol/%s", volId.sStr);
    strncpy(path.sStr, path1.sStr, WFS_MAX_PATH_NAME_SIZE);
    path.nLen = STRNLEN(path.sStr, WFS_MAX_PATH_NAME_SIZE);
    PathCacheFindPath(&path, path1.sStr, &pci);
    //osTPrintf("LINE: %d\n", __LINE__);

    // Check disk for access list
    krnResult = DirCheckDisk(pAreaInfo->ah.aclHdr.aclBlock, &pci.di, fh, pBuffer512, bufferSize, DIR_CHECK_DISK_FOR_ACL_TREE);
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("WFSSrvCheckDisk: failed at acl1\n");
        krnResult = (WFSKrnResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
        goto UNPINBLOCK;
    }
    krnResult = DirCheckDisk(pAreaInfo->ah.aclHdr.nameBlock, &pci.di, fh, pBuffer512, bufferSize,  DIR_CHECK_DISK_FOR_ACL_TREE);
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("WFSSrvCheckDisk: failed at acl2 \n");
        krnResult = (WFSKrnResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
        goto UNPINBLOCK;
    }
    //osTPrintf("LINE: %d\n", __LINE__);
    // Check disk for dir module
    krnResult = DirCheckDisk(pAreaInfo->ah.nRootBlkAdr, &pci.di, fh, pBuffer512, bufferSize,  DIR_CHECK_DISK_FOR_DIR_TREE);
    if (krnResult != WFSKRN_RESULT_OK){
        osTPrintf("WFSSrvCheckDisk: failed at dir module \n");
        krnResult = (WFSKrnResult)WFSKRN_RESULT_CHECKDISK_INTERNAL_ERROR;
        goto UNPINBLOCK;
    }

    //osTPrintf("LINE: %d\n", __LINE__);
    // check if the content of the bit array file is all 1(0xff)
    memset(pBuffer512 + bufferSize/2, 0xff, bufferSize/2);
    for(nPos=0; nPos<nSize; nPos += bufferSize/2){
        int verifySize = bufferSize/2;
        if( nPos+bufferSize/2 > nSize) verifySize = nSize - nPos;     //
        WFSSrvReadFile(fh, pBuffer512, nPos, verifySize);
        if(memcmp(pBuffer512, pBuffer512+bufferSize/2, verifySize) != 0){
            // file system error!
            osTPrintf("Error : Check disk verify : Buffer = %d\n", nPos);
            krnResult = WFSKRN_RESULT_CHECKDISK_INCONSISTENT;
        }
    }

UNPINBLOCK:
    TransUnpinAndFreeBlks(pAreaInfo, 16, 1, &blkAdr, NULL);
    TransInvalidate3(pAreaInfo, blkAdr, NULL);

CLOSEFILE:
    WFSSrvCloseFile(fh, nSize);
    WFSSrvDelete(&bitArrayFilePath);

    return (WFSResult)krnResult;

}

/*
WFSResult WFSSrvCacheDeleteAll(u32 nDevIdx) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    utf8 sVolPath[WFS_MAX_VOLUME_ID_SIZE + 6];
    DeviceInfo *pDevInfo = &wkg.aDevInfo[nDevIdx];
    WFSVolumeId VolumeId;
    VolumeConvertU32ToVolId(&VolumeId, pDevInfo->pVolInfo->vh.nVolId);
    osSNPrintf(sVolPath, WFS_MAX_VOLUME_ID_SIZE + 6, "/vol/%s", VolumeId.sStr);
    VolumeInfo *pVolInfo;
    RxtItr rxi;
    rxi.pStrPtr = sVolPath;
    PathCacheEntry *pEntry = PathCacheFind2(&rxi, (void **)&pVolInfo);
    if (!pEntry) {
        nResult = WFSKRN_RESULT_NOT_FOUND;
        goto Exit;
    }
    if (0 < pEntry->pcd.nOpenHandleCount && pVolInfo->pDevInfo->bInUse) {
#if _DEBUG
        PathCacheDebugOutput();
#endif
        nResult = WFSKRN_RESULT_FILE_OPEN;
        goto Exit;
    }
    nResult = PathCacheDeleteSubTree(pEntry);
    pEntry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, sVolPath);
    if (pEntry){
        pEntry->pcd.pData = pVolInfo;
    }
Exit:;
    return (WFSResult)nResult;
}
*/

#include <private/wfs_Debug.h>

WFSResult WFSSrvExecDebugProcedure(u32 cmd , void *arg0 , void *arg1) {
    WFSResult nResult = WFS_RESULT_OK;

    // convert.
    switch (cmd)
    {
      case ((int) WFSDEBUG_CMD_ATTACH):
        cmd = WFSDEBUG_CMD_DEVICE_ATTACH_DETACH;
        arg0 = (void*) TRUE;
        arg1 = (void*) 0; // [check]
        break;
      case ((int) WFSDEBUG_CMD_DETACH):
        cmd = WFSDEBUG_CMD_DEVICE_ATTACH_DETACH;
        arg0 = (void*) FALSE;
        arg1 = (void*) 0; // [check]
        break;
      case ((int) WFSDEBUG_CMD_ATTACH_WRITE_PROTECTED):
      case ((int) WFSDEBUG_CMD_DELETE_DEVICEROOT):
      case ((int) WFSDEBUG_CMD_DELETE_PATH):
      case ((int) WFSDEBUG_CMD_CLOSE_ALLFILE):
      case ((int) WFSDEBUG_CMD_DUMP_PERMISSION):
      case ((int) WFSDEBUG_CMD_SET_AUTO_FORMAT_ON_ATTACH):
      case ((int) WFSDEBUG_CMD_PUSH_PATH):
      case ((int) WFSDEBUG_CMD_EXEC_BAT):
      case ((int) WFSDEBUG_CMD_TOUCH):
      // default:
        return WFS_RESULT_NOT_IMPLEMENTED;
    }

    switch (cmd) {
/*
        case WFSDEBUG_CMD_KRN_INIT_VIRTUAL_DISK:
#if !_IOP
            gpPtr = (u8*)arg0;
#else
            gpPtr = (u8*)((u32)arg0 & 0x7fffffff);
#endif
            gSize = (u32)arg1;
            break;
*/
//        case WFSDEBUG_CMD_KRN_INIT_AREA_SECTORS:
//            gNumSectors = (u32)arg0;
//            break;

/*
        case WFSDEBUG_CMD_KRN_INIT_CACHE:
#if !_IOP
            gpCachePtr = (u8*)arg0;
#else
            gpCachePtr = (u8*)((u32)arg0 & 0x7fffffff);
#endif
            gCacheSize = (u32)arg1;
            break;
*/

        case WFSDEBUG_CMD_DIR_CHECK_DNS_EMPTY:
            DirNodeStackCheckEmpty;
            break;
/*
        case WFSDEBUG_CMD_PATH_CACHE_DELETE_ALL:{
            u32 nDevIdx = (u32) arg0;
            nResult = WFSSrvCacheDeleteAll(nDevIdx);
            break;
        }
*/
        case WFSDEBUG_CMD_DEVICE_ATTACH_DETACH:
            {
                u32 bAttach = (u32) arg0;
                u32 nDevIdx = (u32) arg1;
                DeviceInfo *pDevInfo = &wkg.aDevInfo[nDevIdx];
                if (bAttach && pDevInfo->bInUse)
                {
                    nResult = WFS_RESULT_DEV_IN_USE;
                }
                else if (!bAttach && !pDevInfo->bInUse)
                {
                    nResult = WFS_RESULT_MEDIA_ERROR;
                }
                else
                {
                    WfsKrnAttachDetachCallback(bAttach, nDevIdx, TEST_LOG2_SECTOR_SIZE, MAX_ALLOC_NUM_SECTORS, WFS_DEVTYPE_USB_MSC_20, 0);
                }
            }
            break;

        case WFSDEBUG_CMD_DEVICE_PSEUDO_DETACH:
            DeviceDebugSetPseudoDetachDevice((s32)arg0);
            break;

        case WFSDEBUG_CMD_DEVICE_PSEUDO_HASH_INCONSISTENT:
            DeviceDebugSetPseudoHashInconsistent((s32)arg0);
            break;

        case WFSDEBUG_CMD_DEVICE_SET_FLAGS:
#if !_IOP
            DeviceGetDeviceInfo((WFSDeviceName*)arg0)->nFlags = (u32)arg1;
#else
            DeviceGetDeviceInfo((WFSDeviceName*)((u32)arg0 & 0x7fffffff))->nFlags = (u32)arg1;
#endif
            break;
        case WFSDEBUG_CMD_AREA_AND_DIR_CHECK_DISK:{
            WFSSrvFileHandle* pfh = (WFSSrvFileHandle*)arg0;
#if !_IOP
            //u8* pCheckDiskArray = (u8*)arg0;
            utf8* pDevString    = (utf8*)arg1;
#else
            //u8* pCheckDiskArray = (u8*)((u32)arg0&0x7fffffff);
            utf8* pDevString    = (utf8*)((u32)arg1&0x7fffffff);
#endif
            nResult = (WFSResult)WFSSrvCheckDisk(pfh, pDevString);
            break;
        }
    }
    return nResult;
}


WFSResult WFSSrvConnectToServer(void) {
    osTPrintf("WFSSrvConnectToServer()\n");
    return WFS_RESULT_OK;
}


WFSResult WFSSrvGetPermissionForUser(const WFSPathName *pAbsPathName, u32 *pFlags, WFSTitleId titleId) {
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    // Check the permission of parent dir and specified entry for executor
    u32 parentPermissions;
    u32 permissions;
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, &permissions, pci.di.tba.pTransInfo->nTitleId));
    if((pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) == 0 || !(permissions & WFS_PERM_LIST))
    {
        WFSKrnExit1OnError(WFSKrnGetParentInheritedPermissionForUserPci(&pci, &parentPermissions, pci.di.tba.pTransInfo->nTitleId));
        if(!(parentPermissions & WFS_PERM_LIST))
        {
            nResult = WFSKRN_RESULT_PERMISSION;
            goto Exit1;
        }
    }
    // Get inherited permissions for user
    WFSKrnExit1OnError(WFSKrnGetInheritedPermissionForUserDi(&pci, pFlags, titleId));
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput4("GetPermissionForUser(\"%s\", pf=%x, t=%x) ", pAbsPathName->sStr, pFlags, titleId);
        if (nResult == WFSKRN_RESULT_OK) {
            WFSKrnDbgOutput2("f=%x ", *pFlags);
        }
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
        );
#endif
    return (WFSResult)nResult;
#else
    osTPrintf("WFSSrvGetPermissionForUser()\n");
    return WFS_RESULT_OK; // dummy
#endif
}


WFSResult WFSSrvSetPermissionForUser(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask, WFSTitleId titleId) {
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    if (!pci.di.tba.pAreaInfo)
    {
        // no one can set permission to the entry which has no area info.
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    WFSKrnExit1OnError(DirGetBlk(&pci.di, TRANS_FLAG_RW_BLK, 0));
    WFSKrnExit1OnError(WFSKrnSetInheritedPermissionForUser(&pci, nFlags, nMask, pci.di.tba.pTransInfo->nTitleId, titleId));
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        // Update acl index data in path cache tree
        pci.pEntry->pcd.nAccessListIdx = pci.di.pAttrHdr->nAccessListIdx;
    }
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput5("SetPermissionForUser(\"%s\", f=%x, m=%x, t=%x) ", pAbsPathName->sStr, nFlags, nMask, titleId);
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
        );
#endif
    return (WFSResult)nResult;
#else
    osTPrintf("WFSSrvSetPermissionForUser()\n");
    return WFS_RESULT_OK; // dummy
#endif
}

WFSResult WFSSrvDeleteAccessListEntry(const WFSPathName *pAbsPathName, WFSTitleId titleId, WFSTitleId entryCreatorId)
{
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    if (!pci.di.tba.pAreaInfo)
    {
        // no one can set permission to the entry which has no area info.
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    WFSKrnExit1OnError(DirGetBlk(&pci.di, TRANS_FLAG_RW_BLK, 0));
    WFSKrnExit1OnError(WFSKrnDeletePermissions(pci.di.tba.pAreaInfo, &(pci.di.pAttrHdr->nAccessListIdx), entryCreatorId, titleId));
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        // Update acl index data in path cache tree
        pci.pEntry->pcd.nAccessListIdx = pci.di.pAttrHdr->nAccessListIdx;
    }
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput4("DeleteAccessListEntry(\"%s\", t=%x, ec=%x) ", pAbsPathName->sStr, titleId, entryCreatorId);
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
        );
#endif
    return (WFSResult)nResult;
#else
    osTPrintf("WFSSrvDeleteAccessListEntry()\n");
    return WFS_RESULT_NOT_IMPLEMENTED; // dummy
#endif
}

s32 WFSSrvGetAccessList(const WFSPathName *pAbsPathName, WFSAccessListEntry *pAccessList, u32 nMaxElements)
{
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    WFSKrnExit1OnError((WFSKrnResult)WFSKrnGetAccessList(pci.di.tba.pAreaInfo, pci.di.pAttrHdr->nAccessListIdx, pAccessList, nMaxElements));
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput4("GetAccessList(\"%s\", pal=%x, me=%x) ", pAbsPathName->sStr, pAccessList, nMaxElements);
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
        );
#endif
    return (s32)nResult;
#else
    osTPrintf("WFSSrvGetAccessList()\n");
    return WFS_RESULT_NOT_IMPLEMENTED; // dummy
#endif
}

WFSResult WFSSrvSetAccessListEntry(const WFSPathName *pAbsPathName, u32 nFlags, u32 nMask, WFSTitleId titleId)
{
#ifdef PERMISSION_ENABLED
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSPathName path;
    WFSKrnExit2OnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, pAbsPathName->sStr, pAbsPathName->nLen, WFS_MAX_PATH_NAME_SIZE));
    PathCacheItr pci;
    pci.di.tba.pTransInfo = TransBegin();
    WFSKrnExit1OnError(PathCacheFindPath(pAbsPathName, path.sStr, &pci));
    if (!pci.di.tba.pAreaInfo)
    {
        // no one can set permission to the entry which has no area info.
        nResult = WFSKRN_RESULT_PERMISSION;
        goto Exit1;
    }
    WFSKrnExit1OnError(DirGetBlk(&pci.di, TRANS_FLAG_RW_BLK, 0));
    if((pci.di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) == 0)
    {
        nMask = nMask & (WFS_PERM_CL_MASK | WFS_PERM_FILE_FULL);
    }
    else
    {
        nMask = nMask & (WFS_PERM_CL_MASK | WFS_PERM_DIR_FULL);
    }
    WFSKrnExit1OnError(WFSKrnSetAccessList(pci.di.tba.pAreaInfo, &(pci.di.pAttrHdr->nAccessListIdx), nFlags, nMask, pci.di.tba.pTransInfo->nTitleId, titleId));
    if (*pci.rxi.pStrPtr==0) { // The pEntry refers to the object (rather than the parent dir)
        // Update acl index data in path cache tree
        pci.pEntry->pcd.nAccessListIdx = pci.di.pAttrHdr->nAccessListIdx;
    }
Exit1:
    DirItrClose(&pci.di);
    TransEnd(pci.di.tba.pTransInfo);
Exit2:
#if _DEBUG
    ++nDbgCommandCount;
    WFSKrnDbgOutputTest(
        WFSKrnDbgOutputCmdNum;
        WFSKrnDbgOutput5("SetAccessListEntry(\"%s\", f=%x, m=%x, t=%x) ", pAbsPathName->sStr, nFlags, nMask, titleId);
        WFSKrnDbgOutput2("WFS_RESULT_%s\n", WFSKrnGetErrorString(nResult));
        );
#endif
    return (WFSResult)nResult;
#else
    osTPrintf("WFSSrvSetAccessList()\n");
    return WFS_RESULT_NOT_IMPLEMENTED; // dummy
#endif
}

#ifdef _DEBUG
#if (defined(ACL_TEST) || defined(FBA_TEST))
void* WFSSrvDebugGetRootAreaInfo(void)
{
    return &wkg.aDevInfo[0].pVolInfo->rootAreaInfo;
}
#endif
#endif // _DEBUG

