/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirApi.cpp - Top layer of the Directory module


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirApi.cpp,v $
  Revision 1.26  2008/11/05 15:08:20  ueno
  Modified DirItrMapBlks() to roll back the update map in case of error.

  Revision 1.25  2008/10/31 09:51:11  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.24  2008/10/30 04:51:28  kondo_masahiro
  Fixed DirDeleteFile

  Revision 1.23  2008/10/20 00:30:54  nakanose_jin
  (none)

  Revision 1.22  2008/10/17 08:50:48  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.21  2008/10/16 09:30:39  ooizumi
  Fixed definitions for debug build.

  Revision 1.20  2008/10/15 00:14:57  kondo_masahiro
  Fixed a bug about 4GB-1

  Revision 1.19  2008/10/14 12:11:49  kondo_masahiro
  Fixed a bug in DirDeleteFile

  Revision 1.18  2008/10/14 10:26:24  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.17  2008/10/10 02:11:20  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.16  2008/10/09 04:16:53  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.15  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.14  2008/09/28 23:31:51  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.13  2008/09/02 23:34:39  ooizumi
  Added several casting.

  Revision 1.12  2008/09/02 23:02:47  ooizumi
  Fixed a bug failed to treat utf8 as unsigned value.

  Revision 1.11  2008/08/05 03:58:49  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.10  2008/07/28 22:05:32  paul
  Improved updateMap processing. Fixed a bug in DirDeleteSubDir().

  Revision 1.9  2008/07/25 02:51:28  paul
  Changed from nTransIdx to *pTransInfo

  Revision 1.8  2008/07/21 22:39:16  paul
  Fixed a bug in DirItrCheckAndUnmapBlks()
  Changed name from DirItrPinBlkAndUpdatePtrs() -> DirItrPinBlkAndFixPtrs()

  Revision 1.7  2008/07/17 22:47:11  paul
  Added functions to support updateMap

  Revision 1.6  2008/07/09 00:30:20  paul
  Changes to make directory code work with block cache.

  Revision 1.5  2008/06/09 17:21:58  paul
  Added WFSKrnValidateCopyStrAndConvertToLowerCase()

  Revision 1.4  2008/05/12 22:35:37  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.3  2008/05/12 19:11:22  paul
  Moved DirModuleInit here from wfskrn_Dir.cpp.
  Updated DirItrClose to remove pDi from list of open iterators.

  Revision 1.2  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.1  2008/04/25 17:29:16  kondo_masahiro
  Initial check-in

  Revision 1.2  2008/04/23 00:25:23  paul
  Modified some functions to take utf8 strings instead of WFSFileName structs

  Revision 1.1  2008/04/21 18:58:30  paul
  *** empty log message ***
*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "DirApi"

#include "wfskrn_Api.h"
#include "wfskrn_Dir.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_Area.h"
#include "wfskrn_PTree.h"
#include "wfskrn_Handles.h"

DirGlobals dirGlobals;


void DirModuleInit() {
    u32 nI=0;
    for(nI=0; nI<256; nI++) {
        dirGlobals.aToLower[nI] = dirGlobals.aToUpper[nI] = (utf8)nI;
    }
    for(nI='A'; nI<='Z'; nI++) {
        dirGlobals.aToLower[nI] = (utf8)(nI + ('a'-'A'));
    }
    for(nI='a'; nI<='z'; nI++) {
        dirGlobals.aToUpper[nI] = (utf8)(nI - ('a'-'A'));
    }
    DirAllocateMoreNodeStackEntries();
    dirGlobals.pOpenDirItrListAnchor = (DirItr *)((u8*)&dirGlobals.openDirItrListAnchor - (size_t)&((DirItr *)0)->link);
    dirGlobals.openDirItrListAnchor.pNext = dirGlobals.openDirItrListAnchor.pPrev = dirGlobals.pOpenDirItrListAnchor;
    PathCacheInit();    //dirGlobals.nUpdateCtr = 0;
    dirGlobals.updateMapAllocator.pBase = (void*)&wkg.pathCache.pRxtHdr->sbah;
    dirGlobals.updateMapAllocator.fpAllocNode  = (PTreeAllocNodeFunc) PTreeAllocNodeFromSba;
    dirGlobals.updateMapAllocator.fpAllocNodes = (PTreeAllocNodesFunc)PTreeAllocNodesFromSba;
    dirGlobals.updateMapAllocator.fpFreeNode   = (PTreeFreeNodeFunc)  PTreeFreeNodeToSba;
    PTreeInit(&dirGlobals.updateMap, &dirGlobals.updateMapAllocator);
}

void WFSKrnCopyFileName(WFSFileName *pDst, WFSFileName *pSrc) {
    if (pSrc->nLen > WFS_MAX_FILE_NAME_SIZE) {
        pSrc->nLen = WFS_MAX_FILE_NAME_SIZE;
    }
    pDst->nLen = pSrc->nLen;
    strncpy(pDst->sStr, pSrc->sStr, pDst->nLen);
    pDst->sStr[pDst->nLen] = 0;
}

inline void WFSKrnCopyStrAndConvertToLowerCase(utf8 *pDstStr, const utf8 *pSrcStr, u32 nLen) {
    const utf8 *pSrcEnd = pSrcStr + nLen;
    for(; pSrcStr<pSrcEnd; ++pSrcStr, ++pDstStr) {
        *pDstStr = dirGlobals.aToLower[(u32)*(u8*)pSrcStr]; // sometimes VC cannot treat 0x80 over utf8 value.
    }
    *pDstStr = 0;
}

WFSKrnResult WFSKrnValidateCopyStrAndConvertToLowerCase(utf8 *pDstStr, const utf8 *pSrcStr, u32 nLen, u32 nMaxLen) {
    if (nLen > nMaxLen) {
        return WFSKRN_RESULT_INVALID;
    }
    WFSKrnCopyStrAndConvertToLowerCase(pDstStr, pSrcStr, nLen);
    return WFSKRN_RESULT_OK;
}

void WFSKrnCopyFileNameAndConvertToLowerCase(WFSFileName *pDst, WFSFileName *pSrc) {
    if (pSrc->nLen > WFS_MAX_FILE_NAME_SIZE) {
        pSrc->nLen = WFS_MAX_FILE_NAME_SIZE;
    }
    pDst->nLen = pSrc->nLen;
    WFSKrnCopyStrAndConvertToLowerCase(pDst->sStr, pSrc->sStr, pSrc->nLen);
}

void WFSKrnCopyFileNameAndRestoreCase(utf8 *pDst, WFSFileName *pSrc, u8 *pCaseBitArray) {
    u32 nCaseBits = (u32)(256 + *pCaseBitArray);
    if (pSrc->nLen > WFS_MAX_FILE_NAME_SIZE) {
        pSrc->nLen = WFS_MAX_FILE_NAME_SIZE;
    }
    utf8 *pDstStr = pDst;
    utf8 *pSrcStr = pSrc->sStr;
    utf8 *pSrcEnd = pSrcStr + pSrc->nLen;
    for(; pSrcStr<pSrcEnd; ++pSrcStr, ++pDstStr) {
        if (nCaseBits & 1) {
            *pDstStr = dirGlobals.aToUpper[(u32)*(u8*)pSrcStr]; // sometimes VC cannot treat 0x80 over utf8 value.
        } else {
            *pDstStr = *pSrcStr;
        }
        nCaseBits >>= 1;
        if (nCaseBits == 1) {
            nCaseBits = (u32)(256 + *(++pCaseBitArray));
        }
    }
    *pDstStr = 0;
}

void WFSKrnStoreCaseInformationStr(u8 *pCaseBitArray, const utf8 *pStr, u32 nLen) {
    u32 nCaseBit = 1;
    u8 nCaseByte = 0;
    const utf8 *pStrEnd = pStr + nLen;
    for(; pStr<pStrEnd; ++pStr) {
        if (*pStr != dirGlobals.aToLower[(u32)*(u8*)pStr]) { // sometimes VC cannot treat 0x80 over utf8 value.
            nCaseByte |= nCaseBit;
        }
        nCaseBit += nCaseBit;
        if (nCaseBit == 256) {
            *pCaseBitArray++ = nCaseByte;
            nCaseByte = 0;
            nCaseBit = 1;
        }
    }
    if (nCaseBit > 1) {
        *pCaseBitArray = nCaseByte;
    }
}


// Update Counter & Update Map Operations:
//
// BlkAdr:
//  * Blocks must be referred to by absolute block address in the update map.
//
// Interaction with the transaction process:
//  * In transactions, the block to be updated is first re-mapped to a different physical block.
//  * When the update has been commited, it is copied to the logical block address.
//  * For the purpose of update map processing, it should be OK to ignore the transaction process.
//
// When a block is initialized as a directory block:
//  - First check the update map to see if any iterator is already referring to it. If so, copy the initial update count from there.
//    Otherwise, the initial update count is initialized to 0.
//
// When a directory meta-data block gets freed:
//  - If the block is in the update-map, increment its counter, then reduce its reference count, removing it if the reference count is zero.
//
// When a block is accessed by a directory iterator during DirFind or DirFindNext
//  - The iterator copies the update counter of the block to its DirNodeStack.
//
// When a block is to be updated in the course of a executing a WFS command
//  - Update the update counter in the block. //ToDo: it would be better if we only updated the counter when an existing sub-block is moved or deleted.
//  - If the counter exists in the update map, increment it. Otherwise (in debug mode at least), reset the counter to 0.
//
// When an iterator is to be left open by a WFS command:
//  - Increment the reference count of the blocks in the update map, or add them to the update map if neccessary.
//  - Update the counter
//
// When an iterator is to re-used after a gap
//  - The blocks it references should be found in the update map, and checked against the values stored in the iterator.
//  - If necessary, the file attributes should be re-located using DirFind.


static DirNodeStack* MapBlks(DirItr *pDi) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=7019){//8194){
        int a;a=0;
    }
    if (wkg.pathCache.pRxtHdr->sbah.aSubBlkList[0].nNumFree > 16) {
        int a;a=0;
    }
#endif
    // When a new iterator is to be left open, we need to register the address of any blocks referenced in its DirNodeStack,
    // and map each of these to the latest value of that block's update counter.
    // However there are some complications:
    //   The same block could be referenced by multiple iterators, so they need to be reference-counted.
    //   An update operation consists of incrementing the reference count, and overwriting the update counter.
    DirNodeStack *pDns = pDi->dni.pDns;
    WFSBlkAdr nAbsStartBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr;
    WFSBlkAdr nBlkAdr = pDi->tba.nBlkAdr;
    u32 nUpdateCtr = (u32)(pDi->nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
    while(nBlkAdr) {
        WFSBlkAdr nAbsBlkAdr = nAbsStartBlkAdr + nBlkAdr;
        PTreeItr pti;
        u32 nData = DIR_BLK_REF_COUNT_INC | (nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
Retry:;
        WFSKrnResult nResult = PTreeInsert(&dirGlobals.updateMap, &pti, nAbsBlkAdr, nData, &dirGlobals.updateMapAllocator);
        if (nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            pti.pLeaf->aData[pti.aEntryIdx[0]] = nData + (pti.pLeaf->aData[pti.aEntryIdx[0]] & (u32)(-DIR_BLK_REF_COUNT_INC));
        } else if (nResult == WFSKRN_RESULT_PTREE_FULL) {
            nResult = PathCacheDeleteLru();
            if (nResult != WFSKRN_RESULT_OK)
            {
                return pDns;
            }
            goto Retry;
        }
#if _DEBUG
        else if (nResult != WFSKRN_RESULT_OK) {
            WFSKrnOutputErrorCode(nResult);
        }
#endif
        if (pDns==0) {
            return pDns;
        }
        while(pDns->nBlkAdr == nBlkAdr) {
            pDns = pDns->pParent;
            if (pDns==0) {
                return pDns;
            }
        }
        nBlkAdr = pDns->nBlkAdr;
        nUpdateCtr = pDns->nUpdateCtr;
    }

    return NULL;
}


static void UnmapBlks(DirItr* pDi, DirNodeStack* pDns)
{
    PTreeItr pti;
    WFSBlkAdr nAbsBlkAdr;
    WFSKrnResult nResult;
    u32 nUpdateCtr;
    WFSBlkAdr nBlkAdr;
    WFSBlkAdr nAbsStartBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr;
    DirNodeStack* pDnsEnd = pDi->dni.pDns;

    while (pDns != pDnsEnd)
    {
        // [check] must check this rollback routine.
        pDns = pDns->pParent;
        nBlkAdr = pDns->nBlkAdr;
        nUpdateCtr = pDns->nUpdateCtr;

        nAbsBlkAdr = nAbsStartBlkAdr + nBlkAdr;
        nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);

#if _DEBUG
        if (nResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND)
        {
            WFSKrnOutputErrorStr("DirItr block not found in updateMap");
        }
        ASSERT(nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND);
#endif
        // Decrement the reference counter
        pti.pLeaf->aData[pti.aEntryIdx[0]] -= DIR_BLK_REF_COUNT_INC;
        if ((pti.pLeaf->aData[pti.aEntryIdx[0]] & (u32)(-DIR_BLK_REF_COUNT_INC)) == 0)
        {
            // The reference counter is zero, so remove the mapping
            PTreeDelete(&dirGlobals.updateMap, &pti, &dirGlobals.updateMapAllocator);
        }
        while (pDns->nBlkAdr == nBlkAdr)
        {
            pDns = pDns->pParent;
        }
    }
}

WFSKrnResult DirItrMapBlks(DirItr *pDi)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DirNodeStack* pDns;
    pDns = MapBlks(pDi);
    if (pDns) // [check] name.
    {
        // roll back.
        nResult = WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED;
        UnmapBlks(pDi, pDns);
#if _DEBUG
        WFSKrnOutputErrorStr("DirItrMapBlks() : WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED\n");
#endif
    }

    return nResult;
}


void DirItrUnmapBlks(DirItr *pDi) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=2284){
        int a;a=0;
    }
#endif
#if _DEBUG
    if (dirGlobals.updateMap.nRootOfs == 0) {
        WFSKrnOutputErrorStr("updateMap invalid");
    }
#endif
    // Reduce the reference count of the blocks referenced by an iterator, and unmap them if the reference count reaches 0
    DirNodeStack *pDns = pDi->dni.pDns;
    WFSBlkAdr nAbsStartBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr;
    WFSBlkAdr nBlkAdr = pDi->tba.nBlkAdr;
    u32 nUpdateCtr = (u32)(pDi->nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
    while(nBlkAdr) {
        WFSBlkAdr nAbsBlkAdr = nAbsStartBlkAdr + nBlkAdr;
        PTreeItr pti;
        WFSKrnResult nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
        if (nResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            WFSKrnOutputErrorStr("DirItr block not found in updateMap");
        }
#if _DEBUG
        else if (nResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            WFSKrnOutputErrorCode(nResult);
        }
#endif
        // Decrement the reference counter
        pti.pLeaf->aData[pti.aEntryIdx[0]] -= DIR_BLK_REF_COUNT_INC;
        if ((pti.pLeaf->aData[pti.aEntryIdx[0]] & (u32)(-DIR_BLK_REF_COUNT_INC)) == 0) {
            // The reference counter is zero, so remove the mapping
            PTreeDelete(&dirGlobals.updateMap, &pti, &dirGlobals.updateMapAllocator);
        }
        if (pDns==0) {
            return;
        }
        while(pDns->nBlkAdr == nBlkAdr) {
            pDns = pDns->pParent;
            if (pDns==0) {
                return;
            }
        }
        nBlkAdr = pDns->nBlkAdr;
        nUpdateCtr = pDns->nUpdateCtr;
    }
}


WFSKrnResult DirItrCheckMapBlks(DirItr *pDi) {
    // Check the updateCounter of blocks referenced by an iterator.
    DirNodeStack *pDns = pDi->dni.pDns;
    WFSBlkAdr nAbsStartBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr;
    WFSBlkAdr nBlkAdr = pDi->tba.nBlkAdr;
    u32 nUpdateCtr = (u32)(pDi->nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
    while(nBlkAdr) {
        WFSBlkAdr nAbsBlkAdr = nAbsStartBlkAdr + nBlkAdr;
        PTreeItr pti;
        WFSKrnResult nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
        if (nResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            WFSKrnOutputErrorStr("DirItr block not found in updateMap");
        }
#if _DEBUG
        else if (nResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            WFSKrnOutputErrorCode(nResult);
        }
#endif
        if ((pti.nData ^ nUpdateCtr) & WFS_MDF_DIR_COUNTER_MASK) {
            return WFSKRN_RESULT_DIR_ITR_INVALID;
        }
        if (pDns==0) {
            return WFSKRN_RESULT_OK;
        }
        while(pDns->nBlkAdr == nBlkAdr) {
            pDns = pDns->pParent;
            if (pDns==0) {
                return WFSKRN_RESULT_OK;
            }
        }
        nBlkAdr = pDns->nBlkAdr;
        nUpdateCtr = pDns->nUpdateCtr;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult DirItrCheckAndUnmapBlks(DirItr *pDi) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=4749){//5045
        int a;a=0;
    }
#endif
    // Check the updateCounter of blocks referenced by an iterator, and decrement the reference count for all blocks.
    DirNodeStack *pDns = pDi->dni.pDns;
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSBlkAdr nAbsStartBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr;
    WFSBlkAdr nBlkAdr = pDi->tba.nBlkAdr;
    u32 nUpdateCtr = (u32)(pDi->nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
    while(nBlkAdr) {
        WFSBlkAdr nAbsBlkAdr = nAbsStartBlkAdr + nBlkAdr;
        PTreeItr pti;
        WFSKrnResult nTempResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
// #if _DEBUG
        if (nTempResult != WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            WFSKrnOutputErrorStr("DirItr block not found in updateMap");
        }
// #endif
        if ((pti.nData ^ nUpdateCtr) & WFS_MDF_DIR_COUNTER_MASK) {
            nResult = WFSKRN_RESULT_DIR_ITR_INVALID;
        }
        // Decrement the reference counter
        pti.pLeaf->aData[pti.aEntryIdx[0]] -= DIR_BLK_REF_COUNT_INC;
        if ((pti.pLeaf->aData[pti.aEntryIdx[0]] & (u32)(-DIR_BLK_REF_COUNT_INC)) == 0) {
            // The reference counter is zero, so remove the mapping
            PTreeDelete(&dirGlobals.updateMap, &pti, &dirGlobals.updateMapAllocator);
        }
        if (pDns==0) {
            break;
        }
        while(pDns->nBlkAdr == nBlkAdr) {
            pDns = pDns->pParent;
            if (pDns==0) {
                goto Done;
            }
        }
        nBlkAdr = pDns->nBlkAdr;
        nUpdateCtr = pDns->nUpdateCtr;
    }
Done:
    return nResult;
}


WFSKrnResult DirItrClose(DirItr *pDi) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    if (pDi->ppDnsRootParent) {
        nResult = TransUnpinBlk(&pDi->tba);
        //MyOSReport("%d: Close(%p)\n", nDirTest, pDi);
        DirNodeStackFree(&pDi->dni, pDi->ppDnsRootParent);
//#if _DEBUG
        pDi->ppDnsRootParent = 0;
//#endif
        pDi->link.pNext->link.pPrev = pDi->link.pPrev;
        pDi->link.pPrev->link.pNext = pDi->link.pNext;
    }
    return nResult;
}


WFSKrnResult DirItrCloseNoUnpin(DirItr *pDi) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    if (pDi->ppDnsRootParent) {
        //MyOSReport("%d: Close(%p)\n", nDirTest, pDi);
        DirNodeStackFree(&pDi->dni, pDi->ppDnsRootParent);
//#if _DEBUG
        pDi->ppDnsRootParent = 0;
//#endif
        pDi->link.pNext->link.pPrev = pDi->link.pPrev;
        pDi->link.pPrev->link.pNext = pDi->link.pNext;
    }
    return nResult;
}


WFSKrnResult DirItrPinBlkAndFixPtrs(DirItr *pDi, u32 nTransFlags) {
    // Ensure that the block referenced to by the directory iterator is in memory,
    // If not, load it into memory and adjust the pDi->pAttrHdr and pDi->dni.pOfs pointers to point to the new location
    s32 nPtrAdjustment = -(s32)pDi->pBlkHdr;
    DirBlkHdr *pBlkHdr;
    WFSKrnReturnOnError(DirGetBlk(pDi, nTransFlags | BCACHE_FLAG_PINNED, &pBlkHdr));
    nPtrAdjustment += (s32)pBlkHdr;
    pDi->dni.pOfs = (u16*)((u8*)pDi->dni.pOfs + nPtrAdjustment);
    pDi->pAttrHdr = (DirEntryAttrHdr *)((u8*)pDi->pAttrHdr + nPtrAdjustment);
    return WFSKRN_RESULT_OK;
}


WFSKrnResult DirFindName(WFSFileName *pName, DirItr *pDi) {
    WFSKrnCopyFileNameAndConvertToLowerCase(&pDi->name, pName);
    pDi->nBlkDepth = 0;
    WFSKrnResult nResult = DirFind(&pDi->name, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
    switch(nResult) {
        case WFSKRN_RESULT_DIR_ENTRY_FOUND:
            WFSKrnCopyFileNameAndRestoreCase(pName->sStr, &pDi->name, pDi->pAttrHdr->aCaseBitArray);
            pName->nLen = pDi->name.nLen;
            return WFSKRN_RESULT_OK;
        case WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH:
        case WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND:
            return WFSKRN_RESULT_NOT_FOUND;
        default:
            return nResult;
    }
}


WFSKrnResult DirFindRaw(DirItr *pDi) {
    pDi->nBlkDepth = 0;
    WFSKrnResult nResult = DirFind(&pDi->name, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
    switch(nResult) {
        case WFSKRN_RESULT_DIR_ENTRY_FOUND:
            return WFSKRN_RESULT_OK;
        case WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH:
        case WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND:
            return WFSKRN_RESULT_NOT_FOUND;
        default:
            return nResult;
    }
}


WFSKrnResult DirFindFirstNamePrefixMatch(WFSFileName *pName, DirItr *pDi) {
    pDi->nBlkDepth = 0;
    WFSKrnCopyFileNameAndConvertToLowerCase(&pDi->name, pName);
    WFSKrnResult nResult = DirFind(&pDi->name, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_PREFIX_SEARCH);
    switch(nResult) {
        case WFSKRN_RESULT_DIR_ENTRY_FOUND:
        case WFSKRN_RESULT_DIR_NODE_STRING_PREFIX:
        case WFSKRN_RESULT_DIR_CHOICE_PREFIX:
            WFSKrnCopyFileNameAndRestoreCase(pName->sStr, &pDi->name, pDi->pAttrHdr->aCaseBitArray);
            pName->nLen = pDi->name.nLen;
            return WFSKRN_RESULT_OK;
        case WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH:
        case WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND:
            return WFSKRN_RESULT_NOT_FOUND;
        default:
            return nResult;
    }
}


WFSKrnResult DirNextName(WFSFileName *pName, DirItr *pDi) {
    pDi->nBlkDepth = 0;
    //WFSKrnCopyFileNameAndConvertToLowerCase(&pDi->name, pName);
    WFSKrnResult nResult = DirNextRaw(&pDi->name, pDi);
    switch(nResult) {
        case WFSKRN_RESULT_OK:
        case WFSKRN_RESULT_DIR_ENTRY_FOUND:
            WFSKrnCopyFileNameAndRestoreCase(pName->sStr, &pDi->name, pDi->pAttrHdr->aCaseBitArray);
            pName->nLen = pDi->name.nLen;
            return WFSKRN_RESULT_OK;
        default:
            return nResult;
    }
}


WFSKrnResult DirInsertName(WFSFileName *pName, DirItr *pDi, u32 nAllocSize) {
    WFSKrnCopyFileNameAndConvertToLowerCase(&pDi->name, pName);
    WFSKrnReturnOnError(DirInsertRaw(&pDi->name, pDi, nAllocSize));
    WFSKrnStoreCaseInformation(pDi->pAttrHdr->aCaseBitArray, pName);
    // ToDo: Add other file attributes to pDi->pAttrHdr
    return WFSKRN_RESULT_OK;
}


void DirInitFileEntryAttrHdr(DirEntryAttrHdr *pAttrHdr) {
    memset(pAttrHdr, 0, sizeof(DirEntryAttrHdr));
}


void DirInitDirEntryAttrHdr(DirEntryAttrHdr *pAttrHdr) {
    memset(pAttrHdr, 0, sizeof(DirEntryAttrHdr));
    pAttrHdr->nAccessListIdx = WFS_FLAG_IS_A_DIRECTORY;
}


WFSKrnResult DirDeleteFile(DirItr *pDi) {
    DirEntryAttrHdr *pAttrHdr = pDi->pAttrHdr;
    TransBlkAdr tba = pDi->tba;
    switch(pAttrHdr->nSizeCategory){
        case FILE_SIZE_CATEGORY_VERY_SMALL:
            return WFSKRN_RESULT_OK;
        case FILE_SIZE_CATEGORY_SMALL:{
            u32 nNumBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> tba.pAreaInfo->ah.nLog2BlkSize) + 1 : 0;
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
            return TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2BlkSize, nNumBlks, (WFSBlkAdr*)pFileBlkPtr, -(s32)sizeof(WFSFileBlkPtr), tba.pTransInfo);
        }
        case FILE_SIZE_CATEGORY_MEDIUM:{
            u32 nNumMediumBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> tba.pAreaInfo->ah.nLog2MediumBlkSize) + 1 : 0;
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
            return TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2MediumBlkSize, nNumMediumBlks, (WFSBlkAdr*)pFileBlkPtr, -(s32)sizeof(WFSFileBlkPtr), tba.pTransInfo);
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            u32 nNumLargeBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> tba.pAreaInfo->ah.nLog2LargeBlkSize) + 1 : 0;
            WFSFileLargeBlkPtr *pFileBlkLargePtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
            return TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2LargeBlkSize, nNumLargeBlks,
                                  (WFSBlkAdr*)pFileBlkLargePtr, -(s32)sizeof(WFSFileLargeBlkPtr), tba.pTransInfo);
        }
        case FILE_SIZE_CATEGORY_VERY_LARGE: {
            u32 nNumLargeBlks    = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> tba.pAreaInfo->ah.nLog2LargeBlkSize) + 1 : 0;
            u32 nNumIndirectBlks = (nNumLargeBlks+tba.pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/tba.pAreaInfo->nNumLargeBlkPtrsPerBlk;
            WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(tba.pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
            s32 nI;
            for(nI=0;nI<nNumIndirectBlks;nI++){
                u8 *pIndirectBlkPtr;
                WFSKrnReturnOnError(TransGetAndPinBlk5(tba.pAreaInfo, pIndirectBlkAdr[-nI], tba.pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                if(nI+1==nNumIndirectBlks){
                    WFSKrnReturnOnError(TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2LargeBlkSize, (nNumLargeBlks-1)%tba.pAreaInfo->nNumLargeBlkPtrsPerBlk+1,
                                                       (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), tba.pTransInfo));
                } else {
                    WFSKrnReturnOnError(TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2LargeBlkSize, tba.pAreaInfo->nNumLargeBlkPtrsPerBlk,
                                                       (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), tba.pTransInfo));
                }
                WFSKrnReturnOnError(TransUnpinBlk3(tba.pAreaInfo, pIndirectBlkAdr[-nI], tba.pTransInfo));
            }
            return TransFreeBlks6(tba.pAreaInfo, tba.pAreaInfo->ah.nLog2BlkSize, nNumIndirectBlks, pIndirectBlkAdr, -(s32)sizeof(WFSBlkAdr), tba.pTransInfo);
        }
        default:
            return WFSKRN_RESULT_NOT_IMPLEMENTED;
    }
}


WFSKrnResult DirDeleteSubDir(DirItr *pDi) {
    DirEntryAttrHdr *pAttrHdr = pDi->pAttrHdr;
    TransBlkAdr tba = pDi->tba;
    tba.nBlkAdr = pAttrHdr->dir.nSubDirBlkAdr;
    DirBlkHdr *pSubDirBlkHdr;

    WFSKrnReturnOnError(TransGetAndPinBlk(&tba, TRANS_FLAG_READ_BLK, (void**)&pSubDirBlkHdr));
/*
    nResult = TransGetAndPinBlk(&tba, TRANS_FLAG_READ_BLK, (void**)&pSubDirBlkHdr);
    if (nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    */
    if (pSubDirBlkHdr->nNumRecs) {
        TransUnpinBlk(&tba);
        return WFSKRN_RESULT_NOT_EMPTY;
    }
    WFSKrnReturnOnError(TransUnpinAndFreeBlks(pDi->tba.pAreaInfo, pDi->tba.pAreaInfo->ah.nLog2BlkSize, 1, &tba.nBlkAdr, pDi->tba.pTransInfo));
    return WFSKRN_RESULT_OK;
}
