/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirFind.cpp - Functions for finding


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirFind.cpp,v $
  Revision 1.15  2008/10/30 04:49:31  kondo_masahiro
  Deleted the function of DirItrCheckList

  Revision 1.14  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.13  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.12  2008/08/27 09:46:48  paul
  Added pAttrHdr->nAttrLog2Size

  Revision 1.11  2008/08/05 03:58:43  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.10  2008/07/28 22:17:10  paul
  Improved updateMap processing.

  Revision 1.9  2008/07/25 02:51:28  paul
  Changed from nTransIdx to *pTransInfo

  Revision 1.8  2008/07/09 00:37:03  paul
  Changes to make directory code work with block cache.

  Revision 1.7  2008/06/09 17:25:31  paul
  Fixed a bug in DirFindNext()

  Revision 1.6  2008/05/27 18:44:44  paul
  Made utf8 variable definitions more consistent, and use (unsigned) for comparisons

  Revision 1.5  2008/05/17 03:54:59  paul
  Fixed a bug in DirFindNext

  Revision 1.4  2008/05/16 18:42:15  paul
  Added DirFindNext to allow WFSSrvSearchDirectoryNext to work while adding/deleting files in the directory

  Revision 1.3  2008/05/12 19:13:34  paul
  Removed dni local variable from DirFind.
  Made DirFind add pDi to a list of open iterators.

  Revision 1.2  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.1  2008/04/25 17:29:11  kondo_masahiro
  Initial check-in

  Revision 1.2  2008/04/22 23:28:12  paul
  WfsArea* -> Area*

  Revision 1.1  2008/04/21 18:58:31  paul
  *** empty log message ***

*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "DirFind"

#include "wfskrn_SubBlkAlloc.h"

#include "wfskrn_Dir.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_DirRxTree.h"

#undef dbg
#if _DEBUG_DIR
    #define dbg(s) s
#else
    #define dbg(s)
#endif


WFSKrnResult DirFind(WFSFileName *pName, DirItr *pDi, u32 nMaxBlkDepth, u32 nMode) {
    // This function searches for a particular named entry in a directory up to a given max depth.
    // If the name is found, its location within the appropriate leaf block is returned in pDi, and the function returns WFSKRN_RESULT_DIR_ENTRY_FOUND.
    // If the name is not found, the location where it should be inserted within the appropriate leaf block, is returned in pDi, and the function returns ....
    // In addition, if the input name is a prefix of one or more entries in the directory, then the characters beyond the end of the input string
    // will be filled with the characters of the first matching entry from the directory. This can be used to help implement pattern searches,
    // or just to get the first entry in a directory (by searching for "").
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=6979){
        int a;a=0;
    }
    if (nDbgCommandCount==2850){
        int a;a=0;
    }
#endif
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    DirBlkHdr *pBlkHdr;
    WFSKrnReturnOnError(DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr));
    DirNodeItr dniTemp;
    utf8 *pStrEnd = pName->sStr + pName->nLen;
    u32 nStrIdx, nPrefixStrIdx;
    utf8 *pChoiceStrPtr = pName->sStr; // To keep track of how many characters are common between adjacent splitters when we move to a child block, so that it can implement prefix suppression.
    utf8 cSubBlkChar, cSearchChar=pChoiceStrPtr[0];
    u16 *pOfsTblEnd;
    u32 nSubBlkStrLen;
    WFSKrnResult nResult = WFSKRN_RESULT_OK;

#if _DEBUG
    if ((pDi->dni.nDnsDepth == 0) && (pDi->ppDnsRootParent != 0)) {
        WFSKrnOutputErrorStr("Directory Iterator already opened");
    }
    // Check pDi is not already on the open iterator list
    DirItr *pListDi = dirGlobals.openDirItrListAnchor.pNext;
    while(pListDi != dirGlobals.pOpenDirItrListAnchor) {
        if (pListDi == pDi) {
            WFSKrnOutputErrorStr("pDi already on open iterator list");
        }
        if (pListDi->link.pNext->link.pPrev != pListDi) {
            WFSKrnOutputErrorStr("pDi link error");
        }
        pListDi = pListDi->link.pNext;
    }
#endif
    //MyOSReport("%d: Open(%p)\n", nDirTest, pDi);
    pDi->link.pNext = dirGlobals.openDirItrListAnchor.pNext;
    pDi->link.pPrev = dirGlobals.pOpenDirItrListAnchor;
    dirGlobals.openDirItrListAnchor.pNext->link.pPrev = pDi;
    dirGlobals.openDirItrListAnchor.pNext = pDi;

    pDi->dni.nDnsDepth = 0;
    WFSKrnReturnOnError(DirNodeStackGetEntry(&pDi->dni));
    pDi->ppDnsRootParent = &pDi->dni.pDns->pParent;

    pDi->dni.pDns->nBlkAdr = 0; //tba.nBlkAdr;
    pDi->dni.pDns->nSubBlkOfs = pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(pDi->dni.pDns->nSubBlkOfs));
    pDi->dni.pDns->nEntryIdx = 1;
    nPrefixStrIdx=0;

    while(1) {
        nStrIdx = nPrefixStrIdx-1;
#if DIR_PREFIX_SUPPRESSION
        pDi->dni.pStrPtr = pChoiceStrPtr; // This should skip any common prefix shared by all file name key fields or splitters in the block
#else
        pDi->dni.pStrPtr = pName->sStr;
#endif
        pDi->dni.pOfs = &pBlkHdr->nRootOfs;
        dbg(SbaCheckOfs(*pDi->dni.pOfs));
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
        if ((pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) || (pDi->nBlkDepth >= nMaxBlkDepth)) {
            break;
        }
        ++pDi->nBlkDepth;
        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        while(1) {
            nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
            nStrIdx += nSubBlkStrLen + 1;
            pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
            if ((pStrEnd - pDi->dni.pStrPtr) < nSubBlkStrLen) {
                do {
                    cSearchChar = *pDi->dni.pStrPtr++;
                    cSubBlkChar = *pDi->dni.pNodeStrPtr++;
                } while(cSearchChar == cSubBlkChar);
                if ((unsigned)cSearchChar < (unsigned)cSubBlkChar) {
                    goto BackTrack;
                }
                goto ChooseMaxDescendent;
            }
            while(nSubBlkStrLen--) {
                cSearchChar = *pDi->dni.pStrPtr++;
                cSubBlkChar = *pDi->dni.pNodeStrPtr++;
                if (cSearchChar != cSubBlkChar) {
                    if ((unsigned)cSearchChar < (unsigned)cSubBlkChar) {
                        goto BackTrack;
                    }
                    goto ChooseMaxDescendent;
                }
            }
            // matched all the characters from the prefix. pDi->dni.pNodeStrPtr should now be pointing to the array of choices following the prefix.
            cSearchChar = *pDi->dni.pStrPtr;
            pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));

            WFSBool bFound = DirBinarySplitSearchChoices(pDi->dni.pNodeStrPtr, pDi->dni.pSubBlkHdr->nNumEntries, cSearchChar, &pDi->dni.nEntryIdx);
#if DIR_PREFIX_SUPPRESSION
            if ((pDi->dni.nEntryIdx>=1) && (pDi->dni.nEntryIdx < pDi->dni.pSubBlkHdr->nNumEntries)) {
                // ToDo: Implement prefix suppression
                nPrefixStrIdx = nStrIdx;
            }
#endif
            WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
            pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
            if (bFound) {
                // Found the search character.
                pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
                if (cSearchChar==0) {
                    // We matched a 0 .. means the file name matched one of the splitters exactly
                    // This may still be  a valid file name.
                    // Note: A splitter can be arbitrary. It does not even have to be a prefix of an existing file name. 
                    //       (This can happen after deleting the filename which was originally chosen as a split point).
                    break;
                }
                if (pDi->dni.pSubBlkHdr->nNumEntries>1) {
                    if (pDi->dni.nEntryIdx>1) {
                        // Update the previous record information
                        dniTemp = pDi->dni;
                        --dniTemp.nEntryIdx;
                    }
                    pChoiceStrPtr = pDi->dni.pStrPtr;
                }
                pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_ote((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pOfsTblEnd);
                dbg(SbaCheckOfs(*pDi->dni.pOfs));
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                ++pDi->dni.pStrPtr;
                continue;
            }

            if (pDi->dni.nEntryIdx==0) {
                // The binary split search determined that the search character was smaller than the smallest choice
BackTrack:
                while(pDi->dni.pDns != dniTemp.pDns) {
                    DirPopNodeStackEntry(&pDi->dni);
                };
                pDi->dni = dniTemp;
                pDi->dni.pDns->nEntryIdx = pDi->dni.nEntryIdx;
                nStrIdx = (u32)(pDi->dni.pDns->nStrIdx);// + pDi->dni.pSubBlkHdr->nStrLen + 1);
                pChoiceStrPtr = pDi->dni.pStrPtr;
                pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
                pDi->dni.pNodeStrPtr = &pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen];
            }
            pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
            if ((pDi->dni.nEntryIdx==1)&&(pDi->dni.pNodeStrPtr[0]==0)) {
                break;
            }
            // Offset is an intra block offset
            pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_ote((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pOfsTblEnd);
            dbg(SbaCheckOfs(*pDi->dni.pOfs));
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            nStrIdx += pDi->dni.pSubBlkHdr->nStrLen + 1;
            pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
ChooseMaxDescendent:
            // Follow intra block offsets until max splitter end is encountered
            if (pDi->dni.pSubBlkHdr->nNumEntries > 1) {
                while(1) {
                    pDi->dni.nEntryIdx = pDi->dni.pSubBlkHdr->nNumEntries;
                    WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
                    pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
                    nStrIdx += pDi->dni.pSubBlkHdr->nStrLen;
                    pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
                    if (pDi->dni.pSubBlkHdr->nNumEntries == 1) {
                        break;
                    }
                    ++nStrIdx;
                    pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                    dbg(SbaCheckOfs(*pDi->dni.pOfs));
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                };
                pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
                break;
            }
            // we should now be at the end of the line
            pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
            pDi->dni.nEntryIdx = 1;
            WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
            pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
            pDi->dni.pDns->nStrIdx = nStrIdx;
            break;
        }
        if (nMode == DIR_FIND_MODE_NODE_SEARCH) {
            if (pDi->nBlkDepth >= nMaxBlkDepth) {
                goto Exit;
            }
        }
        {
            u32 nNewBlkAdr = DirNodeGetBlkAdr_ote(pDi->dni.pSubBlkHdr, pOfsTblEnd);
            TransUnpinBlk(&pDi->tba);
            pDi->tba.nBlkAdr = nNewBlkAdr;
            WFSKrnExitOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
        }
    }


#if (DIR_PREFIX_SUPPRESSION==0)
    // ToDo: Prefix suppression not yet implemented
    pDi->dni.pStrPtr = pName->sStr;
#endif

    // Search the leaf node
    if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    } else {
        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    }
    while(1) {
        nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen + 1;
        pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
        if ((pStrEnd - pDi->dni.pStrPtr) < nSubBlkStrLen) {
            while(*pDi->dni.pStrPtr == *pDi->dni.pNodeStrPtr) {
                ++pDi->dni.pStrPtr; ++pDi->dni.pNodeStrPtr;
            }
            nSubBlkStrLen -= (pDi->dni.pNodeStrPtr - pDi->dni.pSubBlkHdr->aFirstChar);
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        while((nSubBlkStrLen) && (*pDi->dni.pStrPtr == *pDi->dni.pNodeStrPtr)) {
            ++pDi->dni.pStrPtr; ++pDi->dni.pNodeStrPtr; --nSubBlkStrLen;
        }
        if (nSubBlkStrLen) {
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        // matched all the characters from the prefix. pDi->dni.pNodeStrPtr should now be pointing to the array of choices following the prefix.
        cSearchChar = *pDi->dni.pStrPtr;
        if (!DirBinarySplitSearchChoices(pDi->dni.pNodeStrPtr, pDi->dni.pSubBlkHdr->nNumEntries, cSearchChar, &pDi->dni.nEntryIdx)) {
            nResult = WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND;
            break;
        }
        WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
        pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
        pDi->dni.pDns->nStrIdx = nStrIdx;
        // Found the search character.
        if (cSearchChar==0) {
            // We matched a 0 .. which means we have found the filename
            nResult = WFSKRN_RESULT_DIR_ENTRY_FOUND;
            break;
        }
        if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
            pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
            dbg(SbaCheckOfs(*pDi->dni.pOfs));
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        } else {
            pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
            dbg(SbaCheckOfs(*pDi->dni.pOfs));
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        }
        ++pDi->dni.pStrPtr;
    }

    // This is the location passed back to the caller
    pDi->pBlkHdr = pBlkHdr;

    if ((*pDi->dni.pStrPtr) || (pBlkHdr->nNumRecs==0) || (!(pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK))) {
        goto Done;
    }

    if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
        goto GetAttr;
    }

    if (nMode != DIR_FIND_MODE_PREFIX_SEARCH) {
        goto Done;
    }

    if (nResult == WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH) {
        nResult = WFSKRN_RESULT_DIR_NODE_STRING_PREFIX;
    } else {
        nResult = WFSKRN_RESULT_DIR_CHOICE_PREFIX;
    }

    // The exact name was not found, but it was a prefix of one or more exising names.
    // We continue traversing to find the left-most descendent.
    // This is useful for iterating through the directory, and wildcard pattern search.
    pDi->dni.nEntryIdx = 1;
    while(1) {
        while(nSubBlkStrLen--) {
            *pDi->dni.pStrPtr++ = *pDi->dni.pNodeStrPtr++;
        }
        //if (pDi->dni.pSubBlkHdr->nNumEntries > 1) {
            WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
            pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
            pDi->dni.pDns->nStrIdx = nStrIdx;
        //}
        cSubBlkChar = pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen];
        *pDi->dni.pStrPtr++ = cSubBlkChar;
        if (cSubBlkChar==0) {
            //pDi->dni = pDi->dni;
            pName->nLen = nStrIdx;
            goto GetAttr;
        }
        pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
        pDi->dni.pOfs = &pOfsTblEnd[-1];
        dbg(SbaCheckOfs(*pDi->dni.pOfs));
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen + 1;
        pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
    }

GetAttr:
    pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s(1, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
    pDi->pAttrHdr = (DirEntryAttrHdr *)((u8*)pDi->pBlkHdr + *pDi->dni.pOfs);
    pDi->dni.nLog2Size = pDi->pAttrHdr->nAttrLog2Size;
Done:
    return nResult;
Exit:
    pDi->pBlkHdr = pBlkHdr;
    return nResult;
}


WFSKrnResult DirFindNext(WFSFileName *pName, DirItr *pDi) {
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    DirBlkHdr *pBlkHdr;
    WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
    DirNodeItr dniPrev, dniNext;
    utf8 *pStrEnd = pName->sStr + pName->nLen;
    u32 nStrIdx, nPrefixStrIdx;
    utf8 *pChoiceStrPtr = pName->sStr; // To keep track of how many characters are common between adjacent splitters when we move to a child block, so that it can implement prefix suppression.
    utf8 cSubBlkChar, cSearchChar=pChoiceStrPtr[0];
    u16 *pOfsTblEnd;
    u32 nSubBlkStrLen;
    WFSKrnResult nResult = WFSKRN_RESULT_OK;

#if _DEBUG
    if ((pDi->dni.nDnsDepth == 0) && (pDi->ppDnsRootParent != 0)) {
        WFSKrnOutputErrorStr("Directory Iterator already opened");
    }
    // Check pDi is not already on the open iterator list
    DirItr *pListDi = dirGlobals.openDirItrListAnchor.pNext;
    while(pListDi != dirGlobals.pOpenDirItrListAnchor) {
        if (pListDi == pDi) {
            WFSKrnOutputErrorStr("pDi already on open iterator list");
        }
        if (pListDi->link.pNext->link.pPrev != pListDi) {
            WFSKrnOutputErrorStr("pDi link error");
        }
        pListDi = pListDi->link.pNext;
    }
#endif
    //MyOSReport("%d: Open(%p)\n", nDirTest, pDi);
    pDi->link.pNext = dirGlobals.openDirItrListAnchor.pNext;
    pDi->link.pPrev = dirGlobals.pOpenDirItrListAnchor;
    dirGlobals.openDirItrListAnchor.pNext->link.pPrev = pDi;
    dirGlobals.openDirItrListAnchor.pNext = pDi;

    pDi->dni.nDnsDepth = 0;
    WFSKrnReturnOnError(DirNodeStackGetEntry(&pDi->dni));
    pDi->ppDnsRootParent = &pDi->dni.pDns->pParent;

    pDi->dni.pDns->nBlkAdr = 0; //tba.nBlkAdr;
    pDi->dni.pDns->nSubBlkOfs = pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(pDi->dni.pDns->nSubBlkOfs));
    pDi->dni.pDns->nEntryIdx = 1;
    nPrefixStrIdx = 0;
    dniNext.nEntryIdx = 0;

    while(1) {
        nStrIdx = nPrefixStrIdx-1;
#if DIR_PREFIX_SUPPRESSION
        pDi->dni.pStrPtr = pChoiceStrPtr; // This should skip any common prefix shared by all file name key fields or splitters in the block
#else
        pDi->dni.pStrPtr = pName->sStr;
#endif
        pDi->dni.pOfs = &pBlkHdr->nRootOfs;
        dbg(SbaCheckOfs(*pDi->dni.pOfs));
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
        if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
            break;
        }
#if _DEBUG_DIR
    /*if (nDirTest == nDirTestBreakPoint) {
        nDirTest = nDirTestBreakPoint;
        pDi->pBlkHdr = pBlkHdr;
        DirNodeCheckBlk(pDi);
    }*/
#endif
        ++pDi->nBlkDepth;
        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        while(1) {
            nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
            nStrIdx += nSubBlkStrLen + 1;
            pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
            if ((pStrEnd - pDi->dni.pStrPtr) < nSubBlkStrLen) {
                do {
                    cSearchChar = *pDi->dni.pStrPtr++;
                    cSubBlkChar = *pDi->dni.pNodeStrPtr++;
                } while(cSearchChar == cSubBlkChar);
                if ((unsigned)cSearchChar < (unsigned)cSubBlkChar) {
                    goto BackTrackNodePrev;
                }
                goto ChooseMaxDescendent;
            }
            while(nSubBlkStrLen--) {
                cSearchChar = *pDi->dni.pStrPtr++;
                cSubBlkChar = *pDi->dni.pNodeStrPtr++;
                if (cSearchChar != cSubBlkChar) {
                    if ((unsigned)cSearchChar < (unsigned)cSubBlkChar) {
                        goto BackTrackNodePrev;
                    }
                    goto ChooseMaxDescendent;
                }
            }
            // matched all the characters from the prefix. pDi->dni.pNodeStrPtr should now be pointing to the array of choices following the prefix.
            cSearchChar = *pDi->dni.pStrPtr;
            pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));

            WFSBool bFound = DirBinarySplitSearchChoices(pDi->dni.pNodeStrPtr, pDi->dni.pSubBlkHdr->nNumEntries, cSearchChar, &pDi->dni.nEntryIdx);
#if DIR_PREFIX_SUPPRESSION
            if ((pDi->dni.nEntryIdx>=1) && (pDi->dni.nEntryIdx < pDi->dni.pSubBlkHdr->nNumEntries)) {
                // ToDo: Implement prefix suppression
                nPrefixStrIdx = nStrIdx;
            }
#endif
            WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
            pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
            if (pDi->dni.nEntryIdx < pDi->dni.pSubBlkHdr->nNumEntries) {
                dniNext = pDi->dni;
            }

            if (bFound) {
                // Found the search character.
                pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
                if (cSearchChar==0) {
                    // We matched a 0 .. means the file name matched one of the splitters exactly
                    // This may still be  a valid file name.
                    // Note: A splitter can be arbitrary. It does not even have to be a prefix of an existing file name. 
                    //       (This can happen after deleting the filename which was originally chosen as a split point).
                    break;
                }
                if (pDi->dni.pSubBlkHdr->nNumEntries>1) {
                    if (pDi->dni.nEntryIdx>1) {
                        // Update the previous record information
                        dniPrev = pDi->dni;
                    }
                    pChoiceStrPtr = pDi->dni.pStrPtr;
                }
                pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_ote((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pOfsTblEnd);
                dbg(SbaCheckOfs(*pDi->dni.pOfs));
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                ++pDi->dni.pStrPtr;
                continue;
            }

            if (pDi->dni.nEntryIdx==0) {
                // The binary split search determined that the search character was smaller than the smallest choice
BackTrackNodePrev:
                while(pDi->dni.pDns != dniPrev.pDns) {
                    DirPopNodeStackEntry(&pDi->dni);
                };
                pDi->dni = dniPrev;
                pDi->dni.pDns->nEntryIdx = --pDi->dni.nEntryIdx;
                dniNext = pDi->dni;
                nStrIdx = (u32)(pDi->dni.pDns->nStrIdx);// + pDi->dni.pSubBlkHdr->nStrLen + 1);
                pChoiceStrPtr = pDi->dni.pStrPtr;
                pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
                pDi->dni.pNodeStrPtr = &pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen];
            }
            pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
            if ((pDi->dni.nEntryIdx==1)&&(pDi->dni.pNodeStrPtr[0]==0)) {
                break;
            }
            // Offset is an intra block offset
            pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_ote((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pOfsTblEnd);
            dbg(SbaCheckOfs(*pDi->dni.pOfs));
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            nStrIdx += pDi->dni.pSubBlkHdr->nStrLen + 1;
            pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
ChooseMaxDescendent:
            // Follow intra block offsets until max splitter end is encountered
            if (pDi->dni.pSubBlkHdr->nNumEntries > 1) {
                while(1) {
                    pDi->dni.nEntryIdx = pDi->dni.pSubBlkHdr->nNumEntries;
                    WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
                    pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
                    nStrIdx += pDi->dni.pSubBlkHdr->nStrLen;
                    pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
                    if (pDi->dni.pSubBlkHdr->nNumEntries == 1) {
                        break;
                    }
                    ++nStrIdx;
                    pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                    dbg(SbaCheckOfs(*pDi->dni.pOfs));
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                };
                pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
                break;
            }
            // we should now be at the end of the line
            pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
            pDi->dni.nEntryIdx = 1;
            WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
            pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
            pDi->dni.pDns->nStrIdx = nStrIdx;
            break;
        }
        {
            u32 nNewBlkAdr = DirNodeGetBlkAdr_ote(pDi->dni.pSubBlkHdr, pOfsTblEnd);
            TransUnpinBlk(&pDi->tba);
            pDi->tba.nBlkAdr = nNewBlkAdr;
            WFSKrnExitOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
        }
    }


#if (DIR_PREFIX_SUPPRESSION==0)
    // ToDo: Prefix suppression not yet implemented
    pDi->dni.pStrPtr = pName->sStr;
#endif

#if _DEBUG_DIR
    /*if (nDirTest == nDirTestBreakPoint) {
        pDi->pBlkHdr = pBlkHdr;
        DirLeafCheckBlk(pDi);
    }*/
#endif

    // Search the leaf node
    pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    while(1) {
        nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen + 1;
        pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
        if ((pStrEnd - pDi->dni.pStrPtr) < nSubBlkStrLen) {
            while(*pDi->dni.pStrPtr == *pDi->dni.pNodeStrPtr) {
                ++pDi->dni.pStrPtr; ++pDi->dni.pNodeStrPtr;
            }
            nSubBlkStrLen -= (pDi->dni.pNodeStrPtr - pDi->dni.pSubBlkHdr->aFirstChar);
        } else {
            while((nSubBlkStrLen) && (*pDi->dni.pStrPtr == *pDi->dni.pNodeStrPtr)) {
                ++pDi->dni.pStrPtr; ++pDi->dni.pNodeStrPtr; --nSubBlkStrLen;
            }
        }
        if (nSubBlkStrLen) {
            if (*pDi->dni.pStrPtr > *pDi->dni.pNodeStrPtr) {
                goto BackTrack;
            }
            pDi->dni.nEntryIdx = 1;
            goto ChooseMinDescendentLeaf;
        }
        // matched all the characters from the prefix. pDi->dni.pNodeStrPtr should now be pointing to the array of choices following the prefix.
        WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
        pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
        pDi->dni.pDns->nStrIdx = nStrIdx;
        cSearchChar = *pDi->dni.pStrPtr;
        if (DirBinarySplitSearchChoices(pDi->dni.pNodeStrPtr, pDi->dni.pSubBlkHdr->nNumEntries, cSearchChar, &pDi->dni.nEntryIdx)) {
            pDi->dni.pDns->nEntryIdx = pDi->dni.nEntryIdx;
            if (pDi->dni.nEntryIdx < pDi->dni.pSubBlkHdr->nNumEntries) {
                dniNext = pDi->dni;
            }
        } else {
            //*pDi->dni.pStrPtr++ = pDi->dni.pNodeStrPtr[pDi->dni.nEntryIdx];
            pDi->dni.pNodeStrPtr += pDi->dni.nEntryIdx;
            ++pDi->dni.nEntryIdx;
            if (pDi->dni.nEntryIdx > pDi->dni.pSubBlkHdr->nNumEntries) {
                goto BackTrack;
            }
            goto ChooseMinDescendentLeaf;
        }
        // Found the search character.
        if (cSearchChar==0) {
            // We matched a 0 .. which means we have found the filename
            // However an exact match is not what we are looking for. We want the next entry.
BackTrack:
            if (dniNext.nEntryIdx == 0) {
                nResult = WFSKRN_RESULT_NOT_FOUND;
                goto Exit;
            }
            nResult = WFSKRN_RESULT_DIR_ENTRY_FOUND;
            //WFSBlkAdr nLeafBlkAdr = pDi->dni.pDns->nBlkAdr;
            while(pDi->dni.pDns != dniNext.pDns) {
                DirPopNodeStackEntry(&pDi->dni);
            }
            pDi->dni = dniNext;
            if (pDi->tba.nBlkAdr != dniNext.pDns->nBlkAdr) {
                // We have back tracked to a previous node
                TransUnpinBlk(&pDi->tba);
                pDi->tba.nBlkAdr = dniNext.pDns->nBlkAdr;
                WFSKrnExitOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
                ++pDi->dni.pDns->nEntryIdx;
                pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr(pDi->dni.pDns->nEntryIdx, pDi->dni.pSubBlkHdr);
                dbg(SbaCheckOfs(*pDi->dni.pOfs));
                pDi->dni.nEntryIdx=1;
                do {
                    while(1) {
                        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                        WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
                        pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
                        if (pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] == 0) {
                            break;
                        }
                        pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr(pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr);
                        dbg(SbaCheckOfs(*pDi->dni.pOfs));
                    }
                    {
                        WFSBlkAdr nNewBlkAdr = DirNodeGetBlkAdr(pDi->dni.pSubBlkHdr);
                        TransUnpinBlk(&pDi->tba);
                        pDi->tba.nBlkAdr = nNewBlkAdr;
                        WFSKrnExitOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
                    }
                    DirCheckBlk(pDi);
#if _DEBUG_DIR
                    /*if (nDirTest == nDirTestBreakPoint) {
                        pDi->pBlkHdr = pBlkHdr;
                        DirCheckBlk(pDi);
                    }*/
#endif
                    pDi->dni.pOfs = &pBlkHdr->nRootOfs;
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    pDi->dni.nEntryIdx = 1;
                } while(!(pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK));
                nStrIdx = (u32)-1;
                pDi->dni.pStrPtr = pName->sStr;
            } else {
                nStrIdx = dniNext.pDns->nStrIdx;
                *pDi->dni.pStrPtr++ = pDi->dni.pNodeStrPtr[dniNext.nEntryIdx];
                ++dniNext.nEntryIdx;
                pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr(dniNext.nEntryIdx, pDi->dni.pSubBlkHdr);
                dbg(SbaCheckOfs(*pDi->dni.pOfs));
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            }
            nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen;
            nStrIdx += nSubBlkStrLen + 1;
            pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
            pDi->dni.nEntryIdx = 1;

ChooseMinDescendentLeaf:
            ++nSubBlkStrLen;
            while(1) {
                memcpy(pDi->dni.pStrPtr, pDi->dni.pNodeStrPtr, nSubBlkStrLen);
                pDi->dni.pStrPtr += nSubBlkStrLen;
                if(pDi->dni.pDns->nBlkAdr != pDi->tba.nBlkAdr || pDi->dni.pDns->nSubBlkOfs != *pDi->dni.pOfs){
                    WFSKrnReturnOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
                } else {
                    pDi->dni.pDns->nEntryIdx  = (u8)pDi->dni.nEntryIdx;
                }
                //WFSKrnExitOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
                pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;
                pDi->dni.pDns->nStrIdx = (u8)nStrIdx;
                pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr(pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr);
                dbg(SbaCheckOfs(*pDi->dni.pOfs));
                if (pDi->dni.pNodeStrPtr[nSubBlkStrLen-1] == 0) {
                    break;
                }
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                pDi->dni.pNodeStrPtr = pDi->dni.pSubBlkHdr->aFirstChar;
                nSubBlkStrLen = pDi->dni.pSubBlkHdr->nStrLen + 1;
                nStrIdx += nSubBlkStrLen;
                pDi->dni.nEntryIdx = 1;
            };
            break;
        }
        pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
        dbg(SbaCheckOfs(*pDi->dni.pOfs));
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        ++pDi->dni.pStrPtr;
    }
    pDi->pAttrHdr = (DirEntryAttrHdr*)((u8*)pBlkHdr + *pDi->dni.pOfs);
    pDi->name.nLen = nStrIdx;
Exit:
    pDi->pBlkHdr = pBlkHdr;
    return nResult;
}
