/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirNodeStack.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirNodeStack.cpp,v $
  Revision 1.8  2008/10/16 09:30:39  ooizumi
  Fixed definitions for debug build.

  Revision 1.7  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.6  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.5  2008/08/05 03:58:31  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.4  2008/07/09 00:37:52  paul
  Changes to make directory code work with block cache.

  Revision 1.3  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:29:50  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.10  2008/04/23 01:56:07  paul
  Fixed bCheckDirNodeStack to be visible from wfskrn_Dir.cpp

  Revision 1.9  2008/04/22 22:12:40  paul
  Changed WFSArea* to Area*

  Revision 1.8  2008/04/19 05:50:51  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.7  2008/04/15 18:45:09  paul
  Added some debug code to output calling locations for NodeStack functions

  Revision 1.6  2008/04/04 22:53:33  paul
  Changed debug triggers to use bCheckDirNodeStack

  Revision 1.5  2008/04/04 21:16:33  paul
  Fixed loop-detecting code

  Revision 1.4  2008/04/03 23:38:31  kondo_masahiro
  change name of debug parameter

  Revision 1.3  2008/03/05 11:02:13  paul
  Changed stack checking functions to not allocate memory

  Revision 1.2  2008/02/28 06:31:49  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.1  2008/02/27 09:46:53  paul
  *** empty log message ***

  Revision 1.2  2008/02/27 00:38:44  paul
  Updated debug code

  Revision 1.1  2008/02/21 04:35:34  paul
  *** empty log message ***


 *---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "DirNS"

#include "wfskrn_DirNodeStack.h"
#include "wfskrn_Mutex.h"
#include "wfskrn_Api.h"


#if _DEBUG_DIR_NODE_STACK
    #define dbg_check(s) s
    #define dbg_output(s)
#else
    #define dbg_check(s)
    #define dbg_output(s)
#endif


#if _DEBUG_DIR_NODE_STACK
    bool bCheckDirNodeStack = false;  // extern
#endif

typedef struct {
    WFSKrnMutex mxMutex;
    dbg_check(u32 nTotalAllocated;)
    dbg_check(u32 nNumUsed;)
    u32 nNumFreeDnsPoolEntries;
    DirNodeStack *pDnsPoolFreeList;
} DirNodeStackGlobals;

DirNodeStackGlobals dnsg;

#if _DEBUG_DIR
    void DebugCheckNumFreeDnsPoolEntries(char *str) {
        static u32 nPrev=0;
        char *s; s = str; // for warning
        if (nPrev != dnsg.nNumFreeDnsPoolEntries) {
            nPrev = dnsg.nNumFreeDnsPoolEntries;
            dbg_output(MyOSReport("%s: dnsg.nNumFreeDnsPoolEntries=%d\n", str, dnsg.nNumFreeDnsPoolEntries));
            if ((dnsg.nNumFreeDnsPoolEntries!=WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT)&&
                (dnsg.nNumFreeDnsPoolEntries!=(WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT*2))) {
                //int a=0; // for warning
            }
        }
    }
#endif //_DEBUG_DIR


#if _DEBUG_DIR_NODE_STACK

    DirNodeStack *pDebugLastNodeStackAlloc;
    u32 nDebugCountNodeStackCheckAlloc = 0;
    u32 nDebugCountNodeStackCheckFreeList = 0;

    static
    void DirNodeStackCheckAlloc(DirNodeItr *pDni) {
        nDebugCountNodeStackCheckAlloc++;
        dbg_output(MyOSReport("%d. ", nDebugCountNodeStackCheckAlloc));
        DirNodeStack *pDns = pDni->pDns;
        u32 nI;
        for(nI=pDni->nDnsDepth; --nI; ) {
            //s32 nIdx = pDns-pDebugLastNodeStackAlloc; // unused
            dbg_output(MyOSReport("%d,", nIdx));
            if ((u32)pDns==0xbaadf00d) {
                WFSKrnOutputErrorStr("DirNodeStack Error! Bad entry!\n");
            }
            DirNodeStack *pDns2 = pDns;
            u32 nJ;
            for(nJ=nI; --nJ; ) {
                pDns2 = pDns2->pParent;
                if (pDns2 == pDns) {
                    WFSKrnOutputErrorStr("DirNodeStack Error! Same entry appears twice!\n");
                }
            }
            //u32 nOfs = pDns->nSubBlkOfs; // unused
            //u32 nBlkAdr = pDns->nBlkAdr;
            pDns = pDns->pParent;
            /*if ((nI>1) && (pDns->nSubBlkOfs == nOfs) && (pDns->nBlkAdr == nBlkAdr)) {
                WFSKrnOutputErrorStr("DirNodeStack Error! Same offset repeats!\n");
            }*/
        }
        dbg_output(MyOSReport(": "));
    }

    void DirNodeStackCheckFreeList() {
        nDebugCountNodeStackCheckFreeList++;
        DirNodeStack *pDns = dnsg.pDnsPoolFreeList;
        u32 nI;
        for(nI=0; nI<dnsg.nNumFreeDnsPoolEntries; nI++) {
            //s32 nIdx = pDns-pDebugLastNodeStackAlloc; // unused
            dbg_output(MyOSReport("%d,", nIdx));
            if ((u32)pDns==0xbaadf00d) {
                WFSKrnOutputErrorStr("DirNodeStack Error! Bad entry on free list!\n");
            }
            DirNodeStack *pDns2 = dnsg.pDnsPoolFreeList;
            u32 nJ;
            for(nJ=0; nJ<nI; nJ++) {
                if (pDns2 == pDns) {
                    WFSKrnOutputErrorStr("DirNodeStack Error! Same entry appears twice on free list!\n");
                }
                pDns2 = pDns2->pParent;
            }
            pDns = pDns->pParent;
        }
        dbg_output(MyOSReport("\n"));
        //WFSKrnHeapFree(&wkg.heap, ppDns);
    }

#endif //_DEBUG_DIR_NODE_STACK


WFSKrnResult DirAllocateMoreNodeStackEntries() {
    dnsg.pDnsPoolFreeList = (DirNodeStack *)WFSKrnHeapAlloc(&wkg.heap, WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT * sizeof(DirNodeStack));
    dbg_check(pDebugLastNodeStackAlloc = dnsg.pDnsPoolFreeList);
    if (dnsg.pDnsPoolFreeList == 0) {
        return WFSKRN_RESULT_OUT_OF_MEMORY;
    }
    u32 nI;
    for(nI=0; nI<WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT-1; nI++) {
        dnsg.pDnsPoolFreeList[nI].pParent = &dnsg.pDnsPoolFreeList[nI+1];
    }
    dbg_check(dnsg.nTotalAllocated += WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT);
    dnsg.nNumFreeDnsPoolEntries += WFS_DIR_NODE_STACK_ENTRY_POOL_ALLOC_UNIT;
    return WFSKRN_RESULT_OK;
}


#if _DEBUG_DIR_NODE_STACK
static u32 nDebugDirNodeStackGetEntryCount=0;
WFSKrnResult _DirNodeStackGetEntry(char *sFile, u32 nLine, DirNodeItr *pDni) {
    ++nDebugDirNodeStackGetEntryCount;
    ++dnsg.nNumUsed;
    if (bCheckDirNodeStack)
        MyOSReport("Dns:%d, %s Ln:%d GetEntry\n", dnsg.nNumUsed, sFile, nLine);
#else
WFSKrnResult DirNodeStackGetEntry(DirNodeItr *pDni) {
#endif
    WFSKrnLockMutex(&dnsg.mxMutex);
    if ((s32)dnsg.nNumFreeDnsPoolEntries<=0) {
        if ((s32)dnsg.nNumFreeDnsPoolEntries<0) {
            WFSKrnOutputErrorStr("nNumFreeDnsPoolEntries < 0");
        }
        WFSKrnReturnOnError(DirAllocateMoreNodeStackEntries());
    }
    --dnsg.nNumFreeDnsPoolEntries;
    DirNodeStack *pNewDnsEntry = dnsg.pDnsPoolFreeList;
    dnsg.pDnsPoolFreeList = dnsg.pDnsPoolFreeList->pParent;
    WFSKrnUnlockMutex(&dnsg.mxMutex);
    pNewDnsEntry->pParent = pDni->pDns;
    pDni->pDns = pNewDnsEntry;
    ++pDni->nDnsDepth;
#if _DEBUG_DIR_NODE_STACK
    if (bCheckDirNodeStack) {
        dbg_output(MyOSReport("GetEntry: "));
        DirNodeStackCheckAlloc(pDni);
        DirNodeStackCheckFreeList();
    }
#endif
    return WFSKRN_RESULT_OK;
}


#if _DEBUG_DIR_NODE_STACK
void _DirNodeStackFree(char *sFile, u32 nLine, DirNodeItr *pDni, DirNodeStack **ppParent) {
    dnsg.nNumUsed -= pDni->nDnsDepth;
    if (bCheckDirNodeStack)
        MyOSReport("Dns:%d, %s Ln:%d Free\n", dnsg.nNumUsed, sFile, nLine);
    if (bCheckDirNodeStack) {
        dbg_output(MyOSReport("StackFree(%d) { ", pDni->nDnsDepth));
        DirNodeStackCheckAlloc(pDni);
        DirNodeStackCheckFreeList();
    }
#else
void DirNodeStackFree(DirNodeItr *pDni, DirNodeStack **ppParent) {
#endif
    WFSKrnLockMutex(&dnsg.mxMutex);
    dnsg.nNumFreeDnsPoolEntries += pDni->nDnsDepth;
    *ppParent = dnsg.pDnsPoolFreeList;
    dnsg.pDnsPoolFreeList = pDni->pDns;
    WFSKrnUnlockMutex(&dnsg.mxMutex);
#if _DEBUG_DIR_NODE_STACK
    if (bCheckDirNodeStack) {
        dbg_output(MyOSReport("} StackFree: "));
        DirNodeStackCheckAlloc(pDni);
        DirNodeStackCheckFreeList();
    }
#endif
}


#if _DEBUG_DIR_NODE_STACK
WFSKrnResult _DirPushNodeStackEntry(char *sFile, u32 nLine, DirNodeItr *pDni, WFSBlkAdr nBlkAdr) {
    ++dnsg.nNumUsed;
    if (bCheckDirNodeStack)
        MyOSReport("Dns:%d, %s Ln:%d Push\n", dnsg.nNumUsed, sFile, nLine);
    #if _DEBUG_DIR
        if (nDirTest == nDirTestBreakPoint) {
            int a; a=0;
        }
    #endif
#else
WFSKrnResult DirPushNodeStackEntry(DirNodeItr *pDni, WFSBlkAdr nBlkAdr) {
#endif
#if _DEBUG_BREAK_POINT
    static int c=0; c++; if(c == 2601){
        int a; a=0;
    }
    if( pDni->pDns->nBlkAdr == nBlkAdr && pDni->pDns->nSubBlkOfs == *pDni->pOfs ){
        int a; a=0;
    }
#endif
    WFSKrnLockMutex(&dnsg.mxMutex);
    if (dnsg.nNumFreeDnsPoolEntries==0) {
        WFSKrnUnlockMutex(&dnsg.mxMutex);
        WFSKrnReturnOnError(DirAllocateMoreNodeStackEntries());
        WFSKrnLockMutex(&dnsg.mxMutex);
    }
    --dnsg.nNumFreeDnsPoolEntries;
    DirNodeStack *pNewDnsEntry = dnsg.pDnsPoolFreeList;
#if _WIN32
    if ((u32)dnsg.pDnsPoolFreeList == 0xbaadf00d) {
        WFSKrnOutputErrorStr("Bad Node Stack Entry");
    }
#endif
    dnsg.pDnsPoolFreeList = dnsg.pDnsPoolFreeList->pParent;
    pNewDnsEntry->pParent    =  pDni->pDns;
    pDni->pDns = pNewDnsEntry;
    ++pDni->nDnsDepth;
    WFSKrnUnlockMutex(&dnsg.mxMutex);
    pNewDnsEntry->nBlkAdr    =  nBlkAdr;
    pNewDnsEntry->nSubBlkOfs = *pDni->pOfs;
    //}
    pNewDnsEntry->nEntryIdx  = (u8)pDni->nEntryIdx;
#if _DEBUG_DIR_NODE_STACK
    if (bCheckDirNodeStack) {
        dbg_output(MyOSReport("PushEntry: "));
        DirNodeStackCheckAlloc(pDni);
        DirNodeStackCheckFreeList();
    }
#endif
    return WFSKRN_RESULT_OK;
}


#if _DEBUG_DIR_NODE_STACK
void _DirPopNodeStackEntry(char *sFile, u32 nLine, DirNodeItr *pDni) {
    --dnsg.nNumUsed;
    if (bCheckDirNodeStack)
        MyOSReport("Dns:%d, %s Ln:%d Pop\n", dnsg.nNumUsed, sFile, nLine);
#else
void DirPopNodeStackEntry(DirNodeItr *pDni) {
#endif
    WFSKrnLockMutex(&dnsg.mxMutex);
    DirNodeStack *pNewFreeEntry = pDni->pDns;
    pDni->pDns = pNewFreeEntry->pParent;
    pNewFreeEntry->pParent = dnsg.pDnsPoolFreeList;
    dnsg.pDnsPoolFreeList = pNewFreeEntry;
    --pDni->nDnsDepth;
    ++dnsg.nNumFreeDnsPoolEntries;
#if _DEBUG_DIR_NODE_STACK
    if (bCheckDirNodeStack) {
        dbg_output(MyOSReport("PopEntry: "));
        DirNodeStackCheckAlloc(pDni);
        DirNodeStackCheckFreeList();
    }
#endif
    WFSKrnUnlockMutex(&dnsg.mxMutex);
}


void DirConstructEntryNameFromNodeStack(WFSFileName *pKey, DirItr *pDi) {
    // This function parses the node stack constructing the key name
    // ToDo: This function simply concatenates the strings from SubBlks,
    //       but it should take account of the adjustment needed between blocks
#if (_DEBUG_DIR && _DEBUG_DIR_NODE_STACK)
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    DirBlkHdr *pBlkHdr = pDi->pBlkHdr;
    DirNodeStack *pDns = pDi->dni.pDns;
    u32 nI=pDi->dni.nDnsDepth;
    pKey->nLen = pDns->nStrIdx;//nStrLen-1;
    utf8 *pStr = pKey->sStr;
    // Trace back through the leaf
    u32 nMinStrIdx = WFS_MAX_FILE_NAME_SIZE;
    do {
        DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDns->nSubBlkOfs);
        nMinStrIdx = (u32)(pDns->nStrIdx - pSubBlkHdr->nStrLen);
        pStr[pDns->nStrIdx] = pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen + pDns->nEntryIdx-1];
        memcpy(&pStr[nMinStrIdx], pSubBlkHdr->aFirstChar, pSubBlkHdr->nStrLen);
        if (0 == --nI) {
            return;
        }
        pDns = pDns->pParent;
    } while (pDi->tba.nBlkAdr == pDns->nBlkAdr);
    // Trace back through nodes
#if (DIR_PREFIX_SUPPRESSION)
    pDi->tba.nBlkAdr = pDns->nBlkAdr;
#if AREA_ENABLE_BLK_CACHE
    AreaBlkMap(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, TRUE, TRUE, TRUE, (void**) &pBlkHdr);
#else
    TransGetAndPinBlk(&pDi->tba, &pBlkHdr);
#endif
    while(nMinStrIdx) {
        DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDns->nSubBlkOfs);
        utf8 cChoiceChar = pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen + pDns->nEntryIdx-1];
        if (pDns->nStrIdx <= nMinStrIdx) {
            if (pDns->nStrIdx < nMinStrIdx) {
                pStr[pDns->nStrIdx] = cChoiceChar;
            }
            memcpy(&pStr[pDns->nStrIdx - pSubBlkHdr->nStrLen], pSubBlkHdr->aFirstChar, pSubBlkHdr->nStrLen);
            nMinStrIdx = pDns->nStrIdx;
        }
        if (0 == --nI) {
            return;
        }
        pDns = pDns->pParent;
        if (pDi->tba.nBlkAdr != pDns->nBlkAdr) {
            pDi->tba.nBlkAdr = pDns->nBlkAdr;
#if AREA_ENABLE_BLK_CACHE
            AreaBlkUnpin(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr);
            AreaBlkMap(pDi->tba.pAreaInfo, pDns->nBlkAdr, TRUE, TRUE, TRUE, (void**) &pBlkHdr);
#else
            TransGetAndPinBlk(&pDi->tba, &pBlkHdr);
#endif
        }
    }
#if AREA_ENABLE_BLK_CACHE
    AreaBlkUnpin(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr);
#endif
#endif
}


#if _DEBUG_DIR_NODE_STACK
void _DirNodeStackCheckEmpty() {
    if (dnsg.nNumUsed != 0) {
        WFSKrnOutputErrorStr("Dir Node Stack Not Empty");
    }
}
#endif
