/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Errors.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Errors.cpp,v $
  Revision 1.13  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.12  2008/09/04 23:53:04  nakanose_jin
  wfsdev has no pathcache.

  Revision 1.11  2008/08/27 23:23:28  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.10  2008/08/27 09:38:04  paul
  Added error messages. small change to error output.

  Revision 1.9  2008/07/11 18:04:54  paul
  Updated to match latest error defs

  Revision 1.8  2008/07/09 00:43:54  paul
  Renamed some error messages. Added software breakpoint for nDbgCommandCount.

  Revision 1.7  2008/07/02 06:34:57  nakanose_jin
  temporary edit

  Revision 1.6  2008/06/10 20:12:44  kondo_masahiro
  Fixed WFSKrnCommandCountBreak.

  Revision 1.5  2008/06/09 17:51:14  paul
  Added support for logging calls to WFSSrv functions.
  Added new error codes.

  Revision 1.4  2008/05/10 03:59:34  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.3  2008/04/25 17:35:50  kondo_masahiro
  Fixed log information

  Revision 1.2  2008/04/25 12:07:44  nakanose_jin
  add new errors

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.7  2008/04/25 04:35:37  nakanose_jin
  add new errors

  Revision 1.6  2008/04/22 23:22:39  paul
  Made errors match laters wfskrn_Errors.h

  Revision 1.5  2008/04/19 05:50:45  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.4  2008/04/04 21:18:58  paul
  Added \n to error message

  Revision 1.3  2008/02/27 07:24:29  paul
  Updated error strings

  Revision 1.2  2008/02/27 00:37:24  paul
  Fixed a typo

  Revision 1.1  2008/02/21 00:49:14  paul
  *** empty log message ***

 *---------------------------------------------------------------------------*/


#include "wfskrn_Defs.h"
#include "wfskrn_Errors.h"

#if _DEBUG
    u32 nDbgCommandCount = 0;
    u32 nDbgCommandOutputStart = 0; //15520; //(u32)-1;
    u32 nDbgCommandOutputEnd = (u32)-1;
    u32 nDbgCommandCountBreak = (u32)-1;
#if _WIN32
    FILE *fpLog;
#endif
#endif //_DEBUG

const char *WFSKrnGetErrorString( WFSKrnResult nResult ) {
    const char *p;
    switch(nResult) {
        case WFSKRN_RESULT_OK                           :p="OK"                           ; break;
        case WFSKRN_RESULT_BUSY                         :p="BUSY"                         ; break;
        case WFSKRN_RESULT_OUT_OF_MEMORY                :p="OUT_OF_MEMORY"                ; break;
        case WFSKRN_RESULT_INVALID                      :p="INVALID"                      ; break;
        case WFSKRN_RESULT_ACCESS                       :p="ACCESS"                       ; break;
        case WFSKRN_RESULT_LIB_NOT_INITIALIZED          :p="LIB_NOT_INITIALIZED"          ; break;
        case WFSKRN_RESULT_LIB_ALREADY_INITIALIZED      :p="LIB_ALREADY_INITIALIZED"      ; break;
        case WFSKRN_RESULT_FILE_TOO_BIG                 :p="FILE_TOO_BIG"                 ; break;
        case WFSKRN_RESULT_NO_CHANGE_SIZE               :p="NO_CHANGE_SIZE"               ; break;
        case WFSKRN_RESULT_MEDIA_ERROR                  :p="MEDIA_ERROR"                  ; break;
        case WFSKRN_RESULT_DEV_UNUSABLE                 :p="DEV_UNUSABLE"                 ; break;
        case WFSKRN_RESULT_DEV_NOT_INITIALIZED          :p="DEV_NOT_INITIALIZED"          ; break;
        case WFSKRN_RESULT_DEV_IN_USE                   :p="DEV_IN_USE"                   ; break;
        case WFSKRN_RESULT_VOL_ID_ERROR                 :p="VOL_ID_ERROR"                 ; break;
        case WFSKRN_RESULT_WRITE_PROTECTED              :p="WRITE_PROTECTED"              ; break;
        case WFSKRN_RESULT_ALREADY_MOUNTED              :p="ALREADY_MOUNTED"              ; break;
        case WFSKRN_RESULT_PERMISSION                   :p="PERMISSION"                   ; break;
        case WFSKRN_RESULT_PERMISSION_CL                :p="PERMISSION_CL"                ; break;
        case WFSKRN_RESULT_ACL_FULL                     :p="ACL_FULL"                     ; break;
        case WFSKRN_RESULT_ACL_ENTRY_NOT_FOUND          :p="ACL_ENTRY_NOT_FOUND"          ; break;
        case WFSKRN_RESULT_AUTHENTICATION               :p="AUTHENTICATION"               ; break;
        case WFSKRN_RESULT_CORRUPTION                   :p="CORRUPTION"                   ; break;
        case WFSKRN_RESULT_DIRECTORY_QUOTA              :p="DIRECTORY_QUOTA"              ; break;
        case WFSKRN_RESULT_MAX_HANDLES                  :p="MAX_HANDLES"                  ; break;
        case WFSKRN_RESULT_ALREADY_EXISTS               :p="ALREADY_EXISTS"               ; break;
        case WFSKRN_RESULT_NOT_FOUND                    :p="NOT_FOUND"                    ; break;
        case WFSKRN_RESULT_NOT_EMPTY                    :p="NOT_EMPTY"                    ; break;
        case WFSKRN_RESULT_NOT_FILE                     :p="NOT_FILE"                     ; break;
        case WFSKRN_RESULT_NOT_DIRECTORY                :p="NOT_DIRECTORY"                ; break;
        case WFSKRN_RESULT_FILE_OPEN                    :p="FILE_OPEN"                    ; break;
        case WFSKRN_RESULT_LOCKED                       :p="LOCKED"                       ; break;
        case WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED      :p="RESOURCE_LIMIT_EXCEEDED"      ; break;
        case WFSKRN_RESULT_DIR_ENTRY_FOUND              :p="DIR_ENTRY_FOUND"              ; break;
        case WFSKRN_RESULT_DIR_NODE_STRING_PREFIX       :p="DIR_NODE_STRING_PREFIX"       ; break;
        case WFSKRN_RESULT_DIR_CHOICE_PREFIX            :p="DIR_CHOICE_PREFIX"            ; break;
        case WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH     :p="DIR_NODE_STRING_MISMATCH"     ; break;
        case WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND    :p="DIR_NODE_CHOICE_NOT_FOUND"    ; break;
        case WFSKRN_RESULT_DIR_BLK_FULL                 :p="DIR_BLK_FULL"                 ; break;
        case WFSKRN_RESULT_SRV_END_OF_PATH              :p="SRV_END_OF_PATH"              ; break;
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1             :p="SRV_PATH_DEPTH_1"             ; break;
        case WFSKRN_RESULT_SRV_PATH_DEV                 :p="SRV_PATH_DEV"                 ; break;
        case WFSKRN_RESULT_SRV_PATH_VOL                 :p="SRV_PATH_VOL"                 ; break;
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT            :p="SRV_PATH_VOL_ROOT"            ; break;
        case WFSKRN_RESULT_PTREE_ENTRY_FOUND            :p="PTREE_ENTRY_FOUND"            ; break;
        case WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND        :p="PTREE_ENTRY_NOT_FOUND"        ; break;
        case WFSKRN_RESULT_PTREE_FULL                   :p="PTREE_FULL"                   ; break;
        case WFSKRN_RESULT_DEVICE_ERROR                 :p="DEVICE_ERROR"                 ; break;
        case WFSKRN_RESULT_DEVICE_INVALID_PARAMETER     :p="DEVICE_INVALID_PARAMETER"     ; break;
        case WFSKRN_RESULT_DEVICE_NOT_FOUND             :p="DEVICE_INVALID_PARAMETER"     ; break;
        case WFSKRN_RESULT_BLK_CACHE_ALLOC_FAILED       :p="BLK_CACHE_ALLOC_FAILED"       ; break;
        case WFSKRN_RESULT_BCACHE_ERROR                 :p="BCACHE_ERROR"                 ; break;
        case WFSKRN_RESULT_BCACHE_RESOURCE_LIMIT        :p="BCACHE_RESOURCE_LIMIT"        ; break;
        case WFSKRN_RESULT_BCACHE_INVALID_PARAMETER     :p="BCACHE_INVALID_PARAMETER"     ; break;
        case WFSKRN_RESULT_BCACHE_NO_MEMORY             :p="BCACHE_NO_MEMORY"             ; break;
        case WFSKRN_RESULT_BCACHE_NOT_FOUND             :p="BCACHE_NOT_FOUND"             ; break;
        case WFSKRN_RESULT_BCACHE_MAX_DEVICES           :p="BCACHE_MAX_DEVICES"           ; break;
        case WFSKRN_RESULT_BCACHE_INVALID_DEVICE        :p="BCACHE_INVALID_DEVICE"        ; break;
        case WFSKRN_RESULT_BCACHE_INVALID_HANDLE        :p="BCACHE_INVALID_HANDLE"        ; break;
        case WFSKRN_RESULT_BCACHE_INVALID_VOLUME        :p="BCACHE_INVALID_VOLUME"        ; break;
        case WFSKRN_RESULT_BCACHE_ALREADY_MAPPED        :p="BCACHE_ALREADY_MAPPED"        ; break;
        case WFSKRN_RESULT_BCACHE_ALLOC                 :p="BCACHE_ALLOC"                 ; break;
        case WFSKRN_RESULT_BCACHE_HASH_BLK_NOT_MAPPED   :p="BCACHE_PMEM"                  ; break;
        case WFSKRN_RESULT_VOLUME_ERROR                 :p="VOLUME_ERROR"                 ; break;
        case WFSKRN_RESULT_VOLUME_INVALID_PARAMETER     :p="VOLUME_INVALID_PARAMETER"     ; break;
        case WFSKRN_RESULT_VOLUME_BCACHE_ALLOC          :p="VOLUME_BCACHE_ALLOC"          ; break;
        case WFSKRN_RESULT_VOLUME_BCACHE_CONFIG         :p="VOLUME_BCACHE_CONFIG"         ; break;
        //case WFSKRN_RESULT_VOLUME_ID                    :p="VOLUME_ID"                    ; break;
        //case WFSKRN_RESULT_VOLUME_CORRUPT_MR            :p="VOLUME_CORRUPT_MR"            ; break;
        //case WFSKRN_RESULT_VOLUME_AREA_ALLOC            :p="VOLUME_AREA_ALLOC"            ; break;
        //case WFSKRN_RESULT_VOLUME_DEVICE_PARAMETER      :p="VOLUME_DEVICE_PARAMETER"      ; break;
        //case WFSKRN_RESULT_VOLUME_NOT_MAPPED            :p="VOLUME_NOT_MAPPED"            ; break;
        case WFSKRN_RESULT_TRANSACTION_ERROR            :p="TRANSACTION_ERROR"            ; break;
        case WFSKRN_RESULT_TRANSACTION_INVALID_PARAMETER:p="TRANSACTION_INVALID_PARAMETER"; break;
        case WFSKRN_RESULT_ACL_ERROR                    :p="ACL_ERROR"                    ; break;
        case WFSKRN_RESULT_ACL_INVALID_PARAMETER        :p="ACL_INVALID_PARAMETER"        ; break;
        case WFSKRN_RESULT_ACL_MAX_ENTRIES              :p="ACL_MAX_ENTRIES"              ; break;
        case WFSKRN_RESULT_ACL_CACHE                    :p="ACL_CACHE"                    ; break;
        case WFSKRN_RESULT_ACL_FILE                     :p="ACL_FILE"                     ; break;
        case WFSKRN_RESULT_ACL_FILENAME                 :p="ACL_FILENAME"                 ; break;
        case WFSKRN_RESULT_ACL_HANDLE                   :p="ACL_HANDLE"                   ; break;
        case WFSKRN_RESULT_AREA_ERROR                   :p="AREA_ERROR"                   ; break;
        case WFSKRN_RESULT_AREA_INVALID_PARAMETER       :p="AREA_INVALID_PARAMETER"       ; break;
        case WFSKRN_RESULT_AREA_BCACHE_ALLOC            :p="AREA_BCACHE_ALLOC"            ; break;
        case WFSKRN_RESULT_NOT_IMPLEMENTED              :p="NOT_IMPLEMENTED"              ; break;
        case WFSKRN_RESULT_UNKNOWN                      :p="UNKNOWN"                      ; break;
        case WFSKRN_RESULT_FATAL_ERROR                  :p="FATAL_ERROR"                  ; break;
        default: p="???" ;break  ;
    }
    return p;
}

WFSKrnResult WFSKrnOutputErrorCode(WFSKrnResult nResult) {
#if _DEBUG
    MyOSReport("%d. Error: (%d) WFSKRN_RESULT_%s\n", nDbgCommandCount, nResult, WFSKrnGetErrorString(nResult) );
#endif
    if (nResult != WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND) {
        return nResult;
    }
    WFSKrnDbgOutputClose();
    return nResult;
}

void PathCacheDebugOutput();

WFSKrnResult WFSKrnOutputErrorStr(const char *pStr) {
#if _DEBUG
    MyOSReport("%d. %s\n", nDbgCommandCount, pStr);
    WFSKrnDbgOutputClose();
#endif
    return WFSKRN_RESULT_UNKNOWN;
}

void WFSKrnCommandCountBreak() {
#if _DEBUG
    if(nDbgCommandCount >= nDbgCommandCountBreak) {
#ifndef WFSDEV
		PathCacheDebugOutput();
#endif
        int a; a=0;
    }
#endif
}
