/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Heap.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Heap.cpp,v $
  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.3  2008/04/19 05:50:40  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2007/12/27 11:07:04  paul
  fixed some copy-paste errors: changed WFS to WFSKrn

  Revision 1.1  2007/12/27 11:04:05  paul
  similar to wfs_Heap.cpp, but will eventually support IOP, so will eventually be different


*---------------------------------------------------------------------------*/

#include "wfskrn_Heap.h"
#include "firm.h"

#undef dbg
#if _DEBUG_WFSKRN_ALLOC
#define dbg(s) s
#else
#define dbg(s)
#endif

#if _WIN32

dbg(u32 nSizeAvaliable=0;)

void WFSKrnCreateHeap(WFSKrnHeap *pHeap, void *pBase, size_t nSize) {
    *pHeap = HeapCreate(0, nSize, nSize);
    dbg(nSizeAvaliable = nSize);
}

void *WFSKrnHeapAlloc(WFSKrnHeap *pHeap, size_t nSize) {
    void *pAddr = HeapAlloc(*pHeap, 0, nSize);
    dbg(nSizeAvaliable -= nSize;
        MyOSReport("A:Heap:%d\n", nSizeAvaliable));
    return pAddr;
}

void WFSKrnHeapFree(WFSKrnHeap *pHeap, void *pAddr) {
    dbg(nSizeAvaliable += HeapSize(pHeap, NULL, pAddr);
        MyOSReport("F:Heap:%d\n", nSizeAvaliable));
    HeapFree(*pHeap, 0, pAddr);
}

void WFSKrnHeapDestroy(WFSKrnHeap *pHeap) {
    dbg(MyOSReport("WFSKrnHeapDestroy()\n"));
    HeapDestroy(*pHeap);
}

//-- twl modified
#elif _TWL

void WFSKrnCreateHeap(WFSKrnHeap *pHeap, void *pBase, size_t nSize) {
    osTPrintf( "%s needs modify.\n", __FUNCTION__);
  //main.cōsĂ
}

void *WFSKrnHeapAlloc(WFSKrnHeap *pHeap, size_t nSize) {
    void *pAddr = osAlloc( nSize);
    if( pAddr == NULL) {
        osTPrintf( "%s error!\n", __FUNCTION__);
    }
//    void *pAddr = OS_AllocFromHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap, nSize);
    return pAddr;
}

void WFSKrnHeapFree(WFSKrnHeap *pHeap, void *pAddr) {
    osFree( pAddr);
//    OS_FreeToHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap, pAddr);
}

void WFSKrnHeapDestroy(WFSKrnHeap *pHeap) {
    osTPrintf( "%s needs modify.\n", __FUNCTION__);
//    OS_DestroyHeap( OS_ARENA_MAIN_SUBPRIV, *pHeap);
}


//-- twl modified

#elif _RVL

void WFSKrnCreateHeap(WFSKrnHeap *pHeap, void *pBase, size_t nSize) {
    pHeap->handle = MEMCreateExpHeap(pBase, nSize);
    MEMInitAllocatorForExpHeap(&pHeap->allocator, pHeap->handle, WFS_HEAP_DEFAULT_ALIGNMENT);
}

void *WFSKrnHeapAlloc(WFSKrnHeap *pHeap, size_t nSize) {
//    osTPrintf("==> %s:%d pHeap: 0x%08x size: %d\n", __FILE__, __LINE__, pHeap->pAllocator, nSize);
    void *pAddr = MEMAllocFromAllocator(&pHeap->allocator, nSize);
//    osTPrintf("==> %s:%d pAddr: 0x%08x\n", __FILE__, __LINE__, pAddr);
    return pAddr;
}

void WFSKrnHeapFree(WFSKrnHeap *pHeap, void *pAddr) {
    MEMFreeToAllocator(&pHeap->allocator, pAddr);
}

void WFSKrnHeapDestroy(WFSKrnHeap *pHeap) {
    MEMDestroyExpHeap(pHeap->handle);
}

#elif _IOP

void WFSKrnCreateHeap(WFSKrnHeap *pHeap, void *pBase, size_t nSize) {
    *pHeap = IOS_CreateHeap(pBase, nSize);
}

void *WFSKrnHeapAlloc(WFSKrnHeap *pHeap, size_t nSize) {
//    osTPrintf("==> %s:%d pHeap: 0x%08x size: %d\n", __FILE__, __LINE__, pHeap, nSize);
    void *pAddr = IOS_AllocAligned(*pHeap, nSize, 32);
//    osTPrintf("==> %s:%d pAddr: 0x%08x\n", __FILE__, __LINE__, pAddr);
    return pAddr;
}

void WFSKrnHeapFree(WFSKrnHeap *pHeap, void *pAddr) {
    IOS_Free(*pHeap, pAddr);
}

void WFSKrnHeapDestroy(WFSKrnHeap *pHeap) {
    IOS_DestroyHeap(*pHeap);
}


#endif
