/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Mutex.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Mutex.cpp,v $
  Revision 1.2  2008/04/25 17:28:54  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.2  2008/04/19 05:50:34  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2007/12/27 11:11:58  paul
  similar to wfs_Mutex.cpp, but will eventually need to support IOP


*---------------------------------------------------------------------------*/

#include "wfskrn_Mutex.h"

#undef dbg
#if _DEBUG_WFSKRN_MUTEX
#define dbg(s) s
#else
#define dbg(s)
#endif

#if _WIN32

dbg(u32 aSignalCounter[0x2000]);

void WFSKrnInitMutex(WFSKrnMutex *pMx) {
    pMx->hHandle = CreateMutex(NULL, false, NULL);
    ReleaseMutex(pMx->hHandle);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]=1;
        MyOSReport("InitMutex(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
}

void WFSKrnLockMutex(WFSKrnMutex *pMx) {
    dbg(MyOSReport("Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    WaitForSingleObject(pMx->hHandle, INFINITE);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]--);
}

bool WFSKrnTryLockMutex(WFSKrnMutex *pMx) {
    return (WaitForSingleObject(pMx->hHandle, 0) != WAIT_TIMEOUT);
}

void WFSKrnUnlockMutex(WFSKrnMutex *pMx) {
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]++;
        MyOSReport("Release(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    ReleaseMutex(pMx->hHandle);
}

void WFSKrnInitCond(WFSKrnCond *pCond) {
    pCond->hAutoResetEvent = CreateEvent(NULL, false, false, NULL);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]=0;
        MyOSReport("InitCond(%x:%d)\n", pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    //ResetEvent(pCond->hAutoResetEvent);
    //pCond->hManualResetEvent = CreateEvent(NULL, true, false, NULL);
}

void WFSKrnSignalCond(WFSKrnCond *pCond) {
    //PulseEvent(pCond->hAutoResetEvent);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]++;
        MyOSReport("Signal(%x:%d)\n", pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    SetEvent(pCond->hAutoResetEvent);
}

void WFSKrnWaitCond(WFSKrnCond *pCond, WFSKrnMutex* pMx) {
#if 1
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]++;
        MyOSReport("Signal(%x:%d) And Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff], pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    //DWORD nRet = SignalObjectAndWait(pMx->hHandle, pCond->hAutoResetEvent, INFINITE, false);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]--);
#else
    ReleaseMutex(pMx->hHandle);
    WaitForSingleObject(pCond->hAutoResetEvent, INFINITE);
#endif
    dbg(MyOSReport("Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    WaitForSingleObject(pMx->hHandle, INFINITE);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]--);
}

void WFSKrnResetCond(WFSKrnCond *pCond) {
    ResetEvent(pCond->hAutoResetEvent);
}

bool WFSKrnCreateThread(WFSKrnThread *pThread,
                        WFSKrnThreadFunc threadFunc,
                        void*      param,
                        void*      stackBase,
                        u32        stackSize,
                        OSPriority priority)
{
    pThread->hThread = CreateThread(NULL, stackSize, (LPTHREAD_START_ROUTINE)threadFunc, param, CREATE_SUSPENDED, 0);
    SetThreadPriority(pThread->hThread, 16-priority);
    return true;
}

s32 WFSKrnResumeThread(WFSKrnThread *pThread) {
    return ResumeThread(pThread->hThread);
}

void WFSKrnCancelThread(WFSKrnThread *pThread) {
    // Note: On windows, this does not do any cleanup - should avoid using
    TerminateThread(pThread->hThread, 0);
}

void WFSKrnExitThread(void *pVal) {
    ExitThread((DWORD)pVal);
}

s32 WFSKrnSuspendThread(WFSKrnThread *pThread)
{
    return SuspendThread(pThread->hThread);
}

bool WFSKrnIsThreadTerminated(WFSKrnThread *pThread)
{
    switch (WaitForSingleObject(pThread->hThread, 0))
    {
      case WAIT_TIMEOUT:
        return false;
      case WAIT_OBJECT_0:  // the thread has exited.
      case WAIT_ABANDONED: // [check]
      case WAIT_FAILED:    // the thread does not exist. [check] should check error code?
      default:
        break;
    }

    return true;
}

bool WFSKrnJoinThread(WFSKrnThread *pThread, void** val)
{
#pragma unused (val) // exit value is not supported.
    switch (WaitForSingleObject(pThread->hThread, INFINITE))
    {
      case WAIT_OBJECT_0:
        return true;
      case WAIT_TIMEOUT:
      case WAIT_ABANDONED:
      case WAIT_FAILED:
      default:
        break;
    }
    return false;
}

#elif _RVL


bool WFSKrnCreateThread(WFSKrnThread *pThread,
                        WFSKrnThreadFunc threadFunc,
                        void*      param,
                        void*      stackBase,
                        u32        stackSize,
                        OSPriority priority)
{
    return OSCreateThread(pThread, threadFunc, param, (void*)((u32)stackBase+stackSize), stackSize, priority, OS_THREAD_ATTR_DETACH);
}

//--- twl modified
#elif _TWL
typedef void (*MyThreadFunc)(void*);
bool WFSKrnCreateThread(WFSKrnThread *pThread,
                        WFSKrnThreadFunc threadFunc,
                        void*      param,
                        void*      stackBase,
                        u32        stackSize,
                        OSPriority priority)
{
     osCreateThread( pThread, (MyThreadFunc)threadFunc, param, (void*)((u32)stackBase+stackSize), stackSize, priority);
     return true;
}

s32 WFSKrnResumeThread(WFSKrnThread *pThread)
{
    osWakeupThreadDirect( pThread);
#if 1
//    osTPrintf( "Need TWL modifies.\n");
    return 0;
#else
    if(WFSKrnIsThreadTerminated(pThread)){
        return 0;
    }
    else if(--pThread->count == 0){
        IOS_StartThread(pThread->tid);
    }
    return pThread->count;
#endif
}

void WFSKrnCancelThread(WFSKrnThread *pThread)
{
    osDestroyThread( pThread);
}

void WFSKrnExitThread(void *pVal)
{
    osExitThread();
}

s32 WFSKrnSuspendThread(WFSKrnThread *pThread)
{
#if 1
    osTPrintf( "Need TWL modifies.\n");
    return 0;
#else
    if(WFSKrnIsThreadTerminated(pThread)){
        return 0;
    }
    else if(++pThread->count == 1){
        IOS_StopThread(pThread->tid);
    }
    return pThread->count;
#endif
}

bool WFSKrnIsThreadTerminated(WFSKrnThread *pThread)
{
    return( osIsThreadTerminated( pThread));
}

bool WFSKrnJoinThread(WFSKrnThread *pThread, void** val)
{
#pragma unused (val) // exit value is not supported.
    osJoinThread( pThread);
    return true;
}


void WFSKrnInitCond(WFSCond *pCond) {
    osInitEvent( pCond);
}

void WFSKrnSignalCond(WFSCond *pCond) {
    osSignalEvent( pCond, 1);
}

void WFSKrnWaitCond(WFSCond *pCond, WFSMutex* pMx) {
    osUnlockMutex( pMx);
    osWaitEventEx( pCond, 1, OS_EVENT_MODE_OR, 1); //Cxg҂ɃNA𔺂
}

void WFSKrnResetCond(WFSCond *pCond) {
    osClearAllEvent( pCond);
    return;
}
//-- twl modified

#elif _IOP

void WFSKrnInitMutex(WFSKrnMutex *mutex)
{
    mutex->mqid = IOS_CreateMessageQueue(mutex->mq, WFSKRN_MUTEX_Q_SIZE);
    IOS_SendMessage(mutex->mqid, 0xabcd1234, IOS_MESSAGE_NOBLOCK);
}

void WFSKrnLockMutex(WFSKrnMutex *mutex)
{
    IOSMessage msg;
    IOS_ReceiveMessage(mutex->mqid, &msg, IOS_MESSAGE_BLOCK);
}

bool WFSKrnTryLockMutex(WFSKrnMutex *mutex)
{
    IOSMessage msg;
    if(IOS_ReceiveMessage(mutex->mqid, &msg, IOS_MESSAGE_NOBLOCK) == IOS_ERROR_OK )
        return true;
    else 
        return false;
}

void WFSKrnUnlockMutex(WFSKrnMutex *mutex)
{
    IOS_SendMessage(mutex->mqid, 0xabcd1234, IOS_MESSAGE_NOBLOCK);
}

bool WFSKrnCreateThread(WFSKrnThread *pThread,
                        WFSKrnThreadFunc threadFunc,
                        void*      param,
                        void*      stackBase,
                        u32        stackSize,
                        OSPriority priority)
{
     IOSError ie = IOS_CreateThread((IOSEntryProc)threadFunc, param, stackBase, stackSize, (u32)priority, IOS_THREAD_CREATE_JOINABLE);
     if(ie < 0) return false;
     pThread->tid = ie;
     pThread->count = 1;
     return true;
}

s32 WFSKrnResumeThread(WFSKrnThread *pThread)
{
    if(WFSKrnIsThreadTerminated(pThread)){
        return 0;
    }
    else if(--pThread->count == 0){
        IOS_StartThread(pThread->tid);
    }
    return pThread->count;
}

void WFSKrnCancelThread(WFSKrnThread *pThread)
{
    IOS_DestroyThread(pThread->tid, 0);
}

void WFSKrnExitThread(void *pVal)
{
    IOS_DestroyThread(IOS_GetThreadId(), pVal);
}

s32 WFSKrnSuspendThread(WFSKrnThread *pThread)
{
    if(WFSKrnIsThreadTerminated(pThread)){
        return 0;
    }
    else if(++pThread->count == 1){
        IOS_StopThread(pThread->tid);
    }
    return pThread->count;
}

bool WFSKrnIsThreadTerminated(WFSKrnThread *pThread)
{
    if(IOS_GetThreadPriority(pThread->tid) >= 0)
        return true;
    else
        return false;
}

bool WFSKrnJoinThread(WFSKrnThread *pThread, void** val)
{
    if(IOS_JoinThread(pThread->tid, val) == IOS_ERROR_OK)
        return true;
    else
        return false;
}

#endif
