/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfsKrn_Permission.cpp - Functions for permissions.

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_AccessList.cpp,v $
  Revision 1.25  2008/10/20 00:30:54  nakanose_jin
  (none)

  Revision 1.24  2008/10/09 04:39:07  ueno
  Fixed DeletePermission() not to set WFSKRN_ACCESSLIST_FREEINDEX to handle in error case.

  Revision 1.23  2008/10/09 01:44:22  ueno
  Added workarounds.

  Revision 1.22  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.21  2008/09/26 11:07:21  ueno
  Cleanup.

  Revision 1.20  2008/09/24 07:33:21  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.19  2008/08/19 08:12:23  ueno
  Fixed error handling in WFSKrnAccessListCreateFile().

  Revision 1.18  2008/07/23 01:57:53  ueno
  Commented out unused functions.

  Revision 1.17  2008/07/11 08:10:36  ueno
  Fixed control level check.

  Revision 1.16  2008/07/11 01:30:32  ueno
  Added null pointer check to WFSKrnAccessListInit().

  Revision 1.15  2008/07/02 06:34:57  nakanose_jin
  temporary edit

  Revision 1.14  2008/06/05 14:05:22  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.13  2008/06/05 08:07:48  ueno
  Revised permission APIs.

  Revision 1.12  2008/06/02 15:10:02  ueno
  Revised WFSKrnAccessList{} to reduce the size.

  Revision 1.11  2008/05/30 08:50:07  ueno
  Cleanup.

  Revision 1.10  2008/05/28 02:12:49  ueno
  Cleanup.

  Revision 1.9  2008/05/23 06:32:08  ueno
  Cleanup.

  Revision 1.8  2008/05/16 11:16:17  ueno
  Implemented permission APIs with directory module.

  Revision 1.7  2008/05/08 11:43:33  ueno
  Cleanup.

  Revision 1.6  2008/05/06 02:23:58  ueno
  Fixed IsAllowed() to check root operation.

  Revision 1.5  2008/04/29 01:12:54  ueno
  Defined WFS_ROOT_ID and WFS_EVERYONE_ID.

  Revision 1.4  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.3  2008/04/25 04:40:07  ueno
  Added support for WFSDEV.

  Revision 1.27  2008/04/23 08:59:47  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.26  2008/04/15 07:36:01  ueno
  Changed return type to WFSKrnResult.

  Revision 1.25  2008/04/14 13:49:59  ueno
  Cleanup.

  Revision 1.24  2008/04/14 03:00:39  ueno
  Modified WFSKrnCreateRootAccessList() to create an accesslist only for the root.
  Added WFSKrnDebugCreateAccessList() for tests.

  Revision 1.23  2008/04/11 13:21:42  ueno
  Removed  WFSKrnAccessListSetInstallerPermission().
  Fixed WFSKrnAccessListCheckDirectoryPermission() and CheckFilePermission().

  Revision 1.22  2008/04/11 06:16:47  ueno
  Defined WFSAclFileHandle.

  Revision 1.21  2008/04/11 04:53:05  ueno
  Added WFSKRN_ACL_USE_MUTEX to switch mutex.

  Revision 1.20  2008/04/11 02:57:59  ueno
  Renamed WFSKrnCreateAccessList() to make clear the purpose.

  Revision 1.19  2008/04/09 11:30:10  ueno
  Fine-granularity locking.

  Revision 1.18  2008/04/09 03:01:01  ueno
  Modified to initialize mutex in AccessListControlBlock.

  Revision 1.17  2008/04/08 06:39:32  ueno
  Cleanup.

  Revision 1.16  2008/04/08 05:43:21  ueno
  Revised creatorfs inherited permissions in a new accesslist.

  Revision 1.15  2008/04/07 12:46:56  ueno
  Added support for group others.

  Revision 1.14  2008/04/07 09:21:46  ueno
  Added WFSKrnAccessListSetInstallerPermission() to set permissions for the installer by the root.

  Revision 1.13  2008/04/03 11:40:44  ueno
  Replaced accesslist index by accesslist handle.

  Revision 1.12  2008/04/03 10:00:22  ueno
  Cleanup.

  Revision 1.11  2008/04/03 06:39:30  ueno
  Modified GetInheritAccessList() to set proper control levels of the caller.
  Cleanup.

  Revision 1.10  2008/04/02 14:14:53  ueno
  Revised permission inheritance.

  Revision 1.9  2008/04/01 08:24:54  ueno
  Fixed cache reference count.

  Revision 1.8  2008/03/31 14:07:27  ueno
  Removed WFSKrnAccessListAddUser() and RemoveUser().
  Removed WFSKrnAccessListCan*().

  Revision 1.7  2008/03/19 10:18:32  ueno
  Revised WFSKrnAddUser() and RemoveUser() to specify operator title ID.

  Revision 1.6  2008/03/18 13:44:04  ueno
  Fixed refernce count of cache.
  Revised WFSKrnAddEntry(), RemoveEntry() and SetPermission() for cleanup.

  Revision 1.5  2008/03/14 08:02:53  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.4  2008/03/12 06:37:37  ueno
  Added WFSAccessListCanRead(), Write()...

  Revision 1.3  2008/03/11 12:10:41  ueno
  Cleanup.

  Revision 1.2  2008/03/11 11:39:01  ueno
  Cleanup.

  Revision 1.1  2008/03/11 11:10:47  ueno
  Moved accesslist functions into wfskrn_Permission_AccessList.cpp.

  Revision 1.17  2008/03/11 09:51:54  ueno
  Fixed EscapeAccessList() to make a filename from UTF-16 characters (0x3300-0x33ff).

  Revision 1.16  2008/03/11 06:53:05  ueno
  Added WFSAccessListSetPermission() and GetPermission().

  Revision 1.15  2008/03/07 05:43:59  ueno
  Cleanup.

  Revision 1.14  2008/03/07 05:32:07  ueno
  Modified WFSCreateAccessList() to be able to create an accesslist for directories.

  Revision 1.13  2008/03/06 12:25:27  ueno
  Revised cache freelist (signgle -> double link list).

  Revision 1.12  2008/03/06 11:01:38  ueno
  Fixed asserts.

  Revision 1.11  2008/03/06 10:45:44  ueno
  Revised accesslist cache.

  Revision 1.10  2008/02/29 07:29:10  ueno
  Added Windows api wrappers.

  Revision 1.9  2008/02/29 00:48:05  ueno
  Changed the attribute of index to inout (WFSAccessListAddEntry() and RemoveEntry()).

  Revision 1.8  2008/02/28 13:11:01  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.7  2008/02/28 07:01:52  ueno
  Fixed escape sequence for accesslist filename.

  Revision 1.6  2008/02/28 04:36:32  ueno
  Cleanup.

  Revision 1.5  2008/02/27 10:05:36  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.4  2008/02/27 04:29:24  ueno
  Modified WFSAccessListAddEntry() and RemoveEntry() to delete accesslist file when the reference count is zero.

  Revision 1.3  2008/02/26 14:11:43  ueno
  Added WFSAccessListAddEntry() and WFSAccessListRemoveEntry() to edit accesslist.

  Revision 1.2  2008/02/25 09:43:57  ueno
  Added support for accesslist cache.

  Revision 1.1  2008/02/22 05:16:13  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#define WFSKrn_ACCESSLIST_KEEP_REF

#include "wfs_Defs.h"
#include "wfskrn_Permission_AccessList.h"

#ifdef WFSDEV
#include "wfskrn_Permission_Win32.h"
#else // WFSDEV
#include "wfskrn_Permission_Iop.h"
#endif // WFSDEV

#include "wfskrn_Permission_File.h"
#include "wfskrn_Permission_Private.h"

#define FreeBlockAddress(x)        ((x)->accessListCb->freeBlock)    // WFSKrnAccessListBlock*
#define MemoryBlockAddress(x)      ((x)->accessListCb->memoryBlock)  // WFSKrnAccessListBlock*
#define NumMemoryBlock(x)          ((x)->accessListCb->numBlock)
#define AccessListCache(x)         ((x)->accessListCb->cache)
#define AccessListCacheFreeList(x) ((x)->accessListCb->cacheFreeList)
#define Mutex(x)                   ((x)->accessListCb->mxMutex)
#define Initialized(x)             ((x)->accessListCb->initialized)
#define InstallerDirPm(x)          ((x)->accessListCb->installerDirPermission)
#define InstallerFilePm(x)         ((x)->accessListCb->installerFilePermission)

#define CacheIsEmpty()         (AccessListCache(area).tail == NULL)
#define CacheFreeListIsEmpty() (AccessListCacheFreeList(area).tail == NULL)

#define IsCacheEnd(list)   ((list) == NULL)
#define SIZE_OF_LIST(x)    (sizeof(WFSKrnKrnAccessList) + WFSKrnKrnAccessListGetNum(x) * sizeof(WFSKrnAccessListBlock))

#define IsDirectory(x)         ((x)->handle & WFS_FLAG_IS_A_DIRECTORY)

#define GetFirstEntry(list)    ((WFSKrnAccessListEntry*) ((list)+1))
#define GetEntry(list, index)  ((WFSKrnAccessListEntry*) ((list)+1) + (index))

#define BRANCH_MODE_SET     0
#define BRANCH_MODE_ADD     1
#define BRANCH_MODE_MODIFY  2
#define BRANCH_MODE_DELETE  3

#define DP_RESERVED     (~(WFS_PERM_DIR_FULL | WFS_PERM_CL_MASK))
#define FP_RESERVED     (~(WFS_PERM_FILE_FULL | WFS_PERM_CL_MASK))

#define IsGroup(x)      ((x) & WFSKRN_ACCESSLIST_GROUP)

static WFSBool VerifyCacheList(const AreaInfo* area);
static WFSBool VerifyList(const WFSKrnAccessList* list);

// [check] should be macro.
static WFSKrnAccessListBlock* ACL_CACHE_ADDRESS(const AreaInfo* area, u8 blockNumber)
{
    WFSKrnAccessListBlock* address;

    if (blockNumber == ACL_NO_ENTRY)
    {
        address = NULL;
    }
    else
    {
        address = MemoryBlockAddress(area) + blockNumber;
    }
    return address;
}

// [check] should be macro.
static u8 ACL_CACHE_BLOCK(const AreaInfo* area, WFSKrnAccessListBlock* address)
{
    u8 blockNumber;
    if (address)
    {
        blockNumber = (address - MemoryBlockAddress(area));
    }
    else
    {
        blockNumber = ACL_NO_ENTRY;
    }
    return blockNumber;
}

// [check] should be macro.
void WFSKrnAccessListCacheAdd(const AreaInfo* area, WFSKrnAccessList* list)
{
    register WFSKrnAccessList* next;
    next = AccessListCache(area).head;
    if (IsCacheEnd(next))
    {
        AccessListCache(area).tail = list;
    }
    else
    {
        next->prevBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) list);
    }
    list->prevBlock = ACL_NO_ENTRY;
    list->nextBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) next);
    AccessListCache(area).head = list;
}

// [check] should be macro.
void WFSKrnAccessListCacheRemove(const AreaInfo* area, WFSKrnAccessList* list)
{
    register WFSKrnAccessList* next;
    register WFSKrnAccessList* prev;

    next = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextBlock);
    prev = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->prevBlock);

    if (IsCacheEnd(next))
    {
        AccessListCache(area).tail = prev;
    }
    else
    {
        next->prevBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) prev);
    }

    if (IsCacheEnd(prev))
    {
        AccessListCache(area).head = next;
    }
    else
    {
        prev->nextBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) next);
    }
}

// [check] should be macro.
// look up an accesslist cache (0 <= ref) in the cache list,
WFSKrnAccessList* WFSKrnAccessListLookupByIndex(const AreaInfo* area, WFSKrnAclHandle handle)
{
    register WFSKrnAccessList* list;
    list = AccessListCache(area).head;
    while (list != NULL)
    {
        if (list->handle == handle)
        {
            return list;
        }
        list = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextBlock);
    }
    return NULL;
}

// [check] should be macro.
// look up an active accesslist (0 < ref)
WFSKrnAccessList* WFSKrnAccessListLookup(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WFSKrnAccessList* list = WFSKrnAccessListLookupByIndex(area, handle);
    if (!list || list->ref == 0)
    {
        return NULL;
    }
    return list;
}

// [check] should be macro.
static void CacheFreeListInsertTail(const AreaInfo* area, WFSKrnAccessList* list)
{
    register WFSKrnAccessList* prev;
    prev = AccessListCacheFreeList(area).tail;
    if (IsCacheEnd(prev))
    {
        AccessListCacheFreeList(area).head = list;
    }
    else
    {
        prev->nextFreeCacheBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) list);
    }
    list->prevFreeCacheBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) prev);
    list->nextFreeCacheBlock = ACL_NO_ENTRY;
    AccessListCacheFreeList(area).tail = list;
}

// [check] should be macro.
static WFSKrnAccessList* CacheFreeListRemoveHead(const AreaInfo* area)
{
    register WFSKrnAccessList* head;
    register WFSKrnAccessList* next;

    head = AccessListCacheFreeList(area).head;
    if (head)
    {
        next = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, head->nextFreeCacheBlock);
        if (IsCacheEnd(next))
        {
            AccessListCacheFreeList(area).tail = next;
        }
        else
        {
            next->prevFreeCacheBlock = ACL_NO_ENTRY;
        }
        AccessListCacheFreeList(area).head = next;
    }
    return head;
}

// [check] should be macro.
void WFSKrnAccessListCacheFreeListRemove(const AreaInfo* area, WFSKrnAccessList* list)
{
    register WFSKrnAccessList* next;
    register WFSKrnAccessList* prev;

    next = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextFreeCacheBlock);
    prev = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->prevFreeCacheBlock);

    if (IsCacheEnd(next))
    {
        AccessListCacheFreeList(area).tail = prev;
    }
    else
    {
        next->prevFreeCacheBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) prev);
    }

    if (IsCacheEnd(prev))
    {
        AccessListCacheFreeList(area).head = next;
    }
    else
    {
        prev->nextFreeCacheBlock = ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) next);
    }
}

// [check] should be macro.
static void CacheFreeListInvalidate(const AreaInfo* area, WFSKrnAccessList* list)
{
    int ret;
    WFSKrnAccessListCacheFreeListRemove(area, list);
    ret = WFSKrnMemoryBlockFree(area, (WFSKrnAccessListBlock*) list, WFSKRN_BLOCKS_IN_LIST(list));
    WFSKRN_ACL_ASSERT(ret == 0);
}

//
// Memory block
//
static int SetFreeBlockNum(WFSKrnAccessListBlock* free, u32 num)
{
    WFSKRN_ACL_ASSERT(num <= ACL_MAX_MEMORY_BLOCK);
    free->num = (u8) num;
    free->type = WFSKRN_ACCESSLIST_FREEBLOCK;
    return 0;
}

static int GetFreeBlockNum(const WFSKrnAccessListBlock* free)
{
    WFSKRN_ACL_ASSERT(free->type == WFSKRN_ACCESSLIST_FREEBLOCK);
    return free->num;
}

u32 WFSKrnAccessListGetNum(const WFSKrnAccessList* list)
{
    return (u32) (list->num);
}

void WFSKrnAccessListSetNum(WFSKrnAccessList* list, u32 num)
{
    WFSKRN_ACL_ASSERT(num <= ACL_MAX_MEMORY_BLOCK);
    list->num = (u8) num;
}

static int WFSKrnMemoryBlockInit(const AreaInfo* area)
{
    if (!(area->accessListCb) || !MemoryBlockAddress(area))
    {
        return -1;
    }

    // all memory blocks are free.
    FreeBlockAddress(area) = MemoryBlockAddress(area);
    FreeBlockAddress(area)->nextBlock = NumMemoryBlock(area); // end of memory block.
    SetFreeBlockNum(FreeBlockAddress(area), NumMemoryBlock(area));

    if (WFSKrnAccessListIndexListInit(area) < 0)
    {
        return  -1;
    }

    return 0;
}

WFSKrnAccessListBlock* WFSKrnMemoryBlockAlloc(const AreaInfo* area, int numBlocks)
{
    WFSKrnAccessListBlock* free = FreeBlockAddress(area);
    WFSKrnAccessListBlock* prev = NULL;
    WFSKrnAccessListBlock* newFree;

    if (MemoryBlockAddress(area) + NumMemoryBlock(area) <= FreeBlockAddress(area) ||
        numBlocks == 0)
    {
        return NULL; // no freeblock.
    }

    while (GetFreeBlockNum(free) < numBlocks)
    {
        if (free->nextBlock == NumMemoryBlock(area))
        {
            WFSKRN_ACL_ASSERT(VerifyMemoryBlock(area));
            return NULL;
        }
        prev = free;
        free = MemoryBlockAddress(area) + free->nextBlock;
    }

    if (GetFreeBlockNum(free) == numBlocks)
    {
        if (prev)
        {
            prev->nextBlock = free->nextBlock;
        }
        else
        {
            FreeBlockAddress(area) = MemoryBlockAddress(area) + free->nextBlock;
        }
    }
    else
    {
        newFree = free + numBlocks;
        newFree->nextBlock = free->nextBlock;
        SetFreeBlockNum(newFree, GetFreeBlockNum(free) - numBlocks);

        if (prev)
        {
            prev->nextBlock = newFree - MemoryBlockAddress(area);
        }
        else
        {
            FreeBlockAddress(area) = newFree;
        }
    }

    free->nextBlock = 0; // allocated.
    SetFreeBlockNum(free, numBlocks);
    free->type      = 0; // allocated.

    WFSKRN_ACL_ASSERT(VerifyMemoryBlock(area));

    return free;
}

int WFSKrnMemoryBlockFree(const AreaInfo* area, WFSKrnAccessListBlock* block, u32 numBlocks)
{
    if (block < MemoryBlockAddress(area) || MemoryBlockAddress(area) + NumMemoryBlock(area) <= block)
    {
        return -1; // out of memory blocks.
    }
    if (numBlocks == 0)
    {
        return 0;
    }

    if (block < FreeBlockAddress(area))
    {
        if (FreeBlockAddress(area) == block + numBlocks)
        {
            // combine
            if (FreeBlockAddress(area) == MemoryBlockAddress(area) + NumMemoryBlock(area))
            {
                SetFreeBlockNum(block, numBlocks); // set flags.
                block->nextBlock = NumMemoryBlock(area); // the tail of memory block.
            }
            else
            {
                SetFreeBlockNum(block, GetFreeBlockNum(FreeBlockAddress(area)) + numBlocks);
                block->nextBlock = FreeBlockAddress(area)->nextBlock;
            }
        }
        else
        {
            SetFreeBlockNum(block, numBlocks);
            block->nextBlock = FreeBlockAddress(area) - MemoryBlockAddress(area);
        }
        FreeBlockAddress(area) = block;
        WFSKRN_ACL_ASSERT(VerifyMemoryBlock(area));
    }
    else
    {
        WFSKrnAccessListBlock* free;
        WFSKrnAccessListBlock* next;

        SetFreeBlockNum(block, numBlocks);
        free = FreeBlockAddress(area);
        next = MemoryBlockAddress(area) + FreeBlockAddress(area)->nextBlock;

        while (next < block)
        {
            free = next;
            next = MemoryBlockAddress(area) + next->nextBlock;
        }

        if (free + GetFreeBlockNum(free) == block)
        {
            // combine
            numBlocks += GetFreeBlockNum(free);
            SetFreeBlockNum(free, numBlocks);
            block = free;
        }
        else
        {
            free->nextBlock = block - MemoryBlockAddress(area);
        }

        if (block + GetFreeBlockNum(block) == next && next < MemoryBlockAddress(area) + NumMemoryBlock(area))
        {
            // combine
            SetFreeBlockNum(block, GetFreeBlockNum(next) + numBlocks);
            block->nextBlock = next->nextBlock;
        }
        else
        {
            block->nextBlock = next - MemoryBlockAddress(area);
        }
        WFSKRN_ACL_ASSERT(VerifyMemoryBlock(area));
    }

    return 0;
}

static int AccessListCacheInit(const AreaInfo* area)
{
    AccessListCache(area).head = NULL;
    AccessListCache(area).tail = NULL;

    AccessListCacheFreeList(area).head = NULL;
    AccessListCacheFreeList(area).tail = NULL;

    return 0;
}

// create an accesslist file corresponding to the specified list.
static WFSKrnResult WFSKrnInitializeAccessList(const AreaInfo* area, WFSKrnAccessList** list)
{
    int result;
    utf8 filename[WFS_MAX_FILE_NAME_SIZE];
    if ((*list)->ref != 1)
    {
        WFSKRN_ACL_ASSERT((*list)->ref == 1);
    }

    if (ACL_MAX_ENTRIES < WFSKrnAccessListGetNum(*list))
    {
        WFSKRN_ACL_REPORT("WFSKrnInitializeAccessList(): too many entries.\n");
        return WFSKRN_RESULT_ACL_MAX_ENTRIES;
    }

    result = WFSKrnConvertAccessListToFilename((*list), filename, WFS_MAX_FILE_NAME_SIZE);
    if (result < 0)
    {
        return WFSKRN_RESULT_ACL_INVALID_PARAMETER;
    }

    result = WFSKrnAccessListCreateFile(area, filename, list);
    if (result < 0)
    {
        return WFSKRN_RESULT_ACL_FILE;
    }
    return WFSKRN_RESULT_OK;
}

static void CloseAccessList(const AreaInfo* area, WFSKrnAccessList* list)
{
    WFSAclLockMutex(&Mutex(area));
    WFSKrnAccessListFree(area, list);
    WFSAclUnlockMutex(&Mutex(area));
}

static WFSKrnResult CheckCreateParams(u32 aclType, WFSTitleId entryCreatorId, WFSTitleId titleId, u32 nFlags)
{
    s32 cl;
    if (aclType & ~ACL_FLAG_DIRECTORY)
    {
        // support directory flag only.
        return WFSKRN_RESULT_ACL_INVALID_PARAMETER;
    }

    if (entryCreatorId == WFS_ROOT_ID &&
        titleId == WFS_ROOT_ID)
    {
        return WFSKRN_RESULT_OK;
    }

    if (aclType & (u32)ACL_FLAG_DIRECTORY)
    {
        if (nFlags & DP_RESERVED)
        {
            return WFSKRN_RESULT_ACL_INVALID_PARAMETER;
        }
    }
    else
    {
        if (nFlags & FP_RESERVED)
        {
            return WFSKRN_RESULT_ACL_INVALID_PARAMETER;
        }
    }

    if (titleId == WFS_ROOT_ID && entryCreatorId != WFS_ROOT_ID)
    {
        // cannot create root entry by others.
        return WFSKRN_RESULT_PERMISSION;
    }

    if (ACL_GET_CL_RELATIVE(nFlags, (int*)&cl) != WFSKRN_RESULT_OK)
    {
        return WFSKRN_RESULT_PERMISSION;
    }
    if (entryCreatorId == titleId)
    {
        if (cl <= 0)
        {
            return WFSKRN_RESULT_PERMISSION;
        }
    }
    else if (entryCreatorId != WFS_ROOT_ID && 0 < cl)
    {
        return WFSKRN_RESULT_PERMISSION;
    }
    return WFSKRN_RESULT_OK;
}

static WFSKrnResult CreateAccessList(const AreaInfo* area, u32 aclType,
    WFSTitleId entryCreatorId, WFSTitleId titleId, u32 nFlags, WFSKrnAclHandle* newHandle)
{
    WFSKrnAccessList* list = NULL;
    WFSKrnResult result;
    WFSKrnAccessListEntry* entry;

    result = CheckCreateParams(aclType, entryCreatorId, titleId, nFlags);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    WFSAclLockMutex(&Mutex(area));
    if (titleId == WFS_ROOT_ID)
    {
        // The root implicitly has all permissions for all objects.
        result = WFSKrnAccessListAlloc(area, 0, &list);
    }
    else
    {
        result = WFSKrnAccessListAlloc(area, 1, &list);
        if (result == WFSKRN_RESULT_OK)
        {
            entry = GetFirstEntry(list);
            entry->id = titleId;
            entry->ecId = entryCreatorId;
            entry->pm = nFlags;
        }
    }

    if (result == WFSKRN_RESULT_OK)
    {
        list->handle = aclType;
        result = WFSKrnInitializeAccessList(area, &list);
        if (result != WFSKRN_RESULT_OK)
        {
            WFSKrnMemoryBlockFree(area, (WFSKrnAccessListBlock*) (list), WFSKRN_BLOCKS_IN_LIST(list)); // [check] must test this code.
        }
    }

    if (result == WFSKRN_RESULT_OK)
    {
        *newHandle = list->handle;
        CloseAccessList(area, list);
    }
    WFSKRN_ACL_ASSERT(VerifyCacheList(area));
    WFSAclUnlockMutex(&Mutex(area));
    return result;
}

static WFSKrnResult CreateDefaultAccessList(const AreaInfo* area)
{
    WFSKrnResult result;
    WFSKrnAclHandle handle;
    result = CreateAccessList(area, 0, WFS_ROOT_ID, 0, 0, &handle);
    if (result != WFSKRN_RESULT_OK || handle != 0)
    {
        return WFSKRN_RESULT_ACL_HANDLE;
    }
    return WFSKRN_RESULT_OK;
}

WFSKrnResult WFSKrnDebugCreateAccessList(const AreaInfo* area, u32 aclType,
    WFSTitleId entryCreatorId, WFSTitleId titleId, u32 nFlags, WFSKrnAclHandle* newHandle)
{
    return CreateAccessList(area, aclType, entryCreatorId, titleId, nFlags, newHandle);
}

WFSKrnResult WFSKrnAccessListInit(AreaInfo* area)
{
    WFSKrnResult result;

    WFSKRN_ACL_ASSERT(area);
    WFSKRN_ACL_ASSERT(area->accessListCb);

    if (Initialized(area))
    {
        WFSKRN_ACL_REPORT("WARNING: WFSKrnAccessListInit(): ACL is already initialized!\n");
        return WFSKRN_RESULT_LIB_ALREADY_INITIALIZED;
    }

    if (WFSKrnAccessListDirectoryInit(area) < 0)
    {
        return WFSKRN_RESULT_ACL_ERROR;
    }

    // allocator for accesslists.
    if (WFSKrnMemoryBlockInit(area) < 0)
    {
        return WFSKRN_RESULT_OUT_OF_MEMORY;
    }

    WFSAclInitMutex(&Mutex(area));
    AccessListCacheInit(area);

    result = CreateDefaultAccessList(area);
    if (result == WFSKRN_RESULT_OK)
    {
        Initialized(area) = TRUE;
    }
    return result;
}

WFSKrnResult WFSKrnAccessListOpen(const AreaInfo* area)
{
    //WFSKrnResult result;

    WFSKRN_ACL_ASSERT(area);
    WFSKRN_ACL_ASSERT(area->accessListCb);

    if (Initialized(area))
    {
        WFSKRN_ACL_REPORT("WARNING: WFSKrnAccessListOpen(): ACL is already initialized!\n");
        return WFSKRN_RESULT_LIB_ALREADY_INITIALIZED;
    }

    // allocator for accesslists.
    if (WFSKrnMemoryBlockInit(area) < 0)
    {
        return WFSKRN_RESULT_OUT_OF_MEMORY;
    }

    WFSAclInitMutex(&Mutex(area));
    AccessListCacheInit(area);

    Initialized(area) = TRUE;
    return WFSKRN_RESULT_OK;
}



WFSKrnResult WFSKrnAccessListExit(const AreaInfo* area)
{
    if (WFSKrnAccessListIndexListExit(area) < 0)
    {
        return WFSKRN_RESULT_AREA_ERROR;
    }
    return WFSKRN_RESULT_OK;
}

// lookup
static WFSKrnAccessList* OpenAccessList(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WFSKrnAccessList* list;
    WFSAclLockMutex(&Mutex(area));

    list = WFSKrnAccessListLookupByIndex(area, handle);
    if (list)
    {
        if (list->ref == 0)
        {
            WFSKrnAccessListCacheFreeListRemove(area, list);
        }
        ++list->ref;
    }
    else
    {
        list = WFSKrnAccessListLookupFile(area, handle);
        WFSKRN_ACL_ASSERT(!list || list->handle == handle);
    }
    WFSAclUnlockMutex(&Mutex(area));
    return list;
}

static WFSBool LookupEntryInList(const WFSKrnAccessList* list, const WFSKrnAccessListEntry* target, WFSKrnAccessListEntry** editPosition)
{
    WFSKrnAccessListEntry* entry;
    u32 numEntry;
    WFSBool checkEntryCreator = (WFSBool) FALSE;

    entry = (WFSKrnAccessListEntry*) (list+1);
    numEntry = WFSKrnAccessListGetNum(list);

    while (numEntry--)
    {
        if (entry->id == target->id && !checkEntryCreator)
        {
            checkEntryCreator = (WFSBool) TRUE;
        }

        if (checkEntryCreator && entry->ecId == target->ecId)
        {
            *editPosition = entry;
            return TRUE;
        }

        if (target->id < entry->id ||
            (checkEntryCreator && target->ecId < entry->ecId))
        {
            *editPosition = entry;
            return FALSE; // target not found in list.
        }
        ++entry;
    }
    *editPosition = entry;
    return FALSE; // not found in list.
}

static void SetEntry(WFSKrnAccessList* newList, u32 index, WFSKrnAccessListEntry* newEntry, u32 nMask)
{
    WFSKrnAccessListEntry* entry = GetFirstEntry(newList);
    entry += index;
    entry->id = newEntry->id;
    entry->ecId = newEntry->ecId;
    entry->pm = (newEntry->pm & nMask) | (entry->pm & ~nMask);
}

static void ModifyEntry(const WFSKrnAccessList* list, WFSKrnAccessList* newList,
    u32 index, WFSKrnAccessListEntry* entry, u32 nMask)
{
    memmove(GetFirstEntry(newList), GetFirstEntry(list),
        WFSKrnAccessListGetNum(list) * sizeof(WFSKrnAccessListEntry));
    SetEntry(newList, index, entry, nMask);
}

static void RemoveEntry(const WFSKrnAccessList* list, WFSKrnAccessList* newList, u32 index)
{
    if (0 < index)
    {
        memmove(GetFirstEntry(newList), GetFirstEntry(list), index * sizeof(WFSKrnAccessListEntry));
    }
    // shift
    if (index < WFSKrnAccessListGetNum(list))
    {
        memmove(GetFirstEntry(newList) + index, GetFirstEntry(list) + index + 1,
            (WFSKrnAccessListGetNum(newList) - index) * sizeof(WFSKrnAccessListEntry));
    }
}

static void AddEntry(const WFSKrnAccessList* list, WFSKrnAccessList* newList,
    u32 index, WFSKrnAccessListEntry* entry, u32 nMask)
{
    if (0 < index)
    {
        memmove(GetFirstEntry(newList), GetFirstEntry(list), index * sizeof(WFSKrnAccessListEntry));
    }
    SetEntry(newList, index, entry, nMask);
    // shift
    if (index < WFSKrnAccessListGetNum(list))
    {
        memmove(GetEntry(newList, index + 1), GetEntry(list, index),
            (WFSKrnAccessListGetNum(list) - index) * sizeof(WFSKrnAccessListEntry));
    }
}

static WFSKrnResult BranchAccessList(const AreaInfo* area, const WFSKrnAccessList* list,
    WFSKrnAccessListEntry* entry, u32 nMask, int mode, WFSKrnAccessList** newList)
{
    WFSKrnResult result;
    WFSKrnAccessListEntry* target;
    u32 entryIndex; // index of the entry to be modified.
    u32 numUsers;
    u32 newPm;
    WFSBool found;

    numUsers = WFSKrnAccessListGetNum(list);
    found = LookupEntryInList(list, entry, &target);

    if (mode == BRANCH_MODE_SET)
    {
        if (found)
        {
            newPm = (entry->pm & nMask) | (target->pm & ~nMask);
            if (target->pm == newPm)
            {
                *newList = (WFSKrnAccessList*) list;
                return WFSKRN_RESULT_OK;
            }
            mode = BRANCH_MODE_MODIFY;
        }
        else
        {
            newPm = (entry->pm & nMask);
            ++numUsers; // add entry.
            nMask = 0xffffffff; // zero clear except for newPm.
            mode = BRANCH_MODE_ADD;
        }
    }
    else if (found)
    {
        --numUsers; // remove entry.
        mode = BRANCH_MODE_DELETE;
    }
    else
    {
        // the entry to be deleted is not found.
        *newList = NULL;
        return WFSKRN_RESULT_NOT_FOUND;
    }

    result = WFSKrnAccessListAlloc(area, numUsers, newList);
    if (result != WFSKRN_RESULT_OK)
    {
        *newList = NULL;
        return result; // [check]
    }
    (*newList)->handle = ACL_TYPE(list); // only copy accesslist type.
                                         // handle itself is assigned in WFSKrnAccessListCreateFile().
    entryIndex = target - GetFirstEntry(list);
    switch (mode)
    {
      case BRANCH_MODE_ADD:
        AddEntry(list, *newList, entryIndex, entry, nMask);
        break;
      case BRANCH_MODE_MODIFY:
        ModifyEntry(list, *newList, entryIndex, entry, nMask);
        break;
      case BRANCH_MODE_DELETE:
        RemoveEntry(list, *newList, entryIndex);
        break;
    }
    return WFSKRN_RESULT_OK;
}

static WFSKrnResult GetModifiedAccessList(const AreaInfo* area, WFSKrnAccessList* list,
    WFSKrnAccessListEntry* entry, u32 nMask, int mode, WFSKrnAccessList** newList)
{
    WFSKrnAclHandle handle;
    WFSKrnResult result;

    result = BranchAccessList(area, list, entry, nMask, mode, newList);
    if (!(*newList) || (*newList) == list)
    {
        return result;
    }

    result = WFSKrnInitializeAccessList(area, newList);
    if (result != WFSKRN_RESULT_OK)
    {
        WFSKrnMemoryBlockFree(area, (WFSKrnAccessListBlock*) (*newList), WFSKRN_BLOCKS_IN_LIST(*newList));
        *newList = 0;
        return result;
    }

    handle = list->handle;
    // decrement the reference count of the trunk (list).
    WFSKrnAccessListFree(area, list); // reduce the reference count of the cache to zero
                                      // before deleting an accesslist file.
    WFSKrnReleaseAccessList(area, handle);
    return result;
}

WFSKrnResult WFSKrnAccessListAlloc(const AreaInfo* area, u32 numEntries, WFSKrnAccessList** newList)
{
    WFSKrnAccessList* list = NULL;
    WFSKrnAccessList* freeCache;

    if (ACL_MAX_ENTRIES < numEntries)
    {
        return WFSKRN_RESULT_ACL_MAX_ENTRIES;
    }

    while (!list)
    {
        list = (WFSKrnAccessList*) WFSKrnMemoryBlockAlloc(area,
            (int) (sizeof(WFSKrnAccessList)/sizeof(WFSKrnAccessListBlock) +
                numEntries * (sizeof(WFSKrnAccessListEntry)/sizeof(WFSKrnAccessListBlock))));
        if (list)
        {
            WFSKrnAccessListSetNum(list, numEntries);
            list->ref = 1;
            list->handle = WFSKRN_ACCESSLIST_FREEINDEX;
            break;
        }

        // no free memory block.
        // Free an entry in the cache freelist.
        if (CacheFreeListIsEmpty())
        {
            WFSKRN_ACL_REPORT("WFSKrnAccessListAlloc(): cannot allocate momory blocks for the new cache. (numEntries %d)\n", numEntries);
            PrintMemoryBlock(area); // debug code.
            break;
        }
        /* [check] should wait until any list is inserted into the freelist.
        while (CacheFreeListIsEmpty())
        {
            // wait
        }
        */

        freeCache = CacheFreeListRemoveHead(area);
        if (freeCache->handle != WFSKRN_ACCESSLIST_FREEINDEX)
        {
            WFSKrnAccessListCacheRemove(area, freeCache);
        }
        WFSKrnMemoryBlockFree(area, (WFSKrnAccessListBlock*) freeCache, WFSKRN_BLOCKS_IN_LIST(freeCache));

        WFSKRN_ACL_ASSERT(VerifyCacheList(area));
    }

    if (!list)
    {
        return WFSKRN_RESULT_ACL_CACHE;
    }

    *newList = list;
    return WFSKRN_RESULT_OK;
}

int WFSKrnAccessListFree(const AreaInfo* area, WFSKrnAccessList* list)
{
    u32 ref = --list->ref;
    if (ref == 0)
    {
        // insert the list into the cache freelist.
        // The cache remains in the cache list.
        CacheFreeListInsertTail(area, list);
        // [check] NYI. notifyAll.
    }

    return ref;
}

// Remove the access list from disk if the refernce count is zero.
WFSKrnResult WFSKrnReleaseAccessList(const AreaInfo* area, WFSKrnAclHandle handle)
{
    u32 ref;
    WFSKrnResult result = WFSKRN_RESULT_ACL_ERROR;
    WFSKrnAccessList* cache;

    WFSKRN_ACL_ASSERT(Initialized(area) == TRUE);

    if ((handle & ACL_TYPE_MASK) == 0)
    {
        // index 0 is reserved for the empty accesslist.
        return WFSKRN_RESULT_OK;
    }

    WFSAclLockMutex(&Mutex(area));
    result = WFSKrnAccessListGetFileReferenceCount(area, handle, &ref);
    if (result == WFSKRN_RESULT_OK)
    {
        if (ref == 1 && WFSKrnAccessListLookup(area, handle))
        {
            // All caches must be closed before deleting the accesslist file.
            result = WFSKRN_RESULT_ACL_CACHE;
        }
        else if (0 < --ref)
        {
            result = WFSKrnAccessListSetFileReferenceCount(area, handle, ref);
        }
        else
        {
            cache = WFSKrnAccessListLookupByIndex(area, handle);

            result = WFSKrnAccessListDeleteFile(area, handle);
            if (result == WFSKRN_RESULT_OK && cache)
            {
                WFSKRN_ACL_ASSERT(cache->ref == 0);
                WFSKRN_ACL_ASSERT(cache->handle != WFSKRN_ACCESSLIST_FREEINDEX);
                WFSKrnAccessListCacheRemove(area, cache);
                WFSKrnFreeAccessListIndex(area, cache->handle);
                CacheFreeListInvalidate(area, cache); // set handle to WFSKRN_ACCESSLIST_FREEINDEX.
            }
        }
    }
    WFSAclUnlockMutex(&Mutex(area));
    return result;
}

/*
static u32 GetRootPermission(const WFSKrnAccessList* list)
{
    if (IsDirectory(list))
    {
        return WFS_PERM_CL_3 | WFS_PERM_DIR_FULL;
    }
    return WFS_PERM_CL_3 | WFS_PERM_FILE_FULL;
}

static int SearchGroup(const WFSKrnAccessList* list, const WFSTitleId groupId, const WFSTitleId titleId)
{
    if (groupId == WFS_EVERYONE_ID)
    {
        return 0;
    }

    // WFSAclLockMutex(&Mutex(area));
    // NYI.
    // WFSAclUnlockMutex(&Mutex(area));

    return -1;
}
*/

static WFSBool IsAllowed(const AreaInfo* area, WFSKrnAccessList* list,
    WFSTitleId entryCreatorId, WFSTitleId titleId, u32 newPermission, u32 mask)
{
    s32 cl;
    u32 reserved;

    // check mask.
    if (IsDirectory(list))
    {
        reserved = (u32) DP_RESERVED;
    }
    else
    {
        reserved = (u32) FP_RESERVED;
    }

    if (reserved & mask)
    {
        return (WFSBool) FALSE;
    }

    if (mask & WFS_PERM_CL_MASK)
    {
        // check control level.
        if (ACL_GET_CL_RELATIVE(newPermission, (int*)&cl) != WFSKRN_RESULT_OK)
        {
            return (WFSBool) FALSE;
        }

        if (entryCreatorId == titleId)
        {
            if (0 < cl)
            {
                return (WFSBool) FALSE;
            }
        }
        else
        {
            if (entryCreatorId != WFS_ROOT_ID && 0 <= cl)
            {
                return (WFSBool) FALSE;
            }
        }
    }
    return (WFSBool) TRUE;;
}

static WFSKrnResult SetPermission(const AreaInfo* area, WFSKrnAccessList* list,
    WFSTitleId entryCreatorId, WFSTitleId titleId, u32 newPermission, u32 mask, WFSKrnAclHandle* handle)
{
    WFSKrnResult result = WFSKRN_RESULT_PERMISSION;
    WFSKrnAccessList* newList;

    WFSKRN_ACL_ASSERT(VerifyList(list));
    if (IsAllowed(area, list, entryCreatorId, titleId, newPermission, mask))
    {
        WFSKrnAccessListEntry newEntry;
        newEntry.id = titleId;
        newEntry.ecId = entryCreatorId;
        newEntry.pm = newPermission;

        WFSAclLockMutex(&Mutex(area));
        result = GetModifiedAccessList(area, list, &newEntry, mask, BRANCH_MODE_SET, &newList);
        WFSAclUnlockMutex(&Mutex(area));

        if (result == WFSKRN_RESULT_OK && newList)
        {
            *handle = newList->handle;
            CloseAccessList(area, newList);
            WFSAclLockMutex(&Mutex(area));
            WFSKRN_ACL_ASSERT(VerifyList(newList));
            WFSAclUnlockMutex(&Mutex(area));
            return WFSKRN_RESULT_OK;
        }
    }
    CloseAccessList(area, list);
    WFSKRN_ACL_ASSERT(VerifyList(list));
    return result;
}

static WFSKrnResult CheckSetParams(WFSTitleId entryCreatorId, WFSTitleId titleId)
{
    // [check] any restriction?
    if (titleId == WFS_ROOT_ID) // The root always has all permissions.
    {
        return WFSKRN_RESULT_PERMISSION;
    }
    return WFSKRN_RESULT_OK;
}

WFSKrnResult WFSKrnSetAccessList(const AreaInfo* area, WFSKrnAclHandle* handle,
        u32 nFlags, u32 nMask, WFSTitleId entryCreatorId, WFSTitleId titleId)
{
    WFSKRN_ACL_ASSERT(Initialized(area) == TRUE);

    WFSKrnAccessList* list;
    WFSKrnResult result = WFSKRN_RESULT_PERMISSION;

    if (CheckSetParams(entryCreatorId, titleId) == WFSKRN_RESULT_OK)
    {
        list = OpenAccessList(area, *handle);
        if (list)
        {
            result = SetPermission(area, list,
                entryCreatorId, titleId, nFlags, nMask, handle);
        }
    }
    return result;
}

static WFSKrnResult DeletePermissions(const AreaInfo* area, WFSKrnAccessList* list,
    WFSTitleId entryCreatorId, WFSTitleId titleId, WFSKrnAclHandle* handle)
{
    WFSKrnResult result = WFSKRN_RESULT_PERMISSION;
    WFSKrnAccessList* newList;
    WFSKrnAccessListEntry entry;

    // [check] any restriction?

    entry.id = titleId;
    entry.ecId = entryCreatorId;

    WFSAclLockMutex(&Mutex(area));
    result = GetModifiedAccessList(area, list, &entry, 0, BRANCH_MODE_DELETE, &newList);
    WFSAclUnlockMutex(&Mutex(area));

    if (newList)
    {
        *handle = newList->handle;
        CloseAccessList(area, newList);
        WFSAclLockMutex(&Mutex(area));
        WFSKRN_ACL_ASSERT(VerifyList(newList));
        WFSAclUnlockMutex(&Mutex(area));
        return WFSKRN_RESULT_OK;
    }
    result = WFSKRN_RESULT_ACL_ERROR;
    CloseAccessList(area, list);
    return result;
}

WFSKrnResult WFSKrnDeletePermissions(const AreaInfo* area,
    WFSKrnAclHandle* handle, WFSTitleId entryCreatorId, WFSTitleId titleId)
{
    WFSKrnAccessList* list;
    WFSKrnResult result = WFSKRN_RESULT_NOT_FOUND;

    WFSKRN_ACL_ASSERT(Initialized(area) == TRUE);

    // [check] any restriction?

    list = OpenAccessList(area, *handle);
    if (list)
    {
        result = DeletePermissions(area, list,
            entryCreatorId, titleId, handle);
    }
    return result;
}

static s32 GetAccessList(const AreaInfo* area, WFSKrnAclHandle handle, WFSKrnAccessList* list,
    WFSAccessListEntry *pAccessList, u32 nMaxElements)
{
    s32 result;
    u32 num;
    int i;
    WFSKrnAccessListEntry* entry;

    WFSKRN_ACL_ASSERT(Initialized(area) == TRUE);

    num = WFSKrnAccessListGetNum(list);
    if (num <= nMaxElements)
    {
        entry = GetFirstEntry(list);
        for (i = 0; i < num; ++i)
        {
            pAccessList->titleId = entry->id;
            pAccessList->entryCreatorId = entry->ecId;
            pAccessList->permissions = entry->pm;
            ++pAccessList;
            ++entry;
        }
        result = (s32) num;
    }
    else
    {
        result = (s32) WFSKRN_RESULT_ACL_BUFFER_TOO_SMALL;
    }
    return result;
}

s32 WFSKrnGetAccessList(const AreaInfo* area, WFSKrnAclHandle handle, WFSAccessListEntry *pAccessList, u32 nMaxElements)
{
    s32 result;
    WFSKrnAccessList* list;
    
    if (!area)
    {
        return WFSKRN_RESULT_INVALID; // [check] workaround to prevent WFSGetAccessList("/", ...) from throwing exception.
    }

    list = OpenAccessList(area, handle);
    if (list)
    {
        result = GetAccessList(area, handle, list, pAccessList, nMaxElements);
        CloseAccessList(area, list);
    }
    else
    {
        result = (s32) WFSKRN_RESULT_ACL_ENTRY_NOT_FOUND;
    }
    WFSAclLockMutex(&Mutex(area));
    WFSKRN_ACL_ASSERT(VerifyCacheList(area));
    WFSAclUnlockMutex(&Mutex(area));

    return result;
}

void PrintMemoryBlock(const AreaInfo* area)
{
    WFSKrnAccessListBlock* block;
    int count;
    WFSKrnAccessList* list;

    block = MemoryBlockAddress(area);
    while (block < MemoryBlockAddress(area) + NumMemoryBlock(area))
    {
        if (block->type == WFSKRN_ACCESSLIST_FREEBLOCK)
        {
            count = GetFreeBlockNum(block);
            WFSKRN_ACL_ASSERT(0 < count && count <= area->accessListCb->numBlock);
            block += count;

            if (block->nextBlock != ACL_NO_ENTRY)
            {
                while (count--)
                {
                    WFSKRN_ACL_REPORT("_");
                }
            }
            else
            {
                // allocated.
                while (count--)
                {
                    WFSKRN_ACL_REPORT("*");
                }
            }
        }
        else
        {
            list = (WFSKrnAccessList*) block;
            if (list->handle == WFSKRN_ACCESSLIST_FREEINDEX)
            {
                WFSKRN_ACL_REPORT("FFF");
            }
            else
            {
                WFSKRN_ACL_REPORT("AAA");
            }
            block += sizeof(WFSKrnAccessList)/sizeof(WFSKrnAccessListBlock);
            count = WFSKrnAccessListGetNum(list) * (sizeof(WFSKrnAccessListEntry)/sizeof(WFSKrnAccessListBlock));
            WFSKRN_ACL_ASSERT(0 <= count && count <= area->accessListCb->numBlock);
            block += count;
            while (count--)
            {
                WFSKRN_ACL_REPORT("E");
            }
        }
    }
    WFSKRN_ACL_REPORT("\n");
}

WFSBool VerifyMemoryBlock(const AreaInfo* area)
{
    int numFreeBlock;
    WFSKrnAccessListBlock* free;
    WFSBool result = (WFSBool) FALSE;

    free = FreeBlockAddress(area);
    numFreeBlock = 0;

    while (free < MemoryBlockAddress(area) + NumMemoryBlock(area))
    {
        if (free->type != WFSKRN_ACCESSLIST_FREEBLOCK)
        {
            goto VERIFY_ERROR;
        }
        if (NumMemoryBlock(area) < free->nextBlock)
        {
            goto VERIFY_ERROR;
        }
        if (GetFreeBlockNum(free) <= 0)
        {
            goto VERIFY_ERROR;
        }
        if (NumMemoryBlock(area) < GetFreeBlockNum(free))
        {
            goto VERIFY_ERROR;
        }
        numFreeBlock += GetFreeBlockNum(free);
        free = MemoryBlockAddress(area) + free->nextBlock;
    }
    result = (WFSBool) TRUE;

VERIFY_ERROR:
    return result;
}

static WFSBool VerifyList(const WFSKrnAccessList* list)
{
    u32 numEntry = WFSKrnAccessListGetNum(list);
    int i;
    int j;
    WFSKrnAccessListEntry* entry0 = (WFSKrnAccessListEntry*) (list+1);
    for (i = 0; i < numEntry; ++i)
    {
        WFSKrnAccessListEntry* entry1 = (WFSKrnAccessListEntry*) (list+1);
        for (j = 0; j < numEntry; ++j)
        {
            if (entry0 != entry1 && entry0->id == entry1->id && entry0->ecId == entry1->ecId )
            {
                return (WFSBool) FALSE;
            }
            ++entry1;
        }
        ++entry0;
    }
    return (WFSBool) TRUE;
}

static void DumpCache(const AreaInfo* area)
{
    WFSKrnAccessListLinkList* CacheList = &area->accessListCb->cache;
    WFSKrnAccessList* list = CacheList->head;

    osTPrintf("[CacheList]->");
    while (list)
    {
        osTPrintf("[%08x]->", list->handle);
        list = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextBlock);
    }
    osTPrintf("NULL\n");
}

static WFSBool VerifyCacheList(const AreaInfo* area)
{
    WFSKrnAccessList* head = AccessListCache(area).head;
    WFSKrnAccessList* tail = AccessListCache(area).tail;
    WFSKrnAccessList* list = head;
    WFSKrnAccessList* next;

    int count = 256; // [check]
    WFSKRN_ACL_ASSERT(!tail || ACL_CACHE_ADDRESS(area, tail->nextBlock) == NULL);

    while (list)
    {
        next = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextBlock);
        if (next)
        {
            WFSKRN_ACL_ASSERT(next->prevBlock == ACL_CACHE_BLOCK(area, (WFSKrnAccessListBlock*) list));
        }

        if (next == head)
        {
            WFSKRN_ACL_REPORT("*** [Cache List] chain loop!!! *** \n");
            DumpCache(area);
            return (WFSBool) FALSE;
        }
        list = next;
        WFSKRN_ACL_ASSERT(0 < count--);
    }

    head = AccessListCacheFreeList(area).head;
    tail = AccessListCacheFreeList(area).tail;
    list = head;

    count = 256; // [check]
    WFSKRN_ACL_ASSERT(!tail || ACL_CACHE_ADDRESS(area, tail->nextFreeCacheBlock) == 0);
    while (list)
    {
        next = (WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, list->nextFreeCacheBlock);
        if (next)
        {
            WFSKRN_ACL_ASSERT((WFSKrnAccessList*) ACL_CACHE_ADDRESS(area, next->prevFreeCacheBlock) == list);
        }
        if (next == head)
        {
            WFSKRN_ACL_REPORT("*** [Free List] chain loop!!! *** \n");
            return (WFSBool) FALSE;
        }
        list = next;
        WFSKRN_ACL_ASSERT(0 < count--);
    }

    return (WFSBool) TRUE;
}

static void DumpListHeader(const AreaInfo* area, const WFSKrnAccessList* list)
{
    WFSKRN_ACL_REPORT("---------------- List (0x%p) ----------------\n", list);
    WFSKRN_ACL_REPORT("     num    0x%x\n"
                      "     handle 0x%x\n"
                      "     ref    0x%x\n", list->num, list->handle, list->ref);
}

static void DumpAcl(WFSKrnAccessListEntry* entry, u32 num)
{
    int cl;
    WFSKRN_ACL_REPORT    ("ID          EntryCreator CL  D LAA RWC\n");
    while (num--)
    {
		WFSKrnResult ret = ACL_GET_CL_RELATIVE(entry->pm, &cl);
        WFSKRN_ACL_ASSERT( ret == WFSKRN_RESULT_OK );
        WFSKRN_ACL_REPORT("0x%08x  0x%08x  [%2d] %1d %1d%1d%1d %1d%1d%1d\n",
            entry->id, entry->ecId,
			(ret == WFSKRN_RESULT_OK) ? cl : 0,
            entry->pm & WFS_PERM_CHANGE_SIZE     ? 1:0,
            entry->pm & WFS_PERM_WRITE           ? 1:0,
            entry->pm & WFS_PERM_READ            ? 1:0,
            entry->pm & WFS_PERM_ADD_OR_DELETE_FILE ? 1:0,
            entry->pm & WFS_PERM_ADD_OR_DELETE_SUBDIR  ? 1:0,
            entry->pm & WFS_PERM_LIST            ? 1:0,
            entry->pm & WFS_PERM_DELETE_OR_MOVE     ? 1:0);
        ++entry;
    }
    WFSKRN_ACL_REPORT("---------------------------------------------\n");
}

static void DumpList(const AreaInfo* area, const WFSKrnAccessList* list)
{
    u32 num;
    WFSKrnAccessListEntry* entry;

    DumpListHeader(area, list);

    num = WFSKrnAccessListGetNum(list);
    entry = GetFirstEntry(list);
    DumpAcl(entry, num);
}

void WFSKrnDebugPrintAccessList(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WFSKrnAccessList* list;
    WFSAclLockMutex(&Mutex(area));
    list = OpenAccessList(area, handle);
    if (!list)
    {
        WFSKRN_ACL_REPORT("Error: no accesslist %x\n", handle);
        return;
    }
    CloseAccessList(area, list); // reduce ref before dump.
    DumpList(area, list);
    WFSAclUnlockMutex(&Mutex(area));
}

