/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfsKrn_Permission_File.cpp - Functions for permissions.

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_File.cpp,v $
  Revision 1.24  2008/10/08 08:34:15  ueno
  Fixed for WFSDEV.

  Revision 1.23  2008/08/19 08:12:23  ueno
  Fixed error handling in WFSKrnAccessListCreateFile().

  Revision 1.22  2008/08/04 04:02:27  ueno
  Modified to use TransInfo.

  Revision 1.21  2008/07/23 04:17:14  ueno
  Reduced memory size required for accesslist.

  Revision 1.20  2008/06/05 14:05:22  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.19  2008/06/05 08:07:48  ueno
  Revised permission APIs.

  Revision 1.18  2008/05/30 08:50:07  ueno
  Cleanup.

  Revision 1.17  2008/05/28 02:12:49  ueno
  Cleanup.

  Revision 1.16  2008/05/22 06:36:12  ueno
  Ported permission APIs to IOP.

  Revision 1.15  2008/05/20 07:08:38  ueno
  Added debug code.

  Revision 1.14  2008/05/19 11:00:04  ueno
  Updated comments.

  Revision 1.13  2008/05/19 10:11:52  ueno
  Revised not to use hardlink.

  Revision 1.12  2008/05/19 05:30:50  ueno
  Defined accesslist meta-data types.

  Revision 1.11  2008/05/16 11:16:17  ueno
  Implemented permission APIs with directory module.

  Revision 1.10  2008/05/15 13:31:54  ueno
  Updated.

  Revision 1.9  2008/05/08 08:07:44  ueno
  Revised AclReadFile() and AclWriteFile() to set offset.

  Revision 1.8  2008/05/06 07:40:39  ueno
  Modified to support WFS (not WFSDEV) codes.

  Revision 1.7  2008/05/06 03:17:51  ueno
  Added AclGetFileSize() to wrap Windows API..

  Revision 1.6  2008/05/06 02:26:03  ueno
  Fixed ifdef.

  Revision 1.5  2008/04/29 06:25:20  ueno
  Cleanup.

  Revision 1.4  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.3  2008/04/25 04:40:07  ueno
  Added support for WFSDEV.

  Revision 1.26  2008/04/23 08:59:46  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.25  2008/04/15 07:36:01  ueno
  Changed return type to WFSKrnResult.

  Revision 1.24  2008/04/11 06:16:47  ueno
  Defined WFSAclFileHandle.

  Revision 1.23  2008/04/08 06:39:31  ueno
  Cleanup.

  Revision 1.22  2008/04/08 04:28:50  ueno
  Modified to invert bits before escape sequence from accesslist content to filename.

  Revision 1.21  2008/04/03 11:40:44  ueno
  Replaced accesslist index by accesslist handle.

  Revision 1.20  2008/04/03 06:35:08  ueno
  Cleanup.

  Revision 1.19  2008/04/01 08:24:54  ueno
  Fixed cache reference count.

  Revision 1.18  2008/03/14 08:02:53  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.17  2008/03/11 12:10:41  ueno
  Cleanup.

  Revision 1.16  2008/03/11 11:10:57  ueno
  Moved accesslist functions into wfskrn_Permission_AccessList.cpp.

  Revision 1.15  2008/03/11 10:48:22  ueno
  Cleanup.

  Revision 1.14  2008/03/11 09:51:54  ueno
  Fixed EscapeAccessList() to make a filename from UTF-16 characters (0x3300-0x33ff).

  Revision 1.13  2008/03/11 06:53:06  ueno
  Added WFSAccessListSetPermission() and GetPermission().

  Revision 1.12  2008/03/07 05:32:07  ueno
  Modified WFSCreateAccessList() to be able to create an accesslist for directories.

  Revision 1.11  2008/03/06 10:45:28  ueno
  Added debug codes.

  Revision 1.10  2008/02/29 07:29:10  ueno
  Added Windows api wrappers.

  Revision 1.9  2008/02/28 13:11:01  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.8  2008/02/28 07:01:52  ueno
  Fixed escape sequence for accesslist filename.

  Revision 1.7  2008/02/28 04:36:32  ueno
  Cleanup.

  Revision 1.6  2008/02/27 13:12:33  ueno
  Revised to store a filename of an accesslist in a file (*.name).

  Revision 1.5  2008/02/27 10:05:36  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.4  2008/02/27 04:29:26  ueno
  Modified WFSAccessListAddEntry() and RemoveEntry() to delete accesslist file when the reference count is zero.

  Revision 1.3  2008/02/26 14:11:43  ueno
  Added WFSAccessListAddEntry() and WFSAccessListRemoveEntry() to edit accesslist.

  Revision 1.2  2008/02/25 09:43:57  ueno
  Added support for accesslist cache.

  Revision 1.1  2008/02/22 05:16:14  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#include "wfs_Defs.h"
#include "wfskrn_Permission_AccessList.h"

#ifdef WFSDEV
#include "wfskrn_Permission_Win32.h"
#else // WFSDEV
#include "wfskrn_Permission_Iop.h"
#endif // WFSDEV

#include "wfskrn_Permission_File.h"
#include "wfskrn_Permission_Private.h"

#include <string.h>

#define AclIndexType(x) ((x) & ~ACL_TYPE_MASK)
#define AccessListFileHeaderIndex(x)      ((x).index & ACL_TYPE_MASK)

/*
  * Accesslist file
    The name of an accesslist file represents the content of the accesslist (title IDs and the permissions).

    Accesslist file contains the index of the accesslist.

  - WFS
    accesslist file: <null-terminated utf8 strings>
    +-------+
    | index |
    +-------+

  - WFSDEV
    accesslist file: <null-terminated utf16 strings>
    +-------+
    | index |
    +-------+

  * How to access an accesslist file from the accesslist index.
  - WFS and WFSDEV
    accesslist index x (u32)
       |
       | read accesslist name file.
       |
       |      accesslist name file: <x>
       |      +-------+-------+--------------------------+
       |      |  num  |  ref  | filename of accesslist x |
       |      +-------+-------+--------------------------+
       |
       | get the filename of the accesslist file.
       V
    access the accesslist file.

  * How to get the filename of an accesslist.
  - WFS and WFSDEV
    The filename of accesslist x is kept in the file <x>.
             file: <x>
             +-------+-------+--------------------------+
             |  num  |  ref  | filename of accesslist x |
             +-------+-------+--------------------------+

*/

int WFSKrnAccessListGetAclFileInfo(const AreaInfo* area, const utf8* filename, u32 type, AclFileInfo* aclFile)
{
    // memset(&aclFile->di, 0, sizeof(DirItr)); // [check] workaround.
    aclFile->type = type;
#ifdef WFSDEV
    utf8* ptr;
    int len;
    len = GetAccessListPath(area, aclFile);
    if (len <= 0 ||
        WFSKRN_ACCESSLIST_MAX_PATH_LEN <= len)
    {
        return -1;
    }

    ptr = aclFile->path;
    ptr += len;

    switch (aclFile->type)
    {
      case ACL_TYPE_INDEXLIST:
      case ACL_TYPE_NAMEFILE:
        len = OS_SNPrintf(ptr, WFSKRN_ACCESSLIST_MAX_PATH_LEN - len, "\\"NAMEFILE_DIR"\\%s", filename);
        break;
      case ACL_TYPE_ACLFILE:
        len = OS_SNPrintf(ptr, WFSKRN_ACCESSLIST_MAX_PATH_LEN - len, "\\%s", filename);
        break;
      default:
        return -1;
    }

    if (len <= 0)
    {
        return -1;
    }
#else // WFSDEV
    int len;
    len = osSNPrintf(ACL_FILEINFO_NAME(aclFile), WFS_MAX_FILE_NAME_SIZE, "%s", filename);
    if (len <= 0)
    {
        return -1;
    }
#endif // WFSDEV

    return 0;
}

int WFSKrnAccessListOpenList(const AreaInfo* area, const utf8* filename, AclFileInfo* aclFile,
    WFSBool create, WFSAclFileHandle *pFh)
{
    WFSBool exist;

    if (WFSKrnAccessListGetAclFileInfo(area, filename, ACL_TYPE_ACLFILE, aclFile) < 0)
    {
        return -1;
    }

    exist = AclAccessListAlreadyExist(area, aclFile);
    if (!exist && !create)
    {
        return -1;
    }

    if (AclCreateAccessListFile(area, aclFile, create, pFh) != WFS_RESULT_OK)
    {
        return -1;
    }

    return exist ? 1 : 0;
}

static const char* UnescapeFilename(const char* src, char* dst, int count)
{
    while (count--)
    {
        if (*src == ESCAPE)
        {
            ++src;
            if (*src == ESCAPED_NULL)
            {
                *dst++ = 0x00;
            }
            else
            {
                *dst++ = ~*src;
            }
            ++src;
        }
        else
        {
            *dst++ = ~*src++;
        }
    }

    return src;
}

static char* EscapeAccessList(char* buf, const char* bufEnd, void* value, u32 size)
{
    char* ptr = (char*) value;
    const char* end = ptr + size;
    u8 inverted;

    if (bufEnd - buf < size)
    {
        return NULL;
    }


    while (ptr < end)
    {
        inverted = ~*ptr;
        if (inverted == '\0')
        {
            if (bufEnd < buf + 1)
            {
                return NULL;
            }
            *buf++ = ESCAPE;
            *buf++ = ESCAPED_NULL;
        }
        else if (ESCAPE == inverted || strchr("*/?", inverted)) // [check] need not escape '*', '/' and '?'?
        {
            if (bufEnd < buf + 1)
            {
                return NULL;
            }
            *buf++ = ESCAPE;
            *buf++ = inverted;
        }
        else
        {
            if (bufEnd < buf)
            {
                return NULL;
            }
            *buf++ = inverted;
        }
        ++ptr;
    }

    return buf;
}

//
// Index list
//
/*
 Index list is a file.
 It consists of an IndexListHeader and an array of IndexListEntry structures.

 An IndexListEntry keeps the location of an accesslist in a WFS area.

     IndexList
 +----------------+
 | IndexListHeader|
 +----------------+
 | IndexListEntry | the location of accesslist 0.
 +----------------+
 | IndexListEntry | the location of accesslist 1.
 +----------------+
 :                :
 :                :
 :                :
 +----------------+
 | IndexListEntry |
 +----------------+

 If the location field in an IndexListEntry{} is set to WFSKRN_ACCESSLIST_FREEINDEX,
 the accesslist is free.
*/

static int SetIndexList(const AreaInfo* area, WFSAclFileHandle indexListHandle, int page, IndexListHeader* header, IndexListEntry* indexList, u32 listSize)
{
    WFSFileSize offset;

    if (AclWriteFile(area, indexListHandle, header, sizeof(IndexListHeader), 0) != WFS_RESULT_OK)
    {
        return -1;
    }

    offset = WFSKRN_ACCESSLIST_INDEXES_IN_PAGE * page * sizeof(IndexListEntry) + sizeof(IndexListHeader);
    if (AclWriteFile(area, indexListHandle, indexList, listSize, offset) != WFS_RESULT_OK)
    {
        return -1;
    }

    return 0;
}

static int CreateIndexList(const AreaInfo* area, WFSAclFileHandle* indexListHandle, AclFileInfo* indexListInfo)
{
#ifndef WFSDEV
    *indexListHandle = indexListInfo;
#endif // WFSDEV

    if (WFSKrnAccessListGetAclFileInfo(area, INDEXLIST_FILENAME, ACL_TYPE_INDEXLIST, indexListInfo) < 0)
    {
        return -1;
    }
    if (AclCreateFile(area, indexListInfo, (WFSBool) TRUE,  indexListHandle) != WFS_RESULT_OK)
    {
        return -1;
    }
    return 0;
}

int OpenIndexList(const AreaInfo* area, WFSAclFileHandle* indexListHandle, AclFileInfo* indexListInfo)
{
#ifndef WFSDEV
    *indexListHandle = indexListInfo;
#endif // WFSDEV

    if (WFSKrnAccessListGetAclFileInfo(area, INDEXLIST_FILENAME, ACL_TYPE_INDEXLIST, indexListInfo) < 0)
    {
        return -1;
    }
    if (AclCreateFile(area, indexListInfo, (WFSBool) FALSE, indexListHandle) != WFS_RESULT_OK)
    {
        return -1;
    }
    return 0;
}

int CloseIndexList(const AreaInfo* area, WFSAclFileHandle indexListHandle)
{
    int result = -1;
    if (AclCloseFile(indexListHandle) == WFS_RESULT_OK)
    {
        result = 0;
    }
#ifdef _DEBUG
    else
    {
        WFSKRN_ACL_REPORT("CloseIndexList() WARNING: unable to close index list.\n");
    }
#endif // _DEBUG
    return result;
}

int WFSKrnAccessListIndexListInit(const AreaInfo* area)
{
    int ret = -1;
    WFSFileSize fileSize;
    IndexListHeader header;
    u32 index;
    IndexListEntry indexList[WFSKRN_ACCESSLIST_INDEXES_IN_PAGE];

    WFSAclFileHandle indexListHandle;
    AclFileInfo indexListInfo; // [check]

    if (CreateIndexList(area, &indexListHandle, &indexListInfo) < 0)
    {
        return -1;
    }

    if (AclGetFileSize(area, indexListHandle, &fileSize) == WFS_RESULT_OK)
    {
        if (fileSize == 0)
        {
            // create a new indexlist.
            for (index = 0; index < WFSKRN_ACCESSLIST_INDEXES_IN_PAGE; ++index)
            {
                indexList[index].location = WFSKRN_ACCESSLIST_FREEINDEX;
            }

            header.numEntry = WFSKRN_ACCESSLIST_INDEXES_IN_PAGE;
            ret = SetIndexList(area, indexListHandle, 0, &header, indexList, sizeof(indexList));
        }
        else
        {
            // [check] should verify the list?
            ret = 0; // already exists.
        }
    }

#ifdef WFSDEV
    CloseIndexList(area, indexListHandle);
#endif // WFSDEV
    return ret;
}

int WFSKrnAccessListIndexListExit(const AreaInfo* area)
{
    // index list is closed when each access is completed.
    return 0;
}

static int SetAccessListIndex(const AreaInfo* area, WFSAclFileHandle indexListHandle,
    WFSKrnAclHandle handle, const IndexListEntry* entry)
{
    WFSKRN_ACL_ASSERT(handle < WFSKRN_ACCESSLIST_INDEXES_IN_PAGE * WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE);
    WFSFileSize offset = sizeof(IndexListHeader) + handle * sizeof(IndexListEntry);

    if (AclWriteFile(area, indexListHandle, entry, sizeof(IndexListEntry), offset) == WFS_RESULT_OK)
    {
        return 0;
    }
    return -1;
}

int GetAccessListIndex(const AreaInfo* area, WFSAclFileHandle indexListHandle, WFSKrnAclHandle handle, IndexListEntry* entry)
{
    WFSKRN_ACL_ASSERT(handle < WFSKRN_ACCESSLIST_INDEXES_IN_PAGE * WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE);
    WFSFileSize offset = sizeof(IndexListHeader) + handle * sizeof(IndexListEntry);

    if (AclReadFile(area, indexListHandle, entry, sizeof(IndexListEntry), offset) == sizeof(IndexListEntry))
    {
        return 0;
    }
    return -1;
}

static int GetIndexList(const AreaInfo* area, WFSAclFileHandle indexListHandle,
    int page, IndexListHeader* header, IndexListEntry* indexList, u32 size)
{
    u32 offset;

    WFSKRN_ACL_ASSERT(page < WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE);
    if (header)
    {
        if (AclReadFile(area, indexListHandle, header, sizeof(IndexListHeader), 0)
            != sizeof(IndexListHeader) || header->numEntry == 0)
        {
            return -1;
        }
    }

    if (indexList)
    {
        offset = WFSKRN_ACCESSLIST_INDEXES_IN_PAGE * page * sizeof(IndexListEntry) + sizeof(IndexListHeader);
        if (AclReadFile(area, indexListHandle, indexList, size, offset) != (int) size)
        {
            return -1;
        }
    }
    return 0;
}

static int GetIndexListHeader(const AreaInfo* area, WFSAclFileHandle indexListHandle, IndexListHeader* header)
{
    return GetIndexList(area, indexListHandle, 0, header, NULL, 0);
}

static int GetIndexListPage(const AreaInfo* area, WFSAclFileHandle indexListHandle, int page, IndexListEntry* indexList, u32 size)
{
    WFSKRN_ACL_ASSERT(size == sizeof(IndexListEntry) * WFSKRN_ACCESSLIST_INDEXES_IN_PAGE);
    return GetIndexList(area, indexListHandle, page, NULL, indexList, size);
}

static int AllocAccessListIndex(const AreaInfo* area, WFSAclFileHandle indexListHandle, u32 aclType, WFSKrnAclHandle* newIndex, u32 num)
{
    IndexListHeader header;
    IndexListEntry* entry;
    IndexListEntry indexList[WFSKRN_ACCESSLIST_INDEXES_IN_PAGE];
    u32 numEntry;
    u32 page;
    int pageInList;
    u32 index;

    // read the header.
    if (GetIndexListHeader(area, indexListHandle, &header) < 0)
    {
        return -1;
    }

    if (num == 0)
    {
        // index 0 is reserved for default (empty) accesslist.
        IndexListEntry entry;
        entry.location = 0;
        if (SetAccessListIndex(area, indexListHandle, 0, &entry) < 0) // [check] should rename?
        {
            return -1;
        }
        *newIndex = aclType;
        return 0;
    }

    pageInList = header.numEntry / WFSKRN_ACCESSLIST_INDEXES_IN_PAGE;
    for (page = 0; page < pageInList; ++page)
    {
        // read a page.
        if (GetIndexListPage(area, indexListHandle, page, indexList, sizeof(indexList)) < 0)
        {
            return -1;
        }
        numEntry = WFSKRN_ACCESSLIST_INDEXES_IN_PAGE;
        entry = indexList;
        if (page == 0)
        {
            numEntry--;
            ++entry; // index 0 is reserved for default (empty) accesslist.
        }
        while (0 < numEntry--)
        {
            if (entry->location == WFSKRN_ACCESSLIST_FREEINDEX)
            {
                index = (entry - indexList) + page * WFSKRN_ACCESSLIST_INDEXES_IN_PAGE;
                WFSKRN_ACL_ASSERT((aclType & index) == 0 &&
                                  (aclType | index) != WFSKRN_ACCESSLIST_FREEINDEX);

                entry->location = (aclType | index); // save aclType in location field.

                if (SetAccessListIndex(area, indexListHandle, index, entry) < 0) // [check] should rename?
                {
                    return -1;
                }
                *newIndex = (aclType | index);
                return 0;
            }
            ++entry;
        }
    }

    if (WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE <= page)
    {
        WFSKRN_ACL_REPORT("AllocAccessListIndex() Error: no free index\n");
        return -1;
    }

    // add a new page to the list.
    index = header.numEntry;
    WFSKRN_ACL_ASSERT((aclType & index) == 0 &&
                      (aclType | index) != WFSKRN_ACCESSLIST_FREEINDEX);
    indexList[0].location = (aclType | index);
    for (numEntry = 1; numEntry < WFSKRN_ACCESSLIST_INDEXES_IN_PAGE; ++numEntry)
    {
        indexList[numEntry].location = WFSKRN_ACCESSLIST_FREEINDEX;
    }

    header.numEntry += WFSKRN_ACCESSLIST_INDEXES_IN_PAGE;
    if (SetIndexList(area, indexListHandle, page, &header, indexList, sizeof(indexList)) < 0)
    {
        return -1;
    }
    *newIndex = (aclType | index);

    return 0;
}

int WFSKrnAllocAccessListIndex(const AreaInfo* area, u32 aclType, WFSKrnAclHandle* newIndex, u32 num)
{
    int result = -1;
    int openResult;
    WFSAclFileHandle indexListHandle;
    AclFileInfo indexListInfo;

    openResult = OpenIndexList(area, &indexListHandle, &indexListInfo);
    if (openResult == 0)
    {
        result = AllocAccessListIndex(area, indexListHandle, aclType, newIndex, num);
    }

#ifdef WFSDEV
    if (openResult == 0)
    {
        CloseIndexList(area, indexListHandle);
    }
#else
    CloseIndexList(area, indexListHandle); // must call close() even when open() returns an error
                                           // to release dns stack memory.
#endif

    return result;
}

static int FreeAccessListIndex(const AreaInfo* area, WFSAclFileHandle indexListHandle, WFSKrnAclHandle handle)
{
    IndexListHeader header;
    IndexListEntry entry;

    handle &= ACL_TYPE_MASK; // discard flags.

    if (handle == WFSKRN_ACCESSLIST_FREEINDEX)
    {
        return 0;
    }

    if (GetIndexListHeader(area, indexListHandle, &header) < 0)
    {
        return -1;
    }

    if (header.numEntry <= handle)
    {
        return -1; // out of range.
    }

    if (GetAccessListIndex(area, indexListHandle, handle, &entry) < 0)
    {
        return -1;
    }

    if (entry.location == WFSKRN_ACCESSLIST_FREEINDEX)
    {
        return 0;
    }

    entry.location = WFSKRN_ACCESSLIST_FREEINDEX;
    if (SetAccessListIndex(area, indexListHandle, handle, &entry) < 0)
    {
        return -1;
    }
    return 0;
}

int WFSKrnFreeAccessListIndex(const AreaInfo* area, WFSKrnAclHandle handle)
{
    int result = -1;
    int openResult;
    WFSAclFileHandle indexListHandle;
    AclFileInfo indexListInfo;

    openResult = OpenIndexList(area, &indexListHandle, &indexListInfo);
    if (openResult == 0)
    {
        result = FreeAccessListIndex(area, indexListHandle, handle);
    }
#ifdef WFSDEV
    if (openResult == 0)
    {
        CloseIndexList(area, indexListHandle);
    }
#else
    CloseIndexList(area, indexListHandle); // must call close() even when open() returns an error
                                           // to release dns stack memory.
#endif

    return result;
}

// debug code.
static WFSBool VerifyList(const WFSKrnAccessList* list)
{
    u32 numEntry = WFSKrnAccessListGetNum(list);
    int i;
    int j;
    WFSKrnAccessListEntry* entry0 = (WFSKrnAccessListEntry*) (list+1);
    for (i = 0; i < numEntry; ++i)
    {
        WFSKrnAccessListEntry* entry1 = (WFSKrnAccessListEntry*) (list+1);
        for (j = 0; j < numEntry; ++j)
        {
            if (entry0 != entry1 && entry0->id == entry1->id && entry0->ecId == entry1->ecId)
            {
                // double registration.
                return (WFSBool) FALSE;
            }
            ++entry1;
        }
        ++entry0;
    }
    return (WFSBool) TRUE;
}

/*
   Create a filename of an accesslist by encoding
   the accesslist content.

   A filename only represents title IDs, the permissions and aclType (WFSKRN_ACCESSLIST_FLAG_*).
   The handle and reference count of an accesslist are written in a file.
*/
int WFSKrnConvertAccessListToFilename(const WFSKrnAccessList* list, utf8* filename, u32 size)
{
    utf8* filenameEnd = (utf8*) ((char*) filename + (size - 1)); // reserve a byte for '\0'.
    WFSKrnAccessListEntry* entry = (WFSKrnAccessListEntry*) (list+1);
    u32 numEntry;
    u32 aclType;
    u16 pm;

    WFSKRN_ACL_ASSERT(VerifyList(list));
    numEntry = WFSKrnAccessListGetNum(list);

    while (numEntry--)
    {
        // [check] endian.
        filename = EscapeAccessList(filename, filenameEnd, &entry->id, sizeof(u32));
        if (!filename)
        {
            return -1;
        }
        filename = EscapeAccessList(filename, filenameEnd, &entry->ecId, sizeof(u32));
        if (!filename)
        {
            return -1;
        }
        pm = (u16) (entry->pm & 0xffff);
        filename = EscapeAccessList(filename, filenameEnd, &pm, sizeof(u16));
        if (!filename)
        {
            return -1;
        }
        ++entry;
    }

    aclType = ACL_TYPE(list);
    filename = EscapeAccessList(filename, filenameEnd, &aclType, sizeof(u32));
    if (!filename)
    {
        return -1;
    }
    *filename++ = 0;

    return 0;
}

int WFSKrnConvertFilenameToAccessList(const utf8* filename, u32 filenameLen, WFSKrnAccessList* list)
{
    const char* src = filename;
    u32 num;
    u32 aclType;
    WFSKrnAccessListEntry* entry;
    u16 pm;

    entry = (WFSKrnAccessListEntry*) (list + 1);
    num = WFSKrnAccessListGetNum(list);
    while (num--)
    {
        src = UnescapeFilename(src, (char*) &entry->id, sizeof(u32));
        src = UnescapeFilename(src, (char*) &entry->ecId, sizeof(u32));
        src = UnescapeFilename(src, (char*) &pm, sizeof(u16));
        entry->pm = pm;
        ++entry;
    }
    UnescapeFilename(src, (char*) &aclType, sizeof(u32));
    WFSKRN_ACL_ASSERT(!(aclType & ACL_TYPE_MASK));

    WFSKRN_ACL_ASSERT(VerifyList(list));
    return 0;
}

static int AclGetIndex(const AreaInfo* area, WFSAclFileHandle hFile, WFSKrnAclHandle* index)
{
    if (AclReadFile(area, hFile, index, sizeof(u32), 0) != sizeof(u32))
    {
        return -1;
    }
    return 0;
}

static int AclSetIndex(const AreaInfo* area, WFSAclFileHandle hFile, WFSKrnAclHandle handle)
{
    // save the index.
    if (AclWriteFile(area, hFile, &handle, sizeof(WFSKrnAclHandle), 0) != WFS_RESULT_OK)
    {
        return -1;
    }
    return 0;
}

static int AclIncrementFileRef(const AreaInfo* area, WFSKrnAclHandle index) // [check] must be global?
{
    u32 ref;
    if ((index & ACL_TYPE_MASK) == 0)
    {
        // index 0 is reserved for the empty accesslist.
        return 0;
    }

    if (ReadNamefile(area, index, &ref, sizeof(u32), sizeof(u32)) < 0)
    {
        return -1;
    }

    ++ref;
    if (WriteNamefile(area, index, &ref, sizeof(u32), sizeof(u32)) < 0)
    {
        return -1;
    }
    return 0;
}

// increment reference count of the specified accesslist if the file already exists.
// otherwise, create an accesslist file and write information in it.
int WFSKrnAccessListCreateFile(const AreaInfo* area, const utf8* filename, WFSKrnAccessList** list)
{
    int resultOpen;
    int result = -1;
    WFSAclFileHandle hFile;
    AclNameFileHeader header;
    WFSKrnAclHandle index;
    AclFileInfo aclFile;

    resultOpen = WFSKrnAccessListOpenList(area, filename, &aclFile, (WFSBool) TRUE, &hFile);
    if (resultOpen < 0)
    {
        goto WFSCREATEACCESSFILEINIT_RESULT;
    }

    if (resultOpen == 1) // already exist.
    {
        if (AclGetIndex(area, hFile, &index) < 0 ||
            AclIncrementFileRef(area, index) < 0)
        {

            // [check] should return fatal error?
            goto WFSCREATEACCESSFILEINIT_RESULT;
        }

        WFSKrnAccessList* cache = WFSKrnAccessListLookupByIndex(area, index);
        if (cache)
        {
            WFSKRN_ACL_ASSERT(ACL_TYPE(*list) == ACL_TYPE(cache));
            // exchange.
            WFSKrnMemoryBlockFree(area, (WFSKrnAccessListBlock*) (*list), WFSKRN_BLOCKS_IN_LIST(*list));
            if (cache->ref == 0)
            {
                // remove from the FREE cachlist.
                WFSKrnAccessListCacheFreeListRemove(area, cache);
            }
            (*list) = cache;
            ++(*list)->ref;
        }
        else
        {
            WFSKRN_ACL_ASSERT(ACL_TYPE(*list) == AclIndexType(index)); // [check]
            WFSKRN_ACL_ASSERT((*list)->ref == 1);
            // change list index, if no cache.
            (*list)->handle = index;
            WFSKrnAccessListCacheAdd(area, *list);
        }
        result = 1; // list was modified.
    }
    else
    {
    	int ret;
        // create a new file.
        // (*list)->handle only contains flags at this moment.
        if (WFSKrnAllocAccessListIndex(area, (*list)->handle,
            &(*list)->handle, WFSKrnAccessListGetNum(*list)) < 0)
        {
            goto WFSCREATEACCESSFILEINIT_RESULT;
        }

        if (AclSetIndex(area, hFile, (*list)->handle) < 0)
        {
            WFSKrnFreeAccessListIndex(area, (*list)->handle);
            AclDeleteAccessListFile(area, &aclFile); // [check] must test this code.
            goto WFSCREATEACCESSFILEINIT_RESULT;
        }

        header.ref = 1; // note header.ref is not equal to (*list)->ref.
        header.num = WFSKrnAccessListGetNum(*list);
        ret = CreateNamefile(area, (*list)->handle, &header, filename);
        if (ret < 0)
        {
            WFSKrnFreeAccessListIndex(area, (*list)->handle);
            AclDeleteAccessListFile(area, &aclFile); // [check] must test this code.
            goto WFSCREATEACCESSFILEINIT_RESULT;
        }
        WFSKrnAccessListCacheAdd(area, *list);
        result = 0;
    }

WFSCREATEACCESSFILEINIT_RESULT:
#ifdef WFSDEV
    if (result < 0)
    {
        WFSKRN_ACL_REPORT("WFSKrnAccessListCreateFile() Error %ld\n", GetLastError());
    }
    if (0 <= resultOpen)
    {
        AclCloseFile(hFile);
    }
#endif // WFSDEV
    return result;
}

int WFSKrnGetFilename(const AreaInfo* area, WFSKrnAclHandle handle, utf8* filename, u32 size)
{
    return ReadNamefile(area, handle, filename, size, sizeof(AclNameFileHeader));
}

WFSKrnAccessList* WFSKrnAccessListLookupFile(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WFSKrnResult result;
    WFSKrnAccessList* list = NULL;
    AclNameFileHeader header;
    utf8 filename[WFS_MAX_FILE_NAME_SIZE];

    if (ReadNamefile(area, handle, &header, sizeof(header), 0) < 0)
    {
        return NULL;
    }

    result = WFSKrnAccessListAlloc(area, header.num, &list);
    if (result != WFSKRN_RESULT_OK)
    {
        return NULL;
    }
    list->handle = handle; // [check] set flag?

    if (WFSKrnGetFilename(area, list->handle, filename, sizeof(filename)) < 0 ||
        WFSKrnConvertFilenameToAccessList(filename, STRNLEN(filename, WFS_MAX_FILE_NAME_SIZE), list) < 0)
    {
        WFSKrnAccessListFree(area, list);
        return NULL;
    }

    WFSKrnAccessListCacheAdd(area, list);
    return list;
}

WFSKrnResult WFSKrnAccessListGetFileReferenceCount(const AreaInfo* area, WFSKrnAclHandle handle, u32* ref)
{
    if ((handle & ACL_TYPE_MASK) == 0)
    {
        // index 0 is reserved for the empty accesslist.
        *ref = 1;
        return WFSKRN_RESULT_OK;
    }

    if (ReadNamefile(area, handle, ref, sizeof(u32), sizeof(u32)))
    {
        return WFSKRN_RESULT_ACL_FILE;
    }
    return WFSKRN_RESULT_OK;
}

WFSKrnResult WFSKrnAccessListSetFileReferenceCount(const AreaInfo* area, WFSKrnAclHandle handle, u32 ref)
{
    if ((handle & ACL_TYPE_MASK) == 0)
    {
        // index 0 is reserved for the empty accesslist.
        return WFSKRN_RESULT_OK; // just return.
    }

    if (WriteNamefile(area, handle, &ref, sizeof(u32), sizeof(u32)) < 0)
    {
        return WFSKRN_RESULT_ACL_FILE;
    }
    return WFSKRN_RESULT_OK;
}

WFSKrnResult WFSKrnAccessListDeleteFile(const AreaInfo* area, WFSKrnAclHandle handle)
{
    AclFileInfo aclFile;
    utf8 filename[WFS_MAX_FILE_NAME_SIZE];

    if (WFSKrnGetFilename(area, handle, filename, sizeof(filename)) < 0 ||
        WFSKrnAccessListGetAclFileInfo(area, filename, ACL_TYPE_ACLFILE, &aclFile) < 0)
    {
        return WFSKRN_RESULT_ACL_FILE;
    }

    if (AclDeleteAccessListFile(area, &aclFile) != WFS_RESULT_OK ||
        DeleteNamefile(area, handle) < 0)
    {
        WFSKRN_ACL_REPORT("WFSKrnAccessListDeleteFile Error.\n");
        return WFSKRN_RESULT_ACL_FILE;
    }
    return WFSKRN_RESULT_OK;
}
