/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Utils.cpp

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Utils.cpp,v $
  Revision 1.2  2008/12/19 05:45:08  saito_tomoya
  Modified _qsort() to sort correctly.

  Revision 1.1  2008/12/03 00:10:21  kondo_masahiro
  Initial check-in.


*---------------------------------------------------------------------------*/

#include "wfskrn_Utils.h"

#define SWAP(type,a,b) { type temp = a; a = b; b = temp; }


static
void exchangeValues(void *x , void *y, size_t size) {
    while(size >= 4){
        SWAP(u32, *(u32*)x, *(u32*)y)
        x = (void*)((size_t)x+4);
        y = (void*)((size_t)y+4);
        size -= 4;
    }
    while(size){
        SWAP(u8, *(u8*)x, *(u8*)y)
        x = (void*)((size_t)x+1);
        y = (void*)((size_t)y+1);
        --size;
    }
}


void _qsort(void *base, size_t num, size_t size, int (*compare)(const void*, const void*) ){
    int i = 0, j = (num-1), key = ((num-1)/2);
    while(1) {
        while( compare( (const void*)((size_t)base+i*size), (const void*)((size_t)base+key*size)) < 0 ){
            ++i;
        }
        while( compare( (const void*)((size_t)base+key*size), (const void*)((size_t)base+j*size)) < 0 ){
            --j;
        }
        if (i >= j) break;
        exchangeValues( (void*)((size_t)base+i*size), (void*)((size_t)base+j*size), size );
        if(i==key) key = j;
        else if(j==key) key = i;
        i++; j--;
    }
    if (0       < i-1) _qsort( base,                            i,     size, compare);
    if ((num-1) > j+1) _qsort((void*)((size_t)base+(j+1)*size), num-j-1, size, compare);
}


int CompareBlkAdrForward(const void *pA, const void *pB) {
    WFSBlkAdr *pBlkAdrA = (WFSBlkAdr *)pA;
    WFSBlkAdr *pBlkAdrB = (WFSBlkAdr *)pB;
    return (int)(*pBlkAdrA - *pBlkAdrB);
}


int CompareBlkAdrReverse(const void *pA, const void *pB) {
    WFSBlkAdr *pBlkAdrA = (WFSBlkAdr *)pA;
    WFSBlkAdr *pBlkAdrB = (WFSBlkAdr *)pB;
    return (int)(*pBlkAdrB - *pBlkAdrA);
}

