/*---------------------------------------------------------------------------*
  Project:  TwlSDK - STD - include
  File:     string.h

  Copyright 2005-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date::            $
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/

#ifndef FIRM_STD_STRING_H_
#define FIRM_STD_STRING_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include <stdarg.h>

#include <brom/misc.h>
#include <brom/types.h>
#include <brom/mi/memory.h>

//---- aliases
#define stdStrCpy          stdCopyString
#define stdStrLCpy         stdCopyLString
#define stdStrChr          stdSearchChar
#define stdStrRChr         stdSearchCharReverse
#define stdStrStr          stdSearchString
#define stdStrLen          stdGetStringLength
#define stdStrNLen         stdGetStringNLength
#define stdStrCat          stdConcatenateString
#define stdStrLCat         stdConcatenateLString
#define stdStrCmp          stdCompareString
#define stdStrNCmp         stdCompareNString
#define stdStrLCmp         stdCompareLString

#define stdMemCpy          stdCopyMemory
#define stdMemMove         stdMoveMemory
#define stdMemSet          stdFillMemory

/*---------------------------------------------------------------------------*
  Name:         stdCopyString

  Description:  same to strcpy

  Arguments:    destp : destination pointer
                srcp  : src pointer

  Returns:      pointer to destination
 *---------------------------------------------------------------------------*/
extern char *stdCopyString(char *destp, const char *srcp);

/*---------------------------------------------------------------------------*
  Name:         stdCopyLStringZeroFill

  Description:  do not correspond with strlcpy

  Arguments:    destp : destination pointer
                srcp  : src pointer
                n     : copy size + 1

  Returns:      size of src
 *---------------------------------------------------------------------------*/
extern int stdCopyLStringZeroFill(char *destp, const char *srcp, int n);

/*---------------------------------------------------------------------------*
  Name:         stdCopyLString

  Description:  same to strlcpy

  Arguments:    destp : destination pointer
                srcp  : src pointer
                siz   : copy size + 1

  Returns:      size of src
 *---------------------------------------------------------------------------*/
extern int stdCopyLString(char *destp, const char *srcp, int siz);

/*---------------------------------------------------------------------------*
  Name:         stdSearchChar

  Description:  same to strchr

  Arguments:    srcp : src string
                c    : character to search from src pointer

  Returns:      pointer to destination
 *---------------------------------------------------------------------------*/
extern char *stdSearchChar(const char *srcp, int c);

/*---------------------------------------------------------------------------*
  Name:         stdSearchCharReverse

  Description:  same to strrchr

  Arguments:    srcp : src string
                c    : character to search from src pointer

  Returns:      pointer to destination
 *---------------------------------------------------------------------------*/
extern char *stdSearchCharReverse(const char *srcp, int c);

/*---------------------------------------------------------------------------*
  Name:         stdSearchString

  Description:  same to strstr

  Arguments:    srcp : src string
                str  : string to search from src pointer

  Returns:      pointer to destination
 *---------------------------------------------------------------------------*/
extern char *stdSearchString(const char *srcp, const char *str);

/*---------------------------------------------------------------------------*
  Name:         stdGetStringLength

  Description:  get string length. same to strlen

  Arguments:    str : string

  Returns:      string length
 *---------------------------------------------------------------------------*/
extern int stdGetStringLength(const char *str);

/*---------------------------------------------------------------------------*
  Name:         stdGetStringNLength

  Description:  get string length below len. same to strnlen

  Arguments:    str : string
                len : max length

  Returns:      string length
 *---------------------------------------------------------------------------*/
extern int stdGetStringNLength(const char *str, int len);

/*---------------------------------------------------------------------------*
  Name:         stdConcatenateString

  Description:  concatenate strings. same to strcat

  Arguments:    str1 : original string
                str2 : string to concatenate

  Returns:      concatenated string
 *---------------------------------------------------------------------------*/
extern char *stdConcatenateString(char *str1, const char *str2);

/*---------------------------------------------------------------------------*
  Name:         stdConcatenateLString

  Description:  concatenate strings. same to strlcat

  Arguments:    str1 : original string
                str2 : string to concatenate
                size : buffer size of str1

  Returns:      length of str1 + length of str2
 *---------------------------------------------------------------------------*/
extern int stdConcatenateLString(char *str1, const char *str2, int size);

/*---------------------------------------------------------------------------*
  Name:         stdCompareString

  Description:  compare strings. same to strcmp

  Arguments:    str1, str2 : strings

  Returns:      0 if same
 *---------------------------------------------------------------------------*/
extern int stdCompareString(const char *str1, const char *str2);

/*---------------------------------------------------------------------------*
  Name:         stdCompareNString

  Description:  same to strncmp

  Arguments:    str1, str2 : strings
                len        : max length

  Returns:      0 if same
 *---------------------------------------------------------------------------*/
extern int stdCompareNString(const char *str1, const char *str2, int len);

/*---------------------------------------------------------------------------*
  Name:         stdCompareLString

  Description:  same to strlcmp

  Arguments:    str1, str2 : strings
                len        : max length

  Returns:      0 if same
 *---------------------------------------------------------------------------*/
extern int stdCompareLString(const char *str1, const char *str2, int len);

/*---------------------------------------------------------------------------*
  Name:         stdCompareNString

  Description:  same to strncasecmp

  Arguments:    str1, str2 : strings
                len        : max length

  Returns:      0 if same
 *---------------------------------------------------------------------------*/
extern int stdCompareNIString(const char *str1, const char *str2, int len);

/*---------------------------------------------------------------------------*
  Name:         stdTSScanf

  Description:  ȃTCYړIƂ sscanf.
                {Iȏw "%(*?)([lh]{,2})([diouxXpn])" ɑΉ.

  Arguments:    src           ͕
                fmt           䕶

  Returns:      ꂽl̑.
                ł̏I[邢͕so -1.
 *---------------------------------------------------------------------------*/
extern int stdTSScanf(const char *src, const char *fmt, ...);

/*---------------------------------------------------------------------------*
  Name:         stdTVSScanf

  Description:  stdTSScanf  va_list Ή.
                {Iȏw "%(*?)([lh]{,2})[diouxX]" ɑΉ.

  Arguments:    src           ͕
                fmt           䕶
                vlist         p[^

  Returns:      ꂽl̑.
                ł̏I[邢͕so -1.
 *---------------------------------------------------------------------------*/
extern int stdTVSScanf(const char *src, const char *fmt, va_list vlist);

/*---------------------------------------------------------------------------*
  Name:         stdTSPrintf

  Description:  ̌`ȊO stdTVSNPrintf Ɠ.

  Arguments:    dst           ʂi[obt@
                fmt           䕶

  Returns:      stdVSNPrintf Ɠ.
 *---------------------------------------------------------------------------*/
extern int stdTSPrintf(char *dst, const char *fmt, ...);

/*---------------------------------------------------------------------------*
  Name:         stdTVSPrintf

  Description:  ̌`ȊO stdTVSNPrintf Ɠ.

  Arguments:    dst           ʂi[obt@
                fmt           䕶
                vlist         p[^

  Returns:      stdVSNPrintf Ɠ.
 *---------------------------------------------------------------------------*/
extern int stdTVSPrintf(char *dst, const char *fmt, va_list vlist);

/*---------------------------------------------------------------------------*
  Name:         stdTSNPrintf

  Description:  ̌`ȊO stdTVSNPrintf Ɠ.

  Arguments:    dst           ʂi[obt@
                len           obt@
                fmt           䕶
 
  Returns:      stdVSNPrintf Ɠ.
 *---------------------------------------------------------------------------*/
extern int stdTSNPrintf(char *dst, size_t len, const char *fmt, ...);

/*---------------------------------------------------------------------------*
  Name:         stdTVSNPrintf

  Description:  ȃTCYړIƂ sprintf.
                {IȏwɑΉ.
                %([-+# ]?)([0-9]*)(\.?)([0-9]*)([l|ll|h||hh]?)([diouxXpncs%])

  Note:         CodeWarrior  MSL sprintf() ̋ɂ킹
                '+'  '#' ͖ɂĂ.
                { // exsample
                  char buf[5];
                  sprintf(buf, "%-i\n", 45);  // "45"  (OK)
                  sprintf(buf, "%0i\n", 45);  // "45"  (OK)
                  sprintf(buf, "% i\n", 45);  // " 45" (OK)
                  sprintf(buf, "%+i\n", 45);  // "%+i" ("+45" expected)
                  sprintf(buf, "%#x\n", 45);  // "%#x" ("0x2d" expected)
                  // but, this works correctly!
                  sprintf(buf, "% +i\n", 45); // "+45" (OK)
                }

  Arguments:    dst           ʂi[obt@
                len           obt@
                fmt           䕶
                vlist         p[^

  Returns:      𐳂o͂ꍇ̕Ԃ. ('\0'܂܂Ȃ)
                obt@TCY[ȂSĂ̕o͂ďI[t^.
                obt@TCYsȂ؂l߂ dst[len-1] I[Ƃ.
                len  0 ̏ꍇ͉Ȃ.

 *---------------------------------------------------------------------------*/
extern int stdTVSNPrintf(char *dst, size_t len, const char *fmt, va_list vlist);


static inline void* stdCopyMemory(void *destp, const void *srcp, u32 size)
{
    miCpuCopy(srcp, destp, size);
    return destp;
}

static inline void* stdMoveMemory(void *destp, const void *srcp, u32 size)
{
    miCpuMove(srcp, destp, size);
    return destp;
}

static inline void* stdFillMemory(void *destp, u8 data, u32 size)
{
	miCpuFill(destp, data, size);
	return destp;
}

#ifdef __cplusplus
} /* extern "C" */
#endif

/* FIRM_STD_STRING_H_ */
#endif
