/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Dir.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Dir.cpp,v $
  Revision 1.43  2008/12/03 00:07:34  kondo_masahiro
  Removed a function _qsort().

  Revision 1.42  2008/11/26 01:41:34  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.41  2008/11/10 02:18:37  kondo_masahiro
  Fixed DirResize()

  Revision 1.40  2008/11/05 04:45:32  kondo_masahiro
  Fixed the end of line made to newline.

  Revision 1.39  2008/11/04 00:21:44  kondo_masahiro
  Added DirSbaAllocAndRegistSubBlk()

  Revision 1.38  2008/10/30 05:01:36  kondo_masahiro
  Fixed DirResize for decrease size

  Revision 1.37  2008/10/30 04:32:42  ueno
  Modified for release build.

  Revision 1.36  2008/10/24 04:54:32  kondo_masahiro
  Added decrease block process of DirResize.

  Revision 1.35  2008/10/21 09:49:36  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.34  2008/10/17 08:53:52  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB) and fixed a bug about accessing 4GB-2.

  Revision 1.33  2008/10/16 09:30:39  ooizumi
  Fixed definitions for debug build.

  Revision 1.32  2008/10/15 00:14:51  kondo_masahiro
  Fixed a bug about 4GB-1

  Revision 1.31  2008/10/14 12:10:59  kondo_masahiro
  Fixed a bug in DirResize

  Revision 1.30  2008/10/14 10:26:24  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.29  2008/10/10 04:07:01  kondo_masahiro
  Fixed a bug in DirResize

  Revision 1.28  2008/10/10 02:11:19  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.27  2008/10/10 01:20:37  kondo_masahiro
  Fixed a bug in DirResize

  Revision 1.26  2008/10/09 10:58:53  ueno
  Commented out debug code.

  Revision 1.25  2008/10/09 04:16:53  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.24  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.23  2008/10/08 08:33:44  ueno
  Moved debug code to be able to test access list directory.

  Revision 1.22  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.21  2008/09/28 23:31:51  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.20  2008/09/02 07:04:45  kondo_masahiro
  added qsort for IOP compiler.

  Revision 1.19  2008/08/27 23:23:28  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.18  2008/08/27 09:37:02  paul
  Started coding for different size categories

  Revision 1.17  2008/08/26 08:04:01  ooizumi
  Fixed a bug in DirReinitBlk. The update counter value must be kept even after re-initialization.

  Revision 1.16  2008/08/05 04:07:27  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.15  2008/07/29 19:00:08  paul
  Fixed a bug in DirNodeBlkSplit()

  Revision 1.14  2008/07/28 21:52:49  paul
  Improved updateMap processing. Added DirGetBlk() and DirUnpinAndFreeBlk().

  Revision 1.13  2008/07/25 02:51:28  paul
  Changed from nTransIdx to *pTransInfo

  Revision 1.12  2008/07/17 22:46:40  paul
  Fixed DirResizeFile() to use block cache.

  Revision 1.11  2008/07/17 05:32:08  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.10  2008/07/09 00:29:35  paul
  Changes to make directory code work with block cache.

  Revision 1.9  2008/06/09 17:29:06  paul
  Minor changes

  Revision 1.8  2008/05/27 18:43:55  paul
  Made utf8 variable definitions more consistent, and use (unsigned) for comparisons

  Revision 1.7  2008/05/20 00:06:11  paul
  Fixed a bug in DirResizeFile

  Revision 1.6  2008/05/19 17:07:26  paul
  Fixed bugs in DirLeafSplit, DirResizeFile

  Revision 1.5  2008/05/17 04:14:07  paul
  Fixed a bug in DirLeafCompact / DirNodeCompact

  Revision 1.4  2008/05/12 19:08:51  paul
  Replaced some calls to DirNodeStackFree with calls to DirItrClose

  Revision 1.3  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:29:42  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.21  2008/04/22 21:06:23  paul
  Changed API to some functions a little, and moved DirFind* to a separate file

  Revision 1.20  2008/04/19 05:51:03  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.19  2008/04/18 02:10:59  paul
  Fixed DirNext to update pName->nLen

  Revision 1.18  2008/04/18 01:31:27  paul
  Removed wfskrn_Permission.h

  Revision 1.17  2008/04/18 00:08:22  paul
  Changed interface to Dir functions

  Revision 1.16  2008/04/15 22:19:55  paul
  Implemented the resize-in-place case for sub nodes of Leaf blocks in DirDelete. Fixed a bug in DirLeafCompact.

  Revision 1.15  2008/04/15 18:33:53  paul
  Changed the method for finding an internal node in DirFind: Insead of using splitter name search, use normal name search, but stop as soon as the requested depth is reached. Changed DirDelete to use this method and fixed a bug in DirDelete. Changed DirLeafCompact to match DirNodeCompact and provide the functionality to resize one node and delete another node while compacting - this is needed to guarantee DirLeafMergeSubBlkWithItsSingleChild always works provided the result can fit within a block.

  Revision 1.14  2008/04/14 17:07:40  paul
  Added block splitting to DirDelete when necessary. Cleaned up DirFind, and ensured it calculates correct nStrIdx during node traversal so that DirConstructEntryNameFromNodeStack can be used on splitters.

  Revision 1.13  2008/04/09 18:20:20  paul
  Added DirNodeCompact and cleaned up DirLeafCompact

  Revision 1.12  2008/04/08 00:22:33  paul
  Implemented some unfinished cases for Delete

  Revision 1.11  2008/04/04 23:46:18  paul
  Implemented Delete. There are some unfinished cases, but it is mostly complete.

  Revision 1.10  2008/03/07 01:36:52  paul
  Changed DirFind to have a special mode for prefix searches (used to implement wildcard pattern search) DIR_PREFIX_SEARCH.

  Revision 1.9  2008/02/28 06:31:56  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.8  2008/02/27 09:39:34  paul
  Removed unnecessary params from DirFindNext. Added a stub for DirDelete

  Revision 1.7  2008/02/27 07:25:15  paul
  Added code to check intra block offsets, fixed a bug in DirFind

  Revision 1.6  2008/02/27 00:36:22  paul
  Many bug-fixes

  Revision 1.5  2008/02/21 04:31:58  paul
  Moved DirNodeStack functions to wfskr_DirNodeStack.cpp

  Revision 1.4  2008/02/21 00:48:26  paul
  Moved test code to wfskrn_DirTest.cpp

  Revision 1.3  2008/02/21 00:14:38  paul
  Moved SubBlk allocation code from this file to wfskrn_SubBlkAlloc.cpp, and made it more generic so it could be used for multiple metadata block types

  Revision 1.2  2008/02/20 08:04:38  paul
  Many changes. Currently debugging multi-block insert code

  Revision 1.1  2007/12/27 11:20:03  paul
  still at an early stage of development. Allows filenames to be added to a single leaf block. Now working on DirLeafBlkSplit() function.

 *---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "Dir"

#include "wfskrn_Dir.h"
#include "wfskrn_DirRxTree.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_Area.h"
#include "wfskrn_BitField.h"
#include "wfskrn_Heap.h"
#include "wfskrn_Api.h"
#include "wfskrn_Permission_AccessList.h"
#include "wfskrn_Mutex.h"
#include "wfskrn_Utils.h"

#undef dbg
#if _DEBUG_DIR
    #define dbg(s) s
#else
    #define dbg(s)
#endif

#if _DEBUG_DIR
    #if !_WIN32
        u32 nDirTest = 0;
        u32 nDirTestBreakPoint = 0;
    #endif
#endif


#define DIR_MAX_SPLIT_BLK_ALLOC ((DIR_MAX_BLK_DEPTH-1)*2)

// Algorithm: Macro-structure is a B-Tree, with a node per disk-block. Uses a Radix Tree (Trie) within each block.
// * Leaf blocks contain a lexicographically sorted list of string key fields, and associated file records.
//   Only leaf blocks can contain file records.
// * Internal node blocks contain a lexicographically sorted list of string splitter fields, and point to leaf blocks.
// In both leaf nodes and internal nodes, the sorted lists of strings are represented as radix trees.
//
// Algorithm for inserting names into the directory:
//    * Traverse 0 or more internal nodes comparing against splitters to determine the correct leaf node for insertion.
//    * Traverse the leaf node to determine the correct point of insertion.
//    * Perform the insertion to the leaf node, splitting it first if neccessary.
//      If a split is required:
//      * copy the first half and second half of the leaf respectively to two new nodes, then delete the original leaf.
//      * Add a splitter to the parent node, or create a new parent node if none exists.
//      * If adding the splitter causes the parent node to become full, then split it first and repeat recursively.
//    * Internal nodes are not split on the way down like some B-Tree algorithms.
//      since it is not easy to determine when a split would be required due to variable sized nodes.
// 2. Deleting names from a directory:
//    *
//
// In order to be able to insert an internal node, we need be able to get to the node.

// ToDo: We need a fast way to determine  the longest filename, and the longest sub-path of any of directory.
//       This should be stored in each directory and updated as needed.
//       If a filename is shortened, or a file is deleted, we need a way to update its parent directory and
//       all ancestor directories if necessary - we'd like to find the next longest sub-path quickly.
//       - is it necessary to update all parent directories?
//       Needs more thought..

u32 DirLeafGetSubBlkLog2Size_2(u32 nStrLen, u32 nNumEntries) {
    u32 nNumBytesRequired = sizeof(DirSubBlkHdr)-1 + nStrLen + (nNumEntries*3);
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nNumBytesRequired-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


u32 DirNodeGetSubBlkLog2Size_3(u32 nStrLen, u32 nNumEntries, utf8 cFirstChoice) {
    u32 nNumBytesRequired = sizeof(DirSubBlkHdr)-1 + nStrLen + (nNumEntries*3);
    if (cFirstChoice == 0) {
        // Nodes require 2 extra bytes for nodes containing a 0 (string terminator) choice.
        // This is because the pointer for the 0 case is a 4 byte Block pointer instead of a 2-byte intra-block pointer.
        nNumBytesRequired += sizeof(WFSBlkAdr) - sizeof(WFSIntraBlkOfs);
    }
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nNumBytesRequired-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


// Indirect blocks contain pointers to large (512KB) data blocks, plus 8 sets of hash codes per 512KB block.
// So the amount of data per pointer to 512KB block is: 4 byte Pointer + (8 * 20 byte hash code) == 164 bytes.
// Each 8KB Block can contain: (49 * 164) + 156 == 8192 bytes. (The header must be at least 24 Bytes, which is true).
// A max size 4GB file == 2048 * 512KB.
// So it takes 8192/49 == 168 indirect blocks of 8KB each to cover 4GB.

u32 DirCalculateAttrSubBlkLog2Size(AreaInfo *pAreaInfo, u32 nAllocSize, u32 nNameLen, u8 *pSizeCategory) {
    // This function calculates the size of the attribute sub-block necessary to store a file of the given size
    // There are 4 size categories: 0B <= Small <= 964B < Medium <= 320KB < Large <= 2560KB < Very Large < 4GB

    // Arguments:
    //   pAreaInfo      [In] .. A pointer to the AreaInfo struct for the area which contains/will contain the file
    //   nAllocSize     [In] .. The size of the file upon which to base the calculations
    //   nNameLen       [In] .. The Length of the file name in bytes
    //   pSizeCategory [Out] .. A pointer to a u32 variable which will be overwritten with flags indicating the block size & indirection level of the file
    // Returns the log base 2 of the size of the sub-block needed to store the file attributes.

    //u32 nTotalSize = sizeof(DirEntryAttrHdr)-1 + (((nNameLen+31)>>5)<<2);   // A bit per character of the name to represent upper/lower case
    u32 nTotalSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((nNameLen+7)>>3);   // A bit per character of the name to represent upper/lower case
    if (nAllocSize <= (1<<SBA_MAX_LOG2_SUB_BLK_SIZE) - nTotalSize) {
        *pSizeCategory = FILE_SIZE_CATEGORY_VERY_SMALL;
        nTotalSize += nAllocSize;
    } else {
        //u32 nNumSmallBlks = (nAllocSize + pAreaInfo->nBlkSize - 1)>>pAreaInfo->ah.nLog2BlkSize;
        // Note: This calculation needs to work with maximum file size = 4294967295
        u32 nNumSmallBlks = ((nAllocSize-1)>>pAreaInfo->ah.nLog2BlkSize)+1;
        if (nNumSmallBlks >= (1 << (WFS_LOG2_MAX_FILE_SIZE-pAreaInfo->ah.nLog2BlkSize))){
            nNumSmallBlks = 1 << (WFS_LOG2_MAX_FILE_SIZE-pAreaInfo->ah.nLog2BlkSize);
        }
        if (nNumSmallBlks <= DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS) {
            *pSizeCategory = FILE_SIZE_CATEGORY_SMALL;
            nTotalSize += nNumSmallBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
        } else {
            u32 nNumMediumBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerMediumBlk) + 1;
            if(nNumMediumBlks <= DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS){
                *pSizeCategory = FILE_SIZE_CATEGORY_MEDIUM;
                nTotalSize += nNumMediumBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
            } else {
                u32 nNumLargeBlks = ((nNumMediumBlks - 1) >> pAreaInfo->nLog2MediumBlkPerLargeBlk) + 1;
                //u32 nNumLargeBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerLargeBlk) + 1;
                if (nNumLargeBlks <= DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS) {
                    *pSizeCategory = FILE_SIZE_CATEGORY_LARGE;
                    nTotalSize += nNumLargeBlks * pAreaInfo->nLargeBlkPtrSize;  // (sizeof(WFSBlkAdr) + (sizeof(WFSHashCode) * pAreaInfo->nNumHashCodesPerLargeBlk));
                } else {
                    *pSizeCategory = FILE_SIZE_CATEGORY_VERY_LARGE;
                    u32 nNumIndirectBlks = (nNumLargeBlks - 1) / pAreaInfo->nNumLargeBlkPtrsPerBlk + 1;
                    nTotalSize += nNumIndirectBlks * sizeof(WFSBlkAdr);
                }
            }
        }
    }
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nTotalSize-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


u32 DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(AreaInfo *pAreaInfo, u32 *pAllocSize, u32 nNameLen, u8 *pSizeCategory) {
    // This version of the function updates *pAllocSize.
    // On entry, *pAllocSize is the minimum number of bytes that must be allocated.
    // On exit, *pAllocSize is the maximum number of bytes that can be used without re-allocation.
    // (This can be larger than the requested amount because we only allocate power-of-2 chunks within the block,
    //  or whole numbers of blocks outside the block).
    u32 nAttrHdrSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((nNameLen+7)>>3);   // A bit per character of the name to represent upper/lower case
    if (*pAllocSize <= (1<<SBA_MAX_LOG2_SUB_BLK_SIZE) - nAttrHdrSize) {
        *pSizeCategory = FILE_SIZE_CATEGORY_VERY_SMALL;
        u32 nTotalSize = nAttrHdrSize + *pAllocSize;
        u32 nLog2Size = WfsCalculateFieldSizeInBits(nTotalSize-1);
        if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
            nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
        }
        *pAllocSize = (1<<nLog2Size) - nAttrHdrSize;
        return nLog2Size;
    }
    // Note: This calculation needs to work with maximum file size = 4294967295
    u32 nNumSmallBlks = ((*pAllocSize-1)>>pAreaInfo->ah.nLog2BlkSize)+1;
    if (nNumSmallBlks <= DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS) {
        *pSizeCategory = FILE_SIZE_CATEGORY_SMALL;
        nAttrHdrSize += nNumSmallBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
        *pAllocSize = nNumSmallBlks << pAreaInfo->ah.nLog2BlkSize;
    } else {
        u32 nNumMediumBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerMediumBlk) + 1;
        if(nNumMediumBlks <= DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS){
            *pSizeCategory = FILE_SIZE_CATEGORY_MEDIUM;
            nAttrHdrSize += nNumMediumBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
            *pAllocSize = nNumMediumBlks << pAreaInfo->ah.nLog2MediumBlkSize;
        } else {
            u32 nNumLargeBlks = ((nNumMediumBlks - 1) >> pAreaInfo->nLog2MediumBlkPerLargeBlk) + 1;
            //u32 nNumLargeBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerLargeBlk) + 1;
            if (nNumLargeBlks <= DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS) {
                *pSizeCategory = FILE_SIZE_CATEGORY_LARGE;
                nAttrHdrSize += nNumLargeBlks * pAreaInfo->nLargeBlkPtrSize; // (sizeof(WFSBlkAdr) + pAreaInfo->nNumHashCodesPerLargeBlk*sizeof(WFSHashCode));
            } else {
                *pSizeCategory = FILE_SIZE_CATEGORY_VERY_LARGE;
                u32 nNumIndirectBlks = (nNumLargeBlks - 1) / pAreaInfo->nNumLargeBlkPtrsPerBlk + 1;
                nAttrHdrSize += nNumIndirectBlks * sizeof(WFSBlkAdr);
            }
            if (nNumLargeBlks >= 1<<(WFS_LOG2_MAX_FILE_SIZE-WFS_LOG2_LARGE_BLK_SIZE)){
                *pAllocSize = (u32)-1;
            } else {
                *pAllocSize = nNumLargeBlks * pAreaInfo->nLargeBlkSize;
            }
        }
    }
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nAttrHdrSize-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


u32 DirCalculateAttrSubBlkLog2SizeDecrease(AreaInfo *pAreaInfo, u32 nAllocSize, u32 nNameLen, u8 *pSizeCategory) {
    // This function calculates the size of the attribute sub-block necessary to store a file of the given size
    // There are 4 size categories: 0B <= Small <= 964B < Medium <= 320KB < Large <= 2560KB < Very Large < 4GB

    // Arguments:
    //   pAreaInfo      [In] .. A pointer to the AreaInfo struct for the area which contains/will contain the file
    //   nAllocSize     [In] .. The size of the file upon which to base the calculations
    //   nNameLen       [In] .. The Length of the file name in bytes
    //   pSizeCategory [Out] .. A pointer to a u32 variable which will be overwritten with flags indicating the block size & indirection level of the file
    // Returns the log base 2 of the size of the sub-block needed to store the file attributes.

    u32 nTotalSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((nNameLen+7)>>3);   // A bit per character of the name to represent upper/lower case
    u32 nNumSmallBlks = nAllocSize ? ((nAllocSize-1)>>pAreaInfo->ah.nLog2BlkSize)+1 : 1;
    if (nNumSmallBlks >= (1 << (WFS_LOG2_MAX_FILE_SIZE-pAreaInfo->ah.nLog2BlkSize))){
        nNumSmallBlks = 1 << (WFS_LOG2_MAX_FILE_SIZE-pAreaInfo->ah.nLog2BlkSize);
    }
    if (nNumSmallBlks <= DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS_DECREASE) {
        *pSizeCategory = FILE_SIZE_CATEGORY_SMALL;
        nTotalSize += nNumSmallBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
    } else {
        u32 nNumMediumBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerMediumBlk) + 1;
        if(nNumMediumBlks <= DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS_DECREASE){
            *pSizeCategory = FILE_SIZE_CATEGORY_MEDIUM;
            nTotalSize += nNumMediumBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
        } else {
            u32 nNumLargeBlks = ((nNumMediumBlks - 1) >> pAreaInfo->nLog2MediumBlkPerLargeBlk) + 1;
            //u32 nNumLargeBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerLargeBlk) + 1;
            if (nNumLargeBlks <= DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS_DECREASE) {
                *pSizeCategory = FILE_SIZE_CATEGORY_LARGE;
                nTotalSize += nNumLargeBlks * pAreaInfo->nLargeBlkPtrSize;  // (sizeof(WFSBlkAdr) + (sizeof(WFSHashCode) * pAreaInfo->nNumHashCodesPerLargeBlk));
            } else {
                *pSizeCategory = FILE_SIZE_CATEGORY_VERY_LARGE;
                u32 nNumIndirectBlks = (nNumLargeBlks - 1) / pAreaInfo->nNumLargeBlkPtrsPerBlk + 1;
                nTotalSize += nNumIndirectBlks * sizeof(WFSBlkAdr);
            }
        }
    }
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nTotalSize-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


u32 DirCalculateAttrSubBlkLog2SizeDecreaseUpdateAllocSize(AreaInfo *pAreaInfo, u32 *pAllocSize, u32 nNameLen, u8 *pSizeCategory) {
    // This version of the function updates *pAllocSize.
    // On entry, *pAllocSize is the minimum number of bytes that must be allocated.
    // On exit, *pAllocSize is the maximum number of bytes that can be used without re-allocation.
    // (This can be larger than the requested amount because we only allocate power-of-2 chunks within the block,
    //  or whole numbers of blocks outside the block).

    u32 nAttrHdrSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((nNameLen+7)>>3);   // A bit per character of the name to represent upper/lower case
    u32 nNumSmallBlks = *pAllocSize ? ((*pAllocSize-1)>>pAreaInfo->ah.nLog2BlkSize)+1 : 1;
    if (nNumSmallBlks <= DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS_DECREASE) {
        *pSizeCategory = FILE_SIZE_CATEGORY_SMALL;
        nAttrHdrSize += nNumSmallBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
        *pAllocSize = nNumSmallBlks << pAreaInfo->ah.nLog2BlkSize;
    } else {
        u32 nNumMediumBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerMediumBlk) + 1;
        if(nNumMediumBlks <= DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS_DECREASE){
            *pSizeCategory = FILE_SIZE_CATEGORY_MEDIUM;
            nAttrHdrSize += nNumMediumBlks * (sizeof(WFSBlkAdr) + sizeof(WFSHashCode));
            *pAllocSize = nNumMediumBlks << pAreaInfo->ah.nLog2MediumBlkSize;
        } else {
            u32 nNumLargeBlks = ((nNumMediumBlks - 1) >> pAreaInfo->nLog2MediumBlkPerLargeBlk) + 1;
            //u32 nNumLargeBlks = ((nNumSmallBlks - 1) >> pAreaInfo->nLog2BlksPerLargeBlk) + 1;
            if (nNumLargeBlks <= DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS_DECREASE) {
                *pSizeCategory = FILE_SIZE_CATEGORY_LARGE;
                nAttrHdrSize += nNumLargeBlks * pAreaInfo->nLargeBlkPtrSize; // (sizeof(WFSBlkAdr) + pAreaInfo->nNumHashCodesPerLargeBlk*sizeof(WFSHashCode));
            } else {
                *pSizeCategory = FILE_SIZE_CATEGORY_VERY_LARGE;
                u32 nNumIndirectBlks = (nNumLargeBlks - 1) / pAreaInfo->nNumLargeBlkPtrsPerBlk + 1;
                nAttrHdrSize += nNumIndirectBlks * sizeof(WFSBlkAdr);
            }
            if (nNumLargeBlks >= 1<<(WFS_LOG2_MAX_FILE_SIZE-WFS_LOG2_LARGE_BLK_SIZE)){
                *pAllocSize = (u32)-1;
            } else {
                *pAllocSize = nNumLargeBlks * pAreaInfo->nLargeBlkSize;
            }
        }
    }
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nAttrHdrSize-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}


WFSKrnResult DirFindFilePos(DirItr *pDi, WFSFileSize nPos, DirFilePos *pFPos, u32 nTransFlags) {
    // This function calculates the block and offset within that block corresponding to a file offset
    DirEntryAttrHdr *pAttrHdr = pDi->pAttrHdr;
    //pDi->tba = pDi->tba;
    //pFPos->nPos = nPos;
    //AreaInfo *pAreaInfo = pDi->tba.pAreaInfo;

    u32 nAttrHdrSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((pAttrHdr->nNameLen+7)>>3);

    //nTransFlags |= BCACHE_FLAG_PINNED;

    switch(pAttrHdr->nSizeCategory) {
        case FILE_SIZE_CATEGORY_VERY_SMALL: {
            if (pAttrHdr->file.nSize <= (1<<SBA_MAX_LOG2_SUB_BLK_SIZE) - nAttrHdrSize) {
                pFPos->pDataPtr = &pAttrHdr->aCaseBitArray[(pAttrHdr->nNameLen+7)>>3] + nPos;
                pFPos->nRemainingBlkSize = (1<<pDi->dni.nLog2Size) - (pFPos->pDataPtr - (u8*)pAttrHdr);
            }
            break;
        }
        case FILE_SIZE_CATEGORY_SMALL: {
            AreaInfo *pAreaInfo = pDi->tba.pAreaInfo;
            u32 nMaxBlkNum = pAttrHdr->file.nSize>>pAreaInfo->ah.nLog2BlkSize;
            u32 nPosBlkNum = nPos>>pAreaInfo->ah.nLog2BlkSize;
            u32 nPosBlkOfs = nPos%pAreaInfo->nBlkSize;
            u8 *pDataBlk;
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nPosBlkNum);
            u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
            u32 nValidSize = (nMaxBlkNum == nPosBlkNum) ? WFS_ROUND_UP(nPos, pAreaInfo->pVolInfo->pDevInfo->nSectorSize) : pAreaInfo->nBlkSize;
            WFSKrnReturnOnError(TransGetAndPinFileBlk(&pDi->tba, pDi->tba.nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, nTransFlags, nValidSize, (void**)&pDataBlk));
            pFPos->pDataPtr = pDataBlk + nPosBlkOfs;
            pFPos->nRemainingBlkSize = pAreaInfo->nBlkSize - nPosBlkOfs;
            WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, pFileBlkPtr->nBlkAdr, pDi->tba.pTransInfo));
            break;
        }
        case FILE_SIZE_CATEGORY_MEDIUM: {
            AreaInfo *pAreaInfo = pDi->tba.pAreaInfo;
            u32 nMaxMedBlkIdx = pAttrHdr->file.nSize>>pAreaInfo->ah.nLog2MediumBlkSize;
            u32 nMedBlkIdx = nPos>>pAreaInfo->ah.nLog2MediumBlkSize;
            u32 nMedBlkOfs = nPos%pAreaInfo->nMediumBlkSize;
            u8 *pDataBlk;
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nMedBlkIdx);
            u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
            u32 nValidSize = (nMaxMedBlkIdx == nMedBlkIdx) ? WFS_ROUND_UP(nPos, pAreaInfo->pVolInfo->pDevInfo->nSectorSize) : pAreaInfo->nMediumBlkSize;
            WFSKrnReturnOnError(TransGetAndPinFileMediumBlk(&pDi->tba, pDi->tba.nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, nTransFlags, nValidSize, (void**)&pDataBlk));
            pFPos->pDataPtr = pDataBlk + nMedBlkOfs;
            pFPos->nRemainingBlkSize = pAreaInfo->nMediumBlkSize - nMedBlkOfs;
            WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, pFileBlkPtr->nBlkAdr, pDi->tba.pTransInfo));
            break;
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            AreaInfo *pAreaInfo = pDi->tba.pAreaInfo;
            u32 nMaxMedBlkIdx = pAttrHdr->file.nSize>>pAreaInfo->ah.nLog2MediumBlkSize;
            u32 nMedBlkIdx = nPos>>pAreaInfo->ah.nLog2MediumBlkSize;
            u32 nMedBlkOfs = nPos%pAreaInfo->nMediumBlkSize;
            u8 *pDataBlk;
            WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nMedBlkIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
            u32 nValidSize = (nMaxMedBlkIdx == nMedBlkIdx) ? WFS_ROUND_UP(nPos, pAreaInfo->pVolInfo->pDevInfo->nSectorSize) : pAreaInfo->nMediumBlkSize;
            WFSKrnReturnOnError(TransGetAndPinFileMediumBlk(&pDi->tba, pDi->tba.nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, nTransFlags, nValidSize, (void**)&pDataBlk));
            pFPos->pDataPtr = pDataBlk + nMedBlkOfs;
            pFPos->nRemainingBlkSize = pAreaInfo->nMediumBlkSize - nMedBlkOfs;
            WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, fileBlkInfo.nBlkAdr, pDi->tba.pTransInfo));
            break;
        }
        case FILE_SIZE_CATEGORY_VERY_LARGE: {
            AreaInfo *pAreaInfo = pDi->tba.pAreaInfo;
            u32 nMaxMedBlkIdx = pAttrHdr->file.nSize>>pAreaInfo->ah.nLog2MediumBlkSize;
            u32 nMedBlkIdx = nPos>>pAreaInfo->ah.nLog2MediumBlkSize; // 64KB
            u32 nMedBlkLocalIdx = nMedBlkIdx%pAreaInfo->nNumMediumBlkPtrsPerBlk;
            u32 nMedBlkOfs = nPos%pAreaInfo->nMediumBlkSize;
            u8 *pIndirectBlk, *pDataBlk;
            WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndMedBlkIdxForCategoryVeryLarge(pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nMedBlkIdx);
            WFSKrnReturnOnError(TransGetAndPinBlk5(pDi->tba.pAreaInfo, *pIndirectBlkAdr, pDi->tba.pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlk));
            WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pAreaInfo, pIndirectBlk, nMedBlkLocalIdx);
            u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
            u32 nValidSize = (nMaxMedBlkIdx == nMedBlkIdx) ? WFS_ROUND_UP(nPos, pAreaInfo->pVolInfo->pDevInfo->nSectorSize) : pAreaInfo->nMediumBlkSize;
            WFSKrnReturnOnError(TransGetAndPinFileMediumBlk(&pDi->tba, pDi->tba.nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, nTransFlags, nValidSize, (void**)&pDataBlk));
            pFPos->pDataPtr = pDataBlk + nMedBlkOfs;
            pFPos->nRemainingBlkSize = pAreaInfo->nMediumBlkSize - nMedBlkOfs;
            WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, fileBlkInfo.nBlkAdr, pDi->tba.pTransInfo));
            WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, *pIndirectBlkAdr, pDi->tba.pTransInfo));
            break;
        }
    }
#if _DEBUG_DIR
    if (pFPos->nRemainingBlkSize > pDi->tba.pAreaInfo->nBlkSize) {
        WFSKrnOutputErrorStr("File Pos Error");
    }
#endif
    return WFSKRN_RESULT_OK;
}

static inline void DirNodeUpdateParentOfsPtr(DirItr *pDi, u16 nNewOfs) {
    DirNodeStack *pDns = pDi->dni.pDns;
    DirNodeStack *pDnsParent = pDns->pParent;
    pDns->nSubBlkOfs = nNewOfs;
    if (pDnsParent->nBlkAdr != pDns->nBlkAdr) {
        pDi->pBlkHdr->nRootOfs = nNewOfs;
        return;
    }
    DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pDi->pBlkHdr + pDnsParent->nSubBlkOfs);
    DirNodeSetIntraBlkOfs(pDnsParent->nEntryIdx, nNewOfs, pSubBlkHdr);
}

static inline void DirLeafUpdateParentOfsPtr(DirItr *pDi, u16 nNewOfs) {
    DirNodeStack *pDns = pDi->dni.pDns;
    DirNodeStack *pDnsParent = pDns->pParent;
    pDns->nSubBlkOfs = nNewOfs;
    if (pDnsParent->nBlkAdr != pDns->nBlkAdr) {
        pDi->pBlkHdr->nRootOfs = nNewOfs;
        return;
    }
    DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pDi->pBlkHdr + pDnsParent->nSubBlkOfs);
    DirLeafSetIntraBlkOfs(pDnsParent->nEntryIdx, nNewOfs, pSubBlkHdr);
}

inline WFSKrnResult DirUnpinAndFreeBlk(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
    // See "Update Counter & Update Map Operations"
    PTreeItr pti;
    WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
    WFSKrnResult nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
    if (nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
        u32 *pData = &pti.pLeaf->aData[pti.aEntryIdx[0]];
        *pData = (*pData & ~WFS_MDF_DIR_COUNTER_MASK) | ((*pData + 1) & WFS_MDF_DIR_COUNTER_MASK);
    }
    return TransUnpinAndFreeBlks(pAreaInfo, pAreaInfo->ah.nLog2BlkSize, 1, &nBlkAdr, pTransInfo);
}

WFSKrnResult DirGetBlk(DirItr *pDi, u32 nTransFlags, DirBlkHdr **ppBlkHdr) {
    // Opens a directory block for modification.
    // See "Update Counter & Update Map Operations"
    DirBlkHdr *pBlkHdr;
    WFSKrnReturnOnError(TransGetBlk(&pDi->tba, nTransFlags, (void**)&pBlkHdr));
    pDi->pBlkHdr = pBlkHdr;
    if(ppBlkHdr){
        *ppBlkHdr = pBlkHdr;
    }
    if (nTransFlags & TRANS_FLAG_WRITE_BLK) {
        pDi->nUpdateCtr = (pBlkHdr->sbah.mdh.nFlags + 1) & WFS_MDF_DIR_COUNTER_MASK;
        pBlkHdr->sbah.mdh.nFlags = (WFSMetaDataFlags)((pBlkHdr->sbah.mdh.nFlags & ~WFS_MDF_DIR_COUNTER_MASK) | pDi->nUpdateCtr);
        PTreeItr pti;
        WFSBlkAdr nAbsBlkAdr = pDi->tba.pAreaInfo->nAbsStartBlkAdr + pDi->tba.nBlkAdr;
        WFSKrnResult nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
        if (nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
            u32 *pData = &pti.pLeaf->aData[pti.aEntryIdx[0]];
            *pData = (*pData & ~WFS_MDF_DIR_COUNTER_MASK) | (pDi->nUpdateCtr & WFS_MDF_DIR_COUNTER_MASK);
        }
    } else {
        pDi->nUpdateCtr = pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_COUNTER_MASK;
    }
    return WFSKRN_RESULT_OK;
}

inline void DirReinitBlk(AreaInfo *pAreaInfo, DirBlkHdr *pBlkHdr, WFSMetaDataFlags nFlags) {
    u32 nLog2BlkSize = pAreaInfo->ah.nLog2BlkSize;
    u32 nUpdateCtr = pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_COUNTER_MASK;
    SbaBlkInit(pBlkHdr, nLog2BlkSize, WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr)));
    pBlkHdr->sbah.mdh.nFlags = (WFSMetaDataFlags)((WFS_MDF_DIR_BLK | nFlags) | nUpdateCtr);
}

WFSKrnResult DirCreateAndInitNewBlk4(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, WFSMetaDataFlags nFlags, TransInfo *pTransInfo, DirBlkHdr **ppBlkHdr) {
    WFSKrnReturnOnError(TransGetAndPinBlk5(pAreaInfo, nBlkAdr, pTransInfo, TRANS_FLAG_WRITE_BLK, (void**)ppBlkHdr));
    DirBlkHdr *pBlkHdr = *ppBlkHdr;
    DirReinitBlk(pAreaInfo, pBlkHdr, nFlags);
    PTreeItr pti;
    WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
#if _DEBUG
    WFSKrnResult nResult = PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator);
#else
    PTreeFind(&dirGlobals.updateMap, &pti, nAbsBlkAdr, &dirGlobals.updateMapAllocator); // [check] error handling.
#endif
    pBlkHdr->sbah.mdh.nFlags = (WFSMetaDataFlags)(pBlkHdr->sbah.mdh.nFlags & ~WFS_MDF_DIR_COUNTER_MASK);
#if _DEBUG
    if (nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND)
#endif
    {
        pBlkHdr->sbah.mdh.nFlags = (WFSMetaDataFlags)(pBlkHdr->sbah.mdh.nFlags | (pti.nData & WFS_MDF_DIR_COUNTER_MASK));
    }
    return WFSKRN_RESULT_OK;
}

void DirBlkCreateRootNode(DirBlkHdr *pBlkHdr) {
    pBlkHdr->nRootOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, SBA_MIN_LOG2_SUB_BLK_SIZE);
    DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pBlkHdr->nRootOfs);
    pSubBlkHdr->nStrLen = 0;
    pSubBlkHdr->nNumEntries = 0;
    pSubBlkHdr->aFirstChar[0] = (utf8)0xC1; // invalid utf8 character - to help debug
}

extern u32 nDebugNumChecks;
WFSKrnResult DirNextRaw(WFSFileName *pName, DirItr *pDi) {
    // ToDo: Make this function more compatible with DirFind
#if _DEBUG_DIR_NODE_STACK
    bool bOldCheckDirNodeStack = bCheckDirNodeStack;
    bCheckDirNodeStack = false;
#endif
    WFSKrnResult nResult;
    DirNodeItr dni = pDi->dni;
    DirBlkHdr *pBlkHdr = pDi->pBlkHdr;
    u32 nSubBlkStrLen;
    u16 nOfs = 0;

    while(1) {
        if (pDi->tba.nBlkAdr != dni.pDns->nBlkAdr) {
            TransUnpinBlk(&pDi->tba);
            pDi->tba.nBlkAdr = dni.pDns->nBlkAdr;
            nResult = DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr);
            if(nResult < WFSKRN_RESULT_OK){
                goto Done;
            }
        }
        /*
        // [check] This debug code will throw exception when pDi is the root directory(?),
        //         because dni.pDns->pParent is not initialized.
        if(dni.pDns->pParent){
            if(dni.pDns != dni.pDns->pParent && dni.pDns->nBlkAdr == dni.pDns->pParent->nBlkAdr && dni.pDns->nSubBlkOfs == dni.pDns->pParent->nSubBlkOfs){
                int a; a=0;
            }
        }
        */
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + dni.pDns->nSubBlkOfs);
        if (dni.nDnsDepth<=1) {
            pDi->dni = dni;
            nResult = WFSKRN_RESULT_NOT_FOUND;
            goto Done;
        }
        if (dni.pDns->nEntryIdx < dni.pSubBlkHdr->nNumEntries) {
            break;
        }
        DirPopNodeStackEntry(&dni);
    }
    u32 nStrIdx = dni.pDns->nStrIdx;
    dni.pStrPtr = &pName->sStr[nStrIdx];
    *dni.pStrPtr++ = dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen + dni.pDns->nEntryIdx];
    ++dni.pDns->nEntryIdx;
    u16 *pOfsTblEnd;

    if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
        nOfs = DirLeafGetIntraBlkOfs(dni.pDns->nEntryIdx, dni.pSubBlkHdr);
        dbg(SbaCheckOfs(nOfs));
    } else {
        // traverse internal nodes
        dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
        u16 *pOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));
        nOfs = DirNodeGetIntraBlkOfs_ote(dni.pDns->nEntryIdx, dni.pSubBlkHdr, pOfsTblEnd);
        dbg(SbaCheckOfs(nOfs));
        while(1) {
            dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
            dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
            pOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));
            WFSKrnReturnOnError(DirNodeStackGetEntry(&dni));
            dni.pDns->nBlkAdr = pDi->tba.nBlkAdr;
            dni.pDns->nSubBlkOfs = nOfs;
            dni.pDns->nEntryIdx = 1;
            nStrIdx += dni.pSubBlkHdr->nStrLen + 1;
            dni.pDns->nStrIdx = (u8)nStrIdx;
            if (dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen] == 0) {
                // We have found the end of the splitter. Change block.
                WFSBlkAdr nNewBlkAdr = DirNodeGetBlkAdr_ote(dni.pSubBlkHdr, pOfsTblEnd);
                TransUnpinBlk(&pDi->tba);
                pDi->tba.nBlkAdr = nNewBlkAdr;
                WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
                nOfs = pBlkHdr->nRootOfs;
                if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
#if (DIR_PREFIX_SUPPRESSION==0)
                    nStrIdx=0;
#endif
                    dni.pStrPtr = &pName->sStr[nStrIdx];
                    break;
                }
            } else {
                nOfs = DirNodeGetIntraBlkOfs_ote(1, dni.pSubBlkHdr, pOfsTblEnd);
                dbg(SbaCheckOfs(nOfs));
            }
        }
        nStrIdx--;
    }

    // traverse leaf node
    while(1) {
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
        nSubBlkStrLen = dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen + 1;
        //utf8 *pSubBlkStrPtr = dni.pSubBlkHdr->aFirstChar;
        dni.pNodeStrPtr = dni.pSubBlkHdr->aFirstChar;
        while(nSubBlkStrLen--) {
            *dni.pStrPtr++ = *dni.pNodeStrPtr++;
        }
        if (dni.pSubBlkHdr->nNumEntries > 1) {
            WFSKrnReturnOnError(DirNodeStackGetEntry(&dni));
            dni.pDns->nBlkAdr = pDi->tba.nBlkAdr;
            dni.pDns->nSubBlkOfs = nOfs;
            dni.pDns->nEntryIdx = 1;
            dni.pDns->nStrIdx = (u8)nStrIdx;
        }
        utf8 cFirstChoice = *dni.pNodeStrPtr++;
        *dni.pStrPtr++ = cFirstChoice;
        dni.nLog2Size = DirLeafGetSubBlkLog2Size(dni.pSubBlkHdr);
        pOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));
        if (cFirstChoice==0) {
            // We have found the end of the entry name
            pName->nLen = dni.pStrPtr - pName->sStr - 1;
            pDi->dni = dni;
            pDi->pBlkHdr = pBlkHdr;
            pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_ote(1, pOfsTblEnd);
            dbg(SbaCheckOfs(nOfs));
            pDi->pAttrHdr = (DirEntryAttrHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
            pDi->dni.nLog2Size = pDi->pAttrHdr->nAttrLog2Size;
            nResult = WFSKRN_RESULT_OK;
            goto Done;
        }
#if _DEBUG_DIR
        if ((nDirTest >= nDirTestBreakPoint) && (pDi->dni.pDns->nSubBlkOfs == 256)) {
           int a; a=0;
        }
#endif
        nOfs = DirLeafGetIntraBlkOfs_ote(dni.pDns->nEntryIdx, pOfsTblEnd);
        dbg(SbaCheckOfs(nOfs));
    }
Done:
#if _DEBUG_DIR_NODE_STACK
    bCheckDirNodeStack = bOldCheckDirNodeStack;
#endif
    return nResult;
}

static
u16 DirNodePrependStringToSubBlk(DirBlkHdr *pBlkHdr, u16 nOldOfs, const utf8 *pStr, u8 nStrLen, utf8 cJoinChar) {
    // A non-Node node is not allowed to end up with exactly 1 child. This function can be used to merge it with the child.
    // This function prepends the string pStr followed by cJoinChar to the start of the node specified by pBlkHdr and nOldOfs.
    // This function reallocates the node if it grows beyond its existing size.
    // It returns the resultant offset address of the updated node, or 0 if allocation fails.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    u32 nOldLog2Size = DirNodeGetSubBlkLog2Size(pOldSubBlkHdr);
    u32 nNewStrLen = (u32)(pOldSubBlkHdr->nStrLen + nStrLen + 1);
    u32 nNewLog2Size = DirNodeGetSubBlkLog2Size_3(nNewStrLen, pOldSubBlkHdr->nNumEntries, pOldSubBlkHdr->aFirstChar[pOldSubBlkHdr->nStrLen]);
    if (nNewLog2Size == nOldLog2Size) {
        memmove(&pOldSubBlkHdr->aFirstChar[nStrLen+1], pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen + pOldSubBlkHdr->nNumEntries));
        memcpy(pOldSubBlkHdr->aFirstChar, pStr, nStrLen);
        pOldSubBlkHdr->aFirstChar[nStrLen] = cJoinChar;
        pOldSubBlkHdr->nStrLen = (u8)nNewStrLen;
        return nOldOfs;
    }
    // The merged node will require more space than the original, so it must be reallocated
    u32 nNewNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewLog2Size);
    if (nNewNodeOfs == 0) {
        // Could not allocate a larger block
        return 0;
    }
    DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewNodeOfs);
    memcpy(pNewSubBlkHdr->aFirstChar, pStr, nStrLen); // copy includes the terminator
    pNewSubBlkHdr->aFirstChar[nStrLen] = cJoinChar;
    s32 nNumEntries = pNewSubBlkHdr->nNumEntries = pOldSubBlkHdr->nNumEntries;
    memmove(&pNewSubBlkHdr->aFirstChar[nStrLen+1], pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen + nNumEntries));
    pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
    u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pOldSubBlkHdr, nOldLog2Size);
    u8  *pSrcEnd   = (u8*)pOldSubBlkHdr + (1<<nOldLog2Size);
    u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pNewSubBlkHdr, nNewLog2Size);
    memcpy(pDstStart, pSrcStart, (u32)(pSrcEnd - (u8*)pSrcStart));
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewNodeOfs;
}

static
u16 DirLeafPrependStringToSubBlk(DirBlkHdr *pBlkHdr, u16 nOldOfs, const utf8 *pStr, u8 nStrLen, utf8 cJoinChar) {
    // A non-leaf node is not allowed to end up with exactly 1 child. This function can be used to merge it with the child.
    // This function prepends the string pStr followed by cJoinChar to the start of the node specified by pBlkHdr and nOldOfs.
    // This function reallocates the node if it grows beyond its existing size.
    // It returns the resultant offset address of the updated node, or 0 if allocation fails.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    u32 nOldLog2Size = DirLeafGetSubBlkLog2Size(pOldSubBlkHdr);
    u32 nNewStrLen = (u32)(pOldSubBlkHdr->nStrLen + nStrLen + 1);
    u32 nNewLog2Size = DirLeafGetSubBlkLog2Size_2(nNewStrLen, pOldSubBlkHdr->nNumEntries);
    if (nNewLog2Size == nOldLog2Size) {
        memmove(&pOldSubBlkHdr->aFirstChar[nStrLen+1], pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen + pOldSubBlkHdr->nNumEntries));
        memcpy(pOldSubBlkHdr->aFirstChar, pStr, nStrLen);
        pOldSubBlkHdr->aFirstChar[nStrLen] = cJoinChar;
        pOldSubBlkHdr->nStrLen = (u8)nNewStrLen;
        return (u16)nOldOfs;
    }
    // The merged node will require more space than the original, so it must be reallocated
    u32 nNewNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewLog2Size);
    if (nNewNodeOfs == 0) {
        // Could not allocate a larger block
        return 0;
    }
    DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewNodeOfs);
    memcpy(pNewSubBlkHdr->aFirstChar, pStr, nStrLen); // copy includes the terminator
    pNewSubBlkHdr->aFirstChar[nStrLen] = cJoinChar;
    s32 nNumEntries = pNewSubBlkHdr->nNumEntries = pOldSubBlkHdr->nNumEntries;
    memmove(&pNewSubBlkHdr->aFirstChar[nStrLen+1], pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen + nNumEntries));
    pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
    u16 *pNewOfsTblEnd = (u16*)((u8*)pNewSubBlkHdr + (1<<nNewLog2Size));
    u16 *pOldOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nOldLog2Size));
    memcpy(&pNewOfsTblEnd[-nNumEntries], &pOldOfsTblEnd[-nNumEntries], sizeof(u16) * nNumEntries);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewNodeOfs;
}

static
WFSKrnResult DirLeafCompact(DirItr *pDi, u16 nResizeOfs, u16 nNewLog2Size, u16 nSkipOfs, WFSBlkAdr *pTempBlkAdr, DirBlkHdr **ppTempBlkHdr) {
    // This function can be used when a Leaf block is nearly full
    // It compacts the specified Leaf block by temporarily copying to a new block.
    // While the copy is being performed, optionally the sub block at offset nResizeOfs can be resized to nNewLog2Size,
    // and an optional sub block can be skipped (nSkipOfs)
    // This function returns the allocated temporary block in *pTempBlkAdr, *ppTempBlkHdr
    WFSKrnResult nResult;
    DirItr diSrc=*pDi, diDst;
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
    WFSFileName name;
    diSrc.dni.pStrPtr = name.sStr;
#endif
    u32 nStrIdx=0, nSubBlkStrLen;

    WFSBlkAdr nBlkAdr = diSrc.tba.nBlkAdr;

    DirNodeStack *pDnsCheckStart = diSrc.dni.pDns;
    DirNodeStack *pDns = pDnsCheckStart;
    u32 nDnsDepthStart = pDi->dni.nDnsDepth;
    u32 nDnsDepth = nDnsDepthStart;
    while(--nDnsDepth) {
        if (pDns->nBlkAdr != nBlkAdr) {
            nDnsDepthStart -= nDnsDepth;
            break;
        }
        pDns = pDns->pParent;
    }
    diSrc.dni.nDnsDepth = 1;

    // Allocate new blocks
    WFSKrnReturnOnError(AreaAllocBlks(diSrc.tba.pAreaInfo, nBlkAdr, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, &diDst.tba.nBlkAdr, sizeof(WFSBlkAdr)));
    nResult = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, diDst.tba.nBlkAdr, diSrc.pBlkHdr->sbah.mdh.nFlags, diSrc.tba.pTransInfo, &diDst.pBlkHdr);
    if(nResult < WFSKRN_RESULT_OK){
        AreaFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, &diDst.tba.nBlkAdr, sizeof(WFSBlkAdr));
        return nResult;
    }
    // Initialize a stack frame for traversing the destination block
    diDst.dni.nDnsDepth = 0;
    nResult = DirNodeStackGetEntry(&diDst.dni);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diDst.ppDnsRootParent = &diDst.dni.pDns->pParent;
    diSrc.dni.nEntryIdx = 1;
    diSrc.dni.pOfs = &diSrc.pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(*diSrc.dni.pOfs));
    nResult = DirPushNodeStackEntry(&diSrc.dni, nBlkAdr);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diSrc.dni.pDns->nStrIdx = 0;    // ToDo: This value can be initialized to 0 because prefix-suppression is not implemented
    u16 nSrcLeafOfs = diSrc.dni.pDns->nSubBlkOfs; // = diSrc.pBlkHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nRecIdx=0;
    diDst.dni.pOfs = &diDst.pBlkHdr->nRootOfs;
    u32 nSrcLeafStartDepth = diSrc.dni.nDnsDepth;

    // Perform a depth-first traversal of the block, until the middle leaf sub-block is found.
    while(1) {
        diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + nSrcLeafOfs);
        diSrc.dni.nLog2Size = DirLeafGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        // Copy over the whole Leaf from the original block
        nResult = DirNodeStackGetEntry(&diDst.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        u32 nCopySize = 1<<diSrc.dni.nLog2Size;
        if (nSrcLeafOfs == nSkipOfs) {
            nCopySize = 0;
        } else {
            if (nSrcLeafOfs == nResizeOfs) {
                diSrc.dni.nLog2Size = nNewLog2Size;
            }
            *diDst.dni.pOfs = diDst.dni.pDns->nSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, diSrc.dni.nLog2Size);
            if (*diDst.dni.pOfs == 0) {
                WFSKrnOutputErrorStr("Could not allocate sub block");
            }
        }
        diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + *diDst.dni.pOfs);

        // Fix up the ofs pointers on the stack
        pDns = pDnsCheckStart;
        nDnsDepth = nDnsDepthStart;
        while(--nDnsDepth) {
            if (pDns->nBlkAdr != nBlkAdr) {
                break;
            }
            if (pDns->nSubBlkOfs == nSrcLeafOfs) {
                pDns->nSubBlkOfs = *diDst.dni.pOfs;
                --nDnsDepthStart;
            }
            pDns = pDns->pParent;
        }

        memcpy(diDst.dni.pSubBlkHdr, diSrc.dni.pSubBlkHdr, nCopySize);
        diDst.dni.pDns->nEntryIdx = 1;
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        diSrc.dni.pDns->nEntryIdx = 1;
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen];

        nSubBlkStrLen = diSrc.dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen;
        diSrc.dni.pDns->nStrIdx = (u8)nStrIdx;
#if _DEBUG_DIR
        //utf8 *pSubBlkStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        diSrc.dni.pNodeStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        while(nSubBlkStrLen--) {
            *diSrc.dni.pStrPtr++ = *diSrc.dni.pNodeStrPtr++;
        }
#endif
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment our record count
#if _DEBUG_DIR
            *diSrc.dni.pStrPtr = 0;
            ++pChoiceChar;
#endif
            // Copy file attributes to new block
            u32 nAttrOfs = pSrcNodeOfsTblEnd[-1];
            dbg(SbaCheckOfs(nAttrOfs));
            DirEntryAttrHdr *pAttrHdr = (DirEntryAttrHdr *)((u8*)diSrc.pBlkHdr + nAttrOfs);
            u16 *pDstNodeOfsTblEnd;
            if (nSrcLeafOfs == nSkipOfs) {
                pDstNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
            } else {
                pDstNodeOfsTblEnd = (u16*)((u8*)diDst.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
            }
            pDstNodeOfsTblEnd[-1] = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, pAttrHdr->nAttrLog2Size);
            DirEntryAttrHdr *pDstAttrHdr = (DirEntryAttrHdr *)((u8*)diDst.pBlkHdr + pDstNodeOfsTblEnd[-1]);
            memcpy(pDstAttrHdr, pAttrHdr, (u32)(1<<pAttrHdr->nAttrLog2Size));
            ++nRecIdx;
            ++diSrc.dni.pDns->nEntryIdx;
            ++diDst.dni.pDns->nEntryIdx;
        }
        while(diSrc.dni.pDns->nEntryIdx > diSrc.dni.pSubBlkHdr->nNumEntries) {
            if (diSrc.dni.nDnsDepth == nSrcLeafStartDepth) {
                // We reached the top of the stack, so this should be the last record
                DirPopNodeStackEntry(&diSrc.dni);
                diDst.pBlkHdr->nRootOfs = diDst.dni.pDns->nSubBlkOfs;
                DirNodeStackFree(&diDst.dni, diDst.ppDnsRootParent);
                diDst.pBlkHdr->nNumRecs = nRecIdx;
                if (nResult != WFSKRN_RESULT_OK) {
                    goto Error;
                }
                *pTempBlkAdr = diDst.tba.nBlkAdr;
                *ppTempBlkHdr = diDst.pBlkHdr;
                return WFSKRN_RESULT_OK;
            }
            DirPopNodeStackEntry(&diDst.dni);
            DirPopNodeStackEntry(&diSrc.dni);
#if _DEBUG_DIR
            diSrc.dni.pStrPtr += diSrc.dni.pDns->nStrIdx - nStrIdx;
#endif
            nStrIdx = diSrc.dni.pDns->nStrIdx;
            diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + diSrc.dni.pDns->nSubBlkOfs);
            ++diDst.dni.pDns->nEntryIdx;
            ++diSrc.dni.pDns->nEntryIdx;
        }
#if _DEBUG_DIR
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen + diSrc.dni.pDns->nEntryIdx-1];
        *diSrc.dni.pStrPtr = *pChoiceChar;
#endif
        ++nStrIdx;
        diSrc.dni.nLog2Size = DirLeafGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        {
            diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + diDst.dni.pDns->nSubBlkOfs);
            if (diSrc.dni.pDns->nSubBlkOfs == nSkipOfs) {
                diDst.dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diSrc.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
            } else {
                diDst.dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
            }
        }
        nSrcLeafOfs = DirLeafGetIntraBlkOfs_ote(diSrc.dni.pDns->nEntryIdx, pSrcNodeOfsTblEnd);
        dbg(SbaCheckOfs(nSrcLeafOfs));
        nResult = DirNodeStackGetEntry(&diSrc.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        diSrc.dni.pDns->nBlkAdr = nBlkAdr;
        diSrc.dni.pDns->nSubBlkOfs = nSrcLeafOfs;
#if _DEBUG_DIR
        ++diSrc.dni.pStrPtr;
#endif
    }
Error:
    TransUnpinAndFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC, &diDst.tba.nBlkAdr, diSrc.tba.pTransInfo);
    return nResult;
}


static
WFSKrnResult DirNodeCompact(DirItr *pDi, u16 nResizeOfs, u16 nNewLog2Size, u16 nSkipOfs, WFSBlkAdr *pTempBlkAdr, DirBlkHdr **ppTempBlkHdr) {
    // This function can be used when a node block is nearly full
    // It compacts the specified node block by temporarily copying to a new block.
    // While the copy is being performed, optionally the sub block at offset nResizeOfs can be resized to nNewLog2Size,
    // and an optional sub block can be skipped (nSkipOfs)
    // This function returns the allocated temporary block in *pTempBlkAdr, *ppTempBlkHdr
    WFSKrnResult nResult;
    DirItr diSrc=*pDi, diDst;
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
    WFSFileName name;
    diSrc.dni.pStrPtr = name.sStr;
#endif
    u32 nStrIdx=0, nSubBlkStrLen;

    WFSBlkAdr nBlkAdr = diSrc.tba.nBlkAdr;

    DirNodeStack *pDnsCheckStart = diSrc.dni.pDns;
    DirNodeStack *pDns = pDnsCheckStart;
    u32 nDnsDepthStart = pDi->dni.nDnsDepth;
    u32 nDnsDepth = nDnsDepthStart;
    while(--nDnsDepth) {
        if (pDns->nBlkAdr != nBlkAdr) {
            nDnsDepthStart -= nDnsDepth;
            break;
        }
        pDns = pDns->pParent;
    }
    diSrc.dni.nDnsDepth = 1;

    // Allocate new blocks
    WFSKrnReturnOnError(AreaAllocBlks(diSrc.tba.pAreaInfo, nBlkAdr, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, &diDst.tba.nBlkAdr, sizeof(WFSBlkAdr)));
    nResult = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, diDst.tba.nBlkAdr, diSrc.pBlkHdr->sbah.mdh.nFlags, diSrc.tba.pTransInfo, &diDst.pBlkHdr);
    if(nResult < WFSKRN_RESULT_OK){
        AreaFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, &diDst.tba.nBlkAdr, sizeof(WFSBlkAdr));
        return nResult;
    }
    // Initialize a stack frame for traversing the destination block
    diDst.dni.nDnsDepth = 0;
    nResult = DirNodeStackGetEntry(&diDst.dni);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diDst.ppDnsRootParent = &diDst.dni.pDns->pParent;
    diSrc.dni.nEntryIdx = 1;
    diSrc.dni.pOfs = &diSrc.pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(*diSrc.dni.pOfs));
    nResult = DirPushNodeStackEntry(&diSrc.dni, nBlkAdr);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diSrc.dni.pDns->nStrIdx = 0;    // ToDo: This value can be initialized to 0 because prefix-suppression is not implemented
    u16 nSrcNodeOfs = diSrc.dni.pDns->nSubBlkOfs; // = diSrc.pBlkHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nRecIdx=0;
    diDst.dni.pOfs = &diDst.pBlkHdr->nRootOfs;
    u32 nSrcLeafStartDepth = diSrc.dni.nDnsDepth;

    // Perform a depth-first traversal of the block, until the middle leaf sub-block is found.
    while(1) {
        diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + nSrcNodeOfs);
        diSrc.dni.nLog2Size = DirNodeGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        // Copy over the whole node from the original block
        nResult = DirNodeStackGetEntry(&diDst.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        u32 nCopySize = 1<<diSrc.dni.nLog2Size;
        if (nSrcNodeOfs == nSkipOfs) {
            nCopySize = 0;
        } else {
            if (nSrcNodeOfs == nResizeOfs) {
                diSrc.dni.nLog2Size = nNewLog2Size;
            }
            *diDst.dni.pOfs = diDst.dni.pDns->nSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, diSrc.dni.nLog2Size);
            if (*diDst.dni.pOfs == 0) {
                WFSKrnOutputErrorStr("Could not allocate sub block");
            }
        }
        diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + *diDst.dni.pOfs);

        // Fix up the ofs pointers on the stack
        pDns = pDnsCheckStart;
        nDnsDepth = nDnsDepthStart;
        while(--nDnsDepth) {
            if (pDns->nBlkAdr != nBlkAdr) {
                break;
            }
            if (pDns->nSubBlkOfs == nSrcNodeOfs) {
                pDns->nSubBlkOfs = *diDst.dni.pOfs;
                --nDnsDepthStart;
            }
            pDns = pDns->pParent;
        }

        memcpy(diDst.dni.pSubBlkHdr, diSrc.dni.pSubBlkHdr, nCopySize);
        diDst.dni.pDns->nEntryIdx = 1;
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        diSrc.dni.pDns->nEntryIdx = 1;
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen];

        nSubBlkStrLen = diSrc.dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen;
        diSrc.dni.pDns->nStrIdx = (u8)nStrIdx;
#if _DEBUG_DIR
        //utf8 *pSubBlkStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        diSrc.dni.pNodeStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        while(nSubBlkStrLen--) {
            *diSrc.dni.pStrPtr++ = *diSrc.dni.pNodeStrPtr++;
        }
#endif
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment our record count
#if _DEBUG_DIR
            *diSrc.dni.pStrPtr = 0;
            ++pChoiceChar;
#endif
            ++nRecIdx;
            ++diSrc.dni.pDns->nEntryIdx;
            ++diDst.dni.pDns->nEntryIdx;
        }
        while(diSrc.dni.pDns->nEntryIdx > diSrc.dni.pSubBlkHdr->nNumEntries) {
            if (diSrc.dni.nDnsDepth == nSrcLeafStartDepth) {
                // We reached the top of the stack, so this should be the last record
                DirPopNodeStackEntry(&diSrc.dni);
                diDst.pBlkHdr->nRootOfs = diDst.dni.pDns->nSubBlkOfs;
                DirNodeStackFree(&diDst.dni, diDst.ppDnsRootParent);
                diDst.pBlkHdr->nNumRecs = nRecIdx;
                if (nResult != WFSKRN_RESULT_OK) {
                    goto Error;
                }
                *pTempBlkAdr = diDst.tba.nBlkAdr;
                *ppTempBlkHdr = diDst.pBlkHdr;
                return WFSKRN_RESULT_OK;
            }
            DirPopNodeStackEntry(&diDst.dni);
            DirPopNodeStackEntry(&diSrc.dni);
#if _DEBUG_DIR
            diSrc.dni.pStrPtr += diSrc.dni.pDns->nStrIdx - nStrIdx;
#endif
            nStrIdx = diSrc.dni.pDns->nStrIdx;
            diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + diSrc.dni.pDns->nSubBlkOfs);
            ++diDst.dni.pDns->nEntryIdx;
            ++diSrc.dni.pDns->nEntryIdx;
        }
#if _DEBUG_DIR
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen + diSrc.dni.pDns->nEntryIdx-1];
        *diSrc.dni.pStrPtr = *pChoiceChar;
#endif
        ++nStrIdx;
        diSrc.dni.nLog2Size = DirNodeGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        {
            diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + diDst.dni.pDns->nSubBlkOfs);
            if (diSrc.dni.pDns->nSubBlkOfs == nSkipOfs) {
                diDst.dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diSrc.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
            } else {
                diDst.dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
            }
        }
        nSrcNodeOfs = DirNodeGetIntraBlkOfs_ote(diSrc.dni.pDns->nEntryIdx, diSrc.dni.pSubBlkHdr, pSrcNodeOfsTblEnd);
        dbg(SbaCheckOfs(nSrcNodeOfs));
        nResult = DirNodeStackGetEntry(&diSrc.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        diSrc.dni.pDns->nBlkAdr = nBlkAdr;
        diSrc.dni.pDns->nSubBlkOfs = nSrcNodeOfs;
#if _DEBUG_DIR
        ++diSrc.dni.pStrPtr;
#endif
    }
Error:
    TransUnpinAndFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC, &diDst.tba.nBlkAdr, diSrc.tba.pTransInfo);
    return nResult;
}

static
WFSKrnResult DirNodeBlkSplit(DirItr *pDiSrc, WFSBlkAdr *aBlkAdr, u32 *pFirstBlockToFree);

static
u16 DirNodeMergeSubBlkWithItsSingleChild(DirItr *pDi, DirBlkHdr *pBlkHdr, u16 nOldOfs) {
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    // A non-leaf node of the radix tree is not allowed to end up with exactly 1 child. This function can be used to merge it with the child.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    u32 nOldLog2Size = DirNodeGetSubBlkLog2Size(pOldSubBlkHdr);
    u16 *pOldNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nOldLog2Size));
    u32 nOldChildOfs = pOldNodeOfsTblEnd[-1];
    dbg(SbaCheckOfs(nOldChildOfs));
    DirSubBlkHdr *pOldChildHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldChildOfs);
    u32 nOldChildLog2Size = DirNodeGetSubBlkLog2Size(pOldChildHdr);
    u32 nNewStrLen = (u32)(pOldSubBlkHdr->nStrLen + pOldChildHdr->nStrLen + 1);
    u32 nNewNodeLog2Size = DirNodeGetSubBlkLog2Size_3(nNewStrLen, pOldChildHdr->nNumEntries, pOldChildHdr->aFirstChar[pOldChildHdr->nStrLen]);
    if (nOldChildLog2Size == nNewNodeLog2Size) {
        memmove(&pOldChildHdr->aFirstChar[pOldSubBlkHdr->nStrLen+1], pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
        memcpy(pOldChildHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
        pOldChildHdr->nStrLen = (u8)nNewStrLen;
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
        return (u16)nOldChildOfs;
    }
    // The merged node will require more space than the original, so it must be reallocated
    u32 nNewChildNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewNodeLog2Size);
    if (nNewChildNodeOfs == 0) {
        // Could not allocate a larger block
        if (pDi == 0) {
            return 0;
        }
        // This section was added for the delete function - compact node and update node stack
        WFSBlkAdr nTempBlkAdr;
        DirBlkHdr *pTempBlkHdr;
        WFSKrnReturnOnError(DirNodeCompact(pDi, nOldOfs, nNewNodeLog2Size, nOldChildOfs, &nTempBlkAdr, &pTempBlkHdr));
        nNewChildNodeOfs = pDi->dni.pDns->nSubBlkOfs;
        DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pTempBlkHdr + nNewChildNodeOfs);
        memcpy(pNewSubBlkHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
        memcpy(pNewSubBlkHdr->aFirstChar +pOldSubBlkHdr->nStrLen+1, pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
        pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
        pNewSubBlkHdr->nNumEntries = pOldChildHdr->nNumEntries;
        s32 nNumEntries = (s32)pOldChildHdr->nNumEntries;
        u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pOldChildHdr, nOldChildLog2Size);
        u8  *pSrcEnd   = (u8*)pOldChildHdr + (1<<nOldChildLog2Size);
        u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pNewSubBlkHdr, nNewNodeLog2Size);
        memcpy(pDstStart, pSrcStart, (u32)(pSrcEnd - (u8*)pSrcStart));
        memcpy(pBlkHdr, pTempBlkHdr, pDi->tba.pAreaInfo->nBlkSize);
        DirUnpinAndFreeBlk(pDi->tba.pAreaInfo, nTempBlkAdr, pDi->tba.pTransInfo);
        return nNewChildNodeOfs;
    }
    DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewChildNodeOfs);
    memcpy(pNewSubBlkHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
    memcpy(pNewSubBlkHdr->aFirstChar +pOldSubBlkHdr->nStrLen+1, pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
    pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
    pNewSubBlkHdr->nNumEntries = pOldChildHdr->nNumEntries;
    s32 nNumEntries = (s32)pOldChildHdr->nNumEntries;
    u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pOldChildHdr, nOldChildLog2Size);
    u8  *pSrcEnd   = (u8*)pOldChildHdr + (1<<nOldChildLog2Size);
    u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(nNumEntries, pNewSubBlkHdr, nNewNodeLog2Size);
    memcpy(pDstStart, pSrcStart, (u32)(pSrcEnd - (u8*)pSrcStart));
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldChildOfs, nOldChildLog2Size);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewChildNodeOfs;
}

static
u16 DirLeafMergeSubBlkWithItsSingleChild(DirItr *pDi, DirBlkHdr *pBlkHdr, u16 nOldOfs) {
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    // A non-leaf node of the radix tree is not allowed to end up with exactly 1 child. This function can be used to merge it with the child.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    u32 nOldLog2Size = DirLeafGetSubBlkLog2Size(pOldSubBlkHdr);
    u16 *pOldNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nOldLog2Size));
    u32 nOldChildOfs = pOldNodeOfsTblEnd[-1];
    dbg(SbaCheckOfs(nOldChildOfs));
    DirSubBlkHdr *pOldChildHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldChildOfs);
    u32 nOldChildLog2Size = DirLeafGetSubBlkLog2Size(pOldChildHdr);
    u32 nNewStrLen = (u32)(pOldSubBlkHdr->nStrLen + pOldChildHdr->nStrLen + 1);
    u32 nNewNodeLog2Size = DirLeafGetSubBlkLog2Size_2(nNewStrLen, pOldChildHdr->nNumEntries);
    if (nOldChildLog2Size == nNewNodeLog2Size) {
        memmove(&pOldChildHdr->aFirstChar[pOldSubBlkHdr->nStrLen+1], pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
        memcpy(pOldChildHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
        pOldChildHdr->nStrLen = (u8)nNewStrLen;
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
        return (u16)nOldChildOfs;
    }
    // The merged node will require more space than the original, so it must be reallocated
    u32 nNewChildNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewNodeLog2Size);
    if (nNewChildNodeOfs == 0) {
        // Could not allocate a larger block
        if (pDi == 0) {
            return 0;
        }
        // This section was added for the delete function - compact leaf and update node stack
        WFSBlkAdr nTempBlkAdr;
        DirBlkHdr *pTempBlkHdr;
        WFSKrnReturnOnError(DirLeafCompact(pDi, nOldOfs, nNewNodeLog2Size, nOldChildOfs, &nTempBlkAdr, &pTempBlkHdr));
        nNewChildNodeOfs = pDi->dni.pDns->nSubBlkOfs;
        DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pTempBlkHdr + nNewChildNodeOfs);
        memcpy(pNewSubBlkHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
        memcpy(pNewSubBlkHdr->aFirstChar +pOldSubBlkHdr->nStrLen+1, pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
        pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
        pNewSubBlkHdr->nNumEntries = pOldChildHdr->nNumEntries;
        s32 nNumEntries = (s32)pOldChildHdr->nNumEntries;
        u16 *pSrcStart = DirLeafGetIntraBlkOfsPtr_l2s(nNumEntries, pOldChildHdr, nOldChildLog2Size);
        u8  *pSrcEnd   = (u8*)pOldChildHdr + (1<<nOldChildLog2Size);
        u16 *pDstStart = DirLeafGetIntraBlkOfsPtr_l2s(nNumEntries, pNewSubBlkHdr, nNewNodeLog2Size);
        memcpy(pDstStart, pSrcStart, (u32)(pSrcEnd - (u8*)pSrcStart));
        memcpy(pBlkHdr, pTempBlkHdr, pDi->tba.pAreaInfo->nBlkSize);
        DirUnpinAndFreeBlk(pDi->tba.pAreaInfo, nTempBlkAdr, pDi->tba.pTransInfo);
        return nNewChildNodeOfs;
    }
    DirSubBlkHdr *pNewSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewChildNodeOfs);
    memcpy(pNewSubBlkHdr->aFirstChar, pOldSubBlkHdr->aFirstChar, (u32)(pOldSubBlkHdr->nStrLen+1));
    memcpy(pNewSubBlkHdr->aFirstChar +pOldSubBlkHdr->nStrLen+1, pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
    pNewSubBlkHdr->nStrLen = (u8)nNewStrLen;
    pNewSubBlkHdr->nNumEntries = pOldChildHdr->nNumEntries;
    u16 *pNewChildOfsTblEnd = (u16*)((u8*)pNewSubBlkHdr + (1<<nNewNodeLog2Size));
    u16 *pOldChildOfsTblEnd = (u16*)((u8*)pOldChildHdr + (1<<nOldChildLog2Size));
    s32 nNumEntries = (s32)pOldChildHdr->nNumEntries;
    memcpy(&pNewChildOfsTblEnd[-nNumEntries], &pOldChildOfsTblEnd[-nNumEntries], sizeof(u16) * nNumEntries);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldChildOfs, nOldChildLog2Size);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewChildNodeOfs;
}

static
u16 DirNodeDeleteSubBlkTail(DirBlkHdr *pBlkHdr, u16 nOldOfs, u32 nNewNumEntries) {
    // Delete the latter portion of the input node so that it contains nNewNumEntries
    // This function will reallocate the node if its size changes, or merge the node with its child
    // if it would otherwise become an illegal non-terminated node with a single child.
    // It returns the offset to the resultant node.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    if (nNewNumEntries==1) {
        utf8 *pOldChoiceChar = &pOldSubBlkHdr->aFirstChar[pOldSubBlkHdr->nStrLen];
        if (*pOldChoiceChar != 0) {
            // This node would end up as a non-Node node with exactly 1 choice, which is illegal.
            // It must be merged with its child.
            return DirNodeMergeSubBlkWithItsSingleChild(0, pBlkHdr, nOldOfs);
        }
    }
    u32 nOldLog2Size = DirNodeGetSubBlkLog2Size(pOldSubBlkHdr);
    u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr_l2s((s32)nNewNumEntries, pOldSubBlkHdr, nOldLog2Size);
    pOldSubBlkHdr->nNumEntries = (u8)nNewNumEntries;
    u32 nNewLog2Size = DirNodeGetSubBlkLog2Size(pOldSubBlkHdr);
    if (nNewLog2Size == nOldLog2Size) {
        // If the block size does not change there is no need to do anything further.
        // (because nNumEntries has already been updated, and we are removing entries from the end of the list of choices).
        return nOldOfs;
    }
    // The deletion forces the node to change size. Try to allocate a free node of the new size.
    u32 nNewNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewLog2Size);
    //u16 *pOldNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nOldLog2Size));
    if (nNewNodeOfs == 0) {
        // ToDo: This is an untested case
        // allocation failed, so resize in place
        // shift/copy the end of the string and the following choices backwards over the start of the string
        u16 *pNewNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nNewLog2Size));
        u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_ote((s32)nNewNumEntries, pOldSubBlkHdr, pNewNodeOfsTblEnd);
        memmove(pDstStart, pSrcStart, (u32)((u8*)pNewNodeOfsTblEnd - (u8*)pDstStart));
        return nOldOfs;
    }
    // copy the end of the node to the new node
    DirSubBlkHdr *pNewNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewNodeOfs);
    u16 *pNewNodeOfsTblEnd = (u16*)((u8*)pNewNodeHdr + (1<<nNewLog2Size));
    u32 nCopySize = pOldSubBlkHdr->nStrLen + nNewNumEntries + sizeof(DirSubBlkHdr)-1;
    memcpy(pNewNodeHdr, pOldSubBlkHdr, nCopySize);
    u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_ote((s32)nNewNumEntries, pNewNodeHdr, pNewNodeOfsTblEnd);
    memcpy(pDstStart, pSrcStart, (u32)((u8*)pNewNodeOfsTblEnd - (u8*)pDstStart));
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewNodeOfs;
}

static
u16 DirLeafDeleteSubBlkTail(DirBlkHdr *pBlkHdr, u16 nOldOfs, u32 nNewNumEntries) {
    // Delete the latter portion of the input node so that it contains nNewNumEntries
    // This function will reallocate the node if its size changes, or merge the node with its child
    // if it would otherwise become an illegal non-terminated node with a single child.
    // It returns the offset to the resultant node.
    DirSubBlkHdr *pOldSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOldOfs);
    if (nNewNumEntries==1) {
        utf8 *pOldChoiceChar = &pOldSubBlkHdr->aFirstChar[pOldSubBlkHdr->nStrLen];
        if (*pOldChoiceChar != 0) {
            // This node would end up as a non-leaf node with exactly 1 choice, which is illegal.
            // It must be merged with its child.
            return DirLeafMergeSubBlkWithItsSingleChild(0, pBlkHdr, nOldOfs);
        }
    }
    u32 nOldLog2Size = DirLeafGetSubBlkLog2Size(pOldSubBlkHdr);
    pOldSubBlkHdr->nNumEntries = (u8)nNewNumEntries;
    u32 nNewLog2Size = DirLeafGetSubBlkLog2Size(pOldSubBlkHdr);
    if (nNewLog2Size == nOldLog2Size) {
        // If the block size does not change there is no need to do anything further.
        // (because nNumEntries has already been updated, and we are removing entries from the end of the list of choices).
        return (u16)nOldOfs;
    }
    // The deletion forces the node to change size. Try to allocate a free node of the new size.
    u32 nNewNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewLog2Size);
    u16 *pOldNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nOldLog2Size));
    if (nNewNodeOfs == 0) {
        // allocation failed, so resize in place
        // shift/copy the end of the string and the following choices backwards over the start of the string
        u16 *pNewNodeOfsTblEnd = (u16*)((u8*)pOldSubBlkHdr + (1<<nNewLog2Size));
        memmove(&pNewNodeOfsTblEnd[-(s32)nNewNumEntries], &pOldNodeOfsTblEnd[-(s32)nNewNumEntries], sizeof(u16) * nNewNumEntries);
        return (u16)nOldOfs;
    }
    // copy the end of the node to the new node
    DirSubBlkHdr *pNewNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewNodeOfs);
    u16 *pNewNodeOfsTblEnd = (u16*)((u8*)pNewNodeHdr + (1<<nNewLog2Size));
    u32 nCopySize = pOldSubBlkHdr->nStrLen + nNewNumEntries + sizeof(DirSubBlkHdr)-1;
    memcpy(pNewNodeHdr, pOldSubBlkHdr, nCopySize);
    memcpy(&pNewNodeOfsTblEnd[-(s32)nNewNumEntries], &pOldNodeOfsTblEnd[-(s32)nNewNumEntries], sizeof(u16) * nNewNumEntries);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewNodeOfs;
}

static
u16 DirNodeCopySubBlkTail(DirBlkHdr *pSrcBlkHdr, u16 nSrcOfs, DirBlkHdr *pDstBlkHdr, u32 nNewNumEntries, u16 nDstChildOfs, bool bCopyPtrTbl) {
    // This function copies the last nNewNumEntries of a node, to a new node. nNewNumEntries must be >= 1.
    // pSrcBlkHdr is the block containing the source node. nSrcOfs is the offset from the start of the block.
    // The new node is created in pDstBlkHdr, which can be the same or different from pSrcBlkHdr.
    // This function returns the offset of the new node within pDstBlkHdr, or 0 if it could not be created.
    // If the nNewNumEntries == 1, and the sub block is non-leaf, it will be merged with the node at offset nDstChildOfs.
    DirSubBlkHdr *pSrcSubBlkHdr = (DirSubBlkHdr *)((u8*)pSrcBlkHdr + nSrcOfs);
    if (nNewNumEntries==1) {
        utf8 *pSrcChoiceChar = &pSrcSubBlkHdr->aFirstChar[pSrcSubBlkHdr->nStrLen + pSrcSubBlkHdr->nNumEntries - 1];
        if (*pSrcChoiceChar != 0) {
            // This node would end up as a non-leaf node with exactly 1 choice, which is illegal.
            // It must be merged with its child.
            return DirNodePrependStringToSubBlk(pDstBlkHdr, nDstChildOfs, pSrcSubBlkHdr->aFirstChar, pSrcSubBlkHdr->nStrLen, *pSrcChoiceChar);
        }
    }
    u32 nDstLog2Size = DirNodeGetSubBlkLog2Size_3(pSrcSubBlkHdr->nStrLen, nNewNumEntries, 1);
    u32 nDstNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pDstBlkHdr, nDstLog2Size);
    if (nDstNodeOfs == 0) {
        return 0;
    }
    DirSubBlkHdr *pDstSubBlkHdr = (DirSubBlkHdr *)((u8*)pDstBlkHdr + nDstNodeOfs);
    u32 nCopySize = pSrcSubBlkHdr->nStrLen + sizeof(DirSubBlkHdr)-1;
    memcpy(pDstSubBlkHdr, pSrcSubBlkHdr, nCopySize);
    u32 nEntryIdx = pSrcSubBlkHdr->nNumEntries - nNewNumEntries;
    memcpy((u8*)pDstSubBlkHdr + nCopySize, (u8*)pSrcSubBlkHdr + nCopySize + nEntryIdx, nNewNumEntries);
    pDstSubBlkHdr->nNumEntries = (u8)nNewNumEntries;
    u16 *pDstOfsTblEnd = (u16*)((u8*)pDstSubBlkHdr + (1<<nDstLog2Size));
    if (bCopyPtrTbl) {
        u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr(pSrcSubBlkHdr->nNumEntries, pSrcSubBlkHdr);
        memcpy(&pDstOfsTblEnd[-(s32)nNewNumEntries], pSrcStart, sizeof(u16) * nNewNumEntries);
    }
    pDstOfsTblEnd[-1] = nDstChildOfs;
    return (u16)nDstNodeOfs;
}

static
u16 DirLeafCopySubBlkTail(DirBlkHdr *pSrcBlkHdr, u16 nSrcOfs, DirBlkHdr *pDstBlkHdr, u32 nNewNumEntries, u16 nDstChildOfs, bool bCopyPtrTbl) {
    // This function copies the last nNewNumEntries of a node, to a new node. nNewNumEntries must be >= 1.
    // pSrcBlkHdr is the block containing the source node. nSrcOfs is the offset from the start of the block.
    // The new node is created in pDstBlkHdr, which can be the same or different from pSrcBlkHdr.
    // This function returns the offset of the new node within pDstBlkHdr, or 0 if it could not be created.
    // If the nNewNumEntries == 1, and the node is non-leaf, it will be merged with the node at offset nDstChildOfs.
    DirSubBlkHdr *pSrcSubBlkHdr = (DirSubBlkHdr *)((u8*)pSrcBlkHdr + nSrcOfs);
    if (nNewNumEntries==1) {
        utf8 *pSrcChoiceChar = &pSrcSubBlkHdr->aFirstChar[pSrcSubBlkHdr->nStrLen + pSrcSubBlkHdr->nNumEntries - 1];
        if (*pSrcChoiceChar != 0) {
            // This node would end up as a non-leaf node with exactly 1 choice, which is illegal.
            // It must be merged with its child.
            return DirLeafPrependStringToSubBlk(pDstBlkHdr, nDstChildOfs, pSrcSubBlkHdr->aFirstChar, pSrcSubBlkHdr->nStrLen, *pSrcChoiceChar);
        }
    }
    u32 nDstLog2Size = DirLeafGetSubBlkLog2Size_2(pSrcSubBlkHdr->nStrLen, nNewNumEntries);
    u32 nDstNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pDstBlkHdr, nDstLog2Size);
    if (nDstNodeOfs == 0) {
        return 0;
    }
    DirSubBlkHdr *pDstSubBlkHdr = (DirSubBlkHdr *)((u8*)pDstBlkHdr + nDstNodeOfs);
    u32 nCopySize = pSrcSubBlkHdr->nStrLen + sizeof(DirSubBlkHdr)-1;
    memcpy(pDstSubBlkHdr, pSrcSubBlkHdr, nCopySize);
    u32 nEntryIdx = pSrcSubBlkHdr->nNumEntries - nNewNumEntries;
    memcpy((u8*)pDstSubBlkHdr + nCopySize, (u8*)pSrcSubBlkHdr + nCopySize + nEntryIdx, nNewNumEntries);
    pDstSubBlkHdr->nNumEntries = (u8)nNewNumEntries;
    u16 *pDstOfsTblEnd = (u16*)((u8*)pDstSubBlkHdr + (1<<nDstLog2Size));
    if (bCopyPtrTbl) {
        u16 *pSrcStart = DirLeafGetIntraBlkOfsPtr(pSrcSubBlkHdr->nNumEntries, pSrcSubBlkHdr);
        memcpy(&pDstOfsTblEnd[-(s32)nNewNumEntries], pSrcStart, sizeof(u16) * nNewNumEntries);
    }
    pDstOfsTblEnd[-1] = nDstChildOfs;
    return (u16)nDstNodeOfs;
}

u16 DirNodeDeleteSubBlkHead(DirBlkHdr *pBlkHdr, u16 nOfs, u32 nEntryIdx) {
    // This function deletes entries from the start of a sub block leaving  the last nNewNumEntries.
    // pBlkHdr is the block containing the sub block. nOldOfs is the offset from the start of the block.
    // This function returns the offset of the new node within pBlkHdr, or 0 if it could not be created.
    // If the nNewNumEntries == 1, and the sub block is non-leaf, the node will be merged with its single child.
    DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
    u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    //u16 *pOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    u16 nNewOfs;
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)pSubBlkHdr + (1<<nLog2Size));
    u16 *pSrcStart = DirNodeGetIntraBlkOfsPtr_ote(pSubBlkHdr->nNumEntries, pSubBlkHdr, pExistingNodeOfsTblEnd);
    u16 *pSrcEnd = DirNodeGetIntraBlkOfsPtr_ote(nEntryIdx, pSubBlkHdr, pExistingNodeOfsTblEnd);
    utf8 *pSrc = &pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen + nEntryIdx];
    memmove(pSrc-1, pSrc, pSubBlkHdr->nNumEntries - nEntryIdx);
    --pSubBlkHdr->nNumEntries;
    u16 nNewLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
    if (nNewLog2Size != nLog2Size) {
        nNewOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewLog2Size);
        if (nNewOfs == 0) {
            // Could not allocate a new block, so resize in place - ToDo: Test this case more
            u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(pSubBlkHdr->nNumEntries, pSubBlkHdr, nNewLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memmove(pDstStart, pSrcStart, nCopySize2);
            SbaReduceSubBlkSize((WFSSubBlkAllocHdr*)pBlkHdr, nOfs, nLog2Size, nNewLog2Size);
            nNewOfs = nOfs;
        } else {
            // Delete offset pointer while copying to a new sub block
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewOfs);
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + pSubBlkHdr->nStrLen + pSubBlkHdr->nNumEntries;
            memcpy(pUpdatedNodeHdr, pSubBlkHdr, nCopySize1);
            u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(pSubBlkHdr->nNumEntries, pUpdatedNodeHdr, nNewLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memcpy(pDstStart, pSrcStart, nCopySize2);
            // Update parent to point to the new node. Parent may be the nRootOfs of the block
            SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, nOfs, nLog2Size);
            pSubBlkHdr = pUpdatedNodeHdr;
        }
    } else {
        // Delete entry from existing node
        u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(pSubBlkHdr->nNumEntries, pSubBlkHdr, nNewLog2Size);
        u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
        memmove(pDstStart, pSrcStart, nCopySize2);
        nNewOfs = nOfs;
    }
    if ((pSubBlkHdr->nNumEntries==1) && (pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen] != 0)) {
        utf8 *pChoiceChar = &pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen];
        if (*pChoiceChar != 0) {
            // This node would end up as a non-leaf node with exactly 1 choice, which is illegal.
            // It must be merged with its child.
            u16 nNewChildOfs = DirNodeGetIntraBlkOfs_l2s(pSubBlkHdr->nNumEntries, pSubBlkHdr, nNewLog2Size);
            nOfs = nNewOfs;
            nNewOfs = DirNodePrependStringToSubBlk(pBlkHdr, nNewChildOfs, pSubBlkHdr->aFirstChar, pSubBlkHdr->nStrLen, *pChoiceChar);
            SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, nOfs, nNewLog2Size);
        }
    }
    return nNewOfs;
}

WFSKrnResult DirNodeSubBlkSplitPrefixString(DirBlkHdr *pBlkHdr, DirSubBlkHdr *pSubBlkHdr, u32 nOfs, u32 nNewStrLen, utf8 nNewChoiceChar) {
    // This function splits a sub block into two pieces by adding a choice at some point along the prefix string
    // ToDo: Implement this function - to remove a common case from insert and delete to node blocks.
    return WFSKRN_RESULT_OK;
}

static
WFSKrnResult DirNodeInsertEntry(WFSFileName *pName, DirItr *pDi, WFSBlkAdr *aBlkAdr, u32 *pFirstBlockToFree) {
#if _DEBUG_BREAK_POINT
    static int c;++c;if(c>=3){
        int a;a=0;
    }
#endif
    // This function adds a splitter to a parent node when a leaf is split.
    // pDi->tba.nBlkAdr contains the block address of the parent node to which the splitter entry is being inserted.
    // aBlkAdr contains the block addresses of: 0..Left Child, 1..Right Child, 2+..Reserved incase further splits needed
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    if (pDi->nBlkDepth) {
        --pDi->nBlkDepth;
    }
//Retry:
    WFSBlkAdr  nBlkAdr = pDi->tba.nBlkAdr;
    DirBlkHdr *pBlkHdr = pDi->pBlkHdr;
    u32 nStrLen = pName->nLen;
    if (pBlkHdr->nNumRecs==0) {
        // The parent block is currently enpty. Add the first record.
        // Create root node with minimum possible key (NULL key), and add pointer to left node
        u32 nRootNodeLog2Size = DirNodeGetSubBlkLog2Size_3(0, 2, 0);
        pBlkHdr->nRootOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nRootNodeLog2Size);
        DirSubBlkHdr *pRootSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pBlkHdr->nRootOfs);
        pRootSubBlkHdr->nStrLen = 0;
        pRootSubBlkHdr->nNumEntries = 2;
        pRootSubBlkHdr->aFirstChar[0]=0;
        pRootSubBlkHdr->aFirstChar[1]=pName->sStr[0];
        u16 *pRootOfsTblEnd = (u16 *)((u8*)pRootSubBlkHdr + (1<<nRootNodeLog2Size));
        DirNodeSetBlkAdr_ote(aBlkAdr[0], pRootSubBlkHdr, pRootOfsTblEnd);
        // Create second node
        nStrLen--;
        u32 nNodeLog2Size = DirNodeGetSubBlkLog2Size_3(nStrLen, 1, 0);
        u16 nNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNodeLog2Size);
        DirNodeSetIntraBlkOfs_ote(2, nNodeOfs, pRootSubBlkHdr, pRootOfsTblEnd);
        DirSubBlkHdr *pNodeSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNodeOfs);
        pNodeSubBlkHdr->nStrLen = (u8)nStrLen;
        pNodeSubBlkHdr->nNumEntries = 1;
        memcpy(pNodeSubBlkHdr->aFirstChar, pName->sStr+1, nStrLen+1);
        DirNodeSetBlkAdr_l2s(aBlkAdr[1], pNodeSubBlkHdr, nNodeLog2Size);
        pBlkHdr->nNumRecs = 2;
        ++pDi->nBlkDepth;
        TransUnpinBlk(&pDi->tba);
        return WFSKRN_RESULT_OK;
    }

    // Insert the new name into the parent node block.
    // First determine the insertion point.
    DirNodeItr dni = pDi->dni;
    dbg(u32 nChildBlkAdr = dni.pDns->nBlkAdr);
    //if (pDi->tba.nBlkAdr != dni.pDns->nBlkAdr) {
    while (pDi->tba.nBlkAdr != dni.pDns->nBlkAdr) {
        //do {
            DirPopNodeStackEntry(&dni);
        //} while(pDi->tba.nBlkAdr != dni.pDns->nBlkAdr);
    }
    dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + dni.pDns->nSubBlkOfs);
    dbg(if (DirNodeGetBlkAdr(dni.pSubBlkHdr) != nChildBlkAdr) {
        WFSKrnOutputErrorStr("BlkAdr doesn't match child!\n");
    })
    DirNodeSetBlkAdr(aBlkAdr[0], dni.pSubBlkHdr);  // Update the address of the left hand child block
    while(dni.nDnsDepth > 2) {
        if (dni.pDns->pParent->nBlkAdr != pDi->tba.nBlkAdr) {
            break;
        }
        DirPopNodeStackEntry(&dni);
    }

    dni.pStrPtr = pName->sStr;
    u32 nSubBlkStrLen, nStrIdx=0;
    //utf8 *pSubBlkStrPtr,
    utf8 *pStrEnd = pName->sStr + pName->nLen;
    utf8 cSearchChar;
    WFSKrnResult nResult;

    dni.pOfs = &pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(*dni.pOfs));
    dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *dni.pOfs);
    dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
    while(1) {
        nSubBlkStrLen = dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen + 1;
        dni.pNodeStrPtr = dni.pSubBlkHdr->aFirstChar;
        if ((pStrEnd - dni.pStrPtr) < nSubBlkStrLen) {
            while(*dni.pStrPtr == *dni.pNodeStrPtr) {
                ++dni.pStrPtr; ++dni.pNodeStrPtr;
            }
            nSubBlkStrLen -= (dni.pNodeStrPtr - dni.pSubBlkHdr->aFirstChar);
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        while((nSubBlkStrLen) && (*dni.pStrPtr == *dni.pNodeStrPtr)) {
            ++dni.pStrPtr; ++dni.pNodeStrPtr; --nSubBlkStrLen;
        }
        if (nSubBlkStrLen) {
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        // matched all the characters from the prefix. dni.pNodeStrPtr should now be pointing to the array of choices following the prefix.
        cSearchChar = *dni.pStrPtr;
        if (!DirBinarySplitSearchChoices(dni.pNodeStrPtr, dni.pSubBlkHdr->nNumEntries, cSearchChar, &dni.nEntryIdx)) {
            nResult = WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND;
            break;
        }
        //if (dni.pSubBlkHdr->nNumEntries > 1) {
        if(dni.pDns->nBlkAdr != nBlkAdr || dni.pDns->nSubBlkOfs != *dni.pOfs){
            WFSKrnReturnOnError(DirPushNodeStackEntry(&dni, nBlkAdr));
        } else {
            dni.pDns->nEntryIdx  = (u8)dni.nEntryIdx;
        }
        dni.pDns->nStrIdx = nStrIdx;
        //}
        // Found the search character.
        //++nStrIdx;
        if (cSearchChar==0) {
            // We matched a 0 .. which means we have found the filename
            nResult = WFSKRN_RESULT_DIR_ENTRY_FOUND;
            break;
        }
        dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s((s32)dni.nEntryIdx, dni.pSubBlkHdr, dni.nLog2Size);
        dbg(SbaCheckOfs(*dni.pOfs));
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *dni.pOfs);
        dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
        ++dni.pStrPtr;
    }

Retry:;
    u32 nNumMatching = (u32)(dni.pNodeStrPtr - dni.pSubBlkHdr->aFirstChar);
    u32 nNodeNumLeft = dni.pSubBlkHdr->nStrLen - nNumMatching;
    u32 nSuffixLen = (u32)(pName->nLen - (dni.pStrPtr - pName->sStr));
    u16 nNameSubBlkLog2Size = 0;
    u16 nNameSubBlkOfs = 0;
    u16 nChoiceNodeSubBlkLog2Size;
    u16 nChoiceNodeSubBlkOfs;
    u16 *pNodeOfsTblEnd;
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));

    // ToDo: Change from name to splitter
    if (nSuffixLen) {
        // allocate a sub block for the new name entry
        nNameSubBlkLog2Size = (u16)DirNodeGetSubBlkLog2Size_3(nSuffixLen-1, 1, 0);
        nNameSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkLog2Size);
        if (nNameSubBlkOfs == 0) {
NameNotAllocated:
            pDi->dni = dni;
            *pFirstBlockToFree += 2;
            WFSKrnReturnOnError(DirNodeBlkSplit(pDi, &aBlkAdr[2], pFirstBlockToFree));
            while(pDi->dni.nDnsDepth > 2) {
                DirPopNodeStackEntry(&pDi->dni);
            }
            pDi->tba.nBlkAdr = pDi->dni.pDns->nBlkAdr;
            DirItrCloseNoUnpin(pDi);
            u32 nFindMaxDepth = pDi->nBlkDepth;
            pDi->nBlkDepth = 0;
            DirFind(pName, pDi, nFindMaxDepth, DIR_FIND_MODE_NORMAL);
            nBlkAdr = pDi->tba.nBlkAdr;
            WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_RW_BLK, &pBlkHdr) );
            dni = pDi->dni;
            dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
            dbg(if (*dni.pOfs != ((u8*)dni.pSubBlkHdr - (u8*)pDi->pBlkHdr)) {
                WFSKrnOutputErrorStr("Ofs mismatch error!");
            })
            goto Retry;
        }
        DirSubBlkHdr *pNameSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNameSubBlkOfs);
        pNameSubBlkHdr->nNumEntries = 1;
        pNameSubBlkHdr->nStrLen = (u8)(nSuffixLen-1);
        memcpy(pNameSubBlkHdr->aFirstChar, &dni.pStrPtr[1], nSuffixLen);  // includes copying the NULL terminator
        DirNodeSetBlkAdr(aBlkAdr[1], pNameSubBlkHdr);
    } else {
        //nNameSubBlkOfs = nAttrOfs;
    }

    utf8 cNameChar = dni.pStrPtr[0];
    if (nNodeNumLeft == 0) {
        // Add a new option to the existing node
        u16 nUpdatedNodeOfs;
        u32 nUpdatedNodeLog2Size;
        if (cNameChar) {
            nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size_3(dni.pSubBlkHdr->nStrLen, (u32)(dni.pSubBlkHdr->nNumEntries+1), dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen]);
        } else {
            nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size_3(dni.pSubBlkHdr->nStrLen, (u32)(dni.pSubBlkHdr->nNumEntries+1), 0);
        }
        if (nUpdatedNodeLog2Size != dni.nLog2Size) {
            // The node needs to be copied to a larger sub block
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            if (nUpdatedNodeOfs == 0) {
                // Allocation failed, so undo any changes so far, split the block, and try again
                if (nSuffixLen) {
                    // Undo the allocation of the name sub block
                    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkOfs, nNameSubBlkLog2Size);
                }
                goto NameNotAllocated;
            }
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            u16 *pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
            // Copy the old choices to the new sub block while inserting the new choice
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + dni.pSubBlkHdr->nStrLen + dni.nEntryIdx;
            memcpy(pUpdatedNodeHdr, dni.pSubBlkHdr, nCopySize1);
            u8 *pDst = &((u8*)pUpdatedNodeHdr)[nCopySize1];
            *pDst++ = (u8)dni.pStrPtr[0];
            s32 nCopySize2 = (s32)(dni.pSubBlkHdr->nNumEntries - dni.nEntryIdx);
            memcpy(pDst, &((u8*)dni.pSubBlkHdr)[nCopySize1], (u32)nCopySize2);
            ++pUpdatedNodeHdr->nNumEntries;
            // Copy the table of child offset pointers from the old block to the new block while inserting the new offset
            u16 *pDstStart = DirNodeGetOfsTblStartPtr_ote(pUpdatedNodeHdr, pUpdatedNodeOfsTblEnd);
            u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            u16 *pSrcEnd   = DirNodeGetIntraBlkOfsPtr_ote((s32)dni.nEntryIdx, dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            memcpy(pDstStart, pSrcStart, (u32)((u8*)pSrcEnd-(u8*)pSrcStart));
            pDstStart += (pSrcEnd-pSrcStart);
            if (cNameChar) {
                *pDstStart++ = nNameSubBlkOfs;
                memcpy(pDstStart, pSrcEnd, (u32)((u8*)pExistingNodeOfsTblEnd-(u8*)pSrcEnd));
            } else {
                // case where new choice is 0
                *((u32*)pDstStart) = aBlkAdr[1];
            }
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, *dni.pOfs, dni.nLog2Size);
            *dni.pOfs = nUpdatedNodeOfs; // Update the parent pointer
        } else {
            // The new option can be added without changing the size of the node
            utf8 *pCharInsertPoint = dni.pSubBlkHdr->aFirstChar + dni.pSubBlkHdr->nStrLen + dni.nEntryIdx;
            s32 nCopySize = (s32)(dni.pSubBlkHdr->nNumEntries - dni.nEntryIdx);
            u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            memmove(pCharInsertPoint+1, pCharInsertPoint, (u32)(nCopySize));
            pCharInsertPoint[0] = dni.pStrPtr[0];
            if (cNameChar) {
                u16 *pOfsInsertPoint = DirNodeGetIntraBlkOfsPtr_ote((s32)dni.nEntryIdx, dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
                memmove(pSrcStart-1, pSrcStart, (u32)((u8*)pOfsInsertPoint-(u8*)pSrcStart));
                ++dni.pSubBlkHdr->nNumEntries;
                pOfsInsertPoint[-1] = nNameSubBlkOfs;
            } else {
                memmove(pSrcStart-2, pSrcStart, (u32)((u8*)pExistingNodeOfsTblEnd-(u8*)pSrcStart));
                ((u32*)pExistingNodeOfsTblEnd)[-1] = aBlkAdr[1];
                ++dni.pSubBlkHdr->nNumEntries;
            }
        }
    } else {
        // Replace the start of the existing node with a new 2-choice node
        // and update the existing node to contain only the end of the string,
        // resizing the node if its size drops to a smaller power-of-2 allocation size.
        // (note: the resize is compulsory due to the algorithms used elsewhere).
        //
        // First, try to allocate a sub block for the new 2-choice node.
        utf8 cNodeChar = dni.pSubBlkHdr->aFirstChar[nNumMatching];
        nChoiceNodeSubBlkLog2Size = (u16)DirNodeGetSubBlkLog2Size_3(nNumMatching, 2, (s8)cNameChar);
        nChoiceNodeSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nChoiceNodeSubBlkLog2Size);
        if (nChoiceNodeSubBlkOfs == 0) {
            // Allocation failed, so undo any changes so far, split the block, and try again
            if (nSuffixLen) {
                // Undo the allocation of the name sub block
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkOfs, nNameSubBlkLog2Size);
            }
            goto NameNotAllocated;
        }
        DirSubBlkHdr *pChoiceNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nChoiceNodeSubBlkOfs);
        pNodeOfsTblEnd = (u16*)((u8*)pChoiceNodeHdr + (1<<nChoiceNodeSubBlkLog2Size));
        pChoiceNodeHdr->nNumEntries = 2;
        pChoiceNodeHdr->nStrLen = (u8)nNumMatching;
        memcpy(pChoiceNodeHdr->aFirstChar, dni.pSubBlkHdr->aFirstChar, nNumMatching);

        // Now, update the existing node
        u16 nUpdatedNodeOfs;
        u32 nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size_3(nNodeNumLeft-1, dni.pSubBlkHdr->nNumEntries, dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen]);
        DirSubBlkHdr *pUpdatedNodeHdr;
        if (nUpdatedNodeLog2Size != dni.nLog2Size) {
            // The existing node should be reduced in size.
            // First try to allocate a new smaller sub block.
            // If this is not possible, then resize in place.
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            u32 nOldNodeOfs = (u16)((u8*)dni.pSubBlkHdr - (u8*)pBlkHdr);
            if (nUpdatedNodeOfs == 0) {
                // allocation failed, so resize in place
                // shift/copy the end of the string and the following choices backwards over the start of the string
                pUpdatedNodeHdr = dni.pSubBlkHdr;
                nUpdatedNodeOfs = (u16)nOldNodeOfs;
                pUpdatedNodeHdr->nStrLen = (u8)(nNodeNumLeft-1);  // Needs to be done first so that pDstStart calculation is correct
                memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries);
                // shift over the table of pointers at the end of the sub block
                u16 *pDstStart = DirNodeGetOfsTblStartPtr_l2s(pUpdatedNodeHdr, nUpdatedNodeLog2Size);
                u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
                memmove(pDstStart, pSrcStart, (u32)((u8*)pExistingNodeOfsTblEnd-(u8*)pSrcStart));
                SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeOfs, dni.nLog2Size, nUpdatedNodeLog2Size);
            } else {
                // copy the end of the node to the new node, and free the old node
                pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
                *dni.pOfs = nUpdatedNodeOfs; // Update the parent pointer
                pUpdatedNodeHdr->nNumEntries = dni.pSubBlkHdr->nNumEntries;
                pUpdatedNodeHdr->nStrLen = (u8)(nNodeNumLeft-1);  // Needs to be done first so that pDstStart calculation is correct
                memcpy(pUpdatedNodeHdr->aFirstChar, &dni.pSubBlkHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
                u16 *pDstStart = DirNodeGetOfsTblStartPtr_l2s(pUpdatedNodeHdr, nUpdatedNodeLog2Size);
                u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
                memcpy(pDstStart, pSrcStart, (u32)((u8*)pExistingNodeOfsTblEnd-(u8*)pSrcStart));
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldNodeOfs, dni.nLog2Size);
            }
        } else {
            // the sub block remains the same size
            pUpdatedNodeHdr = dni.pSubBlkHdr;
            nUpdatedNodeOfs = (u16)((u8*)pUpdatedNodeHdr - (u8*)pBlkHdr);
            dbg(SbaCheckOfs(nUpdatedNodeOfs));
            // Remove characters from the start of the string
            memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
        }
        pUpdatedNodeHdr->nStrLen = (u8)(nNodeNumLeft-1);
        *dni.pOfs = nChoiceNodeSubBlkOfs;
        dbg(SbaCheckOfs(*dni.pOfs));

        if ((unsigned)cNodeChar < (unsigned)cNameChar) {
            pChoiceNodeHdr->aFirstChar[nNumMatching] = cNodeChar;
            pChoiceNodeHdr->aFirstChar[nNumMatching+1] = cNameChar;
            pNodeOfsTblEnd[-1] = nUpdatedNodeOfs;
            pNodeOfsTblEnd[-2] = nNameSubBlkOfs;
        } else {
            pChoiceNodeHdr->aFirstChar[nNumMatching] = cNameChar;
            pChoiceNodeHdr->aFirstChar[nNumMatching+1] = cNodeChar;
            if (cNameChar==0) {
                ((u32*)pNodeOfsTblEnd)[-1] = aBlkAdr[1];
                pNodeOfsTblEnd[-3] = nUpdatedNodeOfs;
            } else {
                pNodeOfsTblEnd[-1] = nNameSubBlkOfs;
                pNodeOfsTblEnd[-2] = nUpdatedNodeOfs;
            }
        }
    }
    ++pDi->nBlkDepth;
    ++pBlkHdr->nNumRecs;
    pDi->dni = dni;     // To ensure dir node stack gets freed properly later
    dbg(DirNodeCheckBlk(pDi));
    TransUnpinBlk(&pDi->tba);
    return WFSKRN_RESULT_OK;
}

WFSKrnResult DirNodeRenameFirstRec(DirBlkHdr *pBlkHdr) {
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    //DirItr di;
    DirNodeItr dni;
    dni.nDnsDepth = 0;
    WFSKrnReturnOnError(DirNodeStackGetEntry(&dni));
    DirNodeStack **ppDnsRootParent = &dni.pDns->pParent;
    u32 nOfs = pBlkHdr->nRootOfs;
    u16 *pOfsTblEnd;
    WFSBlkAdr nChildBlkAdr;
    while(1) {
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
        dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
        pOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));
        WFSKrnReturnOnError(DirNodeStackGetEntry(&dni));
        dni.pDns->nSubBlkOfs = nOfs;
        if (dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen] == 0) {
            // We have found the end of the splitter. Record the child block for later.
            nChildBlkAdr = DirNodeGetBlkAdr_ote(dni.pSubBlkHdr, pOfsTblEnd);
            break;
        }
        nOfs = DirNodeGetIntraBlkOfs_ote(1, dni.pSubBlkHdr, pOfsTblEnd);
    }
    if (dni.nDnsDepth <= 1) {
        WFSKrnOutputErrorStr("Error! First rec already NULL");
    }
    if (dni.pSubBlkHdr->nNumEntries == 1) {
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOfs, dni.nLog2Size);
        DirPopNodeStackEntry(&dni);
    }
    u16 nDstChildOfs = 0;
    if (dni.nDnsDepth>1) {
        nOfs = dni.pDns->nSubBlkOfs;
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
        dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
        nDstChildOfs = DirNodeDeleteSubBlkHead(pBlkHdr, nOfs, 1);
    }
    if (dni.nDnsDepth>2) {
        DirPopNodeStackEntry(&dni);
        nOfs = dni.pDns->nSubBlkOfs;
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
        dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
        pOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));
        pOfsTblEnd[-1] = nDstChildOfs;
        while(dni.nDnsDepth>2) {
            DirPopNodeStackEntry(&dni);
        }
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + dni.pDns->nSubBlkOfs);
    } else {
        pBlkHdr->nRootOfs = nDstChildOfs;
        dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pBlkHdr->nRootOfs);
    }
    // Insert "" element into root sub block
    u32 nUpdatedNodeLog2Size;
    dni.nLog2Size = DirNodeGetSubBlkLog2Size(dni.pSubBlkHdr);
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)dni.pSubBlkHdr + (1<<dni.nLog2Size));

    if (dni.pSubBlkHdr->nStrLen != 0) {
        // Handle the case where the root sub block has a prefix - requires a new parent sub-node to be added
        //DirNodeSubBlkSplitPrefixString(pBlkHdr, dni.pSubBlkHdr, nOfs, 1, 0);
        utf8 cChoiceChar = dni.pSubBlkHdr->aFirstChar[0];
        nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size_3(dni.pSubBlkHdr->nStrLen-1, dni.pSubBlkHdr->nNumEntries, dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen]);
        if (nUpdatedNodeLog2Size != dni.nLog2Size) {
            // allocate a new sub block
            u32 nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            if (nUpdatedNodeOfs == 0) {
                return WFSKRN_RESULT_FATAL_ERROR;
            }
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            pUpdatedNodeHdr->nStrLen = dni.pSubBlkHdr->nStrLen - 1;
            pUpdatedNodeHdr->nNumEntries = dni.pSubBlkHdr->nNumEntries;
            u32 nCopySize1 = pUpdatedNodeHdr->nStrLen + pUpdatedNodeHdr->nNumEntries;
            memcpy(pUpdatedNodeHdr->aFirstChar, &dni.pSubBlkHdr->aFirstChar[1], nCopySize1);
            u16 *pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
            u16 *pDstStart = DirNodeGetOfsTblStartPtr_ote(pUpdatedNodeHdr, pUpdatedNodeOfsTblEnd);
            u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, pBlkHdr->nRootOfs, dni.nLog2Size);
            pBlkHdr->nRootOfs = nUpdatedNodeOfs;
        } else {
            dni.pSubBlkHdr->nStrLen--;
            u32 nCopySize1 = dni.pSubBlkHdr->nStrLen + dni.pSubBlkHdr->nNumEntries;
            memmove(dni.pSubBlkHdr->aFirstChar, &dni.pSubBlkHdr->aFirstChar[1], nCopySize1);
        }

        u32 nNewRootNodeLog2Size = DirNodeGetSubBlkLog2Size_3(0, 2, 0);
        u16 nNewRootNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNewRootNodeLog2Size);
        if (nNewRootNodeOfs == 0) {
            return WFSKRN_RESULT_FATAL_ERROR;
        }
        DirSubBlkHdr *pNewRootNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNewRootNodeOfs);
        pNewRootNodeHdr->nStrLen = 0;
        pNewRootNodeHdr->nNumEntries = 2;
        pNewRootNodeHdr->aFirstChar[0] = 0;
        pNewRootNodeHdr->aFirstChar[1] = cChoiceChar;
        u16 *pNewRootOfsTblEnd = (u16*)((u8*)pNewRootNodeHdr + (1<<nNewRootNodeLog2Size));
        ((u32*)pNewRootOfsTblEnd)[-1] = nChildBlkAdr;
        pNewRootOfsTblEnd[-3] = pBlkHdr->nRootOfs;
        pBlkHdr->nRootOfs = nNewRootNodeOfs;
    } else {
        // case where there is no prefix string in the root sub block
        nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size_3(dni.pSubBlkHdr->nStrLen, dni.pSubBlkHdr->nNumEntries+1, 0);
        if (nUpdatedNodeLog2Size != dni.nLog2Size) {
            u32 nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            if (nUpdatedNodeOfs == 0) {
                return WFSKRN_RESULT_FATAL_ERROR;
            }
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + dni.pSubBlkHdr->nStrLen;
            memcpy(pUpdatedNodeHdr, dni.pSubBlkHdr, nCopySize1);
            ++pUpdatedNodeHdr->nNumEntries;
            memcpy((u8*)pUpdatedNodeHdr + nCopySize1 + 1, &dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen], dni.pSubBlkHdr->nNumEntries);
            pUpdatedNodeHdr->aFirstChar[pUpdatedNodeHdr->nStrLen] = 0;
            u16 *pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
            u16 *pDstStart = DirNodeGetOfsTblStartPtr_ote(pUpdatedNodeHdr, pUpdatedNodeOfsTblEnd);
            u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            ((u32*)pUpdatedNodeOfsTblEnd)[-1] = nChildBlkAdr;
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, pBlkHdr->nRootOfs, dni.nLog2Size);
            pBlkHdr->nRootOfs = nUpdatedNodeOfs;
        } else {
            // Insert new entry without moving sub block
            utf8 *pSrc = &dni.pSubBlkHdr->aFirstChar[dni.pSubBlkHdr->nStrLen];
            memmove(pSrc+1, pSrc, dni.pSubBlkHdr->nNumEntries);
            u16 *pSrcStart = DirNodeGetOfsTblStartPtr_ote(dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
            *pSrc = 0;
            ++dni.pSubBlkHdr->nNumEntries;
            memmove(pSrcStart-2, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            ((u32*)pExistingNodeOfsTblEnd)[-1] = nChildBlkAdr;
        }
    }
    //dbg(DirNodeCheckBlk(&di));
    DirNodeStackFree(&dni, ppDnsRootParent);
    return WFSKRN_RESULT_OK;
}

u32 nNumNodeBlkSplits=0;
WFSKrnResult DirNodeBlkSplit(DirItr *pDiSrc, WFSBlkAdr *aDstBlkAdr, u32 *pFirstBlockToFree) {
    // Split the specified node block into two pieces, at its mid point.
    // Copy the first half of the records to new block 1
    // Copy the first half of the records to new block 2
    // Delete the original block and rename new block 2 to be this block
    ++nNumNodeBlkSplits;
    WFSKrnResult result;
    DirItr diSrc=*pDiSrc, diDst;
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
    WFSFileName name;
    diSrc.dni.pStrPtr = name.sStr;
#endif
    u32 nStrIdx=0, nSubBlkStrLen;

    WFSBlkAdr nBlkAdr = diSrc.tba.nBlkAdr;
    WFSBlkAdr nParentBlkAdr = 0;
    while(diSrc.dni.nDnsDepth > 1) {
        if (diSrc.dni.pDns->nBlkAdr != nBlkAdr) {
            nParentBlkAdr = diSrc.dni.pDns->nBlkAdr;
            break;
        }
        DirPopNodeStackEntry(&diSrc.dni);
    }
    DirBlkHdr *pLeftBlkHdr;
    DirBlkHdr *pRightBlkHdr;
    result = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, aDstBlkAdr[0], WFS_MDF_NULL, diSrc.tba.pTransInfo, &pLeftBlkHdr);
    if(result < WFSKRN_RESULT_OK){
        return result;
    }
    result = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, aDstBlkAdr[1], WFS_MDF_NULL, diSrc.tba.pTransInfo, &pRightBlkHdr);
    if(result < WFSKRN_RESULT_OK){
        TransUnpinAndFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, aDstBlkAdr, diSrc.tba.pTransInfo);
        return result;
    }
    diDst.pBlkHdr = pLeftBlkHdr;
    // Initialize a stack frame for traversing the destination block
    diDst.dni.nDnsDepth = 0;
    WFSKrnReturnOnError(DirNodeStackGetEntry(&diDst.dni));
    diDst.ppDnsRootParent = &diDst.dni.pDns->pParent;

    diSrc.dni.nEntryIdx = 1;
    diSrc.dni.pOfs = &diSrc.pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(*diSrc.dni.pOfs));
    WFSKrnReturnOnError(DirPushNodeStackEntry(&diSrc.dni, nBlkAdr));
    diSrc.dni.pDns->nStrIdx = 0;    // ToDo: This value can be initialized to 0 because prefix-suppression is not implemented
    u16 nSrcNodeOfs = diSrc.dni.pDns->nSubBlkOfs; // = diSrc.pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(nSrcNodeOfs));
    utf8 *pChoiceChar;
    u32 nRecIdx=0;
    u32 nHalfNumRecs = (u32)(diSrc.pBlkHdr->nNumRecs>>1);
    diDst.dni.pOfs = &diDst.pBlkHdr->nRootOfs;
    s32 nSplitPointDnsDepth=-1;
    u32 nCopied = 1;
    u16 nLeftChildOfs = 0;
    WFSFileName parentKey;
    u32 nSrcNodeStartDepth = diSrc.dni.nDnsDepth;

    // Perform a depth-first traversal of the block, until the middle Node sub-block is found.
    while(1) {
        diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + nSrcNodeOfs);
        diSrc.dni.nLog2Size = DirNodeGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        // Copy over the whole node from the original block
        WFSKrnReturnOnError(DirNodeStackGetEntry(&diDst.dni));
        *diDst.dni.pOfs = diDst.dni.pDns->nSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, diSrc.dni.nLog2Size);
        diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + *diDst.dni.pOfs);
        memcpy(diDst.dni.pSubBlkHdr, diSrc.dni.pSubBlkHdr, (u32)(1<<diSrc.dni.nLog2Size));
        diDst.dni.pDns->nEntryIdx = 1;
        //u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        diSrc.dni.pDns->nEntryIdx = 1;
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen];
        nSubBlkStrLen = diSrc.dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen;
        diSrc.dni.pDns->nStrIdx = (u8)nStrIdx;
#if _DEBUG_DIR
        utf8 *pSubBlkStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        while(nSubBlkStrLen--) {
            *diSrc.dni.pStrPtr++ = *pSubBlkStrPtr++;
        }
#endif
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment our record count
#if _DEBUG_DIR
            *diSrc.dni.pStrPtr = 0;
            //MyOSReport("%d. %s\n", nRecIdx, name.sStr);
            ++pChoiceChar;
#endif
            ++nRecIdx;
            if (nRecIdx==nHalfNumRecs+1) {
                DirConstructEntryNameFromNodeStack(&parentKey, &diSrc);
#if _DEBUG_DIR
                if (strncmp(parentKey.sStr, name.sStr, WFS_MAX_FILE_NAME_SIZE)) {
                    WFSKrnOutputErrorStr("Error constructing entry name from stack\n");
                }
#endif
                // No Shortening for splitters // ToDo: Shortening can be implemented if we ensure the first splitter of all blocks is ""
                //parentKey.nLen -= diSrc.dni.pSubBlkHdr->nStrLen;
                //parentKey.sStr[parentKey.nLen] = 0; // This shortens the name to the minimum splitter
            }
            ++diSrc.dni.pDns->nEntryIdx;
            ++diDst.dni.pDns->nEntryIdx;
        }
        while(diSrc.dni.pDns->nEntryIdx > diSrc.dni.pSubBlkHdr->nNumEntries) {
            if (diSrc.dni.nDnsDepth == nSrcNodeStartDepth) {
                // We reached the top of the stack, so this should be the last record
                pLeftBlkHdr->nRootOfs = nLeftChildOfs;
                pRightBlkHdr->nRootOfs = diDst.dni.pDns->nSubBlkOfs;
                DirNodeStackFree(&diDst.dni, diDst.ppDnsRootParent);
                pLeftBlkHdr->nNumRecs = (u16)nHalfNumRecs;
                pRightBlkHdr->nNumRecs = (u16)(diSrc.pBlkHdr->nNumRecs - nHalfNumRecs);
                if (diSrc.pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_ROOT_BLK) {
                    nParentBlkAdr = nBlkAdr;
                    DirReinitBlk(diSrc.tba.pAreaInfo, diSrc.pBlkHdr, (WFSMetaDataFlags)WFS_MDF_DIR_ROOT_BLK);
                } else {
                    DirUnpinAndFreeBlk(diSrc.tba.pAreaInfo, nBlkAdr, diSrc.tba.pTransInfo);
                    diSrc.tba.nBlkAdr = nParentBlkAdr;
#if _CHECK_BLK_USAGE // for debug
                    if(DeviceDebugGetPseudoDetachDevice() == 0){
                        DeviceDebugSetPseudoDetachDevice(1);
                    }
                    if(DeviceDebugGetPseudoHashInconsistent() == 0){
                        DeviceDebugSetPseudoHashInconsistent(1);
                    }
#endif
                    WFSKrnReturnOnError( DirGetBlk(&diSrc, TRANS_FLAG_RW_BLK | BCACHE_FLAG_PINNED, 0) );  // must be debug
                }
                TransUnpinBlk3(diSrc.tba.pAreaInfo, aDstBlkAdr[0], diSrc.tba.pTransInfo);
                WFSKrnReturnOnError(DirNodeInsertEntry(&parentKey, &diSrc, aDstBlkAdr, pFirstBlockToFree));
                *pDiSrc=diSrc;
                WFSKrnResult result = DirNodeRenameFirstRec(pRightBlkHdr);
                TransUnpinBlk3(diSrc.tba.pAreaInfo, aDstBlkAdr[1], diSrc.tba.pTransInfo);
                return result;
            }
            u16 nDstChildOfs = diDst.dni.pDns->nSubBlkOfs;
            DirPopNodeStackEntry(&diDst.dni);
            DirPopNodeStackEntry(&diSrc.dni);
#if _DEBUG_DIR
            diSrc.dni.pStrPtr += diSrc.dni.pDns->nStrIdx - nStrIdx;
#endif
            nStrIdx = diSrc.dni.pDns->nStrIdx;
            diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + diSrc.dni.pDns->nSubBlkOfs);
            if (nSplitPointDnsDepth >= (s32)diSrc.dni.nDnsDepth) {
                // This is a node which straddles the left and right halves of the source block.
                // At this point it has already been copied in whole to the left block.
                // We need to move part of it to the right block, and remove unused entries from the left block.
                nSplitPointDnsDepth  = (s32)(diSrc.dni.nDnsDepth-1);
                u32 nRightNumEntries = (u32)(diSrc.dni.pSubBlkHdr->nNumEntries - diSrc.dni.pDns->nEntryIdx + 1);
                u32 nLeftNumEntries  = diSrc.dni.pDns->nEntryIdx - nCopied;
                u16 nLeftOfs         = diDst.dni.pDns->nSubBlkOfs;
                nDstChildOfs = diDst.dni.pDns->nSubBlkOfs = DirNodeCopySubBlkTail(pLeftBlkHdr, nLeftOfs, pRightBlkHdr, nRightNumEntries, nDstChildOfs, false);
                dbg(SbaCheckOfs(nDstChildOfs));
                if (nCopied==0) {
                    DirSubBlkHdr *pLeftSubBlkHdr = (DirSubBlkHdr *)((u8*)pLeftBlkHdr + nLeftOfs);
                    diSrc.dni.nLog2Size  = DirNodeGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
                    DirNodeSetIntraBlkOfs_l2s(diSrc.dni.pDns->nEntryIdx, nLeftChildOfs, pLeftSubBlkHdr, diSrc.dni.nLog2Size);
                }
                nLeftChildOfs  = DirNodeDeleteSubBlkTail(pLeftBlkHdr, nLeftOfs, nLeftNumEntries);
                dbg(if (nLeftChildOfs == 0) {
                    // This is not expected to happen since the newly split block should have plenty of free space
                    WFSKrnOutputErrorStr("Fatal Error!\n");
                })
                diDst.dni.pDns->nEntryIdx = 1;
                nCopied = 0;
            }
            ++diDst.dni.pDns->nEntryIdx;
            ++diSrc.dni.pDns->nEntryIdx;
        }
        if ((nRecIdx==nHalfNumRecs)&&(nSplitPointDnsDepth<0)) {
            // We have found the middle record of the Node block.
            // This code should only be executed once during the loop
            nSplitPointDnsDepth = (s32)diSrc.dni.nDnsDepth;
            diDst.pBlkHdr = pRightBlkHdr;
        }
#if _DEBUG_DIR
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen + diSrc.dni.pDns->nEntryIdx-1];
        *diSrc.dni.pStrPtr = *pChoiceChar;
#endif
        ++nStrIdx;
        diSrc.dni.nLog2Size = DirNodeGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        {
            // Calculate the correct address for diDst.dni.pOfs
            if ((nRecIdx < nHalfNumRecs) || (nSplitPointDnsDepth >= diSrc.dni.nDnsDepth)) {
                diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pLeftBlkHdr + diDst.dni.pDns->nSubBlkOfs);
                diDst.dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
                dbg(SbaCheckOfs(*diDst.dni.pOfs));
            } else {
                diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pRightBlkHdr + diDst.dni.pDns->nSubBlkOfs);
                diDst.dni.nLog2Size = DirNodeGetSubBlkLog2Size(diDst.dni.pSubBlkHdr);
                diDst.dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s(diDst.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diDst.dni.nLog2Size);
            }
        }
        nSrcNodeOfs = DirNodeGetIntraBlkOfs_l2s(diSrc.dni.pDns->nEntryIdx, diSrc.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
        dbg(SbaCheckOfs(nSrcNodeOfs));
        WFSKrnReturnOnError(DirNodeStackGetEntry(&diSrc.dni));
        diSrc.dni.pDns->nBlkAdr = nBlkAdr;
        diSrc.dni.pDns->nSubBlkOfs = nSrcNodeOfs;
#if _DEBUG_DIR
        ++diSrc.dni.pStrPtr;
#endif
    }
}

WFSKrnResult DirReGetHashAdr(AreaInfo *pAreaInfo, DirEntryAttrHdr *pAttrHdr, WFSBlkAdr nNewBlkAdr, DirBlkHdr* pNewBlkHdr){
    WFSKrnResult nResult = WFSKRN_RESULT_UNKNOWN;
    switch(pAttrHdr->nSizeCategory) {
        case FILE_SIZE_CATEGORY_SMALL:{
            u32 nNumFileBlks = ((pAttrHdr->nAllocSize-1)>>pAreaInfo->ah.nLog2BlkSize)+1;
            u32 nI;
            for(nI=0;nI<nNumFileBlks;nI++){
                WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                u32 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pNewBlkHdr;
                nResult = BCacheSetHash(pAreaInfo->pVolInfo, pFileBlkPtr->nBlkAdr, nNewBlkAdr, nHashOfs);
                if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                    return nResult;
                }
            }
            break;
        }
        case FILE_SIZE_CATEGORY_MEDIUM:{
            u32 nNumFileMediumBlks = ((pAttrHdr->nAllocSize-1)>>pAreaInfo->ah.nLog2MediumBlkSize)+1;
            u32 nI;
            for(nI=0;nI<nNumFileMediumBlks;nI++){
                WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                u32 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pNewBlkHdr;
                nResult = BCacheSetHash(pAreaInfo->pVolInfo, pFileBlkPtr->nBlkAdr, nNewBlkAdr, nHashOfs);
                if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                    return nResult;
                }
            }
            break;
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            u32 nNumBlks = ((pAttrHdr->nAllocSize-1)>>pAreaInfo->ah.nLog2MediumBlkSize)+1;
            WFSFileLargeBlkPtr *pFileLargeBlkPtr = 0;
            u32 nI;
            for(nI=0;nI<nNumBlks;nI++){
                if(!(nI%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk))){
                    pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nI>>pAreaInfo->nLog2MediumBlkPerLargeBlk);
                }
                WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pAreaInfo, pFileLargeBlkPtr, nI%(1<<pAreaInfo->nLog2MediumBlkPerLargeBlk));
                u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pNewBlkHdr;
                nResult = BCacheSetHash(pAreaInfo->pVolInfo, fileBlkInfo.nBlkAdr, nNewBlkAdr, nHashOfs);
                if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                    return nResult;
                }
            }
        }
        default:
            break;
    }
    return WFSKRN_RESULT_OK;
}


u32 nNumLeafBlkSplits = 0;
static
WFSKrnResult DirLeafBlkSplit(DirItr *pDiSrc) {
    // Split the specified leaf block into two pieces, at its mid point.
    // Copy the first half of the records to new block 1
    // Copy the first half of the records to new block 2
    // Delete the original block and rename new block 2 to be this block
    dbg(DirLeafCheckBlk(pDiSrc));
    DirItr diSrc=*pDiSrc, diDst;
#if _DEBUG_DIR
    ++nNumLeafBlkSplits;
    if (nNumLeafBlkSplits==3086) {
        int a;a=0;
    }
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
    WFSFileName name;
    diSrc.dni.pStrPtr = name.sStr;
#endif
    WFSKrnResult nResult;
    u32 nStrIdx=0, nSubBlkStrLen;
    // Allocate new blocks
    WFSBlkAdr nBlkAdr = diSrc.tba.nBlkAdr;
    WFSBlkAdr aDstBlkAdr[DIR_MAX_SPLIT_BLK_ALLOC];
    WFSKrnReturnOnError(AreaAllocBlks(diSrc.tba.pAreaInfo, nBlkAdr, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC, aDstBlkAdr, sizeof(WFSBlkAdr)));
    WFSBlkAdr nParentBlkAdr = 0;
    while(diSrc.dni.nDnsDepth > 1) {
        if (diSrc.dni.pDns->nBlkAdr != nBlkAdr) {
            nParentBlkAdr = diSrc.dni.pDns->nBlkAdr;
            break;
        }
        DirPopNodeStackEntry(&diSrc.dni);
    }
    DirBlkHdr *pLeftBlkHdr;
    DirBlkHdr *pRightBlkHdr;
    nResult = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, aDstBlkAdr[0], WFS_MDF_DIR_LEAF_BLK, diSrc.tba.pTransInfo, &pLeftBlkHdr);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    nResult = DirCreateAndInitNewBlk4(diSrc.tba.pAreaInfo, aDstBlkAdr[1], WFS_MDF_DIR_LEAF_BLK, diSrc.tba.pTransInfo, &pRightBlkHdr);
    if(nResult < WFSKRN_RESULT_OK){
        TransUnpinAndFreeBlks(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, 1, aDstBlkAdr, diSrc.tba.pTransInfo);
        return nResult;
    }
#if _DEBUG_BREAK_POINT
    if(aDstBlkAdr[0] == 6370 || aDstBlkAdr[1] == 6370){
        int a; a=0;
    }
#endif
    WFSBlkAdr nCurrentDstBlkAdr = aDstBlkAdr[0];
    diDst.pBlkHdr = pLeftBlkHdr;
    // Initialize a stack frame for traversing the destination block
    diDst.dni.nDnsDepth = 0;
    nResult = DirNodeStackGetEntry(&diDst.dni);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diDst.ppDnsRootParent = &diDst.dni.pDns->pParent;
    diSrc.dni.nEntryIdx = 1;
    diSrc.dni.pOfs = &diSrc.pBlkHdr->nRootOfs;
    dbg(SbaCheckOfs(*diSrc.dni.pOfs));
    nResult = DirPushNodeStackEntry(&diSrc.dni, nBlkAdr);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Error;
    }
    diSrc.dni.pDns->nStrIdx = 0;    // ToDo: This value can be initialized to 0 because prefix-suppression is not implemented
    u16 nSrcNodeOfs = diSrc.dni.pDns->nSubBlkOfs; // = diSrc.pBlkHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nRecIdx=0;
    u32 nHalfNumRecs = (u32)(diSrc.pBlkHdr->nNumRecs>>1);
    diDst.dni.pOfs = &diDst.pBlkHdr->nRootOfs;
    //dbg(SbaCheckOfs(*diDst.dni.pOfs));
    s32 nSplitPointDnsDepth=-1;
    u32 nCopied = 1;
    u16 nLeftChildOfs = 0;
    WFSFileName parentKey;
    u32 nSrcLeafStartDepth = diSrc.dni.nDnsDepth;

    // Perform a depth-first traversal of the block, until the middle leaf sub-block is found.
    while(1) {
        diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + nSrcNodeOfs);
        diSrc.dni.nLog2Size = DirLeafGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        // Copy over the whole node from the original block
        nResult = DirNodeStackGetEntry(&diDst.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        *diDst.dni.pOfs = diDst.dni.pDns->nSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, diSrc.dni.nLog2Size);
        diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diDst.pBlkHdr + *diDst.dni.pOfs);
        memcpy(diDst.dni.pSubBlkHdr, diSrc.dni.pSubBlkHdr, (u32)(1<<diSrc.dni.nLog2Size));
        diDst.dni.pDns->nEntryIdx = 1;
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        diSrc.dni.pDns->nEntryIdx = 1;
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen];

        nSubBlkStrLen = diSrc.dni.pSubBlkHdr->nStrLen;
        nStrIdx += nSubBlkStrLen;
        diSrc.dni.pDns->nStrIdx = (u8)nStrIdx;
#if _DEBUG_DIR
        utf8 *pSubBlkStrPtr = diSrc.dni.pSubBlkHdr->aFirstChar;
        while(nSubBlkStrLen--) {
            *diSrc.dni.pStrPtr++ = *pSubBlkStrPtr++;
        }
#endif
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment our record count
#if _DEBUG_DIR
            *diSrc.dni.pStrPtr = 0;
            ++pChoiceChar;
#endif
            // Copy file attributes to new block
            u32 nAttrOfs = pSrcNodeOfsTblEnd[-1];
            dbg(SbaCheckOfs(nAttrOfs));
            DirEntryAttrHdr *pAttrHdr = (DirEntryAttrHdr *)((u8*)diSrc.pBlkHdr + nAttrOfs);
            u16 *pDstNodeOfsTblEnd = (u16*)((u8*)diDst.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
            pDstNodeOfsTblEnd[-1] = SbaAllocSubBlk((WFSSubBlkAllocHdr *)diDst.pBlkHdr, pAttrHdr->nAttrLog2Size);
            DirEntryAttrHdr *pDstAttrHdr = (DirEntryAttrHdr *)((u8*)diDst.pBlkHdr + pDstNodeOfsTblEnd[-1]);
            memcpy(pDstAttrHdr, pAttrHdr, (u32)(1<<pAttrHdr->nAttrLog2Size));
            // Update to the new hash pointer in file attributes
            nResult = DirReGetHashAdr(diSrc.tba.pAreaInfo, pDstAttrHdr, nCurrentDstBlkAdr, diDst.pBlkHdr);
            if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                goto Error;
            }
            ++nRecIdx;
            if (nRecIdx==nHalfNumRecs+1) {
                DirConstructEntryNameFromNodeStack(&parentKey, &diSrc);
#if _DEBUG_DIR
                if (strncmp(parentKey.sStr, name.sStr, WFS_MAX_FILE_NAME_SIZE)) {
                    WFSKrnOutputErrorStr("Error constructing entry name from stack\n");
                };
#endif
                parentKey.nLen -= diSrc.dni.pSubBlkHdr->nStrLen;
                parentKey.sStr[parentKey.nLen] = 0; // This shortens the name to the minimum splitter. ToDo: Can be shortened more in many cases
            }
            ++diSrc.dni.pDns->nEntryIdx;
            ++diDst.dni.pDns->nEntryIdx;
        }
        while(diSrc.dni.pDns->nEntryIdx > diSrc.dni.pSubBlkHdr->nNumEntries) {
            if (diSrc.dni.nDnsDepth == nSrcLeafStartDepth) {
                // We reached the top of the stack, so this should be the last record
                pLeftBlkHdr->nRootOfs = nLeftChildOfs;
                pRightBlkHdr->nRootOfs = diDst.dni.pDns->nSubBlkOfs;
                DirNodeStackFree(&diDst.dni, diDst.ppDnsRootParent);
                pLeftBlkHdr->nNumRecs = (u16)nHalfNumRecs;
                pRightBlkHdr->nNumRecs = (u16)(diSrc.pBlkHdr->nNumRecs - nHalfNumRecs);
                TransUnpinBlk3(diSrc.tba.pAreaInfo, aDstBlkAdr[0], diSrc.tba.pTransInfo);
                TransUnpinBlk3(diSrc.tba.pAreaInfo, aDstBlkAdr[1], diSrc.tba.pTransInfo);
                if (diSrc.pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_ROOT_BLK) {
                    nParentBlkAdr = nBlkAdr;
                    DirReinitBlk(diSrc.tba.pAreaInfo, diSrc.pBlkHdr, WFS_MDF_DIR_ROOT_BLK);
                } else {
                    DirUnpinAndFreeBlk(diSrc.tba.pAreaInfo, nBlkAdr, diSrc.tba.pTransInfo);
                    diSrc.tba.nBlkAdr = nParentBlkAdr;
#if _CHECK_BLK_USAGE // for debug
                    if(DeviceDebugGetPseudoDetachDevice() == 0){
                        DeviceDebugSetPseudoDetachDevice(1);
                    }
                    if(DeviceDebugGetPseudoHashInconsistent() == 0){
                        DeviceDebugSetPseudoHashInconsistent(1);
                    }
#endif
                    nResult = DirGetBlk(&diSrc, TRANS_FLAG_RW_BLK | BCACHE_FLAG_PINNED, 0); // must be debug
                    if (nResult != WFSKRN_RESULT_OK) {
                        goto Error;
                    }
                }
                *pDiSrc=diSrc;
                u32 nFirstBlockToFree = 2;
                nResult = DirNodeInsertEntry(&parentKey, pDiSrc, aDstBlkAdr, &nFirstBlockToFree);
                if (nResult != WFSKRN_RESULT_OK) {
                    goto Error;
                }
                nResult = TransFreeBlks5(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC-nFirstBlockToFree, &aDstBlkAdr[nFirstBlockToFree], diSrc.tba.pTransInfo);
                if (nResult != WFSKRN_RESULT_OK) {
                    WFSKrnOutputErrorCode(nResult);
                }
                return WFSKRN_RESULT_OK;
            }
            u16 nDstChildOfs = diDst.dni.pDns->nSubBlkOfs;
            DirPopNodeStackEntry(&diDst.dni);
            DirPopNodeStackEntry(&diSrc.dni);
#if _DEBUG_DIR
            diSrc.dni.pStrPtr += diSrc.dni.pDns->nStrIdx - nStrIdx;
#endif
            nStrIdx = diSrc.dni.pDns->nStrIdx;
            diSrc.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)diSrc.pBlkHdr + diSrc.dni.pDns->nSubBlkOfs);
            if (nSplitPointDnsDepth >= (s32)diSrc.dni.nDnsDepth) {
                // This is a node which straddles the left and right halves of the source block.
                // At this point it has already been copied in whole to the left block.
                // We need to move part of it to the right block, and remove unused entries from the left block.
                nSplitPointDnsDepth  = (s32)(diSrc.dni.nDnsDepth-1);
                u32 nRightNumEntries = (u32)(diSrc.dni.pSubBlkHdr->nNumEntries - diSrc.dni.pDns->nEntryIdx + 1);
                u32 nLeftNumEntries  = diSrc.dni.pDns->nEntryIdx - nCopied;
                u16 nLeftOfs         = diDst.dni.pDns->nSubBlkOfs;
                nDstChildOfs = diDst.dni.pDns->nSubBlkOfs = DirLeafCopySubBlkTail(pLeftBlkHdr, nLeftOfs, pRightBlkHdr, nRightNumEntries, nDstChildOfs, false);
                dbg(SbaCheckOfs(nDstChildOfs));
                if (nCopied==0) {
                    DirSubBlkHdr *pLeftSubBlkHdr = (DirSubBlkHdr *)((u8*)pLeftBlkHdr + nLeftOfs);
                    diSrc.dni.nLog2Size  = DirLeafGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
                    DirLeafSetIntraBlkOfs_l2s(diSrc.dni.pDns->nEntryIdx, nLeftChildOfs, pLeftSubBlkHdr, diSrc.dni.nLog2Size);
                }
                nLeftChildOfs  = DirLeafDeleteSubBlkTail(pLeftBlkHdr, nLeftOfs, nLeftNumEntries);
                dbg(if (nLeftChildOfs == 0) {
                    // This is not expected to happen since the newly split block should have plenty of free space
                    WFSKrnOutputErrorStr("Fatal Error!\n");
                })
                diDst.dni.pDns->nEntryIdx = 1;
                nCopied = 0;
            }
            ++diDst.dni.pDns->nEntryIdx;
            ++diSrc.dni.pDns->nEntryIdx;
        }
        if ((nRecIdx==nHalfNumRecs)&&(nSplitPointDnsDepth<0)) {
            // We have found the middle record of the leaf block.
            // This code should only be executed once during the loop
            nSplitPointDnsDepth = (s32)diSrc.dni.nDnsDepth;
            nCurrentDstBlkAdr = aDstBlkAdr[1];
            diDst.pBlkHdr = pRightBlkHdr;
        }
#if _DEBUG_DIR
        pChoiceChar = &diSrc.dni.pSubBlkHdr->aFirstChar[diSrc.dni.pSubBlkHdr->nStrLen + diSrc.dni.pDns->nEntryIdx-1];
        *diSrc.dni.pStrPtr = *pChoiceChar;
#endif
        ++nStrIdx;
        diSrc.dni.nLog2Size = DirLeafGetSubBlkLog2Size(diSrc.dni.pSubBlkHdr);
        pSrcNodeOfsTblEnd = (u16*)((u8*)diSrc.dni.pSubBlkHdr + (1<<diSrc.dni.nLog2Size));
        {
            // Calculate the correct address for diDst.dni.pOfs
            if ((nRecIdx < nHalfNumRecs) || (nSplitPointDnsDepth >= diSrc.dni.nDnsDepth)) {
                diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pLeftBlkHdr + diDst.dni.pDns->nSubBlkOfs);
                diDst.dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s(diSrc.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diSrc.dni.nLog2Size);
                dbg(SbaCheckOfs(*diDst.dni.pOfs));
            } else {
                diDst.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pRightBlkHdr + diDst.dni.pDns->nSubBlkOfs);
                diDst.dni.nLog2Size = DirLeafGetSubBlkLog2Size(diDst.dni.pSubBlkHdr);
                diDst.dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s(diDst.dni.pDns->nEntryIdx, diDst.dni.pSubBlkHdr, diDst.dni.nLog2Size);
                //dbg(SbaCheckOfs(*diDst.dni.pOfs));
            }
        }
        nSrcNodeOfs = pSrcNodeOfsTblEnd[-(s32)diSrc.dni.pDns->nEntryIdx];
        dbg(SbaCheckOfs(nSrcNodeOfs));
        nResult = DirNodeStackGetEntry(&diSrc.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            goto Error;
        }
        diSrc.dni.pDns->nBlkAdr = nBlkAdr;
        diSrc.dni.pDns->nSubBlkOfs = nSrcNodeOfs;
#if _DEBUG_DIR
        ++diSrc.dni.pStrPtr;
#endif
    }
Error:
    TransFreeBlks5(diSrc.tba.pAreaInfo, diSrc.tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC, aDstBlkAdr, diSrc.tba.pTransInfo);
    return nResult;
}


static WFSKrnResult _DirSbaAllocSubBlkForResize(DirItr *pDi, u32 nLog2Size, u32 *pNewAttrOfs, DirEntryAttrHdr **ppNewAttrHdr){
    while(1) {
        // Alloc sub block as the new file attributes
        dbg(DirLeafCheckBlk(pDi));
        *pNewAttrOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pDi->pBlkHdr, nLog2Size);
        if (*pNewAttrOfs) {
            break;
        }
        // Split the leaf and retry
        // ToDo: Update the update map.
        // Note: We should not assume that any iterator which performs a resize is going to be registered in the update map.
        //       The file-size may be changed on file close, after which the iterator is no longer needed.
        DirItrUnmapBlks(pDi);
        WFSKrnReturnOnError(DirLeafBlkSplit(pDi));
        s32 nDepth=pDi->dni.nDnsDepth-2;
        DirNodeStack *pDns = pDi->dni.pDns;
        for(; nDepth>0; --nDepth) {
            pDns = pDns->pParent;
        }
        DirItrCloseNoUnpin(pDi);
        if (pDi->tba.nBlkAdr != pDns->nBlkAdr) {
            pDi->tba.nBlkAdr = pDns->nBlkAdr;
        }
        pDi->nBlkDepth = 0;
        WFSKrnResult nResult = DirFind(&pDi->name, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
        if (nResult != WFSKRN_RESULT_DIR_ENTRY_FOUND) {
            return nResult;
        }
        nResult = DirGetBlk(pDi, TRANS_FLAG_RW_BLK, 0);
        if (nResult < WFSKRN_RESULT_OK) {
            return nResult;
        }
        DirItrMapBlks(pDi);
    }
    *ppNewAttrHdr = (DirEntryAttrHdr *)((u8*)pDi->pBlkHdr + *pNewAttrOfs);
    return WFSKRN_RESULT_OK;
}


#define FILE_RESIZE_CATEGORY_VERY_SMALL_TO_VERY_SMALL ((FILE_SIZE_CATEGORY_VERY_SMALL<<4) + FILE_SIZE_CATEGORY_VERY_SMALL)
#define FILE_RESIZE_CATEGORY_VERY_SMALL_TO_SMALL      ((FILE_SIZE_CATEGORY_VERY_SMALL<<4) + FILE_SIZE_CATEGORY_SMALL)
#define FILE_RESIZE_CATEGORY_VERY_SMALL_TO_MEDIUM     ((FILE_SIZE_CATEGORY_VERY_SMALL<<4) + FILE_SIZE_CATEGORY_MEDIUM)
#define FILE_RESIZE_CATEGORY_VERY_SMALL_TO_LARGE      ((FILE_SIZE_CATEGORY_VERY_SMALL<<4) + FILE_SIZE_CATEGORY_LARGE)
#define FILE_RESIZE_CATEGORY_VERY_SMALL_TO_VERY_LARGE ((FILE_SIZE_CATEGORY_VERY_SMALL<<4) + FILE_SIZE_CATEGORY_VERY_LARGE)
#define FILE_RESIZE_CATEGORY_SMALL_TO_VERY_SMALL      ((FILE_SIZE_CATEGORY_SMALL     <<4) + FILE_SIZE_CATEGORY_VERY_SMALL)
#define FILE_RESIZE_CATEGORY_SMALL_TO_SMALL           ((FILE_SIZE_CATEGORY_SMALL     <<4) + FILE_SIZE_CATEGORY_SMALL)
#define FILE_RESIZE_CATEGORY_SMALL_TO_MEDIUM          ((FILE_SIZE_CATEGORY_SMALL     <<4) + FILE_SIZE_CATEGORY_MEDIUM)
#define FILE_RESIZE_CATEGORY_SMALL_TO_LARGE           ((FILE_SIZE_CATEGORY_SMALL     <<4) + FILE_SIZE_CATEGORY_LARGE)
#define FILE_RESIZE_CATEGORY_SMALL_TO_VERY_LARGE      ((FILE_SIZE_CATEGORY_SMALL     <<4) + FILE_SIZE_CATEGORY_VERY_LARGE)
#define FILE_RESIZE_CATEGORY_MEDIUM_TO_VERY_SMALL     ((FILE_SIZE_CATEGORY_MEDIUM    <<4) + FILE_SIZE_CATEGORY_VERY_SMALL)
#define FILE_RESIZE_CATEGORY_MEDIUM_TO_SMALL          ((FILE_SIZE_CATEGORY_MEDIUM    <<4) + FILE_SIZE_CATEGORY_SMALL)
#define FILE_RESIZE_CATEGORY_MEDIUM_TO_MEDIUM         ((FILE_SIZE_CATEGORY_MEDIUM    <<4) + FILE_SIZE_CATEGORY_MEDIUM)
#define FILE_RESIZE_CATEGORY_MEDIUM_TO_LARGE          ((FILE_SIZE_CATEGORY_MEDIUM    <<4) + FILE_SIZE_CATEGORY_LARGE)
#define FILE_RESIZE_CATEGORY_MEDIUM_TO_VERY_LARGE     ((FILE_SIZE_CATEGORY_MEDIUM    <<4) + FILE_SIZE_CATEGORY_VERY_LARGE)
#define FILE_RESIZE_CATEGORY_LARGE_TO_VERY_SMALL      ((FILE_SIZE_CATEGORY_LARGE     <<4) + FILE_SIZE_CATEGORY_VERY_SMALL)
#define FILE_RESIZE_CATEGORY_LARGE_TO_SMALL           ((FILE_SIZE_CATEGORY_LARGE     <<4) + FILE_SIZE_CATEGORY_SMALL)
#define FILE_RESIZE_CATEGORY_LARGE_TO_MEDIUM          ((FILE_SIZE_CATEGORY_LARGE     <<4) + FILE_SIZE_CATEGORY_MEDIUM)
#define FILE_RESIZE_CATEGORY_LARGE_TO_LARGE           ((FILE_SIZE_CATEGORY_LARGE     <<4) + FILE_SIZE_CATEGORY_LARGE)
#define FILE_RESIZE_CATEGORY_LARGE_TO_VERY_LARGE      ((FILE_SIZE_CATEGORY_LARGE     <<4) + FILE_SIZE_CATEGORY_VERY_LARGE)
#define FILE_RESIZE_CATEGORY_VERY_LARGE_TO_VERY_SMALL ((FILE_SIZE_CATEGORY_VERY_LARGE<<4) + FILE_SIZE_CATEGORY_VERY_SMALL)
#define FILE_RESIZE_CATEGORY_VERY_LARGE_TO_SMALL      ((FILE_SIZE_CATEGORY_VERY_LARGE<<4) + FILE_SIZE_CATEGORY_SMALL)
#define FILE_RESIZE_CATEGORY_VERY_LARGE_TO_MEDIUM     ((FILE_SIZE_CATEGORY_VERY_LARGE<<4) + FILE_SIZE_CATEGORY_MEDIUM)
#define FILE_RESIZE_CATEGORY_VERY_LARGE_TO_LARGE      ((FILE_SIZE_CATEGORY_VERY_LARGE<<4) + FILE_SIZE_CATEGORY_LARGE)
#define FILE_RESIZE_CATEGORY_VERY_LARGE_TO_VERY_LARGE ((FILE_SIZE_CATEGORY_VERY_LARGE<<4) + FILE_SIZE_CATEGORY_VERY_LARGE)


WFSKrnResult DirResizeFile(DirItr *pDi, WFSFileSize nNewSize) {

    // Resize file cases
    // The resize file operation is complicated by the fact that files have to be laid out
    // differently depending on which of 5 size categories they fall into.
    // We can go from any category to any other category.
    //
    //          To: Very Small Size Category | Small Size Category        | Medium Size Category            | Large Size Category             | Very Large Size Category        |
    // --From:-----+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+
    //             | Copy data directly from | Allocate new block(s)      | Allocate new medium block(s)    | Allocate new large block(s)     | Allocate new large blocks and   |
    // Very Small  | old attr sub-block to   | -------------------------- | ------------------------------- | ------------------------------- | indirect block(s).              |
    // Size        | new attr sub-block      | Copy data from old attr    | Copy data from old attr sub-    | Copy data from old attr sub-    | ------------------------------- |
    // Category    |                         | sub-block to first new     | block to first large block.     | block to first large block.     | Copy data from old attr sub-    |
    //             |                         | block.                     | ------------------------------- | ------------------------------- | block to first large block.     |
    //             |                         | -------------------------- | Calculate hash for first 64KB   | Calculate hash for first 64KB   | ------------------------------- |
    //             |                         | Calculate hash for new     | and write to new attr sub-block | and write to new attr sub-block | Calculate hash for first 64KB   |
    //             |                         | block and write to new     |                                 |                                 | of large block, and write to    |
    //             |                         | attr sub-block along       |                                 |                                 | first indirect block along with |
    //             |                         | with pointers to new       |                                 |                                 | pointer.                        |
    //             |                         | blocks.                    |                                 |                                 | ------------------------------- |
    //             |                         |                            |                                 |                                 | Write pointers to indirect      |
    //             |                         |                            |                                 |                                 | block(s) to new attr sub-block. |
    // ------------+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+
    // Small Size  | Copy data from first    | Allocate new blocks or     | Allocate new medium blocks.     | Allocate new large blocks.      | Allocate new large blocks and   |
    // Category    | block to new attr       | free old ones depending    | ------------------------------- | ------------------------------- | indirect block(s).              |
    //             | sub-block               | on change in file size     | Copy data from old blocks to    | Copy data from old blocks to    | ------------------------------- |
    //             | ----------------------- | -------------------------- | new large blocks.               | new large blocks.               | Copy data from old blocks to    |
    //             | Free old block(s).      | Copy pointers from old     | ------------------------------- | ------------------------------- | first new large block.          |
    //             |                         | attr sub-block to new      | Re-calculate hash values at     | Re-calculate hash values at     | ------------------------------- |
    //             |                         | attr sub-block             | 64KB granularity, and write to  | 64KB granularity, and write to  | Copy pointers and hash values   |
    //             |                         |                            | new attr sub-block.             | new attr sub-block.             | from old attr sub-block to      |
    //             |                         |                            | ------------------------------- | ------------------------------- | first new indirect block.       |
    //             |                         |                            | Free old blocks.                | Free old blocks.                | ------------------------------- |
    //             |                         |                            |                                 |                                 | Write pointers to indirect      |
    //             |                         |                            |                                 |                                 | block(s) to new attr sub-block. |
    // ------------+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+
    // Medium Size | Copy data from first    | Convert the first medium   | Allocate new medium blocks or   | Allocate new large blocks.      | Allocate new large block(s)     |
    // Category    | medium block to new     | block into small block(s), | free old ones depending on      | ------------------------------- | and indirect block(s).          |
    //             | attr sub-block          | calculate hash values on   | change in file size.            | Copy data from old blocks to    | ------------------------------- |
    //             | ----------------------- | the granularity of small   | ------------------------------- | new large blocks.               | Copy data from old blocks to    |
    //             | Free old large block(s) | blocks, and store with     | copy pointers from old attr     | ------------------------------- | new large blocks.               |
    //             |                         | pointers in the new attr   | sub-block to new attr sub-block | copy pointers from old attr     | ------------------------------- |
    //             |                         | sub-block.                 |                                 | sub-block to new attr sub-block.| copy pointers from old attr     |
    //             |                         | -------------------------- |                                 | ------------------------------- | sub-block to new attr sub-block.|
    //             |                         | Free unused blocks and     |                                 | Free old blocks.                | ------------------------------- |
    //             |                         | unused parts of the first  |                                 |                                 | Free old blocks.                |
    //             |                         | block.                     |                                 |                                 |                                 |
    // ------------+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+
    // Large Size  | Copy data from first    | Convert the first large    | Allocate new large blocks or    | Allocate new large blocks or    | Allocate new large block(s)     |
    // Category    | large block to new      | block into small block(s), | free old ones depending on      | free old ones depending on      | and indirect block(s).          |
    //             | attr sub-block          | calculate hash values on   | change in file size.            | change in file size.            | ------------------------------- |
    //             | ----------------------- | the granularity of small   | ------------------------------- | ------------------------------- | Copy pointers and hash values   |
    //             | Free old large block(s) | blocks, and store with     | copy pointers from old attr     | copy pointers from old attr     | from old attr sub-block to      |
    //             |                         | pointers in the new attr   | sub-block to new attr sub-block | sub-block to new attr sub-block | first new indirect block.       |
    //             |                         | sub-block.                 |                                 |                                 | ------------------------------- |
    //             |                         | -------------------------- |                                 |                                 | Write pointers to indirect      |
    //             |                         | Free unused blocks and     |                                 |                                 | block(s) to new attr sub-block. |
    //             |                         | unused parts of the first  |                                 |                                 |                                 |
    //             |                         | block.                     |                                 |                                 |                                 |
    // ------------+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+
    // Very Large  | Copy data from first    | Convert the first large    | Copy pointers and hash values   | Copy pointers and hash values   | Allocate new large blocks and/  |
    // Size        | large block to new      | block into small block(s), | from old indirect block to new  | from old indirect block to new  | or indirect block(s), or free   |
    // Category    | attr sub-block          | calculate hash values on   | attr sub-block.                 | attr sub-block.                 | old large blocks and/or         |
    //             | ----------------------- | the granularity of small   | ------------------------------- | ------------------------------- | indirect block(s) depending on  |
    //             | Free old large blocks   | blocks, and store with     | Free unused block(s) and        | Free unused block(s) and        | change in size.                 |
    //             | and indirect block(s).  | pointers in the new attr   | unused indirect block(s).       | unused indirect block(s).       | ------------------------------- |
    //             |                         | sub-block                  |                                 |                                 | copy pointers from old attr     |
    //             |                         | -------------------------- |                                 |                                 | sub-block to new attr sub-      |
    //             |                         | Free unused blocks,        |                                 |                                 | block.                          |
    //             |                         | unused parts of the first  |                                 |                                 |                                 |
    //             |                         | block, and unused indirect |                                 |                                 |                                 |
    //             |                         | block(s).                  |                                 |                                 |                                 |
    // ------------+-------------------------+----------------------------+---------------------------------+---------------------------------+---------------------------------+

#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
#if _DEBUG_BREAK_POINT
    static int c=0; c++; if (c == 196) {
        int a; a=0;
    }
#endif

    WFSKrnResult nResult = WFSKRN_RESULT_UNKNOWN;
    TransBlkAdr *pTba = &pDi->tba;

    DirEntryAttrHdr *pAttrHdr = pDi->pAttrHdr;
    u32 nAttrHdrSize = (u32)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + ((pAttrHdr->nNameLen+7)>>3);
    u32 nNewAttrOfs = 0;

    u32 nAllocSize = nNewSize;
    u8  nNewSizeCategory = 0;
    u32 nNewAttrLog2Size = 0;

    u32 nNumBlks = 0;
    u32 nNewNumBlks = 0;
    u32 nStride = 0;
    s32 nNumStartBlks = 0;

    if(nNewSize > pAttrHdr->nAllocSize){
        nNewAttrLog2Size = DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(pDi->tba.pAreaInfo, &nAllocSize, pDi->pAttrHdr->nNameLen, &nNewSizeCategory);
        if(nNewSizeCategory < pAttrHdr->nSizeCategory) {
            nAllocSize = nNewSize;
            nNewAttrLog2Size = DirCalculateAttrSubBlkLog2SizeDecreaseUpdateAllocSize(pDi->tba.pAreaInfo, &nAllocSize, pDi->pAttrHdr->nNameLen, &nNewSizeCategory);
        }
    }
    else{
        /*
        pAttrHdr->file.nSize = nNewSize; 
        return WFSKRN_RESULT_OK;
        */
        if(pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_VERY_SMALL){
            pAttrHdr->file.nSize = nNewSize;
            return WFSKRN_RESULT_OK;
        }
        nNewAttrLog2Size = DirCalculateAttrSubBlkLog2SizeDecreaseUpdateAllocSize(pDi->tba.pAreaInfo, &nAllocSize, pDi->pAttrHdr->nNameLen, &nNewSizeCategory);
        if(nNewSizeCategory > pAttrHdr->nSizeCategory) {
            nAllocSize = nNewSize;
            nNewAttrLog2Size = DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(pDi->tba.pAreaInfo, &nAllocSize, pDi->pAttrHdr->nNameLen, &nNewSizeCategory);
        }
    }

    DirEntryAttrHdr *pNewAttrHdr = ( nNewAttrLog2Size == pAttrHdr->nAttrLog2Size && nNewSizeCategory == pAttrHdr->nSizeCategory ) ? pAttrHdr : 0;

    switch(pAttrHdr->nSizeCategory){
        case FILE_SIZE_CATEGORY_SMALL:
            nNumBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2BlkSize) + 1 : 0;
            break;
        case FILE_SIZE_CATEGORY_MEDIUM:
            nNumBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2MediumBlkSize) + 1 : 0;
            break;
        case FILE_SIZE_CATEGORY_LARGE:
        case FILE_SIZE_CATEGORY_VERY_LARGE:
            nNumBlks = pAttrHdr->nAllocSize ? ((pAttrHdr->nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2LargeBlkSize) + 1 : 0;
            break;
    }

    // Culculate nNumBlks and nNewNumBlks
    if(nAllocSize != pAttrHdr->nAllocSize){
        switch(nNewSizeCategory){
            case FILE_SIZE_CATEGORY_VERY_SMALL:{
                if(pNewAttrHdr != pAttrHdr){
                    // Allocate subblock
                    WFSKrnReturnOnError(_DirSbaAllocSubBlkForResize(pDi, nNewAttrLog2Size, &nNewAttrOfs, &pNewAttrHdr));
                    pAttrHdr = pDi->pAttrHdr;
                }
                break;
            }
            case FILE_SIZE_CATEGORY_SMALL:{
                nStride     = sizeof(WFSFileBlkPtr);
                nNewNumBlks = nAllocSize ? ((nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2BlkSize) + 1 : 0;
                if(pNewAttrHdr != pAttrHdr){
                    // Allocate subblock
                    WFSKrnReturnOnError(_DirSbaAllocSubBlkForResize(pDi, nNewAttrLog2Size, &nNewAttrOfs, &pNewAttrHdr));
                    pAttrHdr = pDi->pAttrHdr;
                }
                if( pAttrHdr->nAllocSize < nAllocSize || nNewSizeCategory != pAttrHdr->nSizeCategory ){
                    // Alloc new blocks
                    nNumStartBlks = (nNewSizeCategory == pAttrHdr->nSizeCategory) ? nNumBlks : 0;
                    WFSBlkAdr nRefBlkAdr = nNumStartBlks ? WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumStartBlks-1)->nBlkAdr : pTba->nBlkAdr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nNewNumBlks-1);
                    memset(pFileBlkPtr, 0, (nNewNumBlks-nNumStartBlks)*sizeof(WFSFileBlkPtr));
                    WFSKrnExitOnError(AreaAllocBlks(pTba->pAreaInfo, nRefBlkAdr, pTba->pAreaInfo->ah.nLog2BlkSize, nNewNumBlks-nNumStartBlks, 
                        (WFSBlkAdr*)&pFileBlkPtr[nNewNumBlks-nNumStartBlks-1], -(s32)nStride));
                    _qsort(pFileBlkPtr, nNewNumBlks-nNumStartBlks, sizeof(WFSFileBlkPtr), CompareBlkAdrReverse);
                } else {
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks);
                    WFSKrnExitOnError(AreaFreeBlks(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumBlks-nNewNumBlks, (WFSBlkAdr*)pFileBlkPtr, -(s32)nStride));
                }
                break;
            }
            case FILE_SIZE_CATEGORY_MEDIUM:{
                nStride     = sizeof(WFSFileBlkPtr);
                nNewNumBlks = nAllocSize ? ((nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2MediumBlkSize) + 1 : 0;
                if(pNewAttrHdr != pAttrHdr){
                    // Allocate subblock
                    WFSKrnReturnOnError(_DirSbaAllocSubBlkForResize(pDi, nNewAttrLog2Size, &nNewAttrOfs, &pNewAttrHdr));
                    pAttrHdr = pDi->pAttrHdr;
                }
                // Alloc new blocks
                if( pAttrHdr->nAllocSize < nAllocSize || nNewSizeCategory != pAttrHdr->nSizeCategory ){
                    nNumStartBlks = (nNewSizeCategory == pAttrHdr->nSizeCategory) ? nNumBlks : 0;
                    WFSBlkAdr nRefBlkAdr = nNumStartBlks ? WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumStartBlks-1)->nBlkAdr : pTba->nBlkAdr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nNewNumBlks-1);
                    memset(pFileBlkPtr, 0, (nNewNumBlks-nNumStartBlks)*sizeof(WFSFileBlkPtr));
                    WFSKrnExitOnError(AreaAllocBlks(pTba->pAreaInfo, nRefBlkAdr, pTba->pAreaInfo->ah.nLog2MediumBlkSize, nNewNumBlks-nNumStartBlks, 
                        (WFSBlkAdr*)&pFileBlkPtr[nNewNumBlks-nNumStartBlks-1], -(s32)nStride));
                    _qsort(pFileBlkPtr, nNewNumBlks-nNumStartBlks, sizeof(WFSFileBlkPtr), CompareBlkAdrReverse);
                } else {
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks);
                    WFSKrnExitOnError(AreaFreeBlks(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2MediumBlkSize, nNumBlks-nNewNumBlks, (WFSBlkAdr*)pFileBlkPtr, -(s32)nStride));
                }
                break;
            }
            case FILE_SIZE_CATEGORY_LARGE:{
                nStride     = pTba->pAreaInfo->nLargeBlkPtrSize;
                nNewNumBlks = nAllocSize ? ((nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2LargeBlkSize) + 1 : 0;
                if(pNewAttrHdr != pAttrHdr){
                    // Allocate subblock
                    WFSKrnReturnOnError(_DirSbaAllocSubBlkForResize(pDi, nNewAttrLog2Size, &nNewAttrOfs, &pNewAttrHdr));
                    pAttrHdr = pDi->pAttrHdr;
                }
                // Alloc new blocks
                if( pAttrHdr->nAllocSize < nAllocSize ){
                    nNumStartBlks = (nNewSizeCategory == pAttrHdr->nSizeCategory) ? nNumBlks : 0;
                    WFSBlkAdr nRefBlkAdr = nNumStartBlks ? WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumStartBlks-1)->nBlkAdr : pTba->nBlkAdr;
                    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pNewAttrHdr, nNewAttrLog2Size, nNewNumBlks-1);
                    memset(pFileLargeBlkPtr, 0, (nNewNumBlks-nNumStartBlks)*sizeof(WFSFileLargeBlkPtr));
                    WFSKrnExitOnError(AreaAllocBlks(pTba->pAreaInfo, nRefBlkAdr, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNewNumBlks-nNumStartBlks, 
                        (WFSBlkAdr*)&pFileLargeBlkPtr[nNewNumBlks-nNumStartBlks-1], -(s32)nStride));
                    _qsort(pFileLargeBlkPtr, nNewNumBlks-nNumStartBlks, sizeof(WFSFileLargeBlkPtr), CompareBlkAdrReverse);
                } else if (nNewSizeCategory == pAttrHdr->nSizeCategory){
                    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks);
                    WFSKrnExitOnError(AreaFreeBlks(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumBlks-nNewNumBlks, (WFSBlkAdr*)pFileLargeBlkPtr, -(s32)nStride));            
                } 
                break;
            }
            case FILE_SIZE_CATEGORY_VERY_LARGE:{
                nStride     = pTba->pAreaInfo->nLargeBlkPtrSize;
                nNewNumBlks = nAllocSize ? ((nAllocSize-1) >> pTba->pAreaInfo->ah.nLog2LargeBlkSize) + 1 : 0;
                if(pNewAttrHdr != pAttrHdr){
                    // Allocate subblock
                    WFSKrnReturnOnError(_DirSbaAllocSubBlkForResize(pDi, nNewAttrLog2Size, &nNewAttrOfs, &pNewAttrHdr));
                    pAttrHdr = pDi->pAttrHdr;
                }
                // Alloc new blocks
                if( pAttrHdr->nAllocSize < nAllocSize ){
                    s32 nNumIndirectBlks = 0;
                    if(nNewSizeCategory == pAttrHdr->nSizeCategory){
                        nNumStartBlks = nNumBlks;
                        nNumIndirectBlks = (nNumStartBlks+pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                        *WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nNumIndirectBlks-1) =
                            *WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nNumIndirectBlks-1);
                    } else if (pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_LARGE){
                        nNumStartBlks = nNumBlks;
                    }
                    s32 nNewNumIndirectBlks = (nNewNumBlks+pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                    if(nNumIndirectBlks < nNewNumIndirectBlks){
                        WFSBlkAdr *pBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nNewNumIndirectBlks-1);
                        WFSBlkAdr nRefBlkAdr = nNumStartBlks ? pBlkAdr[nNewNumIndirectBlks-nNumIndirectBlks] : pTba->nBlkAdr;
                        WFSKrnExitOnError(AreaAllocBlks(pTba->pAreaInfo, nRefBlkAdr, pTba->pAreaInfo->ah.nLog2BlkSize, nNewNumIndirectBlks-nNumIndirectBlks, 
                            &pBlkAdr[nNewNumIndirectBlks-nNumIndirectBlks-1], -(s32)sizeof(WFSBlkAdr)));
                        _qsort(pBlkAdr, nNewNumIndirectBlks-nNumIndirectBlks, sizeof(WFSBlkAdr), CompareBlkAdrReverse);
                    }
                    u32 nFlag;
                    u32 nStartBlkOfs = nNumStartBlks%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                    u32 nEndBlkOfs   = nNewNumBlks ? (nNewNumBlks-1)%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1 : 0;
                    if(nStartBlkOfs == 0 || pAttrHdr->nSizeCategory == FILE_SIZE_CATEGORY_LARGE) {
                        ++nNumIndirectBlks;
                        nFlag = TRANS_FLAG_WRITE_BLK;
                    } else {
                        nFlag = TRANS_FLAG_RW_BLK;
                    }
                    WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nNumIndirectBlks-1);
                    WFSBlkAdr nRefBlkAdr = *pIndirectBlkAdr;
                    for(;nNumIndirectBlks<=nNewNumIndirectBlks;nNumIndirectBlks++){
                        u8 *pIndirectBlkPtr;
                        WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, nFlag, (void**)&pIndirectBlkPtr));
                        WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nStartBlkOfs);
                        if(nFlag == TRANS_FLAG_RW_BLK) { nRefBlkAdr = ((WFSFileLargeBlkPtr*)((u8*)pFileLargeBlkPtr-sizeof(WFSFileLargeBlkPtr)))->nBlkAdr; }
                        u32 nNumAllocBlks = nNumIndirectBlks < nNewNumIndirectBlks ? pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-nStartBlkOfs : nEndBlkOfs-nStartBlkOfs;
                        memset(pFileLargeBlkPtr, 0, nNumAllocBlks*sizeof(WFSFileLargeBlkPtr));
                        WFSKrnExitOnError(AreaAllocBlks(pTba->pAreaInfo, nRefBlkAdr, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumAllocBlks, &pFileLargeBlkPtr->nBlkAdr, (s32)nStride));
                        _qsort(&pFileLargeBlkPtr->nBlkAdr, nNumAllocBlks, sizeof(WFSFileLargeBlkPtr), CompareBlkAdrForward);
                        nRefBlkAdr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nNumAllocBlks+nStartBlkOfs-1)->nBlkAdr;
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                        pIndirectBlkAdr = &pIndirectBlkAdr[-1];
                        nFlag = TRANS_FLAG_WRITE_BLK;
                        nStartBlkOfs = 0;
                    }
                } else if (nNewSizeCategory == pAttrHdr->nSizeCategory){
                    s32 nNumIndirectBlks = (nNumBlks-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1;
                    WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nNumIndirectBlks-1);
                    s32 nNumEndBlks  = (nNumBlks-1)%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1;
                    u32 nNumFreeBlks = nNumBlks-nNewNumBlks;
                    while(nNumFreeBlks > 0){
                        u8 *pIndirectBlkPtr;
                        WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                        WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nNumEndBlks-1);
                        if(nNumFreeBlks >= nNumEndBlks){
                            WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumEndBlks, &pFileLargeBlkPtr->nBlkAdr, -(s32)nStride, pTba->pTransInfo));
                            WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                            WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, 1, pIndirectBlkAdr, -(s32)nStride, pTba->pTransInfo));
                            nNumFreeBlks -= nNumEndBlks;
                        } else {
                            WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumFreeBlks, &pFileLargeBlkPtr->nBlkAdr, -(s32)nStride, pTba->pTransInfo));
                            WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                            nNumFreeBlks = 0;
                        }
                        pIndirectBlkAdr = &pIndirectBlkAdr[1];
                        nNumEndBlks = pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                    }
                }
                break;
            }
        }
    } else {
        nNewNumBlks = nNumBlks;
        nNewSizeCategory = pAttrHdr->nSizeCategory;
    }

    if(pNewAttrHdr != pAttrHdr){
        u32 nCopyForward = nAttrHdrSize;
        u32 nNewSizeCategory2 = (pAttrHdr->nSizeCategory << 4) + nNewSizeCategory;
        switch(nNewSizeCategory2) {
            // Same categories
            case FILE_RESIZE_CATEGORY_VERY_SMALL_TO_VERY_SMALL:{
                // Copy attribute header and data( for only increase )
                nCopyForward += pAttrHdr->file.nSize < nAllocSize ? pAttrHdr->file.nSize : nAllocSize;
                break;
            }
            case FILE_RESIZE_CATEGORY_SMALL_TO_SMALL:{
                // Copy attribute data (nBlkAdr + hash)
                WFSFileBlkPtr *pFileBlkPtr = 0;
                u32 nNumCopyBlks = pAttrHdr->nAllocSize < nAllocSize ? nNumBlks : nNewNumBlks;
                memcpy((u8*)pNewAttrHdr+(1<<nNewAttrLog2Size)-(nStride*nNumCopyBlks), (u8*)pAttrHdr+(1<<pAttrHdr->nAttrLog2Size)-(nStride*nNumCopyBlks), nStride*nNumCopyBlks);
                if(pAttrHdr->nAllocSize < nAllocSize){
                    nNumCopyBlks = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2BlkSize)+1 : 0;
                }
                int nI;
                u32 nHashOfs = 0;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    // If there is the resized data in the block caches, the hash address is updated.  
                    pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr,nNewAttrLog2Size, nI);
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    nResult = BCacheSetHash(pTba->pAreaInfo->pVolInfo, pFileBlkPtr->nBlkAdr, pTba->nBlkAdr, nHashOfs);
                    if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                        goto Exit;
                    }
                }
                if(pAttrHdr->nAllocSize > nAllocSize && nNumCopyBlks){
                    WFSKrnExitOnError(TransTrancateFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, pTba->pAreaInfo->nBlkSize, ((nNewSize-1)&(pTba->pAreaInfo->nBlkSize-1))+1));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_MEDIUM_TO_MEDIUM:{
                // Copy attribute data (nBlkAdr + hash)
                WFSFileBlkPtr *pFileBlkPtr = 0;
                u32 nNumCopyBlks = pAttrHdr->nAllocSize < nAllocSize ? nNumBlks : nNewNumBlks;
                memcpy((u8*)pNewAttrHdr+(1<<nNewAttrLog2Size)-(nStride*nNumCopyBlks), (u8*)pAttrHdr+(1<<pAttrHdr->nAttrLog2Size)-(nStride*nNumCopyBlks), nStride*nNumCopyBlks);
                if(pAttrHdr->nAllocSize < nAllocSize){
                    nNumCopyBlks = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)+1 : 0;
                }
                int nI;
                u32 nHashOfs = 0;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    // If there is the resized data in the block caches, the hash address is updated.  
                    pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr,nNewAttrLog2Size, nI);
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    nResult = BCacheSetHash(pTba->pAreaInfo->pVolInfo, pFileBlkPtr->nBlkAdr, pTba->nBlkAdr, nHashOfs);
                    if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                        goto Exit;
                    }
                }
                if(pAttrHdr->nAllocSize > nAllocSize && nNumCopyBlks){
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, pTba->pAreaInfo->nMediumBlkSize, ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1 ));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_LARGE_TO_LARGE: {
               // Copy attribute data (nBlkAdr + hash)
                WFSFileLargeBlkPtr *pFileLargeBlkPtr = 0;
                u32 nNumCopyBlks = pAttrHdr->nAllocSize < nAllocSize ? nNumBlks : nNewNumBlks;
                memcpy((u8*)pNewAttrHdr+(1<<nNewAttrLog2Size)-(nStride*nNumCopyBlks), (u8*)pAttrHdr+(1<<pAttrHdr->nAttrLog2Size)-(nStride*nNumCopyBlks), nStride*nNumCopyBlks);
                if(pAttrHdr->nAllocSize < nAllocSize){
                    nNumCopyBlks = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2LargeBlkSize)+1 : 0;
                }
                int nI, nJ;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pNewAttrHdr, nNewAttrLog2Size, nI);
                    for(nJ=0;nJ<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk;nJ++){
                        WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pFileLargeBlkPtr, nJ);
                        u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                        nResult = BCacheSetHash(pTba->pAreaInfo->pVolInfo, fileBlkInfo.nBlkAdr, pTba->nBlkAdr, nHashOfs);
                        if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                            goto Exit;
                        }
                    }
                }
                if(pAttrHdr->nAllocSize > nAllocSize && nNumCopyBlks){
                    u32 nMedBlkIdx = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)&((1<<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk)-1);
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pFileLargeBlkPtr, nMedBlkIdx);
                    u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, pTba->pAreaInfo->nMediumBlkSize, 
                                                                 ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_LARGE_TO_VERY_LARGE:{
                u32 nCopyBytes = sizeof(WFSBlkAdr)*((nNumBlks+pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk);
                if(nCopyBytes){
                    memcpy((u8*)pNewAttrHdr+(1<<nNewAttrLog2Size)-nCopyBytes, (u8*)pAttrHdr+(1<<pAttrHdr->nAttrLog2Size)-nCopyBytes, nCopyBytes);
                }
                if(pAttrHdr->nAllocSize > nAllocSize && nNewSize){
                    u8 *pIndirectBlkPtr;
                    u32 nTrancateSize = ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1;
                    if(nTrancateSize == pTba->pAreaInfo->nMediumBlkSize){ break; }
                    u32 nIndirectBlkIdx = (nNewNumBlks-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                    u32 nMedBlkIdx = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)%(pTba->pAreaInfo->nNumMediumBlkPtrsPerBlk);
                    WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nIndirectBlkIdx);
                    WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr));
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nMedBlkIdx);
                    u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, pTba->pAreaInfo->nMediumBlkSize, nTrancateSize));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                }
                break;
            }
            // Different categories
            case FILE_RESIZE_CATEGORY_VERY_SMALL_TO_SMALL:{
                if(pAttrHdr->file.nSize){
                    u8 *pBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, 0);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileBlk(pTba,  pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, pAttrHdr->file.nSize, (void**)&pBlkPtr));
                    memcpy(pBlkPtr, (u8*)pAttrHdr+nAttrHdrSize, pAttrHdr->file.nSize);
                    WFSKrnExitOnError(TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_SMALL_TO_MEDIUM:{
                if(pAttrHdr->file.nSize){
                    u8 *pBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, 0);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba,  pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, pAttrHdr->file.nSize, (void**)&pBlkPtr));
                    memcpy(pBlkPtr, (u8*)pAttrHdr+nAttrHdrSize, pAttrHdr->file.nSize);
                    WFSKrnExitOnError(TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_SMALL_TO_LARGE:{
                if(pAttrHdr->file.nSize){
                    u8 *pBlkPtr;
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, 0);
                    u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK,  pAttrHdr->file.nSize, (void**)&pBlkPtr));
                    memcpy(pBlkPtr, (u8*)pAttrHdr+nAttrHdrSize, pAttrHdr->file.nSize);
                    WFSKrnExitOnError(TransStoreBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_SMALL_TO_VERY_LARGE:{
                if(pAttrHdr->file.nSize){
                    u8 *pBlkPtr;
                    u8 *pIndirectBlkPtr;
                    WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, 0);
                    WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr));
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, 0);
                    u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK,  pAttrHdr->file.nSize, (void**)&pBlkPtr));
                    memcpy(pBlkPtr, (u8*)pAttrHdr+nAttrHdrSize, pAttrHdr->file.nSize);
                    WFSKrnExitOnError(TransStoreBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileBlkInfo.nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                }
                break;
            }
            case FILE_RESIZE_CATEGORY_SMALL_TO_MEDIUM:{
                // Copy file block data
                u32 nI;
                u32 nMediumBlkIdx = 0;
                u32 nSize = pAttrHdr->file.nSize;
                u8 *pDstBlkPtr = 0;
                WFSFileBlkPtr *pFileMediumBlkPtr = NULL;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        pFileMediumBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nMediumBlkIdx);
                        u16 nHashOfs = (u8*)&pFileMediumBlkPtr->hash - (u8*)pDi->pBlkHdr;
                        u32 nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileMediumBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    }
                    else{
                        pDstBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pSrcBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK,  pTba->pAreaInfo->nBlkSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK,  nSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileMediumBlkPtr->nBlkAdr, pTba->pTransInfo));
                        ++nMediumBlkIdx;
                    }
                }
                for(nI=0;nI<nMediumBlkIdx;nI++){
                    pFileMediumBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileMediumBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_SMALL_TO_LARGE:{
                // Copy file block data
                u32 nI;
                u32 nMaxHashBlkId = 0;
                u32 nSize = pAttrHdr->file.nSize;
                u8 *pDstBlkPtr = 0;
                WFSFileBlkInfo fileLargeBlkInfo;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        fileLargeBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nMaxHashBlkId);
                        u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                        u32 nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    }
                    else{
                        pDstBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pSrcBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK,  pTba->pAreaInfo->nBlkSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK,  nSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                        ++nMaxHashBlkId;
                    }
                }
                WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pNewAttrHdr, nNewAttrLog2Size, 0);
                for(nI=0;nI<(nMaxHashBlkId<<pTba->pAreaInfo->nLog2BlksPerMediumBlk);nI+=(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)){
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileLargeBlkPtr->nBlkAdr+nI, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_SMALL_TO_VERY_LARGE:{
                // Copy file block data
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr));
                u32 nI;
                u32 nMaxHashBlkId = 0;
                u32 nSize = pAttrHdr->file.nSize;
                u8 *pDstBlkPtr = 0;
                WFSFileBlkInfo fileLargeBlkInfo;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        fileLargeBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nMaxHashBlkId);
                        u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                        u32 nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK,  nValidSize, (void**)&pDstBlkPtr));
                    }
                    else{
                        pDstBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pSrcBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK, pTba->pAreaInfo->nBlkSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK, nSize, (void**)&pSrcBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                        ++nMaxHashBlkId;
                    }
                }
                WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, 0);
                for(nI=0;nI<(nMaxHashBlkId<<pTba->pAreaInfo->nLog2BlksPerMediumBlk);nI+=(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)){
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileLargeBlkPtr->nBlkAdr+nI, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_MEDIUM_TO_SMALL:{
                // Copy file block data
                u32 nI;
                u32 nMediumBlkIdx = 0;
                u32 nOldSize = pAttrHdr->file.nSize;
                u32 nSize = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                u8 *pSrcBlkPtr = 0;
                WFSFileBlkPtr *pFileMediumBlkPtr = NULL;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        pFileMediumBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nMediumBlkIdx);
                        u16 nHashOfs = (u8*)&pFileMediumBlkPtr->hash - (u8*)pDi->pBlkHdr;
                        u32 nValidSize = nOldSize < pTba->pAreaInfo->nMediumBlkSize ? nOldSize : pTba->pAreaInfo->nMediumBlkSize;
                        nOldSize -= pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileMediumBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    }
                    else{
                        pSrcBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pDstBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, pTba->pAreaInfo->nBlkSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileMediumBlkPtr->nBlkAdr, pTba->pTransInfo));
                        ++nMediumBlkIdx;
                    }
                }
                for(nI=0;nI<nNewNumBlks;nI++){
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2MediumBlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));

                break;
            }
            case FILE_RESIZE_CATEGORY_MEDIUM_TO_LARGE:{
                // Copy file block data
                u32 nI;
                u32 nSize = pAttrHdr->file.nSize;
                u32 nNumCopyBlks = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)+1 : 0;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    u8 *pDstBlkPtr, *pSrcBlkPtr;
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nI);
                    u32 nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                    nSize -= pTba->pAreaInfo->nMediumBlkSize;
                    u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nMediumBlkSize);
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                }
                for(nI=0;nI<nNumCopyBlks;nI++){
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2MediumBlkSize, nNumBlks, 
                                                  (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                                  -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_MEDIUM_TO_VERY_LARGE:{
                // Copy file block data
                u32 nSize = pAttrHdr->file.nSize;
                u32 nNumCopyBlks = pAttrHdr->file.nSize ? ((pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)+1 : 0;
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr));
                u32 nI;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    u8 *pDstBlkPtr, *pSrcBlkPtr;
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nI);
                    u32 nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                    nSize -= pTba->pAreaInfo->nMediumBlkSize;
                    u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nMediumBlkSize);
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                }
                for(nI=0;nI<nNumCopyBlks;nI++){
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2MediumBlkSize, nNumBlks, 
                                                  (WFSBlkAdr*)WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                                  -(s32)sizeof(WFSFileBlkPtr), pTba->pTransInfo));
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_LARGE_TO_SMALL:{
                // Copy file block data
                u32 nI;
                u32 nMaxHashBlkId = 0;
                u32 nOldSize = pAttrHdr->file.nSize;
                u32 nSize = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                u8 *pSrcBlkPtr = 0;
                WFSFileBlkInfo fileLargeBlkInfo;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        fileLargeBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nMaxHashBlkId);
                        u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                        u32 nValidSize = nOldSize < pTba->pAreaInfo->nMediumBlkSize ? nOldSize : pTba->pAreaInfo->nMediumBlkSize;
                        nOldSize -= pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    }
                    else{
                        pSrcBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pDstBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, pTba->pAreaInfo->nBlkSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                        ++nMaxHashBlkId;
                    }
                }
                for(nI=0;nI<nNewNumBlks;nI++){
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_LARGE_TO_MEDIUM:{
                // Copy file block data
                u32 nI;
                u32 nOldSize = pAttrHdr->file.nSize;
                u32 nSize = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                u32 nNumCopyBlks = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                if (nNumCopyBlks) {nNumCopyBlks = ((nNumCopyBlks-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)+1;}
                for(nI=0;nI<nNumCopyBlks;nI++){
                    u8 *pDstBlkPtr, *pSrcBlkPtr;
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge4(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    u32 nValidSize = nOldSize < pTba->pAreaInfo->nMediumBlkSize ? nOldSize : pTba->pAreaInfo->nMediumBlkSize;
                    nOldSize -= pTba->pAreaInfo->nMediumBlkSize;
                    u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nI);
                    nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                    nSize -= pTba->pAreaInfo->nMediumBlkSize;
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nMediumBlkSize);
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                }
                for(nI=0;nI<nNumCopyBlks;nI++){
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumBlks, (WFSBlkAdr*)WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, 0), 
                                   -(s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_LARGE_TO_VERY_LARGE:{
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pNewAttrHdr, nNewAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_WRITE_BLK, (void**)&pIndirectBlkPtr));
                WFSFileLargeBlkPtr *pSrcFileLargeBlkPtr = 0, *pDstFileLargeBlkPtr = 0;
                int nI, nJ;
                for(nI=0;nI<nNumBlks;nI++){
                    pSrcFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nI);
                    pDstFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nI);
                    *pDstFileLargeBlkPtr = *pSrcFileLargeBlkPtr;
                    for(nJ=0;nJ<(1<<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk);nJ++){
                        WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pDstFileLargeBlkPtr, nJ);
                        u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                        nResult = BCacheSetHash(pTba->pAreaInfo->pVolInfo, fileBlkInfo.nBlkAdr, *pIndirectBlkAdr, nHashOfs);
                        if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                            goto Exit;
                        }
                    }
                }
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_LARGE_TO_SMALL:{
                // Copy file block data
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                s32 nI;
                u32 nMaxHashBlkId = 0;
                u32 nOldSize = pAttrHdr->file.nSize;
                u32 nSize = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                u8 *pSrcBlkPtr = 0;
                WFSFileBlkInfo fileLargeBlkInfo;
                for(nI=0;nSize>0;nI++){
                    if(!(nI%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk))){
                        fileLargeBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nMaxHashBlkId);
                        u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                        u32 nValidSize = nOldSize < pTba->pAreaInfo->nMediumBlkSize ? nOldSize : pTba->pAreaInfo->nMediumBlkSize;
                        nOldSize -= pTba->pAreaInfo->nMediumBlkSize;
                        WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    }
                    else{
                        pSrcBlkPtr += pTba->pAreaInfo->nBlkSize;
                    }
                    u8 *pDstBlkPtr;
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    u16 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    if(nSize > pTba->pAreaInfo->nBlkSize){
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, pTba->pAreaInfo->nBlkSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nBlkSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize -= pTba->pAreaInfo->nBlkSize;
                    } else {
                        WFSKrnExitOnError(TransGetAndPinFileBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nSize, (void**)&pDstBlkPtr));
                        memcpy(pDstBlkPtr, pSrcBlkPtr, nSize);
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                        nSize = 0;
                    }
                    if(!((nI+1)%(1<<pTba->pAreaInfo->nLog2BlksPerMediumBlk)) || !nSize){
                        WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                        ++nMaxHashBlkId;
                    }
                }
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                for(nI=0;nI<nNewNumBlks;nI++){
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                u32 nNumIndirectBlks = (nNumBlks+pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                for(nI=0;nI<nNumIndirectBlks;nI++){
                    WFSKrnExitOnError(TransGetAndPinBlk5(pTba->pAreaInfo, pIndirectBlkAdr[-nI], pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                    if(nI+1==nNumIndirectBlks){
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, (nNumBlks-1)%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1,
                            (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                    } else {
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk,
                            (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                    }
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pIndirectBlkAdr[-nI], pTba->pTransInfo));
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumIndirectBlks, pIndirectBlkAdr, -(s32)sizeof(WFSBlkAdr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_LARGE_TO_MEDIUM:{
                // Copy file block data
                u32 nOldSize = pAttrHdr->file.nSize;
                u32 nSize = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                u32 nNumCopyBlks = pAttrHdr->file.nSize < nNewSize ? pAttrHdr->file.nSize : nNewSize;
                if (nNumCopyBlks) {nNumCopyBlks = ((nNumCopyBlks-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)+1;}
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                s32 nI;
                for(nI=0;nI<nNumCopyBlks;nI++){
                    u8 *pDstBlkPtr, *pSrcBlkPtr;
                    WFSFileBlkInfo fileLargeBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nI);
                    u32 nValidSize = nOldSize < pTba->pAreaInfo->nMediumBlkSize ? nOldSize : pTba->pAreaInfo->nMediumBlkSize;
                    nOldSize -= pTba->pAreaInfo->nMediumBlkSize;
                    u16 nHashOfs = (u8*)fileLargeBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileLargeBlkInfo.nBlkAdr, TRANS_FLAG_READ_BLK, nValidSize, (void**)&pSrcBlkPtr));
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nI);
                    nValidSize = nSize < pTba->pAreaInfo->nMediumBlkSize ? nSize : pTba->pAreaInfo->nMediumBlkSize;
                    nSize -= pTba->pAreaInfo->nMediumBlkSize;
                    nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransGetAndPinFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, TRANS_FLAG_WRITE_BLK, nValidSize, (void**)&pDstBlkPtr));
                    memcpy(pDstBlkPtr, pSrcBlkPtr, pTba->pAreaInfo->nMediumBlkSize);
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, fileLargeBlkInfo.nBlkAdr, pTba->pTransInfo));
                }
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                for(nI=0;nI<nNumCopyBlks;nI++){
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pNewAttrHdr, nNewAttrLog2Size, nI);
                    WFSKrnResult nResult = TransStoreBlk3(pTba->pAreaInfo, pFileBlkPtr->nBlkAdr, pTba->pTransInfo);
                    if(nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND){
                        goto Exit;
                    }
                }
                u32 nNumIndirectBlks = (nNumBlks+pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                for(nI=0;nI<nNumIndirectBlks;nI++){
                    WFSKrnExitOnError(TransGetAndPinBlk5(pTba->pAreaInfo, pIndirectBlkAdr[-nI], pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                    if(nI+1==nNumIndirectBlks){
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, (nNumBlks-1)%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1,
                            (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                    } else {
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk,
                            (WFSBlkAdr*)(pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr), pTba->pTransInfo));
                    }
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, pIndirectBlkAdr[-nI], pTba->pTransInfo));
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumIndirectBlks, pIndirectBlkAdr, -(s32)sizeof(WFSBlkAdr), pTba->pTransInfo));
                break;
            }
            case FILE_RESIZE_CATEGORY_VERY_LARGE_TO_LARGE:{
                u8 *pIndirectBlkPtr;
                WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, 0);
                WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                WFSFileLargeBlkPtr *pSrcFileLargeBlkPtr = 0, *pDstFileLargeBlkPtr = 0;
                int nI, nJ;
                for(nI=0;nI<nNewNumBlks;nI++){
                    pDstFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pNewAttrHdr, nNewAttrLog2Size, nI);                          
                    pSrcFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nI);
                    *pDstFileLargeBlkPtr = *pSrcFileLargeBlkPtr;
                    for(nJ=0;nJ<(1<<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk);nJ++){
                        WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pDstFileLargeBlkPtr, nJ);
                        u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                        nResult = BCacheSetHash(pTba->pAreaInfo->pVolInfo, fileBlkInfo.nBlkAdr, pTba->nBlkAdr, nHashOfs);
                        if (nResult != WFSKRN_RESULT_OK && nResult != WFSKRN_RESULT_BCACHE_NOT_FOUND) {
                            goto Exit;
                        }
                    }
                }
                WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                s32 nNumIndirectBlks = (nNumBlks-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1;
                pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nNumIndirectBlks-1);
                s32 nNumEndBlks  = (nNumBlks-1)%pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk+1;
                u32 nNumFreeBlks = nNumBlks-nNewNumBlks;
                while(nNumFreeBlks > 0){
                    u8 *pIndirectBlkPtr;
                    WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_READ_BLK, (void**)&pIndirectBlkPtr));
                    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nNumEndBlks-1);
                    if(nNumFreeBlks >= nNumEndBlks){
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumEndBlks, &pFileLargeBlkPtr->nBlkAdr, -(s32)nStride, pTba->pTransInfo));
                        nNumFreeBlks -= nNumEndBlks;
                    } else {
                        WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2LargeBlkSize, nNumFreeBlks, &pFileLargeBlkPtr->nBlkAdr, -(s32)nStride, pTba->pTransInfo));
                        nNumFreeBlks = 0;
                    }
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                    pIndirectBlkAdr = &pIndirectBlkAdr[1];
                    nNumEndBlks = pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                }
                WFSKrnExitOnError(TransFreeBlks6(pTba->pAreaInfo, pTba->pAreaInfo->ah.nLog2BlkSize, nNumIndirectBlks, &pIndirectBlkAdr[-1], -(s32)sizeof(WFSBlkAdr), pTba->pTransInfo));
                if(pAttrHdr->nAllocSize > nAllocSize && nNewNumBlks){
                    u32 nMedBlkIdx = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)&((1<<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk)-1);
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pDstFileLargeBlkPtr, nMedBlkIdx);
                    u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, pTba->pAreaInfo->nMediumBlkSize, 
                                                                 ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1));
                }
                break;
            }
            default:
                return WFSKRN_RESULT_NOT_IMPLEMENTED;
        }
        memcpy(pNewAttrHdr, pAttrHdr, nCopyForward);
    } else{
        if(pAttrHdr->file.nSize > nNewSize && nNewSize){
            switch(nNewSizeCategory){
                case FILE_SIZE_CATEGORY_SMALL:{
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks-1);
                    u32 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    u32 nOldBlkSize = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2BlkSize == (pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2BlkSize) ? 
                                      ((pAttrHdr->file.nSize-1)&(pTba->pAreaInfo->nBlkSize-1))+1 : pTba->pAreaInfo->nBlkSize;
                    WFSKrnExitOnError(TransTrancateFileBlk( pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, 
                                                            nOldBlkSize, ((nNewSize-1)&(pTba->pAreaInfo->nBlkSize-1))+1) );
                    break;
                }
                case FILE_SIZE_CATEGORY_MEDIUM:{
                    WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks-1);
                    u32 nHashOfs = (u8*)&pFileBlkPtr->hash - (u8*)pDi->pBlkHdr;
                    u32 nOldBlkSize = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize == (pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize) ?
                                      ((pAttrHdr->file.nSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1 : pTba->pAreaInfo->nMediumBlkSize;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, pFileBlkPtr->nBlkAdr, 
                                                                 nOldBlkSize, ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1));
                    break;
                }
                case FILE_SIZE_CATEGORY_LARGE:{
                    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nNewNumBlks-1);
                    u32 nMedBlkIdx = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)&((1<<pTba->pAreaInfo->nLog2MediumBlkPerLargeBlk)-1);
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pTba->pAreaInfo, pFileLargeBlkPtr, nMedBlkIdx);
                    u32 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pDi->pBlkHdr;
                    u32 nOldBlkSize = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize == (pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize) ?
                                      ((pAttrHdr->file.nSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1 : pTba->pAreaInfo->nMediumBlkSize;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, pTba->nBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, 
                                                                 nOldBlkSize, ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1));
                    break;
                }
                case FILE_SIZE_CATEGORY_VERY_LARGE:{
                    u8 *pIndirectBlkPtr;
                    u32 nTrancateSize = ((nNewSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1;
                    if(nTrancateSize == pTba->pAreaInfo->nMediumBlkSize){ break; }
                    u32 nIndirectBlkIdx = (nNewNumBlks-1)/pTba->pAreaInfo->nNumLargeBlkPtrsPerBlk;
                    u32 nMedBlkIdx = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize)%(pTba->pAreaInfo->nNumMediumBlkPtrsPerBlk);
                    WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pTba->pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nIndirectBlkIdx);
                    WFSKrnReturnOnError(TransGetAndPinBlk5(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo, TRANS_FLAG_RW_BLK, (void**)&pIndirectBlkPtr));
                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoFromIndirectBlkForCategoryVeryLarge(pTba->pAreaInfo, pIndirectBlkPtr, nMedBlkIdx);
                    u16 nHashOfs = (u8*)fileBlkInfo.pHash - (u8*)pIndirectBlkPtr;
                    u32 nOldBlkSize = ((nNewSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize == (pAttrHdr->file.nSize-1)>>pTba->pAreaInfo->ah.nLog2MediumBlkSize) ?
                                      ((pAttrHdr->file.nSize-1)&(pTba->pAreaInfo->nMediumBlkSize-1))+1 : pTba->pAreaInfo->nMediumBlkSize;
                    WFSKrnExitOnError(TransTrancateFileMediumBlk(pTba, *pIndirectBlkAdr, nHashOfs, fileBlkInfo.nBlkAdr, nOldBlkSize, nTrancateSize));
                    WFSKrnExitOnError(TransUnpinBlk3(pTba->pAreaInfo, *pIndirectBlkAdr, pTba->pTransInfo));
                    break;
                }
            }            
        }
    }

    // Update the attributes data.
    if(nAllocSize == pAttrHdr->nAllocSize){
        pAttrHdr->file.nSize = nNewSize;
    } else {
        pNewAttrHdr->nAllocSize = nAllocSize;
        pNewAttrHdr->file.nSize = nNewSize;
        pNewAttrHdr->nAttrLog2Size = nNewAttrLog2Size;
        pNewAttrHdr->nSizeCategory = nNewSizeCategory;
        if(pNewAttrHdr != pAttrHdr){
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pDi->pBlkHdr, *pDi->dni.pOfs, pDi->dni.nLog2Size);
            pDi->pAttrHdr = pNewAttrHdr;
            pDi->dni.nLog2Size = nNewAttrLog2Size;
            *pDi->dni.pOfs = nNewAttrOfs;
            dbg(DirLeafCheckBlk(pDi));
        }
    }
    return WFSKRN_RESULT_OK;

Exit:;
    if(pNewAttrHdr != pAttrHdr){
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pDi->pBlkHdr, nNewAttrOfs, nNewAttrLog2Size);
        dbg(DirLeafCheckBlk(pDi));
    }

    return nResult;

}


WFSKrnResult DirInsertRaw(WFSFileName *pName, DirItr *pDi, u32 nAllocSize) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=0x76){
        int a;a=0;
    }
#endif
    // Inserts a new name into the directory
    // pDi->tba.pAreaInfo    [In] .. pointer to area info which contains details of are location on disk
    // pDi->tba.nBlkAdr      [In] .. logical block address relative to the start of the area
    // pDi->tba.pTransInfo    [In] .. index specifying the transaction

    // Specifies a block as a combination of pAreaInfo, relative
    // ToDo: Make sorted sequential insertion give most compact packing. Currently gives least compact packing!
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif

   // WFSBlkAdr nDirRootBlkAdr = pDi->tba.nBlkAdr;
    u8 nSizeCategory;
    DirEntryAttrHdr *pAttrHdr;
    u8 nAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2SizeUpdateAllocSize(pDi->tba.pAreaInfo, &nAllocSize, pName->nLen, &nSizeCategory);
    WFSKrnResult nResult = DirSbaAllocAndRegistSubBlk(pName, pDi, nAttrLog2Size, &pAttrHdr);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }

    pAttrHdr->nAccessListIdx = 0;
    pAttrHdr->nAllocSize = nAllocSize;
    pAttrHdr->nNameLen  = pName->nLen;
    pAttrHdr->nAttrLog2Size = nAttrLog2Size;
    pAttrHdr->nSizeCategory = nSizeCategory;

    switch(nSizeCategory) {
        case FILE_SIZE_CATEGORY_VERY_SMALL: {
    #if _DEBUG_DIR
            u8 nTestSizeCategory;
            u8 nTestAttrLog2Size = (u8)DirCalculateAttrSubBlkLog2Size(pDi->tba.pAreaInfo, nAllocSize, pName->nLen, &nTestSizeCategory);
            if ((nTestAttrLog2Size != nAttrLog2Size) || (nTestSizeCategory != nTestSizeCategory)) {
                WFSKrnOutputErrorStr("Error: AllocSize does not fit in allocated block");
            }
            nTestAttrLog2Size = (u16)DirCalculateAttrSubBlkLog2Size(pDi->tba.pAreaInfo, nAllocSize+1, pName->nLen, &nTestSizeCategory);
            if ((nTestAttrLog2Size == nAttrLog2Size) && (nTestSizeCategory == nTestSizeCategory)) {
                WFSKrnOutputErrorStr("Error: AllocSize+1 fits in allocated block");
            }
    #endif
            nResult = WFSKRN_RESULT_OK;
            break;
        }
        case FILE_SIZE_CATEGORY_SMALL: {
            u32 nNumFileBlks = nAllocSize ? ((nAllocSize-1)>>pDi->tba.pAreaInfo->ah.nLog2BlkSize)+1 : 0;
            // Set 0 to the all of WFSFileBlkPtr.
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumFileBlks-1);
            memset(pFileBlkPtr, 0, nNumFileBlks*sizeof(WFSFileBlkPtr));
            // Alloc new block address for UserData.
            nResult = AreaAllocBlks(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pAreaInfo->ah.nLog2BlkSize, 
                                    nNumFileBlks, &pFileBlkPtr[nNumFileBlks-1].nBlkAdr, -(signed)sizeof(WFSFileBlkPtr));
            // Sort the allocated block addresses into reverse order (since we access the WFSFileBlkPtr in reverse order).
            _qsort(&pFileBlkPtr->nBlkAdr, nNumFileBlks, sizeof(WFSFileBlkPtr), CompareBlkAdrReverse);
            break;
        }
        case FILE_SIZE_CATEGORY_MEDIUM: {
            u32 nNumFileMediumBlks = nAllocSize ? ((nAllocSize-1)>>pDi->tba.pAreaInfo->ah.nLog2MediumBlkSize)+1 : 0;
            // Set 0 to the all of WFSFileBlkPtr.
            WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumFileMediumBlks-1);
            memset(pFileBlkPtr, 0, nNumFileMediumBlks*sizeof(WFSFileBlkPtr));
            // Alloc new block address for UserData.
            nResult = AreaAllocBlks(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pAreaInfo->ah.nLog2MediumBlkSize, 
                                    nNumFileMediumBlks, &pFileBlkPtr[nNumFileMediumBlks-1].nBlkAdr, -(signed)sizeof(WFSFileBlkPtr));
            // Sort the allocated block addresses into reverse order (since we access the WFSFileBlkPtr in reverse order).
            _qsort(&pFileBlkPtr->nBlkAdr, nNumFileMediumBlks, sizeof(WFSFileBlkPtr), CompareBlkAdrReverse);
            break;
        }
        case FILE_SIZE_CATEGORY_LARGE: {
            u32 nNumFileLargeBlks = nAllocSize ? ((nAllocSize-1)>>pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize)+1 : 0;
            // Set 0 to the all of WFSFileBlkPtr.
            WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pAttrHdr, pAttrHdr->nAttrLog2Size, nNumFileLargeBlks-1);
            memset(pFileLargeBlkPtr, 0, nNumFileLargeBlks*sizeof(WFSFileLargeBlkPtr));
            // Alloc new block address for UserData.
            nResult = AreaAllocBlks(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize, 
                                    nNumFileLargeBlks, &pFileLargeBlkPtr[nNumFileLargeBlks-1].nBlkAdr, -(signed)sizeof(WFSFileLargeBlkPtr));
            // Sort the allocated block addresses into reverse order (since we access the WFSFileBlkPtr in reverse order).
            _qsort(&pFileLargeBlkPtr->nBlkAdr, nNumFileLargeBlks, sizeof(WFSFileLargeBlkPtr), CompareBlkAdrReverse);
            break;
        }
        case FILE_SIZE_CATEGORY_VERY_LARGE: {
            u32 nNumFileLargeBlks = nAllocSize ? ((nAllocSize-1)>>pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize)+1 : 0;
            u32 nNumIndirectBlks = (nNumFileLargeBlks+pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk-1)/pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk;
            // Set 0 to the all of indirect block addresses.
            WFSBlkAdr *pIndirectBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pDi->tba.pAreaInfo, pAttrHdr, pAttrHdr->nAttrLog2Size, nNumIndirectBlks-1);
            //memset(pIndirectBlkAdr, 0, nNumIndirectBlks*sizeof(WFSBlkAdr));
            pIndirectBlkAdr = &pIndirectBlkAdr[nNumIndirectBlks-1];
            // Alloc new block addresses for indirect blocks.
            nResult = AreaAllocBlks(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pAreaInfo->ah.nLog2BlkSize, nNumIndirectBlks, pIndirectBlkAdr, -(s32)sizeof(WFSBlkAdr));
            // Sort the allocated block addresses into reverse order (since we access the WFSFileBlkPtr in reverse order).
            _qsort(&pIndirectBlkAdr[1-nNumIndirectBlks], nNumIndirectBlks, sizeof(WFSBlkAdr), CompareBlkAdrReverse);
            // Alloc new block addresses for file data blocks.
            u32 nI, nRefBlkAdr = *pIndirectBlkAdr;
            for(nI=0;nI<nNumIndirectBlks;nI++){
                u8 *pIndirectBlkPtr;
                // Map the indirect block to the cache.
                WFSKrnReturnOnError(TransGetAndPinBlk5(pDi->tba.pAreaInfo, *pIndirectBlkAdr, pDi->tba.pTransInfo, TRANS_FLAG_WRITE_BLK, (void**)&pIndirectBlkPtr));
                // Set 0 to the all of indirect block addresses.
                // Alloc new block address for indirect blocks.
                u32 nNumAllocLargeBlks = (nI+1 == nNumIndirectBlks) ? (nNumFileLargeBlks-1)%pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk+1 : pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk;
                memset((u8*)pIndirectBlkPtr+sizeof(WFSMetaDataHdr), 0, nNumAllocLargeBlks*sizeof(WFSFileLargeBlkPtr));
                nResult = AreaAllocBlks(pDi->tba.pAreaInfo, nRefBlkAdr, pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize, nNumAllocLargeBlks, 
                                        (WFSBlkAdr*)((u8*)pIndirectBlkPtr+sizeof(WFSMetaDataHdr)), (s32)sizeof(WFSFileLargeBlkPtr));
                // Sort the allocated block addresses into reverse order (since we access the WFSFileBlkPtr in reverse order).
                _qsort((u8*)pIndirectBlkPtr+sizeof(WFSMetaDataHdr), nNumAllocLargeBlks, sizeof(WFSFileLargeBlkPtr), CompareBlkAdrForward);
                // Update the reference block address
                WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectBlkPtr, nNumAllocLargeBlks-1);
                nRefBlkAdr = pFileLargeBlkPtr->nBlkAdr;
                // Unpin the indirect block
                WFSKrnReturnOnError(TransUnpinBlk3(pDi->tba.pAreaInfo, *pIndirectBlkAdr, pDi->tba.pTransInfo));
                pIndirectBlkAdr = &pIndirectBlkAdr[-1];
            }
            break;
        }
    }

    dbg(DirLeafCheckBlk(pDi));

    return nResult;
}


WFSKrnResult DirDeleteEntry(DirItr *pDi) {
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        DirLeafCheckBlk(pDi);
    }
#endif
    DirBlkHdr *pBlkHdr;
    WFSKrnResult nResult = DirGetBlk(pDi, TRANS_FLAG_RW_BLK, &pBlkHdr);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, *pDi->dni.pOfs, pDi->pAttrHdr->nAttrLog2Size);
    --pBlkHdr->nNumRecs;
    // Backtrack within leaf block
    while(1) {
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDi->dni.pDns->nSubBlkOfs);
        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        pDi->dni.nEntryIdx = pDi->dni.pDns->nEntryIdx;
        if (pDi->dni.pSubBlkHdr->nNumEntries!=1) {
            break;
        }
        // This is the only entry in the sub-block. Delete it, and delete the corresponding entry from its parent.
        // If this leaves the parent empty, delete the parent and repeat with earlier ancestors.
        SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size);
        if (pDi->dni.pDns->nSubBlkOfs == pBlkHdr->nRootOfs) {
            // This was the last sub block of the block
            goto DeleteFromNode;
        }
        DirPopNodeStackEntry(&pDi->dni);
    }
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
    u16 *pSrcStart = DirLeafGetIntraBlkOfsPtr_ote(pDi->dni.pSubBlkHdr->nNumEntries, pExistingNodeOfsTblEnd);
    u16 *pSrcEnd = DirLeafGetIntraBlkOfsPtr_ote(pDi->dni.nEntryIdx, pExistingNodeOfsTblEnd);
    // Delete the choice character from the original block
    utf8 *pSrc = &pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.nEntryIdx];
    memmove(pSrc-1, pSrc, pDi->dni.pSubBlkHdr->nNumEntries - pDi->dni.nEntryIdx);
    --pDi->dni.pSubBlkHdr->nNumEntries;
    u32 nUpdatedNodeLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    if (nUpdatedNodeLog2Size != pDi->dni.nLog2Size) {
        u32 nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
        if (nUpdatedNodeOfs == 0) {
            // Could not allocate a new block, so resize in place
            u16 *pDstStart = DirLeafGetIntraBlkOfsPtr_l2s(pDi->dni.pSubBlkHdr->nNumEntries, pDi->dni.pSubBlkHdr, nUpdatedNodeLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memmove(pDstStart, pSrcStart, nCopySize2);
            if (pDi->dni.nEntryIdx != 1) {
                pDstStart = (u16*)((u8*)pDstStart + nCopySize2);
                pSrcStart = &pSrcEnd[1];
                memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            }
            SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size, nUpdatedNodeLog2Size);
        } else {
            // Delete offset pointer while copying to a new sub block
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.pSubBlkHdr->nNumEntries;
            memcpy(pUpdatedNodeHdr, pDi->dni.pSubBlkHdr, nCopySize1);
            u16 *pDstStart = DirLeafGetIntraBlkOfsPtr_l2s(pDi->dni.pSubBlkHdr->nNumEntries, pUpdatedNodeHdr, nUpdatedNodeLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memcpy(pDstStart, pSrcStart, nCopySize2);
            pDstStart = (u16*)((u8*)pDstStart + nCopySize2);
            pSrcStart = &pSrcEnd[1];
            memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            // Update parent to point to the new node. Parent may be the nRootOfs of the block
            SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size);
            DirLeafUpdateParentOfsPtr(pDi, nUpdatedNodeOfs);
            pDi->dni.pSubBlkHdr = pUpdatedNodeHdr;
        }
    } else {
        // Delete entry from existing node
        u16 *pDstStart = DirLeafGetIntraBlkOfsPtr_ote(pDi->dni.pSubBlkHdr->nNumEntries, pExistingNodeOfsTblEnd);
        u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
        memmove(pDstStart, pSrcStart, nCopySize2);
    }
    if ((pDi->dni.pSubBlkHdr->nNumEntries==1) && (pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] != 0)) {
        // The sub block has been left with one entry after deletion.
        // If it has a child pointer, it must be concatenated with the child.
        u16 nNewOfs = DirLeafMergeSubBlkWithItsSingleChild(pDi, pBlkHdr, pDi->dni.pDns->nSubBlkOfs);
        if (nNewOfs == 0) {
            // ToDo: This case may need to be handled properly
            WFSKrnOutputErrorStr("No space to merge sub-blocks!");
        }
        DirLeafUpdateParentOfsPtr(pDi, nNewOfs);
    }
    dbg(DirLeafCheckBlk(pDi));
    goto Done;

    // Backtrack within node blocks
    while(1) {
        pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDi->dni.pDns->nSubBlkOfs);
        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        pDi->dni.nEntryIdx = pDi->dni.pDns->nEntryIdx;
        if (pDi->dni.pSubBlkHdr->nNumEntries!=1) {
            break;
        }
        // This is the only entry in the sub-block. Delete it, and delete the corresponding entry from its parent.
        // If this leaves the parent empty, delete the parent and repeat with earlier ancestors.
        SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size);
        //if (pDi->dni.pOfs == &pBlkHdr->nRootOfs) {
        if (pDi->dni.pDns->nSubBlkOfs == pBlkHdr->nRootOfs) {
            // This was the last sub block of the block
            // If a parent node exists, we should delete this block, and proceed to remove the splitter from the parent.
            // Otherwise, if this is the only block in the directory, we should leave it in the empty state.
DeleteFromNode:
            if (pDi->dni.nDnsDepth==2) {
                DirReinitBlk(pDi->tba.pAreaInfo, pBlkHdr, (WFSMetaDataFlags)(pBlkHdr->sbah.mdh.nFlags | WFS_MDF_DIR_LEAF_BLK));
                //DirCreateAndInitNewBlk(&pDi->tba, (WFSMetaDataFlags)(pBlkHdr->sbah.mdh.nFlags | WFS_MDF_DIR_LEAF_BLK));
                DirBlkCreateRootNode(pBlkHdr);
                dbg(DirNodeCheckBlk(pDi));
                goto Done;
            }
            --pDi->nBlkDepth;
            DirUnpinAndFreeBlk(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pTransInfo);
            pDi->tba.nBlkAdr = pDi->dni.pDns->pParent->nBlkAdr;
          osTPrintf( "%s, %d : %d, %d\n", __FUNCTION__, __LINE__, pDi->tba.nBlkAdr, pDi->dni.pDns->pParent->nBlkAdr);
            WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_RW_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
            dbg(DirNodeCheckBlk(pDi));

            // Deletion sometimes requires the used space in a node to increase.
            // This is due to the constraint that nodes with exactly one child must be merged with their child.
            // In such cases the merged node may require more space than the total of the pre-merged nodes beacuse
            // it must be allocated as a contiguous power-of-two sized sub block.
            // We must detect cases where the deletion would cause the used space to increase beyond the size of
            // the node, and split the node in advance.

            u32 nExtraSize = 0;
            DirNodeStack *pDns = pDi->dni.pDns->pParent;
            DirSubBlkHdr *pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDns->nSubBlkOfs);
            if (pSubBlkHdr->nNumEntries == 1) {
                pDns = pDns->pParent;
                pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDns->nSubBlkOfs);
            }
            if (pDns->nBlkAdr == pDi->tba.nBlkAdr) {
                if (pSubBlkHdr->nNumEntries == 2) {
                    if ((pDns->nEntryIdx == 1) || (pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen] != 0)) {
                        u32 nLog2Size = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
                        u16 nOfs = DirNodeGetIntraBlkOfs(pDns->nEntryIdx^3, pSubBlkHdr);
                        pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
                        u32 nLog2Size2 = DirNodeGetSubBlkLog2Size(pSubBlkHdr);
                        nExtraSize = (u32)(- (1<<nLog2Size2) - (1<<nLog2Size));
                        if (nLog2Size2 < nLog2Size) {
                            nLog2Size2 = nLog2Size;
                        }
                        nExtraSize += 1<<(nLog2Size2+1);
                    }
                }
            }
            u32 nTotalFree = SbaCountFreeSpace((WFSSubBlkAllocHdr *)pBlkHdr);
            if (nExtraSize > nTotalFree) {
                DirPopNodeStackEntry(&pDi->dni);
                u32 nFirstBlockToFree = 0;
                WFSBlkAdr aDstBlkAdr[DIR_MAX_SPLIT_BLK_ALLOC-2];
                pDi->tba.nBlkAdr = pDi->dni.pDns->nBlkAdr;
                WFSKrnReturnOnError(AreaAllocBlks(pDi->tba.pAreaInfo, pDi->tba.nBlkAdr, pDi->tba.pAreaInfo->ah.nLog2BlkSize, DIR_MAX_SPLIT_BLK_ALLOC-2, aDstBlkAdr, sizeof(WFSBlkAdr)));
                WFSKrnReturnOnError(DirNodeBlkSplit(pDi, aDstBlkAdr, &nFirstBlockToFree));
                u32 nFindMaxDepth = pDi->nBlkDepth;   // Note: This must be after the call to DirNodeBlkSplit
                while(pDi->dni.nDnsDepth > 2) {
                    DirPopNodeStackEntry(&pDi->dni);
                }
                pDi->tba.nBlkAdr = pDi->dni.pDns->nBlkAdr;
                DirItrClose(pDi);
                pDi->nBlkDepth = 0;
                nResult = DirFind(&pDi->name, pDi, nFindMaxDepth+1, DIR_FIND_MODE_NODE_SEARCH);
                if (nResult != WFSKRN_RESULT_OK) {
                    return nResult;
                }
                pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                dbg(if (*pDi->dni.pOfs != ((u8*)pDi->dni.pSubBlkHdr - (u8*)pBlkHdr)) {
                    WFSKrnOutputErrorStr("Ofs mismatch error!");
                })
                --pBlkHdr->nNumRecs;
                continue;
            }
            --pBlkHdr->nNumRecs;
        }
        DirPopNodeStackEntry(&pDi->dni);
    }
    // Calculate srouce pointer location first (because it would change if we are deleting choice 0)
    pExistingNodeOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
    if ((pDi->dni.pDns->nSubBlkOfs == pBlkHdr->nRootOfs) && (pDi->dni.nEntryIdx==1)) {
        if (pDi->dni.pSubBlkHdr->aFirstChar[0]!=0) {
            WFSKrnOutputErrorStr("First entry of node non-zero!");
        }
        // Instead of deleting the "" entry, then renaming the next entry to be "",
        // we can make the "" entry point to same as the next entry, and delete the next entry.
        // This method would have less problems with handling almost-full nodes.
        pDi->dni.pDns->nEntryIdx = 2;
        u16 nOfs = DirNodeGetIntraBlkOfs_ote(pDi->dni.pDns->nEntryIdx, pDi->dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
        u16 *pOfsTblEnd;
        while(1) {
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
            pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
            pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
            WFSKrnReturnOnError(DirNodeStackGetEntry(&pDi->dni));
            pDi->dni.pDns->nEntryIdx = 1;
            pDi->dni.pDns->nSubBlkOfs = nOfs;
            pDi->dni.pDns->nBlkAdr = pDi->tba.nBlkAdr;
            if (pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] == 0) {
                break;
            }
            nOfs = DirNodeGetIntraBlkOfs_ote(1, pDi->dni.pSubBlkHdr, pOfsTblEnd);
        }
        // We have found the end of the splitter. Record the child block off.
        ((WFSBlkAdr*)pExistingNodeOfsTblEnd)[-1] = DirNodeGetBlkAdr_ote(pDi->dni.pSubBlkHdr, pOfsTblEnd);
        if (pDi->dni.pSubBlkHdr->nNumEntries==1) {
            SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size);
            DirPopNodeStackEntry(&pDi->dni);
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDi->dni.pDns->nSubBlkOfs);
            pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
            pDi->dni.nEntryIdx = pDi->dni.pDns->nEntryIdx;
        }
        pExistingNodeOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
    }
    pSrcStart = DirNodeGetIntraBlkOfsPtr_ote(pDi->dni.pSubBlkHdr->nNumEntries, pDi->dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
    pSrcEnd = DirNodeGetIntraBlkOfsPtr_ote(pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
    // Delete the choice character from the original block
    pSrc = &pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.nEntryIdx];
    memmove(pSrc-1, pSrc, pDi->dni.pSubBlkHdr->nNumEntries - pDi->dni.nEntryIdx);
    --pDi->dni.pSubBlkHdr->nNumEntries;
    nUpdatedNodeLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    if (nUpdatedNodeLog2Size != pDi->dni.nLog2Size) {
        u32 nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
        if (nUpdatedNodeOfs == 0) {
            // Could not allocate a new block, so resize in place
            u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(pDi->dni.pSubBlkHdr->nNumEntries, pDi->dni.pSubBlkHdr, nUpdatedNodeLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memmove(pDstStart, pSrcStart, nCopySize2);
            if (pDi->dni.nEntryIdx != 1) {
                pDstStart = (u16*)((u8*)pDstStart + nCopySize2);
                pSrcStart = &pSrcEnd[1];
                memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            }
            SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size, nUpdatedNodeLog2Size);
        } else {
            // Delete offset pointer while copying to a new sub block
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.pSubBlkHdr->nNumEntries;
            memcpy(pUpdatedNodeHdr, pDi->dni.pSubBlkHdr, nCopySize1);
            u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_l2s(pDi->dni.pSubBlkHdr->nNumEntries, pUpdatedNodeHdr, nUpdatedNodeLog2Size);
            u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
            memcpy(pDstStart, pSrcStart, nCopySize2);
            if (pDi->dni.nEntryIdx != 1) {
                pDstStart = (u16*)((u8*)pDstStart + nCopySize2);
                pSrcStart = &pSrcEnd[1];
                memcpy(pDstStart, pSrcStart, (u8*)pExistingNodeOfsTblEnd - (u8*)pSrcStart);
            }
            // Update parent to point to the new node. Parent may be the nRootOfs of the block
            SbaFreeSubBlk((WFSSubBlkAllocHdr*)pBlkHdr, pDi->dni.pDns->nSubBlkOfs, pDi->dni.nLog2Size);
            DirNodeUpdateParentOfsPtr(pDi, nUpdatedNodeOfs);
#if _DEBUG
            if (pBlkHdr != pDi->pBlkHdr) {
                WFSKrnOutputErrorStr("pDi->pBlkHdr changed");
            }
#endif
            pDi->dni.pSubBlkHdr = pUpdatedNodeHdr;
        }
    } else {
        // Delete entry from existing node
        u16 *pDstStart = DirNodeGetIntraBlkOfsPtr_ote(pDi->dni.pSubBlkHdr->nNumEntries, pDi->dni.pSubBlkHdr, pExistingNodeOfsTblEnd);
        u32 nCopySize2 = (u8*)pSrcEnd - (u8*)pSrcStart;
        memmove(pDstStart, pSrcStart, nCopySize2);
    }
    if ((pDi->dni.pSubBlkHdr->nNumEntries==1) && (pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] != 0)) {
        // The sub block has been left with one entry after deletion.
        // If it has a child pointer, it must be concatenated with the child.
        u16 nNewOfs = DirNodeMergeSubBlkWithItsSingleChild(pDi, pBlkHdr, pDi->dni.pDns->nSubBlkOfs);
        if (nNewOfs == 0) {
            // ToDo: This case may need to be handled properly
            WFSKrnOutputErrorStr("No space to merge sub-blocks!");
        }
        DirNodeUpdateParentOfsPtr(pDi, nNewOfs);
#if _DEBUG
            if (pBlkHdr != pDi->pBlkHdr) {
                WFSKrnOutputErrorStr("pDi->pBlkHdr changed");
            }
#endif
    }
    DirSubBlkHdr *pRootSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pBlkHdr->nRootOfs);
    if (pRootSubBlkHdr->aFirstChar[0] != 0) {   // Note: Testing first char, not first choice since there must be a "" entry in every node
        nResult = DirNodeRenameFirstRec(pBlkHdr);
        if (nResult != WFSKRN_RESULT_OK) {
            WFSKrnOutputErrorCode(nResult);
        }
    }
    dbg(DirNodeCheckBlk(pDi));
Done:
    return nResult;
}


WFSKrnResult DirSbaAllocAndRegistSubBlk(WFSFileName *pName, DirItr *pDi, u32 nAttrLog2Size, DirEntryAttrHdr **ppNewAttrHdr){

    WFSBlkAdr nDirRootBlkAdr = pDi->tba.nBlkAdr;
Retry:
    pDi->tba.nBlkAdr = nDirRootBlkAdr;
    pDi->nBlkDepth = 0;
    WFSKrnResult nResult = DirFind(pName, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
    if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
        // File name already exists
        return WFSKRN_RESULT_ALREADY_EXISTS;
    }
    else if ((nResult == WFSKRN_RESULT_MEDIA_ERROR) || (nResult == WFSKRN_RESULT_CORRUPTION)) {
        return nResult;
    }
    DirBlkHdr *pBlkHdr;
    WFSKrnReturnOnError(DirGetBlk(pDi, TRANS_FLAG_RW_BLK, &pBlkHdr));
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
    // Format of name node sub blocks:
    // [ StrLen, NumEntries, Char0, Char1, .. Char(StrLen-1), Choice1, .. Choice N, |<--gap-->|, Link N, Link N-1, .. Link 1 ]
    // [ StrLen, NumEntries, Char0, Char1, .. Char(StrLen-1), Choice1, .. Choice N, |<--gap-->|, Link N, Link N-1, .. Link 1 ]
    u16 nAttrOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nAttrLog2Size);
    if (nAttrOfs == 0) {
AttrNotAllocated:
        WFSKrnReturnOnError(DirLeafBlkSplit(pDi));
        DirItrCloseNoUnpin(pDi);
        goto Retry;
    }
    DirEntryAttrHdr *pAttrHdr= pDi->pAttrHdr = (DirEntryAttrHdr *)((u8*)pBlkHdr + nAttrOfs);
    // Some fields of the attributes block must be initialized here incase a split is required, which needs the size of
    // the attribute block to be calculatable. Also, making sure to overwrite the start of the attributes block so that
    // it no longer contains the magic number that is used to recognize free blocks in "wfskrn_SubBlkAlloc.cpp".

    u32 nSuffixLen = (u32)(pName->nLen - (pDi->dni.pStrPtr - pName->sStr));
    u32 nLog2Size;
    if (pBlkHdr->nNumRecs==0) {
        // No records in this directory yet. Add the first record.
        pDi->dni.pSubBlkHdr->nStrLen = pName->nLen;
        pDi->dni.pSubBlkHdr->nNumEntries = 1;
        nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
        //SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pBlkHdr, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size, nLog2Size);
        if (nLog2Size != pDi->dni.nLog2Size) {
            u32 nOfs = (u32)((u8*)pDi->dni.pSubBlkHdr - (u8*)pBlkHdr);
            dbg(SbaCheckOfs(nOfs));
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOfs, pDi->dni.nLog2Size);
            pDi->dni.nLog2Size = nLog2Size;
            nOfs = *pDi->dni.pOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, pDi->dni.nLog2Size);
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nOfs);
        }
        pDi->dni.pSubBlkHdr->nStrLen = pName->nLen; //(u8)(pName->nLen - (pDi->dni.pStrPtr - pName->sStr));
#if _DEBUG_DIR
        if (pDi->dni.pStrPtr != pName->sStr) {
            WFSKrnOutputErrorStr("Expected suffix == full name");
        }
#endif
        pDi->dni.pSubBlkHdr->nNumEntries = 1;
        memcpy(pDi->dni.pSubBlkHdr->aFirstChar, pDi->dni.pStrPtr, (size_t)(pDi->dni.pSubBlkHdr->nStrLen+1));
        u16 *pOfsTblEnd = (u16*)((u8*)pDi->dni.pSubBlkHdr + (1<<pDi->dni.nLog2Size));
        pDi->dni.pOfs = &pOfsTblEnd[-1];
        *pDi->dni.pOfs = nAttrOfs;
        ++pBlkHdr->nNumRecs;
        //DirNodeStackFree(&di);
        pDi->dni.nLog2Size = nAttrLog2Size;
        goto Exit;
        //return WFSKRN_RESULT_OK;
    }

    // Insert the new name at the location found
    //
    // To insert a new file name into a directory block requires to allocate up to two new sub blocks
    // for name nodes, resize an existing name node, and add a new sub block for file attributes and
    // block pointers. It needs to be determined up-front whether this will require splitting the block.
    //
    // Cases:                     nNodeNumLeft   nNumMatching   nSuffixLen   Result
    // "",?      + "zyx",0     ->      0              0              3          "",z("yx",0),?
    // "",?      + "",0        ->      0              0              0          "",0,?
    // "abc",?   + "abc",0     ->      0              3              0          "abc",0,?
    // "abc",?   + "abczyx",0  ->      0              3              3          "abc",z("yx",0),?
    // "abc",0,? + "abczyx",0  ->      0              3              3          "abc",0,z("yx",0),?
    // "abc",?   + "ab",0      ->      1              2              0          "ab",0,c("",?)
    // "abc",?   + "abzyx",0   ->      1              2              3          "ab",c("",?),z("yx",0)
    // "abc",?   + "a",0       ->      2              1              0          "a",0,b("c",?)
    // "abc",?   + "azyx",0    ->      2              1              3          "a",b("c",?),z("yx",0)
    // "abc",?   + "zyx",0     ->      3              0              3          "",a("bc",?),z("yx",0)

    u32 nNumMatching = (u32)(pDi->dni.pNodeStrPtr - pDi->dni.pSubBlkHdr->aFirstChar);
    u32 nNodeNumLeft = pDi->dni.pSubBlkHdr->nStrLen - nNumMatching;
    u16 nNameSubBlkLog2Size = 0;
    u16 nNameSubBlkOfs;
    u16 nChoiceNodeSubBlkLog2Size;
    u16 nChoiceNodeSubBlkOfs;
    u16 *pNodeOfsTblEnd;
    u16 *pAttrOfs = 0;

    if (nSuffixLen) {
        // allocate a sub block for the new name entry
        nNameSubBlkLog2Size = (u16)DirLeafGetSubBlkLog2Size_2(nSuffixLen-1, 1);
        nNameSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkLog2Size);
        if (nNameSubBlkOfs == 0) {
NameNotAllocated:
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nAttrOfs, nAttrLog2Size);
            goto AttrNotAllocated;
        }
        DirSubBlkHdr *pNameSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nNameSubBlkOfs);
        u16 *pNameOfsTblEnd = (u16*)((u8*)pNameSubBlkHdr + (1<<nNameSubBlkLog2Size));
        pNameSubBlkHdr->nNumEntries = 1;
        pNameSubBlkHdr->nStrLen = (u8)(nSuffixLen-1);
        memcpy(pNameSubBlkHdr->aFirstChar, &pDi->dni.pStrPtr[1], nSuffixLen);  // includes copying the NULL terminator
        pAttrOfs = &pNameOfsTblEnd[-1];
        pNameOfsTblEnd[-1] = nAttrOfs;
    } else {
        nNameSubBlkOfs = nAttrOfs;
    }

    if (nNodeNumLeft == 0) {
        // Add a new option to the existing node
        u16 nUpdatedNodeOfs;
        u16 nUpdatedNodeLog2Size = (u16)DirLeafGetSubBlkLog2Size_2(pDi->dni.pSubBlkHdr->nStrLen, (u32)(pDi->dni.pSubBlkHdr->nNumEntries+1));
        if (nUpdatedNodeLog2Size != pDi->dni.nLog2Size) {
            // The node needs to be copied to a larger sub block
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            if (nUpdatedNodeOfs == 0) {
                // Allocation failed, so undo any changes so far, split the block, and try again
                if (nSuffixLen) {
                    // Undo the allocation of the name sub block
                    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkOfs, nNameSubBlkLog2Size);
                }
                goto NameNotAllocated;
            }
            DirSubBlkHdr *pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
            u16 *pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
            // Copy the old choices to the new sub block while inserting the new choice
            u32 nCopySize1 = sizeof(DirSubBlkHdr)-1 + pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.nEntryIdx;
            memcpy(pUpdatedNodeHdr, pDi->dni.pSubBlkHdr, nCopySize1);
            u8 *pDst = &((u8*)pUpdatedNodeHdr)[nCopySize1];
            *pDst++ = (u8)pDi->dni.pStrPtr[0];
            s32 nCopySize2 = (s32)(pDi->dni.pSubBlkHdr->nNumEntries - pDi->dni.nEntryIdx);
            memcpy(pDst, &((u8*)pDi->dni.pSubBlkHdr)[nCopySize1], (u32)nCopySize2);
            ++pUpdatedNodeHdr->nNumEntries;
            // Copy the table of child offset pointers from the old block to the new block while inserting the new offset
            u16 *pDst2 = &pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries];
            memcpy(pDst2, &pExistingNodeOfsTblEnd[-(s32)pDi->dni.pSubBlkHdr->nNumEntries], nCopySize2 * sizeof(u16));
            pDst2+=nCopySize2;
            if (pAttrOfs == 0) {
                pAttrOfs = pDst2;
            }
            *pDst2++ = nNameSubBlkOfs;
            memcpy(pDst2, &pExistingNodeOfsTblEnd[nCopySize2-pDi->dni.pSubBlkHdr->nNumEntries], pDi->dni.nEntryIdx * sizeof(u16));
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, *pDi->dni.pOfs, pDi->dni.nLog2Size);
            *pDi->dni.pOfs = nUpdatedNodeOfs; // Update the parent pointer
        } else {
            // The new option can be added without changing the size of the node
            utf8 *pCharInsertPoint = pDi->dni.pSubBlkHdr->aFirstChar + pDi->dni.pSubBlkHdr->nStrLen + pDi->dni.nEntryIdx;
            s32 nCopySize = (s32)(pDi->dni.pSubBlkHdr->nNumEntries - pDi->dni.nEntryIdx);
            memmove(pCharInsertPoint+1, pCharInsertPoint, (u32)nCopySize);
            pCharInsertPoint[0] = pDi->dni.pStrPtr[0];
            u16 *pPtrArray = &pExistingNodeOfsTblEnd[-(s32)pDi->dni.pSubBlkHdr->nNumEntries];
            memmove(pPtrArray-1, pPtrArray, sizeof(u16) * nCopySize);
            if (pAttrOfs == 0) {
                pAttrOfs = &pPtrArray[nCopySize-1];
            }
            pPtrArray[nCopySize-1] = nNameSubBlkOfs;
            ++pDi->dni.pSubBlkHdr->nNumEntries;
        }
    } else {
        // Replace the start of the existing node with a new 2-choice node
        // and update the existing node to contain only the end of the string,
        // resizing the node if its size drops to a smaller power-of-2 allocation size.
        // (note: the resize is compulsory due to the algorithms used elsewhere).
        //
        // First, try to allocate a sub block for the new 2-choice node.
        nChoiceNodeSubBlkLog2Size = (u16)DirLeafGetSubBlkLog2Size_2(nNumMatching, 2);
        nChoiceNodeSubBlkOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nChoiceNodeSubBlkLog2Size);
        if (nChoiceNodeSubBlkOfs == 0) {
            // Allocation failed, so undo any changes so far, split the block, and try again
            if (nSuffixLen) {
                // Undo the allocation of the name sub block
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nNameSubBlkOfs, nNameSubBlkLog2Size);
            }
            goto NameNotAllocated;
        }
        DirSubBlkHdr *pChoiceNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nChoiceNodeSubBlkOfs);
        pNodeOfsTblEnd = (u16*)((u8*)pChoiceNodeHdr + (1<<nChoiceNodeSubBlkLog2Size));
        pChoiceNodeHdr->nNumEntries = 2;
        pChoiceNodeHdr->nStrLen = (u8)nNumMatching;
        memcpy(pChoiceNodeHdr->aFirstChar, pDi->dni.pSubBlkHdr->aFirstChar, nNumMatching);
        utf8 cNodeChar = pDi->dni.pSubBlkHdr->aFirstChar[nNumMatching];
        utf8 cNameChar = pDi->dni.pStrPtr[0];

        // Now, update the existing node
        u16 nUpdatedNodeOfs;
        u16 nUpdatedNodeLog2Size = (u16)DirLeafGetSubBlkLog2Size_2(nNodeNumLeft-1, pDi->dni.pSubBlkHdr->nNumEntries);
        DirSubBlkHdr *pUpdatedNodeHdr;
        if (nUpdatedNodeLog2Size != pDi->dni.nLog2Size) {
            // The existing node should be reduced in size.
            // First try to allocate a new smaller sub block.
            // If this is not possible, then resize in place.
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeLog2Size);
            u16 *pUpdatedNodeOfsTblEnd;
            u32 nOldNodeOfs = (u16)((u8*)pDi->dni.pSubBlkHdr - (u8*)pBlkHdr);
            if (nUpdatedNodeOfs == 0) {
                // allocation failed, so resize in place
                // shift/copy the end of the string and the following choices backwards over the start of the string
                pUpdatedNodeHdr = pDi->dni.pSubBlkHdr;
                nUpdatedNodeOfs = (u16)nOldNodeOfs;
                pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
                memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries);
                // shift over the table of pointers at the end of the sub block
                memmove(&pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], &pExistingNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], sizeof(u16) * pUpdatedNodeHdr->nNumEntries);
                SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pBlkHdr, nUpdatedNodeOfs, pDi->dni.nLog2Size, nUpdatedNodeLog2Size);
            } else {
                // copy the end of the node to the new node, and free the old node
                pUpdatedNodeHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nUpdatedNodeOfs);
                *pDi->dni.pOfs = nUpdatedNodeOfs; // Update the parent pointer
                pUpdatedNodeHdr->nNumEntries = pDi->dni.pSubBlkHdr->nNumEntries;
                memcpy(pUpdatedNodeHdr->aFirstChar, &pDi->dni.pSubBlkHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
                pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
                memcpy(&pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], &pExistingNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], sizeof(u16) * pUpdatedNodeHdr->nNumEntries);
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBlkHdr, nOldNodeOfs, pDi->dni.nLog2Size);
            }
        } else {
            // the sub block remains the same size
            pUpdatedNodeHdr = pDi->dni.pSubBlkHdr;
            nUpdatedNodeOfs = (u16)((u8*)pUpdatedNodeHdr - (u8*)pBlkHdr);
            dbg(SbaCheckOfs(nUpdatedNodeOfs));
            // Remove characters from the start of the string
            memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
        }
        pUpdatedNodeHdr->nStrLen = (u8)(nNodeNumLeft-1);
        *pDi->dni.pOfs = nChoiceNodeSubBlkOfs;

        if ((unsigned)cNodeChar < (unsigned)cNameChar) {
            pChoiceNodeHdr->aFirstChar[nNumMatching] = cNodeChar;
            pChoiceNodeHdr->aFirstChar[nNumMatching+1] = cNameChar;
            pNodeOfsTblEnd[-1] = nUpdatedNodeOfs;
            if (pAttrOfs == 0) {
                pAttrOfs = &pNodeOfsTblEnd[-2];
            }
            pNodeOfsTblEnd[-2] = nNameSubBlkOfs;
        } else {
            pChoiceNodeHdr->aFirstChar[nNumMatching] = cNameChar;
            pChoiceNodeHdr->aFirstChar[nNumMatching+1] = cNodeChar;
            if (pAttrOfs == 0) {
                pAttrOfs = &pNodeOfsTblEnd[-1];
            }
            pNodeOfsTblEnd[-1] = nNameSubBlkOfs;
            pNodeOfsTblEnd[-2] = nUpdatedNodeOfs;
        }
    }
    ++pBlkHdr->nNumRecs;
    pDi->dni.pOfs = pAttrOfs;
    pDi->dni.nLog2Size = nAttrLog2Size;

Exit:;

    if(ppNewAttrHdr){
        *ppNewAttrHdr = pAttrHdr;
    }

    //DirLeafCheckBlk(pDi);
    return WFSKRN_RESULT_OK;
}
